#ifndef LTLC_H
#define LTLC_H

#include "time.h"

#include <sicstus/sicstus.h>
#include "ltlmem.h"
#include "ltlarray.h"
#include "bitoperations.h"

/* debugging */
#ifdef LTLDEBUG
#define IFLTLDEBUG(Instr) Instr
#else
#define IFLTLDEBUG(Instr)
#endif

#ifdef LTLTIMING
#define IFLTLTIMING(Instr) Instr
#else
#define IFLTLTIMING(Instr)
#endif

int ltl_raise(const char *format, ...);
int ltl_assert(int pred, const char *format, ...);

#define node_uninitialised (-3)
#define node_not_opened    (-2)
#define node_unexplorable  (-1)

/*
  An ltl_nodetype indicates if a node id (state id) is
  an integer (that should normally be the case, int_node)
  or any other prolog term (term_node).
 */
typedef enum {int_node, term_node} ltl_nodetype;

/*
  A node id is either an integer (in case of an int_node, see
  tlt_nodetype above) or an SP_term_ref.
*/
typedef union {
  unsigned int id;
  SP_term_ref ref;
} ltl_nodeid;

/*
  An ltlnode represents a state in the system.
  Informations about the state (atomic prepositions,
  transitions to other states) are received from ProB's
  prolog part exactly once.
  (see ?? for details)
*/
typedef struct {
  /*
    Number of atoms of this node.
    If number_of_atoms<0, the information of the
    node has not yet been retrieved from the prolog site.
  */
  int number_of_atoms;
  /*
    The id of the first atom belonging to this node.
    Only defined if number_of_atoms>0.
    The ids of all atoms of the node are the intervall
    [first_atom_id,first_atom_id+number_of_atoms-1].
  */
  unsigned int first_atom_id;
  /*
    The number of transitions starting in this node.
    If number_of_transitions<0, then the information about
    transitions has not yet been retrieved (see number_of_atoms).
    If number_of_transitions==0, then the node is a deadlock.
  */
  int number_of_transitions;
  /*
    The id of the first transition, only defined if 
    number_of_transitions>=0. The ids of all transitions of the
    node are the intervall
    [first_transid,first_transid+number_of_transitions-1].
  */
  unsigned int first_transid;
  /*
    The type of the node ID.
    Nodes with integer IDs and arbitrary prolog terms as IDs are
    destinguished.
  */
  ltl_nodetype type;
  /*
    The ID of the node. Either an integer of a reference to an
    prolog term.
  */
  ltl_nodeid id;
} ltlnode;

/*
  A nodestore_item is used for the linked list of nodes with
  a general prolog term ID.
 */
typedef struct nodestore_item {
  /* The reference to the prolog term */
  SP_term_ref id;
  /* The corresponding node */
  ltlnode node;
  /* The next element in the list, NULL if this is the last */
  struct nodestore_item *next;
}  nodestore_item;

/*
  The node store is used to store C representations of system states.
  It is optimised to store a large number of states that have integer
  IDs in a certain range, and only a small number of states that
  have arbitrary prolog terms as IDs (stored as a linked list).

  Internally three stores are used: Two arrays (for nodes with
  integer IDs). Roughly speaking, one for those IDs >= the first
  one stored, and one < the first one.
  Additionally a linked list is used for states whose ID is another
  prolog term.
*/
typedef struct {
  /*
    The ID of the first inserted element is used to as an offset
    (rounded down).
  */
  long offset;
  
  /* The two arrays for elements >=offset and <offset */
  ltlarray pos,neg;
  
  /* 
     The start of the linked list for nodes with a prolog therm ID.
     NULL, if empty.
  */
  nodestore_item *list;
  
  /* The number of elements in the nodestore */
  unsigned int size;
} nodestore;

typedef enum {ltlap,
              ltltp,
              ltlnext,
              ltlnextnot,
              ltlyesterday,
              ltlyesterdaynot,
              ltlor,
              ltluntil,
              ltlsince} ltl_operator;

/*
  ltl_pform_ref contains an argument of an operator (see ltl_part_formula)
  which can also be negated. This data type is normally used in
  a part of a closure (ltl_part_formula) to refer to another part.
*/
typedef struct {
  /* reference to a part of the closure */
  int ref;
  /* the signum indicates if the argument should be negated or not */
  unsigned int signum;
} ltl_pform_ref;

/*
  An ltl_part_formula is a part of a closure of a formula.
  The field op indicates which kind of operator is used,
  the field args contains - depending on the op - the
  arguments of the operator. All operators are references
  to other parts of the closure (by integers) or references
  with an optional negation (by ltl_pform_ref).
  An array of this is used in ltl_formula.
*/
typedef struct {
  ltl_operator op;
  union {
    int case_next;
    int case_nextnot;
    int case_yesterday;
    int case_yesterdaynot;
    struct {ltl_pform_ref a,b;} case_or;
    struct {ltl_pform_ref inv,goal; int next;} case_until;
    struct {ltl_pform_ref inv,goal; int yest;} case_since;
  } args;
} ltl_part_formula;

/*
  An atom is a node+possible evaluation of the closure.
*/
typedef struct {
  /* A pointer to the node this atom belongs to. Must not be NULL */
  ltlnode *node;
  /*
    The number of transitions starting from this atom.
    If number_of_transitions<0, the atom is not yet explored.
  */
  int number_of_transitions;
  /*
    The ID of the first transition starting from this atom.
    first_transistion_id is only defined if number_of_transitions > 0.
    The corresponding transitions are stored in the field
    atomtransstore in ltl_formula.
  */
  unsigned int first_transition_id;
  /* The tarjan_index is used in Tarjan's algorithm */
  int tarjan_number;
  /* tarjan_on_stack is a flag used in Tarjan's algorithm */
  char tarjan_on_stack;
  /* the lowlink is only stored for debugging */
  #ifdef LTLSTORELOWLINK
  int tarjan_lowlink;
  #endif
} ltlatom;

typedef struct {
  ltlnode *dest;
  SP_term_ref operation;
} ltlnodetrans;

/*
  An equal_evaluation_positions object describes which
  two positions a and b of a closure evaluation must be equal
  (or distinct) if their is an transition between two atoms
 */
typedef struct {
  /* positions in the first (a) and the second (b) atom */
  int a,b;
  /* should the evaluations be equal (TRUE) or distinct (FALSE) */
  int signum;
} equal_evaluation_positions;

/*
  ltl_formula represents the static and run-time properties of a
  formula. This is the central data structure of the
  LTL model checking procedure.
*/
typedef struct {
  /* The number of elements in the LTL closure */
  int closure_size;
  /*
    A pointer to the array of the parts of the formula,
    the closure. The size of the closure is given
    by closure_size.

    The closure is divided into four parts:
    1) The atomic propositions (number_of_aps contains their number)
    2) The formulas of the form X(...) (number_of_nexts contains their number)
    3) The formulas of the form X(not(...)) (same number as formulas of the form X(...))
    4) Other formulas like or, etc.
  */
  ltl_part_formula *closure;

  /* The number of atomic propositions in the closure */
  int number_of_aps;
  /* The number of transition propositions */
  int number_of_tps;
  /* The number of formulas of the form X(...) in the closure */
  int number_of_nexts;
  /* The number of formulas of the form Y(...) in the closure */
  int number_of_yesterdays;
  /* The number of formulas of the form or(...), until(...) or since(...) in the closure */
  int number_of_generics;

  /* Where in the closure start the following information: */
  int index_tps;
  int index_nexts;
  int index_nextnots;
  int index_yesterdays;
  int index_yesterdaynots;
  int index_generics;

  ltl_pform_ref mainref;

  /*
    If a transition from one atom to another exists depends on the
    comparision of the evaluations of several formulas in the
    closure. The comparisions are in an array equal_evaluations
    of the size number_of_equal_evaluations.
  */
  int number_of_equal_evaluations;
  equal_evaluation_positions *equal_evaluations;

  /* the nodes */
  nodestore nstore;

  /* the array of all transitions between nodes */
  ltlarray nodetrans;
  unsigned int nodetrans_number;

  /* the array of all atoms */
  ltlarray astore;
  unsigned int number_of_atoms;

  /* the array of all transitions between atoms */
  ltlarray atomtransstore;
  unsigned int number_of_atomtrans;

  /* flag to indicate that the whole search space was not explored */
  int unexplored_nodes;

  /* callback predicate to prolog */
  SP_pred_ref callback_predicate;
  SP_term_ref ap_list,tp_list;

  /* callback predicate to check if an SCC is valid 
     (e.g. by fairness conditions) */
  int check_found_scc_activated;
  SP_pred_ref check_found_scc_predicate;
  SP_term_ref check_found_scc_argument;

  /* update calls to prolog */
  int update_predicate_activated;
  SP_pred_ref update_predicate;
  clock_t last_update_call;  
  int update_countdown;
  int canceled;

  /* untils are often handled special */
  int number_of_untils;
  /* an array (size number_of_untils) of indices, referencing those
   * items of the closure which are untils */
  unsigned int *untils;
  

  /* temporary storage */
  bitfield *tmp_evaluation,*tmp_evaluation2;
  ltlarray tmp_transeval_store;
  bitfield *zero_transeval;

  IFLTLTIMING( int callback_calls;
               clock_t callback_time; );
} ltl_formula;

/*
  The type of a found SCC.
*/
typedef struct {
  /* The size of the SCC */
  int number_of_scc_atoms;
  /* An array containing the atoms of the SCC */
  ltlatom **scc;
  /* The atom that is the entry point from the path into the SCC
     The atom itself if in the SCC */
  ltlatom *scc_entry;
  /* The length of the path leading to the SCC */
  int number_of_path_elements;
  /* An array containing the path to the SCC */
  ltlatom **path;
} found_scc_type;

/*
  Initialialise the node store.
  This method must be called before the first use of the node store.
  
  st must not be NULL.
*/
void init_nodestore(nodestore *st);

/*
  Delete the node store.
  This method must be called after the last use of the node store.
  All used memory will be freed.
  
  st must not be NULL.
*/
void delete_nodestore(nodestore *st);

/*
  Get a node from the node store.
  If the is not present in the store before the function call,
  it will be created.
  The return value indicates if a new node has been created.
  
  st is the node store (must not be NULL).
  id is the prolog term that identifies the state (node).
     Must be a valid prolog term (and not NULL).
  node points to a pointer of the node. The pointer is set by the function.
  node must not be NULL.
 */
int get_node_by_term(nodestore *st, SP_term_ref id, ltlnode **node);

/*
  The same as get_node_by_term, except that instead of a prolog term
  a type and id will be used.
 */
int get_node_by_id(nodestore *st, ltl_nodetype type, ltl_nodeid id, ltlnode **node);

size_t ltl_nodestore_memory_usage(nodestore *st);
void print_prolog_node(FILE *file, ltlnode *node);
void print_prolog_nodestore(FILE *file, nodestore *nstore);

int explore_atom(ltl_formula *formula, ltlatom *atom);
int explore_node(ltl_formula *formula, ltlnode *node, int is_start_node);
void get_atom(ltlarray *array, unsigned int id, ltlatom **atom, bitfield **evaluation);
bitfield *get_atom_evaluation(ltlatom *atom);
int eval(bitfield *evaluation, ltl_pform_ref *ref);
ltlatom *get_atom_trans(ltl_formula *formula, unsigned int id);
void set_atom_trans(ltl_formula *f, unsigned int id, ltlatom *dest);
SP_term_ref find_transition_operation(ltl_formula *f, ltlatom *a, ltlatom *b);
SP_term_ref find_transition_operations(ltl_formula *f, ltlatom *a, ltlatom *b);

void init_formula(ltl_formula *f);
void delete_formula(ltl_formula *f);
size_t ltl_formula_memory_usage(ltl_formula *f);

int start_tarjan(ltl_formula *f, ltlnode **starts, int size, 
                 size_t *used_memory, found_scc_type **found_scc);
unsigned int find_atom_id(ltl_formula *f, ltlatom *atom);

void print_atom_pointer(ltl_formula *f);
void ltl_printnode(ltlnode *node);
void ltl_snapshot(ltl_formula *f);
void ltl_print_graph(ltl_formula *f, found_scc_type *found_scc, int number);
void ltldeb_inc(void);
void ltldeb_dec(void);
int ltldeb_get(void);
void ltldeb_set(int depth);
void ltlmsg(const char *msg, ...);
void ltl_print_prolog_graph(ltl_formula *f, found_scc_type *found_scc);
void debug_print_trace(ltl_formula *f, ltlatom **result,int resultlength,int is_deadlock);

void get_path_through_scc(ltl_formula *f, found_scc_type *found, ltlatom ***path, int *length, int *isdeadlock, int *loop_length);

int shorten_path(ltl_formula *f, ltlatom **path, int length);

int is_safety_property(ltl_formula *f, ltl_pform_ref *ref);

void ltl_reraise_exception(void);

extern const clock_t ltl_update_interval;
extern const int update_countdown_init;

#endif
