% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(ltlc, [ltlc_init/0,c_ltl_modelcheck/4,c_ltl_modelcheck/7,c_ltl_aptp/2]).

:- use_module(library(lists)).
:- use_module('../../src/module_information.pl').

:- module_info(group,ltl).
:- module_info(description,'This is the interface to the LTL model checker\'s C code.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').


:- meta_predicate c_ltl_modelcheck(-,-,-,5).

/* the C interface */
foreign_resource(ltlc, [ltlc_modelcheck]).
foreign(ltlc_modelcheck,ltlc_modelcheck(+term,+term,+term,+term,+term,+term,+term,-term)).


:- dynamic loaded/0.

ltlc_init :- loadfr.

loadfr :- (loaded -> true
 ; (assert(loaded), 
    load_foreign_resource(library(ltlc)))
   ).


/* The central predicate */
c_ltl_modelcheck(Formula, Startnodes, Result, CModule:Callback) :-
    c_ltl_modelcheck(Formula, Startnodes, Result, CModule:Callback, none, none, []).

c_ltl_modelcheck(Formula, Startnodes, Result, Callback, Update, CheckSCC, CheckArg) :-
    c_ltl_closure(Formula,Closure,Main,_),
	ltlc_modelcheck(Closure, Startnodes, Callback, Update, CheckSCC, CheckArg, Main, Result).

% Return the list of atomic propositions and transition propositions
% of a formula. c_ltl_modelcheck returns counter-examples, where the
% evaluation part of the atoms begins with the evaluation of the
% formulas corresponding to ApTp.
c_ltl_aptp(Formula,ApTp) :-
    c_ltl_closure(Formula,_,_,ApTp).

/* Take a formula (with all supported operators), normalise it
 * (rewrite to basic operators) and convert it into a list of
 * atomic propositions (APs), transition propositions (TPs),
 * and basic operators with references to other parts of the
 * closure.
 * Additionally a reference to the original formula is given
 * Each reference has the form pos(N) or neg(N) where N is
 * the position of the subformula in the closure (starting with 0).
 * The position -1 refers to "true".
 * The closure always start with the APs and TPs.
 * Example:
 *  Formula: not(globally(ap(AP)))
 *  Closure: [ap(AP),next(3),nextnot(3),until(pos(-1),neg(0),1)
 *  Main:    pos(3)
*/
c_ltl_closure(Formula,Closure,Main,Atomics) :-
	( ltl_norm(Formula, Normalised) ->
	    true
	; otherwise ->
	    write('internal error: ltl_norm failed for formula '),
	    write(Formula),nl,fail),
	positive_closure(Normalised, PClosure),
	stripClosure(PClosure, Normalised, Main, Closure, Atomics).

stripClosure(Unsorted, MainIn, MainOut, Result, Atomics) :-
	filter_parts(Ap,Tp,Next,Nextnot,Yest,Yestnot,Unsorted,Generic),
	append([Ap,Tp,Next,Nextnot,Yest,Yestnot,Generic], Sorted),
    append(Ap,Tp,Atomics),
	findref(MainIn,Sorted,MainOut),
	numeric_refs(Sorted,Sorted,Result).

filter_parts(Ap,Tp,Next,Nextnot,Yest,Yestnot) -->
		     filter_terms(Ap, ap(_)),
		     filter_terms(Tp, action(_)),
		     filter_terms(Nextnot, next(not(_))),
		     filter_terms(Next, next(_)),
		     filter_terms(Yestnot, yesterday(not(_))),
		     filter_terms(Yest, yesterday(_)).

filter_terms(Filtered,Pattern,In,Out) :- filter_terms2(In,Pattern,Filtered,Out).
		    
filter_terms2([],_,[],[]).
filter_terms2([E|Rest],Term,A,B) :-
    ( copy_term(Term, E) ->
        A=[E|ARest],B=BRest
    ;
        A=ARest,B=[E|BRest]),
    filter_terms2(Rest,Term,ARest,BRest).

numeric_refs([],_,[]).
numeric_refs([F|Rest],L,[R|RRest]) :-
	xconv(F,L,R),numeric_refs(Rest,L,RRest).

xconv(ap(AP),_,ap(AP)).
xconv(action(TP),_,tp(TP)).
xconv(next(not(X)),L,nextnot(R)) :- !,finduref(X,L,R),!.
xconv(next(X),L,next(R)) :- finduref(X,L,R),!.
xconv(yesterday(not(X)),L,yesterdaynot(R)) :- !,finduref(X,L,R),!.
xconv(yesterday(X),L,yesterday(R)) :- finduref(X,L,R),!.
xconv(or(A,B),L,or(RA,RB)) :- findref(A,L,RA),findref(B,L,RB).
xconv(until(A,B),L,until(RA,RB,RN)) :-
	findref(A,L,RA),
	findref(B,L,RB),
	findref(next(until(A,B)),L,pos(RN)).
xconv(since(A,B),L,since(RA,RB,RY)) :-
	findref(A,L,RA),
	findref(B,L,RB),
	findref(yesterday(since(A,B)),L,pos(RY)).
findref(not(F),L,neg(N)) :- !,finduref(F,L,N).
findref(F,L,pos(N)) :- finduref(F,L,N).

finduref(true,_,-1) :- !.
finduref(F,L,N) :- nth0(N,L,F).


%*******************************************************************************
% Closure of a formula

% the positive closure has only the positive (without leading nots)
% formulas of the closure.
% Also its elements are sorted
positive_closure(F,NDPositive) :-
	closure(F,Closure),
	sort_closure(Closure,Sorted),
	remove_negatives(Sorted,Positive),
	remove_duplicates(Positive,NDPositive).

remove_duplicates(In,Out) :- remove_duplicates2(In,[],Out).
remove_duplicates2([],_,[]).
remove_duplicates2([E|Rest],Before,Out) :-
	(member(E,Before) ->
	    !,remove_duplicates2(Rest,Before,Out);
	    Out=[E|DRest],remove_duplicates2(Rest,[E|Before],DRest)).

closure(F,C) :- closure([F],[],C).

closure([],Acc,Acc).
closure([true|Rest],Acc,Closure) :-
	!,closure(Rest,Acc,Closure).
closure([F|Rest],Acc,Closure) :-
	findall(G,dclosure(F,G),Gen),
	filternew(Gen,Acc,Rest,New),
	append(Rest,New,Next),
	closure(Next,[F|Acc],Closure).

dclosure(F,Neg) :- negate(F,Neg).
dclosure(or(F,_),F).
dclosure(or(_,F),F).
dclosure(next(F),F).
dclosure(yesterday(F),F).
dclosure(not(next(F)),next(Neg)) :- negate(F,Neg).
dclosure(not(yesterday(F)),yesterday(Neg)) :- negate(F,Neg).
dclosure(until(F,_),F).
dclosure(until(_,F),F).
dclosure(until(A,B),next(until(A,B))).
dclosure(since(F,_),F).
dclosure(since(_,F),F).
dclosure(since(A,B),yesterday(since(A,B))).

negate(not(F),F) :- !.
negate(F,not(F)).

filternew([],_,_,[]).
filternew([F|Rest],Ex1,Ex2,Result) :-
	(\+ member(F,Ex1), \+ member(F,Ex2)
	-> Result = [F|FRest]
	;  Result = FRest),
	filternew(Rest,Ex1,Ex2,FRest).

remove_negatives([],[]).
remove_negatives([F|Rest],Positive) :-
	(F = not(_)
	  -> !,Positive = PRest
	  ;  Positive = [F|PRest]),
	remove_negatives(Rest,PRest).

sort_closure(In,Out) :- sort_closure2(In,[],Out).
sort_closure2([],_,[]) :- !.
sort_closure2(Unsorted,Sorted,[F|RestSorted]) :-
	select_formula(Unsorted,Sorted,F,URest),
	sort_closure2(URest,[F|Sorted],RestSorted).

select_formula([F|FRest],Sorted,F,FRest) :-
	deps(F,Deps),
	subset(Deps,Sorted),!.
select_formula([F|FRest],Sorted,S,[F|Rest]) :-
	select_formula(FRest,Sorted,S,Rest).

deps(true,[]).
deps(not(F),[F]).
deps(ap(_),[]).
deps(or(A,B),[A,B]).
deps(next(F),D) :- (F=not(N) -> D=[next(N)]; D=[]).
deps(yesterday(F),D) :- (F=not(N) -> D=[yesterday(N)]; D=[]).
deps(until(A,B),[A,B,next(until(A,B))]).
deps(since(A,B),[A,B,yesterday(since(A,B))]).
deps(action(_),[]).

% subset(A,B): A is a subset of B
subset([],_).
subset([F|Rest],Superset) :-
	(F=true;member(F,Superset)),!,subset(Rest,Superset).

%*******************************************************************************
% normalise a formula (LTL)

% part I: translating syntatic sugar
ltl_norm(false,not(true)).
ltl_norm(and(A,B),N) :- ltl_norm(not(or(not(A),not(B))),N).
ltl_norm(implies(A,B),N) :- ltl_norm(or(not(A),B),N).
ltl_norm(finally(F),N) :- ltl_norm(until(true,F),N).
ltl_norm(globally(F),N) :- ltl_norm(not(finally(not(F))),N).
ltl_norm(weakuntil(A,B),N) :- ltl_norm(or(globally(A),until(A,B)),N).
ltl_norm(release(A,B),N) :- ltl_norm(not(until(not(A),not(B))),N).
ltl_norm(once(F),N) :- ltl_norm(since(true,F),N).
ltl_norm(historically(F),N) :- ltl_norm(not(once(not(F))),N).
ltl_norm(trigger(A,B),N) :- ltl_norm(not(since(not(A),not(B))),N).

% part II: mostly normalising arguments recursivly, removing not(not(...))
ltl_norm(true,true).
ltl_norm(next(F),next(NF)) :- ltl_norm(F,NF).
ltl_norm(yesterday(F),yesterday(NF)) :- ltl_norm(F,NF).
ltl_norm(or(A,B),or(NA,NB)) :-
	ltl_norm(A,NA),ltl_norm(B,NB).
ltl_norm(not(F),N) :- ltl_norm(F,N1), (N1=not(P) -> N=P; N=not(N1)).
ltl_norm(ap(X),ap(X)).
ltl_norm(until(A,B),until(NA,NB)) :-
	ltl_norm(A,NA),ltl_norm(B,NB).
ltl_norm(since(A,B),since(NA,NB)) :-
	ltl_norm(A,NA),ltl_norm(B,NB).
ltl_norm(action(Op),action(Op)).

