#include "ltlc.h"

typedef struct {
  ltl_formula *f;
  /* The fields store already computed values.
   * There is an entry for each partial formula (i.e. each entry in f->closure),
   * each position in the counter-example and for each case true or false.
   * The function fieldposition returns an index.
   * Initial values are -1.
   */
  int *fields;
  int length;
  ltlatom **path;
} sdata;

static int shorten_next(sdata *data, int offset, unsigned int signum, int ref);
static int shorten_yest(sdata *data, int offset, unsigned int signum, int ref);
static int validate(sdata *data, int offset, ltl_pform_ref *ref);
static int shorten(sdata *data, int offset, ltl_pform_ref *main);
static int shorten_ref(sdata *data, int offset, unsigned int signum, int ref);
static int shorten_negated(sdata *data, int offset, ltl_pform_ref *ref);
static int minimum(int a, int b);
static int maximum(int a, int b);
static int fieldposition(sdata *data, int offset, ltl_pform_ref *ref);
void print_sdata(sdata *data);
static int shorten_positive_formula(sdata *data, ltl_part_formula *mr, int offset, int longest);
static int shorten_negative_formula(sdata *data, ltl_part_formula *mr, int offset, int longest);

/*
 * shorten_path(ltl_formula *f, ltlatom **path, int length):
 *    f: the LTL formula
 *    path: the (unshortened) path
 *    length: the length of the path
 *  returns the number of needed elements that are sufficient for a
 *  counter-example. It is -1 if the complete loop should be used.
 *  The difference between a return value (a) length and (b) -1 is that 
 *  in case (a) all elements are used in the counter-example but the end is not
 *  needed to be connected for a loop. In case (b) the full loop is used, but
 *  the number of elements is still the same.
 */
int shorten_path(ltl_formula *f, ltlatom **path, int length) {
  ltl_pform_ref *mainref = &(f->mainref);
  int fsize,i,s;
  sdata *data = (sdata*) malloc(sizeof(sdata));
  if (data) {
    fsize = f->closure_size * length * 2;
    data->length = length;
    data->path = path;
    data->fields = (int*) malloc(sizeof(int) * fsize);
    if (data->fields) {
      data->f = f;
      for (i=0; i<fsize; i++) {
        data->fields[i] = -1;
      }
      /* shorten returns the index of the last relevant position in the path 
         or the length of the counter-example if it's a loop
       */
      s = shorten(data,0,mainref);
      IFLTLDEBUG( print_sdata(data); );
      free(data->fields);
    } else {
      fprintf(stderr, "malloc() returned NULL for fields in shorten_path\n");
      s = length;
    }
    free(data);
  } else {
    fprintf(stderr, "malloc() returned NULL in shorten_path\n");
    s = length;
  }
  return s>=length ? -1 : s+1;
}

int shorten(sdata *data, int offset, ltl_pform_ref *main) {
  int result;
  int longest = data->length;
  if (offset >= longest || offset<0) {
    /* we ran out of the counter-example somehow */
    result = longest;
  } else {
    if (main->ref < 0) {
      /* the basic formula true or false */ 
      if (validate(data, offset, main)) {
        result = minimum(offset, longest);
      } else {
        result = longest;
      }
    } else {
      int fpos = fieldposition(data,offset,main);
      if (data->fields[fpos] >= 0) {
        result = data->fields[fpos];
      } else {
        if (!validate(data, offset, main)) {
          result = longest;
        } else {
          ltl_part_formula *mr = &(data->f->closure[main->ref]);
          if (main->signum) {
            result = shorten_positive_formula(data,mr,offset,longest);
          } else {
            result = shorten_negative_formula(data,mr,offset,longest);
          }
        }
        data->fields[fpos] = result;
      }
    }
  }
  return result;
}

static int shorten_positive_formula(sdata *data, ltl_part_formula *mr, int offset, int longest) {
  int a,b,goal,inv,next,future,yest,past;
  switch(mr->op) {
  case ltlap:
    return offset;
  case ltltp:
    return minimum(offset+1, longest);
  case ltlnext:
    return shorten_next(data,offset,1,mr->args.case_next);
  case ltlnextnot:
    return shorten_next(data,offset,0,mr->args.case_nextnot);
  case ltlyesterday:
    return shorten_yest(data,offset,1,mr->args.case_yesterdaynot);
  case ltlyesterdaynot:
    return shorten_yest(data,offset,0,mr->args.case_yesterdaynot);
  case ltlor:
    a = shorten(data,offset,&(mr->args.case_or.a));
    b = shorten(data,offset,&(mr->args.case_or.b));
    return minimum(a,b);
  case ltluntil:
    goal = shorten(data,offset,&(mr->args.case_until.goal));
    inv = shorten(data,offset,&(mr->args.case_until.inv));
    next = shorten_ref(data,offset,1,mr->args.case_until.next);
    future = maximum(inv,next);
    return minimum(goal, future);
  case ltlsince:
    goal = shorten(data,offset,&(mr->args.case_since.goal));
    inv = shorten(data,offset,&(mr->args.case_since.inv));
    yest = shorten_ref(data,offset,1,mr->args.case_since.yest);
    past = maximum(inv,yest);
    return minimum(goal, past);
  default:
    fprintf(stderr, "unexpected case in shorten: pos. switch\n");
    return longest;
  }
}

static int shorten_negative_formula(sdata *data, ltl_part_formula *mr, int offset, int longest) {
  int a,b,goal,inv,next,future,yest,past;
  switch(mr->op) {
  case ltlap:
    return offset;
  case ltltp:
    return minimum(offset+1, longest);
  case ltlnext:
    return shorten_next(data,offset,0,mr->args.case_next);
  case ltlnextnot:
    return shorten_next(data,offset,1,mr->args.case_nextnot);
  case ltlyesterday:
    return shorten_yest(data,offset,0,mr->args.case_yesterday);
  case ltlyesterdaynot:
    return shorten_yest(data,offset,1,mr->args.case_yesterdaynot);
  case ltlor:
    a = shorten_negated(data,offset,&(mr->args.case_or.a));
    b = shorten_negated(data,offset,&(mr->args.case_or.b));
    return maximum(a,b);
  case ltluntil:
    goal = shorten_negated(data,offset,&(mr->args.case_until.goal));
    inv = shorten_negated(data,offset,&(mr->args.case_until.inv));
    next = shorten_ref(data,offset,0,mr->args.case_until.next);
    future = minimum(inv,next);
    return maximum(goal,future);
  case ltlsince:
    goal = shorten_negated(data,offset,&(mr->args.case_since.goal));
    inv = shorten_negated(data,offset,&(mr->args.case_since.inv));
    yest = shorten_ref(data,offset,0,mr->args.case_since.yest);
    past = minimum(inv,yest);
    return maximum(goal, past);
  default:
    fprintf(stderr, "unexpected case in shorten: neg. switch\n");
    return longest;
  }
}
int shorten_next(sdata *data, int offset, unsigned int signum, int ref) {
  return maximum(offset, shorten_ref(data,offset+1,signum,ref));
}

int shorten_yest(sdata *data, int offset, unsigned int signum, int ref) {
  int result;
  if (offset==0 && !signum) {
    result = 0;
  } else {
    result = maximum(offset, shorten_ref(data,offset-1,signum,ref));
  }
  return result;
}

int validate(sdata *data, int offset, ltl_pform_ref *ref) {
  bitfield *evaluation = get_atom_evaluation(data->path[offset]);
  return eval(evaluation, ref);
}

int shorten_ref(sdata *data, int offset, unsigned int signum, int ref) {
  ltl_pform_ref pfref;
  pfref.ref = ref;
  pfref.signum = signum;
  return shorten(data,offset,&pfref);
}

int shorten_negated(sdata *data, int offset, ltl_pform_ref *ref) {
  return shorten_ref(data,offset,ref->signum ? 0 : 1, ref->ref);
}

int minimum(int a, int b) {
  return a<b ? a : b;
}

int maximum(int a, int b) {
  return a>b ? a : b;
}

int fieldposition(sdata *data, int offset, ltl_pform_ref *ref) {
  int line = ref->ref + (ref->signum ? data->f->closure_size : 0);
  return line * data->length + offset;
}

static void print_pform_ref(ltl_pform_ref *ref) {
  if (ref->ref < 0) {
    printf(ref->signum?" T ":" F ");
  } else {
    printf("%s%02i", ref->signum?"+":"-", ref->ref);
  }
}

static void print_binop(const char *functor, ltl_pform_ref *a, ltl_pform_ref *b) {
  printf("%s(",functor);
  print_pform_ref(a);
  printf(",");
  print_pform_ref(b);
  printf(") ");
}


static void print_part_formula(ltl_formula *f, int pos, int signum) {
  ltl_part_formula *pf = &(f->closure[pos]);
  if (f->mainref.ref==pos && f->mainref.signum==signum) {
    printf(" *");
  } else {
    printf("  ");
  }
  switch(pf->op) {
  case ltlap:           printf("          ap%02i ",pos); break;
  case ltltp:           printf("          tp%02i ",pos-f->number_of_aps); break;
  case ltlnext:         printf("      next(%02i) ",pf->args.case_next); break;
  case ltlnextnot:      printf("  not next(%02i) ",pf->args.case_nextnot); break;
  case ltlyesterday:    printf("      yest(%02i) ",pf->args.case_yesterday); break;
  case ltlyesterdaynot: printf("  not yest(%02i) ",pf->args.case_yesterdaynot); break;
  case ltlor:           print_binop("   or",&(pf->args.case_or.a),&(pf->args.case_or.b)); break;
  case ltluntil:        print_binop("until",&(pf->args.case_until.inv),&(pf->args.case_until.goal));
    break;
  case ltlsince:        print_binop("since",&(pf->args.case_since.inv),&(pf->args.case_since.goal));
    break;
  default:              printf("????????????"); break;
  }
}

static void print_sdata_part(sdata *data, int signum) {
  ltl_pform_ref pref;
  int ref,offset,value;
  pref.signum = signum;
  for (ref=0; ref < data->f->closure_size; ref++) {
    pref.ref = ref;
    print_part_formula(data->f,ref,signum);
    for (offset=0; offset < data->length; offset++) {
      value = data->fields[fieldposition(data,offset,&pref)];
      if (value==-1) {
        printf("  -");
      } else {
        printf("%3i", value);
      }
    }
    printf("\n");
  }
}

void print_sdata(sdata *data) {
  printf("Positive:\n");
  print_sdata_part(data,1);
  printf("Negative:\n");
  print_sdata_part(data,0);
}
