:- module(pltables_export_latex, [
							latex_header/2,
							latex_footer/2,
							latex_table_header/2,
							latex_table_footer/2,
							latex_table_rows/2
					 	   ]).

:- use_module(pltables_export_tools).

:- use_module(probsrc(module_information)).

:- module_info(group,pltables).
:- module_info(description,'This module is used to export coverage tables to latex.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

% ----
% table headers
% ----
latex_header(S, TableName) :-
	pltables:pltable(TableName, Attributes),
	write(S,'\\documentclass[a4paper,10pt]{article}\n'),
	write(S,'\\usepackage{a4wide}\n\\usepackage{longtable}\n\\usepackage{xcolor}\n\\begin{document}\n'),
	(
	  member(headline(XHU), Attributes) ->
	      latex_escape(XHU,XH),
	      format(S, '\\textbf{~w} \\\\\\\\\n', [XH])
	      ; true
	).

% ----
% table footers
% ----
latex_footer(S, _TableName) :-
  write(S,'\\end{document}').
  
% ----
% table header / footer
% ----
latex_table_header(S, TableName) :-
    pltables:pltable(TableName, TableAttributes),
	pltables:pltable_header(TableName, Columns, Attributes),
	(
	  member(description(XDU, DescAttrs), TableAttributes) ->
	      latex_escape(XDU,XD),
	      (member(bold, DescAttrs) -> write(S, '\\textbf{') ; true),
	      format(S, '~w\n', [XD]),
	      (member(bold, DescAttrs) -> write(S, '}') ; true)
	      ; true
	),
	write(S, '\\begin{longtable}'),
	write(S, '{|'),
	write_latex_table_header1(S, Attributes),
	write(S, '}\n\\hline\n'),
	write_latex_table_header2(S, Columns, Attributes).

write_latex_table_header1(_S, []).
write_latex_table_header1(S, [Attributes|TAttributes]) :-
    (member(align(left), Attributes) -> write(S, 'l|') ; true),
    (member(align(right), Attributes) -> write(S, 'r|') ; true),
    (member(align(center), Attributes) -> write(S, 'c|') ; true),
    (\+member(align(_), Attributes) -> write(S, 'l|') ; true),
	write_latex_table_header1(S, TAttributes).
write_latex_table_header2(S, [Column], _Attributes) :-
	latex_escape(Column, EscapedColumn),
	format(S, '~w \\\\\n\\hline\n', [EscapedColumn]).
write_latex_table_header2(S, [Column | Columns], Attributes) :-
    latex_escape(Column, EscapedColumn),
	format(S, '~w & ', [EscapedColumn]),
	write_latex_table_header2(S, Columns, Attributes).
	
latex_table_footer(S, _TableName) :-
    write(S,'\\end{longtable}\n').
	

% ----
% table rows
% ----
latex_table_rows(S, TableName) :-
	findall(Entries, pltables:pltable_row(TableName, Entries, Attributes), ListOfEntries),
	findall(Attributes, pltables:pltable_row(TableName, Entries, Attributes), ListOfAttributes),
	pltables:pltable(TableName, TableAttributes),
	pltables:pltable_header(TableName, _, ColumnsAttributes),
	write_latex_table_rows(S, ListOfEntries, ListOfAttributes, TableAttributes, ColumnsAttributes).
	
write_latex_table_rows(_S, [], [], _TableAttributes, _ColumnsAttributes).
write_latex_table_rows(S, [Entries | MoreEntries], [RowAttributes | MoreAttributes], TableAttributes, ColumnsAttributes) :-
	write_latex_table_row(S, Entries, RowAttributes, TableAttributes, ColumnsAttributes),
	write(S, ' \\\\\n\\hline\n'),
	write_latex_table_rows(S, MoreEntries, MoreAttributes, TableAttributes, ColumnsAttributes).
	
write_latex_table_row(S, [Entry], RowAttributes, TableAttributes, [ColumnAttribute]) :-
	merge_attributes(TableAttributes, ColumnAttribute, TmpAttributes),
	merge_attributes(TmpAttributes, RowAttributes, MergedAttributes),
	(member(color(C), MergedAttributes) -> format(S, '\\textcolor[HTML]{~w}{', [C]) ; true),
	(member(bold, MergedAttributes) -> write(S, '\\textbf{') ; true),
	latex_escape(Entry, EscapedEntry),
	format(S, '~w', [EscapedEntry]),
	(member(bold, MergedAttributes) -> write(S, '}') ; true),
	(member(color(C), MergedAttributes) -> write(S, '}') ; true).
write_latex_table_row(S, [Entry | Entries], RowAttributes, TableAttributes, [ColumnAttribute | ColumnsAttributes]) :-
	merge_attributes(TableAttributes, ColumnAttribute, TmpAttributes),
	merge_attributes(TmpAttributes, RowAttributes, MergedAttributes),
	(member(color(C), MergedAttributes) -> format(S, '\\textcolor[HTML]{~w}{', [C]) ; true),
	(member(bold, MergedAttributes) -> write(S, '\\textbf{') ; true),
	latex_escape(Entry, EscapedEntry),
	format(S, '~w', [EscapedEntry]),
	(member(bold, MergedAttributes) -> write(S, '}') ; true),
	(member(color(C), MergedAttributes) -> write(S, '}') ; true),
	write(S, ' & '),
	write_latex_table_row(S, Entries, RowAttributes, TableAttributes, ColumnsAttributes).
	
% ----
% escaping characters
% ----
latex_escape(ToEscape, Escaped) :-
  atom(ToEscape), !,
  atom_codes(ToEscape, CodeList),
  latex_escape2(CodeList, EscapedCodeList),
  atom_codes(Escaped, EscapedCodeList).
latex_escape(ToEscape, ToEscape).

latex_escape2([],[]).
latex_escape2([95|T], [92,95|EscT]) :-
  latex_escape2(T, EscT).
latex_escape2([35|T], [92,35|EscT]) :-
  latex_escape2(T, EscT).
latex_escape2([36|T], [92,36|EscT]) :-
  latex_escape2(T, EscT).
latex_escape2([37|T], [92,37|EscT]) :-
  latex_escape2(T, EscT).
latex_escape2([H|T], [H|EscT]) :-
  H \= 95,
  latex_escape2(T, EscT).