:- module(worker, [start_worker/2,load/1]).

:- use_module(library(fastrw),[ fast_buf_read/2,fast_buf_write/3 ]).
:- use_module('../../src/module_information.pl').
:- use_module(library(lists)).

:- module_info(group,experimental).
:- module_info(description,'This is the interface to C code for distributed model checking.').
:- module_info(revision,'$Rev: 9609 $').
:- module_info(lastchanged,'$LastChangedDate: 2011-11-18 10:43:02 +0100 (Fr, 18 Nov 2011) $').


foreign_resource('worker', [worker_main,put_result, put_successor]).
foreign(worker_main, c, worker_main(+string, +string)).
foreign(put_result, put_result(+address(char),+integer)).
foreign(put_successor, put_successor(+address(char),+integer,+integer)).

% The worker gets some initialisation code, e.g., a B model and the instruction what to check.
load(Init) :-
    fast_buf_read(Initialisation,Init), 
    callback_load(Module,Pred),
    call(Module:Pred,Initialisation).


:-dynamic assertion_state/1.
:-dynamic assertion_task/2.


% The worker gets a Workpackage. It must produce a list of computation results 
% (which are being forwarded to the master) and a list of successor Workpackages. 
% The successor workpackages are sent to this predicate at a later point in time. 
% It is not guaranteed that it will be sent to this process, so they must contain all  
% required information to process the workpackage. 
% It can be assumed that all processes that might receive this workpackage executed 
% the same initialisation code.
%process(_,[some_result(4), another_result(x)], [foo(baz), doh(dude)]).



process(assertion,[],S) :- findall(check(N),assertion_task(N,_),S).

process(check(N),R,[]) :- assertion_task(N,A), 
                          check_assertion(A,R).


process(root,R,S) :- process(state(root),R,S).
process(state(State),Res,Successors) :- !, % print(treating(State)),nl,
 % statistics(walltime,[T1,_]),
 % statistics(runtime,[T3,_]),
  % prolog_flag(gc,G),print(gc(G)),nl,
   specfile:prepare_state_for_specfile_trans(State,PreparedState), % will perform unpacking of constants just once, for example
   (debug_mode->count(X),assert(state(X,State)),inc;true),
   (specfile:state_corresponds_to_initialised_b_machine(PreparedState,S2) ->
      bmachine:b_get_invariant_from_machine(Invariant),
    ( b_interpreter:b_test_boolean_expression_for_ground_state(Invariant,[],S2)
      % TO DO: add time_out, specialized invariants,...
      -> Res = []  %, print(invariant_ok),nl
      ; Res = [invariant_violated(state(S2))]
    )
  %  ,print(checked_inv(Res)),nl
    ; Res = []  %, print(not_initialised(PreparedState)),nl
   ),
   
   findall(state(NewState),
           compute_successors(PreparedState,_ActionName,NewState),
           S), 
  (debug_mode->dump_trans(PreparedState);true),   
 %  print(successors(S)),nl,
%statistics(walltime,[T2,_]),
%statistics(runtime,[T4,_]),
%TW is T2 - T1,
%TR is T4 - T3,

%(TW > 10 -> ( print(time(TW,TR)),nl,statistics,print(x),nl);true),
    Successors = S.
process(max_reached,Res,[]) :- !,
   (succeed_max:max_reached -> Res=true ; Res=false).
process(X,_,S) :- print(unhandled(X)),nl, S=[].


check_assertion(A,R) :- assertion_state(State),
    ( b_interpreter:b_test_boolean_expression_for_ground_state(A,[],State) -> R = [] ; R = [assertion_violated(A)] ).




compute_successors(CurState,ActionName,NewState) :-
   specfile:specfile_possible_trans_name(CurState,ActionName),
   specfile:specfile_trans(CurState,ActionName,_Act,NewState,_TransInfo,_Residue).
     % TO DO: check Residue, add time-out, ...

% dump transitons for debugging
dump_trans(S) :- findall(trans(ActionName,NewState),
                         compute_successors(S,ActionName,NewState), List),
                 dump_trans(List,S).
dump_trans([],_).
dump_trans([trans(Action,H)|T],SS) :- assert(trans(Action,SS, H)), dump_trans(T,SS).

:- dynamic callback_load/2.

:- dynamic is_initialised/0.
:- dynamic count/1.
:- dynamic state/2.
:- dynamic trans/3.
:- dynamic count_transitions/1.
:- dynamic debug_mode/0.
:- dynamic local_mode/0.
:- dynamic ids/1.
:- dynamic queue/2.

set_debug :- assert(debug_mode).
set_local :- assert(local_mode).

:- assert(count(0)).
inc :- count(X), retractall(count(_)), X1 is X + 1, assert(count(X1)).

:- assert(ids(1)).
next_id :- ids(X), retractall(ids(_)), X1 is X + 1, assert(ids(X1)).

:- assert(count_transitions(0)).
inc_transitions :- count_transitions(X), retractall(count_transitions(_)), X1 is X + 1, assert(count_transitions(X1)).


init_zmq_worker :- is_initialised,!.
init_zmq_worker :-
    % set_prolog_flag(gc,off), % disable garbage collection. 
    load_foreign_resource(library(worker)),
    assert(is_initialised).

:- meta_predicate start_worker(+,1).
start_worker(Host,LoadCallbackModule:LoadCallback) :-
    init_zmq_worker,
    retractall(callback_load(_,_)), assert(callback_load(LoadCallbackModule,LoadCallback)),
    model_checker:depth_breadth_first_mode(MODE),
    worker_main(Host, MODE).

local_zmq_check(N) :-
  queue(N, Workpackage),
  process(Workpackage, R, S), 
  add_result(R), 
  add_successor(S).

zmq_check(_WP) :- 
	fast_buf_read(Workpackage,_WP), 
  process(Workpackage, R, S), 
  successors_without_deadlock(S,_W,R,R2),
  add_result(R2), 
  add_successor(S).

successors_with_deadlock([],W,R,[deadlock(W)|R]).
successors_with_deadlock([_,_],_,R,R).

successors_without_deadlock(_,_,R,R).


	
add_result([]).
add_result([H|T]) :-  fast_buf_write(H,Len,Addr), put_result(Addr,Len), add_result(T).

add_successor([]).
add_successor([H|T]) :-
  (debug_mode->inc_transitions; true),
  (local_mode->ids(X), next_id, assert(queue(X, H)) ; X = 0),
  fast_buf_write(H,_Len,_Addr),
  put_successor( _Addr,_Len, X),
  add_successor(T).


dump :- 
  portray_clause(start(0)), print('.'),nl,
  (succeed_max:max_reached -> print('/* max_reached */'),nl ; true),
  portray_clause((prop(S,'='(Var,Val)) :- state(S,L), member(bind(Var,Val),L))), print('.'),nl,
  tools:print_dynamic_pred(worker,trans,3),
  tools:print_dynamic_pred(worker,state,2)
.
