% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(probvm, [plugin_info/1]).

plugin_info([
             % The name of the Plugin
             name = 'ProB Virtual Machine'
             % Author is not used yet
            ,author = 'Daniel Plagge'
             % marks the plugin as unstable or stable
            ,status = unstable
             % A list of supported file extensions together with a file description
            %,file_extensions = [ ('Plugin demonstration spec', ['.bvm']) ]
             % The predicate that is used to load a file. The first argument is the
             % filename, the second the file extension (like '.mch').
            %,load_file = load_spec/1
             % The transition skeleton initialises a transition variable in such
             % way that only a part of all transitions match it. It is used to
             % separate time outs and maximum number of computations for several
             % transition classes (like different operations).
            %,transition_skeleton = trans_skel/1
             % The predicate that computes a transition in the state-space, its
             % arguments are SourceState,Operation,DestinationState,ListOfOperationInfos.
            ,transition = do_transition/4
             % The predicate that computes an initial state, its arguments are
             % Operation,DestinationState,ListOfOperationInfos.
            ,initialisation = do_initialisation/3
             % A predicate that returns a property for a given state, its arguments
             % are State,Property.
            ,state_property = find_state_property/2
             % A predicate that receives a list of expressions,predicates and
             % predicates on transitions as input (list of codes) and returns
             % the parsed expressions.
             % Its arguments are ListOfExpressionStrings,ListOfPredicateStrings,
             % ListOfTransitionPredicateStrings,ListOfExpressions,ListOfPredicates,
             % ListOfTransitionPredicates.
            %,parser = parse_spec/6
             % A predicate that computes a value of an expression (returned by
             % the parser above) in a state. Its arguments are Expression,State
            %,compute_expression = compute/2
             % A predicate that evaluates a predicate (returned by the parser above)
             % in a state. Its arguments are Predicate,State
            %,evaluate_predicate = eval/2
             % A predicate that evaluates a predicate (returned by the parser above)
             % for a transition. Its arguments are Predicate,Transition
            %,evaluate_transition = evaltrans/2
             % A predicate that checks if a state is initialised. Its argument is a State
            ,is_initialised_state = is_init/1
             % Pretty-print a transition, the first argument is the transition, the
             % second -the result- is a list of codes.
            %,prettyprint_transition = pp_trans/2
             % Pretty-print a property, the first argument is the property, the
             % second -the result- is a list of codes.
            %,prettyprint_property = pp_prop/2
            ,plugin_for_modes = [b]
            ,plugin_init=plugin_init/0
            ]).

:- dynamic state_schema/2, init/3, operation/4.

:- use_module(library(lists)).
:- use_module(library(terms),[term_variables/2]).

:- use_module(probsrc(tools)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(translate),[translate_bvalue/2]).
:- use_module(probsrc(state_space),[state_space_initialise/0]).
:- use_module(probsrc(error_manager),[add_error_and_fail/3]).

:- use_module(probvm_compiler).
:- use_module(probvm_interpreter).
:- use_module(probvm_optimize).

:- use_module(probsrc(module_information)).
:- module_info(group,plugin_vm).
:- module_info(description,'This is an experimental plugin that translates B to an internal (byte-code like) format before interpreting it.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

plugin_init :-
    state_space_initialise,
    reset,
    register_variables,
    register_initialisation,
    register_ops,
    pretty_print.

do_initialisation(Operation,State,[]) :-
    init(Operation,Spec,State),
    interpret(Spec).

do_transition(State,Operation,NewState,[]) :-
    operation(Operation,Spec,State,NewState),
    interpret(Spec).

find_state_property(State,Prop) :-
    state_schema(Access,State),
    member(Name=Value,Access),
    translate_bvalue(Value,PPValue),
    ajoin([Name,' = ',PPValue],Prop).

reset :-
    retractall(state_schema(_,_)),
    retractall(init(_,_,_)),
    retractall(operation(_,_,_,_)).

register_variables :-
    b_get_machine_constants(Consts),
    b_get_machine_variables(Vars),
    create_state_schema(Vars,VarAccess,VarState),
    ( Consts = [] ->
        assert( state_schema(VarAccess,initialised_b(VarState)) )
    ; Consts = [_|_] ->
        create_state_schema(Consts,ConstAccess,ConstState),
        append(ConstAccess,VarAccess,Access),
        assert( state_schema(ConstAccess,constants(ConstState)) ),
        assert( state_schema(Access,initialised_b(ConstState,VarState)) )).

register_initialisation :-
    ( b_machine_has_constants ->
        CSState = constants(_),
        state_schema(CSAccess,CSState),
        create_constants_setup(SetupSkel,CSAccess,CSInstructions),
        assert_init(SetupSkel,CSInstructions,CSState),

        CState = constants(ConstState),
        IState = initialised_b(ConstState,_VarState),
        state_schema(CAccess,CState),
        state_schema(IAccess,IState),
        create_initialisation(IOpSkel,CAccess,IAccess,InitInstructions),
        assert_op(IOpSkel,InitInstructions,CState,IState)
    ; otherwise ->
        State = initialised_b(_),
        state_schema(Access,State),
        create_initialisation(IOpSkel,none,Access,Instructions),
        assert_init(IOpSkel,Instructions,State)).

create_constants_setup('$setup_constants'(Access),Access,Instructions) :-
    create_cmp_env(Access,none,Env),
    b_get_properties_from_machine(Predicate),
    ( compile_predicate(Predicate,Env,Instructions1,[]) ->
        optimize_instructions(Instructions1,Instructions)
    ; otherwise ->
        add_error_and_fail(probvm,'Unable to compile properties',properties)).

% TODO: PostAccess includes the constants, too
create_initialisation('$initialisation'(PostAccess),PreAccess,PostAccess,Instructions) :-
    create_cmp_env(PreAccess,PostAccess,Env),
    b_get_initialisation_from_machine(Subst,_),
    ( compile_substitution(Subst,Env,Instructions1,[]) ->
        optimize_instructions(Instructions1,Instructions)
    ; otherwise ->
        add_error_and_fail(probvm,'Unable to compile initialisation',initialisation)).

assert_init(OpSkel,Instructions,State) :-
    assert( init(OpSkel,Instructions,State) ).
assert_op(OpSkel,Instructions,PreState,PostState) :-
    assert( operation(OpSkel,Instructions,PreState,PostState) ).

register_ops :-
    findall( op(Name,Results,Parameters,Body,Info),
             ( b_get_machine_operation(Name,Results,Parameters,Body),
                 get_operation_info(Name,Info)),
             Ops),
    compile_ops(Ops).
compile_ops([]).
compile_ops([Op|Orest]) :-
    compile_op(Op),
    compile_ops(Orest).
compile_op(op(Name,[],Parameters,Body,Info)) :-
    memberchk(modifies(Modifies),Info),
    operation_state_schema(PreAccess,PreState),
    operation_state_schema(PostAccess,PostState),
    ( op_instructions(Name,Parameters,Body,PreAccess,PostAccess,Modifies,Skeleton,Instructions1) ->
        optimize_instructions(Instructions1,Instructions),
        assert_op(Skeleton,Instructions,PreState,PostState)
    ; otherwise ->
        write('Unable to compile operation '),write(Name),nl).
op_instructions(Name,Parameters,Body,PreAccess,PostAccess,Modifies,Operation,Instructions) :-
    unmodified_vars(PreAccess,PostAccess,Modifies),
    create_cmp_env(PreAccess,PostAccess,Env),
    op_skeleton(Name,Parameters,Operation,Env,SubEnv,Instructions,BodyInstructions),
    balance_substitutions(Body,Parameters,BalBody),
    compile_substitution(BalBody,SubEnv,BodyInstructions,[]).

operation_state_schema(Access,State) :-
    ( b_machine_has_constants -> State = initialised_b(_Const,_Vars)
    ; otherwise               -> State = initialised_b(_Vars)),
    state_schema(Access,State),!.

unmodified_vars([],[],_).
unmodified_vars([Id=V1|R1],[Id=V2|R2],Modifies) :-
    ( memberchk(Id,Modifies) -> true
    ; otherwise -> V1 = V2),
    unmodified_vars(R1,R2,Modifies).

op_skeleton(Name,Parameters,Op,PreState,State) -->
    { length(Parameters,L),
      functor(Op,Name,L),
      Op =.. [Name|PValues]},
    add_enumerated_identifier(Parameters,PValues,PreState,State).
    

create_state_schema([],[],[]).
create_state_schema([Id|Irest],[Name=Value|Arest],[Value|Vrest]) :-
    get_texpr_id(Id,Name),
    create_state_schema(Irest,Arest,Vrest).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% balance ifs: each branch should modify the same set of variables

balance_substitutions(TSubst,TLocals,TBal) :-
    get_texpr_ids(TLocals,Locals),
    balance_substitutions2(TSubst,Locals,TBal).
balance_substitutions2(TSubst,Locals,TBal) :-
    remove_bt(TSubst,Subst,Bal,TBal),
    balance_substitutions3(Subst,Locals,Bal).
balance_substitutions3(assign(L,R),_,assign(L,R)).
balance_substitutions3(becomes_element_of(L,R),_,becomes_element_of(L,R)).
balance_substitutions3(parallel(Ss),Locals,parallel(Ns)) :-
    balance_substitutions_l(Ss,Locals,Ns).
balance_substitutions3(any(Ids,P,S),Locals,any(Ids,P,N)) :-
    balance_substitutions2(S,Locals,N).
balance_substitutions3(precondition(Cond,Subst),Locals,precondition(Cond,BSubst)) :-
    balance_substitutions2(Subst,Locals,BSubst).
balance_substitutions3(select([When]),Locals,select([BWhen])) :-
    balance_substitutions2(When,Locals,BWhen).
balance_substitutions3(select_when(P,S),Locals,select_when(P,B)) :-
    balance_substitutions2(S,Locals,B).
balance_substitutions3(if(IfList),Locals,if(NewIfList)) :-
    create_texpr(if(IfList),subst,[],TExpr),
    btypechecker:get_accessed_vars(TExpr,Locals,Modifies,_),
    balance_iflist(IfList,Locals,Modifies,NewIfList1),
    ( has_else(IfList) -> NewIfList1 = NewIfList
    ; otherwise ->
        append(NewIfList1,[Else],NewIfList),
        create_else(Modifies,Else)).
balance_substitutions_l([],_,[]).
balance_substitutions_l([S|Srest],Locals,[N|Nrest]) :-
    balance_substitutions2(S,Locals,N),
    balance_substitutions_l(Srest,Locals,Nrest).

has_else(IfList) :-
    last(IfList,TElse),
    get_texpr_expr(TElse,if_elsif(P,_)),
    is_truth(P).
create_else(Ids,Else) :-
    create_texpr(truth,pred,[],Truth),
    create_texpr(if_elsif(Truth,StaySame),subst,[],Else),
    create_subst(Ids,StaySame).

balance_iflist([],_,_,[]).
balance_iflist([TIE|Trest],Locals,Modifies,[New|Nrest]) :-
    balance_iflist2(TIE,Locals,Modifies,New),
    balance_iflist(Trest,Locals,Modifies,Nrest).
balance_iflist2(TIE,Locals,AllModifies,TNew) :-
    get_texpr_expr(TIE,if_elsif(P,S)),
    btypechecker:get_accessed_vars(S,Locals,Modifies,_),
    remove_all(AllModifies,Modifies,Unmodified),
    ( Unmodified == [] ->
        TNew = TIE
    ; otherwise ->
        create_subst(Unmodified,Subst),
        create_texpr(if_elsif(P,SP),subst,[],TNew),
        create_texpr(parallel([S,Subst]),subst,[],SP)).
create_subst(Ids,Subst) :-
    create_texpr(assign(TIds,TIds),subst,[],Subst),
    make_pseudotyped(Ids,TIds).
make_pseudotyped([],[]).
make_pseudotyped([Id|Rest],[TId|Trest]) :-
    create_texpr(identifier(Id),_,[],TId),
    make_pseudotyped(Rest,Trest).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pretty-print
pretty_print :-
    pp_init.
pretty_print :-
    write('Operations:\n'),
    fail.
pretty_print :-
    pp_operations.
pretty_print.

pp_init :-
    init(Op,Spec,State),
    write('Initialisation:\n'),
    state_names('State_',State),
    local_names('Local',Spec),
    pp_op(Op),
    pp_state(state,State),
    pp_spec(Spec),
    fail.

pp_operations :-
    operation(Op,Spec,Pre,Post),
    parameter_names(Op),
    state_names('Pre_',Pre),
    state_names('Post_',Post),
    local_names('Local',Spec),
    pp_op(Op),
    pp_state(pre_state,Pre),
    pp_state(post_state,Post),
    pp_spec(Spec),
    fail.

state_names(Prefix,State) :-
    state_schema(Access,State),
    state_names2(Access,Prefix).
state_names2([],_).
state_names2([Name=Var|Vrest],Prefix) :-
    ( var(Var) -> atom_concat(Prefix,Name,Id), Var = '$VAR'(Id)
    ; true),
    state_names2(Vrest,Prefix).

parameter_names(Op) :-
    Op =.. [Name|Args],
    b_get_machine_operation(Name,_,TParams,_),
    get_texpr_ids(TParams,Ids),
    create_name_var_tuples(Ids,Args,Access),
    state_names2(Access,'Arg_'),!.
parameter_names(_).
create_name_var_tuples([],[],[]).
create_name_var_tuples([N|Nrest],[V|Vrest],[N=V|Trest]) :-
    create_name_var_tuples(Nrest,Vrest,Trest).

local_names(Prefix,Expr) :-
    term_variables(Expr,Vars),
    local_names2(Vars,Prefix,1).
local_names2([],_,_).
local_names2([V|Vrest],Prefix,N) :-
    ( var(V) ->
        atom_codes(Prefix,PC),
        number_codes(N,NC),
        append(PC,NC,Codes),
        atom_codes(Id,Codes),
        V='$VAR'(Id)
    ; otherwise -> true),
    N2 is N+1,
    local_names2(Vrest,Prefix,N2).

pp_op(Op) :-
    write_term(Op,[numbervars(true)]),
    write(' :-\n').
pp_state(Desc,State) :-
    functor(F,Desc,1),
    arg(1,F,State),
    pp_spec(F).
pp_spec([]) :- !.
pp_spec([I|Irest]) :- !,pp_spec(I),pp_spec(Irest).
pp_spec(I) :-
    write('    '),
    write_term(I,[numbervars(true)]),
    nl.
