% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(units_domain, [
                           same_units/2,
                           lub/3,
                           units_domain_multiplication/3,
                           units_domain_multiplication_conversion/3,
                           units_domain_division/3,
                           units_domain_division_conversion/3,
                           units_domain_subtraction_conversion/3,
                           units_domain_addition_conversion/3,
                           units_domain_power_of/3,
                           units_domain_member/2,
                           units_domain_relation/3,
			   involved_in_constraint/1,
  			   power_of_multiply_units/3
                        ]).

:- use_module(library(lists)).
:- use_module(library(clpfd)).
:- use_module(library(sets)).

:- use_module(probsrc(self_check)).

:- use_module(units_conversions).

:- use_module(probsrc(module_information)).
:- module_info(group,plugin_units).
:- module_info(description,'Units Plugin: The abstract Domain.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- attribute mfactors/2, dfactors/2, pfactors/2, constraint/0.

involved_in_constraint(X) :- var(X), !, get_atts(X, constraint).
involved_in_constraint([H|T]) :-
    !, (involved_in_constraint(H) ; involved_in_constraint(T)).
involved_in_constraint(Pred) :-
    !, Pred =.. [Head|Tail],
    Tail = [_|_],
    (involved_in_constraint(Head) ; involved_in_constraint(Tail)).

add_constraint_marker(A,B) :-
    (var(A) -> put_atts(A, constraint) ; true),
    (var(B) -> put_atts(B, constraint) ; true).

verify_attributes(Var, Value, Goal) :-
    % the variable might be the result of a multiplication
    % -> propagate inwards
    get_atts(Var, mfactors(A,B)), !,
    (var(Value)         -> Goal = [] ;
    (A==B, var(A))      -> Goal=[power_of_multiply_units(2,A,Value)] ;
    (var(A), nonvar(B)) -> Goal=[units_domain_division(Value,B,A)] ;
    (var(B), nonvar(A)) -> Goal=[units_domain_division(Value,A,B)] ;
    otherwise           -> Goal=[]).
verify_attributes(Var, Value, Goal) :-
    % the variable might be the result of a division
    % -> propagate inwards
    get_atts(Var, dfactors(A,B)), !,
    (var(Value)          -> Goal=[] ;
    (var(A), nonvar(B))  -> Goal=[units_domain_multiplication(Value,B,A)] ;
    (var(B), nonvar(A))  -> Goal=[units_domain_division(A,Value,B)] ;
    otherwise            -> Goal=[]).
verify_attributes(Var, Value, Goal) :-
    % the variable might be the result of a exponentiation
    % -> propagate inwards
    get_atts(Var, pfactors(A,B)), !,
    (var(Value)         -> Goal=[] ;
    (var(A), nonvar(B)) -> Goal=[power_of_multiply_units(B,A,Value)] ;
    otherwise           -> Goal=[]).
% rescue constraint attribute when unifying
verify_attributes(Var, Value, []) :-
    get_atts(Var, constraint), !,
    (var(Value) -> put_atts(Value, constraint) ; true).
%verify_attributes(Var, Value, []) :-
    % not a result of multiplication or division
%    true.

one_is_var(A,B) :- var(A) ; var(B).

% special case "power_of":
% A is the unit part of an integer. This might be a variable.
units_domain_power_of(A,b(integer(B,_,_)),C) :- var(A), put_atts(A,constraint), put_atts(C, pfactors(A,B)), !.
% B is a raw syntax element. if it is a known integer literal, compute exponent
units_domain_power_of(A, b(integer(X),_,_),C) :-
    power_of_multiply_units(X,A,C), !.
% A might be top containing a list of known units
units_domain_power_of(top(A), b(integer(X),_,_),top(C)) :-
    maplist(power_of_multiply_units(X),A,C), !.
% otherwise, we can not infer the unit anymore
units_domain_power_of(_A,_B,top([])) :- !.

power_of_multiply_units(_,[],[]).
power_of_multiply_units(X,[[Factor1,Unit,Pow1]|T1],[[Factor2,Unit,Pow2]|T2]) :-
    Factor2 #= Factor1 * X, Pow2 #= Pow1 * X,
    power_of_multiply_units(X,T1,T2).
    
% Conversions between units. Conversion_factor and conversion_add is supplied in its own module.

units_domain_addition_conversion(_, X, _) :- var(X), !.
units_domain_addition_conversion(b(integer(X),_,_), [Unit|Units], [Unit2|Units]) :-
	repeat_conversion_add(X,Unit2,Unit).
units_domain_addition_conversion(BInt,[Unit|Units],[Unit|Units2]) :-
	units_domain_addition_conversion(BInt,Units,Units2).
	
units_domain_subtraction_conversion(_, X, _) :- var(X), !.
units_domain_subtraction_conversion(b(integer(X),_,_), [Unit|Units], [Unit2|Units]) :-
	repeat_conversion_add(X,Unit,Unit2).
units_domain_subtraction_conversion(BInt,[Unit|Units],[Unit|Units2]) :-
	units_domain_subtraction_conversion(BInt,Units,Units2).

:- assert_must_succeed((units_domain_multiplication_conversion(b(integer(10),_,_),
                       _,X),var(X))).
:- assert_must_succeed((units_domain_multiplication_conversion(b(integer(10),_,_),
                       [[1,m,1]],X),X=[[0,m,1]])).
:- assert_must_succeed((units_domain_multiplication_conversion(b(integer(60),_,_),
                       [[1,mins,1]],X),X=[[1,s,1]])).
:- assert_must_succeed((units_domain_multiplication_conversion(b(integer(10),_,_),
                       [[1,m,1],[1,s,1]],X),X=[[0,m,1],[1,s,1]])).
:- assert_must_succeed((units_domain_multiplication_conversion(b(integer(10),_,_),
                       [[1,s,1],[1,m,1]],X),X=[[1,s,1],[0,m,1]])).
:- assert_must_succeed((units_domain_multiplication_conversion(b(integer(60),_,_),
                       [[1,m,1],[1,mins,1]],X),X=[[1,m,1],[1,s,1]])).
:- assert_must_succeed((units_domain_multiplication_conversion(b(integer(60),_,_),
                       [[1,mins,1],[1,m,1]],X),X=[[1,s,1],[1,m,1]])).
units_domain_multiplication_conversion(_, X, _) :- var(X), !.
units_domain_multiplication_conversion(b(integer(X),_,_), [Unit|Units], [Unit2|Units]) :-
    repeat_conversion_factor(X,Unit,Unit2).
units_domain_multiplication_conversion(BInt,[Unit|Units],[Unit|Units2]) :-
	units_domain_multiplication_conversion(BInt,Units,Units2).

:- assert_must_succeed((units_domain_division_conversion(b(integer(10),_,_),
                       _,X),var(X))).
:- assert_must_succeed((units_domain_division_conversion(b(integer(10),_,_),
                       [[1,m,1]],X),X=[[2,m,1]])).
units_domain_division_conversion(_, X, _) :- var(X), !.
units_domain_division_conversion(b(integer(X),_,_), [Unit|Units], [Unit2|Units]) :-
	repeat_conversion_factor(X,Unit2,Unit).
units_domain_division_conversion(BInt,[Unit|Units],[Unit|Units2]) :-
	units_domain_division_conversion(BInt,Units,Units2).
    
repeat_conversion_factor(X,_,_) :- X < 1, !, fail.
repeat_conversion_factor(1.0,Unit,Unit).
repeat_conversion_factor(Factor,Unit,UnitOut) :-
    var(UnitOut)
    -> conversion_factor(Unit2,Factor2,Unit),
       FactorOut is Factor / Factor2,
       repeat_conversion_factor(FactorOut,Unit2,UnitOut)
    ;  conversion_factor(UnitOut,Factor2,Unit2),
       FactorOut is Factor / Factor2,
       repeat_conversion_factor(FactorOut, Unit, Unit2).
       
repeat_conversion_add(X,_,_) :- X < 0, !, fail.
repeat_conversion_add(0,Unit,Unit).
repeat_conversion_add(Factor,Unit,UnitOut) :-
    var(UnitOut)
    -> conversion_add(Unit2,Factor2,Unit),
       FactorOut is Factor - Factor2,
       repeat_conversion_add(FactorOut,Unit2,UnitOut)
    ;  conversion_add(UnitOut,Factor2,Unit2),
       FactorOut is Factor - Factor2,
       repeat_conversion_add(FactorOut, Unit, Unit2).
    

% Operations on Integers
:- assert_must_succeed(units_domain_multiplication(top([]),[[1,m,1]],top([]))).
:- assert_must_succeed(units_domain_multiplication(top([[[1,m,1]],[[2,m,1]]]),[[1,m,1]],top([[[2,m,2]],[[3,m,2]]]))).
:- assert_must_succeed(units_domain_multiplication(top([[[1,m,1]],[[2,m,1]]]),top([[[1,m,1]],[[2,m,1]]]),
						   top([[[2,m,2]],[[3,m,2]],[[3,m,2]],[[4,m,2]]]))).
:- assert_must_succeed(units_domain_multiplication([[1,m,1]],[[1,m,1]],[[2,m,2]])).
:- assert_must_succeed((units_domain_multiplication(A,B,C),var(A),var(B),var(C))).
:- assert_must_succeed((units_domain_multiplication(_,[[1,m,1]],Y), var(Y))).
:- assert_must_succeed((units_domain_multiplication([[1,m,1]],_,Y), var(Y))).
:- assert_must_succeed(units_domain_multiplication([[1,m,1],[1,a,2]],[[1,m,1]],[[2,m,2],[1,a,2]])).
:- assert_must_succeed((units_domain_multiplication([[0,m,1],[0,s,-1]],[[0,s,1]],Res), Res = [[0,m,1]])).
units_domain_multiplication(A,B,C) :- one_is_var(A,B), !, add_constraint_marker(A,B), put_atts(C, mfactors(A,B)).
units_domain_multiplication(top(A),top(B),top(C)) :- !,
    map_product(units_domain_multiplication,A,B,C).
units_domain_multiplication(top(A),B,top(C)) :- !,
    maplist(units_domain_multiplication(B),A,C).
units_domain_multiplication(A,top(B),top(C)) :- !,
    maplist(units_domain_multiplication(A),B,C).
units_domain_multiplication(Op1,[],Op1) :- !.
units_domain_multiplication([],Op2,Op2) :- !.
units_domain_multiplication([[Factor1,Unit1,Pow1]|Units], UnitsOp2, Res2) :- !,
    (selectchk([Factor2,Unit1,Pow2],UnitsOp2,UnitsOp22)
    ->  Factor3 is Factor1+Factor2, Pow3 is Pow1+Pow2,
        Res1 = [Factor3,Unit1,Pow3], units_domain_multiplication(Units,UnitsOp22,UnitsResult)
    ;   Res1 = [Factor1,Unit1,Pow1], units_domain_multiplication(Units,UnitsOp2,UnitsResult)
    ),
    (Res1 = [0,_,0] -> Res2 = UnitsResult ; Res2 = [Res1|UnitsResult]).

:- assert_must_succeed(units_domain_division(top([]),[[1,m,1]],top([]))).
:- assert_must_succeed(units_domain_division([[1,m,1]],top([]),top([]))).
:- assert_must_succeed(units_domain_division(top([[[1,m,1],[1,a,2]],[[1,m,1],[2,a,2]]]),[[1,m,1]],top([[[1,a,2]],[[2,a,2]]]))).
:- assert_must_succeed(units_domain_division([[1,m,1]],top([[[1,m,2]],[[1,m,3]]]),top([[[0,m,-1]],[[0,m,-2]]]))).
:- assert_must_succeed(units_domain_division(top([[[1,m,1]],[[2,m,1]]]),top([[[1,m,2]],[[1,m,3]]]),
					     top([[[0,m,-1]],[[0,m,-2]],[[1,m,-1]],[[1,m,-2]]]))).
:- assert_must_succeed(units_domain_division([[1,m,1]],[[1,m,1]],[])).
:- assert_must_succeed(units_domain_division([[1,m,1],[1,a,2]],[[1,m,1]],[[1,a,2]])).
:- assert_must_succeed(units_domain_division([[1,m,1]],[[1,m,1],[1,a,2]],[[-1,a,-2]])).
:- assert_must_succeed((units_domain_division(A,B,C),var(A),var(B),var(C))).
:- assert_must_succeed((units_domain_division(A,A,C),var(A),C==[])).
:- assert_must_succeed((units_domain_division(_,[[1,m,1]],Y), var(Y))).
:- assert_must_succeed((units_domain_division([[1,m,1]],_,Y), var(Y))).
:- assert_must_succeed((units_domain_division([[0,m,1],[0,s,1]],[[0,s,1]],Res), Res = [[0,m,1]])).
units_domain_division(A,B,C) :- var(A), A==B, !, C=[]. 
units_domain_division(A,B,C) :- one_is_var(A,B), !, add_constraint_marker(A,B), put_atts(C, dfactors(A,B)).
units_domain_division(top(A),top(B),top(C)) :-
    !, map_product(units_domain_division,A,B,C).
units_domain_division(top(A),B,top(C)) :-
    !, maplist(units_domain_division_reversed_arguments(B),A,C).
units_domain_division(A,top(B),top(C)) :-
    !, maplist(units_domain_division(A),B,C).
units_domain_division(Op1,[],Op1) :- !.
units_domain_division([],Op2,Op3) :- !,
    division_reverse_signs(Op2,Op3).
units_domain_division([[Factor1,Unit1,Pow1]|Units], UnitsOp2, Res2) :- !,
    (selectchk([Factor2,Unit1,Pow2],UnitsOp2,UnitsOp22)
    ->  Factor3 is Factor1-Factor2, Pow3 is Pow1-Pow2,
        Res1 = [Factor3,Unit1,Pow3], units_domain_division(Units,UnitsOp22,UnitsResult)
    ;   Res1 = [Factor1,Unit1,Pow1], units_domain_division(Units,UnitsOp2,UnitsResult)
    ),
    (Res1 = [0,_,0] -> Res2 = UnitsResult ; Res2 = [Res1|UnitsResult]).

division_reverse_signs([],[]).
division_reverse_signs([[Factor,Unit,Pow]|T],[[Factor2,Unit,Pow2]|T2]) :-
    Factor2 is -Factor, Pow2 is -Pow,
    division_reverse_signs(T,T2).

% reverse argument order - for maplist
units_domain_division_reversed_arguments(A,B,C) :-
    units_domain_division(B,A,C).

% Operations on Sets
:- assert_must_succeed(units_domain_member(type,set(type))).
:- assert_must_fail(units_domain_member(integer([[1,m,1]]), set(integer([[2,m,1]])))).
units_domain_member(A,B) :- set(A) = B, !.

:- assert_must_succeed(units_domain_relation(set(type),set(type2),set(set(couple(type,type2))))).
units_domain_relation(set(A),set(B),set(set(couple(A,B)))).



% COMPARISON OF UNITS
% same_units/2 checks if two units are equal
% lub/3 infers least upper bound - a common unit higher in the lattice than the first ones

% same_units/2 - test cases
:- assert_must_fail(same_units(type(_), other_type)).
:- assert_must_fail(same_units(type(type2(_)), type(type3(_)))).
:- assert_must_succeed(same_units(type([[1,m,1],[1,s,2]]), type([[1,s,2],[1,m,1]]))).
:- assert_must_succeed(same_units(type([[-1,m,1],[1,s,2]]), type([[-1,s,2],[1,m,1]]))).
:- assert_must_succeed(same_units(type([[2,m,1],[1,s,2]]), type([[2,s,2],[1,m,1]]))).
:- assert_must_succeed(same_units(type([[-1,m,1],[0,s,2]]), type([[-2,s,2],[1,m,1]]))).
:- assert_must_succeed(same_units(type(top([])), type(top([])))).
:- assert_must_succeed(same_units(type(X), type(X))).

% lub/3 - test cases
:- assert_must_succeed(lub(type([[1,m,1],[1,s,2]]), type([[1,s,2],[1,m,1]]), type([[1,m,1],[1,s,2]]))).
:- assert_must_succeed(lub(type([[-1,m,1],[1,s,2]]), type([[-1,s,2],[1,m,1]]), type([[-1,m,1],[1,s,2]]))).
:- assert_must_succeed(lub(type(top([])), type([[1,m,1]]), type(top([[[1,m,1]]])))).
:- assert_must_succeed(lub(type(top([])), type([[1,s,2],[1,m,1]]), type(top([[[1,s,2],[1,m,1]]])))).
:- assert_must_succeed(lub(type([[1,m,1],[5,s,2]]), type(top([[[1,m,1]]])), type(top([[[1,m,1],[5,s,2]],[[1,m,1]]])))).
:- assert_must_succeed(lub(type(X), type(X), type(X))).
:- assert_must_succeed(lub([[3,m,1]],[[3,m,1],[0,h,-1]],top([[[3,m,1]],[[3,m,1],[0,h,-1]]]))).
:- assert_must_succeed(lub(integer(_),integer(top([[[0,m,1]],[[0,m,1],[0,s,-1]],[[0,m,1]]])),
			   integer(top([[[0,m,1]],[[0,m,1],[0,s,-1]],[[0,m,1]]])))).
:- assert_must_succeed(lub(integer(top([[[0,m,1]],[[0,m,1],[0,s,-1]],[[0,m,1]]])),integer(_),
			   integer(top([[[0,m,1]],[[0,m,1],[0,s,-1]],[[0,m,1]]])))).
:- assert_must_succeed(lub(top([[[0,m,1]],[[0,m,1],[0,s,-1]]]),[[0,m,1],[0,s,-1]],
			   top([[[0,m,1]],[[0,m,1],[0,s,-1]]]))).
:- assert_must_succeed(lub([[0,m,1],[0,s,-1]],top([[[0,m,1]],[[0,m,1],[0,s,-1]]]),
			   top([[[0,m,1]],[[0,m,1],[0,s,-1]]]))).
:- assert_must_succeed(lub(top([[[0,m,1]],[[0,m,1],[0,s,-1]]]),top([[[0,m,1],[0,s,-1]],[[0,m,1]]]),
			   top([[[0,m,1],[0,s,-1]],[[0,m,1]]]))).


% two units are identical, if the lub is again one of the two units
same_units(A,B) :- lub(A,B,C), !, (A=C ; B=C).

% lub/3 - implementation
lub(A,B,C) :- A=B, !, B=C.
lub(A,B,C) :- lub_domain_level(A,B,C), !.
lub(top(A),top(B),top(C)) :-
    !, union(A,B,C).
lub(top(A),B,top(C)) :-
    !, union(A,[B],C).
lub(A,top(B),top(C)) :-
    !, union([A],B,C).
lub(A,B,C) :-
    \+ is_list(A), \+ is_list(B), \+ is_list(C),
    A =.. [Type|ArgsA],
    B =.. [Type|ArgsB],
    lub_list(ArgsA,ArgsB,ArgsC), !,
    C =.. [Type|ArgsC].
lub(A,B,top([A,B])) :- !.

lub_list([],[],[]).
lub_list([H1|T1],[H2|T2],[H3|T3]) :-
    lub(H1,H2,H3), !,
    lub_list(T1,T2,T3).

lub_domain_level(U1,U2,U3) :-
    check_factor(U1,U2,0,0),
    lub_domain_level2(U1,U2,U3).

lub_domain_level2([],[],[]) :- !.
lub_domain_level2([[Factor1,Unit1,Pow1]|Units], UnitsOp2,[[Factor1,Unit1,Pow1]|ResUnits]) :-
    selectchk([_Factor1,Unit1,Pow1],UnitsOp2,UnitsOp22), !,
    lub_domain_level2(Units,UnitsOp22,ResUnits).

check_factor([],[],X,X).
check_factor([[Factor1,_,_]|T1], [[Factor2,_,_]|T2],X,Y) :-
    !, X2 is X + Factor1, Y2 is Y + Factor2, check_factor(T1,T2,X2,Y2).

