% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(units_prettyprint, [
                           units_pp_trans/2,
                           units_pp_prop/2,
    		           units_pp_value/2,
                           unit_from_internal/2,
                           pp/3,
			   pp_fixpoint_result/3
                        ]).

:- use_module(library(codesio)).
:- use_module(library(lists)).

:- use_module(probsrc(tools)).
:- use_module(probsrc(error_manager)).

:- use_module(units_alias).
:- use_module(units_tools).
                        
:- use_module(probsrc(module_information)).
:- module_info(group,plugin_units).
:- module_info(description,'Units Plugin: Prettyprinter for properties and transitions').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').
                        
units_pp_trans(Transition,Pretty) :-
    %write(units_pp_trans(Transition,Pretty)),
    pp(classic,Transition,PrettyTerm),
    write_to_codes(PrettyTerm,Pretty).

units_pp_prop(Property,Pretty) :-
    pp(classic,Property,PrettyTerm),
    write_to_codes(PrettyTerm,Pretty).

units_pp_value(Value,Pretty) :-
    pp(classic,Value,PrettyTerm),
    write_to_codes(PrettyTerm,Pretty).


pp(_ClassicOrEventB,X,unknown) :- var(X), !.
pp(ClassicOrEventB,integer(X),Y) :- !, pp(ClassicOrEventB,X,Y).
pp(ClassicOrEventB,set(X),Y) :- !, pp(ClassicOrEventB,X,Y).
pp(ClassicOrEventB,couple(X,Y),Pretty) :- !,
    pp(ClassicOrEventB,X,PPX), pp(ClassicOrEventB,Y,PPY),
    ajoin([PPX, ' -> ', PPY],Pretty).
pp(ClassicOrEventB,global(X,_Y),Pretty) :- !,
    pp(ClassicOrEventB,X,Pretty).
% in the tcltk interface, the error is shown in a different way
% thus, we do not need to include it in the prettyprint
pp(classic,X,error) :- X=top(_), !.
% in rodin, we show an error listing all infered units
pp(eventb,X,PPOut) :-
    X=top(List), !,
    maplist(pp_unit(eventb),List,PPList),
    ajoin([multiple, '(', PPList, ')'], PPOut).
pp(_ClassicOrEventB,'$setup_constants','SETUP_CONSTANTS') :- !.
pp(_ClassicOrEventB,'$initialise_machine','INITIALISATION') :- !.
pp(_ClassicOrEventB,X,X) :- atom(X), !.
pp(_ClassicOrEventB,X,X) :- number(X), !.
pp(ClassicOrEventB,'-->'(Operation,ResultValues),PrettyTransition) :- !,
    pp(ClassicOrEventB,Operation,PrettyOperation),
    pp(ClassicOrEventB,ResultValues,PrettyResults),
    ajoin([PrettyOperation, '-->', PrettyResults],PrettyTransition).
pp(ClassicOrEventB,L,Result) :-
    is_unit_list(L),
    pp_unit(ClassicOrEventB,L,Result),!.
pp(_ClassicOrEventB,[],[]) :- !.
pp(ClassicOrEventB,[H|T],[PH|PT]) :- !,
    pp(ClassicOrEventB,H,PH),
    pp(ClassicOrEventB,T,PT).
pp(ClassicOrEventB,Compound,Pretty) :-
    Compound =.. [Head|Args], !,
    pp(ClassicOrEventB,Head,PrettyHead),
    pp(ClassicOrEventB,Args,PrettyArgs),
    Pretty =.. [PrettyHead|PrettyArgs].
pp(_ClassicOrEventB,Term,_Pretty) :-
    !, add_error_fail(units_prettyprint, 'Term not implemented: ', [Term]).

pp_unit(ClassicOrEventB,L,Result) :-
    subseq(L,Com,Sub),
    unit_alias(Aliased,Sub), !,
    (Com = []
    -> Result = Aliased
    ;  pp_unit(ClassicOrEventB,Com,ComPP),
       ajoin_with_sep([Aliased,ComPP],' * ',Result)).
pp_unit(ClassicOrEventB,[[EtoT,Symbol,EtoS]|T],Result) :-
    exponentiation_operator(ClassicOrEventB,Exponentiation),
    (EtoT = 0
    -> EtoTPP = ''
    ;  (EtoT = 1 -> EtoTPP = '10*' ; ajoin([10,Exponentiation,EtoT,'*'],EtoTPP))),
    (EtoS = 1 -> EtoSPP = Symbol ; ajoin([Symbol,Exponentiation,EtoS],EtoSPP)),
    ajoin([EtoTPP,EtoSPP],FirstUnitPP),
    (T = []
     -> Result = FirstUnitPP
     ;  pp_unit(ClassicOrEventB,T,TPP),
        ajoin_with_sep([FirstUnitPP,TPP],' * ',Result)).

exponentiation_operator(classic,'**').
exponentiation_operator(eventb,'^').


% predicates below should only be called from tcltk
pp_fixpoint_result(FixPointState,UnboundInConstraint,PP) :-
    pp_state(FixPointState,PPFixPointState),
    evaluate_fixpoint(FixPointState,UnboundInConstraint,Part2),
    units:iterations(Iterations),
    evaluate_operation(OperationsResult),
    ajoin(['Fixpoint reached after ', Iterations,' Iterations\n\nState:\n',PPFixPointState,'\n\n',Part2,'\n',OperationsResult],PP).

pp_state([],'').
pp_state([bind(Key,Val)|T],Res) :-
    pp_state(T,PPT),
    pp(classic,Val,PPVal),
    ajoin([Key, ' = ', PPVal, '\n', PPT],Res).

evaluate_operation(OperationsResult) :-
    findall(Name, units_interpreter:top_in_parameters(Name), ListOfNames1),
    findall(Name, units_interpreter:top_in_results(Name), ListOfNames2),
    (ListOfNames1 = []
     -> OperationsResult1 = ''
     ; ajoin(['Multiple units inferred for parameters of: ', ListOfNames1, '\n\n'], OperationsResult1)
    ),
    (ListOfNames2 = []
     -> OperationsResult = OperationsResult1
     ; ajoin([OperationsResult1,'Multiple units inferred for results of: ', ListOfNames2], OperationsResult)
    ).

evaluate_fixpoint([],_Unbound,'').
evaluate_fixpoint([bind(Id,Val)|State],Unbound,AtomOut) :-
    \+(ground(Val)), member(Id,Unbound), !,
    ajoin(['No unit inferred for Variable / Constant, but used in constraint: ',Id,'\n'],Out1),
    evaluate_fixpoint(State,Unbound,SubOut),
    ajoin([Out1,SubOut],AtomOut).
evaluate_fixpoint([bind(Id,Val)|State],Unbound,AtomOut) :-
    \+(ground(Val)), !,
    ajoin(['No unit inferred for Variable / Constant: ',Id,'\n'],Out1),
    evaluate_fixpoint(State,Unbound,SubOut),
    ajoin([Out1,SubOut],AtomOut).
evaluate_fixpoint([bind(Id,Val)|State],Unbound,AtomOut) :-
    contains_top(Val), !,
    ajoin(['Multiple units inferred for Variable / Constant: ',Id,'\n'],Out1),
    evaluate_fixpoint(State,Unbound,SubOut),
    ajoin([Out1,SubOut],AtomOut).
evaluate_fixpoint([bind(_Id,_Val)|State],Unbound,SubOut) :-
    evaluate_fixpoint(State,Unbound,SubOut).