% (c) 2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(ample_sets, [ compute_dependendency_relation_of_all_events_in_the_model/0,
		        compute_ample_set2/2,enabled_graph/1,
			dependent_act/2,stutter_action/1,visible_action/1,enabled_act/2,
			visited_state/1,dependency_enabling_computed/0,
			clear_dynamic_predicates_for_POR/0]).

:- use_module(library(ordsets)).
:- use_module(library(random)).
:- use_module(library(lists)).

:- use_module(probsrc(self_check)).
:- use_module(probsrc(error_manager)).

:- use_module(probsrc(state_space)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(specfile)).
:- use_module(probsrc(b_interpreter)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(b_state_model_check)).

:- use_module(module_information).

:- use_module(library(ugraphs)).

:- module_info(group,model_checker).
:- module_info(description,'This module contains predictes for calculating ample sets for partial order reduction.').
:- module_info(revision,'$Rev: 12605 $').
:- module_info(lastchanged,'$LastChangedDate: 2012-12-10 10:30:00 +0100 (Tue, 10 Dec 2012) $').

:- dynamic stutter_action/1,visible_action/1,dependent_act/2,enabled_act/2.
:- dynamic visited_state/1,dependency_enabling_computed/0,enabled_graph/1.

clear_dynamic_predicates_for_POR :- 
   retractall(dependent_act(_,_)),
   retractall(enabled_act(_,_)),
   retractall(stutter_action(_)),
   retractall(visible_action(_)),
   retractall(visited_state(_)),
   retractall(enabled_graph(_)),
   retractall(dependency_enabling_computed).

/* This implementation of the ample sets approach sastisfies all four conditions of the ample sets theory (not proved yet).
   The conditions are taken from the book "Principles of Model Checking" written by C.Baier and J.P.Katoen.

For every state s in the state space, where Act(s) stands for all enabled actions in state s, the following conditions 
must be satisfied:

1. Nonemptiness condition (A1) and terminal states
   (A 1.1) ample(s) <: Act(s) (set equality is possible)
   (A 1.2) ample(s) = {} <=> Act(s) = {}

2. Dependency condition
   (A 2) Let s -a_{1}> s_{1} -a_{2}> ... -a_{n}> s_{n} -a> s' be a finite execution
         fragment in the original state space. If a depends on ample(s), then a_{i}:ample(s)
	 for some 0<i<=n.

3. Stutter condition
   (A 3) If ample(s) /= Act(s) then any action t:ample(s) is a stutter action.

4. Cycle condition
   (A 4) For any cycle s_{0}s_{1}...s_{n} in the reduced state space and actions t:Act(s_{i}), for some 
         0 < i <= n, there exists j:{1..n} such that t:ample(s_{j}). (s_{0},...,s_{n} are states)

 */

/* compute all outgoing tranitions from the ample set calculated for the particular node */
:- use_module(specfile,[prepare_state_for_specfile_trans/2]).
compute_ample_set2(CurID,CurState) :- 
   user: reset_max_reached,
   specfile: prepare_state_for_specfile_trans(CurState,PreparedCurState),
   get_all_possible_action_names_from_current_state(PreparedCurState,Actions),
   get_action_names_only(Actions,ActionNames),
   assert(visited_state(CurID)),
   list_to_ord_set(ActionNames,AllActions),    % we want only unique action names for analyzing and computing of ample sets
   compute_ample_actions(AllActions,AmpleSet), % compute a sound ample set from all actions enabled in state
   %print(sets(allActions(AllActions),ampleSet(AmpleSet))),nl,
   ( ord_seteq(AmpleSet,AllActions) -> 
        /* if there are no reductions done in the particular state,
	   we then compute all transtions */
        %print('compute all transitions'),nl,
        retract(visited_state(CurID)),
	user: compute_all_transitions2(CurID,CurState)
   ; (/*\+ord_seteq(AmpleSet,AllActions),*/ AmpleSet \= []) -> 
	% compute just the actions from the ample set (reduction of the state space)
        %print('compute only ample set transitions'),nl,
	add_transitions(CurID,CurState,AmpleSet,Actions,_NewIDs)
   ; otherwise -> true /* DEADLOCK*/ % what should we do here?
   ),
   (user: max_reached -> user: assert_max_reached_for_node(CurID),MaxReached=true ; MaxReached=false),
   (CurID==root -> user: save_constants(MaxReached) ; user: add_new_transitions_to_cache(CurID)).

% computing a sound ample set of actions for the current state.
compute_ample_actions(AllActions,AmpleSet) :- 
   enabled_graph(EnGraph),
   get_ample_actions(AllActions,EnGraph,AmpleSet).

/************************** ADD TRANSITIONS (BEGIN) ******************************/

% Adding transitions form the particular ample set to the state space.
add_transitions(_CurID,_CurState,[],_Actions,[]).
add_transitions(CurID,CurState,[ActionName|ActNames],Actions,_NewIDs) :-
   add_action_name_transitions(CurID,CurState,ActionName,Actions),
   add_transitions(CurID,CurState,ActNames,Actions,_NewIDs).

% An event can have parameters, hence there can be multiple actions with the same ActionName.
% e.g. act1(1),act1(2),act1(3),...
% Therefore we have to add every single action with same action name as a transition to the state space.
add_action_name_transitions(_CurID,_CurState,_ActionName,[]).
add_action_name_transitions(CurID,CurState,ActionName,[(ActionName1,Act,NewExpression,TransInfo,Residue)|R]) :-
   (ActionName == ActionName1 -> 
	user: add_trans_id(CurID,Act,NewExpression,Residue,NewId,TransInfo,_)
   ; otherwise -> 
	true
   ),
   % if successor state already visited we have to explore all actions
   ((\+var(NewId),visited_state(NewId),(CurID >= NewId)) ->             % Case in which Condition 4 can be violited,
	user: compute_all_transitions2(CurID,CurState) % therefore we must explore all passible actions in the current state.
	%,print('********** CYCLE FOUND ---> COMPUTE ALL ACTIONS ****************'),nl
   ; otherwise -> 
	add_action_name_transitions(CurID,CurState,ActionName,R) % add only the actions from the computed ample set
   ).

/************************** ADD TRANSITIONS (END) ******************************/

/* Auxiliary predicates for getting action names enabled at the current state */ 

% getting action names from the action tuples
get_action_names_only([],[]).
get_action_names_only([(ActionName,_A,_NewE,_TrInf,_Residue)|R],[ActionName|Res]) :-
   get_action_names_only(R,Res).

% Getting all possible actions from the current state (check performance)
get_all_possible_action_names_from_current_state(CurState,AllActions) :- 
   findall((ActionName,Act,NewExpression,TransInf,Residue),
	(specfile: specfile_possible_trans_name(CurState,ActionName),
	 specfile: specfile_trans(CurState,ActionName,Act,NewExpression,TransInf,Residue)),
		AllActions).
/* Auxiliary predicates for getting action names enabled at the curent state */ 


cputime(T) :- statistics(runtime,[T,_]).

/************************** DEPENDENCY RELATION (BEGIN) ******************************/
    
/* predicate which determines the dependency graph of all actions in the model,
   it should be called only once (at the beginning of the model checking) */

compute_dependendency_relation_of_all_events_in_the_model :- 
   %print('***********'),print(compute_dependendency_relation_of_all_events_in_the_model),print('******************'),nl,
   findall(Action, (b_interpreter: b_get_machine_operation(Action,_,_,_)),AllActions),
   %retractall(dependency_graph(_)),
   retractall(enabled_graph(_)),
   print('**********DETERMINE ACTION DEPENDENCIES*************'),nl,
   cputime(T1),
   determine_dependent_actions(AllActions,_DepGraph),
   cputime(T2),
   print('%% Dependencies Checking Time: '), D is T2-T1, print(D),print('ms.'),nl,
   print('**********DETERMINE ACTION ENABLING*************'),nl,
   cputime(T3),
   determine_enabled_actions(AllActions,EnGraph),
   cputime(T4),
   print('%% Enabling Checking Time: '), D1 is T4-T3, print(D1),print('ms.'),nl,
   print('**********FINISHED ACTION ANALYSES*************'),nl,
   %print('EnGraph: '),print(EnGraph),nl,
   assert(dependency_enabling_computed),
   assert(enabled_graph(EnGraph)).

determine_dependent_actions(AllActions,DepGraph) :- 
   vertices_edges_to_ugraph([],[],Graph),
   determine_dependent_actions_graph(AllActions,AllActions,Graph,DepGraph).

/* auxiliary predicates (determine_dependent_actions_graph/4 and determine_dependency_graph_wrt_one_action/4) 
   for creating the dependency graph */
determine_dependent_actions_graph([],_AllActions,Graph,Graph).
determine_dependent_actions_graph([Action|Actions],AllActions,Graph,DepGraph) :- 
   determine_dependency_graph_wrt_one_action(Action,AllActions,Graph,ResGraph),
   determine_dependent_actions_graph(Actions,AllActions,ResGraph,DepGraph).

determine_dependency_graph_wrt_one_action(_Act,[],Graph,Graph).
determine_dependency_graph_wrt_one_action(Act1,[Act2|Actions],Graph,ResGraph) :- 
   ( dependent_act(Act1,Act2) -> 
	Graph2=Graph % edges and vertices should be already added to the graph
   ; dependent_actions(Act1,Act2) -> 
	add_vertices(Graph,[Act1,Act2],Graph1),
	add_edges(Graph1,[Act1-Act2,Act2-Act1],Graph2), % dependency relation is equivalence relation (reflexive, symmetric and transitiv)
	assert(dependent_act(Act1,Act2)), 
	assert(dependent_act(Act2,Act1))
   ; otherwise -> 
	Graph2=Graph
   ),determine_dependency_graph_wrt_one_action(Act1,Actions,Graph2,ResGraph).

determine_enabled_actions(AllActions,EnGraph) :- 
   vertices_edges_to_ugraph([],[],Graph),
   determine_enabled_actions_graph(AllActions,AllActions,Graph,EnGraph).
   
determine_enabled_actions_graph([],_AllActions,Graph,Graph).
determine_enabled_actions_graph([Action|Actions],AllActions,Graph,EnGraph) :- 
   determine_enabled_graph_wrt_one_action(Action,AllActions,Graph,ResGraph),
   determine_enabled_actions_graph(Actions,AllActions,ResGraph,EnGraph).

determine_enabled_graph_wrt_one_action(_Act,[],Graph,Graph).
determine_enabled_graph_wrt_one_action(Act1,[Act2|Actions],Graph,ResGraph) :- 
   debug_println(9,check(Act1,Act2)),
   ( enabled_act(Act1,Act2) -> 
	Graph2=Graph % edges and vertices should be already added to the graph
   ; enabled_actions(Act1,Act2) -> 
	add_vertices(Graph,[Act1,Act2],Graph1),
	add_edges(Graph1,[Act1-Act2],Graph2), 
	assert(enabled_act(Act1,Act2))
   ; otherwise -> 
	Graph2=Graph
   ),determine_enabled_graph_wrt_one_action(Act1,Actions,Graph2,ResGraph).

enabled_actions(Act1,Act2) :- 
   enabling_analysis: enable_analysis(Act1,Act2,Enable),
   %print(enable_analysis(Act1,Act2,Enable)),nl,
   ( (Enable == guaranteed)
   ; (Enable == timeout)  % pessimistic approach
   ; (Enable == possible)
   ; (Enable == possible_enable)
   ).

/* 
   dependent_actions/2 determine statically if two actions are dependent
   two events/actions are dependent in the following three cases:
	1. If both events modify at least one common variable.
	2. Act1 modifies a variable in the guard of Act2. (Act1 can enable or disable Act2)
	3. Act2 modifies a variable in the guard of Act1. (Act2 can enable or disable Act1)
*/

%%% Act1 and Act2 are enabled in current state State,
%%% we don't have to check if both actions are enabled in 
%%% some state
dependent_actions(Act1,Act2) :- 
	enabling_analysis: dependent_actions(Act1,Act2,Res),
	((Res == '-') -> 
		dependent_actions(Act2,Act1)
	; (Res == static_independent) -> 
		fail
	; (Res == independent) ->
		fail
	; otherwise ->
		true
	).

/************************** DEPENDENCY RELATION (END) ******************************/


/************************** DETERMINING POSSIBLE AMPLE SETS (BEGIN) ******************************/
/************************** PREDICATES COMPUTING AMPLE SETS WITH RESPECT TO CONDITION (A2) *************************/

get_sound_ample_set_if_possible([],AllActions,AllActions).
get_sound_ample_set_if_possible(AmpleSets,AllActions,AmpleActs) :- 
   random_member(Set,AmpleSets),
   delete(AmpleSets,Set,AmpleSets1),
   (ord_seteq(Set,AllActions) -> AmpleActs=AllActions % if all actions enabled in s are added to the ample set, we don't have to check condition (A3)
   ; otherwise ->  % condition (A3): if ample(s) \= Actions(s) then any t:ample(s) is a stutter action
	( stutter_actions(Set) -> AmpleActs=Set
	; get_sound_ample_set_if_possible(AmpleSets1,AllActions,AmpleActs)
	)
   ).

/* 

   implementation of the ample-sets approach satisfying the dependency condition (A2) 
   by using enabling analysis (see module enabling analysis) 

*/

get_ample_actions(AllActions,EnGraph,AmpleActs) :- 
   /* compute possible ample sets by adding all interdependent actions in separate sets 
      and actions that can enable dependent actions at some point */
   get_possible_ample_action_sets(AllActions,EnGraph,AmpleSets),
   /* check if assummed ample set contains only stutter actions if state not fully expanded */
   get_sound_ample_set_if_possible(AmpleSets,AllActions,AmpleActs).

get_possible_ample_action_sets([],_EnGraph,[]).
get_possible_ample_action_sets([Act|Actions],EnGraph,[DepClass|AmpleSets]) :- 
   add_dependent_class(Act,Actions,EnGraph,DepClass,RemActions),
   get_possible_ample_action_sets(RemActions,EnGraph,AmpleSets).

add_dependent_class(Act,Actions,EnGraph,Res,RemActions) :-
   get_dependent_actions_from_actions_list(Act,[Act|Actions],EnGraph,DepActions),
   ord_subtract(Actions,DepActions,RemActions),
   ord_add_element(DepActions,Act,Res).


get_dependent_actions_from_actions_list(Act,Actions,EnabledGraph,AmpleSet) :- 
   setof(Act1,
	(member(Act1,Actions),(dependent_act(Act1,Act);Act1==Act)),
		DepActions), % first, all dependent actions should be in the ample set
   ord_subtract(Actions,DepActions,IndActions), % determine all independent actions to DepActions
   /* add all independent actions from IndActions that either can an action, which depends on DepActions,  
      or path of actions that can enable an action, which depends on DepActions */
   (setof(Act2, 
	(member(Act2,IndActions),can_enable_dep_action(Act2,DepActions,EnabledGraph,_L)/*,print_path(L)*/),
		EnabledDep) -> true
   ; otherwise -> EnabledDep = []),
   ord_union(DepActions,EnabledDep,AmpleSet).

/* used for debugging only
print_path(P) :- 
   print('******************'),print(path(P)),print('***********************'),nl.
*/

/* searching for enabling path that enables action dependents on AmpleSet */
can_enable_dep_action(Act,AmpleSet,EnabledGraph,P) :- 
   vertices(EnabledGraph,Vertices),
   member(Act1,Vertices),
   Act1 \= Act,
   min_path(Act,Act1,EnabledGraph,P,_L),
   \+member(Act1,AmpleSet),
   depends_on_ample_set(Act1,AmpleSet),!.

depends_on_ample_set(Act,AmpleSet) :- 
   member(Act1,AmpleSet),
   Act\=Act1,
   dependent_act(Act,Act1).

/* Determining ample sets by assuming that the dependency relation ist transitive (strong condition) */
/*

can_enable_dep_action1(Act,AmpleSet,EnabledGraph,P) :- 
   member(Act1,AmpleSet),
   dependent_act(Act1,Act2),
   Act1 \= Act2,
   \+ member(Act2,AmpleSet),
   min_path(Act,Act2,EnabledGraph,P,_L),!.

get_ample_actions(AllActions,EnGraph,AmpleActs) :- 
   get_possible_ample_action_sets_graph(AllActions,EnGraph,AmpleSets),
   get_sound_ample_set_if_possible(AmpleSets,AllActions,AmpleActs).

get_possible_ample_action_sets_graph(AllActions,Graph,AmpleSetsRes) :- 
   get_possible_ample_action_sets(AllActions,Graph,AmpleSetsRes),
   print(ample_sets_res(AmpleSetsRes)),nl.

get_possible_ample_action_sets([],_Graph,[]).
get_possible_ample_action_sets([Act|Actions],Graph,[DepClass|AmpleSets]) :- 
   add_dependent_class(Act,Actions,Graph,DepClass,RemActions),
   get_possible_ample_action_sets(RemActions,Graph,AmpleSets).

add_dependent_class(Act,Actions,Graph,Res,RemActions) :-
   get_dependent_actions_from_actions_list(Act,Actions,Graph,DepActions),
   ord_subtract(Actions,DepActions,RemActions),
   ord_add_element(DepActions,Act,Res).

get_dependent_actions_from_actions_list(Act,Actions,Graph,DepActions) :- 
   findall(Act1,
       (member(Act1,Actions),min_path(Act,Act1,Graph,P,_L),print(path(Act,Act1,P)),nl),
	     DepActions).
*/

/************************** DETERMINING POSSIBLE AMPLE SETS (END) ******************************/

/************************** CHECKING IF ALL ACTIONS FROM AMPLE SET ARE STUTTER ACTIONS (BEGIN) ******************************/
/************************** PREDICATES WHICH SATISFY CONDITION (A3) ******************************/

stutter_actions([]).
stutter_actions([Act|Acts]) :- 
   is_stutter_action(Act),
   stutter_actions(Acts).

% visible actions are non stutter actions
% how do we want to name the non stutter actions? visible?
is_stutter_action(Act) :- 
   ( stutter_action(Act) -> /*print(stutter_action(Act)),nl,*/true
   ; visible_action(Act) -> /*print(visible_action(Act)),nl,*/fail
   ; otherwise -> check_if_action_is_stutter_wrt_inv(Act)
   ).

:- use_module(debug).
check_if_action_is_stutter_wrt_inv(Act) :- 
    bmachine:b_operation_preserves_full_invariant(Act),!,
    assert(stutter_action(Act)), debug_println(9,stutter_according_to_proof_info(Act)).
check_if_action_is_stutter_wrt_inv(Act) :- 
   b_get_invariant_from_machine(Invariant),
   conjunction_to_list(Invariant,Conjunctions), % getting all conjunctions from the invariant
   enabling_analysis: b_get_read_write(Act,_Reads,Writes),
   ( action_does_not_modify_important_variables_in_the_inv(Writes,Conjunctions) 
	-> assert(stutter_action(Act)), debug_println(9,stutter_according_to_read_write_info(Act))
   ; otherwise 
	-> assert(visible_action(Act)),fail
   ).

action_does_not_modify_important_variables_in_the_inv([],_Conjuncions).
action_does_not_modify_important_variables_in_the_inv([Var|WrVars],Conjunctions) :- 
   var_not_changing_visible_condition_in_the_inv(Var,Conjunctions),
   action_does_not_modify_important_variables_in_the_inv(WrVars,Conjunctions).

var_not_changing_visible_condition_in_the_inv(_Var,[]).
var_not_changing_visible_condition_in_the_inv(Var,[Conj|Conjs]) :- 
   ( occurs_in_expr(Var,Conj) 
	-> (\+ trivial_predicate(Var,Conj) -> fail; true)
   ; otherwise -> true
   ),
   var_not_changing_visible_condition_in_the_inv(Var,Conjs).

trivial_predicate(Var,Conj) :- 
   \+var(Conj),
   trivial_member_predicate(Var,Conj).

/* TODO: Define formal when an action in Event-B/B can be seen as a stutter action.
   Maybe we should use here the predicate b_specialized_invariant_for_op/2 from bmachine.pl.
    */

/*trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('NATURAL'),set(integer),_)),pred,_).
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('NAT'),set(integer),_)),pred,_).*/
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('INT'),set(integer),_)),pred,_). 
trivial_member_predicate(Var,Conj) :-
   Conj=b(member(b(identifier(Var),integer,_),b(integer_set('INTEGER'),set(integer),_)),pred,_).

/************************** CHECKING IF ALL ACTIONS FROM AMPLE SET ARE STUTTER ACTIONS (END) ******************************/

