% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

:- module(b_ast_cleanup, [clean_up/3, clean_up_l/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module implements transformations/simplifications on the AST.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(tools, [safe_atom_chars/3,exact_member/2]).
:- use_module(error_manager).
:- use_module(self_check).
:- use_module(bsyntaxtree).
:- use_module(btypechecker, [unify_types_strict/2]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% clean up some code afterwards
clean_up(Expr,NonGroundExceptions,CExpr) :-
	get_texpr_type(Expr,Type1),
	ground_type(Type1,NonGroundExceptions),
    cleanups(pre,Expr,[],TPExpr),
    remove_bt(TPExpr,PExpr,LExpr,TLExpr),
	syntaxtransformation(PExpr,Subs,_,NewSubs,LExpr),
	clean_up_l(Subs,NonGroundExceptions,NewSubs),
	cleanups(post,TLExpr,[],CExpr).

% apply the clean-up rules to an expression until all
% applicable rules are processed
% cleanups(Phase,Expr,AppliedRules,Result):
%   Phase: pre or post
%   Expr:  the expression to clean up
%   AppliedRules: a list of clean up rules that have been already applied
%                 and must only be apply once ("single" mode)
%   Result: the cleaned-up expression
cleanups(Phase,Expr,AppliedRules,Result) :-
    % set up co-routines that ensure that "Rule" is not applied if
    % is in the list AppliedRules
    assure_single_rules(AppliedRules,Rule),
    ( cleanup_phase(Phase,Expr,NExpr,Mode/Rule) ->       % try to apply a rule (matching the current phase)
        ( Mode==single -> AppRules = [Rule|AppliedRules] % if the rule is marked as "single", we add to the list of already applied rules
        ; Mode==multi  -> AppRules = AppliedRules        % if "multi", we do not add it to the list, the rule might be applied more than once
        ; otherwise    -> add_error_fail(b_ast_cleanup,'Unexpected rule mode ',Mode)),
        cleanups(Phase,NExpr,AppRules,Result)            % continue recursively with the new expression
    ; otherwise ->              % if no rule matches anymore,
        Result = Expr,          %   we leave the expression unmodified
        Rule = none).           %   and unblock the co-routine (see assure_single_rules/2)

assure_single_rules([],_Rule) :- !.
assure_single_rules(AppliedRules,Rule) :-
        assure_single_rules2(AppliedRules,Rule).
:- block assure_single_rules2(?,-).
assure_single_rules2(_AppliedRules,none) :- !.
assure_single_rules2(AppliedRules,Rule) :-
        \+ member(Rule, AppliedRules).

cleanup_phase(Phase,OTExpr,NTExpr,Mode/Rule) :-
        create_texpr(OExpr,OType,OInfo,OTExpr),
        create_texpr(NExpr,NType,NInfo,NTExpr),
        cleanup_phase2(Phase,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule).
cleanup_phase2(pre,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule) :-
        cleanup_pre(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule).
cleanup_phase2(post,OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule) :-
        cleanup_post(OExpr,OType,OInfo,NExpr,NType,NInfo,Mode/Rule).

clean_up_l([],_,[]).
clean_up_l([Expr|Rest],NonGroundExceptions,[CExpr|CRest]) :-
	clean_up(Expr,NonGroundExceptions,CExpr),
	clean_up_l(Rest,NonGroundExceptions,CRest).

:- use_module(typing_tools,[is_infinite_type/1]).
:- use_module(specfile,[animation_minor_mode/1]).
% cleanup_pre(OldExpr,OldType,OldInfo,NewExpr,NewType,NewInfo,Mode/Rule)
% replace finite by truth
cleanup_pre(finite(S),pred,I,truth,pred,[was(finite)|I],multi/remove_finite) :- 
    get_texpr_type(S,Type), 
    \+ is_infinite_type(Type),!.
    %!,(is_infinite_type(Type) -> print(keep_finite(Type)),nl,fail ; print(removed_finite(Type)),nl).
%% implement fnc by rewriting using fnc(r) == %x.(x:dom(r)|r[{x}])
cleanup_pre(trans_function(Rel),Type,Info,NewExpr,Type,[was(trans_function)|Info],multi/trans_function) :-
    Type = set(couple(T1,set(T2))),
    %print(trans_function(T1,T2,Rel)),nl,
    NewExpr = lambda([IDx],MemberPred,ImageExpr),
    create_texpr(domain(Rel),set(T1),[],RangeSet),
    create_texpr(identifier('_zzzz_unary'),T1,[generated],IDx),
    create_texpr(member(IDx,RangeSet),pred,[],MemberPred),
    create_texpr(image(Rel,SingletonX),set(T2),[],ImageExpr), % r[{x}]
    create_texpr(set_extension([IDx]),set(T1),[],SingletonX). % {x}
    %print('Rewriting fnc: '),translate:print_bexpr(NewExpr),nl.
%% implement rel by rewriting using rel(r) == {x,y| x:dom(r) & y: union(r[{x}])}
cleanup_pre(trans_relation(Rel),Type,Info,NewExpr,Type,[was(trans_relation)|Info],multi/trans_relation) :-
    Type = set(couple(T1,T2)),
    %print(trans_relation(T1,T2,Rel)),nl,
    NewExpr = comprehension_set([IDx,IDy],MemberPred),
    create_texpr(conjunct(MemberPredX,MemberPredY),pred,[generated],MemberPred),
    create_texpr(identifier('_zzzz_unary'),T1,[],IDx), % was [generated]
    create_texpr(identifier('_zzzz_binary'),T2,[],IDy), % was [generated]
    create_texpr(domain(Rel),set(T1),[],RangeSet),
    create_texpr(member(IDx,RangeSet),pred,[],MemberPredX), % x:dom(r)
    create_texpr(set_extension([IDx]),set(T1),[],SingletonX), % {x}
    create_texpr(image(Rel,SingletonX),set(set(T2)),[],ImageExpr), % r[{x}]
    create_texpr(general_union(ImageExpr),set(T2),[],UnionSet), % union(r[{x}])
    create_texpr(member(IDy,UnionSet),pred,[],MemberPredY). % y: union(r[{x}])
    %print('Rewriting rel: '),translate:print_bexpr(NewExpr),nl.
% exchange quantified union by generalized union
cleanup_pre(quantified_union(Ids,Pred,Expr),Type,I,general_union(Set),Type,I,multi/quant_union) :-
	!,quantified_set_op(Ids,Pred,Expr,quantified_union,Type,Set).
% exchange quantified intersection by generalized intersection
cleanup_pre(quantified_intersection(Ids,Pred,Expr),Type,I,general_intersection(Set),Type,I,multi/quant_inter) :-
	!,quantified_set_op(Ids,Pred,Expr,quantified_intersection,Type,Set).
% In reply to PROB-240: Check if arguments of Prj1/2 are types only using is_just_type:
cleanup_pre(function(TProjection,Argument),Type,I,Result,Type,I,multi/projection_call) :-
	get_texpr_expr(TProjection,Projection),
	( Projection=first_projection(A,B) ->
	    check_is_just_type(B),
	    Result = first_of_pair(Argument)
	; Projection=second_projection(A,B) ->
	    check_is_just_type(B),
	    Result = second_of_pair(Argument)
	; Projection=event_b_first_projection(A) ->
	    Result = first_of_pair(Argument)
	; Projection=event_b_second_projection(A) ->
	    Result = second_of_pair(Argument)),
	check_is_just_type(A),
	!.
cleanup_pre(function(Lambda,Argument),Type,I,assertion_expression(Cond,ErrMsg,Expr),Type,I,multi/lambda_guard) :-
	get_texpr_expr(Lambda,lambda([TId],TPre,TVal)),!,
	get_texpr_id(TId,Id),
	ErrMsg = 'function called outside of domain: ',
	~~mnf( replace_id_by_expr(TPre,Id,Argument,Cond) ),
	~~mnf( replace_id_by_expr(TVal,Id,Argument,Expr) ).
	%print(assertion(expression)),nl,translate:print_bexpr(Cond),nl, translate:print_bexpr(Expr),nl.
cleanup_pre(first_projection(A,B),Type,I,Set,Type,[was(prj1),prob_annotation('SYMBOLIC')|I],multi/first_projection) :-
	!,create_projection_set(A,B,first,Set).
cleanup_pre(second_projection(A,B),Type,I,Set,Type,[was(prj2),prob_annotation('SYMBOLIC')|I],multi/second_projection) :-
	!,create_projection_set(A,B,second,Set).
cleanup_pre(event_b_first_projection(Rel),Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev_first_projection) :-
	!,create_event_b_projection_set(Rel,first,Set).
cleanup_pre(event_b_second_projection(Rel),Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev_second_projection) :-
	!,create_event_b_projection_set(Rel,second,Set).
cleanup_pre(event_b_first_projection_v2,Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev2_first_projection) :-
	!,create_event_b_projection_set_v2(Type,first,Set). % TO DO: Daniel, does this work ???
cleanup_pre(event_b_second_projection_v2,Type,I,Set,Type,[prob_annotation('SYMBOLIC')|I],multi/ev2_second_projection) :-
	!,create_event_b_projection_set_v2(Type,second,Set). % TO DO: Daniel, does this work ???
cleanup_pre(function(Fun,Arg),integer,I,ArithOp,integer,I,multi/succ_pred_optimisation) :-
	get_texpr_expr(Fun,PS),
	( PS=predecessor -> Op=minus
	; PS=successor   -> Op=add),
	ArithOp =.. [Op,Arg,Integer],
	create_texpr(integer(1),integer,[],Integer).
cleanup_pre(event_b_comprehension_set(IDs,CoupleExpr,Pred),T,I,NewExpr,T,
	    [was(event_b_comprehension_set)|I],multi/ev_compset_lambda_single_id) :- 
	    % detect lambda expressions in classical B style
	    nested_couple_to_list(CoupleExpr,List),
	    check_ids(IDs,List,Expr),!,
	    %print(rewriting_event_b1(IDs,Pred,Expr)),nl,
	    NewExpr = lambda(IDs,Pred,Expr).
cleanup_pre(event_b_comprehension_set(IDList,CoupleExpr,Pred),T,I,comprehension_set(IDList,NewPred),T,
	    [was(event_b_comprehension_set)|I],multi/ev_compset_ids) :-
	     % Event_B_Comprehension with a several IDs which are also used as the couple expression
	     nested_couple_to_list(CoupleExpr,List), %print(List),nl,
	     List = IDList,
	     !,
	     NewPred=Pred.
/* cleanup_pre(event_b_comprehension_set([ID],ID,Pred),T,I,comprehension_set([Result],NewPred),T,
	    [was(event_b_comprehension_set)|I],multi/ev_compset_single_id) :-
	     % Event_B_Comprehension with a single ID which is also the expression
	     % TO DO: expand for multiple IDs
	     !,
	     Result = ID, NewPred=Pred. */
cleanup_pre(event_b_comprehension_set(Ids,Expr,Pred),T,I,comprehension_set([Result],NewPred),T,
	    [was(event_b_comprehension_set)|I],multi/ev_compset) :-
	unify_types_strict(T,set(Type)),
	% print(event_b_comprehension_set(Ids,Expr,Pred)),nl,
	(select(Expr,Ids,RestIds)
	  -> % the Expr is an identifier which is part of Ids: we can avoid complicated translation below
	   % example {f,n•n:INT & f:1..n-->Digit|f} --translated-> {f|#(n).(n:INT & f:1..n-->Digit)}
	  % print(remove(Expr,RestIds)),nl,
	     ExPred=Pred, Result=Expr
	  ; get_unique_id('_comp_result',ResultId),
	    create_texpr(identifier(ResultId),Type,[],Result),
	    create_texpr(equal(Result,Expr),pred,[],Equal),
	    conjunct_predicates([Equal,Pred],ExPred),
	    RestIds=Ids
	),
    create_exists(RestIds,ExPred,NewPred), print(done),nl, translate:print_bexpr(NewPred),nl.
cleanup_pre(domain_restriction(A,B),T,I,identity(A),T,I,multi/event_b_to_normal_identity1) :-
  /* translate S <| id into id(S) */
   B = b(event_b_identity,_,_).
cleanup_pre(range_restriction(B,A),T,I,identity(A),T,I,multi/event_b_to_normal_identity2) :-
  /* translate id |> S into id(S) */
   B = b(event_b_identity,_,_).
cleanup_pre(ring(A,B),T,I,composition(B,A),T,I,multi/ring_composition).
cleanup_pre(minus_or_set_subtract(A,B),integer,I,minus(A,B),integer,I,multi/remove_ambiguous_minus_int).
cleanup_pre(minus_or_set_subtract(A,B),set(T),I,set_subtraction(A,B),set(T),I,multi/remove_abiguous_minus_set).
cleanup_pre(minus_or_set_subtract(A,B),seq(T),I,set_subtraction(A,B),seq(T),I,multi/remove_abiguous_minus_seq).
cleanup_pre(mult_or_cart(A,B),integer,I,multiplication(A,B),integer,I,multi/remove_ambiguous_times_int).
cleanup_pre(mult_or_cart(A,B),set(T),I,cartesian_product(A,B),set(T),I,multi/remove_ambiguous_times_set).
cleanup_pre(mult_or_cart(A,B),seq(T),I,cartesian_product(A,B),seq(T),I,multi/remove_ambiguous_times_seq).
cleanup_pre(E,T,Iin,E,T,Iout,single/remove_rodinpos) :-
    selectchk(nodeid(rodinpos(_,[],_)),Iin,Iout). % reomve rodinpos information with Name=[]
cleanup_pre(exists(AllIds,P),pred,I,let_predicate(LetIds,Exprs,NewP),pred,I,multi/exists_to_let) :-
    % rewrite predicates of the form #x.(x=E & P(x)) into (LET x==E IN P(x))
    % side condition for #(ids).(id=E & P(ids)): no identifiers of ids occur in E
    conjunction_to_list(P,Preds),
    % The ids are needed to build a "black list"
    get_texpr_ids(AllIds,UnsortedIds), sort(UnsortedIds,Ids),
    find_one_point_rules(AllIds,Preds,Ids,LetIds,Exprs,RestIds,NewPreds),
    % only succeed if we found at least one id which can be rewritten as let
    LetIds = [ID1|_],!,
    %print(generate_let(LetIds)),translate:print_bexpr(P),nl,
    (atomic(ID1) -> add_internal_error(cleanup_pre,unwrapped_let_identifier(ID1)) ; true),
    create_exists_opt(RestIds,NewPreds,NewP).
cleanup_pre(exists(AllIds,P),pred,I,disjunct(NewP1,NewP2),pred,I,multi/partition_exists) :-
    is_a_disjunct(P,D1,D2),
    %print('### Warning: body of exists is a disjunction: '), translate:print_bexpr(P),nl,
    translate:translate_bexpression(P,PS),
    (animation_minor_mode(eventb) -> true
      ; add_warning(b_ast_cleanup,'Warning: body of existential quantifier is a disjunction (not allowed by Atelier-B): ',PS)),
    /* note that even if D2 is only well-defined in case D1 is false; it is ok to seperate this out
       into two existential quantifiers:  #x.(x=0 or 1/x=10) is ok to transform into #x.(x=0) or #x.(1/x=10) */
    create_exists_opt(AllIds,[D1],NewP1), % print('P1: '),translate:print_bexpr(NewP1),nl,
    create_exists_opt(AllIds,[D2],NewP2). %, print('P2: '),translate:print_bexpr(NewP1),nl.

:- use_module(preferences,[get_preference/2]).
check_is_just_type(_A) :- preferences:get_preference(ignore_prj_types,true),!.
check_is_just_type(A) :- (is_just_type(A) -> true ; debug_println(9,not_type_for_prj(A)),fail).

is_empty_set(b(empty_set,_,_)).
is_empty_set(b(empty_sequence,_,_)).
% create a lambda expression for a projection
create_projection_set(A,B,_Switch,Res) :- 
    (is_empty_set(A) ; is_empty_set(B)),!,
    Res = empty_set.
create_projection_set(A,B,Switch,comprehension_set(Ids,SPred)) :-
	Ids = [TArg1,TArg2,TArg3],
	( Switch==first  -> Expr = TArg1, Type1 = Type3
	; Switch==second -> Expr = TArg2, Type2 = Type3),
	get_texpr_type(A,TA1), unify_types_strict(TA1,set(Type1)), 
	get_texpr_type(B,TB2), unify_types_strict(TB2,set(Type2)), 
	get_unique_id(prj_arg__,Arg1), get_unique_id(prj_arg__,Arg2),
	((\+ occurs_in_expr('_lambda_result_',A),
	  \+ occurs_in_expr('_lambda_result_',B))
	   -> Arg3 = '_lambda_result_'
	   ;  get_unique_id(prj_arg__,Arg3)
	),
	create_texpr(identifier(Arg1),Type1,[generated(Switch)],TArg1),
	create_texpr(identifier(Arg2),Type2,[generated(Switch)],TArg2),
	create_texpr(identifier(Arg3),Type3,[generated(Switch)],TArg3),
	create_texpr(member(TArg1,A),pred,[],MembA),
	create_texpr(member(TArg2,B),pred,[],MembB),
	create_texpr(equal(TArg3,Expr),pred,[],Equal),
	conjunct_predicates([MembA,MembB,Equal],Pred), SPred=Pred.
%	b_interpreter:mark_bexpr_as_symbolic(Pred,SPred). % TO DO: put mark code into another module; maybe only mark as symbolic if types large enough ??
create_event_b_projection_set(Rel,Switch,comprehension_set(Ids,SPred)) :-
	Ids = [TArg1,TArg2,TArg3],
	( Switch==first  -> Expr = TArg1, Type1 = Type3
	; Switch==second -> Expr = TArg2, Type2 = Type3),
	get_texpr_type(Rel,RT),unify_types_strict(RT,set(couple(Type1,Type2))),
	get_unique_id(prj_arg__,Arg1), get_unique_id(prj_arg__,Arg2),
	((\+ occurs_in_expr('_lambda_result_',Rel))
	   -> Arg3 = '_lambda_result_'
	   ;  get_unique_id(prj_arg__,Arg3)
	),
	create_texpr(identifier(Arg1),Type1,[generated(Switch)],TArg1),
	create_texpr(identifier(Arg2),Type2,[generated(Switch)],TArg2),
	create_texpr(identifier(Arg3),Type3,[generated(Switch)],TArg3),
	create_texpr(couple(TArg1,TArg2),couple(Type1,Type2),[],Couple),
	create_texpr(member(Couple,Rel),pred,[],Member),
	create_texpr(equal(TArg3,Expr),pred,[],Equal),
	conjunct_predicates([Member,Equal],Pred), SPred=Pred.
	%b_interpreter:mark_bexpr_as_symbolic(Pred,SPred).
	
create_event_b_projection_set_v2(RelType,Switch,comprehension_set(Ids,SPred)) :-
	Ids = [TArg1,TArg2,TArg3],
	( Switch==first  -> Expr = TArg1, Type1 = Type3
	; Switch==second -> Expr = TArg2, Type2 = Type3),
	unify_types_strict(RelType,set(couple(couple(Type1,Type2),T3))),
	(T3==Type3 -> true ; add_error(create_event_b_projection_set,'Unexpected return type: ',T3)),
	%print(types(Type1,Type2,Type3)),nl,
	get_unique_id(prj_arg__,Arg1), get_unique_id(prj_arg__,Arg2), 
	Arg3 = '_lambda_result_', % no way this can occur inside TArg1/TArg2 ; was get_unique_id(prj_arg__,Arg3),
	create_texpr(identifier(Arg1),Type1,[generated(Switch)],TArg1),
	create_texpr(identifier(Arg2),Type2,[generated(Switch)],TArg2),
	create_texpr(identifier(Arg3),Type3,[generated(Switch)],TArg3),
	create_texpr(equal(TArg3,Expr),pred,[],Equal),
	conjunct_predicates([Equal],Pred), SPred=Pred.
%	b_interpreter:mark_bexpr_as_symbolic(Pred,SPred).
	%,print(Pred),nl.
	

:- use_module(btypechecker,[couplise_list/2,prime_identifiers/2]).
% create a comprehension set for quantified union or intersection
quantified_set_op(Ids,Pred,Expr,Loc,set(Type),Set) :-
	Info = [generated(Loc)],
	create_texpr(range(Lambda),set(set(Type)),Info,Set),
	create_texpr(lambda(Ids,Pred,Expr),set(couple(ArgType,set(Type))),Info,Lambda),
	get_texpr_types(Ids,Types),couplise_list(Types,ArgType).
	
:- use_module(debug).
:- use_module(kernel_records,[normalise_record_type/2]).
cleanup_post(Expr,Type,I,Expr,Type,I2,single/remove_erroneous_info_field) :-
    \+ ground(I),!, functor(Expr,F,_N),
    add_internal_error('Information field not ground: ',I:F),
    I2=[].
cleanup_post(block(TS),subst,_,Subst,subst,Info,multi/remove_block) :-
	!,get_texpr_expr(TS,Subst),get_texpr_info(TS,Info).
cleanup_post(Expr,Type,I,Expr,NType,I,single/normalise_record_type) :-
	nonvar(Type),Type=record(Fields),!,
	normalise_record_type(record(Fields),NType).
cleanup_post(conjunct(b(truth,pred,_),b(B,pred,I)),pred,_,B,pred,I,multi/remove_truth_conj1) :- !.
cleanup_post(conjunct(b(A,pred,I),b(truth,pred,_)),pred,_,A,pred,I,multi/remove_truth_conj2) :- !.
% detect certain tautologies/inconsistencies
cleanup_post(member(_,B),pred,I,truth,pred,I,multi/remove_type_member) :- is_just_type(B),!.
cleanup_post(member(E,b(fin_subset(E2),_,_)),pred,I,finite(E),pred,I,multi/introduce_finite) :- 
    (same_texpr(E,E2); is_just_type(E2)),!. % print(introduce(finite(E))),nl.  
cleanup_post(subset(_,B),pred,I,truth,pred,I,multi/remove_type_subset) :- is_just_type(B),!.
cleanup_post(subset(A,B),pred,I,NewPred,pred,I,multi/replace_subset_by_element) :- 
     get_texpr_expr(A,set_extension(List)),!,
     % applying rule {x1,x2,...} <: B  <=>  x1:B & x2:B & ...
     maplist(gen_member_predicates(B),List,Conjuncts),
     conjunct_predicates(Conjuncts,TNewPred),
     % translate:print_bexpr(TNewPred),nl,
     get_texpr_expr(TNewPred,NewPred).
cleanup_post(subset(A,B),pred,I,NewPred,pred,[generated_conjunct|I],multi/replace_union_subset) :- 
     % mark conjunct as generated: used e.g. by flatten_conjunct in predicate_evaluator
     get_texpr_expr(A,union(_,_)),!,
     % applying rule A1 \/ A2 <: B  <=>  A1 <: B & A2 <: B
     % could be detrimental if checking that something is an element of B is expensive
     extract_unions(A,As),
     findall(b(subset(Ai,B),pred,I),member(Ai,As),Conj),
     conjunct_predicates(Conj,TNewPred),
     %translate:print_bexpr(TNewPred),nl,
     get_texpr_expr(TNewPred,NewPred).
cleanup_post(equal(A,B),pred,I,truth,pred,I,multi/remove_equality) :- 
   same_texpr(A,B),always_well_defined(A),!. % ,print(removed_equal(A,B)),nl. 
cleanup_post(equal(A,B),pred,I,falsity,pred,I,multi/remove_equality_false) :- 
   different_texpr_values(A,B),!. %,print(removed_equal_false(A,B)),nl. 
cleanup_post(equal(A,B),pred,I,equal(A,B),pred,Info2,multi/detect_recursion) :- 
   get_texpr_id(A,ID), \+ member(prob_annotation(recursive(ID)),I),
   %occurs_in_expr(ID,B), 
   find_recursive_usage(B,ID),
   debug_println(9,recursion_detected(ID)),
   !, Info2 = [prob_annotation(recursive(ID))|I].
cleanup_post(equal(A,B),pred,_I,ResultExpr,pred,Info2,multi/simplify_bool_true) :- 
   % simplify bool(X)=TRUE -> X
   ( get_texpr_expr(A,convert_bool(X)), get_texpr_expr(B,boolean_true);
     get_texpr_expr(A,boolean_true),get_texpr_expr(B,convert_bool(X))
   ),
   !, debug_println(9,simplify_bool_equal_TRUE(X)),
   get_texpr_expr(X,ResultExpr), get_texpr_info(X,Info2).
cleanup_post(not_equal(A,B),pred,I,falsity,pred,I,multi/remove_disequality) :- 
   same_texpr(A,B),always_well_defined(A),!. % ,print(removed_not_equal(A,B)),nl. 
   % sometimes one uses & TRUE=TRUE to finish off guards, invariants, ...
   % exchange lambda expressions by a comprehension set
% TO DO: also add rule for bool(X)=FALSE -> not(X)
cleanup_post(not_equal(A,B),pred,I,truth,pred,I,multi/remove_disequality_false) :- 
   different_texpr_values(A,B),!. %  ,print(removed_not_equal_false(A,B)),nl.
cleanup_post(negation(A),pred,I,falsity,pred,I,multi/remove_negation_truth) :- 
   is_truth(A),!. % ,print(negation(A)),nl. 
cleanup_post(negation(A),pred,I,truth,pred,I,multi/remove_negation_falsity) :- 
   is_falsity(A),!. % ,print(negation(A)),nl. 
cleanup_post(convert_bool(A),boolean,I,boolean_true,boolean,I,multi/remove_convert_bool) :- 
   is_truth(A),!. %  ,print(convert_bool(A)),nl. 
cleanup_post(convert_bool(A),boolean,I,boolean_false,boolean,I,multi/remove_convert_bool_false) :- 
   is_falsity(A),!. %  ,print(convert_bool_false(A)),nl. 
cleanup_post(assertion_expression(Cond,_ErrMsg,Expr),_T,_I,BE,TE,IE,multi/remove_assertion_expression) :- 
   is_truth(Cond),!,  %print(remove_assertion_expression(Cond)),nl,
   Expr = b(BE,TE,IE).
cleanup_post(lambda(Ids,Pred,Expr),Type,I, comprehension_set(CompIds,CompPred),Type,[was(lambda)|I],multi/remove_lambda) :- !,
	unify_types_strict(Type,set(couple(_ArgType,ResType))),
	ResultId = '_lambda_result_',
	% get_unique_id('_lambda_result_',ResultId),
	def_get_texpr_id(Result,ResultId),
	get_texpr_type(Result,ResType),
	get_texpr_info(Result,[lambda_result]),
	append(Ids,[Result],CompIds),
	get_texpr_expr(Equal,equal(Result,Expr)),
	get_texpr_type(Equal,pred),
	extract_important_info_from_subexpression(Expr,EqInfo), % mark equality with wd condition if Expr has wd condition
	get_texpr_info(Equal,[lambda_result|EqInfo]),
	conjunct_predicates([Pred,Equal],CompPred).
cleanup_post(forall(Ids,Lhs,Rhs),pred,IOld,forall(Ids,Lhs,Rhs),pred,INew,single/forall_used_identifier) :- !,
	conjunct_predicates([Lhs,Rhs],P),
	add_used_identifier_info(Ids,P,IOld,INew).
cleanup_post(exists(Ids,P),pred,IOld,exists(Ids,P),pred,INew,single/exists_used_identifier) :- !,
	add_used_identifier_info(Ids,P,IOld,INew).
cleanup_post(precondition(TP,TS),subst,_,S,subst,I,multi/remove_triv_precondition) :-
	% remove trivial preconditions
	get_texpr_expr(TP,truth),!,
	get_texpr_expr(TS,S),get_texpr_info(TS,I).
	
cleanup_post(parallel(Statements),subst,I,Result,subst,I, single/parallel_to_multi_assign) :- 
 % this merges multiple assignments into a single one: advantage: only one waitflag set up
    %print(parallel(Statements)),nl,
    extract_assignments(Statements,LHS,RHS,Rest,Nr), Nr>1,!,
    (debug_mode(on) -> 
       print('Parallel to Assignment: '), translate:print_subst(b(parallel(Statements),subst,I)),nl
      ; true),
    (Rest == []  ->   Result  =  assign(LHS,RHS)
     ;  Result = parallel([b(assign(LHS,RHS),subst,[])|Rest])).
    %translate:print_subst(b(Result,subst,I)),nl.
cleanup_post(assign([b(identifier(ID),TYPE,INFO)],[EXPR]),subst,I,
             assign_single_id(b(identifier(ID),TYPE,INFO),EXPR),subst,I,single/assign_single_id) :- 
    \+ animation_minor_mode(eventb), % there is no support in the Event-B interpreter for assign_single_id yet
    simple_expression(EXPR), % the assign_single_id is not guarded by a waitflag; EXPR should not be too expensive too calculate
    !,
	(debug_mode(on) -> print('Single Assignment: '),
	    translate:print_subst(b(assign([b(identifier(ID),TYPE,INFO)],[EXPR]),subst,I)),nl 
	 ;  true).

cleanup_post(evb_becomes_such(Ids,Pred),subst,I,any(Primed,Pred,Assignment),subst,I,multi/ev2_becomes_such_any) :-
	prime_identifiers(Ids,Primed),
	create_texpr(assign(Ids,Primed),subst,[],Assignment).
cleanup_post(evb2_becomes_such(Ids,Pred),subst,I,becomes_such(Ids,Pred),subst,I,multi/ev2_becomes_such) :- !.
cleanup_post(successor,Type,I,Compset,Type,[was(successor)|I],multi/successor) :- !,
    % translation of succ
	pred_succ_compset(add,Compset).
cleanup_post(predecessor,Type,I,Compset,Type,[was(predecessor)|I],multi/predecessor) :- !,
    % translation of pred
	pred_succ_compset(minus,Compset).
cleanup_post(becomes_such(Ids1,Pred),subst,I,becomes_such(Ids2,Pred),subst,I,single/becomes_such) :- !,
	find_used_primed_ids(Pred,BeforeAfter),
	add_before_after_info(Ids1,BeforeAfter,Ids2).
cleanup_post(select([CHOICE|Rest]),subst,_,S,subst,I,single/remove_select) :-
    CHOICE=b(select_when(TRUTH,Subst),subst,_),
    is_truth(TRUTH),!, 
    (Rest=[] -> true ; print('Removing useless SELECT'),nl),
    get_texpr_expr(Subst,S),get_texpr_info(Subst,I).
cleanup_post(select([CHOICE|_],_ELSE),subst,_,S,subst,I,single/remove_select_else) :-
    CHOICE=b(select_when(TRUTH,Subst),subst,_),
    is_truth(TRUTH),!, print('Removing useless SELECT ELSE'),nl,
    get_texpr_expr(Subst,S),get_texpr_info(Subst,I).
cleanup_post(Expr,Type,I,Expr,Type,[contains_wd_condition|I],single/possibly_undefined) :-
    nonmember(contains_wd_condition,I),is_possibly_undefined(Expr).
% if a substitution has a sub-expression that is a substitution with that refers to the original
% value of a variable, we mark this substitution, too.
cleanup_post(Subst,subst,I,Subst,subst,NI,single/refers_to_old_state) :-
    safe_syntaxelement(Subst,Subs,_,_,_),
    % check if a child contains the refers_to_old_state flag
    setof(Reference, (member(Sub,Subs),
                         get_texpr_info(Sub,SubInfo),
                         memberchk(refers_to_old_state(References),SubInfo),
                         member(Reference,References)), ReferedIds),
    sort(ReferedIds,SortedIds),
    NI = [refers_to_old_state(SortedIds)|I].
% If the substitution of an operation contains a while whose invariant contains references x$0
% to the original value of a variable x, we must insert a LET substitution to preserve the original value.
cleanup_post(operation(Id,Results,Args,Body),Type,I,operation(Id,Results,Args,NewBody),Type,I,single/refers_to_old_state_let) :-
    get_texpr_info(Body,BodyInfo),
    memberchk(refers_to_old_state(References),BodyInfo),!,
    create_equalities_for_let(References,Ids,Equalities),
    conjunct_predicates(Equalities,P),
    insert_let(Body,Ids,P,NewBody).
cleanup_post(let_expression([],[],TExpr),Type,_,Expr,Type,I,multi/remove_let) :-
    % remove trivial let expressions without any introduced identifiers
    % this rule makes only sense in combination with the next rule which removes
    % simple let identifiers
    get_texpr_expr(TExpr,Expr),
    get_texpr_info(TExpr,I).
cleanup_post(let_expression(Ids,Exprs,Expr),Type,I,
             let_expression(NIds,NExprs,NExpr),Type,I,multi/remove_let) :-
    % remove let identifiers whose definitions are very simple, i.e. identifiers
    nth0(N,Exprs,TId,NExprs),get_texpr_id(TId,_),
    nth0(N,Ids,TI,NIds),get_texpr_id(TI,ToReplace),
    replace_id_by_expr(Expr,ToReplace,TId,NExpr).

% detect whether there is a patter of a recursive usage of the identifier:  ID(x) or ID[x] or x:ID
find_recursive_usage(TExpr,ID) :- 
    syntaxtraversion(TExpr,Expr,_,_,Subs,TNames), % print(try_id(ID,Expr,Subs,TNames)),nl,
    ((Expr = function(Fun,_), get_texpr_id(Fun,ID)) -> true
     ; (Expr = image(Rel,_), get_texpr_id(Rel,ID)) -> true
     ; (Expr = member(_,Set), get_texpr_id(Set,ID)) -> true
     ; \+ (member(ID1,TNames),get_texpr_id(ID1,ID)), % new local variable with same name
        member(Sub,Subs), find_recursive_usage(Sub,ID)
     ).


extract_unions(A,R) :- get_texpr_expr(A,union(A1,A2)),!,
     extract_unions(A1,R1), extract_unions(A2,R2),
     append(R1,R2,R).
extract_unions(A,[A]).

gen_member_predicates(B,SEl,TExpr) :- create_texpr(member(SEl,B),pred,[],TExpr).

% when constructing an expression/predicate: important to ripple wd information up
extract_important_info_from_subexpression(b(_,_,Info),NewInfo) :-
    (member(contains_wd_condition,Info) -> NewInfo = [contains_wd_condition] ; NewInfo=[]).

check_ids([],[CoupleExpr],CoupleExpr). % we terminate with a single expression
check_ids([ID|T],[CoupleExprID|CT],Rest) :-
  same_id(ID,CoupleExprID,_),
  check_ids(T,CT,Rest).

% extract all assignments from a list of statements; last arg is the number of assignments extracted
extract_assignments([],[],[],[],0).
extract_assignments([b(assign_single_id(LHS,RHS),subst,_I)|T],ResLHS,ResRHS,Rest,Nr) :- !,
    ResLHS=[LHS|LT],ResRHS=[RHS|RT],
    extract_assignments(T,LT,RT,Rest,N), Nr = N.
extract_assignments([b(assign(LHS,RHS),subst,Info)|T],ResLHS,ResRHS,Rest,Nr) :- 
    \+ member(assigned_after(_),Info), % these assignments are treated in a special way
    \+ member(modifies(_),Info),
    !,
    extract_assignments(T,LT,RT,Rest,N),
    append(LHS,LT,ResLHS), append(RHS,RT,ResRHS), Nr is N+1.
extract_assignments([H|T],LT,RT,[H|Rest],Nr) :- 
    extract_assignments(T,LT,RT,Rest,Nr).
    % assigned_after(Primed),modifies(Var)

% check if we have a simple expression which will not be complicated to calculate
simple_expression(b(E,_,_)) :- simple2(E).
 %  (simple2(E) -> true ; print(not_simple(E)),nl).
simple2(bool_set).
simple2(boolean_false).
simple2(boolean_true).
simple2(empty_sequence).
simple2(empty_set).
simple2(identifier(_)).
simple2(integer_set(_)).
simple2(integer(_)).
simple2(max_int).
simple2(min_int).
simple2(string_set).
simple2(string(_)).
simple2(value(_)).
simple2(first_of_pair(_)).  simple2(second_of_pair(_)).
simple2(add(A,B)) :- simple_expression(A), simple_expression(B).
simple2(minus(A,B)) :- simple_expression(A), simple_expression(B).
simple2(multiplication(A,B)) :- simple_expression(A), simple_expression(B).
simple2(unary_minus(A)) :- simple_expression(A).
simple2(convert_bool(A)) :- simple_predicate(A).

simple_predicate(b(E,_,_)) :- simplep2(E).
simplep2(equal(A,B)) :- simple_expression(A), simple_expression(B). % could be slightly more expensive if set type
simplep2(not_equal(A,B)) :- simple_expression(A), simple_expression(B). % ditto
simplep2(less(A,B)) :- simple_expression(A), simple_expression(B).
simplep2(less_equal(A,B)) :- simple_expression(A), simple_expression(B).
simplep2(greater(A,B)) :- simple_expression(A), simple_expression(B).
simplep2(greater_equal(A,B)) :- simple_expression(A), simple_expression(B).

:- use_module(self_check).
:- assert_must_succeed(b_ast_cleanup:nested_couple_to_list(b(couple(b(couple(a,b),x,[]),c),xx,[]),[a,b,c])).
:- assert_must_succeed(b_ast_cleanup:nested_couple_to_list(b(couple(a,c),x,[]),[a,c])).
:- assert_must_succeed(b_ast_cleanup:nested_couple_to_list(b(couple(a,b(couple(b,c),x,[])),xx,[]),[a,b(couple(b,c),x,[])])).
nested_couple_to_list(NC,L) :- nested_couple_to_list_dcg(NC,L,[]).
nested_couple_to_list_dcg(b(couple(A,B),_,_)) --> !,
   nested_couple_to_list_dcg(A), [B].
nested_couple_to_list_dcg(E) --> [E].


create_equalities_for_let([],[],[]).
create_equalities_for_let([oref(PrimedId,OrigId,Type)|RefRest],[TPrimed|Prest],[Equality|Erest]) :-
    create_texpr(identifier(OrigId),Type,[],TOrig),
    create_texpr(identifier(PrimedId),Type,[],TPrimed),
    create_texpr(equal(TPrimed,TOrig),pred,[],Equality),
    create_equalities_for_let(RefRest,Prest,Erest).


% inserts a let statement. If the original statement is a precondition or any, the let is moved
% inside the original statement to prevent strange side-effects. This can be used for other,
% non-value changing substitutions as well.
insert_let(TExpr,Ids,P,NTExpr) :-
    remove_bt(TExpr,Expr,NewExpr,NTExpr),
    move_let_inside(Expr,Old,New,NewExpr),!,
    insert_let(Old,Ids,P,New).
insert_let(TExpr,Ids,P,NTExpr) :-
    create_texpr(let(Ids,P,TExpr),subst,[],NTExpr).
move_let_inside(precondition(Cond,Old),Old,New,precondition(Cond,New)).
move_let_inside(any(Any,Where,Old),Old,New,any(Any,Where,New)).

% find_one_point_rules(+TypedIds,+Blacklist,+Predicates,
%                      -LetIds,-LetExprs,-RemainingIds,-RemainingPredicates)
% TypedIds: The ids that are quantified in the exists clause
% Blacklist: All ids that must not be used in the found expression
% Predicates: The predicates of the exists (without already used id=E predicates)
% LetIds: The ids that can be introduced as LET
% LetExprs. For each id (in LetIds) the corresponding expression
% RemainingIds: The ids that are not converted into LETs
% RemainingPredicates: The predicates after removing the id=E predicates
find_one_point_rules([],Preds,_,[],[],[],Preds).
find_one_point_rules([TId|Irest],Preds,Blacklist,LetIds,Exprs,RestIds,NewPreds) :-
    ( select_equality(TId,Preds,Blacklist,Expr,RestPred) ->
        LetIds = [TId|Lrest], Exprs = [Expr|Erest],
        RestIds = RestIds2, Preds2 = RestPred
    ; otherwise ->
        LetIds = Lrest, Exprs = Erest,
        RestIds = [TId|RestIds2], Preds2 = Preds),
    find_one_point_rules(Irest,Preds2,Blacklist,Lrest,Erest,RestIds2,NewPreds).
% select a predicate from Preds of the form id=Expr (or Expr=id) where Expr does not contain
% references to identifiers in Blacklist. Rest is Preds without id=Expr
select_equality(TId,Preds,Blacklist,Expr,Rest) :-
    get_texpr_id(TId,Id),
    get_texpr_expr(TEqual,equal(TA,TB)),
    select(TEqual,Preds,Rest),
    ( get_texpr_id(TA,Id),TB=Expr ; get_texpr_id(TB,Id),TA=Expr ),
    find_identifier_uses(Expr,[],UsedIds),
    ord_disjoint(Blacklist,UsedIds).

pred_succ_compset(Op,comprehension_set([A,B],Pred)) :-
	create_texpr(equal(B,TArithOp),pred,[],Pred),
	%get_unique_id('_a_',AId),get_unique_id('_b_',BId),
	BId = '_lambda_result_',
	(Op = add -> AId = '_succ_' ; AId='_pred_'),
	create_texpr(identifier(AId),integer,[],A),
	create_texpr(identifier(BId),integer,[],B),
	create_texpr(integer(1),integer,[],Integer),
	create_texpr(ArithOp,integer,[],TArithOp),
	ArithOp =.. [Op,A,Integer].

add_used_identifier_info(Ids,P,IOld,[used_ids(FoundIds)|IOld]) :-
	% add used identifiers to information
	get_texpr_ids(Ids,Ignore),
	find_identifier_uses(P,Ignore,FoundIds).

% generation of unique identifiers
:- dynamic unique_id_counter/1.
unique_id_counter(1).

get_unique_id(Prefix,Id) :-
	retract(unique_id_counter(Old)),
	New is Old + 1,
	assert(unique_id_counter(New)),
	safe_atom_chars(Prefix,CPrefix,get_unique_id1),
	number_chars(Old,CNumber),
	append(CPrefix,CNumber,CId),
	safe_atom_chars(Id,CId,get_unique_id2).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% replace all variables in types by "any"

ground_type(T,_) :- ground(T),!.
ground_type(T,Exceptions) :- var(T),!,
	( exact_member(T,Exceptions) -> !
	; otherwise -> T=any).
ground_type(T,Exceptions) :-
	functor(T,_,Arity),
	ground_type_args(Arity,T,Exceptions).
ground_type_args(0,_T,_Exceptions) :- !.
ground_type_args(N,T,Exceptions) :-
	arg(N,T,Arg),ground_type(Arg,Exceptions),
	N2 is N-1,ground_type_args(N2,T,Exceptions).

% put optional before/after usage into the information of the identifiers
% makes only sense in the context of becomes_such substitutions
add_before_after_info([],_,[]).
add_before_after_info([TId|IRest],BeforeAfter,[TId2|ORest]) :-
	def_get_texpr_id(TId,Id),
	( member(ba(Id,BeforeId),BeforeAfter) ->
	    get_texpr_type(TId,Type),
	    get_texpr_info(TId,Info),
	    create_texpr(identifier(Id),Type,[before_substitution(Id,BeforeId)|Info],TId2)
	; otherwise ->
	    TId = TId2 ),
	add_before_after_info(IRest,BeforeAfter,ORest).

% find all used pairs of before/after variables, e.g. ba(x,'x$0')
% see becomes_such substitutions
find_used_primed_ids(TExpr,Uses) :-
	find_used_primed_ids2(TExpr,[],Uses).
find_used_primed_ids2(TExpr,In,Out) :-
	syntaxtraversion(TExpr,Expr,_,Infos,Subs,_),
	( Expr = identifier(_), member(before_substitution(Id,FullId),Infos) ->
	    ord_add_element(In,ba(Id,FullId),Uses)
	; otherwise ->
	    In = Uses ),
	find_used_primed_ids_l(Subs,Uses,Out).
find_used_primed_ids_l([],Uses,Uses).
find_used_primed_ids_l([Expr|Rest],In,Out) :-
	find_used_primed_ids2(Expr,In,Uses),
	find_used_primed_ids_l(Rest,Uses,Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% find expressions which are possibly not well-defined

is_possibly_undefined(Expr) :-
        syntaxtraversion(_,Expr,_,_,Subs,_),
        member(Sub,Subs),
        get_texpr_info(Sub,Info),
        member(contains_wd_condition,Info),!.
is_possibly_undefined(Expr) :-
        functor(Expr,F,_),has_wd_condition(F,Expr).  %%,print(wd(F,Expr)),nl.

% division and module must not divide by zero
has_wd_condition(div,Expr) :- arg(2,Expr,DIV),
    \+ definitely_not_zero(DIV).
has_wd_condition(modulo,Expr) :- arg(2,Expr,DIV),
    (\+ definitely_not_zero(DIV) ;
     \+ definitely_not_negative(DIV) ;
      arg(1,Expr,A1), \+ definitely_not_negative(A1)).
% power_of must have a non-negative exponent (?)
has_wd_condition(power_of,Expr) :- arg(2,Expr,EXP),
   \+ definitely_not_negative(EXP).
% min and max need a non-empty set
has_wd_condition(max,Expr) :- arg(1,Expr,S), \+ definitely_not_empty(S).
has_wd_condition(min,Expr) :- arg(1,Expr,S), \+ definitely_not_empty(S).
% all the sequence operations must not be applied to non-sequence
% relations and some must not be applied to empty-sequences
has_wd_condition(size,E) :- arg(1,E,S), \+ definitely_sequence(S).
has_wd_condition(first,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(last,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(front,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(tail,E) :- arg(1,E,S), \+ definitely_not_empty_sequence(S).
has_wd_condition(rev,E) :- arg(1,E,S), \+ definitely_sequence(S).
has_wd_condition(concat,_).
has_wd_condition(insert_front,_).
has_wd_condition(insert_tail,_).
has_wd_condition(restrict_front,_).
has_wd_condition(restrict_tail,_).
has_wd_condition(general_concat,_).
% functions must not be applied to values outside their domain
has_wd_condition(function,_).
% the general intersection must not be applied to an empty set of sets
has_wd_condition(general_intersection,_).
has_wd_condition(quantified_intersection,_). % gets translated to general_intersection
% card must not be applied to infinite sets:
has_wd_condition(card,Expr) :- arg(1,Expr,S), \+ definitely_finite(S).

definitely_not_zero(b(integer(X),integer,_)) :- number(X), X \= 0.
definitely_not_negative(b(integer(X),integer,_)) :- number(X), X >= 0.
% to do add more ?: card(_), ...

definitely_not_empty(b(S,_,_)) :- definitely_not_empty2(S).
definitely_not_empty2(set_extension(_)).
definitely_not_empty2(sequence_extension(_)).

definitely_not_empty_sequence(b(S,_,_)) :- definitely_not_empty_sequence2(S).
definitely_not_empty_sequence2(sequence_extension(_)).

definitely_sequence(b(S,_,_)) :- definitely_sequence2(S).
definitely_sequence2(empty_sequence).
definitely_sequence2(sequence_extension(_)).

definitely_finite(b(S,_,_)) :- definitely_finite2(S).
definitely_finite2(empty_set).
definitely_finite2(empty_sequence).
definitely_finite2(set_extension(_)).
definitely_finite2(sequence_extension(_)).


