% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(b_global_sets, 
          [b_get_global_constants/1, b_get_global_enumerated_sets/1, b_get_global_sets/1,
           lookup_global_constant/2,
           is_b_global_constant/3,
           is_used_b_global_constant/3, is_unused_b_global_constant/3,
           
           b_global_set/1, b_non_empty_global_set/1, b_empty_global_set/1,
           b_global_deferred_set/1,
           b_global_set_with_potential_symmetry/1, b_global_deferred_set_with_card_gt1/1,
           b_partially_deferred_set/1, unfixed_deferred_set/1,
           b_exists_global_set_with_potential_symmetry/0,
           b_global_set_cardinality/2,
           b_reset_global_set_type_definitions/0,
           b_check_and_precompile_global_set_type_definitions/0,
           %b_precompile_used_global_set_information/0,
           b_get_prob_deferred_set_elements/1,add_prob_deferred_set_elements_to_store/2,
           
           find_inequal_global_set_identifiers/4,
           
           b_integer_set/1, b_integer_or_string_set/1,
           b_type2global_set/2, try_b_type2global_set/2,
           
           b_fd_type/3, b_fd_card/2,
           is_global_set_of_cardinality_one/2,
           global_type/2, global_type_wf/3,
           enum_global_type/2,
           
           all_elements_of_type/2,
           
           set_user_defined_scope/2, 
           
           add_new_global_set/1
           %,add_global_set_with_named_constants/2  /* for use by Z,... */
           ]).
           
:- use_module(debug).
:- use_module(tools).
:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(self_check).
:- use_module(preferences).
:- use_module(error_manager).
:- use_module(bsyntaxtree).
:- use_module(library(between),[between/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides support for deferred/enumerated set elements in B.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(bmachine).

/* what we call global sets here usually called "given sets", which can be
   either enumerated or deferred */


:- dynamic b_global_constant/3. % named element of a deferred or enumerated set
  % for enumerated sets S = {a,b,c}  we would have entries a,b,c as b_global_constant
  % for deferred set S, we may add CONSTANTS which are element of S as b_global_constant
:- dynamic unused_b_global_constant/3.  % TO DO: compute properly again !!

b_get_global_enumerated_sets(GSets) :-
  findall(GS, (b_global_set(GS),  is_b_global_constant(GS,_Nr,_Cst)), GSets).
b_get_global_sets(GSets) :-
  findall(GS, b_global_set(GS), GSets).

b_get_global_constants(Csts) :- findall(Cst,is_b_global_constant(_,_,Cst),Csts).

is_b_global_constant(Set,Nr,Cst) :- b_global_constant(Set,Nr,Cst).
is_unused_b_global_constant(Set,Nr,Cst) :- unused_b_global_constant(Set,Nr,Cst).
is_used_b_global_constant(Set,Nr,Cst) :-
   b_global_constant(Set,Nr,Cst), \+ unused_b_global_constant(Set,Nr,_).
   
lookup_global_constant(Id,Val) :-
  b_global_constant(Set,Nr,Id),
  Val = fd(Nr,Set).
/*  debug_print(lookup(Id,Val)),debug_nl. */


:- dynamic b_precompiled_global_set/1.
b_precompiled_global_set(_) :- print_error('*** b_global_set not precompiled'),fail.

b_global_set(GS) :- b_precompiled_global_set(GS).

 
b_global_deferred_set(GS) :-
  b_global_set(GS),
  \+ is_b_global_constant(GS,_Nr,_Cst). % note : some deferred sets are translated into partially_deferred_set
  
% either an enumerated set with unused constants or a deferred sets where some constants were lifted into the deferred set
b_partially_deferred_set(GS) :-
   b_global_set(GS),
   (is_b_global_constant(GS,_Nr,_Cst) -> true), % GS is in principle enumerated
   (unused_b_global_constant(GS,_X,_) -> true).


b_global_set_with_potential_symmetry(GS) :-
   b_global_set(GS),
   (b_global_deferred_set_with_card_gt1(GS) -> true
    ; ((unused_b_global_constant(GS,X,_),
        unused_b_global_constant(GS,Y,_), X\=Y) -> true % At least two unused constants exist
        ; fail)).  

b_global_deferred_set_with_card_gt1(GS) :-
   b_global_deferred_set(GS),
   extract_setsize_from_machine_cache(GS,Low,Up),
   Up>Low.   
  
:- volatile b_exists_global_set_with_potential_symmetry/0.
:- dynamic b_exists_global_set_with_potential_symmetry/0.

b_check_and_precompile_global_set_type_definitions :-
                      /* must be called before the animator can run !! */
  b_reset_global_set_type_definitions,
  retractall(b_exists_global_set_with_potential_symmetry),
  debug_println(9,'Checking existence of global type definitions: '),
  debug_print(9,'%  '),
  b_get_machine_set(Set),
  add_new_global_set(Set),
  (b_fd_type(Set,LowBnd,UpBnd)  /* also computes & caches the size of the SET */
    -> debug_print(9,' '),debug_print(9,Set),
       debug_print(9,'=='),debug_print(9,LowBnd),debug_print(9,'..'),debug_print(9,UpBnd)
    ;  add_internal_error('No b_fd_type/2 definition for global set: ',b_fd_type(Set,_,_))
   ),debug_print(9,'  '),
   fail.
b_check_and_precompile_global_set_type_definitions :-  
   find_all_used_enumerated_elements(UsedEnumeratedElements),
   b_fd_type(GS,Low,Up),
   (is_b_global_constant(GS,_Nr,_Cst) -> true), % GS is in principle enumerated
   (user_forced_symmetry(GS)
     -> ForceSymm=true, print('FORCING SYMMETRY: '), print(GS), print(' ')
     ;  ForceSymm=false
   ),
   enum_fd(X,Low,Up), /* enumerate all elements of GS */
   can_symmetry_reduc_applied_to_constant(ForceSymm,GS,X,UsedEnumeratedElements,Display),
   assert(unused_b_global_constant(GS,X,deferred)),
   debug_println(9,unused_b_global_constant(Display)),
   fail.
b_check_and_precompile_global_set_type_definitions :- 
  (b_global_set_with_potential_symmetry(_)
   -> assert(b_exists_global_set_with_potential_symmetry),
      debug_nl,print('% Symmetry is potentially useful for this machine'),nl
   ; true),
  debug_nl(9),
  /* clean up: */
  reset_global_set_user_defined_scope.


can_symmetry_reduc_applied_to_constant(true,GS,X,_Enums,GS/X). % if symmetry enforced, all elements are
                                                               % subject to symmetry reduction
can_symmetry_reduc_applied_to_constant(false,GS,X,UsedEnums,Display) :-
    ( is_b_global_constant(GS,X,Name) ->  % sym-reduction can be applied to an enumerated set member
        nonmember(Name,UsedEnums),        % if it's not used anywhere
        Display=Name
    ; otherwise -> Display=GS/X).         % sym-reduction can be applied to deffered set elements


find_all_used_enumerated_elements(Elements) :-
    b_get_all_used_identifiers(Identifiers),
    b_get_machine_variables(TVariables),get_texpr_ids(TVariables,Variables),
    % We assume that b_get_all_used_identifiers/1 is a proper ordset. (sorted with sort/2)
    % As an optimisation, we remove the variables, they cannot be enumerated elements.
    list_to_ord_set(Variables,OVariables),
    ord_subtract(Identifiers,OVariables,Elements),!.
find_all_used_enumerated_elements(_Elements) :-
    add_error_and_fail(b_global_sets, 'find_all_used_enumerated_elements failed').
    
   
reset_global_set_user_defined_scope :-
   retract(global_set_user_defined_scope(GS,_Scope)),
   (b_fd_type(GS,_Low,_Up)
     -> true
     ;  add_error(b_global_sets,'Trying to set scope of unknown SET: ',GS)
   ),fail.
reset_global_set_user_defined_scope.

% Return those constants which are implicitly introduced by ProB, but not really
% part of the model. These are the elements of a deferred set.
b_get_prob_deferred_set_elements(TIds) :-
    findall( TId,
             ( prob_deferred_set_element(GS,_Nr,Id),
                 create_texpr(identifier(Id),global(GS),[],TId)),
             TIds).

add_prob_deferred_set_elements_to_store(OldStore,NewStore) :-
    %print(add_prob_deferred_set_elements_to_store(OldStore)),nl,
    findall(bind(Id,fd(Nr,GS)),(prob_deferred_set_element(GS,Nr,Id),
                                \+ member(bind(Id,_),OldStore)),
                                NewStore, OldStore).

prob_deferred_set_element(GlobalSet,Elem,Id) :-
    b_global_set(GlobalSet),
    atom_codes(GlobalSet,GlobalSetCodes),append(GlobalSetCodes,ECodes,NameCodes),
    b_fd_type(GlobalSet,Low,Up),
    between(Low,Up,Elem),
    \+ is_b_global_constant(_,Elem,_), % the identifier is not used somewhere else
    \+ is_b_global_constant(Elem,_,_), % not used as another SET name
    number_codes(Elem,ECodes),
    atom_codes(Id,NameCodes).
                 

user_forced_symmetry(GS) :- % check if the user defined FORCE_SYMMETRY_GS == TRUE
     string_concatenate('FORCE_SYMMETRY_',GS,DefName),
     b_get_typed_definition(DefName,[],F), get_texpr_expr(F,boolean_true).

% TO DO: re-add this feature, so that unused enumerated set elements can be detected
% :- use_module(bmachine,[used_identifier/1]).
%b_precompile_used_global_set_information :- 
%  b_precompute_used_identifiers,
%  b_global_constant(GlobalSetName,Nr,CstName),
%  (used_identifier(CstName) -> true
%   ; (assert(unused_b_global_constant(GlobalSetName,Nr,CstName)),
%      debug_print(9,unused_b_global_constant(GlobalSetName,Nr,CstName)), debug_nl(9))
%  ),
%  fail.
%b_precompile_used_global_set_information :- 
%   retractall(b_exists_deferred_set_or_unused_csts_with_card_gt_1),
%   ((b_global_deferred_set_with_card_gt1(_GS) ; 
%     (is_unused_b_global_constant(GS,Nr,_),
%      is_unused_b_global_constant(GS,Nr2,_), Nr2 \= Nr))
%       -> (assert(b_exists_deferred_set_or_unused_csts_with_card_gt_1),
%           print_message('symmetry can be used for this machine'))
%     ; true
%   ).

:- volatile is_global_set_of_cardinality_one/2.
:- dynamic is_global_set_of_cardinality_one/2.
b_reset_global_set_type_definitions :- 
  % print_message('resetting global sets'),
  retractall(b_precompiled_global_set(_)),
  retractall(enumerated_set(_)),
  retractall(fixed_deferred_set_size(_,_)),
  retractall(extract_setsize_from_machine_cache(_,_,_)),
  retractall(is_global_set_of_cardinality_one(_,_)),
  retractall(b_global_constant(_,_,_)),
  retractall(unused_b_global_constant(_,_,_)).
  
add_new_global_set(Set) :-
  (b_precompiled_global_set(Set)
   -> add_internal_error('Global set already exists: ',add_new_global_set(Set))
   ;  assert(b_precompiled_global_set(Set))).
  
/* --------------------------------------------------------- */
/* Extracting Finite Domain type information from the B machine */
/* --------------------------------------------------------- */



/* below treats annotations in the form:
DEFINITIONS
    scope_Name == 1..3;
    scope_Code == 4..8
 which inform us about which finidte domain ranges we should
  give to global sets defined in SETS
*/


:- dynamic extract_setsize_from_machine_cache/3.
extract_setsize_from_machine_cache(_,_,_) :-
   print_error('*** extract_setsize_from_machine not precompiled'),fail.

extract_setsize_from_machine(GlobalSetName,LowBound,UpBound) :- 
  extract_setsize_from_machine_cache(GlobalSetName,L,U),!,
  LowBound=L,UpBound=U.
extract_setsize_from_machine(GlobalSetName,LowBound,UpBound) :-
  b_get_named_machine_set(GlobalSetName,ListOfConstants),!,
  ~~pp_mnf(assert_enumerated_set(GlobalSetName,ListOfConstants,LowBound,UpBound)),
  ((get_user_defined_scope(GlobalSetName,DL,DU), DU-DL =\= UpBound-LowBound)
    -> add_error(extract_setsize_from_machine,'Conflict between scope_ Definition and explicit enumeration of SET:',GlobalSetName)
    ; true
  ).
extract_setsize_from_machine(GlobalSetName,L,U) :-
   /* we have a DEFERRED SET */
   extract_DEFERRED_setsize_from_machine(GlobalSetName,LowBound,UpBound),
   (LowBound=UpBound -> assert(is_global_set_of_cardinality_one(GlobalSetName,LowBound)) ; true),
   assert_extract_setsize_from_machine_cache(GlobalSetName,LowBound,UpBound),
   (b_get_disjoint_constants_of_type(GlobalSetName, DisjointConstants)
    -> /* add those constants to the deferred set as though they were enumerated set elements: improves performance by avoiding redundant permutations; is a kind of symmetry breaking */
     print(adding_disjoint_constants(GlobalSetName,DisjointConstants)),nl,
     ~~mnf(add_named_constants_to_global_set(GlobalSetName,DisjointConstants))
    ; true
   ),
   LowBound=L,UpBound=U.

/* for Deferred SETS : */
extract_DEFERRED_setsize_from_machine(GlobalSetName,L,U) :-
   b_get_properties_from_machine(Properties), member_in_conjunction(C,Properties),
  % check if the PROPERTIES contain an expression of the form card(GS) = Nr
   is_equality_card_global_set(C,GlobalSetName,Card),!,
   assert(fixed_deferred_set_size(GlobalSetName,Card)),
   LowBound=1, UpBound = Card,
   ((get_user_defined_scope(GlobalSetName,DL,DU),
      DU-DL =\= UpBound-LowBound)
    -> add_error(extract_DEFERRED_setsize_from_machine,'Conflict between scope_ Definition and PROPERTIES:',GlobalSetName)
    ; true
   ),
   LowBound=L,UpBound=U.
extract_DEFERRED_setsize_from_machine(GlobalSetName,L,U) :-  /* check if there is a scope_ DEFINITION */
  get_user_defined_scope(GlobalSetName,LowBound,UpBound),!,
  LowBound=L,UpBound=U.
extract_DEFERRED_setsize_from_machine(GlobalSetName,LowBound,UpBound) :- 
   get_preference(globalsets_fdrange,DefUpBound),
   find_minimum_cardinality(GlobalSetName,Card),Card>=DefUpBound,!,
   nl,print_message(inferred_minimum_set_size(GlobalSetName,Card)),
   LowBound=1, UpBound = Card.
extract_DEFERRED_setsize_from_machine(GlobalSetName,LowBound,UpBound) :-
   LowBound=1, get_preference(globalsets_fdrange,UpBoundPref),
   ((find_maximum_cardinality(GlobalSetName,MaxCard),UpBoundPref>MaxCard)
    -> print(reducing_deferred_set_size(GlobalSetName,MaxCard)),nl,
       UpBound=MaxCard
    ; UpBound=UpBoundPref).

find_maximum_cardinality(GlobalSetName,MaxCard) :-
   b_get_machine_constants(Csts),
   b_get_properties_from_machine(Properties),
   findall(MC2,find_maximum_cardinality2(Csts,Properties,GlobalSetName,MC2),List),
   debug_println(9,maxcard_list(GlobalSetName,List)),
   min_member(MaxCard,List). 
find_maximum_cardinality2(_Csts,Properties,GlobalSetName,MaxCard) :-
    member_in_conjunction(C,Properties),
    get_texpr_expr(C,equal(LHS,RHS)),
    (get_texpr_expr(LHS,identifier(GlobalSetName)), R=RHS
     ;
     get_texpr_expr(RHS,identifier(GlobalSetName)), R=LHS
    ),
    get_texpr_expr(R,set_extension(SetExtElements)),
    /* We have an axiom of the form  GlobalSetName = { El1, ..., Eln} */
    length(SetExtElements,MaxCard).
    
find_minimum_cardinality(GlobalSetName,MinCard) :-
 /* try to find out from the Properties whether there is an implied minimum cardinality */
   b_get_machine_constants(Csts),
   b_get_properties_from_machine(Properties),
   findall(MinCard2,find_minimum_cardinality2(Csts,Properties,GlobalSetName,MinCard2),List),
   debug_println(9,mincard_list(GlobalSetName,List)),
   max_member(MinCard,List).

find_minimum_cardinality2(_,Properties,GlobalSetName,MinCard) :-
   member_in_conjunction(C,Properties),
   find_minimum_cardinality3(C,Properties,GlobalSetName,MinCard).
find_minimum_cardinality2(_Constants,_Properties,GlobalSetName,MinCard) :-  
   %print(trying_find_min_card(GlobalSetName,Constants)),nl,
   b_get_disjoint_constants_of_type(GlobalSetName, DisjointConstants),
   length(DisjointConstants,MinCard), MinCard>1.

find_minimum_cardinality3(C,_Properties,GlobalSetName,MinCard) :-
  % check if the PROPERTIES contain an expression of the form card(GS) >= Nr or similar
   is_greaterequalthan_card_global_set(C,GlobalSetName,MinCard).
find_minimum_cardinality3(C,Properties,GlobalSetName,MinCard) :-
  % find things like partition(GlobalSetName,S1,...,SN)
   get_texpr_expr(C,partition(Set,DisjSets)),
   global_set_identifier(Set,GlobalSetName),
   % print(partition(Set,DisjSets,GlobalSetName)),nl,
   find_min_card_of_disjoint_sets(DisjSets,Properties,MinCard). % ,print(min(MinCard)),nl.
find_minimum_cardinality3(C,_Properties,GlobalSetName,MinCard) :-
   get_texpr_expr(C,member(_Something,SURJ)),
   is_surjection(SURJ,Set,OTHERSET),
   global_set_identifier(Set,GlobalSetName), 
   (minimum_cardinality(OTHERSET,MinCard) -> true ; print('*** could not compute mincard: '), print(OTHERSET),nl,fail),
   (MinCard=inf -> add_error(b_global_sets,'Set has to be of infinite cardinality: ',GlobalSetName),fail ; true),
   print(total_surjection(OTHERSET)),nl.
 
is_surjection(SURJ,FromSet,ToSet) :-  get_texpr_expr(SURJ,total_surjection(FromSet,ToSet)).
is_surjection(SURJ,FromSet,ToSet) :-  get_texpr_expr(SURJ,partial_surjection(FromSet,ToSet)).
is_surjection(SURJ,FromSet,ToSet) :-  get_texpr_expr(SURJ,total_bijection(FromSet,ToSet)).
is_surjection(SURJ,FromSet,ToSet) :-  get_texpr_expr(SURJ,partial_bijection(FromSet,ToSet)).
   
% compute the minmum cardinality of an expression
minimum_cardinality(ID,MinCard) :- global_set_identifier(ID,GlobalSet),!,
  (b_fd_type(GlobalSet,LowBnd,UpBnd) -> MinCard is 1+UpBnd-LowBnd
   ; print('*** Do not know card of : '), print(GlobalSet),nl,
   % TO DO: do full blown cardinality analysis of deferred sets
    fail).
minimum_cardinality(b(EXPR,TYPE,_INFO),MinCard) :-
  minimum_cardinality2(EXPR,TYPE,MinCard).

%minimum_cardinality2(A,T,R) :- print(mincard(A,T,R)),nl,fail.
minimum_cardinality2(bool_set,_,2).
minimum_cardinality2(cartesian_product(A,B),_,Res) :- minimum_cardinality(A,CA),
 minimum_cardinality(B,CB), kernel_objects:safe_mul(CA,CB,Res).
% TO DO: enumerated sets, ... seq, relations, ... ?


 find_min_card_of_disjoint_sets([],_Properties,0).
 find_min_card_of_disjoint_sets([H|T],Properties,R) :-
    find_min_card_of_disjoint_sets(T,Properties,TN),
    (minimum_cardinality_of_set(H,Properties,MC)
      -> true % print(min_card(MC,H)),nl 
       ; add_internal_error('Call failed: ',minimum_cardinality_of_set(H,Properties,MC)), MC=1),
    R is TN+MC.


minimum_cardinality_of_set(b(Expr,set(Type),_),Properties,MC) :-
  minimum_cardinality_of_set_aux(Expr,Type,Properties,MC).

%minimum_cardinality_of_set_aux(P,Type,_,_MC) :- print(min_card(P,Type)),nl,fail.
minimum_cardinality_of_set_aux(set_extension(List),GlobalSetName,Properties,MinCard) :- !,
   get_texpr_id_names(List,AtomicIDList),
   find_inequal_global_set_identifiers(AtomicIDList,GlobalSetName,Properties,DisjointConstants),
   length(DisjointConstants,MinCard).
%   print(disj(MinCard,DisjointConstants)),nl,nl.
minimum_cardinality_of_set_aux(identifier(CST),Type,Properties,MC) :-
   EMPTYSET = b(empty_set,set(Type),_),
   get_texpr_expr(NEQ,not_equal(LHS,RHS)), 
   member_in_conjunction(NEQ,Properties),
   (LHS=b(identifier(CST),set(Type),_),RHS=EMPTYSET ;
    RHS=b(identifier(CST),set(Type),_),LHS=EMPTYSET ),
   %print(neq(NEQ,CST,RHS)),nl,
   !,
   MC=1.
% TO DO: add more cases like explicit set {a}
minimum_cardinality_of_set_aux(_,_,_,0).
   
% the following could be further improved
% try and find out how many different identifiers are in a list in the context of a predicate Prop
%find_inequal_global_set_identifiers(L,GS,P,_R) :- print(find_inequal_global_set_identifiers(L,GS)),nl,fail.
   
find_inequal_global_set_identifiers([],_,_,[]).
find_inequal_global_set_identifiers([ID|T],GS,Prop,Res) :-
    (atomic(ID) -> true
     ; add_internal_error('Not atomic id: ',find_inequal_global_set_identifiers([ID|T],GS,Prop,Res)),fail),
    findall(ID2, (member_in_conjunction(C,Prop), %print(chck(C)),nl,
                  inequality(C,GS,ID,ID2),
                  member(ID2,T)), DiffIDs),
    ((DiffIDs=[],T=[_,_|_]) % then try to proceed with T, ignoring ID
      -> find_inequal_global_set_identifiers(T,GS,Prop,Res)
      ;  find_inequal_global_set_identifiers(DiffIDs,GS,Prop,RT), Res = [ID|RT]
     ).

inequality(b(P,pred,_),GS,ID,OtherID) :-
  inequality_aux(P,GS,ID,OtherID).
inequality_aux(not_equal(I1,I2),GS,ID,OtherID) :-
  gs_identifier(I1,GS,ID1), gs_identifier(I2,GS,ID2),
 (ID=ID1 -> OtherID=ID2 ; ID=ID2,OtherID=ID1).
inequality_aux(negation( b(equal(I1,I2),pred,_)),GS,ID,OtherID) :-
  gs_identifier(I1,GS,ID1), gs_identifier(I2,GS,ID2),
  (ID=ID1 -> OtherID=ID2 ; ID=ID2,OtherID=ID1).
inequality_aux(partition(_Set,DisjSets),GS,ID,OtherID) :- %print(part(Set,DisjSets,GS,ID,OtherID)),nl,
 % detect such things as partition(_,{...ID...},...{...OtherID...}...)
   select(Set1,DisjSets,Rest), set_extension_list(Set1,List1),
   member(O1,List1), gs_identifier(O1,GS,ID),
   select(Set2,Rest,_), set_extension_list(Set2,List2),
   member(O2,List2), gs_identifier(O2,GS,OtherID).

set_extension_list(b(set_extension(List),_,_),List).
gs_identifier(b(identifier(ID),GS,_),GS,ID).

:- assert_must_succeed(( b_global_sets:is_equality_card_global_set(b(equal(b(card(b(identifier('FACES'),set(global('FACES')),[])),integer,[]),b(integer(6),integer,[])),pred,[]),GS,Card),GS='FACES',Card=6)
).
:- assert_must_succeed(( b_global_sets:is_equality_card_global_set(b(equal(b(card(b(value(global_set('TITLE')),set(global('TITLE')),[])),integer,[]),b(integer(4),integer,[])),pred,[]),GS,Card),
 GS=='TITLE', Card=4)
).

is_equality_card_global_set(TE,GlobalSet,Card) :- get_texpr_expr(TE,equal(LHS,RHS)),
   get_texpr_expr(LHS,L), get_texpr_expr(RHS,R),
   ((L=card(C),R=integer(Card)) ; (R=card(C),L=integer(Card))),
   global_set_identifier(C,GlobalSet),
   get_texpr_type(C,set(global(GlobalSet))).
   
global_set_identifier(C,GS) :- get_texpr_expr(C,BE), global_set_identifier2(BE,GS).
global_set_identifier2(identifier(GlobalSet),GlobalSet).
global_set_identifier2(value(global_set(GlobalSet)),GlobalSet). % generated by Z
 
is_greaterequalthan_card_global_set(TE,GlobalSet,Card) :- 
   (get_texpr_expr(TE,greater_equal(LHS,RHS)) ;
    get_texpr_expr(TE,less_equal(RHS,LHS)) ),
   get_texpr_expr(LHS,card(C)), get_texpr_expr(RHS,integer(Card)),
   global_set_identifier(C,GlobalSet),
   get_texpr_type(C,set(global(GlobalSet))).
is_greaterequalthan_card_global_set(TE,GlobalSet,Card1) :- 
   (get_texpr_expr(TE,greater(LHS,RHS)) ;
    get_texpr_expr(TE,less(RHS,LHS)) ),
   get_texpr_expr(LHS,card(C)), get_texpr_expr(RHS,integer(Card)),
   number(Card), Card1 is Card+1,
   global_set_identifier(C,GlobalSet),
   get_texpr_type(C,set(global(GlobalSet))).
is_greaterequalthan_card_global_set(TE,GlobalSet,2) :- 
  %% preferences:get_preference(allow_empty_global_sets,false), preference no longer exists
  /* GlobalSet /= { ANY } -> as GlobalSet cannot be empty & as ANY must be well-typed:
     we need at least one more element of GlobalSet */
   get_texpr_expr(TE,not_equal(LHS,RHS)),
   ( get_texpr_expr(LHS,identifier(GlobalSet))
     -> get_texpr_expr(RHS,set_extension([b(_X,global(GlobalSet),_)]))
     ;  get_texpr_expr(RHS,identifier(GlobalSet)),
        get_texpr_expr(LHS,set_extension([b(_X,global(GlobalSet),_)]))
   ).
   %%print(not_equal(GlobalSet,_X)),nl.
%  not_equal(S_CHEMIN_ID,b(set_extension([b(identifier(C_NULL_CHEMIN_ID),global(S_CHEMIN_ID),[nodeid(pos(35,1,80,27,80,42)),loc(local,emi,concrete_constants)])]),set(global(S_CHEMIN_ID)),[nodeid(pos(34,1,80,26,80,43))]))


:- volatile global_set_user_defined_scope/2.
:- dynamic global_set_user_defined_scope/2.

/* allow to set scope for specific sets, e.g., via command-line */
set_user_defined_scope(GS,X) :- \+ number(X),!,
  add_error(set_user_defined_scope,'Scope must be number: ',GS:X).
set_user_defined_scope(GS,X) :-
  print('% Setting user defined scope: '), print(GS), print(' == '), print(X),nl,
  assert(global_set_user_defined_scope(GS,X)).

get_user_defined_scope(GlobalSetName,LowBound,UpBound) :- 
   (var(GlobalSetName)
    -> add_error(get_user_defined_scope,'Arg is variable: ',get_user_defined_scope) ; true),
   (global_set_user_defined_scope(GlobalSetName,UpBound)
    -> LowBound=1
    ;  extract_setsize_from_defs(GlobalSetName,LowBound,UpBound)).


extract_setsize_from_defs(GlobalSetName,LowBound,UpBound) :-
  b_get_machine_setscope(GlobalSetName,DefTerm),
  get_texpr_expr(DefTerm,DefExpr),
  extract_integer_range(DefExpr,LowBound,UpBound).
  
extract_integer_range(interval(LB,UB), LowBound, UpBound) :-
     get_texpr_expr(LB,integer(LowBound)),  get_texpr_expr(UB,integer(UpBound)).
extract_integer_range(set_extension(CstList),LowBound,UpBound) :-
     LowBound = 1, length(CstList,UpBound). /* TO DO: extract names of constants in list for pretty printing */
extract_integer_range(integer(UpBound),LowBound,UpBound) :-
     LowBound = 1, UpBound>0.
   

assert_extract_setsize_from_machine_cache(GlobalSetName,LowBound,UpBound) :-
  (retract(extract_setsize_from_machine_cache(GlobalSetName,L,U))
    -> print(overriding_set_size(GlobalSetName,LowBound-UpBound,L-U)),nl
    ; true),
  assert(extract_setsize_from_machine_cache(GlobalSetName,LowBound,UpBound)),
  (UpBound<LowBound
   -> print_error('*** illegal empty global set :'),print_error(GlobalSetName)
    ; true
  ).
  
  
  %% b_try_update_fd_cardinality has been removed

:- volatile enumerated_set/1, fixed_deferred_set_size/2.
:- dynamic enumerated_set/1, fixed_deferred_set_size/2.
% true if GS is a deferred set whose size was not fixed; i.e., ProB has not inspected all possible instantiations
unfixed_deferred_set(GS) :- 
   b_global_set(GS), \+ enumerated_set(GS), \+ fixed_deferred_set_size(GS,_).

:- assert_pre(b_global_sets:assert_enumerated_set(_GS,_L,_,_),true).
% TODO(DP, 7.8.2008)	     
%                (atomic(GS),type_check(L,list(xml_identifier)))).
:- assert_post(b_global_sets:assert_enumerated_set(_,_,L,U),(number(L),number(U))).

assert_enumerated_set(GlobalSetName,ListOfConstants,LowBound,UpBound) :-
  assert(enumerated_set(GlobalSetName)),
 % print_message(enum_set(GlobalSetName,ListOfConstants)),
  LowBound = 1,
  length(ListOfConstants,UpBound),
  assert_extract_setsize_from_machine_cache(GlobalSetName,LowBound,UpBound),
  add_named_constants_to_global_set(GlobalSetName,ListOfConstants).

add_named_constants_to_global_set(GlobalSetName,ListOfConstants) :-
  (b_global_constant(GlobalSetName,CurNr,CurCst)
    -> add_error_fail(add_named_constants_to_global_set,'Globalset already has constants: ',
                      b_global_constant(GlobalSetName,CurNr,CurCst))
    ; true),
  nth1(Nr,ListOfConstants,Cst),
  (b_global_constant(GS2,_,Cst)
    -> (GS2=GlobalSetName 
         -> add_error(b_global_sets,'Duplicate element in global set:',(Cst,GS2))
         ;  add_error(b_global_sets,'Element appears in multiple sets:',(Cst,GS2,GlobalSetName))
       )
    ; true
  ),
  assert(b_global_constant(GlobalSetName,Nr,Cst)),fail.
add_named_constants_to_global_set(_,_).


%add_global_set_with_named_constants(GlobalSetName,ListOfConstants) :-
%  ~~pp_mnf(b_global_sets:assert_enumerated_set(GlobalSetName,ListOfConstants,_,_)),
%  add_new_global_set(GlobalSetName).
 

  
:- assert_pre(b_global_sets:b_fd_type(_G,_L,_U),true).
:- assert_post(b_global_sets:b_fd_type(G,L,U),(atomic(G),(integer(L),integer(U)))).

b_fd_type(GlobalSetName,LowBound,UpBound) :-
  b_get_machine_set(GlobalSetName),
  extract_setsize_from_machine(GlobalSetName,LowBound,UpBound).


b_global_set_cardinality('STRING',Card) :- !, Card=inf.
b_global_set_cardinality('NAT',Card) :- !, get_preference(maxint,MaxInt), Card is MaxInt+1. 
b_global_set_cardinality('NATURAL',Card) :- !,  Card=inf. %get_preference(maxint,MaxInt), Card is MaxInt+1. 
b_global_set_cardinality('NAT1',Card) :- !, get_preference(maxint,Card). 
b_global_set_cardinality('NATURAL1',Card) :- !, Card=inf. % get_preference(maxint,Card). 
b_global_set_cardinality('INT',Card) :- !, get_preference(maxint,MaxInt),
   get_preference(minint,MinInt),
   (MaxInt >= MinInt -> Card is MaxInt-MinInt+1
      ; add_error(b_global_set_cardinality,'MININT larger than MAXINT',(MinInt:MaxInt)),
        Card = 0
   ).
b_global_set_cardinality('INTEGER',Card) :- !, Card=inf. % b_global_set_cardinality('INT',Card).
b_global_set_cardinality(GlobalSet,Card) :- %b_global_set(GlobalSet),
  b_fd_type(GlobalSet,Low,Up), Card is 1+ Up - Low.
  
b_fd_card(GlobalSet,Card) :- b_fd_type(GlobalSet,Low,Up), Card is 1+ Up - Low.

b_non_empty_global_set(_). % allow_empty_global_sets no longer exists; all global_sets are non-empty
%b_non_empty_global_set(IntSet) :- b_integer_or_string_set(IntSet).
%b_non_empty_global_set(GlobalSet) :- b_global_set(GlobalSet),
%   b_fd_type(GlobalSet,Low,Up), Up >= Low.

b_empty_global_set(_) :- fail.  % allow_empty_global_sets no longer exists; all global_sets are non-empty
%b_empty_global_set(GlobalSet) :- %b_global_set(GlobalSet),
%   b_fd_type(GlobalSet,Low,Up), Up < Low.
  

b_integer_or_string_set('STRING').
b_integer_or_string_set(X) :- b_integer_set(X).

b_integer_set('NAT').  b_integer_set('NATURAL').
b_integer_set('NAT1').  b_integer_set('NATURAL1').
b_integer_set('INT').  b_integer_set('INTEGER').

% convert a type term into a global_set term
try_b_type2global_set(string,R) :- !, R= global_set('STRING').
try_b_type2global_set(integer,R) :- !, R= global_set('INTEGER').
try_b_type2global_set(global(GS),R) :- b_global_set(GS),!, R=global_set(GS).
try_b_type2global_set(freetype(ID),R) :- !, R=freetype(ID).
try_b_type2global_set(bool,R) :- !,
   %custom_explicit_sets:construct_avl([pred_false  /* bool_false */,pred_true  /* bool_true */],R).
    R = avl_set(node(pred_false,true,1,empty,node(pred_true,true,0,empty,empty))).

b_type2global_set(GS,Res) :- try_b_type2global_set(GS,GR),!, Res=GR.
b_type2global_set(GS,_) :- add_error_and_fail(b_type2global_set,'Type cannot be converted to global_set: ',GS).

%:- use_module(fd_utils).
:- use_module(fd_utils_clpfd).

global_type(fd(X,GlobalSet),GlobalSet) :- %b_global_set(GlobalSet),
   b_fd_type(GlobalSet,Low,Up),
   % print(in_fd(X,Low,Up)),nl,
   in_fd(X,Low,Up).
   
global_type_wf(fd(X,GlobalSet),GlobalSet,WF) :- %b_global_set(GlobalSet),
   b_fd_type(GlobalSet,Low,Up),
   in_fd_wf(X,Low,Up,WF).
     
     
enum_global_type(fd(X,GlobalSet),GlobalSet) :- %b_global_set(GlobalSet),
   (nonvar(GlobalSet) -> 
      (b_fd_type(GlobalSet,Low,Up) -> enum_global_type_aux(X,Low,Up,GlobalSet)
        ; add_internal_error('Illegal value: ',enum_global_type(fd(X,GlobalSet),GlobalSet)),fail)
    ; b_fd_type(GlobalSet,Low,Up),
      enum_global_type_aux(X,Low,Up,GlobalSet)
   ).
   
enum_global_type_aux(X,Low,Up,GlobalSet) :- atomic(X),!,
  ((number(X),X >= Low, X =< Up) -> true
    ; add_internal_error('Illegal value: ', enum_global_type_aux(X,Low,Up,GlobalSet))).
enum_global_type_aux(X,Low,Up,_GlobalSet) :- enum_fd(X,Low,Up).

/* ------------------------------------- */


:- block all_elements_of_type(-,?).
all_elements_of_type(Type,Set) :- % print(all_elements(Type,Set)),nl,
    all_elements_of_type2(Type,Set).
    
:- use_module(kernel_objects,[all_strings/1, enum_warning/3]).
all_elements_of_type2('STRING',Res) :- !, all_strings(Res).
all_elements_of_type2('NAT',Res) :- !, get_preference(maxint,MaxInt),
      gen_int_set(0,MaxInt,Res).
all_elements_of_type2('NATURAL',Res) :- !, get_preference(maxint,MaxInt),
      enum_warning('NATURAL',0:sup,0:MaxInt), %print('### Warning: expanding NATURAL'),nl,
      gen_int_set(0,MaxInt,Res).
all_elements_of_type2('NAT1',Res) :- !, get_preference(maxint,MaxInt),
      gen_int_set(1,MaxInt,Res).
all_elements_of_type2('NATURAL1',Res) :- !, get_preference(maxint,MaxInt),
      enum_warning('NATURAL1',1:sup,1:MaxInt), %print('### Warning: expanding NATURAL1'),nl,
      gen_int_set(1,MaxInt,Res).
all_elements_of_type2('INT',Res) :- !, get_preference(maxint,MaxInt),
      get_preference(minint,MinInt),
      gen_int_set(MinInt,MaxInt,Res).
all_elements_of_type2('INTEGER',Res) :- !, get_preference(maxint,MaxInt),
      get_preference(minint,MinInt),
      enum_warning('INTEGER',inf:sup,MinInt:MaxInt), %print('### Warning: expanding INTEGER'),nl,
      gen_int_set(MinInt,MaxInt,Res).      
all_elements_of_type2(Type,Set) :- b_global_set(Type),
       findall(El,enum_global_type(El,Type),Set).
       
 

gen_int_set(Min,Max,Set) :-
   ((Min>Max) -> (Set = [])
              ;  (Set = [int(Min)|RSet],
                  M1 is Min+1,
                  gen_int_set(M1,Max,RSet))).
