% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(b_state_model_check,
   [b_set_up_valid_state/1,
    b_check_valid_state/1,

    state_model_check/4,
    
    cbc_deadlock_freedom_check/3, get_all_guards_false_pred/1,
    get_unsorted_all_guards_false_pred/1, get_negated_guard/3,

    cbc_static_assertions_check/1,

    set_up_transition/7,
    set_up_initialisation/5,

    negate/2
    ]).



:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,cbc).
:- module_info(description,'This module provides various tools for constraint-based checking of B machines.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(self_check).
:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(store).
:- use_module(translate).
:- use_module(b_enumerate,[  b_tighter_enumerate_values/2]).
:- use_module(b_interpreter,
            [b_test_boolean_expression/4, b_test_boolean_expression/6,
             b_test_boolean_expression_wf/3, b_det_test_boolean_expression/6,
             b_not_test_boolean_expression/4,  b_not_test_boolean_expression/6,
             b_execute_top_level_operation_wf/6,
             b_execute_top_level_statement/7,
             set_up_typed_localstate/5]).
:- use_module(b_interpreter_eventb, [b_event_constraints/7]).
:- use_module(b_global_sets).
:- use_module(bmachine).
:- use_module(bsyntaxtree).

:- use_module(kernel_objects).
:- use_module(bsets_clp).

:- use_module(debug).
:- use_module(typechecker).
:- use_module(bmachine).
:- use_module(kernel_waitflags).
:- use_module(error_manager).
:- use_module('kodkod/kodkod',[apply_kodkod/3,apply_kodkod_to_counterexample_search/5]).

/* --------------------------------------------------------- */


state_model_check(OpName,State,Operation,NewState) :- 
  print_message('===> Starting Constraint-Based Checking'),
  statistics(runtime,[_,_]),
  cbc_model_check(OpName,State,Operation,NewState),
  nl,nl,
  statistics(runtime,[_,T]),
   print_message('===> found counter example'),
     print('% ===> Operation: '), translate_event(Operation,OpStr),print(OpStr),nl,
     print('% ===> State: '),translate_bstate(State,StateStr),print(StateStr),nl,
     print('% ===> NewState: '),translate_bstate(NewState,NStateStr),print(NStateStr),nl,
   print('% RunTime: '), print(T), print(' ms'),nl,nl.

:- use_module(library(avl)).
cbc_model_check(OpName,State,Operation,NewState) :-
   (nonvar(OpName)
     -> (b_get_machine_operation(OpName,_,_,_) 
          -> print_message(operation_name(OpName))
          ;  (add_error(constraint_based_check,'Operation does not exist: ',OpName),fail)
         )
      ; true
   ),
   b_get_properties_from_machine(Properties),
   b_get_invariant_from_machine(Invariant),
   b_get_machine_variables(Variables),
   b_get_machine_constants(Constants),
  % print_message(extracted_invariant(Variables,Invariant)),
   
   empty_state(EmptyState),
   set_up_typed_localstate(Constants,_FreshVars1,TypedVals1,EmptyState,State1),
   set_up_typed_localstate(Variables,_FreshVars2,TypedVals2,State1,State),

   init_wait_flags(WF),
   
   print_message(' --------------- '),
   print_message(' set up start state:'),  print_message(State),
   translate:print_bexpr(Properties),
   empty_avl(Ai),
   b_test_boolean_expression(Properties,[],State,WF,Ai,A2),
    print_message(' set up properties for constants'),
   b_test_boolean_expression(Invariant,[],State,WF,A2,_A3),
    print_message(' set up invariant for start state'),

   set_up_transition(OpName,Operation,State1,State,NewState,_TransInfo,WF),
   print_message(' set up target state for operation:'),  print_message(OpName),
   print_message(' target state:'), print_message(NewState),
   
   % Retrieve Information on proof status of the invariant
 
   (get_proven_invariant(OpName,ProvenInvariant) -> true
    ;   create_texpr(truth,pred,[],ProvenInvariant), print(no_proven_invariant_info),nl), 
   (b_specialized_invariant_for_op(OpName,UncheckedInvariant)
     -> true ; UncheckedInvariant=Invariant, print(no_relevant_invariant_info),nl),

   %print_message('proven invariant:'),
   %print(proven(ProvenInvariant)),
   translate:print_bexpr(ProvenInvariant),nl,
   b_test_boolean_expression(ProvenInvariant,[],NewState,WF,A2,A4),  % not A3 but A2; as different State
    print_message(' finished set up invariant of proven invariants for target state'),
    
   translate:print_bexpr(UncheckedInvariant),nl,
   b_not_test_boolean_expression(UncheckedInvariant,[],NewState,WF,A4,_Ao),
    print_message(' finished set up negation of invariant for target state'),
   
  %  print_message(' Enumerated Start State: '), print_message(State),
    
   
   %b_execute_operation_wo_normalise(OpName,Operation,State,NewState),
  % print(enum1(TypedVals1)),nl,
   b_tighter_enumerate_values(TypedVals1,WF),
  % print(enum2(TypedVals2)),nl,
   b_tighter_enumerate_values(TypedVals2,WF),  
   
   (debug_mode(on) -> portray_waitflags(WF) ; true),
   print_bt_message(grounding_wait_flags),
   ground_wait_flags(WF).

set_up_transition(OpName,Operation,ConstantsState,State,NewState,TransInfo,WF) :-
    is_event_b_transition(OpName),!,
    b_get_machine_variables(Variables),
    create_target_state(Variables,_Values,ConstantsState,NewState,WF),
    copy_unmodified_variables(Variables,OpName,State,NewState),
    event_b_transition(OpName,Operation,State,NewState,TransInfo,WF).
set_up_transition(OpName,Operation,ConstantsState,State,NewState,TransInfo,WF) :-
    % classical B
    append(ConstantsState,State,InState),
    append(ConstantsState,NewState,OutState),
    b_get_machine_variables(Variables),
    set_up_target_store(InState,OutState),
    copy_unmodified_variables(Variables,OpName,InState,OutState),
    b_execute_top_level_operation_wf(OpName,Operation,InState,OutState,TransInfo,WF).
is_event_b_transition(OpName) :-
    b_get_machine_operation(OpName,[],_Parameters,_TEvent,eventb(_ChangeSet,_Values,_Operation)),!.
event_b_transition(OpName,Operation,State,NewState,TransInfo,WF) :-
    b_get_machine_operation(OpName,[],Parameters,TEvent,eventb(ChangeSet,Values,Operation)),
    copy_variable_store(State,ChangeSet,_,NewState),
    b_event_constraints(TEvent,Parameters,Values,State,NewState,TransInfo,WF).

create_target_state(Variables,Values,ConstantsState,NewState,WF) :-
    set_up_typed_localstate(Variables,Values,TypedVals,ConstantsState,NewState),
    b_tighter_enumerate_values(TypedVals,WF).

% TODO[DP, 7.4.2011]: refactor, see b_interpreter:b_execute_operation3/13
set_up_initialisation(ConstantsState,NewState,Operation,TransInfo,WF) :-
    b_get_machine_variables(Variables),
    create_target_state(Variables,Values,ConstantsState,NewState,WF),
    store:l_expand_and_normalise_values(Values,NormalisedValues),
    safe_univ(Operation,['$initialise_machine'|NormalisedValues]),
    b_get_initialisation_from_machine(Body,OType),
    set_up_initialisation2(OType,Body,Variables,ConstantsState,NewState,TransInfo,WF).
set_up_initialisation2(classic,Body,Variables,ConstantsState,NewState,[path(Path)],WF) :-
    set_up_undefined_localstate(Variables,ConstantsState,PreInitState),
    empty_state(EmptyState),
    b_execute_top_level_statement(Body,EmptyState,PreInitState,Updates,WF,Path,output_required),
    save_updates_in_existing_store(Updates,NewState).
set_up_initialisation2(eventb(_,_,_),Body,_Variables,ConstantsState,NewState,TransInfo,WF) :-
    b_event_constraints(Body,[],_Values,ConstantsState,NewState,TransInfo,WF).

set_up_target_store([],[]).
set_up_target_store([bind(V,_)|In],[bind(V,_)|Out]) :-
    set_up_target_store(In,Out).
    
% copy over the variables that are not changed by the operation from OldStore to NewStore
copy_unmodified_variables(Variables,OpName,OldStore,NewStore) :-
    ( get_operation_info(OpName,SubstitutionInfo) ->
        ( memberchk(modifies(ModIDs),SubstitutionInfo) ->
            get_texpr_ids(Variables,Ids),
            exclude(is_modified_var(ModIDs),Ids,UnmodifiedIds),
            maplist(copy_var(OldStore,NewStore),UnmodifiedIds)
        ; otherwise ->
            add_internal_error('Operation/Event has not modifies info: ',
                               copy_unmodified_variables(OpName)),
            fail)
    ; otherwise ->
        add_internal_error('Illegal Operation/Event:', copy_unmodified_variables(OpName)),
        fail
    ).
is_modified_var(ModIDs,Id) :-
    memberchk(Id,ModIDs).
copy_var(OldStore,NewStore,Id) :- 
    lookup_value(Id,OldStore,V),lookup_value(Id,NewStore,VNew),
    equal_object(V,VNew).

/* --------------------------------------------------------- */

   
b_set_up_valid_state(State) :-
   init_wait_flags(WF),
   b_set_up_valid_state(State,WF),
   %print(grounding(WF)),nl,
   ground_wait_flags(WF).
   
b_set_up_valid_state(State,WF) :-
   b_get_properties_from_machine(Properties),
   b_get_invariant_from_machine(Invariant),
   b_get_machine_variables(Variables),
   b_get_machine_constants(Constants),
   
   empty_state(EmptyState),
   set_up_typed_localstate(Constants,_FreshVars1,TypedVals1,EmptyState,State1), 
   set_up_typed_localstate(Variables,_FreshVars2,TypedVals2,State1,State),
   
   %translate:print_bexpr(Properties),nl,
  % b_interpreter:b_tracetest_boolean_expression(Properties,[],State,WF,one), % should be det; but is not for CrCtl_Comb1_Final ! Investigate
   empty_avl(Ai),
   b_det_test_boolean_expression(Properties,[],State,WF,Ai,Aii),
   %print(state(State)),nl,
    print_message(' set up properties for constants'),
   b_test_boolean_expression(Invariant,[],State,WF,Aii,_Ao),
    print_message(' set up invariant for start state'),
   
   b_tighter_enumerate_values(TypedVals1,WF),
   b_tighter_enumerate_values(TypedVals2,WF).

/* ---------------------- */
/* DEADLOCK FREEDOM CHECK */
/* ---------------------- */

 :- use_module(bsyntaxtree,[predicate_identifiers/2]).
:- use_module(partition_detection,[detect_all_partitions_in_predicate/2]).
 
cbc_deadlock_freedom_check(State,Goal,Filter) :-
    print(cbc_deadlock_freedom_check(State,Filter)),nl,flush_output,
    ( is_truth(Goal) ->
        get_all_guards_false_pred(Target)
    ; otherwise ->
        print('Adding additional goal predicate: '),nl,
        print_bexpr(Goal),nl,
        ( Filter=1 ->
            get_all_guards_false_pred(AllGFalse,Goal)
        ; otherwise ->
            get_all_guards_false_pred(AllGFalse)),
        conjunct_predicates([AllGFalse,Goal],Target)
    ),!,flush_output,
    statistics(runtime,[Start,_]),
    statistics(walltime,[WStart,_]),
    cbc_deadlock_freedom_check2(Target,RState,Start,WStart,false,TO), % TO DO: control SkipIrrelevantComponents by preference or GUI
    (TO==time_out -> State=time_out ; State=RState).
cbc_deadlock_freedom_check2(Target,State,Start,WStart,SkipIrrelevantComponents,TO) :-
   get_texpr_expr(Target,T), T \= falsity,
   b_get_properties_from_machine(Properties),
   b_get_invariant_from_machine(Invariant),
   conjunct_predicates([Properties,Invariant,Target],Pred1),
   b_get_machine_variables(Variables),
   b_get_machine_constants(Constants), append(Constants,Variables,CV),

   apply_kodkod(CV,Pred1,Pred2),
   detect_all_partitions_in_predicate(Pred2,Pred), % detect set partitions
   
   set_up_typed_localstate_for_pred(CV,Pred,TypedVals,State),
   b_interpreter_components:reset_component_info(false),
   (debug_mode(on) -> translate:nested_print_bexpr(Pred) ; true),
   (SkipIrrelevantComponents==true
     -> predicate_identifiers(Target,TargetIds)  % Warning: can skip over obvious inconsistencies TRUE:BOOL
     ; TargetIds=all),
   print(target(TargetIds)),nl,
   time_out_with_factor_call(
        b_interpreter_components:b_trace_test_components(Pred,State,TypedVals,TargetIds),
        10, (nl,TO=time_out)),
   statistics(runtime,[Stop,_]), Time is Stop - Start,
   statistics(walltime,[WStop,_]), WTime is WStop - WStart,
   (TO==time_out -> print(time_out_during_deadlock_checking(Time)),nl
     ; TO=no_time_out, print(deadlock_counter_example_found(Time,WTime)),nl).
cbc_deadlock_freedom_check2(Target,_State,Start,WStart,_,no_time_out) :-
   (get_texpr_expr(Target,falsity) -> print('Disjunction of guards obviously true'),nl ; true),
   statistics(runtime,[Stop,_]), Time is Stop - Start,
   statistics(walltime,[WStop,_]), WTime is WStop - WStart,
   print(no_deadlock_counter_example_found(Time,WTime)),nl,fail.

set_up_typed_localstate_for_pred(CV,Pred,TypedVals,State) :-
   (preferences:preference(use_static_ordering,true)
     -> static_ordering:sort_ids_by_usage(CV,Pred,SortedVarlist,no_warnings)
        %,reverse(Constants1,Constants)
        %,print(sorted_by_usage(Constants)),nl
     ;  SortedVarlist =  CV),
   set_up_typed_localstate(SortedVarlist,_FreshVars,TypedVals,[],State).

% test if a predicate can be satisfied for some instantiation of constants and variables
predicate_satisfiable(P) :- predicate_satisfiable(P,_),!, print('SATISFIABLE'),nl,nl.
predicate_satisfiable(_) :- print('*UNSATISFIABLE*'),nl,nl,fail.
predicate_satisfiable(Predicate,State) :- 
   b_get_properties_from_machine(Properties),
   b_get_invariant_from_machine(Invariant),
   conjunct_predicates([Properties,Invariant,Predicate],TotPred),
   b_get_machine_variables(Variables),
   b_get_machine_constants(Constants), append(Constants,Variables,CV),
   set_up_typed_localstate_for_pred(CV,TotPred,TypedVals,State),
   b_interpreter_components:reset_component_info(false),
   translate:nested_print_bexpr(Predicate),
   % TO DO: provide option to only check components which relate directly to Predicate predicate_identifiers(Target,TargetIds)?
   predicate_identifiers(Predicate,RelevantIds),
   b_interpreter_components:b_trace_test_components(TotPred,State,TypedVals,RelevantIds).

   
% set up a predicate which is true if all guards are false
get_all_guards_false_pred(AllGuardsFalsePred) :-
   findall(NegGuard,
          (b_get_machine_operation(OpName,_,_,_),
           get_negated_guard(OpName,_,NegGuard)), ListOfNegGuards),
   ~~mnf(conjunct_and_sort_smt_predicates(ListOfNegGuards,AllGuardsFalsePred)).
get_all_guards_false_pred(AllGuardsFalsePred,FilterPred) :-
   findall(NegGuard,
      (b_get_machine_operation(OpName,_,_,_),
       get_negated_guard(OpName,Guard,NegGuard),
       conjunct_predicates([Guard,FilterPred],FGuard),
       nl,print('---> CHECKING IF GUARD SATISFIABLE: '), print(OpName),nl,
       predicate_satisfiable(FGuard) % if Guard not satisfiable given FilterPred: remove it
       ), ListOfNegGuards),
   ~~mnf(conjunct_and_sort_smt_predicates(ListOfNegGuards,AllGuardsFalsePred)).

% the following is useful, e.g., for graphical visualization
get_unsorted_all_guards_false_pred(AllGuardsFalsePred) :-
   findall(NegGuard,
          (b_get_machine_operation(OpName,_,_,_),
           get_negated_guard(OpName,_,NegGuard)), ListOfNegGuards),
   conjunct_predicates(ListOfNegGuards,AllGuardsFalsePred).
   
conjunct_and_sort_smt_predicates(List,Result) :-
   print('Counting basic predicates'),nl,
   empty_avl(Ai), count(List,2,Ai,Ao),
   % portray_avl(Ao),nl,
   sort_smt_predicates(List,Ao,SList),
   %nl,print(sorted(SList)),nl,
   conjunct_smt_predicates(SList,Result) , print(finished),nl.

conjunct_smt_predicates([],b(truth,pred,[])).
conjunct_smt_predicates([P|T],Res) :- conjunct_smt_predicates(T,TRes),
   conjoin(P,TRes,[try_smt],Res).

% Count the number of occurences of given basic predicates
count(b(Pred,pred,_Info),Rel,Ai,Ao) :- !,count2(Pred,Rel,Ai,Ao).
count([],_,Ai,Ao) :- !,Ao=Ai.
count([H|T],Rel,Ai,Ao) :- !,count(H,Rel,Ai,Aii), count(T,Rel,Aii,Ao).
count(R,_,_,_) :- add_error_fail(count,'Illegal argument to count: ',R).

count2(negation(Pred),Rel,Ai,Ao) :- !, count(Pred,Rel,Ai,Ao).
count2(conjunct(A,B),Rel,Ai,Ao) :- !, count(A,Rel,Ai,Aii), count(B,Rel,Aii,Ao).
count2(Pred,Rel,Ai,Ao) :- disjunctive_pred(Pred,A,B),!,
  (Rel>1
    -> R1 is Rel-1, count(A,R1,Ai,Aii), count(B,R1,Aii,Ao)
    ;  Ai=Ao).
count2(Pred,Rel,Ai,Ao) :- smt_predicate(Pred),!, norm_pred(Pred,NP),
  (avl_fetch(NP,Ai,C1) -> Count is C1+Rel ; Count = Rel),
  avl_store(NP,Ai,Count,Ao).
count2(_Pred,_Rel,A,A).

disjunctive_pred(disjunct(A,B),A,B).
disjunctive_pred(implication(A,B),A,B).
disjunctive_pred(equivalence(A,B),A,B).


sort_smt_predicates(b(Pred,pred,Info),Ai,b(SP,pred,Info)) :- sort_smt_predicates2(Pred,Ai,SP).
sort_smt_predicates([],_,[]).
sort_smt_predicates([H|T],Ai,[SH|ST]) :- sort_smt_predicates(H,Ai,SH), sort_smt_predicates(T,Ai,ST).

sort_smt_predicates2(negation(Pred),AVL,negation(SPred)) :- !,sort_smt_predicates(Pred,AVL,SPred).
sort_smt_predicates2(implication(A,B),AVL,implication(SA,SB)) :- !,
   sort_smt_predicates(A,AVL,SA), sort_smt_predicates(B,AVL,SB).
sort_smt_predicates2(equivalence(A,B),AVL,equivalence(SA,SB)) :- !,
   sort_smt_predicates(A,AVL,SA), sort_smt_predicates(B,AVL,SB).
sort_smt_predicates2(disjunct(A,B),AVL,disjunct(SA,SB)) :- !,  % To do: also sort disjuncts
   sort_smt_predicates(A,AVL,SA), sort_smt_predicates(B,AVL,SB).
sort_smt_predicates2(conjunct(A,B),AVL,Res) :- !,
   conjunction_to_count_list(b(conjunct(A,B),pred,[]),AVL,CountList),
   sort(CountList,SC),
   reverse(SC,SortedCList), % to do : use samsort with custom order
   %print(sorted(SortedCList)),nl,
   project_count_list(SortedCList,PR),
   conjunct_smt_predicates(PR,BRes),
   translate:print_bexpr(BRes),nl,
   BRes = b(Res,pred,_).
sort_smt_predicates2(X,_,X).

project_count_list([],[]).
project_count_list([count(_,P)|T],[P|PT]) :- project_count_list(T,PT).

conjunction_to_count_list(C,AVL,List) :- is_a_conjunct(C,LHS,RHS),!,
   conjunction_to_count_list(LHS,AVL,L1),
   conjunction_to_count_list(RHS,AVL,R1),
   append(L1,R1,List).  % TO DO: improve, use Difference Lists
conjunction_to_count_list(b(Pred,pred,Info),AVL,[count(Count,b(Pred,pred,Info))]) :- smt_predicate(Pred),!,
  norm_pred(Pred,NX),
  (avl_fetch(NX,AVL,Count) -> true ; Count=0).
conjunction_to_count_list(P,_AVL,[count(0,P)]).

smt_predicate(equal(_,_)).
smt_predicate(not_equal(_,_)).
smt_predicate(member(_,_)).
smt_predicate(not_member(_,_)).
smt_predicate(less(_,_)). 
smt_predicate(less_equal(_,_)).
smt_predicate(greater(_,_)).
smt_predicate(greater_equal(_,_)).

norm_pred(equal(A,B),Res) :- !,norm_equal(A,B,Res).
norm_pred(not_equal(A,B),Res) :- !,norm_equal(A,B,Res).
norm_pred(member(A,B),member(AA,BB)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(not_member(A,B),member(AA,BB)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(less(A,B),less(AA,BB)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(greater(A,B),less(BB,AA)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(less_equal(A,B),less(BB,AA)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(greater_equal(A,B),less(AA,BB)) :- !,norm_pred(A,AA),norm_pred(B,BB).
norm_pred(b(B,_,_),Res) :- !, norm_pred(B,Res).
norm_pred(set_extension(L),Res) :- !, Res=set_extension(NL), norm_pred(L,NL).
norm_pred(integer(A),Res) :- !,Res=A.
norm_pred(identifier(A),Res) :- !,Res=A.
norm_pred([],Res) :- !, Res=[].
norm_pred([H|T],Res) :- !, Res=[NH|NT], norm_pred(H,NH), norm_pred(T,NT).
norm_pred(X,X).


norm_equal(b(identifier(ID),_,_),_B,Res) :- !,Res=ID.
norm_equal(A,B,Res) :- norm_pred(A,AA),norm_pred(B,BB),
  (AA==boolean_false -> AAA=boolean_true ; AAA=AA),
  (BB==boolean_false -> BBB=boolean_true ; BBB=BB),
  (BBB @< AAA -> Res = equal(AAA,BBB) ; Res= equal(BBB,AAA)).
/*
operation_satisfiable(OpName,FilterPred) :- get_machine_operation(OpName,_,_,_),
       get_negated_guard(OpName,Guard,_NegGuard),
       conjunct_predicates([Guard,FilterPred],FGuard),
       print('CHECKING IF GUARD SATISFIABLE: '), print(OpName),nl,
       predicate_satisfiable(FGuard).*/
   
% Optionally: Remove certain complicated parts from the guards
simplify_guard(b(Pred,pred,Info),SPred) :- simplify_guard2(Pred,Info,SPred).
simplify_guard2(conjunct(A,B),Info,Res) :- !,
    simplify_guard(A,SA), simplify_guard(B,SB),
    conjoin(SA,SB,Info,Res).
simplify_guard2(exists(P,B),_Info,R) :- !, R=b(truth,pred,[]),
     print('REMOVED: '),translate:print_bexpr(b(exists(P,B),pred,[])),nl.
simplify_guard2(X,Info,b(X,pred,Info)).
      
conjoin(b(truth,_,_),B,_,R) :- !,R=B.
conjoin(b(falsity,_,_),_B,_,R) :- !,R=b(falsity,pred,[]).
conjoin(A,b(truth,_,_),_,R) :- !,R=A.
conjoin(_A,b(falsity,_,_),_,R) :- !,R=b(falsity,pred,[]).
conjoin(A,B,Info,b(conjunct(A,B),pred,Info)).


get_negated_guard(OpName,Guard,NegationOfGuard) :-
    get_guard(OpName,Guard), 
    SGuard=Guard,   
    %simplify_guard(Guard,SGuard), %% comment in to remove complicated parts
	negate(SGuard,NegationOfGuard).

negate(b(truth,_,_),R) :- !, create_texpr(falsity,pred,[],R).
negate(SGuard,R) :- !, create_texpr(negation(SGuard),pred,[try_smt],R).

:- use_module(predicate_debugger,[get_operation_enabling_condition/6]).
%:- use_module(b_ast_cleanup, [clean_up/3]).
:- use_module(b_interpreter_components,[construct_optimized_exists/3]).

% get a guard, by translating all parameters into existential quantifiers
get_guard(OpName,Guard) :-
  get_operation_enabling_condition(OpName,Parameters,EnablingCondition,_BecomesSuchVars,false,true),
  % TO DO: partition; avoid lifting becomes_such_that conditions 
  %   (in EventB the feasibility PO will ensure that Guard => BecomeSuchThat is ok)
  construct_optimized_exists(Parameters,EnablingCondition,Guard).

/* --------------------------------------------------------- */


b_check_valid_state(State) :-
   (b_valid_state(State) -> true
     ; (print_error('Invalid State:'),
        print_error(State))
   ).

b_valid_state(State) :-
   ~~mnf(bmachine:b_get_machine_variables(Variables)),
  ~~pp_cll(b_state_model_check:b_check_variable_types(Variables,State)).


% TODO(DP,6.5.2008): removed type check
:- assert_pre(b_state_model_check:b_check_variable_types(V,S), (list_skeleton(V),list_skeleton(S))).
%              (list_skeleton(V),list_skeleton(VT),list_skeleton(S))).
	      
:- assert_post(b_state_model_check:b_check_variable_types(_V,_S),true).
  
b_check_variable_types([],_).
b_check_variable_types([Var|VT],State) :-
   debug_println(9,checking(Var,Type)),
   def_get_texpr_id(Var,VarID), get_texpr_type(Var,Type),
   ((~~pp_mnf(store:lookup_value_for_existing_id(VarID,State,Val)),Val\==fail)
    -> ((~~pp_cll(kernel_objects:basic_type(Val,Type))
          -> debug_println(check_basic_type(Val,Type))
          ;  print_error('Type Error ! Variable,Value,Type: '),print_error(VarID),
             print_error(Val),print_error(Type)
        ),
        b_check_variable_types(VT,State))
    ;   add_error(b_check_variable_types,'Variable not defined in state: ', VarID),
        add_error(b_check_variable_types,'Be sure that INITIALISATION initialises the variable: ', VarID),
        print_error(State)
   ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% constraint based assertion check
:- use_module(preferences).

cbc_static_assertions_check(Result) :-
    ( time_out_with_factor_call(cbc_static_assertions_check2(State),10,TO=time_out) ->
        ( TO==time_out -> Result = time_out
        ; otherwise    -> Result = counterexample_found(State))
    ; otherwise ->
        Result = no_counterexample_found).


cbc_static_assertions_check2(concrete_constants(NormalisedState)) :-
    b_get_properties_from_machine(Properties),
    b_get_machine_constants(Constants),
    b_get_static_assertions_from_machine(Assertions),
   
    empty_state(EmptyState),
    set_up_typed_localstate(Constants,_FreshVars1,TypedVals,EmptyState,State),
    init_wait_flags(WF),
    cbc_static_assertions_check3(Constants,Properties,Assertions,State,WF),
    b_tighter_enumerate_values(TypedVals,WF),
   
    print_bt_message(grounding_wait_flags),
    ground_wait_flags(WF),

    normalise_store(State,NormalisedState).

cbc_static_assertions_check3(Constants,Properties,Assertions,State,WF) :-
    conjunct_predicates(Assertions,Assertion),
    apply_kodkod_to_counterexample_search(Constants,Properties,Assertion,Kodkod,Rest),!,
    print_message(' setting up Kodkod part '),
    b_test_boolean_expression(Kodkod,[],State,WF),
    print_message(' setting up not translated part '),
    b_test_boolean_expression(Rest,[],State,WF).

cbc_static_assertions_check3(_Constants,Properties,Assertions,State,WF) :-
 % TO DO: find a way to apply detect_all_partitions_in_predicate
    empty_avl(Ai),
    b_test_boolean_expression(Properties,[],State,WF,Ai,A1),
    print_message(' finished set properties'),

    % TO DO: filter out proven Assertions ; for this the ProB Rodin translator needs to transmit this information

    member(Assertion,Assertions),  %% comment in if you want to check assertions one-by-one
    %conjunct_predicates(Assertions,Assertion),  %% here we check all assertions in one go; can be more efficient if there is overlap, or some later assertions are easy to disprove
    ((get_preference(use_po,true),bmachine:is_discharged_assertion(Assertion))
      -> %debug_mode(on),
         print('Already proven: '),translate:print_bexpr(Assertion),nl,fail
       ; true), 
    
    print_message(' set up negation of assertion'),
    (debug_mode(on) -> translate:print_bexpr(Assertion),nl ; true), %%
    
    b_not_test_boolean_expression(Assertion,[],State,WF,A1,_Ao).
