% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% multi-level variant of bmachine_eventb
% this module should replace bmachine_eventb in the near future

:- module(bmachine_eventb,[check_event_b_project/4,prime_variable/2,prime_variables/2]).

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(avl)).

:- use_module(error_manager).
:- use_module(tools,[remove_all/3,foldl/4,foldl/5,ajoin_with_sep/3]).
:- use_module(self_check).
:- use_module(btypechecker).
:- use_module(b_ast_cleanup).
:- use_module(bsyntaxtree).
:- use_module(debug).
:- use_module(bmachine_construction).
:- use_module(preferences,[get_preference/2]).
:- use_module(input_syntax_tree,[extract_raw_identifiers/2]).
:- use_module(translate,[translate_bexpression/2]).
:- use_module(pragmas,[extract_pragmas_from_event_b_extensions/2,apply_pragmas_to_event_b_machine/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module contains the rules for loading a Event-B machine or context. ').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').


check_event_b_project([],RawContextes,Extensions,MachineWithPragmas) :-
    !,
    check_extensions(Extensions,Freetypes,InitEnv),
    check_contextes(InitEnv,RawContextes,Contextes),
    convert_event_b(context,Freetypes,Contextes,[],[],Machine),
    extract_pragmas_from_event_b_extensions(Extensions,Pragmas),
    apply_pragmas_to_event_b_machine(Pragmas,Machine,MachineWithPragmas).
check_event_b_project(RawModels,RawContextes,Extensions,MachineWithPragmas) :-
    compute_animation_chain(RawModels,AnimationChain),
    check_event_b_project2(RawModels,RawContextes,Extensions,
                           AnimationChain,Machine),
    extract_pragmas_from_event_b_extensions(Extensions,Pragmas),
    apply_pragmas_to_event_b_machine(Pragmas,Machine,MachineWithPragmas).

check_event_b_project2(RawModels,RawContextes,Extensions,AnimationChain,Machine) :-
    RawModels=[event_b_model(_,Main,_)|_],
    check_extensions(Extensions,Freetypes,InitEnv),
    % generate type-checked contextes
    check_contextes(InitEnv,RawContextes,Contextes),
    % generate type-checked models
    check_models(InitEnv,RawModels,Contextes,Models),
    % merge all the constants/vars/predicates and events into one machine
    convert_event_b(Main,Freetypes,Contextes,Models,AnimationChain,Machine).

% the filter chain contains the name of all models that
% should be contained in the animation
compute_animation_chain(Models,AnimationChain) :-
    get_preference(number_animated_abstractions,N),
    debug_println(9,compute_animation_chain(N)),
    length(Models,L),
    N1 is min(N+1,L),
    prefix_length(Models, AnimatedModels, N1),
    extract_model_names(AnimatedModels,AnimationChain).
extract_model_names([],[]).
extract_model_names([event_b_model(_,Name,_)|MRest],[Name|NRest]) :-
    extract_model_names(MRest,NRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% context

check_contextes(InitEnv,RawContextes,Contextes) :-
    debug_println(9,'Checking Contextes'),
    % make sure that a extended context is treated before
    % the extending context
    topological_sort(RawContextes,SortedContextes),
    check_contextes2(SortedContextes,InitEnv,[],Contextes).
check_contextes2([],_,_,[]).
check_contextes2([RawContext|Rrest],InitEnv,PrevContextes,[Context|Crest]) :-
    check_context(RawContext,InitEnv,PrevContextes,Context),
    check_contextes2(Rrest,InitEnv,[Context|PrevContextes],Crest).
    
check_context(RawContext,InitEnv,PrevContextes,Context) :-
    extract_ctx_sections(RawContext,Name,Extends,
                         RawSets,RawConstants,RawAxioms,RawTheorems),
    Context = context(Name,Extends,Sets,Constants,Axioms,Theorems),

    % create deferred sets
    create_sets(RawSets,Name,Sets),
    % and constants
    create_identifiers(RawConstants,[section(Name)],Constants),

    % with the constants and set of seen contextes,
    create_ctx_env(Extends,PrevContextes,InitEnv,PrevEnv),
    % and the sets and constants
    add_unique_variables(Sets,PrevEnv,SetEnv),
    add_unique_variables(Constants,SetEnv,Env),

    % now typecheck the axioms and theorems,
    % this should determine the constant's types
    typecheck_l(RawAxioms,Env,pred,Axioms),
    typecheck_l(RawTheorems,Env,pred,Theorems).

extract_ctx_sections(RawContext,Name,Extends,RawSets,RawConstants,RawAxioms,RawTheorems) :-
    RawContext = event_b_context(_Pos,Name,Sections),
    optional_rawmachine_section(extends,Sections,[],Extends),
    optional_rawmachine_section(sets,Sections,[],RawSets),
    optional_rawmachine_section(constants,Sections,[],RawConstants),
    optional_rawmachine_section(axioms,Sections,[],RawAxioms),
    optional_rawmachine_section(theorems,Sections,[],RawTheorems).


create_sets([],_,[]).
create_sets([RawSet|RSrest],CtxName,[Set|Srest]) :-
    ext2int(RawSet,deferred_set(SetName),_,set(global(SetName)),
            identifier(SetName),[section(CtxName),given_set],Set),
    create_sets(RSrest,CtxName,Srest).

create_identifiers([],_,[]).
create_identifiers([RawIdentifier|RIrest],Infos,[Identifier|Irest]) :-
    ext2int(RawIdentifier,I,_,_,I,Infos,Identifier),
    I=identifier(_),
    create_identifiers(RIrest,Infos,Irest).

create_ctx_env(Extends, Contextes, Old, New) :-
    % find all seen contextes (and their extended contextes,
    % and again their extended contextes, ...)
    transitive_contextes(Extends,Contextes,TransExt),
    % and all variables of extended machines
    add_env_vars_for_extended2(TransExt,Contextes,Old,New).
transitive_contextes(Extends,Contextes,All) :-
    findall(E, transext(Extends,Contextes,E), All1), sort(All1, All).
transext(Extends,_,E) :- member(E,Extends).
transext(Extends,Contextes,E) :-
    member(Name,Extends),
    internal_context_extends(Context,Name,NewExtends),
    member(Context,Contextes),
    transext(NewExtends,Contextes,E).
% add_env_vars_for_extended2(Names,Contextes,OldEnv,NewEnv):
%  Names: a list of the names of the contextes whose constants and sets
%     should be added to the type environment
%  Contextes: a list of all existing contextes
%  OldEnv: the input environment
%  NewEnv: the input environment plus the constants and sets
add_env_vars_for_extended2([],_,Env,Env).
add_env_vars_for_extended2([Name|Erest],Contextes,Old,New) :-
    internal_context_sets(Context,Name,Sets),
    internal_context_constants(Context,Name,Constants),
    member(Context,Contextes),!,
    add_unique_variables(Sets,Old,Env1),
    add_unique_variables(Constants,Env1,Env),
    add_env_vars_for_extended2(Erest,Contextes,Env,New).

% copy&paste from bmachine_construction
add_unique_variables([],Env,Env).
add_unique_variables([Var|Rest],Old,New) :-
    def_get_texpr_id(Var,Name),
    get_texpr_type(Var,Type),
    get_texpr_info(Var,Infos),
    ( env_lookup_type(Name,Old,_) ->
        add_error(bmachine_eventb,'Identifier declared twice',Name),
        fail
    ; otherwise -> true),
    env_store(Name,Type,Infos,Old,Env),
    add_unique_variables(Rest,Env,New).

% add the primed (') versions of the variables to the
% type environment
add_primed_variables(Variables,OrigEnv,NewEnv) :-
    prime_variables(Variables,Primed),
    add_unique_variables(Primed,OrigEnv,NewEnv).
% add a prime (') to each typed identifier
prime_variables([],[]).
prime_variables([V|Vrest],[P|Prest]) :-
    prime_variable(V,P),
    prime_variables(Vrest,Prest).
% add a prime (') to a typed identifier
prime_variable(V,P) :-
    def_get_texpr_id(V,Id),
    prime_id(Id,PId),
    rename_bt(V,[rename(Id,PId)],P).
% add (or remove) a prime (') to a plain identifier (without syntax structure)
prime_id(UnPrimed,Primed) :-
    atom_concat(UnPrimed,'\'',Primed).

check_for_errors(Errors) :-
    ( Errors = [] -> true
    ; otherwise ->
        add_all_perrors(Errors),fail).

typecheck_l(Untyped,Env,Type,Typed) :-
    all_same(Untyped,Type,Types),
    btype_ground_dl(Untyped,Env,[],Types,Typed1,Errors,[]),
    check_for_errors(Errors),
    clean_up_l(Typed1,[],Typed),!.
typecheck_l(Untyped,Env,Type,Typed) :- 
     add_failed_call_error(typecheck_l(Untyped,Env,Type,Typed)),fail.
all_same([],_,[]).
all_same([_|R],E,[E|T]) :- all_same(R,E,T).

typecheck(Untyped,Env,Type,Typed) :-
    typecheck_with_freevars(Untyped,Env,[],Type,Typed).
typecheck_with_freevars(Untyped,Env,Freevars,Type,Typed) :-
    btype_ground_dl([Untyped],Env,Freevars,[Type],[Typed1],Errors,[]),
    clean_up(Typed1,Freevars,Typed),
    check_for_errors(Errors).





%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% models

check_models(InitEnv,RawModels,Contextes,Models) :- 
    debug_println(9,'CHECKING MODELS'),
    topological_sort(RawModels,SortedModels),
    check_models2(SortedModels,InitEnv,[],Contextes,Models).

check_models2([],_,_,_,[]).
check_models2([RawModel|RMrest],InitEnv,PrevModels,Contextes,[Model|Mrest]) :-
    (check_model(RawModel,InitEnv,PrevModels,Contextes,Model) -> true
      ; add_failed_call_error(check_model(RawModel,InitEnv,PrevModels,Contextes,Model)),
        fail),
    check_models2(RMrest,InitEnv,[Model|PrevModels],Contextes,Mrest).

check_model(RawModel,InitEnv,PrevModels,Contextes,Model) :- 
    extract_model_sections(RawModel,Name,Refines,Sees,
                           RawVariables,RawInvariant,
                           RawTheorems,RawVariant,RawEvents),
    Model = model(Name,Refines,Sees,ConcreteVariables,NewVariables,Invariant,Theorems,Events),
    get_refined_model(Refines,PrevModels,RefinedModel),

    get_all_abstract_variables(Refines,PrevModels,AbstractionVariables),
    length(PrevModels,Level),
    create_identifiers(RawVariables,[section(Name),level(Level)],ConcreteVariables),
    split_abstract_vars(AbstractionVariables,ConcreteVariables,
                        NewVariables,DroppedAbstractVars),

    create_ctx_env(Sees,Contextes,InitEnv,CtxEnv),
    add_unique_variables(ConcreteVariables,CtxEnv,ModelEnv),
    add_unique_variables(DroppedAbstractVars,ModelEnv,InvEnv),
    % add primed versions of the concrete variables to
    % permit before-after predicates in the witnesses
    add_primed_variables(ConcreteVariables,InvEnv,BeforeAfterEnv),

    typecheck_l(RawInvariant,InvEnv,pred,Invariant),
    typecheck_l(RawTheorems,InvEnv,pred,Theorems),

    typecheck_variant(RawVariant,InvEnv,Variant),

    get_abstract_events(RefinedModel,AbstractEvents),

    check_events(RawEvents,Name,ModelEnv,BeforeAfterEnv,
                 DroppedAbstractVars,ConcreteVariables,AbstractEvents,
                 NewVariables,Variant,Events).

typecheck_variant(-, _Env, -) :- !.
typecheck_variant(RawVariant, Env, Variant) :-
    typecheck(RawVariant, Env, Type, Variant),
    ( Type==integer       -> true
    ; functor(Type,set,1) -> true
    ; otherwise ->
        add_error(bmachine_eventb,'Invalid type of variant ',Type),
        fail).

remove_linking_parts([],_AnimationChain,[]).
remove_linking_parts([Inv|Irest],AnimationChain,Invariants) :-
    ( abstraction_var_occures(Inv,AnimationChain) ->
        Invariants = RestInvariants
    ; otherwise ->
        Invariants = [Inv|RestInvariants]),
    remove_linking_parts(Irest,AnimationChain,RestInvariants).
    
abstraction_var_occures(TExpr,AnimationChain) :-
    get_texpr_expr(TExpr,identifier(_)),!,
    get_texpr_section(TExpr,Section),
    \+ member(Section,AnimationChain).
abstraction_var_occures(TExpr,AnimationChain) :-
    syntaxtraversion(TExpr,_,_,_,Subs,_),
    abstraction_var_occures_l(Subs,AnimationChain).
abstraction_var_occures_l([Sub|_],AnimationChain) :-
    abstraction_var_occures(Sub,AnimationChain),!.
abstraction_var_occures_l([_|Srest],AnimationChain) :-
    abstraction_var_occures_l(Srest,AnimationChain).

get_texpr_section(TExpr,Section) :-
    get_texpr_info(TExpr,Info),member(section(Section),Info),!.

extract_model_sections(RawModel,Name,Refines,Sees,Variables,Invariant,Theorems,Variant,Events) :-
    RawModel = event_b_model(_Pos,Name,Sections),
    optional_rawmachine_section(refines,Sections,-,Refines),
    optional_rawmachine_section(sees,Sections,[],Sees),
    optional_rawmachine_section(variables,Sections,[],Variables),
    optional_rawmachine_section(invariant,Sections,[],Invariant),
    optional_rawmachine_section(theorems,Sections,[],Theorems),
    optional_rawmachine_section(variant,Sections,-,Variant),
    optional_rawmachine_section(events,Sections,[],Events).

%get_abstraction_variables(-,[]).
%get_abstraction_variables(Model,Variables) :-
%    internal_model_vars(Model,_Name,Variables).

get_abstract_events(-,[]).
get_abstract_events(Model,Events) :- internal_model_events(Model,_Name,Events).

get_refined_model(-,_Models,-) :- !.
get_refined_model(Name,Models,Model) :-
    internal_model(M,Name),
    ( memberchk(M,Models) ->
        M=Model
    ; otherwise ->
        add_error(bmachine_eventb,'Unable to find refined model ',Name),
        fail).

split_abstract_vars(AbstractVars,ConcreteVars,NewVars,DroppedVars) :-
    split_abstract_vars2(AbstractVars,ConcreteVars,DroppedVars),
    find_new_variables(ConcreteVars,AbstractVars,NewVars).
split_abstract_vars2([],_,[]).
split_abstract_vars2([AbstractVar|AVrest],ConcreteVars,Dropped) :-
    ( find_identifier_in_list(AbstractVar,ConcreteVars,ConcreteVar) ->
        % Both variables have the same type
        get_texpr_type(AbstractVar,Type),
        get_texpr_type(ConcreteVar,Type),
        Dropped = DroppedRest
    ; otherwise ->
        Dropped = [AbstractVar|DroppedRest]),
    split_abstract_vars2(AVrest,ConcreteVars,DroppedRest).
find_new_variables([],_,[]).
find_new_variables([ConcreteVar|CVrest],AbstractVars,NewVars) :-
    ( find_identifier_in_list(ConcreteVar,AbstractVars,_AbstractVar) -> NewVars = Nrest
    ; otherwise -> NewVars = [ConcreteVar|Nrest]),
    find_new_variables(CVrest,AbstractVars,Nrest).

% find_identifier_in_list(TypedIdToFind,List,Elem):
%  search for a typed identifier Elem in List wich
%  has the same id as TypedIdToFind, but type and informations
%  may be different
find_identifier_in_list(TypedIdToFind,List,Found) :-
    def_get_texpr_id(TypedIdToFind,Id),
    get_texpr_id(Found,Id),
    member(Found,List).

get_all_abstract_variables(Refines,AbstractModels,AbstractVariables) :-
    get_all_abstract_variables2(Refines,AbstractModels,_EncounteredIds,AbstractVariables,[]).
get_all_abstract_variables2(-,_AbstractModels,EncounteredIds) --> 
    {!,empty_avl(EncounteredIds)}.
get_all_abstract_variables2(Name,AbstractModels,EncounteredIds) -->
    {get_refined_model(Name,AbstractModels,Model),
     internal_model_refines(Model,_,Refines)},
    get_all_abstract_variables2(Refines,AbstractModels,EncounteredIdsAbs),
    get_all_abstract_variables3(Model,EncounteredIdsAbs,EncounteredIds).
get_all_abstract_variables3(Model,EncounteredIdsIn,EncounteredIdsOut,
                            FoundVarsIn,FoundVarsOut) :-
    internal_model_vars(Model,_,Vars),
    extract_new_variables(Vars,EncounteredIdsIn,NewVars),
    get_texpr_ids(NewVars,NewIds),
    add_all_to_avl(NewIds,EncounteredIdsIn,EncounteredIdsOut),
    add_all(NewVars,FoundVarsIn,FoundVarsOut).
add_all_to_avl([],Avl,Avl).
add_all_to_avl([H|T],In,Out) :-
    avl_store(H,In,1,Inter),
    add_all_to_avl(T,Inter,Out).
add_all([],L,L).
add_all([H|T],[H|L1],L2) :- add_all(T,L1,L2).
extract_new_variables([],_EncounteredIdsIn,[]).
extract_new_variables([Var|Vrest],EncounteredIds,NewVars) :-
    get_texpr_id(Var,Id),
    ( avl_fetch(Id,EncounteredIds) ->
        NewVars = NewRest
    ; otherwise ->
        NewVars = [Var|NewRest]),
    extract_new_variables(Vrest,EncounteredIds,NewRest).

    
    
    

check_events([],_ModelName,_Env,_GEnv,_Dropped,_ConcreteVars,_AbstractEvents,_NewVars,_Variant,[]).
check_events([RawEvent|RawRest],ModelName,Env,GEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVars,Variant,[Event|Erest]) :-
    (check_event(RawEvent,ModelName,Env,GEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVars,Variant,Event)
      -> true
      ;  add_failed_call_error(check_event(RawEvent,ModelName,Env,GEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVars,Variant,Event))),
    check_events(RawRest,ModelName,Env,GEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVars,Variant,Erest).


check_event(RawEvent,ModelName,Env,GEnv,DroppedVars,ConcreteVars,AbstractEvents,NewVariables,Variant,
            event(EvName,ModelName,Status,TParameters,TGuard,TTheorems,MActions,TVarWitnesses,TParamWitnesses,Unmodifiables,Refines)) :-
    
    raw_event(RawEvent,_,EvName,RawStatus,Refines,Params,Guards,Theorems,Actions,Witnesses),
    check_status(RawStatus, Variant, Status, Refines, AbstractEvents, eventlocation(ModelName,EvName)),
    get_abstract_parameters(Refines,AbstractEvents,AbstractParameters),
    create_identifiers(Params,[],TParameters),
    split_abstract_vars(AbstractParameters,TParameters,
                        _NewParameters,DroppedParameters1),
    annotate_dropped_parameters(DroppedParameters1,Refines,AbstractEvents,DroppedParameters),
    conjunct_guards(Guards,Guard),
    adapt_becomes_such(Actions,Actions1),
    add_identifiers_to_environment(TParameters,Env,Subenv),
    add_identifiers_to_environment(TParameters,GEnv,SubGenv),
    typecheck(Guard,Subenv,pred,TGuard),
    typecheck_l(Theorems,Subenv,pred,TTheorems),
    typecheck_l(Actions1,Subenv,subst,TActions),
    check_witnesses(Witnesses,SubGenv,DroppedParameters,DroppedVars,TVarWitnesses,TParamWitnesses),
    % add checks that some variables may not be modified, e.g
    % if skip is refined
    get_abstract_actions(Refines,AbstractEvents,AbstractActions),
    ~~mnf( add_modified_vars_to_actions(TActions,MActions1) ),
    add_unmodified_assignments(AbstractActions,MActions1,ConcreteVars,MActions),
    add_equality_check(MActions,AbstractActions,NewVariables,ConcreteVars,Unmodifiables).

% add assignments of the form x:=x if a not dropped abstract variable
% is modified by the abstract event, but not the dropped event
add_unmodified_assignments(AbstractActions,ConcreteActions,ConcreteVars,NewActions) :-
    get_modified_vars_of_actions(AbstractActions,AbstractModified),
    get_modified_vars_of_actions(ConcreteActions,ConcreteModified),
    remove_all(AbstractModified,ConcreteModified,OnlyAbstractModified),
    findall( Var,
             ( member(Id,OnlyAbstractModified), get_texpr_id(Var,Id), memberchk(Var,ConcreteVars)),
             ToKeep),
    ( ToKeep = []    -> NewActions = ConcreteActions
    ; ToKeep = [_|_] ->
        create_texpr(assign(ToKeep,ToKeep),subst,[],Action1),
        add_modified_vars_to_action(Action1,Action),
        NewActions = [Action|ConcreteActions]).

check_status(RawStatus, Variant, TConv, Refines, AbstractEvents, Location) :-
    remove_pos(RawStatus,Status),
    check_status2(Status, Variant, Refines, AbstractEvents, Location, Conv, Infos),
    create_texpr(Conv, status, Infos, TConv).
check_status2(ordinary, _Variant, _Refines, _AbstractEvents, _Location, ordinary, []).
check_status2(convergent, _Variant, Refines, AbstractEvents, _Location, ordinary, [convergence_proved]) :-
     % if the refined event is also convergent, we can (or must because the
     % variant may be missing in the refinement) omit the check for decreasing the
     % variant. We do this by treating the event as ordinary.
     convergence_ensured_by_abstract_event(Refines, AbstractEvents),!.
check_status2(convergent, Variant, _Refines, _AbstractEvents, Location, convergent(RVariant), [convergence_proved]) :-
     variant_must_be_set(Variant,Location,RVariant).
check_status2(anticipated, Variant, _Refines, _AbstractEvents, _Location, anticipated(RVariant), []) :- 
     get_variant(Variant,RVariant).

:- use_module(tools,[ajoin/2]).
variant_must_be_set(-,Location,RVariant) :- !,
    Location = eventlocation(Model,Name),
    ajoin(['No variant defined for ', Name, ' in model ', Model],Msg),
    add_error(bmachine_eventb,Msg),
    get_variant(-,RVariant).
    % removed fail to allow animation
variant_must_be_set(V,_,V).

% get_variant(ModelVariant,Variant) checks profides a default variant (0) if
% the model does not have a variant
get_variant(-,V) :- !, create_texpr(integer(0),integer,[default_variant],V).
get_variant(V,V).

convergence_ensured_by_abstract_event(Refines, AbstractEvents) :-
    % find a refined event which status is convergent
    internal_event_status(Event,Ref,Status),
    get_texpr_info(Status,StatusInfo),
    member(Event, AbstractEvents),
    member(Ref, Refines),
    member(convergence_proved,StatusInfo),!.

annotate_dropped_parameters([],_,_,[]).
annotate_dropped_parameters([P|PRest],Refines,AbstractEvents,[A|ARest]) :-
    def_get_texpr_id(P,Id),
    add_texpr_infos(P,[droppedby(DropList)],A),
    findall(Event, event_drops_param(Event,Refines,AbstractEvents,Id), DropList),
    annotate_dropped_parameters(PRest,Refines,AbstractEvents,ARest).
event_drops_param(Event,Refines,AbstractEvents,Id) :-
    get_texpr_id(Pattern,Id),
    member(Event,Refines),
    get_event_parameters(Event,AbstractEvents,AbParams),
    member(Pattern,AbParams).

get_abstract_parameters([],_,R) :- !,R=[].
get_abstract_parameters([Refines|Rest],Events,Params) :- !,
    get_event_parameters(Refines,Events,LocalParams),
    append(LocalParams,RestParams,Params),
    get_abstract_parameters(Rest,Events,RestParams).
get_abstract_parameters(Refines,_,Params) :- add_error(get_abstract_parameters,'Illegal Refines List: ',Refines),
    Params=[].
get_event_parameters(Refines,Events,Params) :-
    internal_event_params(Event,Refines,Params),
    member(Event,Events),!.

get_abstract_actions([],_,[]).
get_abstract_actions([Refines|Rest],Events,Actions) :-
    get_event_action(Refines,Events,LocalActions),
    append(LocalActions,RestActions,Actions),
    get_abstract_actions(Rest,Events,RestActions).
get_event_action(Refines,Events,Actions) :-
    internal_event_actions(Event,Refines,Actions),
    member(Event,Events),!.

check_witnesses([],_Env,_DroppedParams,_DroppedVars,[],[]).
check_witnesses([Witness|Rest],Env,DroppedParams,DroppedVars,TVarWitnesses,TParamWitnesses) :-
    check_witness(Witness,Env,DroppedParams,DroppedVars,TWitness,WType),
    ( WType = droppedvar ->
        TVarWitnesses = [TWitness|TWVRest],
        TParamWitnesses = TWPRest
    ; WType = droppedparam(_DropList) ->
        TVarWitnesses = TWVRest,
        TParamWitnesses = [TWitness|TWPRest]),
    check_witnesses(Rest,Env,DroppedParams,DroppedVars,TWVRest,TWPRest).
check_witness(witness(_Pos,RawId,Pred),Env,DroppedParams,DroppedVars,TWitness,WType) :-
    RawId = identifier(_,WitnessId),
    get_witness_identifier(WitnessId,DroppedParams,DroppedVars,TId,Info,WType),
    get_texpr_type(TId,Type),
    env_store(WitnessId,Type,Info,Env,Subenv),
    typecheck(Pred,Subenv,pred,TPred),
    create_texpr(witness(TId,TPred),witness,[],TWitness).
get_witness_identifier(PrimedId,_DroppedParams,DroppedVars,TPrimeId,[],droppedvar) :-
    prime_id(Id,PrimedId),
    get_texpr_id(TId,Id),
    member(TId,DroppedVars),!,
    prime_variables([TId],[TPrimeId]).
get_witness_identifier(Id,DroppedParams,_DroppedVars,TId,Info,droppedparam(DropList)) :-
    get_texpr_id(TId,Id),
    member(TId,DroppedParams),!,
    get_texpr_info(TId,Info),
    member(droppedby(DropList),Info),!.

conjunct_guards([],truth(none)).
conjunct_guards([H|T],Result) :- conjunct_guards2(T,H,Result).
conjunct_guards2([],G,G).
conjunct_guards2([H|T],G,Result) :- conjunct_guards2(T,conjunct(none,G,H),Result).

parallel_actions(Actions,Pos,Action) :-
    ( Actions = [] -> Action = skip(Pos)
    ; Actions = [S] -> Action = S
    ; otherwise -> Action = parallel(Pos,Actions)).

adapt_becomes_such([],[]).
adapt_becomes_such([Subst|Srest],[NSubst|Nsubst]) :-
    adapt_becomes_such2(Subst,NSubst),
    adapt_becomes_such(Srest,Nsubst).
adapt_becomes_such2(becomes_such(Pos,Ids,Pred),evb2_becomes_such(Pos,Ids,Pred)) :- !.
adapt_becomes_such2(Subst,Subst).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% convert B project to classical B machine

convert_event_b(Main,Freetypes,Contextes,Models,AnimationChain,Machine) :-
    debug_println(9,'Converting to classical B'),
    select_models(AnimationChain,Models,AnimatedModels),
    % todo: access to private elements of bmachine_construction
    bmachine_construction:create_machine(Main,Empty),
    convert_event_b2(Freetypes,AnimatedModels,Contextes,AnimationChain,
                     Empty,Machine),!.
convert_event_b(Main,Freetypes,Contextes,Models,AnimationChain,Machine) :-
    add_failed_call_error(convert_event_b(Main,Freetypes,Contextes,Models,AnimationChain,Machine)),
    fail.
convert_event_b2(Freetypes,AnimatedModels,Contextes,AnimationChain) -->
    create_freetypes(Freetypes),
    create_constants(AnimatedModels,Contextes,AnimatedContextNames),
    {append(AnimatedContextNames,AnimationChain,AnimatedSectionNames)},
    create_variables(AnimatedModels,AnimatedSectionNames),
    create_events(AnimatedModels,AnimatedSectionNames),
    create_animated_section_info(AnimatedContextNames,AnimationChain).

create_freetypes(Freetypes) -->
    append_to_section(freetypes,Freetypes).

create_animated_section_info(Contextes,Models) -->
    {reverse(Models,FromAbstractToConcrete),
     maplist(mark_as_model,FromAbstractToConcrete,MarkedModels),
     maplist(mark_as_context,Contextes,MarkedContextes),
     append(MarkedContextes,MarkedModels,Sections)},
    append_to_section(meta,[animated_sections(Sections)]).

mark_as_model(Name,model(Name)).
mark_as_context(Name,context(Name)).

get_operation_identifiers([],[]).
get_operation_identifiers([TExpr|Rest],[Id|IdRest]) :-
    get_texpr_expr(TExpr,operation(Id,_,_,_)),
    get_operation_identifiers(Rest,IdRest).

create_variables(Models,AnimatedSectionNames) -->
    {collect_model_infos(Models,AnimatedSectionNames,Variables,Invariants,Theorems)},
    append_to_section(abstract_variables,Variables),
    conjunct_to_section(invariant,Invariants),
    append_to_section(assertions,Theorems).

select_contexts([],_Contexts,[]).
select_contexts([Name|NRest],Contexts,[Context|CRest]) :-
    internal_context(Context,Name),
    memberchk(Context,Contexts),
    select_contexts(NRest,Contexts,CRest).

select_models([],_Models,[]).
select_models([Name|NRest],Models,[Model|MRest]) :-
    internal_model(Model,Name), memberchk(Model,Models),
    select_models(NRest,Models,MRest).

% collect_model_infos(Models,Variables,Invariants,Theorems):
%  collect all variables, invariants and theorems of the given models
collect_model_infos([],_,[],[],[]) :- !.
collect_model_infos(Models,AnimatedSectionNames,Vars,Invs,Thms) :-
    collect_model_infos2(Models,AnimatedSectionNames,AllVars,Vars1,Invs,Thms),
    % add to every variable the information in which models
    % it is present
    add_var_model_occurrences(Vars1,AllVars,Vars).
collect_model_infos2([],_,[],[],[],[]).
collect_model_infos2([Model|MRest],AnimatedSectionNames,Vars,NewVars,Invs,Thms) :-
    internal_model_vars(Model,_,LVars), append(LVars,RVars,Vars),
    % use only newly introduced variables of a model, to prevend
    % double declarations
    % in case of a limited animation chain, it is important to introduce
    % all variables for the most abstract model
    ( MRest == [] -> LNewVars = LVars
    ; otherwise   -> internal_model_newvars(Model,_,LNewVars)),
    append(LNewVars,RNewVars,NewVars),
    internal_model_invs(Model,_,LInvs1), append_stripped_predicates(LInvs1,AnimatedSectionNames,RInvs,Invs),
    internal_model_thms(Model,_,LThms1), append_stripped_predicates(LThms1,AnimatedSectionNames,RThms,Thms),
    collect_model_infos2(MRest,AnimatedSectionNames,RVars,RNewVars,RInvs,RThms).
add_var_model_occurrences([],_,[]).
add_var_model_occurrences([Var|Vrest],AllVars,[MVar|Mrest]) :-
    add_var_model_occurrence(Var,AllVars,MVar),
    add_var_model_occurrences(Vrest,AllVars,Mrest).
add_var_model_occurrence(Var,AllVars,MVar) :-
    findall(Section,
            ( find_identifier_in_list(Var,AllVars,M),
                get_texpr_info(M,Infos),
                member(section(Section),Infos)),
            Sections),
    reverse(Sections,FromAbstractToConcrete),
    add_texpr_infos(Var,[occurrences(FromAbstractToConcrete)],MVar).
append_stripped_predicates(PredsIn,AnimatedSectionNames,RestPreds,Preds) :-
    remove_linking_parts(PredsIn,AnimatedSectionNames,LPreds),
    append(LPreds,RestPreds,Preds).

all_used_contextes(AnimatedModels,Contextes,Sees) :-
    ( AnimatedModels = [] ->
        all_context_names(Contextes,Sees)
    ; otherwise ->
        collect_sees(AnimatedModels,Sees)).
all_context_names([],[]).
all_context_names([Context|Crest],[Name|Nrest]) :-
    internal_context(Context,Name),
    all_context_names(Crest,Nrest).
collect_sees([],[]).
collect_sees([Model|MRest],Sees) :-
    internal_model_sees(Model,_,LocalSees), append(LocalSees,RestSees,Sees),
    collect_sees(MRest,RestSees).

create_constants(AnimatedModels,Contextes,AnimatedContextNames) -->
    {all_used_contextes(AnimatedModels,Contextes,Sees),
     get_constants(Sees,Contextes,DefSets,EnumSets,EnumElems,Constants,Axioms,Theorems,AnimatedContextNames)},
    append_to_section(deferred_sets,DefSets),
    append_to_section(enumerated_sets,EnumSets),
    append_to_section(enumerated_elements,EnumElems),
    append_to_section(concrete_constants,Constants),
    conjunct_to_section(properties,Axioms),
    append_to_section(assertions,Theorems).

get_constants(Sees,Contextes,DefSets,EnumSets,EnumElems,Constants,Axioms,Theorems,AllContextNames) :-
    transitive_contextes(Sees,Contextes,AllContextNames),
    select_contexts(AllContextNames,Contextes,AllContextes),
    collect_context_infos(AllContextes,Sets,Constants1,Axioms1,Theorems),
    enumerated_deferred_sets(Sets,Constants1,Axioms1,DefSets,Constants,Axioms,
                             EnumSets,EnumElems).

collect_context_infos([],[],[],[],[]).
collect_context_infos([context(Name,_Ex,LSets1,LCsts1,LAxms,LThms)|Rest],Sets,Constants,Axioms,Theorems) :-
    add_context_occurrences(LSets1,Name,LSets),
    add_context_occurrences(LCsts1,Name,LCsts),
    append(LSets,RSets,Sets), 
    append(LCsts,RCsts,Constants),
    append(LAxms,RAxms,Axioms),
    append(LThms,RThms,Theorems),
    collect_context_infos(Rest,RSets,RCsts,RAxms,RThms).
add_context_occurrences([],_,[]).
add_context_occurrences([Constant|Crest],Name,[CC|CCrest]) :-
    add_texpr_infos(Constant,[occurrences([Name])],CC),
    add_context_occurrences(Crest,Name,CCrest).


% merge the events and write them to a machine
create_events(AnimatedModels,AnimatedSectionNames) -->
    {get_events(AnimatedModels,AnimatedSectionNames,Events,Initialisation,Promoted)},
    append_to_section(operation_bodies,Events),
    bmachine_construction:write_section(initialisation,Initialisation),
    append_to_section(promoted,Promoted).

% merge the events of multiple refinement levels
get_events([],_,[],Initialisation,[]) :- !,
    create_texpr(skip,subst,[],Initialisation).
get_events(AnimatedModels,AnimatedSectionNames,Events,Initialisation,Promoted) :-
    get_events_to_animate(AnimatedModels,AnimEvents),
    merge_refinement_levels(AnimEvents,AnimatedModels,AnimatedSectionNames,MergedEvents,Promoted1),
    remove_initialisation(Promoted1,Promoted2),
    add_op_parenthesis(Promoted2,Promoted),
    select_initialisation(MergedEvents,PreOps,Initialisation),
    create_operations(PreOps,Events).
remove_initialisation(In,Out) :-
    get_texpr_id(Expr,'INITIALISATION'),
    select(Expr,In,Out),!.
remove_initialisation(In,Out) :-
    % add_error(check_event_b_project,'Model has no INITIALISATION'), %err msg already generated below
    Out=In.
add_op_parenthesis([],[]).
add_op_parenthesis([Old|ORest],[New|NRest]) :-
    def_get_texpr_id(Old,Id),
    rename_bt(Old,[rename(Id,op(Id))],New),
    add_op_parenthesis(ORest,NRest).

select_initialisation(In,Ops,Init) :-
    select(eventop('INITIALISATION',Section,[],Init1,op([],[]),_),In,Ops),!,
    add_texpr_infos(Init1,[eventb(Section)],Init).
select_initialisation(In,Ops,Init) :-
    add_error(select_initialisation,'Model has no INITIALISATION'),
    Init = b(skip,subst,[]), % replace skip by non-det assign
    Ops=In.
create_operations([],[]).
create_operations([E|ERest],[O|ORest]) :-
    create_operation(E,O),
    create_operations(ERest,ORest).
create_operation(eventop(Name,Section,Parameters,Subst,Type,Refines),Operation) :-
    create_texpr(identifier(op(Name)),Type,[],Id),
    get_texpr_info(Subst,SubstInfo),
    Mod=modifies(_),memberchk(Mod,SubstInfo),
    Reads=reads(_),memberchk(Reads,SubstInfo),
    ( Refines == [] -> RefInfo = []
    ; otherwise -> RefInfo = [refines(Refines)]),
    create_texpr(operation(Id,[],Parameters,Subst),Type,[eventb(Section),Mod,Reads|RefInfo],Operation).

get_events_to_animate([Model|_],Events) :- internal_model_events(Model,_,Events).

merge_refinement_levels([],_,_,[],[]).
merge_refinement_levels([Event|ERest],AnimModels,AnimatedSectionNames,[Merged|MRest],[Prom|PRest]) :-
    merge_refinement_levels2(Event,AnimModels,AnimatedSectionNames,Merged,Prom),
    merge_refinement_levels(ERest,AnimModels,AnimatedSectionNames,MRest,PRest).

merge_refinement_levels2(Event,[CurrentLevel|Abstractions],AnimatedSectionNames,
                         eventop(Name,Section,Parameters,Subst,OpType,Refines),Prom) :-
    AnimatedLevels = [CurrentLevel|Abstractions],
    internal_model(CurrentLevel,Section),
    internal_event_params(Event,Name,Parameters),
    internal_event_refined(Event,Name,Refines),
    %merge_parameters(Refines,Parameters,Abstractions,MergedParameters),
    merge_events_to_subst(Name,AnimatedLevels,AnimatedSectionNames,[],Subst1),
    optimise_events(Subst1,Subst),
    get_texpr_types(Parameters,Types),OpType = op(Types,[]),
    create_texpr(identifier(Name),OpType,[],Prom).
merge_events_to_subst(Name,[CurrentLevel|Abstractions],AnimatedSectionNames,RefinedParameters,TEvent) :-
    create_texpr(RefLevelEvent,subst,[modifies(Vars),reads(Reads)],TEvent),
    RefLevelEvent = rlevent(Name,Section,Status,Params,Guard,Theorems,Actions,VWitnesses,PWitnesses,Unmodifiables,AbstractEvents),
    InternalEvent = event(Name,Section,Status,AllParams,Guard,Theorems,Actions,VWitnesses1,PWitnesses1,Unmodifiables,Refines),
    get_event_from_model(Name,CurrentLevel,InternalEvent),
    strip_witnesses(VWitnesses1,AnimatedSectionNames,VWitnesses),
    strip_witnesses(PWitnesses1,AnimatedSectionNames,PWitnesses),
    remove_refined_parameters(AllParams,RefinedParameters,Params),
    ( Abstractions = [] ->
        AbstractEvents = []
    ; otherwise ->
        append(RefinedParameters,Params,NewRefinedParameters),
        merge_events_to_subst_l(Refines,Abstractions,AnimatedSectionNames,NewRefinedParameters,AbstractEvents1),
        normalise_merged_events(AbstractEvents1,AbstractEvents)),
    collect_modified_vars(Actions,AbstractEvents,Vars),
    collect_read_vars(AllParams,Guard,Theorems,Actions,VWitnesses,PWitnesses,AbstractEvents,Reads).
merge_events_to_subst_l([],_,_,_,[]).
merge_events_to_subst_l([Name|NRest],Abstractions,AnimatedSectionNames,RefinedParameters,[TEvent|TRest]) :-
    merge_events_to_subst(Name,Abstractions,AnimatedSectionNames,RefinedParameters,TEvent),
    merge_events_to_subst_l(NRest,Abstractions,AnimatedSectionNames,RefinedParameters,TRest).

% removes a witness completely if the witnessed variable belongs
% to a refinement level that is not animated or removes part of the
% witness-predicate if references to not animated levels are made.
strip_witnesses([],_AnimatedSectionNames,[]).
strip_witnesses([Witness|WRest],AnimatedSectionNames,Result) :-
    strip_witness(Witness,AnimatedSectionNames,StrippedWitness),
    append(StrippedWitness,SWRest,Result),
    strip_witnesses(WRest,AnimatedSectionNames,SWRest).
strip_witness(Witness,AnimatedSectionNames,StrippedWitness) :-
    get_texpr_expr(Witness,witness(W,P)),
    ( abstraction_var_occures(W,AnimatedSectionNames) ->
        % The abstract variable is not animated, skip the complete witness
        StrippedWitness = []
    ; otherwise ->
        conjunction_to_list(P,Preds),
        remove_linking_parts(Preds,AnimatedSectionNames,SPreds),
        conjunct_predicates(SPreds,SP),
        get_texpr_type(Witness,Type),
        get_texpr_info(Witness,Info),
        create_texpr(witness(W,SP),Type,Info,TWitness),
        StrippedWitness = [TWitness]).

remove_refined_parameters([],_RefinedParameters,[]).
remove_refined_parameters([Param|APRest],RefinedParameters,Parameters) :-
    ( find_identifier_in_list(Param,RefinedParameters,_RefParam) ->
        Parameters = PRest
    ; otherwise ->
        Parameters = [Param|PRest]),
    remove_refined_parameters(APRest,RefinedParameters,PRest).

merge_parameters([],Parameters,_Abstractions,Parameters).
merge_parameters([_|_],Parameters,[],Parameters) :- !.
merge_parameters([Refines|Rest],OldParameters,Abstractions,Parameters) :-
    Abstractions = [Abstract|FurtherAbstractions],
    get_event_from_model(Refines,Abstract,Event),
    internal_event_params(Event,_,NewParameters),
    internal_event_refined(Event,_,NewRefines),
    add_unique_parameters(NewParameters,OldParameters,Parameters2),
    merge_parameters(NewRefines,Parameters2,FurtherAbstractions,Parameters3),
    merge_parameters(Rest,Parameters3,Abstractions,Parameters).
add_unique_parameters([],Parameters,Parameters).
add_unique_parameters([P|PRest],InParameters,OutParameters) :-
    ( find_identifier_in_list(P,InParameters,_) ->
        !,Parameters = InParameters
    ; otherwise ->
        Parameters = [P|InParameters]),
    add_unique_parameters(PRest,Parameters,OutParameters).

get_event_from_model(Name,Model,Event) :-
    internal_model_events(Model,_,Events),
    internal_event(Event,Name),
    member(Event,Events),!.
    
collect_modified_vars(Actions,Events,Vars) :-
    ~~mnf( get_modified_vars_of_actions(Actions,M1) ),
    append_modified_vars_of_events(Events,M2),
    append(M1,M2,M3),sort(M3,Vars).
get_modified_vars_of_actions([],[]).
get_modified_vars_of_actions([Action|Arest],Vars) :-
    get_modified_vars_of_action(Action,Vlocal),
    append(Vlocal,Vrest,Vars),
    get_modified_vars_of_actions(Arest,Vrest).
get_modified_vars_of_action(Action,Vars) :-
    get_texpr_info(Action,Infos),
    member(modifies(Vars), Infos),!.

add_modified_vars_to_actions([],[]).
add_modified_vars_to_actions([TAction|ARest],[MAction|MRest]) :-
    add_modified_vars_to_action(TAction,MAction),
    add_modified_vars_to_actions(ARest,MRest).
add_modified_vars_to_action(TAction,MAction) :-
    get_lhs_ids(TAction,Vars),
    maplist(prime_id,Vars,Primed),
    add_texpr_infos(TAction,[assigned_after(Primed),modifies(Vars)],MAction).

get_lhs_ids(TAction,Ids) :-
    get_lhs_tids(TAction,Vars),
    get_texpr_ids(Vars,Ids).
get_lhs_tids(TAction,Vars) :-
    get_texpr_expr(TAction,Action),
    split_lhs_rhs(Action,Lhs,_),
    extract_lhs_ids(Lhs,Vars).

split_lhs_rhs(assign(Ids,Exprs),Ids,Exprs).
split_lhs_rhs(assign_single_id(Id,Expr),[Id],[Expr]).
split_lhs_rhs(becomes_element_of(Ids,Expr),Ids,[Expr]).
split_lhs_rhs(becomes_such(Ids,Pred),Ids,[Pred]).
extract_lhs_ids([],[]).
extract_lhs_ids([TId|TRest],[Var|VRest]) :-
    extract_lhs_id(TId,Var),
    extract_lhs_ids(TRest,VRest).
extract_lhs_id(Id,Var) :-
    get_texpr_id(Id,_),!,Var=Id.
extract_lhs_id(Func,Var) :-
    get_texpr_expr(Func,function(Id,_)),
    get_texpr_id(Id,Var).

append_modified_vars_of_events(Events,Vars) :-
    append_modified_vars_of_events2(Events,Unsorted),
    sort(Unsorted,Vars).
append_modified_vars_of_events2([],[]).
append_modified_vars_of_events2([TEvent|ERest],Vars) :-
    get_modified_vars_of_event(TEvent,Local),
    append(Local,VRest,Vars),
    append_modified_vars_of_events2(ERest,VRest).
get_modified_vars_of_event(TEvent,Local) :-
    get_texpr_info(TEvent,Info),
    memberchk(modifies(Local),Info).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get all the variables that are accessed in an event
collect_read_vars(AllParams,Guard,Theorems,Actions,VWitnesses,PWitnesses,AbstractEvents,Reads) :-
    get_texpr_ids(AllParams,Params), % access to parameters is not reading a variable
    find_identifier_uses(Guard,Params,GrdVars),
    find_identifier_uses_l(Theorems,Params,ThmVars),
    maplist(find_all_action_reads(Params),Actions,ActVars0),      ord_union(ActVars0, ActVars),
    maplist(find_all_witness_reads(Params),VWitnesses,VWitVars0), ord_union(VWitVars0,VWitVars),
    maplist(find_all_witness_reads(Params),PWitnesses,PWitVars0), ord_union(PWitVars0,PWitVars),
    join_abstract_reads(AbstractEvents,AbstrVars),
    ord_union([GrdVars,ThmVars,ActVars,VWitVars,PWitVars,AbstrVars],Reads).
find_all_identifiers(Ignore,Expr,Vars) :-
    find_identifier_uses(Expr,Ignore,Vars). % Just switch arguments to allow the use of maplist
find_all_action_reads(Ignore,TAction,Vars) :-
    get_texpr_expr(TAction,Action),
    find_action_reads2(Action,Ignore,Vars).
find_action_reads2(assign(_,Exprs),Ignore,Vars) :- find_identifier_uses_l(Exprs,Ignore,Vars).
find_action_reads2(assign_single_id(_,Expr),Ignore,Vars) :- find_identifier_uses(Expr,Ignore,Vars).
find_action_reads2(becomes_element_of(_,Expr),Ignore,Vars) :- find_identifier_uses(Expr,Ignore,Vars).
find_action_reads2(becomes_such(Ids,Pred),Ignore,Vars) :-
    prime_variables(Ids,PIds),get_texpr_ids(PIds,Primes),
    append(Primes,Ignore,Ignore0),sort(Ignore0,Ignore1),
    find_identifier_uses(Pred,Ignore1,Vars).
find_all_witness_reads(Ignore,Witness,Vars) :-
    get_texpr_expr(Witness,witness(I,P)),
    get_texpr_id(I,Id),
    sort([Id|Ignore],Ignore2),
    find_identifier_uses(P,Ignore2,Vars1),
    % a small hack to remove the primed variables that refer to the after-state
    % without the need to put all variables in the Ingore list:
    exclude(is_primed_id,Vars1,Vars).
is_primed_id(Id) :- atom_codes(Id,Codes),[Prime]="'",last(Codes,Prime).
join_abstract_reads([],[]) :- !.
join_abstract_reads(AbstractEvents,Reads) :-
    maplist(get_read_from_info,AbstractEvents,AbsReads),
    ord_union(AbsReads,ReadBySome),
    maplist(get_modifies_from_info,AbstractEvents,AbsModifies),
    ord_union(AbsModifies,ModifiedBySome),
    ord_intersection(AbsModifies,ModifiedByAll),
    ord_subtract(ModifiedBySome,ModifiedByAll,ModifiedByNotAll),
    ord_union(ReadBySome,ModifiedByNotAll,Reads).
get_read_from_info(AbstrEvents,Reads) :-
    get_texpr_info(AbstrEvents,Info),memberchk(reads(Reads),Info).
get_modifies_from_info(AbstrEvents,Modifies) :-
    get_texpr_info(AbstrEvents,Info),memberchk(modifies(Modifies),Info).

add_equality_check(Actions,AbstractActions,NewVars,Variables,Unmodifiables) :-
    get_modified_vars_of_actions(Actions,Modified),
    get_modified_vars_of_actions(AbstractActions,AbstractModified),
    unmodifiables(Modified,AbstractModified,NewVars,Variables,Unmodifiables).
unmodifiables([],_AbstractModified,_NewVars,_Variables,[]).
unmodifiables([UM|Modified],AbstractModified,NewVars,Variables,Unmodifiables) :-
    ( is_modifiable(UM,AbstractModified,NewVars) ->
        Unmodifiables = UMrest  
    ; otherwise ->
        get_texpr_id(TUM,UM),
        member(TUM,Variables),!,
        Unmodifiables = [TUM|UMrest]),
    unmodifiables(Modified,AbstractModified,NewVars,Variables,UMrest).
% the abstract event modifies it, so we can do it, too
is_modifiable(UM,AbstractModified,_NewVars) :- member(UM,AbstractModified),!.
% the variables is newly introduced, so we can modifiy it
is_modifiable(UM,_AbstractModified,NewVars) :-
    get_texpr_id(TUM,UM), member(TUM,NewVars).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% normalising merged events means that if two refined events A and B
% change a different set of variables then the event A gets an additional
% assignment of the form x:=x with x as a variable that is changed in B
% but not in A.
% This can happen even though merged events have the same assignments because
% they again can refine events that have different assignments.

% None or one refined event do not have to be normalised
normalise_merged_events([],[]) :- !.
normalise_merged_events([E],[E]) :- !.
normalise_merged_events(Events,Normalised) :-
    append_modified_vars_of_events(Events,AllModifiedVars),
    find_typed_version_in_events(AllModifiedVars,Events,TypedVars),
    normalise_merged_events2(Events,AllModifiedVars,TypedVars,Normalised).
normalise_merged_events2([],_AllModifiedVars,_TypedVars,[]).
normalise_merged_events2([Event|Erest],AllModifiedVars,TypedVars,[Normalised|Nrest]) :-
    normalise_merged_event(Event,AllModifiedVars,TypedVars,Normalised),
    normalise_merged_events2(Erest,AllModifiedVars,TypedVars,Nrest).
normalise_merged_event(Event,AllModifiedVars,TypedVars,Normalised) :-
    get_modified_vars_of_event(Event,Modified),
    remove_all(AllModifiedVars,Modified,MissingVars),
    add_missing_assignments(MissingVars,Event,AllModifiedVars,TypedVars,Normalised).
add_missing_assignments([],Event,_,_,Event) :- !.
add_missing_assignments(MissingVars,TEvent,AllModifiedVars,TypedVars,Normalised) :-
    get_texpr_expr(TEvent,Event),
    get_texpr_type(TEvent,Type),
    get_texpr_info(TEvent,Infos),
    selectchk(modifies(_),Infos,RestInfos), % TO DO: also add reads Information
    create_texpr(NewEvent,Type,[modifies(AllModifiedVars)|RestInfos],Normalised),
    Event = rlevent(Name,Sec,St,Prm,Grd,Thm,Actions,VWit,PWit,Unmod,AbsEvents),
    NewEvent = rlevent(Name,Sec,St,Prm,Grd,Thm,NewActions,VWit,PWit,Unmod,AbsEvents),
    append(Actions,NewAssignments,NewActions),
    create_missing_assignments(MissingVars,TypedVars,NewAssignments).
create_missing_assignments([],_TypedVars,[]).
create_missing_assignments([MissingVar|Mrest],TypedVars,[Assign|Arest]) :-
    create_missing_assignment(MissingVar,TypedVars,Assign),
    create_missing_assignments(Mrest,TypedVars,Arest).
create_missing_assignment(Var,TypedVars,Assignment) :-
    get_texpr_id(TId,Var),
    memberchk(TId,TypedVars),
    prime_id(Var,Primed),
    create_texpr(assign([TId],[TId]),subst,[assigned_after(Primed),modifies(Var)],Assignment).

find_typed_version_in_events([],_Events,[]).
find_typed_version_in_events([Name|Nrest],Events,[Typed|Trest]) :-
    find_typed_version_in_events2(Name,Events,Typed),
    find_typed_version_in_events(Nrest,Events,Trest).
find_typed_version_in_events2(Name,Events,Typed) :-
    get_texpr_id(Typed,Name),
    member(Event,Events),
    get_texpr_expr(Event,rlevent(_Name,_Sec,_St,_Prm,_Grd,_Thm,Actions,_VWit,_PWit,_Unmod,AbsEvents)),
    find_typed_version_in_events3(Name,Actions,AbsEvents,Typed),!.
find_typed_version_in_events3(_Name,Actions,_AbsEvents,Typed) :-
    member(Action,Actions),
    get_lhs_tids(Action,Vars),
    member(Typed,Vars),!.
find_typed_version_in_events3(Name,_Actions,AbsEvents,Typed) :-
    find_typed_version_in_events2(Name,AbsEvents,Typed).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% optimisation of guards:
% A guard in a refinement does not need to be evaluated in in an abstract event
% again. This happens quiet frequently, especially if "extends" is used.
% The same applies for assignments: If in a refinement an action is performed,
% we do not have to check if the same action can be performed in a more abstract
% level.
optimise_events(TEvent,NTEvent) :-
    empty_avl(Empty),
    optimise_events2(Empty,TEvent,NTEvent).
optimise_events2(KnownElements,TEvent,NTEvent) :-
    create_texpr(Event,Type,Infos,TEvent),
    create_texpr(NewEvent,Type,Infos,NTEvent),
    Event = rlevent(Name,Section,Status,Params,OldGuard,NewTheorems,OldActions,
                    VWitnesses,PWitnesses,Unmodifiables,OldAbstractEvents),
    NewEvent = rlevent(Name,Section,Status,Params,NewGuard,NewTheorems,NewActions,
                    VWitnesses,PWitnesses,Unmodifiables,NewAbstractEvents),
    optimise_guards(OldGuard,NewTheorems,KnownElements,NewGuard,NewTheorems,KnownElements1),
    optimise_formulas(OldActions,KnownElements1,NewActions,NewKnownElements),
    maplist(optimise_events2(NewKnownElements),OldAbstractEvents,NewAbstractEvents).
optimise_guards(OldGuard,OldTheorems,KnownElements,NewGuard,NewTheorems,NewKnownElements) :-
    conjunction_to_list(OldGuard,OldGuards),
    optimise_formulas(OldGuards,KnownElements,NewGuards,KnownElements1),
    % If no elements were removed, we just use the old guard
    ( OldGuards=NewGuards -> OldGuard=NewGuard
    ; otherwise           -> conjunct_predicates(NewGuards,NewGuard)),
    optimise_formulas(OldTheorems,KnownElements1,NewTheorems,NewKnownElements).
% optimise_formulas(+AllFormulas,+KnownElements,-NewFormulas,-NewNormed)
%   AllFormulas:      A list of formulas
%   KnownElementes:   A list of normed versions of already encountered elements
%   NewFormulas:      A sub-list of AllFormulas which are new (not in KnownElements)
%   NewKnownElements: The updated list of already encountered elements
% If no formulas are removed, it is guaranteed that NewFormulas=AllFormulas
optimise_formulas(AllFormulas,KnownElements,NewFormulas,NewKnownElements) :-
    % convert formulas to a list: Formula/Normed where Normed is a normalised form
    maplist(add_stripped_ast,AllFormulas,AllFormNormed),
    % remove the already known formulas
    exclude(is_duplicate_formula(KnownElements),AllFormNormed,NewFormNormed),
    % split the list [Form/Normed] into [Form] and [Normed]
    maplist(unzipslash,NewFormNormed,NewFormulas1,NewNormed),
    ( AllFormNormed = NewFormNormed -> % No formulas removed, keep everything as it is
        NewFormulas = AllFormulas
    ; otherwise ->
        NewFormulas = NewFormulas1
    ),
    foldl(add_to_known_elements,NewNormed,KnownElements,NewKnownElements).
add_stripped_ast(Formula,Formula/Stripped) :-
    strip_and_norm_ast(Formula,Stripped).
is_duplicate_formula(KnownElements,_Formula/Stripped) :-
    avl_fetch(Stripped,KnownElements).
unzipslash(A/B,A,B).
add_to_known_elements(New,In,Out) :-
    avl_store(New,In,true,Out).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% internal event handling (just data structure)
internal_event(event(Name,_Section,_Status,_Params,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_Ref),Name).
internal_event_status(event(Name,_Section,Status,_Params,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_Ref),Name,Status).
internal_event_params(event(Name,_Section,_Status,Params,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,_Ref),Name,Params).
internal_event_actions(event(Name,_Section,_Status,_Params,_Grd,_Thm,Act,_VWit,_PWit,_Unmod,_Ref),Name,Act).
internal_event_refined(event(Name,_Section,_Status,_Params,_Grd,_Thm,_Act,_VWit,_PWit,_Unmod,Ref),Name,Ref).

internal_context(context(Name,_Extends,_Sets,_Constants,_Axioms,_Theorems),Name).
internal_context_extends(context(Name,Extends,_Sets,_Constants,_Axioms,_Theorems),Name,Extends).
internal_context_sets(context(Name,_Extends,Sets,_Constants,_Axioms,_Theorems),Name,Sets).
internal_context_constants(context(Name,_Extends,_Sets,Constants,_Axioms,_Theorems),Name,Constants).

internal_model(model(Name,_Refines,_Sees,_Variables,_NewVariables,_Invariants,_Theorems,_Events),Name).
internal_model_refines(model(Name,Refines,_Sees,_Variables,_NewVariables,_Invariants,_Theorems,_Events),Name,Refines).
internal_model_sees(model(Name,_Refines,Sees,_Variables,_NewVariables,_Invariants,_Theorems,_Events),Name,Sees).
internal_model_vars(model(Name,_Refines,_Sees,Variables,_NewVariables,_Invariants,_Theorems,_Events),Name,Variables).
internal_model_newvars(model(Name,_Refines,_Sees,_Variables,NewVariables,_Invariants,_Theorems,_Events),Name,NewVariables).
internal_model_invs(model(Name,_Refines,_Sees,_Variables,_NewVariables,Invariants,_Theorems,_Events),Name,Invariants).
internal_model_thms(model(Name,_Refines,_Sees,_Variables,_NewVariables,_Invariants,Theorems,_Events),Name,Theorems).
internal_model_events(model(Name,_Refines,_Sees,_Variables,_NewVariables,_Invariants,_Theorems,Events),Name,Events).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% analyse deferred sets and generate enumerated sets if possible




%% :- include(bmachine_eventb_common).
%% code below used to be in separate file, as there was the single-level and multi-level mode:



extract_name(event_b_model(_,Name,_Sections),Res) :- !, Res=Name.
extract_name(event_b_context(_,Name,_Sections),Res) :- !, Res=Name.
extract_name(MODEL,_) :- add_error(extract_name,'Could not extract model name: ',MODEL),fail.

extract_sections(event_b_model(_,_Name,Sections),Sections).
extract_sections(event_b_context(_,_Name,Sections),Sections).

% topological sort

topological_sort(List,Sorted) :-
    get_dependencies(List,Dependencies),
    topolsort(Dependencies,[],SortedNames),
    get_sorted(SortedNames,List,Sorted),!.
topological_sort(List,Sorted) :- add_failed_call_error(topological_sort(List,Sorted)),fail.

get_dependencies([],[]).
get_dependencies([E|Rest],[dep(Name,Deps)|Drest]) :-
    extract_name(E,Name),
    findall(D,dependency(E,D),Deps1),
    sort(Deps1,Deps),
    get_dependencies(Rest,Drest).

dependency(Elem,Dependend) :-
    extract_sections(Elem,Sections),
    ( rawmachine_section(extends,Sections,Extends),
        member(Dependend,Extends)
    ; rawmachine_section(refines,Sections,Dependend)).

topolsort([],_,[]) :- !.
topolsort(Deps,Fulfilled,[First|Rest]) :-
    select(dep(First,Before),Deps,Drest),
    is_sublist(Before,Fulfilled),!,
    sort([First|Fulfilled],NewFulfilled),
    topolsort(Drest,NewFulfilled,Rest).
is_sublist([],_).
is_sublist([H|T],[F|R]) :-
    ( H=F -> is_sublist(T,R)
    ; otherwise -> is_sublist([H|T],R)).

get_sorted([],_,[]).
get_sorted([Name|Nrest],Elems,[First|Srest]) :-
    member(First,Elems),
    extract_name(First,Name),!,
    get_sorted(Nrest,Elems,Srest).
    
% copy&paste from bmachine_construction
rawmachine_section(Elem,List,Result) :-
    functor(Pattern,Elem,2),
    arg(2,Pattern,Result),
    member(Pattern,List),!.

optional_rawmachine_section(Elem,List,Default,Result) :-
    ( rawmachine_section(Elem,List,Result1) -> true
    ; Result1=Default),
    Result1=Result.

% ----------------------------------------


% older versions of the .eventb files do not have a status included
raw_event(event(Pos,Name,Refines,Params,Guards,Actions,Witnesses),
      Pos,Name,ordinary(none),Refines,Params,Guards,[],Actions,Witnesses) :-
    (\+((Refines==[] ; Refines=[_|_]))
     -> add_error(b_machine_eventb,'Warning: outdated .eventb format') ; true).
raw_event(event(Pos,Name,Status,Refines,Params,Guards,Actions,Witnesses),
      Pos,Name,Status,Refines,Params,Guards,[],Actions,Witnesses).
raw_event(event(Pos,Name,Status,Refines,Params,Guards,Theorems,Actions,Witnesses),
      Pos,Name,Status,Refines,Params,Guards,Theorems,Actions,Witnesses).

% code common to bmachine_eventb.pl and bmachine_eventb_ml.pl

append_to_section(Section,List,Old,New) :-
    % todo: access to private elements of bmachine_construction
    bmachine_construction:select_section(Section,OldList,NewList,Old,New),
    append(OldList,List,NewList).
conjunct_to_section(Section,Preds,Old,New) :-
    % todo: access to private elements of bmachine_construction
    bmachine_construction:select_section(Section,OldPred,NewPred,Old,New),
    conjunct_predicates([OldPred|Preds],NewPred).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% analyse Event-B deferred sets and generate enumerated sets if possible

enumerated_deferred_sets(Sets,Constants,Axioms,NewSets,NewConstants,NewAxioms,TSet,TElements) :-
    conjunct_predicates(Axioms,Axiom),
    conjunction_to_list(Axiom,Axiom2),
    enumerated_deferred_sets2(Sets,Constants,Axiom2,NewSets,NewConstants,NewAxioms,TSet,TElements).

enumerated_deferred_sets2([],Csts,Axioms,[],Csts,Axioms,[],[]).
enumerated_deferred_sets2([Set|SetRest],Csts,Axioms,NewSets,NewCsts,NewAxioms,[TSet|TSetRest],TElementsAll) :-
    % case 1: the set is an enumerated set
    enumerated_deferred_set(Set,Csts,Axioms,ICsts,IAxioms,TSet,TElements),!,
    append(TElements,TElementsRest,TElementsAll),
    enumerated_deferred_sets2(SetRest,ICsts,IAxioms,NewSets,NewCsts,NewAxioms,TSetRest,TElementsRest).
enumerated_deferred_sets2([Set|SetRest],Csts,Axioms,[Set|NewSets],NewCsts,NewAxioms,TSets,TElements) :-
    % case 2: the set is a normal deferred set
    enumerated_deferred_sets2(SetRest,Csts,Axioms,NewSets,NewCsts,NewAxioms,TSets,TElements).

enumerated_deferred_set(Set,Constants,Axioms,NewConstants,NewAxioms,Set,TElements) :-
   %% print(checking_for_enumerated_deferred_sets(Axioms)),nl,
    has_enumeration_axioms(Set,Elements,Axioms,NewAxioms),
 %% print(found_enumerated_set(Set)),nl, %%
    % remove Elem1,...,ElemN from the constants
    remove_constants(Elements,Constants,NewConstants),
    % add enumerated set information to the elements
    get_texpr_type(Set,SType),
    (SType=set(global(Type)) -> true
     ; add_error_and_fail(enumerated_deferred_set,'Unexpected Set: ',(Set:SType))),
    (typeset_enumerated(Elements,Type,TElements) -> true
     ; add_error_and_fail(enumerated_deferred_set,'Could not typeset: ',(Type,Elements))).


has_enumeration_axioms(Set,Elements,AxiomsIn,AxiomsOut) :-
    has_partition_axioms(Set,Elements,AxiomsIn,AxiomsOut),!.
has_enumeration_axioms(Set,Elements,AxiomsIn,AxiomsOut) :-
    % there in an axiom Set = {Elem1,...,ElemN}
    is_equal_extension_set(Set,AxiomsIn,Axioms1,Elements),
    % check if all axioms needed for an enumeration are
    % present, and remove them
    % second case: All elements of the set are mutually unequal
    
    % generate the unequalities we have to check for
    generate_unequals(Elements,Unequals),
    % check if they are all in the axioms, and remove them from the axioms
    all_unequals_present(Unequals,Axioms1,AxiomsOut).

has_partition_axioms(Set,PartitionElements,AxiomsIn,AxiomsOut) :-
    % print(checking_for_enumeration_axioms(Set)),nl,
    % first case: A partion predicate is present with
    % all elements of the set as its arguments
    %  ( e.g. partition(S,{a},{b},...) )

    % create a pattern for the partition:
    get_texpr_id(Set,SetId),get_texpr_id(SetRef,SetId),
    create_texpr(partition(SetRef,Singletons),pred,_,Partition),
    % remove the partition from the axioms
    % print(looking_for_partition(SetId,Partition)),nl,nl,
    select(Partition,AxiomsIn,AxiomsOut1),
    %%print(found_partition(SetID,Partition)),nl,
    % check if all arguments are singletons and get the
    % singleton elements
    all_singletons(Singletons,PartitionElements,AxiomsOut1,AxiomsOut2,Recursive),
    (Recursive==true
     -> AxiomsOut=AxiomsIn  /* then we still need the partition axioms */
     ; AxiomsOut=AxiomsOut2).
%%    ,print(all_singletons(Singletons)),nl.
    %% print(singletons(PartitionElements)),nl,
    % check if all arguments and the identifiers
    % of the set are the same
    % same_elements(Elements,PartitionElements).


% all elements of a list are singletons, the second argument
% returns the list of elements without the set around it
all_singletons([],[],Ax,Ax,_).
all_singletons([Set|SRest],[Expr|ERest],AxiomsIn,AxiomsOut,Recursive) :-
    get_texpr_expr(Set,set_extension([Expr])),!,
    all_singletons(SRest,ERest,AxiomsIn,AxiomsOut,Recursive).
all_singletons([ID1|SRest],Result,AxiomsIn,AxiomsOut,Recursive) :- 
    ID1 = b(identifier(_ID),set(global(_)),_Info),
    Recursive=true,
    % print(trying_to_look_for_recursive_partition(_ID)),nl, %%
    has_enumeration_axioms(ID1,PartitionElements,AxiomsIn,AxiomsOut1),
   % print(recursive_partition(ID,PartitionElements)),nl,
    append(PartitionElements,ERest,Result),
    all_singletons(SRest,ERest,AxiomsOut1,AxiomsOut,_).
    
% the two lists of typed identifiers have the same identifiers
same_elements(ElemsA,ElemsB) :-
    get_texpr_ids(ElemsA,IdsA),sort(IdsA,A),
    get_texpr_ids(ElemsB,IdsB),sort(IdsB,B),
    A=B.

typeset_enumerated([],_Type,[]).
typeset_enumerated([Elem|Erest],Type,[TElem|Trest]) :-
    create_texpr(identifier(Id),global(Type),Infos,Elem),
    create_texpr(identifier(Id),global(Type),[enumerated_set_element|Infos],TElem),
    typeset_enumerated(Erest,Type,Trest).

remove_constants([],Cst,Cst).
remove_constants([TId|Erest],In,Out) :-
    get_texpr_id(TId,Id),
    get_texpr_id(Pattern,Id),
    select(Pattern,In,Env),
    remove_constants(Erest,Env,Out).


is_equal_extension_set(Set,In,Out,Elements) :-
    get_texpr_id(Set,SetId),
    get_texpr_id(Expr,SetId),
    create_texpr(set_extension(Elements),_,_,Ext),
    create_texpr(Equal,_,_,TEqual),
    ( Equal = equal(Expr,Ext); Equal = equal(Ext,Expr) ),
    select(TEqual,In,Out),!.

generate_unequals(Elements,Unequals) :-
    get_texpr_ids(Elements,Ids),
    findall(unequal(A,B),two_ids(Ids,A,B),Unequals).
two_ids([A|Rest],A,B) :- member(B,Rest).
two_ids([_|Rest],A,B) :- two_ids(Rest,A,B).

all_unequals_present([],Axioms,Axioms).
all_unequals_present([unequal(A,B)|Rest],InAxioms,OutAxioms) :-
    remove_unequalities(InAxioms,A,B,IntAxioms,0,N), N>0,!, % at least one occurrence
    all_unequals_present(Rest,IntAxioms,OutAxioms).
    
remove_unequalities([],_,_,[],N,N).
remove_unequalities([TAxiom|T],A,B,Rest,NrRemoved,R) :-
    get_texpr_id(AExpr,A), get_texpr_id(BExpr,B),
    get_texpr_expr(TAxiom,Axiom),
    is_unequality(Axiom,AExpr,BExpr),!, N1 is NrRemoved+1,
    remove_unequalities(T,A,B,Rest,N1,R).
remove_unequalities([Axiom|T],A,B,[Axiom|TR],N,R) :- % not an unequality between A and B
    remove_unequalities(T,A,B,TR,N,R).    

is_unequality(not_equal(A,B),A,B) :- !.
is_unequality(not_equal(B,A),A,B) :- !.
is_unequality(negation(TEqual),A,B) :-
    get_texpr_expr(TEqual,Equal),
    is_equality(Equal,A,B).
is_equality(equal(A,B),A,B) :- !.
is_equality(equal(B,A),A,B) :- !.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% creating a type environment with information about theories inside

:- meta_predicate env_store_operator(+,8,+,-).
% store a callback predicate in the environment
env_store_operator(Id,Module:Callback,In,Out) :-
    % the type is ignored, we just use the info field
    env_store(operator(Id),boolean,[Module:Callback],In,Out).

check_extensions(Infos,Freetypes,TheoryEnv) :-
    findall( T, (member(T,Infos),functor(T,theory,Arity),memberchk(Arity,[3,4])), Theories),
    env_empty(EmptyEnv),
    foldl(check_theory,Theories,FreetypesL,EmptyEnv,TheoryEnv),
    append(FreetypesL,Freetypes).
check_theory(theory(PT,Datatypes,OP),Freetypes,InEnv,OutEnv) :-
    % just to be backward-compatible to a version without axiomatic definitions
    check_theory(theory(PT,Datatypes,OP,[]),Freetypes,InEnv,OutEnv).
check_theory(theory(PT,Datatypes,OP,AxDefs),Freetypes,InEnv,OutEnv) :-
    extract_freetypes(Datatypes,InEnv,Freetypes,Env0),
    foldl(store_operator(PT),OP,Env0,Env1),
    foldl(store_axiomatic_operator(PT),AxDefs,Env1,OutEnv).
store_free_type(Id,In,Out) :-
    % store a parameter like a global set
    env_store(Id,set(global(Id)),[theory_parameter_type],In,Out).

store_operator(PT,operator(Id,Parameters,_WD,DirectDef,RecursiveDef),In,Out) :-
    ( RecursiveDef = [_|_] ->
        Op=unsupported_operator('Recursive definitions are not supported yet')
    ; DirectDef=[Def] ->
        % TODO [DP, 15.11.2012]: currently, we ignore the WD-condition
        Op=direct_definition(Parameters,Def,PT)
    ; otherwise ->
        Op=unsupported_operator('Operators without direct definition are currently unsupported')),
    env_store_operator(Id,Op,In,Out).

extract_freetypes(DataTypes,Env,Freetypes,OutEnv) :-
    extract_freetypes2(DataTypes,DataTypes,Env,Freetypes,Env,OutEnv).
extract_freetypes2([],_AllDataTypes,_ParamEnv,[],Env,Env).
extract_freetypes2([DataType|Drest],AllDataTypes,ParamEnv,[Freetype|Frest],InEnv,OutEnv) :-
    extract_freetype(DataType,AllDataTypes,ParamEnv,Freetype,InEnv,InterEnv),
    extract_freetypes2(Drest,AllDataTypes,ParamEnv,Frest,InterEnv,OutEnv).
extract_freetype(datatype(Id,TypeParams,Constructors),AllDataTypes,Env,
                 freetype(FreetypeId,Cases),InEnv,OutEnv) :-
    foldl(add_datatype_to_env,AllDataTypes,Env,Env1),
    foldl(add_type_parameter,TypeParams,Types,Env1,ParamEnv),
    create_freetype_typeid(Id,Types,FreetypeId),
    Type = freetype(FreetypeId),
    extract_used_type_parameters(Constructors,TypeParams,Usage),
    env_store_operator(Id,freetype_operator(FreetypeId,Types,TypeParams,Usage),InEnv,Env2),
    extract_cases(Constructors,ParamEnv,Type,Types,Cases,Env2,OutEnv).
add_datatype_to_env(datatype(Id,Params,_),In,Out) :-
    length(Params,NumberOfParams),
    functor(FreetypeId,Id,NumberOfParams),
    env_store(Id,set(freetype(FreetypeId)),[],In,Out).
add_type_parameter(TypeParam,Type,In,Out) :-
    extract_id(TypeParam,ParameterId),
    env_store(ParameterId,set(Type),[],In,Out).
extract_id(identifier(_,Id),Id).
extract_cases([],_,_,_,[],Env,Env).
extract_cases([Constructor|Rrest],ParamEnv,Type,Types,[Case|Srest],InEnv,OutEnv) :-
    extract_case(Constructor,ParamEnv,Type,Types,Case,InEnv,InterEnv),
    extract_cases(Rrest,ParamEnv,Type,Types,Srest,InterEnv,OutEnv).
extract_case(constructor(Id,[]),_ParamEnv,freetype(TypeId),_Types,
             case(Id,constant([Id])),InEnv,OutEnv) :- !,
    env_store_operator(Id,constructor_operator(TypeId,[]),InEnv,OutEnv).
extract_case(constructor(Id,Destructors),ParamEnv,Freetype,ParamTypes,
             case(Id,Type),InEnv,OutEnv) :-
    extract_destructors(Destructors,Projections,ParamEnv,Freetype,ParamTypes,Id,Type,Types,InEnv,Env),
    Freetype=freetype(FreetypeId),
    combine_typelist_with_prj(Types,Type,Projections), 
    env_store_operator(Id,constructor_operator(FreetypeId,Types),Env,OutEnv).
extract_destructors([],[],_,_,_,_,_,[],Env,Env).
extract_destructors([Destructor|Drest],[Projection|Prest],Env,Freetype,ParamTypes,
                    Case,CType,
                    [Type|Trest],In,Out) :-
    extract_destructor(Destructor,Projection,Env,Freetype,ParamTypes,Case,CType,Type,In,Inter),
    extract_destructors(Drest,Prest,Env,Freetype,ParamTypes,Case,CType,Trest,Inter,Out).
extract_destructor(destructor(Name,TypeExpr),Projection,Env,freetype(FreetypeId),ParamTypes,Case,CType,Type,In,Out) :-
    typecheck_with_freevars(TypeExpr,Env,ParamTypes,set(Type),_),
    env_store_operator(Name,destructor_operator(FreetypeId,Projection,Case,CType),In,Out).

create_freetype_typeid(Id,Types,FreetypeId) :-
    FreetypeId =.. [Id|Types].

:- assert_must_succeed(( combine_typelist_with_prj([a,b],T,P),
                         T == couple(a,b), P == [[prj1],[prj2]] )).
:- assert_must_succeed(( combine_typelist_with_prj([a,b,c],T,P),
                         T == couple(a,couple(b,c)),
                         P == [[prj1],[prj2,prj1],[prj2,prj2]] )).
:- assert_must_succeed(( combine_typelist_with_prj([a,b,c,d],T,P),
                         T == couple(couple(a,b),couple(c,d)),
                         P == [[prj1,prj1],[prj1,prj2],[prj2,prj1],[prj2,prj2]] )).
combine_typelist_with_prj([T],R,Projections) :- !,R=T,Projections=[[]].
combine_typelist_with_prj(Types,couple(TypeL,TypeR),Projections) :-
    split_list_in_middle(Types,TypesLeft,TypesRight),
    same_length(TypesLeft,PrjLeft),same_length(TypesRight,PrjRight),
    append(PrjLeft,PrjRight,Projections),
    maplist(cons(prj1),PrjLeft1,PrjLeft),
    maplist(cons(prj2),PrjRight1,PrjRight),
    combine_typelist_with_prj(TypesLeft,TypeL,PrjLeft1),
    combine_typelist_with_prj(TypesRight,TypeR,PrjRight1).
split_list_in_middle(List,Left,Right) :-
    length(List,N),
    Half is floor(N / 2),
    append_length(Left,Right,List,Half).

combine_exprlist_to_couple([T],R) :- !,T=R.
combine_exprlist_to_couple(Exprs,R) :-
    split_list_in_middle(Exprs,ExprsLeft,ExprsRight),
    combine_exprlist_to_couple(ExprsLeft,TL),
    combine_exprlist_to_couple(ExprsRight,TR),
    get_texpr_types([TL,TR],[TypeL,TypeR]),
    create_texpr(couple(TL,TR),couple(TypeL,TypeR),[],R).

combine_exprlist_to_cprod([T],R) :- !,T=R.
combine_exprlist_to_cprod(Exprs,R) :-
    split_list_in_middle(Exprs,ExprsLeft,ExprsRight),
    combine_exprlist_to_cprod(ExprsLeft,TL),
    combine_exprlist_to_cprod(ExprsRight,TR),
    get_texpr_types([TL,TR],[set(TypeL),set(TypeR)]),
    create_texpr(cartesian_product(TL,TR),set(couple(TypeL,TypeR)),[],R).

extract_used_type_parameters(Constructors,TypeParams,Usages) :-
    maplist(extract_id,TypeParams,TypeParamIds),
    convlist(used_type_parameters_of_constructor(TypeParamIds),Constructors,Usages).
used_type_parameters_of_constructor(TypeParamIds,constructor(Id,Destructors),
                                    used_param_types(Id,UsedParamTypes,DestructorExprs)) :-
    maplist(destructor_expr,Destructors,DestructorExprs),
    maplist(used_type_parameters_of_destructor(TypeParamIds),
            DestructorExprs,UsedParamTypeL),
    UsedParamTypeL = [_|_],   % store the information only for constructors that actually use
                              % type parameters
    ord_union(UsedParamTypeL,UsedParamTypes).
used_type_parameters_of_destructor(TypeParamIds,TypeExpr,UsedTypes) :-
    extract_raw_identifiers(TypeExpr,UsedIds),
    findall( N, (nth1(N,TypeParamIds,T),memberchk(T,UsedIds)), UsedTypes1),
    sort(UsedTypes1,UsedTypes).
destructor_expr(destructor(_,Expr),Expr).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% callback predicates for operators

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

unsupported_operator(Msg,Id,_Args,Pos,_Env,ExPr,Result,TRin,TRout) :-
    store_typecheck_error(Msg,Pos,TRin,TRout),
    failure_syntax_element(ExPr,Id,F,Type), % just to make sure that the result
    create_texpr(F,Type,[],Result).         % is nonvar, to prevent later errors
failure_syntax_element(expr,Id,identifier(Id),_).
failure_syntax_element(pred,_Id,falsity,pred).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: direct definitions
direct_definition(Params,DirectDef,PT,
                  _Id,TArgs,_Pos,Env,ExPr,Result,TRin,TRout) :-
    apply_direct_definition(TArgs,Params,DirectDef,PT,Env,Type,TParams,TDef,TRin,TRout),
    ( TArgs=[]  -> TDef=Result
    ; ExPr=expr -> create_texpr(let_expression(TParams,TArgs,TDef),Type,[],Result)
    ; ExPr=pred -> create_texpr(let_predicate(TParams,TArgs,TDef),pred,[],Result)).

% Args are the arguments of the operator where it is used
% Params are the parameters in the operator definition
apply_direct_definition(TArgs,Params,DirectDef,PT,Env,
                        Type,TParams,TDef,TRin,TRout) :-
    same_length(TArgs,Params),!,
    % type check the arguments given in the operator usage
    get_texpr_types(TArgs,ArgTypes),
    % type check the parameters' type expressions of the operator definition, they should have the same type
    % as the arguments (except being a set)
    % E.g. an operator "o" has one argument "arg" and type expression "T**BOOL" (with T being a parameter type of the theory),
    %   then for "o(5|->TRUE)" we have "couple(integer,boolean)" as the argument's type,
    %   we add "T" to the type environment (with type set(_)),
    %   then we type check the parameter expression "T**BOOL", expecting the type set(coule(integer,boolean)),
    %   the type check results in T being of type set(integer).
    maplist(split_arg_and_typedef,Params,ParamIds,ParamTypeExprs),    
    foldl(add_as_unbounded_type,PT,Env,SubEnv),
    maplist(map_to_set_type,ArgTypes,SetArgTypes),
    btype_l(ParamTypeExprs,SubEnv,SetArgTypes,_TParamTypeExprs,TRin,TR2),
    % type check the direct definition of the operator. To do this we need the parameters of the operators
    % in the type environment. Their type is determined by the arguments of the operator call.
    maplist(create_param_identifier,ParamIds,ArgTypes,TParams),
    add_identifiers_to_environment(TParams,SubEnv,DefEnv),
    btype(apply_direct_definition,DirectDef,DefEnv,Type,TDef,TR2,TRout).
apply_direct_definition(_TArgs,_Arguments,_DirectDef,_PT,_Env,
                        _Type,_TParams,_TDef,TRin,TRout) :-
    store_typecheck_error('Wrong number of operator arguments',TRin,TRout).
split_arg_and_typedef(argument(Id,TypeExpr),Id,TypeExpr).
add_as_unbounded_type(Id,In,Out) :- env_store(Id,set(_),[],In,Out).
map_to_set_type(Type,set(Type)).
create_param_identifier(ParamId,Type,TParam) :-
    create_texpr(identifier(ParamId),Type,[],TParam).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: datatype sets
freetype_operator(FreetypeId1,Types1,TypeParams1,Usage1,
                  _Id,TArgs,_Pos,Env,_ExPr,Result,TR,TR) :-
    % make a fresh copy to prevent that several instances of the freetype propagate
    % their types to each other (resulting in spurious typing errors)
    copy_term(freetype(FreetypeId1,Types1,TypeParams1,Usage1),
              freetype(FreetypeId, Types, TypeParams, Usage)),
    maplist(map_to_set_type,Types,SetTypes),
    get_texpr_types(TArgs,SetTypes),
    create_freetype_compset(TArgs,FreetypeId,TypeParams,Usage,Env,Result).
create_freetype_compset(TArgs,FreetypeId,TypeParams,Usage,Env,Result) :-
    find_constructors_with_complex_types(TArgs,Usage,Constructors),
    ( Constructors = [] ->
        create_texpr(freetype_set(FreetypeId),set(freetype(FreetypeId)),[],Result)
    ; otherwise ->
        create_freetype_compset2(TArgs,Constructors,FreetypeId,TypeParams,Env,Result)).
create_freetype_compset2(TArgs,Constructors,FreetypeId,TypeParams,Env,Result) :-
    btypechecker:add_ext_variables(TypeParams,Env,TParams,Subenv,TRin,TRout),
    get_texpr_types(TParams,Types),get_texpr_types(TArgs,Types),
    Type=freetype(FreetypeId),
    create_texpr(identifier(freetype_element),Type,[],TId),
    do_prettyprint_freetype(FreetypeId,TArgs,PP),
    create_texpr(comprehension_set([TId],Let),set(Type),
                 [prob_annotation('SYMBOLIC'),freetype(PP)],Result),
    create_texpr(let_predicate(TParams,TArgs,Cond),pred,[],Let),
    foldl(create_cons_check(FreetypeId,Subenv,TId),Constructors,Checks,TRin,TRout),
    conjunct_predicates(Checks,Cond).
create_cons_check(FreetypeId,Env,TId,cons(Case,Exprs),Check,TRin,TRout) :-
    btype_l(Exprs,Env,_Types,TExprs,TRin,TRout),
    combine_exprlist_to_cprod(TExprs,Set),
    get_texpr_type(Set,set(Type)),
    create_texpr(freetype_case(FreetypeId,Case,TId),pred,[],CheckCase),
    create_texpr(freetype_destructor(FreetypeId,Case,TId),Type,[],Value),
    create_texpr(member(Value,Set),pred,[],Membercheck),
    create_texpr(implication(CheckCase,Membercheck),pred,[],Check).
% this is used to pretty-print the resulting comprehension set
do_prettyprint_freetype(Id,Params,R) :-
    maplist(translate_bexpression,Params,Strs),
    ajoin_with_sep(Strs,',',S),
    functor(Id,I,_),
    ajoin([I,'(',S,')'],R).

find_constructors_with_complex_types(TArgs,Usage,Constructors) :-
    find_complex_type_params(TArgs,ComplexTP),
    findall(cons(Id,DestructorExprs),
            ( member(used_param_types(Id,UsedParamTypes,DestructorExprs),Usage),
                ord_intersect(UsedParamTypes,ComplexTP)),
            Constructors).
% returns the indeces of the type arguments that are not only types
find_complex_type_params(TArgs,ComplexTP) :-
    findall(N,(nth1(N,TArgs,TArg),\+ is_just_type(TArg)),ComplexTP1),
    sort(ComplexTP1,ComplexTP).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: datatype constructors
constructor_operator(FreetypeId1,[],Id,[],_Pos,_Env,_ExPr,Result,TRin,TRout) :-
    !, copy_term(FreetypeId1,FreetypeId),
    create_texpr(value(freeval(FreetypeId,Id,term(Id))),
                 freetype(FreetypeId),[],Result),
    TRin = TRout.
constructor_operator(FreetypeId1,ParamTypes1,
                     Id,TArgs,_Pos,_Env,_ExPr,Result,TR,TR) :-
    copy_term(constructor(FreetypeId1,ParamTypes1),constructor(FreetypeId,ParamTypes)),
    get_texpr_types(TArgs,ParamTypes),
    combine_exprlist_to_couple(TArgs,TArg),
    create_texpr(freetype_constructor(FreetypeId,Id,TArg),freetype(FreetypeId),[],Result).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: datatype destructors
destructor_operator(FreetypeId1,Projections,Case,CType1,
                    _Id,[TArg],_Pos,_Env,_ExPr,Result,TR,TR) :-
    copy_term(destructor(FreetypeId1,CType1),destructor(FreetypeId,CType)),
    get_texpr_type(TArg,freetype(FreetypeId)),
    create_texpr(freetype_destructor(FreetypeId,Case,TArg),CType,[],Destructed),
    apply_projections(Projections,Destructed,Result).

% create an projection expression (or a combination of arbitrary many) to
% access a part of a couple. The input is a list of prj1/prj2 atoms,
% e.g. [prj1,prj1,prj2]: Take the first of the first of the second element.
apply_projections([],Expr,Expr).
apply_projections([P|Rest],Expr,Result) :-
    get_texpr_type(Expr,couple(A,B)),
    apply_projections2(P,A,B,Result,Inner),
    apply_projections(Rest,Expr,Inner).
apply_projections2(prj1,A,_B,Result,Inner) :-
    create_texpr(first_of_pair(Inner),A,[],Result).
apply_projections2(prj2,_A,B,Result,Inner) :-
    create_texpr(second_of_pair(Inner),B,[],Result).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Theory plug-in: axiomatic definitions of operators
store_axiomatic_operator(PT,axiomatic_def_block(Id,Types,OpDefs,Axioms),In,Out) :-
    % look if we recognize the axiomatic defined operator
    lookup_predefined_axdef(Id,Types,OpDefs,Axioms,PT,In,Out),!.
store_axiomatic_operator(_PT,axiomatic_def_block(_Id,_Types,OpDefs,_Axioms),In,Out) :-
    % in case we do not recognize the operator, we store an error message
    % that is shown as soon the operator is used somewhere
    foldl(store_axdef_error,OpDefs,In,Out).
store_axdef_error(opdef(Id,_Arguments,_WD),In,Out) :-
    ajoin(['Axiomatic defined operator \"',Id,'\" not recognized.'],Msg),
    env_store_operator(Id,unsupported_operator(Msg),In,Out).

lookup_predefined_axdef(_Id,[],OpDefs,_Axioms,PT,In,Out) :-
    % SUM and PRODUCT
    OpDefs = [_,_], % Two operator definitions,
    memberchk(opdef('SUM',    [SumArg] ,_SumWD), OpDefs), % SUM operator with one argument
    memberchk(opdef('PRODUCT',[ProdArg],_ProdWD),OpDefs), % PRODUCT operator with one argument
    env_store_operator('SUM',    direct_definition([SumArg], Sigma,PT),In,Env1),
    env_store_operator('PRODUCT',direct_definition([ProdArg],Pi,   PT),Env1,Out),
    create_raw_aggop(general_sum,    SumArg, Sigma),
    create_raw_aggop(general_product,ProdArg,Pi).
create_raw_aggop(Op,argument(Arg,_ArgType),Result) :-
    I = identifier(none,agg_op1),             % I is the (not used) index variable
    N = identifier(none,agg_op2),             % N is the integer to sum or multiply
    A = identifier(none,Arg),                 % User-given argument
    Result =.. [Op,none,[I,N],Pred,N],        % SIGMA(I,N . I|->N:Arg | N)  (or PI)
    Pred = member(none,couple(none,I,N),A).   % I|->N : Arg

    
    
              
    