:- module(bmachine_static_checks, [static_check_machine/1]).

:- use_module(probsrc(module_information)).
:- module_info(group,repl).
:- module_info(description,'Static checking of B machines upon construction. Detection of name clashes and uninitalised variables.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(bmachine_structure, [get_section/3]).
:- use_module(btypechecker, [get_accessed_vars/4]).
:- use_module(bsyntaxtree, [get_texpr_expr/2,get_texpr_id/2,get_texpr_ids/2]).
:- use_module(error_manager, [add_warning/2]).
:- use_module(tools, [ajoin/2]).
:- use_module(library(lists)).

static_check_machine(Machine) :-
    initialises_all_variables(Machine),
    parameter_name_clashes(Machine).

% ---------------------
% Checks is all variables are initialised by the machine
% ---------------------
initialises_all_variables(Machine) :-
    get_section(initialisation,Machine,Initialisation),
    % check wich variables are read / modified by the initialisation
    get_accessed_vars(Initialisation,[],Modifies,_Reads),
    % get all variables that should be initialised
    get_section(abstract_variables,Machine,AbsVars),
    get_section(concrete_variables,Machine,ConcVars),
    append(AbsVars,ConcVars,TAllVars),
    get_texpr_ids(TAllVars,AllVars),
    % check for each variable, if the initialisation modifies it
    exclude(is_initialised(Modifies),AllVars,Uninitialised),
    % generate a warning if unitialised is not empty
    generate_uninitialised_warning(Uninitialised).

is_initialised(Modifies,Id) :-
    memberchk(Id,Modifies).

generate_uninitialised_warning([]) :- !.
generate_uninitialised_warning(Vars) :-
    ajoin(['Machine may not initialise some of its variables: ', Vars], Msg),
    add_warning(bmachine_static_checks,Msg).


% ---------------------
% Checks if an operations parameter clashes with a variable
% ---------------------
parameter_name_clashes(Machine) :-
    get_section(operation_bodies,Machine,Operations),
    % get all variables and constants that might clash
    get_section(abstract_variables,Machine,AbsVars),
    get_section(concrete_variables,Machine,ConcVars),
    get_section(abstract_constants,Machine,AbsCons),
    get_section(concrete_constants,Machine,ConcCons),
    append([AbsVars,ConcVars,AbsCons,ConcCons],AllVarsAndConstants),
    get_texpr_ids(AllVarsAndConstants,AllIds),
    % for each operation, compare the parameter names with existing vars/constants
    maplist(name_clashes(AllIds),Operations).

name_clashes(AllIds,Operation) :-
    get_texpr_expr(Operation,operation(IdFull,Results,Params,_Subst)),
    get_texpr_id(IdFull,op(Id)),
    get_texpr_ids(Results,ResultIds), get_texpr_ids(Params,ParamIds),
    include(clash_warnings(parameter,AllIds,Id),ParamIds,_ParamsCausingWarning),
    include(clash_warnings(result,AllIds,Id),ResultIds,_ResultsCausingWarning).

clash_warnings(ResultOrParameter,AllIds,Operation,Name) :-
    member(Name,AllIds), !,
    ajoin(['Operation ', ResultOrParameter, ' ', Name, ' has the same name as a variable or constant in operation ', Operation,'.'], Msg),
    add_warning(bmachine_static_checks,Msg).
