:- module(check_ast,[check_parse_unit/2,ast_eq/2]).
:- use_module(probsrc(module_information)).
:- module_info(group,repl).
:- module_info(description,'check that a prolog term is a valid B AST according to the SableCC grammar').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(bparser(ast_BParser)).
:- use_module(bparser(parser_call)).
:- use_module(bparser(pre_parser)).
:- use_module(bparser(ast_tools)).
:- use_module(probsrc(error_manager),[add_error/3]).
:- use_module(probsrc(debug),[debug_println/2]).
:- use_module(library(samsort)).
:- use_module(library(lists)).

check_parse_unit(FileName,complete_machine(Main,Machines,Files,Pragmas)):-!,
        print('Checking against ADT definition'),nl,
	maplist(check_parseUnit_catch,Machines),
        print('Parsing machine with Prolog'),nl,
	loadMachineHierarchy(FileName,complete_machine(Main,PrologRes,Files,Pragmas)),
        maplist(testParseError,PrologRes),
	print('comparing Prolog-parser with Java-parser'),nl,
        samsort(compareMachine,Machines,MachinesSorted),
	samsort(compareMachine,PrologRes,PrologSorted),
	maplist(compareProlog,MachinesSorted,PrologSorted,_).


testParseError(parse_errors(Err)) :- !,
	check_parse_unit_error(check_parser_unit_ast,'error while parsing with prolog parser',Err).
testParseError(_).

compareProlog(M1,M2,_) :- !,
	catch((ast_eq(M1,M2),!,print('Prolog AST == Java AST'),nl),
	      ast_diff(Msg,T),
	      check_parse_unit_error(check_parse_unit_ast,Msg,T)).

check_parse_unit_error(Info,Msg,Term):-
	print(check_parse_unit_error(Info,Msg,Term)),nl,
        true.
%	add_error(Info,Msg,Term).
	      
% all special cases that are introduced in trim_ast.pl
% are also special in check_ast

check_parseUnit_catch(Term):-
	catch(check_ast('ParseUnit',Term),
	      ast_error(Msg,Type,Term),
	      check_parse_unit_error(check_parse_unit_ast,Msg,type_term(Type,Term))).

check_ast(Type,Term) :- rewrite(Term,NTerm),!,check_ast(Type,NTerm). % some exceptions that are not easy to model

check_ast('IntegerLiteral',Term):-!,integer(Term).
check_ast(list('IdentifierLiteral'),Term):-!,atom(Term).
check_ast('IdentifierLiteral',Term):-!,atom(Term). % todo check where this happens
check_ast('DefLiteralPredicate',Term):-!,atom(Term). % todo check where this happens
check_ast('DefLiteralSubstitution',Term):-!,atom(Term). % todo check where this happens
check_ast('StringLiteral',Term):-!,atom(Term).
check_ast(optional(_),nothing).
check_ast(optional(Type),Term):-!,check_ast(Type,Term).
check_ast(list(Type),List):-!,
	(check_list(Type,List) -> true ;  throw(ast_error('type error in AST:',list(Type),List))).
check_ast(Type,Term) :- !,
	Term =.. [Functor,_Pos|Childs],
	lookup_types(Type,Functor,ChildTypes),
	(check_types(ChildTypes,Childs) -> true ; throw(ast_error('type error in AST:',Type,Term))).

check_types([],[]).
check_types([T|TRest],[C|CRest]) :-
	(check_ast(T,C)  -> true; throw(ast_error('type error in AST:',T,C))),
	check_types(TRest,CRest).

check_list(_,[]).
check_list(Type,[H|Rest]) :-
	check_ast(Type,H),
	check_list(Type,Rest).

%overwrite some of the AST data type definitions from Sablecc
%these are all special cases that differ from the definitions in the grammar
% see ast.pl for the definition in the grammar

%lookup_types(Type,operation_call,Res):-!, lookup_types(Type,op,Res).
lookup_types('Substitution',operation,Res):-!, lookup_types('Substitution',op,Res).
lookup_types(Type,comprehension_set,Res):-!, lookup_types(Type,prover_comprehension_set,Res).

% java swaps the return values with the identifier and wrapps the identifier in a extra 'identifier' functor
% which (just from a type-perspective makes it look like an expression)
% todo: fix this
lookup_types('Operation', operation, [ 'Expression' , list('Expression') ,  list('Expression') ,  'Substitution' ]).
%adt('Operation', operation, [ list('Expression') ,  list('IdentifierLiteral') ,  list('Expression') ,  'Substitution' ]).


%removing of optiontypes (see trim_ast.pl)
lookup_types('Substitution', if, [ 'Predicate' ,  'Substitution' ,  list('Substitution') ,  'Substitution' ]).
lookup_types('Substitution', case, [ 'Expression' ,  list('Expression') ,  'Substitution' ,  list('Substitution') , 'Substitution' ]).
lookup_types('Substitution', case, [ 'Expression' ,  list('Expression') ,  'Substitution' ,  list('Substitution') , 'Substitution' ]).

lookup_types(Type,Term,Res):- adt(Type,Term,Res),!.
lookup_types(Type,Term,_) :- throw(ast_error('cannot find constructor:',Type,Term)).


% special cases that are difficult to model
% this is similar to trim_ast:trimNode
rewrite(select(Pos,Cond,Then,When),Res) :- Res = select(Pos,Cond,Then,When,skip(none)).


% compare two ASTs
% throw exception if the two ASTs are not equal
% some special cases for lists, to giver better error messages
ast_eq(M1,M2):-ast_eq_p((M1/M2),M1,M2). % remember parent of node for better messages

ast_eq_p(_,pos(_,_,_,_,_,_),pos(_,_,_,_,_,_)). % ignore positions
ast_eq_p(_,none,none).
ast_eq_p(_,pos(_,_,_,_,_,_),none).
ast_eq_p(_,none,pos(_,_,_,_,_,_)).
ast_eq_p(P,definitions(_Pos1,L1),definitions(_Pos2,L2)):-
	ast_eq_definition_list(P,L1,L2).
ast_eq_p(_,[],[]).
ast_eq_p(P,[H1|T1],(H2|T2)):-
       	(ast_eq_list(P,[H1|T1],[H2|T2]) -> true ;  throw(ast_diff('ASTs differ ',P))).
ast_eq_p(P,A1,A2):-!,
	A1 =.. [F1|C1],
	A2 =.. [F2|C2],
	(F1 = F2 -> true ; throw(ast_diff('ASTs differ ',P))),
	(ast_eq_list((A1/A2),C1,C2) ; throw(ast_diff('ASTs differ ',P))).

% compare lists.
ast_eq_list(_,[],[]).
ast_eq_list(P,[H1|T1],[H2|T2]):-
	ast_eq_p(P,H1,H2),
	ast_eq_list(P,T1,T2).

% java returns definitions in some unknown order
% we need to sort definitions, before comparing them
ast_eq_definition_list(P,L1,L2):-
	samsort(compareName,L1,S1),
	samsort(compareName,L2,S2),
	ast_eq_list(P,S1,S2).

compareName(T1,T2):- !,
	getDefinitionName(T1,N1),!,
	getDefinitionName(T2,N2),!,
	N1 @< N2.

compareMachine(M1,M2):-!,
	getMachineName(M1,N1),
	getMachineName(M2,N2),
	N1 @< N2.