:- module(pre_parser,
	  [initDefinitionsFromGlobalMachine/0
	  ,initDefinitionsFromTokens/2
          ,resetDefinitions/0
          ,patchIdentifierTokens/2
          ,assertIdentifierTypes/1
	  ,isDefinitionIdentifier/1
	  ,definitionArity/2
	  ,definitionFileLoaded/1
	  ,fileContainsDefinition/2
	]).
                                                         
:- use_module(probsrc(module_information)).
:- module_info(group,repl).
:- module_info(description,'Utilities for parsing definitions.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(bparser(parser_tables_BParser)).
:- use_module(bparser(sablecc_engine)).
:- use_module(bparser(parser_call)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(bmachine_structure)).
:- use_module(library(lists)).
:- use_module(library(samsort)).

initDefinitionsFromGlobalMachine :-
	bmachine:machine(N,M),!,
        get_section(definitions,machine(N,M),Definitions),
        assertIdentifierTypes(Definitions).
initDefinitionsFromGlobalMachine :-
	print('patchIdentifierTokens: no machine available'),
	true.

initDefinitionsFromTokens(FileName,Tokens) :-
        (initFromTokens(FileName,Tokens) -> true
	; throw(parse_errors(['initDefinitionsFromTokens failed']))).
initFromTokens(FileName,Tokens) :-
        %use DCG parser to split definitions
	splitDefinitions(Definitions,Tokens,[]),!,
	%assert which identifier are defined and the arity of of the definitions
	%this gives limited support for cyclic recursive definitions
	%see
	maplist(assertDefinitionIdentiferArgs,Definitions),!,
	maplist(parseDefinition(FileName),Definitions).

%	findall(def(X,Y),identifierType(X,Y),L1),samsort(L1,Def1),
%	findall(def(X,Y),identifierType2(X,Y),L2),samsort(L2,Def2),
%	(Def1 = Def2 -> true
%	;throw(definitions_differ(Def1,Def2))).

parseDefinition(FileName,unparsed_definition(IdentToken,Tokens)) :-!,
	(tryParseDefinition(predicate     ,IdentToken,Tokens,AST) -> assertIdType(AST)
	;tryParseDefinition(expression    ,IdentToken,Tokens,AST) -> assertIdType(AST)
        ;tryParseDefinition(substitution  ,IdentToken,Tokens,AST) -> assertIdType(AST)
	;otherwise -> throw(parse_errors([cannot_parse_definition(IdentToken)]))        
	),
	assert(fileContainsDefinition(FileName,direct_definition(AST))).

parseDefinition(FileName,include_definition(IncludeName)) :-
	(
	  absolute_file_name(IncludeName,AbsoluteName,[relative_to(FileName),access(read)])
	  -> true
	  ; throw(parse_errors([('cannot find File',IncludeName)]))
	),!,
        loadDefinitionFileCaching(AbsoluteName),
	assert(fileContainsDefinition(FileName,include_definition(AbsoluteName))).

tryParseDefinition(Variant,IdentToken,Tokens,AST) :-!,
	lookupNewTokenClass(Variant,NewTokenID),	
	token(X,token_identifier_literal,IdentName) = IdentToken,
	NewIdentToken = token(X,NewTokenID,IdentName),
	Input=[token(none,token_kw_definition,dummy_token),NewIdentToken|Tokens],
	patchIdentifierTokens(Input,Input2),
	on_exception(parse_errors(_Err),
		     (parseTokens(pstate_BParser_0,Input2,ParseUnit),parse_unit_definition(_POS,AST)=ParseUnit)
		    ,fail ).

:- dynamic identifierType/3.
:- dynamic definitionIdentifierArgs/2.
:- dynamic fileContainsDefinition/2.
:- dynamic definitionFileLoaded/1.
resetDefinitions :-
	retractall(identifierType(_,_,_)),
	retractall(definitionIdentifierArgs(_,_)),
        retractall(fileContainsDefinition(_,_)),
        retractall(definitionFileLoaded(_)).

% patch those identifierTokens which are used as definition-identifier
% this dynamically asserted identifierType/3
patchIdentifierTokens(In,Out):- maplist(patchId,In,Out).

patchId(InToken,OutToken) :-
	token(Pos,token_identifier_literal,IdentName) = InToken,
	identifierType(IdentName,DefType,_), !,
	lookupNewTokenClass(DefType,NewTokenID),
	OutToken=token(Pos,NewTokenID,IdentName).
patchId(InToken,InToken).



isDefinitionIdentifier(Name):- definitionIdentifierArgs(Name,_).
definitionArity(Name,Num):-
	definitionIdentifierArgs(Name,Args),length(Args,Num).

lookupNewTokenClass(predicate,'token_def_literal_predicate').
lookupNewTokenClass(substitution,'token_def_literal_substitution').
lookupNewTokenClass(expression,'token_identifier_literal').


assertIdentifierTypes(Definitions) :- maplist(assertIdType,Definitions).

setIdentifierType(ID,Type,Args) :-
	assert(identifierType(ID,Type,Args)).

assertIdType(definition_decl(ID,DefType,_Pos,Paras,_Body,_Deps)) :- !,
	setIdentifierType(ID,DefType,Paras).
assertIdType(file(_)):- !.
assertIdType(X) :-
	throw(unknownDefinition(X)).


/*
store the name of the definition and the arity
*/
assertDefinitionIdentiferArgs(include_definition(_IncludeName)):-!.
assertDefinitionIdentiferArgs(unparsed_definition(IdentToken,Tokens)):-
	parseDefinitionLHS(ParseResult,[IdentToken|Tokens],_UnParsedTokens),
	assert(ParseResult).

	
splitDefinitions([]) --> tokenEOF,!.
splitDefinitions(List) --> skipToken('token_definitions'),!,definitionList(List),!,skipRest.
splitDefinitions(List) --> skipAnyToken,!,splitDefinitions(List).

definitionList([]) --> otherClauseBegin,!.
definitionList([]) --> nextTokenEOF,!.
definitionList([Def|Rest]) --> definition(Def),!,definitionList2(Rest),!.
definitionList2([]) --> otherClauseBegin,!.
definitionList2([]) --> nextTokenEOF,!.
definitionList2([]) --> skipToken('token_semicolon'),otherClauseBegin,!.
definitionList2([]) --> skipToken('token_semicolon'),nextTokenEOF,!.
definitionList2([Def|Rest]) -->
        skipToken('token_semicolon'),!,
	definition(Def),!,
	definitionList2(Rest),!.

otherClauseBegin --> skipToken('token_abstract_constants'),!.
otherClauseBegin --> skipToken('token_abstract_variables'),!.
otherClauseBegin --> skipToken('token_assertions'),!.
otherClauseBegin --> skipToken('token_concrete_constants'),!.
otherClauseBegin --> skipToken('token_concrete_variables'),!.
otherClauseBegin --> skipToken('token_constants'),!.
otherClauseBegin --> skipToken('token_constraints'),!.
otherClauseBegin --> skipToken('token_end'),!.
otherClauseBegin --> skipToken('token_implementation'),!.
otherClauseBegin --> skipToken('token_imports'),!.
otherClauseBegin --> skipToken('token_includes'),!.
otherClauseBegin --> skipToken('token_initialisation'),!.
otherClauseBegin --> skipToken('token_invariant'),!.
otherClauseBegin --> skipToken('token_local_operations'),!.
otherClauseBegin --> skipToken('token_operations'),!.
otherClauseBegin --> skipToken('token_promotes'),!.
otherClauseBegin --> skipToken('token_properties'),!.
otherClauseBegin --> skipToken('token_refines'),!.
otherClauseBegin --> skipToken('token_refinement'),!.
otherClauseBegin --> skipToken('token_sees'),!.
otherClauseBegin --> skipToken('token_sets'),!.
otherClauseBegin --> skipToken('token_uses'),!.
otherClauseBegin --> skipToken('token_values'),!.
otherClauseBegin --> skipToken('token_variables'),!.

skipRest --> tokenEOF,!.
skipRest --> skipAnyToken,!,skipRest.

skipAnyToken --> [_].
skipToken(TokenClass,[token(_,TokenClass,_String)|Out],Out).

parseToken(TokenClass,token(Pos,TokenClass,String),[token(Pos,TokenClass,String)|Out],Out).

parseAnyToken(X) --> [X].

tokenEOF([token(_,'token_eof',_String)],[]).
nextTokenEOF([token(_,'token_eof',String)],[token(_,'token_eof',String)]).

definition(include_definition(FileName)) -->
	parseToken('token_string_literal',token(_,_,FileName)),!.

definition(unparsed_definition(IdentifierToken,Tokens)) -->
	parseToken('token_identifier_literal',IdentifierToken),
	definitionLHS(Tokens-A),
	definitionRHS([noNesting],A-B),
	% append an eof-token for later parsing
	{B=[token(_,token_eof,dummy_eof_token_for_definitios)]}.

% copy every token including ==
definitionLHS(I-O) -->
	parseToken('token_double_equal',Token),!,
	{I=[Token|O]}.
definitionLHS(I-O) --> !,
	parseAnyToken(H),
	{I=[H|B]},
	definitionLHS(B-O).

definitionRHS([noNesting],L-L,In,In):-
        otherClauseBegin(In,_),!.
definitionRHS(_,L-L,In,In):-
        tokenEOF(In,_),!.
definitionRHS([noNesting],L-L,In,In):-
	skipToken('token_semicolon',In,_),!.
definitionRHS(Nesting,I-O)-->
	beginNesting(I-A),!,
	definitionRHS([begin|Nesting],A-O).
definitionRHS(Nesting,I-O)-->
	parseToken('token_left_par',X),!,
	{I=[X|A]},
	definitionRHS([leftPar|Nesting],A-O).
definitionRHS([begin|Nesting],I-O) -->
	parseToken('token_end',X),!,
	{I=[X|A]},
	definitionRHS(Nesting,A-O).
definitionRHS([leftPar|Nesting],I-O) -->
	parseToken('token_right_par',X),!,
	{I=[X|A]},
	definitionRHS(Nesting,A-O).
definitionRHS(Nesting,I-O) --> !,
	parseAnyToken(H),
	{I=[H|B]},
	definitionRHS(Nesting,B-O).

beginNesting(I-O) --> parseToken('token_begin',X) ,!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_pre',X)   ,!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_assert',X),!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_choice',X),!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_if',X)    ,!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_select',X),!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_case',X)  ,!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_any',X)   ,!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_var',X)   ,!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_while',X) ,!,{I=[X|O]}.
beginNesting(I-O) --> parseToken('token_let',X)   ,!,{I=[X|O]}.


parseDefinitionLHS(definitionIdentifierArgs(Name,Args)) -->
        parseDefinitionName(Name),
        parseDefinitionArgs(Args),
        skipToken('token_double_equal').

parseDefinitionName(Name) -->
        parseToken('token_identifier_literal',token(_,_,Name)).

parseDefinitionArgs([],In,In) :- skipToken('token_double_equal',In,_),!.
parseDefinitionArgs(L) -->
        skipToken('token_left_par'),!,
        parseDefinitionArgsBody(L),
        skipToken('token_right_par').

parseDefinitionArgsBody([Var|Rest]) -->
        parseToken('token_identifier_literal',token(_,_,Var)),
        parseDefinitionArgsBodyCont(Rest).
parseDefinitionArgsBodyCont([],In,In) :- skipToken('token_right_par',In,_).
parseDefinitionArgsBodyCont(L) -->
	skipToken('token_comma'),!,
	parseDefinitionArgsBody(L).
