:- module(sablecc_engine,[parseTokens/3,mapTokenIDs/2,lexString/3]).
/*
A parser engine for parsing tables generated by SableCC.
This calls out to trim_ast for handcoded special cases.
The lexer-mode for nested comments is also handcoded (as in the Java-version).
*/
:- use_module(probsrc(module_information)).
:- module_info(group,repl).
:- module_info(description,'A prolog parser-engine for parser tables generated by SableCC').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(bparser(parser_tables_BParser)).
:- use_module(bparser(trim_ast)).

:- use_module(library(lists)).

% :- load_files(library(detcheck),
%                   [when(compile_time), if(changed)]).

% parse a list of Tokens
% setup an inital LR-stack, call the LR-parser loop and extract the AST from the final stack.
parseTokens(InitialState,Tokens,AST):-
        ParserPosition=parserPosition(dummyStartPos,dummyEndPos),
        ParserStack=[parserState(InitialState,dummyNode,ParserPosition)],
	parserLoop(ParserStack,Tokens,AST),!.

/*
parserLoop is called approx. three times for each token (one shift, one reduce, and some internal steps).
THERE MUST BE STRICTLY NO CHOICEPOINT in parserLoop,
otherwise the parser will run out of memory for big machines.
*/
parserLoop(Stack,Input,Res) :-
	findAction(Stack,Input,Action),!,
	parserLoop2(Action,Stack,Input,Res).
parserLoop2(accept,Stack,_,Res):-!,
	Stack=[parserState(_,Res,_ParserPosition),_].
parserLoop2(error(ErrorID),_Stack,Input,_Res) :- !,
	[token(Pos,_,_)|_]=Input,
	errorMessage(ErrorID,Msg),name(N,Msg),throw(parse_errors([error(N,Pos)])).
parserLoop2(shift(_),_Stack,[],_Res) :- !,throw(parserBug('shift action with empty input')).
parserLoop2(shift(NewStateID),Stack,[H|Rest],Res) :- !,
	token(Pos,_TokenID,TokenNode)=H,
	TopStack=parserState(NewStateID,TokenNode,parserPosition(Pos,Pos)),
	parserLoop([TopStack|Stack],Rest,Res).
parserLoop2(reduce(ID,_PopCount,GotoIndex),Stack,Input,Res) :- !,
	performReduce(ID,Stack,AST,RestStack,NewPosition),!,
	findGoto(GotoIndex,RestStack,NewStateID),!,
	TopStack=parserState(NewStateID,AST,NewPosition),!,
	parserLoop([TopStack|RestStack],Input,Res).

findAction([], _,_) :- !, throw(parserBug('findAction : empty Stack')).
findAction(_ ,[],_) :- !, throw(parserBug('findAction : empty Input')).
findAction([parserState(StateID,_,_ParserPosition)|_],Input,Res) :- !,findAction_aux(Input,StateID,Res).
findAction(_,_,_) :- throw(parserBug('findAction : lookup failed')).

findAction_aux([token(_,TokenID,_)|_],StateID,Res) :- action(StateID, TokenID, Res),!.
findAction_aux(_,StateID,Res) :- defaultAction(StateID,Res).

findGoto(_GI,[],_Res) :- !, throw(parserBug('findGoto : empty stack')).
findGoto(GI,[parserState(StateID,_,_ParserPosition)|_],Res) :- goto(GI,StateID,Res),!.
findGoto(GI,_,Res) :- !,defaultGoto(GI,Res).
findGoto(_,_,_) :- throw(parserBug('findGoto : look failed')). % This will never be called !! <--


unifyStack([],Stack,Res):-Res=Stack.
unifyStack([H|T],[parserState(_,AstNode,_ParserPosition)|TStack],Res):-!,
	H=AstNode,
	unifyStack(T,TStack,Res).

findStartPos([],_,_):-!,throw('reduce-zero-nodes').
findStartPos([_],[parserState(_,_AstNode,parserPosition(StartPos,_))|_],Res):-!,Res=StartPos.
findStartPos([_H1|T1],[_H2|T2],Res):-!,findStartPos(T1,T2,Res).

performReduce(ID,Stack,AST,RestStack,NewParserPosition):-!,
	reductionRule(ID,Nodes,Term),!,
	unifyStack(Nodes,Stack,RestStack),!,
	findStartPos(Nodes,Stack,StartPos),!,
	[parserState(_,_,parserPosition(_,EndPos))|_]=Stack,
        NewParserPosition=parserPosition(StartPos,EndPos),
        (evalReduction(NewParserPosition,Term,AST)
	  -> true
	  ;throw(parserBug('error in reduce step',ID,Nodes,AST))).

evalReduction(_Env,var(X),X):- !.
evalReduction(_Env,emptyList,[]):- !.
evalReduction(Env,singletonList(L),[Res]):-!,evalReduction(Env,L,Res).
evalReduction(Env,appendList(L1,L2),Res):- !,
  evalReduction(Env,L1,R1),
  evalReduction(Env,L2,R2),
  append(R1,R2,Res).
evalReduction(Env,appendNode(L,N),Res):-!,
  evalReduction(Env,L,R1),
  evalReduction(Env,N,R2),
  append(R1,[R2],Res).
evalReduction(Env,newNode(Constructor1,_Constructor2,L),Res):-!, % todo refactor generator for reduction rules Constructor2 is no longer needed
	reduceList(Env,L,Nodes),
	toAstPosition(Env,NodePosition),
	AstNode=..[Constructor1|[NodePosition|Nodes]],
	trimNode(AstNode,Res).
evalReduction(Env,head(E),Res):-!,
  evalReduction(Env,E,[Res]).
evalReduction(_Env,resultNothing,nothing):-!.
evalReduction(_Env,Reduction,Res):-throw(parserBug('cannot eval reduction',Reduction,Res)).

reduceList(_Env,[],[]).
reduceList(Env,[H|T],[H2|T2]) :-
	evalReduction(Env,H,H2),
	reduceList(Env,T,T2).

toAstPosition(parserPosition(lineCol(StartLine,StartCol),lineCol(EndLine,EndCol)),Res):-!,
	Res=pos(nan,1,StartLine,StartCol,EndLine,EndCol).
toAstPosition(_,none).

mapTokenIDs([],[token(end_of_input,token_eof,'dummy eof token')]):-!.
mapTokenIDs([token(state_BParser_COMMENT_0,_Pos,_,_)|Rest],X):-!,
	mapTokenIDs(Rest,X).
mapTokenIDs([token(_,_Pos,token_white_space,_)|Rest],X):-!, % TODO: fix hardcoded token_white_space
	mapTokenIDs(Rest,X).
mapTokenIDs([token(_,_Pos,token_comment,_)|Rest],X):-!, % TODO: fix hardcoded token_white_space
	mapTokenIDs(Rest,X).
mapTokenIDs([token(_Mode,Pos,TokenName,TokenString)|Rest],[token(Pos,TokenName,TokenString)|R]) :-!,
	mapTokenIDs(Rest,R).

% call the main lexerLoop (pass lexer-mode, nest-count and accumulator for seen prefix).
lexString(InitialMode,String,Tokens):-
	LexerState=lexerState(String,InitialMode,0,1,1),
	lexerLoop(LexerState,Tokens),!.

lexerLoop(LexerState,TokenListTail):- isInputEmpty(LexerState),!,TokenListTail=[].
lexerLoop(LexerState,TokenListTail) :-!,
	lexerState(Input,Mode,_Nest,_Line,_Col) = LexerState,
	scanOneToken(Mode,Input,ScanResult),
	lexerLoop2(ScanResult,LexerState,TokenListTail).

lexerLoop2(right(NewInput,TokenAtom,TokenID,RevTokenString),LexerState,TokenListTail) :-!,
	lexerState(_,OldMode,OldNest,OldLine,OldCol)=LexerState,
	updateMode(OldMode/OldNest,TokenID,NewMode/NewNest),!,
	advancePosition(OldLine,OldCol,RevTokenString,NewLine,NewCol),!,
	Token=token(OldMode,lineCol(OldLine,OldCol),TokenID,TokenAtom),
	TokenListTail=[Token|NewTokenListTail],
	NewLexerState=lexerState(NewInput,NewMode,NewNest,NewLine,NewCol),
	lexerLoop(NewLexerState,NewTokenListTail).
lexerLoop2(left(Err),_LexerState,_TokenListTail) :- throw(Err).


isInputEmpty(lexerState([],_,_,_,_)).

scanOneToken(State,Input,ScanResult):-!,
	  checkState(State,left('error:no token recognized'),Input,[],ScanResult).

scanChar([],_,_,_,_Res) :- !,throw('internalLexerError').
scanChar([Char|Rest],LastAcceptedToken,State,Acc,Res) :-
  transition(State,ITable),
  lookupInterval(ITable,Char,Rest,LastAcceptedToken,Acc,Res).

checkState(State,LastAcceptedToken,Input,Acc,Res) :-
  accepts(State,Accept),!,
  checkState2(Accept,LastAcceptedToken,State,Input,Acc,Res).

checkState2(nothing,LastAcceptedToken,State,Input,Acc,Res) :- !,
	(Input = [] -> Res=LastAcceptedToken
	;scanChar(Input,LastAcceptedToken,State,Acc,Res)
	).
checkState2(just(TokenID),_LastAcceptedToken,State,Input,Acc,Res) :- !,
        makeTokenAtom(TokenID,Acc,Atom),
	(
	Input = [] -> Res=right([],Atom,TokenID,Acc)
	; scanChar(Input,right(Input,Atom,TokenID,Acc),State,Acc,Res)
	).




lookupInterval(leaf,_Char,_Rest,LastAcceptedToken,_Acc,Res):-
	Res=LastAcceptedToken.
lookupInterval(node(LowTree,LowChar,State,HighChar,HighTree),Char,Rest,LastAcceptedToken,Acc,Res) :-
	( Char =< HighChar, Char >= LowChar -> (
		epsilon(NewState) = State -> scanChar([Char|Rest],LastAcceptedToken,NewState,Acc,Res)
	       ;notEpsilon(NewState) = State  -> checkState(NewState,LastAcceptedToken,Rest,[Char|Acc],Res))
	; Char < LowChar  -> lookupInterval(LowTree,Char,Rest,LastAcceptedToken,Acc,Res)
	; Char > HighChar -> lookupInterval(HighTree,Char,Rest,LastAcceptedToken,Acc,Res)
	; otherwise -> throw('internal lexer-error: lookupInterval')). 

%
% hardcoded handling of nested comments
%

updateMode(Mode/Nest,TokenName,Res) :-
  modeTransition(TokenName, Mode,TMode),
  updateMode2(TMode/Nest, TokenName, Res).

updateMode2(Mode/Nest, token_comment, Res) :- !,
  NewNest is Nest + 1,
  updateMode3(Mode/NewNest, token_comment, Res).

updateMode2(Mode/Nest, token_comment_end, Res) :- !,
  NewNest is Nest - 1,
  updateMode3(Mode/NewNest, token_comment_end, Res).

updateMode2(ModeNest,TokenName,Res) :- updateMode3(ModeNest,TokenName,Res).

updateMode3(_/0, token_comment_end, state_BParser_NORMAL_0/0):- !.
updateMode3(ModeNest,_,ModeNest).

% first approximation: only set start-positions :
% todo : also set end-positions

advancePosition(Line,Col,TokenString,ResLine,ResCol):-
	(TokenString="\n" -> (ResLine is Line +1, ResCol =1)
	;TokenString="\t" -> (ResLine = Line, ResCol is (Col +8) mod 8)
	;otherwise -> (ResLine = Line, length(TokenString,L), ResCol is Col + L)).
