:- module(sablecc_engine_backtracking,[parseTokensAlt/3]).
/*
Experimental backtracking parser-engine.
*/
:- use_module(probsrc(module_information)).
:- module_info(group,repl).
:- module_info(description,'backtracking parser engine').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(bparser(parser_tables_BParser)).
:- use_module(bparser(trim_ast)).

:- use_module(library(lists)).

parseTokensAlt(InitialState,Tokens,Result):-
	findall(ParseResult, (toNonDetStream(Tokens,Stream),parseTokens(InitialState,Stream,ParseResult)), ResultList),
	partitionResults(ResultList,ASTs,ParseErrors),
        (ASTs=[Result] -> true
	;ASTs=[_,_|_] -> throw(parserBug('internal error: ambiguous parse'))
	;ParseErrors=[_|_] -> (reverse(ParseErrors,[LastError|_]),throw(LastError))
	;otherwise -> throw(parserBug('internal error: cannot determine parseResult'))).

partitionResults([],[],[]).
partitionResults([H|T],ASTs,Errors):-
	(H=parse_errors(_)
	-> (Errors=[H|Rest],partitionResults(T,ASTs,Rest))
	; (ASTs=[H|Rest],partitionResults(T,Rest,Errors))).

% parse a list of Tokens
% setup an inital LR-stack, call the LR-parser loop and extract the AST from the final stack.
parseTokens(InitialState,Tokens,AST):-
        ParserPosition=parserPosition(dummyStartPos,dummyEndPos),
        ParserStack=[parserState(InitialState,dummyNode,ParserPosition)],
	parserLoop(ParserStack,Tokens,AST).

parserLoop(Stack,Input,Res) :-
	findAction(Stack,Input,Action),
	parserLoop2(Action,Stack,Input,Res).

parserLoop2(accept,Stack,_,Res):-!,
	Stack=[parserState(_,Res,_ParserPosition),_].
parserLoop2(error(ErrorID),_Stack,Input,Res) :- !,
	stream(token(Pos,_,_),_)=Input,
        errorMessage(ErrorID,Msg),
        name(N,Msg),
        Res=parse_errors([error(N,Pos)]).
parserLoop2(shift(_),_Stack,emptyStream,_Res) :- !,throw(parserBug('shift action with empty input stream')).
parserLoop2(shift(NewStateID),Stack,stream(H,Rest),Res) :- !,
	token(Pos,_TokenID,TokenNode)=H,
	TopStack=parserState(NewStateID,TokenNode,parserPosition(Pos,Pos)),
	toNonDetStream(Rest,NewStream),
	parserLoop([TopStack|Stack],NewStream,Res).
parserLoop2(reduce(ID,_PopCount,GotoIndex),Stack,Input,Res) :- !,
	performReduce(ID,Stack,AST,RestStack,NewPosition),!,
	findGoto(GotoIndex,RestStack,NewStateID),!,
	TopStack=parserState(NewStateID,AST,NewPosition),!,
	parserLoop([TopStack|RestStack],Input,Res).

findAction([], _,_) :- !, throw(parserBug('findAction : empty Stack')).
findAction(_ ,emptyStream,_) :- !, throw(parserBug('findAction : empty Input stream')).
findAction([parserState(StateID,_,_ParserPosition)|_],Input,Res) :- !,findAction_aux(Input,StateID,Res).
findAction(_,_,_) :- throw(parserBug('findAction : lookup failed')).

findAction_aux(stream(token(_,TokenID,_),_),StateID,Res) :- action(StateID, TokenID, Res),!.
findAction_aux(_,StateID,Res) :- defaultAction(StateID,Res).

findGoto(_GI,[],_Res) :- !, throw(parserBug('findGoto : empty stack')).
findGoto(GI,[parserState(StateID,_,_ParserPosition)|_],Res) :- goto(GI,StateID,Res),!.
findGoto(GI,_,Res) :- !,defaultGoto(GI,Res).
findGoto(_,_,_) :- throw(parserBug('findGoto : look failed')). % This will never be called !! <--

toNonDetStream([],emptyStream):-!.
toNonDetStream([H|Rest],Stream):-!,
	token(TokenStr,TokenID,Pos)=H,
	nonDetTokenAlias(TokenID,NewID),
	Stream=stream(token(TokenStr,NewID,Pos),Rest).

unifyStack([],Stack,Res):-Res=Stack.
unifyStack([H|T],[parserState(_,AstNode,_ParserPosition)|TStack],Res):-!,
	H=AstNode,
	unifyStack(T,TStack,Res).

findStartPos([],_,_):-!,throw('reduce-zero-nodes').
findStartPos([_],[parserState(_,_AstNode,parserPosition(StartPos,_))|_],Res):-!,Res=StartPos.
findStartPos([_H1|T1],[_H2|T2],Res):-!,findStartPos(T1,T2,Res).

performReduce(ID,Stack,AST,RestStack,NewParserPosition):-!,
	reductionRule(ID,Nodes,Term),!,
	unifyStack(Nodes,Stack,RestStack),!,
	findStartPos(Nodes,Stack,StartPos),!,
	[parserState(_,_,parserPosition(_,EndPos))|_]=Stack,
        NewParserPosition=parserPosition(StartPos,EndPos),
        (evalReduction(NewParserPosition,Term,AST)
	  -> true
	  ;throw(parserBug('error in reduce step',ID,Nodes,AST))).

evalReduction(_Env,var(X),X):- !.
evalReduction(_Env,emptyList,[]):- !.
evalReduction(Env,singletonList(L),[Res]):-!,evalReduction(Env,L,Res).
evalReduction(Env,appendList(L1,L2),Res):- !,
  evalReduction(Env,L1,R1),
  evalReduction(Env,L2,R2),
  append(R1,R2,Res).
evalReduction(Env,appendNode(L,N),Res):-!,
  evalReduction(Env,L,R1),
  evalReduction(Env,N,R2),
  append(R1,[R2],Res).
evalReduction(Env,newNode(Constructor1,_Constructor2,L),Res):-!, % todo refactor generator for reduction rules Constructor2 is no longer needed
	reduceList(Env,L,Nodes),
	toAstPosition(Env,NodePosition),
	AstNode=..[Constructor1|[NodePosition|Nodes]],
	trimNode(AstNode,Res).
evalReduction(Env,head(E),Res):-!,
  evalReduction(Env,E,[Res]).
evalReduction(_Env,resultNothing,nothing):-!.
evalReduction(_Env,Reduction,Res):-throw(parserBug('cannot eval reduction',Reduction,Res)).

reduceList(_Env,[],[]).
reduceList(Env,[H|T],[H2|T2]) :-
	evalReduction(Env,H,H2),
	reduceList(Env,T,T2).

toAstPosition(parserPosition(lineCol(StartLine,StartCol),lineCol(EndLine,EndCol)),Res):-!,
	Res=pos(nan,1,StartLine,StartCol,EndLine,EndCol).
toAstPosition(_,none).
