:- module(trim_ast,[trimNode/2,nonDetTokenAlias/2,makeTokenAtom/3]).
:- use_module(probsrc(module_information)).
:- module_info(group,repl).
:- module_info(description,'AST transformations for the pure-Prolog B-Parser').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(bparser(parser_tables_BParser)).
:- use_module(bparser(pre_parser)).

:- use_module(library(lists)).
%Translate dedicated nodes in the AST to a make the AST compatible with the internal ProB representation

% nonDetTokenAlias defines non-deterministic token-rewrites for the back-tracking sable-engine.
% With non-Deterministic token-rewrites it is easy to hack around shift-reduce-conflicts in the grammar.
%


% If the lexer finds a token_identifier_literal, it could as well be stand for an token_predicate_identifier
nonDetTokenAlias('token_identifier_literal','token_predicate_identifier').

nonDetTokenAlias('token_true','token_truth_predicate'). % allow TRUE in Predicates
nonDetTokenAlias('token_false','token_bfalse').         % allow FALSE in Predicates (synonym for bflase)

nonDetTokenAlias(X,X). %Any token stands for itself


%makeTokenAtom is called from the lexer to create the an atom for the token content
makeTokenAtom('token_identifier_literal'      ,Acc,Atom)   :- !,makeIdentToken(Acc,Atom).
makeTokenAtom('token_def_literal_predicate'   ,Acc,Atom)   :- !,makeIdentToken(Acc,Atom).
makeTokenAtom('token_def_literal_substitution',Acc,Atom)   :- !,makeIdentToken(Acc,Atom).
makeTokenAtom('token_integer_literal'         ,Acc,Number) :- !,reverse(Acc,X),number_codes(Number,X).
makeTokenAtom('token_predicate_identifier'    ,Acc,Atom)   :- !,makeIdentToken(Acc,Atom).
makeTokenAtom('token_string_literal'          ,Acc,Atom)   :- !,
        reverse(Acc,List),
	append([34|WithoutQuotes],[34],List),
	atom_codes(Atom,WithoutQuotes).
makeTokenAtom(_,_,dummy_token_content).

makeIdentToken(Acc,Atom):-
	reverse(Acc,String),
	atom_codes(Atom,String).

% trimNode rewrites one AST - Node
% The parser calls trimNode on the fly in a buttom-up fashion after executing
% an LR-reductions which generates an AST node.

trimNode(machine_header(Pos,[Id],Args),machine_header(Pos,Id,Args)) :- !.
trimNode(deferred_set(Pos,[Id]),deferred_set(Pos,Id)) :- !.
trimNode(enumerated_set(Pos,[Id],L),enumerated_set(Pos,Id,L)) :- !.

% todo: fix for qualified names
trimNode(identifier(Pos,[Name]),Res):- !,
	(isDefinitionIdentifier(Name)
	    -> Res=definition(Pos,Name,[])
	    ;  Res=identifier(Pos,Name)). 

trimNode(primed_identifier(Pos,[Name],Grade),Res):- !,
	Res=primed_identifier(Pos,Name,Grade).

trimNode(function(Pos,definition(Pos,Name,[]),Args),Node):-
	length(Args,Arity),
	definitionArity(Name,Arity),!, % just guessing
	Node=definition(Pos,Name,Args).

trimNode(operation(Pos,Expr1,[Id],Expr2,Subs),operation(Pos,identifier(none,Id),Expr1,Expr2,Subs)):-!.

 %serious bullshit, because operation_call with different arity is also in the grammar
trimNode(op(Pos,Expr,L),operation_call(Pos,Expr,L)):-!.

trimNode(prover_comprehension_set(Pos,E,P),comprehension_set(Pos,E,P)):-!.

% Todo: refactor in ProB use 'nothing' for optional parts
% In these substitutions the else-part is optional.
% The parser returns 'nothing' for optional parts
% replace 'nothing' with a 'skip(none)'.

trimNode(if(Pos,Cond,Then,ElseIf,nothing),Res) :- !,Res=if(Pos,Cond,Then,ElseIf, SKIP),dummySkip(SKIP).
%trimNode(select(Pos,Cond,Then,When,nothing),Res) :- Res = select(Pos,Cond,Then,When,SKIP),dummySkip(SKIP).
trimNode(select(Pos,Cond,Then,When,nothing),Res) :- !,Res = select(Pos,Cond,Then,When). % inconsistency in Java

trimNode(case(Pos,Expr,Either,Subs,OrSubs,nothing),Res) :- !,Res = case(Pos,Expr,Either,Subs,OrSubs,SKIP),dummySkip(SKIP).

% todo: change the rules in the grammar for 'forall,exists',.. use identifierlist
trimNode(forall(P,L,T)             ,forall(P,R,T)             ):- !,trimCouple(L,R).
trimNode(exists(P,L,T)             ,exists(P,R,T)             ):- !,trimCouple(L,R).
trimNode(general_sum(P,L,T1,T2)    ,general_sum(P,R,T1,T2)    ):- !,trimCouple(L,R).
trimNode(general_product(P,L,T1,T2),general_product(P,R,T1,T2)):- !,trimCouple(L,R).
trimNode(lambda(P,L,T1,T2)         ,lambda(P,R,T1,T2)         ):- !,trimCouple(L,R).

% rewrite bool-identifier predicates  x => x=TRUE
trimNode(predicate_identifier(Pos,Ident),NewNode):-!,
	NewNode=equal(Pos,identifier(Pos,Ident),boolean_true(none)).

trimNode(predicate_function(Pos,Ident,Args),NewNode):-!,
	NewNode=equal(Pos,function(Pos,identifier(Pos,Ident),Args),boolean_true(none)).

% todo rewrite if then else 

trimNode(X,X).

trimCouple([couple(_POS,LIST)],R) :- !, R=LIST.
trimCouple(X,X).

dummySkip(skip(none)).
