% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(bsyntaxtree,
          [is_texpr/1, % checks if the given argument is a typed expression

           get_texpr_expr/2, % get the expression part of a typed expression
           get_texpr_type/2, % get the type of a typed expression
           get_texpr_info/2, % get the list of information of a typed expression
           get_texpr_id/2, % get the id of a typed identifier, fails if it's not an identifier
           def_get_texpr_id/2, % same as above, but raises error if arg1 is not an identifier
           create_typed_id/3, % create a typed identifier
           get_texpr_id_names/2, % translate a typed id list into a plain list of id names
           same_id/3, % true if two typed identifiers have the same id
           get_texpr_exprs/2, % list variant of the above
           get_texpr_types/2, % list variant of the above
           get_texpr_infos/2, % list variant of the above
           get_texpr_ids/2,   % list variant of the above
           get_texpr_pos/2,   % get the position of a typed expression

           same_texpr/2, % check if two typed expressions are equal modulo Info fields
           different_texpr_values/2, % check if two type expressions are WD and definitely denote different values

           create_texpr/4, % creates a typed expression by giving expression, type and infos
           add_texpr_infos/3, % adds some more information to a typed expression

           get_rodin_name/2, get_texpr_label/2,
           get_texpr_output_pragma/3,
           always_well_defined/1,
           is_truth/1, is_falsity/1,
           conjunct_predicates/2,
           is_a_conjunct/3, is_a_disjunct/3,
           conjunction_to_list/2, member_in_conjunction/2,
           disjunction_to_list/2,

           predicate_components/2, % split a predicate into components which use distinct identifiers
           predicate_components_with_restriction/3,
           predicate_identifiers/2,

           project_predicate_on_identifiers/4,

           find_identifier_uses/3, find_identifier_uses_l/3,
           find_typed_identifier_uses/3, find_typed_identifier_uses_l/3,
           occurs_in_expr/2,
           create_exists/3, create_exists_opt/3,
		   create_forall/3,
		   
           replace_id_by_expr/4,

           rename_bt/3,
           rename_bt_l/3,
           remove_bt/4,
           syntaxtransformation/5,
           syntaxtraversion/6,
           safe_syntaxelement/5,

           remove_all_infos/2, extract_info/2, extract_info/3,
           bsyntax_pattern/2,

           check_if_typed_predicate/1,
           check_if_typed_expression/1,
           check_if_typed_substitution/1,

           strip_and_norm_ast/2,
           is_set_type/2, get_texpr_set_type/2,
           is_just_type/1
          ]).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module provides operations on the type-checked AST.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).
:- use_module(library(ordsets)).
:- use_module(library(avl)).
:- use_module(library(terms)).

:- use_module(self_check).
:- use_module(error_manager).

:- use_module(typing_tools,[is_infinite_type/1]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% basic access to enriched syntax tree

is_texpr(b(_,_,_)).

get_texpr_expr(b(Name,_,_),R) :- !,R=Name.
get_texpr_expr(E,_) :- add_error_fail(get_texpr_expr,'B Expression not properly wrapped: ',E).

get_texpr_type(b(_,Type,_),R) :- !, R=Type.
get_texpr_type(E,_) :- add_error_fail(get_texpr_type,'B Expression not properly wrapped: ',E).
get_texpr_info(b(_,_,Info),R) :- !, R=Info.
get_texpr_info(E,_) :- add_error_fail(get_texpr_info,'B Expression not properly wrapped: ',E).

get_texpr_id(b(N,_,_),Id) :- !, N=identifier(Id).
get_texpr_id(E,_) :- add_error_fail(get_texpr_id,'B Expression not properly wrapped: ',E).
/* same as above: but must succeed */
def_get_texpr_id(b(identifier(Id),_,_),R) :- !,R=Id.
def_get_texpr_id(ID,_) :- add_error_fail(def_get_texpr_id,'Could not extract identifier: ',ID).

create_typed_id(IDName,Type,b(identifier(IDName),Type,[])).

same_id(X,Y,ID) :- get_texpr_id(X,ID), get_texpr_id(Y,ID).


% translate a list of ids with the b/3 wrapper into a plain id list
get_texpr_id_names([],[]).
get_texpr_id_names([H|TT],[ID|T]) :- def_get_texpr_id(H,ID), get_texpr_id_names(TT,T).

% check if two wrapped expressions are equal (modulo associated Info, e.g. source loc info)
same_texpr(b(E1,Type,_),b(E2,Type,_)) :-
  same_functor(E1,E2),
  safe_syntaxelement(E1,Subs1,_Names1,_List1,Constant),
  safe_syntaxelement(E2,Subs2,_Names2,_List2,Constant2),
  Constant==Constant2, % in case we have values with variables inside !
  same_sub_expressions(Subs1,Subs2).

same_sub_expressions([],[]).
same_sub_expressions([H1|T1],[H2|T2]) :- same_texpr(H1,H2), same_sub_expressions(T1,T2).

% check if two wrapped expressions definitely denote a different value (and are well-defined)
% only detects certain cases !
different_texpr_values(b(E1,Type,_),b(E2,Type,_)) :- 
 different_value(E1,E2).
different_value(boolean_false,boolean_true) :- !.
different_value(boolean_true,boolean_false) :- !.
different_value(integer(X),integer(Y)) :- !, X\=Y.
different_value(string(X),string(Y)) :- !, X\=Y.


get_texpr_exprs([],[]).
get_texpr_exprs([E|Rest],[N|NRest]) :- get_texpr_expr(E,N),get_texpr_exprs(Rest,NRest).
get_texpr_types([],[]).
get_texpr_types([E|Rest],[T|TRest]) :- get_texpr_type(E,T),get_texpr_types(Rest,TRest).
get_texpr_infos([],[]).
get_texpr_infos([E|Rest],[I|IRest]) :- get_texpr_info(E,I),get_texpr_infos(Rest,IRest).
get_texpr_ids([],[]).
get_texpr_ids([E|Rest],[I|IRest]) :- get_texpr_id(E,I),get_texpr_ids(Rest,IRest).

get_texpr_pos(TExpr,Pos) :-
    get_texpr_info(TExpr,Infos),
    ( member(nodeid(Pos1),Infos) -> !,Pos=Pos1
    ; Pos = none).

create_texpr(Expr,Type,Info,b(Expr,Type,Info)).

add_texpr_infos(b(Expr,Type,Old),Infos,b(Expr,Type,New)) :-
    append(Infos,Old,New).


% try and extract the Rodin name of a predicate or expression
get_rodin_name(Expression,Name) :- get_texpr_pos(Expression,rodinpos(Name,_)), Name \= [].
get_rodin_name(Expression,Name) :- get_texpr_pos(Expression,rodinpos(_Model,Name,_)), Name \= []. % new rodinpos


% try and extract the Rodin name or label pragma of a predicate or expression
get_texpr_label(TExpr,Label) :-
    get_texpr_info(TExpr,Infos),
    member(I,Infos), info_label(I,Label).
info_label(nodeid(rodinpos(Name,_)),[Name]) :- Name \= [].
info_label(nodeid(rodinpos(_,Name,_)),[Name]) :- Name \= []. % new rodinpos
info_label(label(LabelList),LabelList).

get_texpr_output_pragma(TExpr,[Pragma],[Content]) :-
    get_texpr_info(TExpr,Infos),
    member(output_pragma(Pragma,Content),Infos).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- assert_must_succeed( (E = "1/2",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "3 mod 2",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "1/0",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2*(3 mod card({}))",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "1/(2+1-3)",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "max({1,2,3,0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "min({1,2,3,0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "min({1,2,3,0}-(0..4))",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2+max({1,2,3, 1/0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none,\+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2-card({1,2,3})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none, always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "2-card({1,2,3, 1/0})",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==integer, Err==none,\+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "bool(2-size([1,2,3, 1/0])=3)",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==boolean, Err==none, \+ always_well_defined(ET) ) ).
:- assert_must_succeed( (E = "bool(2-size([1,2,3, 1/2, 3 mod 2, 3**2])=size([]))",
       bmachine:b_parse_machine_expression_from_codes(E,ET,Type,false,Err),
       Type==boolean, Err==none, always_well_defined(ET) ) ).


always_well_defined(b(E,_,Infos)) :- !,
  (always_wd(E) -> true  % some special rules
   ; nonmember(contains_wd_condition,Infos)).
always_well_defined(E) :- add_error_fail(always_well_defined,'Illegal call: ',always_well_defined(E)).


:- assert_must_succeed( always_wd(div( b(integer(2),integer,[]), b(integer(2),integer,[]) )) ).
:- assert_must_fail( always_wd(div( b(integer(2),integer,[]), b(integer(0),integer,[]) )) ).
:- assert_must_succeed( always_wd(mod( b(integer(2),integer,[]), b(integer(2),integer,[]) )) ).
:- assert_must_fail( always_wd(mod( b(integer(2),integer,[]), b(integer(0),integer,[]) )) ).
:- assert_must_fail( always_wd(mod( b(integer(2),integer,[]), b(integer(-2),integer,[]) )) ).
:- assert_must_fail( always_wd(mod( b(integer(-1),integer,[]), b(integer(2),integer,[]) )) ).
:- assert_must_succeed( always_wd(power_of( b(integer(2),integer,[]), b(integer(3),integer,[]) )) ).
:- assert_must_succeed( always_wd(power_of( b(integer(2),integer,[]), b(integer(0),integer,[]) )) ).
:- assert_must_fail( always_wd(power_of( b(integer(2),integer,[]), b(integer(-1),integer,[]) )) ).

% catch cases where the construct is currently so instantiated that we can determine that it is well-defined
% can happen e.g. during closure compilation
:- use_module(custom_explicit_sets,[check_in_domain_of_avlset/2]).
always_wd(power_of(X,Y)) :- integer_value(Y,Val), number(Val),Val>=0, always_well_defined(X).
always_wd(div(X,Y)) :-  integer_value(Y,Val), number(Val),Val\=0, always_well_defined(X).
always_wd(mod(X,Y)) :- integer_value(Y,Val), number(Val),Val>0, 
   integer_value(X,ValX), number(ValX), ValX>=0. % TO DO: maybe treat this differently in TLA / Event-B minor mode ?
always_wd(function(X,Y)) :- X= b(value(avl_set(AVL)),_,_), Y= b(value(Val),_,_),
   ground(Val), check_in_domain_of_avlset(Val,AVL). % ,print(ok),nl.
always_wd(min(X)) :- non_empty_set_value(X).
always_wd(max(X)) :- non_empty_set_value(X).
always_wd(card(X)) :- finite_set_value(X).
always_wd(size(X)) :- finite_seq_value(X).
always_wd(general_intersection(X)) :- non_empty_set_value(X).
% other candidates: size(_), first(_) last(_) tail(_) front(_) restrict_front(_,_) restrict_tail(_,_) rel_iterate(_,_)

integer_value(b(E,integer,_),Val) :- integer_value2(E,Val).
integer_value2(integer(X),R) :- !, R=X.
integer_value2(value(int(X)),R) :- !, R=X.
%integer_value2(X,_) :- print(not_integer_value2(X)),nl,fail.

non_empty_set_value(b(E,_,_)) :- non_empty_set2(E).
non_empty_set2(value(X)) :- definitely_not_empty(X).
non_empty_set2(set_extension(S)) :- l_always_well_defined(S). % the set_extension could contain wd_errors !!
non_empty_set2(sequence_extension(S)) :- l_always_well_defined(S). % ditto
definitely_not_empty(X) :- var(X),!,fail.
definitely_not_empty(avl_set(_)).
definitely_not_empty([_|_]).

finite_set_value(b(E,_,_)) :- finite_set2(E).
finite_set2(empty_set).
finite_set2(empty_sequence).
finite_set2(value(X)) :- X==[] -> true ; definitely_not_empty(X).
finite_set2(set_extension(S)) :- l_always_well_defined(S). % the set_extension could contain wd_errors !!
finite_set2(sequence_extension(S)) :- l_always_well_defined(S). % ditto

finite_seq_value(b(E,_,_)) :- finite_seq2(E).
finite_seq2(empty_set).
finite_seq2(empty_sequence).
finite_seq2(value(X)) :- X==[]. % it could be expensive to check if non empty value is a sequence ??
finite_seq2(sequence_extension(S)) :- l_always_well_defined(S).

l_always_well_defined([]).
l_always_well_defined([H|T]) :- always_well_defined(H), l_always_well_defined(T).

is_truth(b(truth,pred,_)).
is_falsity(b(falsity,pred,_)).

% conjunction of a list of predicates
conjunct_predicates([],b(truth,pred,[])).
conjunct_predicates([P|Rest],Result) :- conjunct2(Rest,P,Result).
conjunct2([],P,P).
conjunct2([Q|Rest],P,Result) :- conjunct3(P,Q,R), conjunct2(Rest,R,Result).
conjunct3(b(truth,_,_),P,P) :- !.
conjunct3(P,b(truth,_,_),P) :- !.
conjunct3(A,B,b(conjunct(A,B),pred,NewInfo)) :- extract_info(A,B,NewInfo).

extract_info(b(_,_,Info1),NewInfo) :-
   extract_info_aux(Info1,[],I1),!, NewInfo = I1.
extract_info(A,R) :- add_internal_error('Could not extract info: ',extract_info(A,R)),
 R=[].
extract_info(b(_,_,Info1),b(_,_,Info2),NewInfo) :-
   extract_info_aux(Info1,[],I1),
   extract_info_aux(Info2,I1,I2),!, NewInfo = I2.
%:- use_module(library(ordsets),[ord_intersection/3]).
%   ord_intersection(Info1,Info2,NewInfo),!.
extract_info(A,B,R) :- add_internal_error('Could not extract info: ',extract_info(A,B,R)),
 R=[].

extract_info_aux([],Acc,Acc).
extract_info_aux([H|T],Acc,Res) :-
  ((important_info(H), \+ member(H,Acc)) -> extract_info_aux(T,[H|Acc],Res)
    ; extract_info_aux(T,Acc,Res)).
important_info(contains_wd_condition).
important_info(prob_annotation(_)).

is_a_conjunct(b(conjunct(A,B),pred,_),A,B).

conjunction_to_list(C,L) :- var(C),!,
   add_error_fail(conjunction_to_list,'Internal error: var :',conjunction_to_list(C,L)).
conjunction_to_list(C,List) :- is_a_conjunct(C,LHS,RHS),!,
   conjunction_to_list(LHS,L1),
   conjunction_to_list(RHS,R1),
   append(L1,R1,List).  % TO DO: improve, use Difference Lists
conjunction_to_list(b(truth,pred,_),R) :- !, R=[].
conjunction_to_list(X,[X]).

member_in_conjunction(M,Conj) :- is_a_conjunct(Conj,LHS,RHS),!,
   (member_in_conjunction(M,LHS) ; member_in_conjunction(M,RHS)).
member_in_conjunction(M,M).

is_a_disjunct(b(disjunct(A,B),pred,_),A,B).

disjunction_to_list(C,L) :- var(C),!,
   add_error_fail(disjunction_to_list,'Internal error: var :',disjunction_to_list(C,L)).
disjunction_to_list(C,List) :- is_a_disjunct(C,LHS,RHS),!,
   disjunction_to_list(LHS,L1),
   disjunction_to_list(RHS,R1),
   append(L1,R1,List).  % TO DO: improve, use Difference Lists
disjunction_to_list(X,[X]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove all info fields
remove_all_infos(TExpr,TNExpr) :-
    var(TExpr),!,TExpr=TNExpr.
remove_all_infos(TExpr,TNExpr) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_type(TExpr,Type),
    syntaxtransformation(Expr,Subs,_,NSubs,NExpr),
    create_texpr(NExpr,Type,_,TNExpr),
    remove_all_infos_l(Subs,NSubs).
remove_all_infos_l([],[]).
remove_all_infos_l([Sub|Srest],[NSub|Nrest]) :-
    remove_all_infos(Sub,NSub),
    remove_all_infos_l(Srest,Nrest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% create_exists_opt(TIds,Preds,NewPred)
%   creating an existential quantification with some optimisations.
% TIds: The typed identifiers that are quantified
% Preds: A list of predicates
% NewPred: The (optimised) quantified expression
% Basically two optimisations are performed:
%  - identifiers that are not used at all are removed from the quantifier
%  - predicates that do not use one of the quantified identifiers are moved
%    outside the quantification, resulting in a predicate of the form "P & #x.Q(x)"
create_exists_opt([],Preds,NewPred) :- !,
    conjunct_predicates(Preds,NewPred).
create_exists_opt(TIds,Preds,NewPred) :-
    get_texpr_ids(TIds,UnsortedIds), sort(UnsortedIds,Ids),
    create_exists_opt2(Preds,Ids,[],Inner,Outer,UsedIds),
    ( Inner = []    -> create_texpr(truth,pred,[],Exists)
    ; Inner = [_|_] ->
        ceo_filter_used_ids(TIds,UsedIds,UsedTIds),
        conjunct_predicates(Inner,I),
        create_exists(UsedTIds,I,Exists)),
    append(Outer,[Exists],All),
    conjunct_predicates(All,NewPred).
create_exists_opt2([],_Ids,UsedIds,[],[],UsedIds) :- !.
create_exists_opt2([Pred|Prest],Ids,UsedIdsIn,Inner,Outer,UsedIdsOut) :- !,
    find_identifier_uses(Pred, [], LocalUses),
    ord_intersection(Ids,LocalUses,LocalExUses),
    ( LocalExUses = [] ->
        Inner = Irest,
        Outer = [Pred|Orest],
        UsedIdsIn = Urest
    ; LocalExUses = [_|_] ->
        Inner = [Pred|Irest],
        Outer = Orest,
        ord_union(UsedIdsIn,LocalExUses,Urest)
    ),
    create_exists_opt2(Prest,Ids,Urest,Irest,Orest,UsedIdsOut).
create_exists_opt2(Pred,Ids,UsedIdsIn,Inner,Outer,UsedIdsOut) :-
   add_error(create_exists_opt,'Expexting predicate list: ',Pred),
   create_exists_opt2([Pred],Ids,UsedIdsIn,Inner,Outer,UsedIdsOut).
ceo_filter_used_ids([],_UsedIds,[]).
ceo_filter_used_ids([TId|Irest],UsedIds,Filtered) :-
    get_texpr_id(TId,Id),
    ( ord_member(Id,UsedIds) -> Filtered = [TId|Frest]
    ; otherwise              -> Filtered = Frest),
    ceo_filter_used_ids(Irest,UsedIds,Frest).

create_exists([],Pred,NewPred) :- !,Pred=NewPred.
create_exists(Ids,Pred,NewPred) :-
    find_identifier_uses(Pred, [], Used),
    create_texpr(exists(Ids,Pred),pred,[used_ids(Used)],NewPred).
create_forall([],Pred,NewPred) :- !,Pred=NewPred.
create_forall(Ids,Pred,NewPred) :-
    find_identifier_uses(Pred, [], Used),
    split_forall_body(Pred,LHS,RHS),
    create_texpr(forall(Ids,LHS,RHS),pred,[used_ids(Used)],NewPred).
split_forall_body(b(implication(LHS,RHS),_,_),LHS,RHS) :- !.
split_forall_body(RHS,b(truth,pred,[]),RHS).

% find uses of identifiers
occurs_in_expr(ID,TExpr) :-
 (find_identifier_uses(TExpr,[],Used) -> member(ID,Used)
   ; add_failed_call_error(find_identifier_uses(TExpr,[],_)),fail).
  % TO DO: optimize so that we only look for ID; we don't have to keep track of other IDs

:- use_module(tools,[safe_sort/3]).
find_identifier_uses(TExpr,Ignored,Ids) :-
    find_typed_identifier_uses(TExpr,Ignored,TIds),
    get_texpr_ids(TIds,Ids).
find_identifier_uses_l(TExprs,Ignored,Ids) :-
    find_typed_identifier_uses_l(TExprs,Ignored,TIds),
    get_texpr_ids(TIds,Ids).

find_typed_identifier_uses(TExpr,Ignored,Ids) :-
    find_typed_identifier_uses_l([TExpr],Ignored,Ids).
find_typed_identifier_uses2(TExpr,Ignored,Ids,Rest) :-
    syntaxtraversion(TExpr,Expr,_,_,Subs,TNames),
    ( Expr = identifier(Id) ->
        ( memberchk(Id,Ignored) -> Ids=Rest
        ; otherwise             ->
            get_texpr_type(TExpr,Type),
            create_texpr(identifier(Id),Type,[],TId),
            Ids = [TId|Rest])
    ; otherwise ->
        get_texpr_ids(TNames,Names),
        append(Names,Ignored,Ignored2),
        find_typed_identifier_uses2_l(Subs,Ignored2,Ids,Rest)).
find_typed_identifier_uses2_l([],_) --> [].
find_typed_identifier_uses2_l([Expr|Rest],Ignored) -->
    find_typed_identifier_uses2(Expr,Ignored),
    find_typed_identifier_uses2_l(Rest,Ignored).

find_typed_identifier_uses_l(TExprs,Ignored,Ids) :-
    find_typed_identifier_uses2_l(TExprs,Ignored,Unsorted,[]),
    safe_sort(find_typed_identifier_uses,Unsorted,Ids).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% break predicate into components with disjunct identifiers
%  Pred: a predicate
%  Components: a list of terms component/2 where the first argument
%    is a predicate, the second is the set of used identifiers
predicate_components(Pred,Res) :-
   predicate_components_with_restriction(Pred,all,Res).

predicate_components_with_restriction(Pred,RestrictionList,Res) :-
    conjunction_to_list(Pred,Preds),
    l_predicate_identifiers(Preds,PredIds),
    ( RestrictionList=all -> R=all
    ; otherwise           -> list_to_ord_set(RestrictionList,R)),
    find_components(PredIds,R,Components),!, Components=Res.
predicate_components_with_restriction(Pred,_,[component(Pred,[])]) :-
    add_internal_error('predicate_components failed: ',Pred).

% get the list of used identifiers for each predicate
l_predicate_identifiers([],[]).
l_predicate_identifiers([Pred|PRest],[pred(Pred,Ids,_Selected)|IRest]) :-
    predicate_identifiers(Pred,Ids),
    l_predicate_identifiers(PRest,IRest).

% project a predicate : keep only those Predicates that are directly or
%  indirectly relevant for Ids; FullIds are all predicates used by ProjectedPredicate
project_predicate_on_identifiers(Pred,Ids,ProjectedPredicate,FullIds) :-
   (debug:debug_mode(on)
     -> print('project: '),translate:print_bexpr(Pred),nl, print('  on : '), print(Ids),nl
     ;  true),
   conjunction_to_list(Pred,Preds),
    l_predicate_identifiers(Preds,PredIds),
   % print(predids(PredIds)),nl,
   extract_all_predicates(Ids,all,Ids,PredIds,ProjectedPredicates,_Rest,FullIds),
   conjunct_predicates(ProjectedPredicates,ProjectedPredicate),
   (debug:debug_mode(on)
     -> print('*result: '),translate:print_bexpr(ProjectedPredicate),nl,
        print('  on : '),    print(FullIds),nl
     ; true).

%print_preds([]).
%print_preds([pred(P,_IDs,_)|T]) :- translate:print_bexpr(P), nl, print('   '), print_preds(T).

:- use_module(b_global_sets,[b_get_global_constants/1, b_get_global_sets/1]).
predicate_identifiers(Pred,Ids) :-
    b_global_sets:b_get_global_constants(EnumeratedSetCsts),
    % b_global_sets:b_get_global_enumerated_sets(GSets), % is there a reason to exclude deferred sets ??; cardinality inference,... are all done before partitioning ?
    b_global_sets:b_get_global_sets(GSets),
    append(GSets,EnumeratedSetCsts,Ignore),
    find_identifier_uses(Pred,Ignore,Ids1),  % Ignore enumerated set names
    list_to_ord_set(Ids1,Ids).

% find_components(ListOf_pred, Restrict, Out:ListOfComponents)
% role of Restrict: all if we do normal partition or List of VariableIDs on which we restrict our attention to (for Existential quantifier construction)
find_components([],_,[]).
find_components([pred(P,PIds,true)|PRest],Restrict,[component(Pred,Ids)|CRest]) :-
    % find all predicates which are using identifiers occuring in PIds
    % (and additionally those which use common identifiers )
    ( Restrict=all -> InterIDs = PIds
    ; otherwise    -> ord_intersection(PIds,Restrict,InterIDs)),
    ( InterIDs =[] ->
        % we simply copy this predicate; not with the scope of Restricted IDs
        % print(skip(PIds,P)),nl, %
        Pred=P, Ids=PIds, PRest=Rest
    ; otherwise ->
        extract_all_predicates(PIds,Restrict,PIds,PRest,Preds,Rest,Ids),
        conjunct_predicates([P|Preds],Pred)
    ),
    find_components(Rest,Restrict,CRest).
extract_all_predicates([],_,_,Preds,Found,Rest,[]) :- !,  % selecting done at end: keep same order of conjuncts
   select_predicates(Preds,Found,Rest).
extract_all_predicates(Ids,Restrict,OldIds,Preds,Found,Rest,ResultIds) :-
    % search for all predicates that directly use one of the
    % identifiers in "Ids"
    select_all_using_preds(Preds,Ids,FoundIds),
    ord_subtract(FoundIds,OldIds,NewIds),
    ( Restrict=all -> NewIdsToExtract=NewIds
    ; otherwise    -> ord_intersection(NewIds,Restrict,NewIdsToExtract)),
    ord_union(OldIds,FoundIds,OldIds2),
    % now recursively do this which the new identifiers that we have found
    extract_all_predicates(NewIdsToExtract,Restrict,OldIds2,Preds,Found,Rest,RestIds),
    ord_union([Ids,NewIds,RestIds],ResultIds).
select_all_using_preds([],_,[]).
select_all_using_preds([pred(_P,PIds,Selected)|PRest],ComponentIds,FoundIds) :-
    ( (Selected==true ; ord_disjoint(PIds,ComponentIds)) ->
        FoundIds1 = []
    ; otherwise ->         % we select the predicate for the component
        FoundIds1 = PIds, Selected=true
    ),
    select_all_using_preds(PRest,ComponentIds,FoundIds2),
    ord_union(FoundIds1,FoundIds2,FoundIds).

select_predicates([],[],[]).
select_predicates([pred(P,PIds,Selected)|T],FoundPreds,OtherPreds) :-
   (ground(Selected)
     -> FoundPreds=[P|FT], OtherPreds=OT
     ;  FoundPreds=FT, OtherPreds=[pred(P,PIds,Selected)|OT]),
   select_predicates(T,FT,OT).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% replace an identifier by an expression
replace_id_by_expr(TExpr,Id,Inserted,Replaced) :-
    remove_bt(TExpr,Expr,NewExpr,TNewExpr),
    ( Expr = identifier(Id) ->
        Replaced = Inserted
    ; otherwise ->
        Replaced = TNewExpr,
        syntaxtransformation(Expr,Subs,Names,NSubs,NewExpr),
        get_texpr_id(TId,Id),
        ( memberchk(TId,Names) ->
            NSubs=Subs
        ; otherwise ->
            replace_id_by_expr_l(Subs,Id,Inserted,NSubs))).
replace_id_by_expr_l([],_,_,[]).
replace_id_by_expr_l([Old|Orest],Id,Inserted,[New|Nrest]) :-
    replace_id_by_expr(Old,Id,Inserted,New),
    replace_id_by_expr_l(Orest,Id,Inserted,Nrest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% rename identifier

rename_bt(Expr,[],Expr) :- !.
rename_bt(OExpr,Renamings,NExpr) :-
    create_texpr(Old,Type,OInfo,OExpr),
    create_texpr(New,Type,NInfo,NExpr),
    rename_in_infos(OInfo,Renamings,NInfo),
    rename_bt2(Old,Renamings,New).
rename_bt2(identifier(Old),Renamings,identifier(New)) :-
    !, rename_id(Old,Renamings,New).
rename_bt2(OExpr,Renamings,NExpr) :-
    syntaxtransformation(OExpr,Subs,TNames,NSubs,NExpr),
    get_texpr_exprs(TNames,Names),
    remove_renamings(Names,Renamings,NRenamings),
    rename_bt_l(Subs,NRenamings,NSubs).
rename_bt_l([],_,[]).
rename_bt_l([Old|ORest],Renamings,[New|NRest]) :-
    rename_bt(Old,Renamings,New),
    rename_bt_l(ORest,Renamings,NRest).

remove_renamings([],Renamings,Renamings).
remove_renamings([identifier(Name)|Rest],Old,New) :-
    ( select(rename(Name,_),Old,Inter) -> !
    ; otherwise -> Old = Inter),
    remove_renamings(Rest,Inter,New).

rename_in_infos(Old,Renamings,New) :-
    ( has_info_to_rename(Old) ->
        maplist(rename_in_infos2(Renamings),Old,New)
    ; otherwise ->
        Old = New).
rename_in_infos2(Renamings,OInfo,NInfo) :-
    ( infos_to_rename(OInfo,OIds,NIds,NInfo) ->
        rename_ids(OIds,Renamings,NIds)
    ; otherwise ->
        OInfo = NInfo).
rename_ids([],_,[]).
rename_ids([OId|Orest],Renamings,[NId|Nrest]) :-
    rename_id(OId,Renamings,NId),
    rename_ids(Orest,Renamings,Nrest).
rename_id(Old,Renamings,New) :-
    ( memberchk(rename(Old,New),Renamings) -> true
    ; otherwise -> Old=New).
has_info_to_rename(Infos) :-
    member(I,Infos),infos_to_rename(I,_,_,_),!.
infos_to_rename(modifies(O),O,N,modifies(N)).
infos_to_rename(reads(O),O,N,reads(N)).

%remove_renamings([rename(Old,New)|Rest],Names,Renamings) :-
%    ( member(identifier(Old),Names) ->
%        Renamings = RenamingsRest
%    ; otherwise ->
%        Renamings = [rename(Old,New)|RenamingsRest]),
%    remove_renamings(Rest,Names,RenamingsRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% remove type information for transformations

remove_bt(b(Expr,Type,Infos),Expr,NExpr,b(NExpr,Type,Infos)).

%remove_bt_l([],[],[],[]).
%remove_bt_l([OT|OTRest],[O|ORest],[N|NRest],[NT|NTRest]) :-
%   remove_bt(OT,O,N,NT),
%    remove_bt_l(OTRest,ORest,NRest,NTRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% traversations
% Takes a predicate or expression or substitution and extracts:
%  the expression itself Expr, its type Type, the syntaxnode information Infos
%  + the subexpressions as a list Subs and the local identifiers declared 
syntaxtraversion(b(Expr,Type,Infos),Expr,Type,Infos,Subs,Names) :- !,
    safe_syntaxelement(Expr,Subs,Names,_,_).
syntaxtraversion(IExpr,Expr,Type,Infos,Subs,Names) :-
    add_internal_error('Not properly wrapped', syntaxtraversion(IExpr,Expr,Type,Infos,Subs,Names)),
    fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% some checks
check_if_typed_predicate(b(Pred,X,_)) :- ground(X), X=pred,
    syntaxelement(Pred,_,_,_,_,pred).
check_if_typed_expression(b(Expr,Type,_)) :-
    syntaxelement(Expr,_,_,_,_,expr),
    Type \== pred, Type \== subst, ground(Type).
check_if_typed_substitution(b(Subst,X,_)) :- ground(X), X=subst,
    syntaxelement(Subst,_,_,_,_,subst).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% transformations

syntaxtransformation(Expr,Subs,Names,NSubs,NExpr) :-
    functor(Expr,F,Arity),
    functor(NExpr,F,Arity),
    safe_syntaxelement(Expr,Subs,Names,Lists,Constant),
    all_same_length(Lists,NLists),
    syntaxelement(NExpr,NSubs,_,NLists,Constant,_).

safe_syntaxelement(Expr,Subs,Names,Lists,Constant) :-
    ( syntaxelement(Expr,SubsX,NamesX,Lists,ConstantX,_) ->
        Subs=SubsX, Names=NamesX, Constant=ConstantX
    ; otherwise ->
        functor(Expr,F,Arity),
        add_error_fail(bsyntaxtree,'Uncovered syntax element: ', F/Arity)
        ).

% syntaxelement(Expr,SubExprs,Identifiers,Lists,Constant,Type):
% Expr: the expression itself
% SubExprs: a list of sub-expressions
% Identifiers: a list of identifiers that are newly introduced (e.g. by a quantifier)
% Lists: A list of lists in the expression, to prevent infinite loops when having variable parts
% Constant: A part of the expression that is not a sub-expression (e.g. the number in integer(...))
% Type: Fundamental type of the element (predicate, expression, etc)

% predicates
syntaxelement(truth,                 [],   [], [], [], pred).
syntaxelement(falsity,               [],   [], [], [], pred).
syntaxelement(conjunct(A,B),         [A,B],[], [], [], pred).
syntaxelement(negation(A),           [A],  [], [], [], pred).
syntaxelement(disjunct(A,B),         [A,B],[], [], [], pred).
syntaxelement(implication(A,B),      [A,B],[], [], [], pred).
syntaxelement(equivalence(A,B),      [A,B],[], [], [], pred).
syntaxelement(equal(A,B),            [A,B],[], [], [], pred).
syntaxelement(not_equal(A,B),        [A,B],[], [], [], pred).
syntaxelement(member(A,B),           [A,B],[], [], [], pred).
syntaxelement(not_member(A,B),       [A,B],[], [], [], pred).
syntaxelement(subset(A,B),           [A,B],[], [], [], pred).
syntaxelement(subset_strict(A,B),    [A,B],[], [], [], pred).
syntaxelement(not_subset(A,B),       [A,B],[], [], [], pred).
syntaxelement(not_subset_strict(A,B),[A,B],[], [], [], pred).
syntaxelement(less_equal(A,B),       [A,B],[], [], [], pred).
syntaxelement(less(A,B),             [A,B],[], [], [], pred).
syntaxelement(greater_equal(A,B),    [A,B],[], [], [], pred).
syntaxelement(greater(A,B),          [A,B],[], [], [], pred).
syntaxelement(forall(Ids,D,P),       [D,P|Ids],Ids,[Ids], [], pred).
syntaxelement(exists(Ids,P),         [P|Ids],  Ids,[Ids], [], pred).
syntaxelement(finite(A),             [A],  [], [], [], pred/only_typecheck).
syntaxelement(partition(S,Es),       [S|Es],[],[Es],[],pred).
syntaxelement(kodkod(PId,Ids),       Ids,[],[Ids],PId, pred).
syntaxelement(external_pred_call(F,Args),Args,[],[Args],F,pred).

% expressions
syntaxelement(value(V),                [],   [], [], V,  expr).
syntaxelement(boolean_true,            [],   [], [], [], expr).
syntaxelement(boolean_false,           [],   [], [], [], expr).
syntaxelement(max_int,                 [],   [], [], [], expr).
syntaxelement(min_int,                 [],   [], [], [], expr).
syntaxelement(empty_set,               [],   [], [], [], expr).
syntaxelement(bool_set,                [],   [], [], [], expr).
syntaxelement(string_set,              [],   [], [], [], expr).
syntaxelement(convert_bool(A),         [A],  [], [], [], expr).
syntaxelement(add(A,B),                [A,B],[], [], [], expr).
syntaxelement(minus(A,B),              [A,B],[], [], [], expr).
syntaxelement(minus_or_set_subtract(A,B),[A,B],[], [], [], expr/only_typecheck).
syntaxelement(unary_minus(A),          [A],  [], [], [], expr).
syntaxelement(multiplication(A,B),     [A,B],[], [], [], expr).
syntaxelement(mult_or_cart(A,B),       [A,B],[], [], [], expr/only_typecheck).
syntaxelement(cartesian_product(A,B),  [A,B],[], [], [], expr).
syntaxelement(div(A,B),                [A,B],[], [], [], expr).
syntaxelement(modulo(A,B),             [A,B],[], [], [], expr).
syntaxelement(power_of(A,B),           [A,B],[], [], [], expr).
syntaxelement(successor,               [],   [], [], [], expr).
syntaxelement(predecessor,             [],   [], [], [], expr).
syntaxelement(max(A),                  [A],  [], [], [], expr).
syntaxelement(min(A),                  [A],  [], [], [], expr).
syntaxelement(card(A),                 [A],  [], [], [], expr).
syntaxelement(couple(A,B),             [A,B],[], [], [], expr).
syntaxelement(pow_subset(A),           [A],  [], [], [], expr).
syntaxelement(pow1_subset(A),          [A],  [], [], [], expr).
syntaxelement(fin_subset(A),           [A],  [], [], [], expr).
syntaxelement(fin1_subset(A),          [A],  [], [], [], expr).
syntaxelement(interval(A,B),           [A,B],[], [], [], expr).
syntaxelement(union(A,B),              [A,B],[], [], [], expr).
syntaxelement(intersection(A,B),       [A,B],[], [], [], expr).
syntaxelement(set_subtraction(A,B),    [A,B],[], [], [], expr).
syntaxelement(general_union(A),        [A],  [], [], [], expr).
syntaxelement(general_intersection(A), [A] , [], [], [], expr).
syntaxelement(relations(A,B),          [A,B],[], [], [], expr).
syntaxelement(identity(A),             [A],  [], [], [], expr).
syntaxelement(event_b_identity,        [],  [], [], [], expr). % for Rodin 1.0, TO DO: Daniel please check
syntaxelement(reverse(A),              [A],  [], [], [], expr).
syntaxelement(first_projection(A,B),   [A,B],[], [], [], expr/only_typecheck).
syntaxelement(first_of_pair(A),        [A],  [], [], [], expr).
syntaxelement(event_b_first_projection(A),[A],  [], [], [], expr/only_typecheck).
syntaxelement(event_b_first_projection_v2,[],  [], [], [], expr/only_typecheck). % for Rodin 1.0, TO DO: Daniel please check
syntaxelement(second_projection(A,B),  [A,B],[], [], [], expr/only_typecheck).
syntaxelement(event_b_second_projection_v2,[],  [], [], [], expr/only_typecheck). % for Rodin 1.0, TO DO: Daniel please check
syntaxelement(second_of_pair(A),       [A],  [], [], [], expr).
syntaxelement(event_b_second_projection(A),[A],  [], [], [], expr/only_typecheck).
syntaxelement(composition(A,B),        [A,B],[], [], [], expr).
syntaxelement(ring(A,B),               [A,B],[], [], [], expr/only_typecheck).
syntaxelement(direct_product(A,B),     [A,B],[], [], [], expr).
syntaxelement(parallel_product(A,B),   [A,B],[], [], [], expr).
syntaxelement(trans_function(A),       [A],  [], [], [], expr).
syntaxelement(trans_relation(A),       [A],  [], [], [], expr).
syntaxelement(iteration(A,B),          [A,B],[], [], [], expr).
syntaxelement(reflexive_closure(A),    [A],  [], [], [], expr).
syntaxelement(closure(A),              [A],  [], [], [], expr).
syntaxelement(domain(A),               [A],  [], [], [], expr).
syntaxelement(range(A),                [A],  [], [], [], expr).
syntaxelement(image(A,B),              [A,B],[], [], [], expr).
syntaxelement(domain_restriction(A,B), [A,B],[], [], [], expr).
syntaxelement(domain_subtraction(A,B), [A,B],[], [], [], expr).
syntaxelement(range_restriction(A,B),  [A,B],[], [], [], expr).
syntaxelement(range_subtraction(A,B), [A,B],[], [], [], expr).
syntaxelement(overwrite(A,B),          [A,B],[], [], [], expr).
syntaxelement(partial_function(A,B),   [A,B],[], [], [], expr).
syntaxelement(total_function(A,B),     [A,B],[], [], [], expr).
syntaxelement(partial_injection(A,B),  [A,B],[], [], [], expr).
syntaxelement(total_injection(A,B),    [A,B],[], [], [], expr).
syntaxelement(partial_surjection(A,B), [A,B],[], [], [], expr).
syntaxelement(total_surjection(A,B),   [A,B],[], [], [], expr).
syntaxelement(total_bijection(A,B),    [A,B],[], [], [], expr).
syntaxelement(partial_bijection(A,B),  [A,B],[], [], [], expr).
syntaxelement(total_relation(A,B),     [A,B],[], [], [], expr).
syntaxelement(surjection_relation(A,B),[A,B],[], [], [], expr).
syntaxelement(total_surjection_relation(A,B),[A,B],[], [], [], expr).
syntaxelement(seq(A),                  [A],  [], [], [], expr).
syntaxelement(seq1(A),                 [A],  [], [], [], expr).
syntaxelement(iseq(A),                 [A],  [], [], [], expr).
syntaxelement(iseq1(A),                [A],  [], [], [], expr).
syntaxelement(perm(A),                 [A],  [], [], [], expr).
syntaxelement(empty_sequence,          [],   [], [], [], expr).
syntaxelement(size(A),                 [A],  [], [], [], expr).
syntaxelement(first(A),                [A],  [], [], [], expr).
syntaxelement(last(A),                 [A],  [], [], [], expr).
syntaxelement(front(A),                [A],  [], [], [], expr).
syntaxelement(tail(A),                 [A],  [], [], [], expr).
syntaxelement(rev(A),                  [A],  [], [], [], expr).
syntaxelement(concat(A,B),             [A,B],[], [], [], expr).
syntaxelement(insert_front(A,B),       [A,B],[], [], [], expr).
syntaxelement(insert_tail(A,B),        [A,B],[], [], [], expr).
syntaxelement(restrict_front(A,B),     [A,B],[], [], [], expr).
syntaxelement(restrict_tail(A,B),      [A,B],[], [], [], expr).
syntaxelement(general_concat(A),       [A],  [], [], [], expr).
syntaxelement(function(A,B),           [A,B],[], [], [], expr).
syntaxelement(external_function_call(F,Args),Args,[],[Args],F,expr).
syntaxelement(identifier(I),           [],   [], [], I,  expr).
syntaxelement(integer(I),              [],   [], [], I,  expr).
syntaxelement(integer_set(T),          [],   [], [], T,  expr).
syntaxelement(string(S),               [],   [], [], S,  expr).
syntaxelement(set_extension(L),        L,    [], [L], [], expr).
syntaxelement(sequence_extension(L),   L,    [], [L], [], expr).
syntaxelement(comprehension_set(Ids,P),[P|Ids],  Ids,[Ids], [], expr).
syntaxelement(event_b_comprehension_set(Ids,E,P),[E,P|Ids],  Ids,[Ids], [], expr/only_typecheck).
syntaxelement(lambda(Ids,P,E),         [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(general_sum(Ids,P,E),             [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(general_product(Ids,P,E),         [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(quantified_union(Ids,P,E),        [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(quantified_intersection(Ids,P,E), [P,E|Ids],Ids,[Ids], [], expr).
syntaxelement(struct(Rec), [Rec], [], [], [], expr).
syntaxelement(rec(Fields), FContent, [], [FContent], FNames, expr) :- syntaxfields(Fields,FContent, FNames).
syntaxelement(record_field(R,I), [R], [], [], I, expr).
syntaxelement(assertion_expression(Cond,ErrMsg,Expr), [Cond,Expr], [], [], ErrMsg, expr).

% substitutions
syntaxelement(skip,                    [],   [], [], [], subst).
syntaxelement(precondition(A,B),       [A,B],[], [], [], subst).
syntaxelement(assertion(A,B),          [A,B],[], [], [], subst).
syntaxelement(if_elsif(A,B),           [A,B],[], [], [], subst/elsif).
syntaxelement(while(A,B,C,D),          [A,B,C,D],[], [], [], subst).
% used only internally in the interpreter, contains last value of variant:
syntaxelement(while1(A,B,C,D,E),          [A,B,C,D],[], [], E, subst).
syntaxelement(select_when(A,B),        [A,B],[], [], [], subst/when).
syntaxelement(block(S),[S],[],[],[], subst/only_typecheck).
syntaxelement(assign(Lhs,Rhs),Exprs,[],[Lhs,Rhs], [], subst) :- append(Lhs,Rhs,Exprs).
syntaxelement(assign_single_id(Id,Rhs),[Id,Rhs],[],[], [], subst).
syntaxelement(any(Ids,P,S),[P,S|Ids],Ids,[Ids], [], subst).
syntaxelement(var(Ids,S),  [S|Ids],  Ids,[Ids], [], subst).
syntaxelement(if(Ifs),   Ifs,  [], [Ifs], [], subst).
syntaxelement(parallel(Ss), Ss, [], [Ss], [], subst).
syntaxelement(sequence(Ss), Ss, [], [Ss], [], subst).
syntaxelement(becomes_element_of(Ids,E), [E|Ids], [], [Ids], [], subst).
syntaxelement(becomes_such(Ids,P), [P|Ids], [], [Ids], [], subst).
syntaxelement(evb_becomes_such(Ids,P), [P|Ids], [], [Ids], [], subst/only_typecheck).
syntaxelement(evb2_becomes_such(Ids,P), [P|Ids], [], [Ids], [], subst/only_typecheck).
syntaxelement(let(Ids,P,S), [P,S|Ids], Ids, [Ids], [], subst).
syntaxelement(operation_call(Id,Rs,As), [Id|Exprs], [], [Rs,As], [], subst) :- append(Rs,As,Exprs).
syntaxelement(case(E,Eithers,Else), [E,Else|Eithers], [], [Eithers], [], subst).
syntaxelement(case_or(Es,S), [S|Es], [], [Es], [], subst/caseor).
syntaxelement(choice(Ss), Ss, [], [Ss], [], subst).
syntaxelement(select(Whens), Whens, [], [Whens], [], subst).
syntaxelement(select(Whens,Else), [Else|Whens], [], [Whens], [], subst).
syntaxelement(operation(I,Rs,As,B), [I,B|Ids], Ids, [Rs,As], [], subst) :- append(Rs,As,Ids).
syntaxelement(external_subst_call(F,Args),Args,[],[Args],F,subst). 

% syntax for Event-B events
syntaxelement(rlevent(I,Sec,St,Ps,G,Ts,As,VWs,PWs,Ums,Rs), Subs, [], [Ps,Ts,As,VWs,PWs,Ums,Rs], [I,Sec], subst) :-
    append([[St],Ps,[G],Ts,As,VWs,PWs,Ums,Rs],Subs).
syntaxelement(witness(I,P), [I,P], [], [], [], witness).

% extended syntax for Z
syntaxelement(let_predicate(Ids,As,Pred), Exprs, Ids, [Ids,As], [], pred) :- append([Ids,As,[Pred]],Exprs).
syntaxelement(let_expression(Ids,As,Expr), Exprs, Ids, [Ids,As], [], expr) :- append([Ids,As,[Expr]],Exprs).
syntaxelement(if_then_else(If,Then,Else),[If,Then,Else], [], [], [], expr).
syntaxelement(compaction(A), [A],  [], [], [], expr).
syntaxelement(mu(A),         [A],  [], [], [], expr).
syntaxelement(bag_items(A),  [A],  [], [], [], expr).

syntaxelement(freetype_set(Id), [], [], [], Id, expr).
syntaxelement(freetype_case(Type,Case,Expr), [Expr], [], [], [Type,Case], pred).
syntaxelement(freetype_constructor(Type,Case,Expr), [Expr], [], [], [Type,Case], expr).
syntaxelement(freetype_destructor(Type,Case,Expr), [Expr], [], [], [Type,Case], expr).

syntaxelement(ordinary,    [], [], [], [], status).
syntaxelement(anticipated(Variant), [Variant], [], [], [], status).
syntaxelement(convergent(Variant),  [Variant], [], [], [], status).


% fields of records
syntaxfields([],[],[]).
syntaxfields([field(N,C)|Rest],[C|CRest],[N|NRest]) :- syntaxfields(Rest,CRest,NRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% pattern definitions (not yet finished)

bsyntax_pattern(Expr,TExpr) :-
    var(Expr),!,Expr=TExpr.
bsyntax_pattern(-Expr,TExpr) :-
    !,remove_all_infos(Expr,TExpr).
bsyntax_pattern(Expr,TExpr) :-
    functor(Expr,b,3),!,Expr=TExpr.
bsyntax_pattern(Expr:Type/Info,TExpr) :-
    !,bsyntax_pattern2(Expr,Type,Info,TExpr).
bsyntax_pattern(Expr:Type,TExpr) :-
    !,bsyntax_pattern2(Expr,Type,_Info,TExpr).
bsyntax_pattern(Expr/Info,TExpr) :-
    !,bsyntax_pattern2(Expr,_Type,Info,TExpr).
bsyntax_pattern(Expr,TExpr) :-
    !,bsyntax_pattern2(Expr,_Type,_Info,TExpr).

bsyntax_pattern2(Pattern,Type,Info,TExpr) :-
    functor(Pattern,Functor,Arity),
    functor(R,Functor,Arity),
    create_texpr(R,Type,Info,TExpr),
    syntaxelement(Pattern,PSubs,_,PLists,Const,EType),
    syntaxelement(R,RSubs,_,RLists,Const,EType),
    ( EType==pred -> Type=pred
    ; EType==subst -> Type=subst
    ; otherwise -> true),
    all_same_length(PLists,RLists),
    maplist(bsyntax_pattern,PSubs,RSubs).

all_same_length([],[]).
all_same_length([A|Arest],[B|Brest]) :-
    ( var(A),var(B) ->
        add_error_fail(bsyntaxtree,'At least one list should be nonvar',[A,B])
    ; otherwise ->
        same_length(A,B)),
    all_same_length(Arest,Brest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% strip an AST into a more compact form (without b/3 terms)

strip_and_norm_ast(TExpr,SNExpr) :-
    get_texpr_expr(TExpr,Expr),
    syntaxtransformation(Expr,Subs,_,NSubs,SExpr),
    strip_and_norm_ast_l(Subs,NSubs),
    norm_strip(SExpr,SNExpr).
strip_and_norm_ast_l([],[]).
strip_and_norm_ast_l([TExpr|Trest],[NExpr|Nrest]) :-
    strip_and_norm_ast(TExpr,NExpr),
    strip_and_norm_ast_l(Trest,Nrest).
norm_strip(greater_equal(A,B),lesser_equal(B,A)) :- !.
norm_strip(greater(A,B),lesser(B,A)) :- !.
norm_strip(Old,New) :-
    functor(Old,Functor,2),
    is_commutative(Functor),!,
    Old =.. [Functor,OA,OB],
    New =.. [Functor,NA,NB],
    ( OA @>= OB -> OA=NB,OB=NA
    ; otherwise -> OA=NA,OB=NB).
norm_strip(Old,Old).

is_commutative(conjunct).
is_commutative(disjunct).
is_commutative(equivalence).
is_commutative(equal).
is_commutative(add).
is_commutative(multiplication).
is_commutative(union).
is_commutative(intersection).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% common subexpression detection

:- assert_must_succeed(
   (   find_common_subexpressions(b(identifier(i),integer,[]),0,R),
       R == b(identifier(i),integer,[ref(0,1,top)])
   )).
:- assert_must_succeed(
   (   I = b(identifier(i),integer,[]),
       O = b(identifier(i),integer,[ref(0,2,top)]),
       find_common_subexpressions(b(add(I,I),integer,[]),0,R),
       R == b(add(O,O),integer,[ref(1,1,top)])
   )).
% two quantifiers with the same variable name must not share
% expressions that contain the variable, other expressions should be
% shared
:- assert_must_succeed(
   (   C = b(identifier(c),integer,[]),
       X = b(identifier(x),integer,[]),
       I = b(identifier(i),set(integer),[]),
       O = b(integer(1),integer,[]),
       %  !x.(x:I => x<c+1)
       F = b(forall([X],
                    b(member(X,I),pred,[]),
                    b(less(X,b(add(C,O),integer,[])),pred,[])),pred,[]),
       %  !x.(x:I => x>c+1)
       Fx = b(forall([X],
                    b(member(X,I),pred,[]),
                    b(greater(X,b(add(C,O),integer,[])),pred,[])),pred,[]),
       C1 = b(identifier(c),integer,[ref(3,2,top)]),
       X1 = b(identifier(x),integer,[ref(0,3,7)]),
       I1 = b(identifier(i),set(integer),[ref(1,2,top)]),
       O1 = b(integer(1),integer,[ref(4,2,top)]),
       F1 = b(forall([X1],
                     b(member(X1,I1),pred,[ref(2,1,7)]),
                     b(less(X1,b(add(C1,O1),integer,[ref(5,2,top)])),pred,[ref(6,1,7)])),pred,[ref(7,1,top)]),
       % e.g. c+1 should be shared (ref. 5)
       % x and x:I must not be shared, they get a new reference (8 and 9)
       X2 = b(identifier(x),integer,[ref(8,3,11)]),
       F2 = b(forall([X2],
                     b(member(X2,I1),pred,[ref(9,1,11)]),
                     b(greater(X2,b(add(C1,O1),integer,[ref(5,2,top)])),pred,[ref(10,1,11)])),pred,[ref(11,1,top)]),
       find_common_subexpressions(b(conjunct(F,Fx),pred,[]),0,R),
       R == b(conjunct(F1,F2),pred,[ref(12,1,top)])
   )).

% find_common_subexpressions/3 searches for common subexpressions and adds an
%   information field of the form ref(Id,Counter,Level) to each AST node.
% find_common_subexpressions(+TExpr,+Start,-NTExpr):
%   TExpr: the original AST
%   Start: at which number the IDs should start (usually 0)
%   NTExpr: the AST, each node annotated with an ref(ID,Counter,SubId) information field
%      two syntax nodes share the same ID when they should yield always the same
%      value. Counter counts the number of occurrences of the same expression in
%      the given AST. SubId is the unique reference of the quantified expression
%      where a used ID is introduced, top there is no such quantifier.
%      (Please note: The declaration of a variable in a quantified expression
%       is an occurrence of its own.)
find_common_subexpressions(TExpr,Start,NTExpr) :-
    create_stores(Start,EmptyStores),
    find_common_subexpressions2(TExpr,EmptyStores,NTExpr,Stores,_Stripped,_UsedIds),
    ground_reference_counters(Stores).
find_common_subexpressions2(TExpr,InStores,NTExpr,OutStores,Stripped,UsedIds) :-
    % decompose the typed expression into it components (Subs), look up the newly introduced
    % identifiers (Ids) and create a new typed expression (NTExpr) with the modified components
    % (NSubs) and the new Info field (Info)
    fc_parts(TExpr,Expr,Subs,Ids,NSubs,Info,NTExpr),
    % introduce a map from expression to reference that is only visible to this node
    % and it subnodes if new identifiers (Ids) are introduced
    % OutStores1 and OutStores contain a patter such that if OutStores1 is set,
    % OutStores is exactly like OutStores1 but without visibility of the newly introduced
    % identifiers.
    create_substore(Ids,ID,InStores,SubStores),
    % apply subexpression detection on the subexpressions
    find_common_subexpressions_l(Subs,SubStores,NSubs,SubStores1,StrippedSubs,UsedSubIds),
    % create a stripped (i.e. no info field, no types) version of the syntax node for look ups
    create_stripped(Expr,StrippedSubs,Stripped),
    % merge the used ids of the subnodes with the used ids in this node,
    % subtract newly introduced identifiers because they are not visible from outside
    create_used_ids(Expr,Ids,UsedSubIds,UsedIds),
    % if a substore was introduced due to newly introduced identifiers, remove it
    remove_last_substore(Ids,SubStores1,Stores1),
    % try to find the expression in the stores
    ( lookup_expression(Stripped,UsedIds,Stores1,OutStores,Info,ID) ->
        true
    ; otherwise ->
        % if not found, store the expression
        store_expression(Stripped,UsedIds,Stores1,OutStores,Info,ID)).

fc_parts(TExpr,Expr,Subs,Ids,NSubs,Info,NTExpr) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_type(TExpr,Type),
    get_texpr_info(TExpr,Infos),
    syntaxtransformation(Expr,Subs,Names,NSubs,NExpr), get_texpr_ids(Names,Ids),
    create_texpr(NExpr,Type,[Info|Infos],NTExpr).

find_common_subexpressions_l([],Stores,[],Stores,[],[]).
find_common_subexpressions_l([Expr|Erest],InStores,[NExpr|Nrest],OutStores,[Stripped|Srest],UsedIds) :-
    find_common_subexpressions2(Expr,InStores,NExpr,InterStores,Stripped,UsedIds1),
    find_common_subexpressions_l(Erest,InterStores,Nrest,OutStores,Srest,UsedIds2),
    ord_union(UsedIds1,UsedIds2,UsedIds).

% the stores data type has the following form:
% stores(Counter,References,Top,Substores)
% - Counter is the next ID for new expressions, the same expressions
%   share the same id
% - Complete is a mapping from each expression to its id (is this sensible?)
% - References is a mapping from an ID to a term rc(C,R) where C
%   contains the number of occurrences of the expressions so far and R the
%   final number of occurrences of this expression. Until the end, R is a variable
%   that will be unified with the last value of C
% - Top is a mapping from a stripped expression to an ID. This mapping is for all
%   expressions that do not contain references to quantified variables
% - Substores is a list of terms substore(Ids,SubStore) where Ids is a list of
%   introduced identifiers in a qualified expression and Substore is a mapping from
%   a stripped expression to its ID. The expression contains references to identifiers
%   in Ids. In case of nested quantified expressions, the inner expression has
%   a substore entry before the outer expression in the list.
create_stores(Start,stores(Start,Empty,Empty,[])) :-
    empty_avl(Empty).
create_substore([],_SubId,Stores,Stores) :- !.
create_substore(Ids,SubId,
                stores(Counter,IR,Top, InSubs),
                stores(Counter,IR,Top, [substore(OrdIds,Empty,SubId)|InSubs])) :-
    list_to_ord_set(Ids,OrdIds),
    empty_avl(Empty).

remove_last_substore([],Stores,Stores).
remove_last_substore([_|_],
                     stores(Counter,IR,Top,[_Substore|Subs]),
                     stores(Counter,IR,Top,Subs)).

lookup_expression(Stripped,UsedIds,InStores,OutStores,ref(Id,RefCounter,Level),Id) :-
    InStores =  stores(Counter,InRefCounter, TopStore,SubStores),
    OutStores = stores(Counter,OutRefCounter,TopStore,SubStores),
    lookup_expression2(SubStores,TopStore,Stripped,UsedIds,Id,Level),!,
    avl_fetch(Id,InRefCounter,rc(OC,RefCounter)),
    NC is OC+1,
    avl_store(Id,InRefCounter,rc(NC,RefCounter),OutRefCounter).
lookup_expression2([],TopStore,Stripped,_UsedIds,Id,top) :-
    avl_fetch(Stripped,TopStore,Id).
lookup_expression2([substore(_Names,Store,SubId)|_Rest],_TopStore,Stripped,_UsedIds,Id,SubId) :-
    avl_fetch(Stripped,Store,Id),!.
lookup_expression2([substore(Names,_Store,_SubId)|Rest],TopStore,Stripped,UsedIds,Id,Level) :-
    ord_disjoint(UsedIds,Names),!,
    lookup_expression2(Rest,TopStore,Stripped,UsedIds,Id,Level).

store_expression(Stripped,UsedIds,InStores,OutStores,ref(Id,RefCounter,Level),Id) :-
    InStores = stores(Id,InRefCounter,InTopStore,InSubStores),
    Counter is Id+1,
    OutStores = stores(Counter,OutRefCounter,OutTopStore,OutSubStores),
    store_expression2(InSubStores,InTopStore,Stripped,UsedIds,Id,OutSubStores,OutTopStore,Level),
    avl_store(Id,InRefCounter,rc(1,RefCounter),OutRefCounter).
store_expression2([],InTopStore,Stripped,_UsedIds,Id,[],OutTopStore,top) :-
    avl_store(Stripped,InTopStore,Id,OutTopStore).
store_expression2([substore(Names,InStore,SubId) |Srest],TopStore,Stripped,UsedIds,Id,
                  [substore(Names,OutStore,SubId)|Srest],TopStore,SubId) :-
    ord_intersect(Names,UsedIds),!,
    avl_store(Stripped,InStore,Id,OutStore).
store_expression2([Substore|ISrest],InTopStore,Stripped,UsedIds,Id,
                  [Substore|OSrest],OutTopStore,Level) :-
    store_expression2(ISrest,InTopStore,Stripped,UsedIds,Id,OSrest,OutTopStore,Level).

create_used_ids(Expr,Ids,UsedSubIds,UsedIds) :-
    create_used_ids2(Expr,UsedSubIds,UsedIds1),
    ord_subtract(UsedIds1,Ids,UsedIds).
create_used_ids2(identifier(Id),UsedSubIds,UsedIds) :- !,
    ord_add_element(UsedSubIds,Id,UsedIds).
create_used_ids2(_Expr,UsedSubIds,UsedIds) :- !, UsedSubIds=UsedIds.

create_stripped(Expr,StrippedSubs,Stripped) :-
    syntaxtransformation(Expr,_Subs,_Names,StrippedSubs,Stripped).

ground_reference_counters(stores(_Counter,RefCounter,_TopStore,_SubStores)) :-
    avl_to_list(RefCounter,List),
    maplist(ground_reference_counter,List).
ground_reference_counter(_Id-rc(C,C)).



% check if a ProB type is a set type:
is_set_type(set(Type),Type).
is_set_type(seq(Type),couple(integer,Type)).

get_texpr_set_type(X,Res) :- get_texpr_type(X,TypeX), 
  (is_set_type(TypeX,Res) -> true ; add_error_fail(get_texpr_set_type,'Not a set type: ',X)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check if a set contains all elements of a type, i.e., maximal type
is_just_type(b(E,T,I)) :- is_just_type3(E,T,I).
is_just_type3(identifier(G),set(global(G)),Infos) :- member(given_set,Infos),!.
is_just_type3(freetype_set(_),_,_).
is_just_type3(pow_subset(E),_,_) :- is_just_type(E).
is_just_type3(fin_subset(E),T,_) :- \+ is_infinite_type(T),is_just_type(E).
is_just_type3(integer_set('INTEGER'),set(integer),_).
is_just_type3(bool_set,set(boolean),_).
is_just_type3(string_set,set(string),_).
is_just_type3(cartesian_product(A,B),_,_) :-
	is_just_type(A),is_just_type(B).
is_just_type3(mult_or_cart(A,B),_,_) :- % is_just_type/1 could be called before the cleanup
	is_just_type(A),is_just_type(B).    % of an expression has finished
is_just_type3(relations(A,B),_,_) :-
	is_just_type(A),is_just_type(B).
is_just_type3(struct(b(rec(Fields),_,_)),_,_) :-
	all_fields_are_just_types(Fields).
all_fields_are_just_types([]).
all_fields_are_just_types([field(_,Set)|Rest]) :-
	is_just_type(Set),
	all_fields_are_just_types(Rest).
