% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(cbc_refinement_checks, [cbc_refinement_check/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,cbc).
:- module_info(description,'This module provides constraint-based checking of B and Event-B refinements.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).
:- use_module(library(sets), [subtract/3]).

:- use_module(specfile, [animation_minor_mode/1]).
:- use_module(bmachine, [full_b_machine/1,
			 b_get_refined_machine/1,
			 b_get_machine_operation/5,
			 b_get_invariant_from_machine/1,
			 b_get_linking_invariant_from_machine/1,
			 b_get_properties_from_machine/1,
			 b_get_machine_constants/1,
			 b_get_machine_variables/1,
			 b_machine_name/1]).
:- use_module(bsyntaxtree, [get_texpr_expr/2,get_texpr_info/2,
			    get_texpr_id/2,get_texpr_ids/2,
			    get_rodin_name/2,
			    conjunct_predicates/2,
			    conjunction_to_list/2,
			    create_texpr/4,
			    rename_bt/3,
			    create_exists_opt/3,
		            find_typed_identifier_uses/3]).
:- use_module(b_interpreter, [set_up_typed_localstate/5]).
:- use_module(bmachine_structure, [get_section/3]).
:- use_module(bmachine_eventb, [prime_variable/2,
				prime_variables/2]).
:- use_module(tools, [ajoin/2,flatten/2,assert_once/1]).
:- use_module(self_check, [time_out_with_factor_call/3]).
:- use_module(error_manager, [add_error_fail/3]).
:- use_module(b_state_model_check, [negate/2]).
:- use_module(weakest_preconditions, [weakest_precondition/3]).
:- use_module(tools,[safe_sort/3]).

:- dynamic errors_found/1.

cbc_refinement_check(List,ErrorsWereFound) :-
    retractall(errors_found(_)),
    (animation_minor_mode(eventb)
    -> eventb_refinement_check(List)
    ; classical_refinement_check(List)),
    (errors_found(true) -> ErrorsWereFound = true ;
     errors_found(timeout) -> ErrorsWereFound = time_out ;
     otherwise -> ErrorsWereFound = false).

check_predicate(Pred,State,Res) :-
    % find all identifiers used in the predicate and create a state
    % was done by the caller, leading to code duplication
    find_typed_identifier_uses(Pred,[],Identifiers),
    set_up_typed_localstate_for_pred(Identifiers,Pred,TypedVals,State),
    b_interpreter_components:reset_component_info(false),
    time_out_with_factor_call(
        (b_interpreter_components:b_trace_test_components(Pred,State,TypedVals) -> Res = yes ; Res = no),
        10, Res='timeout'),
    (Res = yes -> assert_once(errors_found(true)) ;
     Res = timeout -> assert_once(errors_found(timeout)) ;
     otherwise -> true).

set_up_typed_localstate_for_pred(CV,Pred,TypedVals,State) :-
   (preferences:preference(use_static_ordering,true)
     -> static_ordering:sort_ids_by_usage(CV,Pred,SortedVarlist,no_warnings)
        %,reverse(Constants1,Constants)
        %,print(sorted_by_usage(Constants)),nl
     ;  SortedVarlist =  CV),
   set_up_typed_localstate(SortedVarlist,_FreshVars,TypedVals,[],State).

% -------------------------------------------
% eventb po generation
% -------------------------------------------

eventb_refinement_check(ResultList) :-
    % fetch Axioms and Invariant/properties first
    full_b_machine(Machine),
    (get_section(axioms,Machine,Axioms) -> true ; Axioms = []),
    b_get_invariant_from_machine(Invariant),
    b_get_properties_from_machine(Properties),
    % evt/p/WFIS
    witness_feasibility(Axioms,Invariant,Properties,ListWF),
    % evt/grd/GRD proof obligations
    guard_strengthening(Axioms,Invariant,Properties,ListGS),
    % evt/act/SIM
    action_simulation(Axioms,Invariant,Properties,ListAS),
    flatten(ListWF,List1), flatten(ListGS,List2), flatten(ListAS,List3),
    append([List1,['\n'],List2,['\n'],List3],ResultList).

witness_feasibility(Axioms,Invariant,Properties,['Witness Feasibility:'|Results]) :-
    % action simulation:
    % A & I & J & H & W & Tau => Q_i
    % find all candidates for evt/act/SIM proof obligations
    findall(RLEVENT,
            (b_get_machine_operation(_Name,_R,_P,Body,_OType),
             get_texpr_expr(Body,RLEVENT)),
             ListOfOperations),
    maplist(witness_feasibility_aux(Axioms,Invariant,Properties),ListOfOperations,Results).

witness_feasibility_aux(Axioms,Invariant,Properties,
		  rlevent(Name,_Section,_Stat,Parameters,Guards,_Thm,_Actions,_VWit,PWit,_Unmod,AbsEvents),
		  ResultsFlat) :-
    % feasibility for parameters - the abstract events have a parameter
    % that is not a parameter of the concrete events
    all_abstract_parameters(AbsEvents,AllAbsParameters),
    subtract(AllAbsParameters,Parameters,ParametersForWfisPos),

    conjunct_predicates([Invariant,Properties,Guards|Axioms],LHSPred),

    % map over all abstract events
    maplist(witness_feasibility_parameters(Name,LHSPred,PWit),ParametersForWfisPos,ResultsParameters),
    flatten(ResultsParameters,ResultsFlat).

witness_feasibility_parameters(Name,LHSPred,PWit,Parameter,Result) :-
    % find matching witness - same identifier
    get_texpr_id(Parameter,Identifier),
    member(Witness,PWit), get_texpr_expr(Witness,witness(WitnessIdentifier,Pred)),
    get_texpr_id(WitnessIdentifier,Identifier), !,

    % set up PO: A & I & J & H => exists p : W_p
    create_exists_opt([Parameter],[Pred],ExistsPred),
    negate(ExistsPred,NegatedExistsPred),
    conjunct_predicates([LHSPred,NegatedExistsPred],CounterExamplePred),

    check_predicate(CounterExamplePred,State,Res),

    % result string
    ajoin([Name,'/',Identifier,'/WFIS: ','Counter-Example found: ', Res],Result),
    % print
    write(Result), nl, translate:nested_print_bexpr(Pred),
    write(State),nl.

action_simulation(Axioms,Invariant,Properties,['Action Simulation:'|Results]) :-
    % action simulation:
    % A & I & J & H & W & Tau => Q_i
    % find all candidates for evt/act/SIM proof obligations
    findall(RLEVENT,
            (b_get_machine_operation(_Name,_R,_P,Body,_OType),
             get_texpr_expr(Body,RLEVENT)),
             ListOfOperations),
    maplist(action_simulation(Axioms,Invariant,Properties),ListOfOperations,Results).

action_simulation(Axioms,Invariant,Properties,
		  rlevent(Name,_Section,_Stat,_Parameters,Guards,_Thm,Actions,VWit,PWit,_Unmod,AbsEvents),
		  ResultsFlat) :-
    % concrete before-after-predicate
    maplist(before_after_predicate,Actions,Tau),
    maplist(extract_witness_predicate,VWit,VWitPreds),
    maplist(extract_witness_predicate,PWit,PWitPreds),
    flatten([Axioms,Invariant,Properties,Guards,VWitPreds,PWitPreds,Tau],LHSPreds),
    conjunct_predicates(LHSPreds,LHS),

    % map over all abstract events
    maplist(action_simulation2(Name,LHS),AbsEvents,Results),
    flatten(Results,ResultsFlat).

action_simulation2(Name,LHS,AbsEvent,Result) :-
    get_texpr_expr(AbsEvent,rlevent(_AbsName,_Section,_Stat,_AbsParameters,_Guards,_Thm,Actions,_VWit,_PWit,_Unmod,_AbsEvents)),
    % map over all abstract actions
    maplist(action_simulation3(Name,LHS),Actions,Result).

action_simulation3(Name,LHS,Action,Result) :-
    modifies_concrete_var(Action), !,
    get_rodin_name(Action,ActionName),

    % before after predicate of abstract action, negate, solve
    before_after_predicate(Action,BAPred),
    negate(BAPred,NegatedBAPred),
    conjunct_predicates([NegatedBAPred,LHS],Pred),

    check_predicate(Pred,State,Res),

    % result string
    ajoin([Name,'/',ActionName,'/SIM: ','Counter-Example found: ', Res],Result),
    % print
    write(Result), nl, translate:nested_print_bexpr(Pred),
    write(State),nl.

% abstract action does not modify a concrete variable -> no PO
action_simulation3(_Name,_LHS,_Action,[]).

guard_strengthening(Axioms,Invariant,Properties,['Guard Strengthening:'|Results]) :-
    % guard strengthening:
    % A & I & J & H & W => G_i
    % guard strengthening (merge):
    % A & I & J & H & W & Tau => G_1 or ... or G_n 
    % find all candidates for evt/grd/GRD proof obligations
    findall(RLEVENT,
            (b_get_machine_operation(_Name,_R,_P,Body,_OType),
             get_texpr_expr(Body,RLEVENT)),
             ListOfOperations),
    maplist(guard_strengthening_aux(Axioms,Invariant,Properties),ListOfOperations,Results).

% guard_strengthening: not a refinement
guard_strengthening_aux(_Axioms,_Invariant,_Properties,
		    rlevent(_Name,_Section,_Stat,_Parameters,_Guards,_Thm,_Actions,_VWit,_PWit,_Unmod,[]),
		    []) :- !. % nothing to check


% guard_strengthening: refining only one event
guard_strengthening_aux(Axioms,Invariant,Properties,
		    rlevent(Name,_Section,_Stat,_Parameters,Guards,_Thm,_Actions,VWit,PWit,_Unmod,[AbsEvent]),
		    Results) :- !,
    get_texpr_expr(AbsEvent,ABSRLEVENT),
    ABSRLEVENT = rlevent(_AbsName,_Absection,_AbsStat,_AbsParameters,AbsGrds,_AbsThm,_AbsAct,_AbsVWit,_AbsPWit,_AbsUnmod,_AbsEvents),
    conjunction_to_list(AbsGrds,AbsGrdsList),
    % map over all abstract guards
    maplist(guard_strengthening_single_event(Name,Axioms,Invariant,Properties,VWit,PWit,Guards),AbsGrdsList,Results).

% guard_strengthening: refining multiple events
guard_strengthening_aux(Axioms,Invariant,Properties,
		    rlevent(Name,_Section,_Stat,_Parameters,Guards,_Thm,Actions,VWit,PWit,_Unmod,AbsEvents),
		    Result) :-
    % concrete before-after-predicate
    maplist(before_after_predicate,Actions,Tau),
    % collect all abstract guards G_i and negate
    findall(AbsGrds,
	    (member(Evt,AbsEvents),
	     get_texpr_expr(Evt,rlevent(_AbsName,_AbsSection,_AbsStat,_AbsParams,AbsGrds,_AbsThm,_AbsAct,_AbsVWit,_AbsPWit,_AbsUnmod,_AbsAbsEvents))),
	    ListOfAbstractGuards),
    maplist(negate,ListOfAbstractGuards,NegatedAbstractGuards),

    % build up predicate A & I & J & H & W & Tau => G_1 or ... or G_n
    maplist(extract_witness_predicate,VWit,VWitPreds),
    maplist(extract_witness_predicate,PWit,PWitPreds),

    flatten([Axioms,Invariant,Properties,Guards,VWitPreds,PWitPreds,Tau,NegatedAbstractGuards],LHSPred),
    conjunct_predicates(LHSPred,Pred),
    check_predicate(Pred,State,Res),

    % result string and print
    ajoin([Name,'/MRG: ','Counter-Example found: ', Res],Result),
    write(Result), nl, translate:nested_print_bexpr(Pred),
    write(State),nl.

guard_strengthening_single_event(OpName,_Axioms,_Invariant,_Properties,_VWit,_PWit,_Guards,AbsGrd,Result) :-
    get_rodin_name(AbsGrd,GuardName),
    format('~nChecking single event guard strengthening for: ~w of ~w~n',[GuardName,OpName]),
    b_machine_name(MainMachineName), % we assume that we prove the guard strengthening for the main machine !
    bmachine:discharged_guard_strengthening(MainMachineName,AbstractEvent,GuardName,OpName),
    format('Guard ~w of ~w is already proven for event ~w:~w !~n',[GuardName,AbstractEvent,MainMachineName,OpName]),
    preferences:get_preference(use_po,true),
    !,
    ajoin([OpName,'/',GuardName,'/GRD: ','already proven'],Result).
guard_strengthening_single_event(OpName,Axioms,Invariant,Properties,VWit,PWit,Guards,AbsGrd,Result) :-
    get_rodin_name(AbsGrd,GuardName),
    % build up predicate and check
    maplist(extract_witness_predicate,VWit,VWitPreds),
    maplist(extract_witness_predicate,PWit,PWitPreds),
    negate(AbsGrd,NegAbsGrd),

    flatten([Axioms,Invariant,Properties,Guards,VWitPreds,PWitPreds,NegAbsGrd],Predicates),
    conjunct_predicates(Predicates,Pred),

    check_predicate(Pred,State,Res),

    % result string and print
    ajoin([OpName,'/',GuardName,'/GRD: ','Counter-Example found: ', Res],Result),
    write(Result), nl, translate:nested_print_bexpr(Pred),
    write(State),nl.

extract_witness_predicate(Witness,Pred) :-
   get_texpr_expr(Witness,witness(_Id,Pred)).

modifies_concrete_var(Action) :-
    get_texpr_info(Action,Info),
    member(modifies(ListOfVars), Info),
    some(is_concrete_var,ListOfVars).

is_concrete_var(VarName) :-
    b_get_machine_variables(Variables),
    b_machine_name(TopLevelName),
    member(b(identifier(VarName),_Type,Infos),Variables),
    member(occurrences(Machines),Infos),
    member(TopLevelName,Machines).

all_abstract_parameters([],[]).
all_abstract_parameters([Event|FurtherEvents],ParametersOut) :-
    get_texpr_expr(Event,rlevent(_Name,_Section,_Stat,Parameters,_Guards,_Thm,_Actions,_VWit,_PWit,_Unmod,_AbsEvents)),
    all_abstract_parameters(FurtherEvents,FurtherParameters), append(Parameters,FurtherParameters,ParametersOut).

% -------------------------------------------
% before-after-predicates used in event-b pos
% -------------------------------------------

before_after_predicate(Action,Predicate) :-
    get_texpr_expr(Action,assign(Ids,Preds)), !,
    ba_predicate_equal(Ids,Preds,Predicates),
    conjunct_predicates(Predicates,Predicate).
before_after_predicate(Action,Predicate) :-
    get_texpr_expr(Action,becomes_element_of(Ids,Preds)), !,
    ba_predicate_member(Ids,Preds,Predicates),
    conjunct_predicates(Predicates,Predicate).
before_after_predicate(Action,Predicate) :-
    get_texpr_expr(Action,becomes_such(Ids,Pred)), !,
    ba_predicate_becomes_such(Ids,Pred,Predicate).
before_after_predicate(Action,Predicate) :- !,
    add_error_fail(before_after_predicate,'before-after-predicate implementation missing: ',[Action,Predicate]).

ba_predicate_equal([],[],[]).
ba_predicate_equal([Id|Ids],[Pred|Preds],Predicates) :-
    ba_predicate_equal(Ids,Preds,SubPreds),
    % prime identifiers on the lhs
    prime_variable(Id,PrimedId),
    create_texpr(equal(PrimedId,Pred),pred,[],TPred),
    Predicates = [TPred|SubPreds].
ba_predicate_member([],[],[]).
ba_predicate_member([Id|Ids],[Pred|Preds],Predicates) :-
    ba_predicate_member(Ids,Preds,SubPreds),
    % prime identifiers on the lhs
    prime_variable(Id,PrimedId),
    create_texpr(member(PrimedId,Pred),pred,[],TPred),
    Predicates = [TPred|SubPreds].

ba_predicate_becomes_such(Ids,Pred,Predicate) :-
    list_renamings_to_primed_vars(Ids,RenamingList),
    rename_bt(Pred,RenamingList,Predicate).
list_renamings_to_primed_vars([],[]).
list_renamings_to_primed_vars([Var|Vars],[rename(VarId,PrimedId)|T]) :-
    get_texpr_id(Var,VarId),
    atom_concat(VarId,'\'',PrimedId),
    list_renamings_to_primed_vars(Vars,T).

% -------------------------------------------
% classical b po generation
% -------------------------------------------

classical_refinement_check(ResultList) :-
    % fetch Axioms and Invariant/properties first
    full_b_machine(Machine),
    b_get_refined_machine(AbsMachine)
    -> b_get_invariant_from_machine(Invariant),
       b_get_linking_invariant_from_machine(LinkingInvariant),
       % check initialisation po
       initialisation_pos(LinkingInvariant,Machine,AbsMachine,ListIniPos),
       % check all operation pos
       get_all_operations_from_machine(Machine,ConcreteOperations),
       get_all_operations_from_machine(AbsMachine,AbstractOperations),
       operations_pos(ConcreteOperations,Invariant,LinkingInvariant,AbstractOperations,ListOpPos),
       flatten(ListIniPos,List), flatten(ListOpPos,List2),
       append([List,['\n'],List2],ResultList)
    ;  ResultList = ['Not a Refinement Machine'].

get_all_operations_from_machine(Machine,Operations) :-
    get_section(operation_bodies,Machine,Ops),
    findall(operation(Name,Results,Parameters,Body),
	    (get_texpr_expr(Op,operation(TId,Results,Parameters,Body)),
		get_texpr_id(TId,op(Name)),member(Op,Ops)),
	    Operations).

initialisation_pos(LinkingInvariant,Machine,AbsMachine,Result) :-
    % get the two initialisations
    get_section(initialisation,Machine,ConcreteInitialisation),
    get_section(initialisation,AbsMachine,AbstractInitialisation),

    % set up proof obligation [T1] not [T] not J (Schneider p.222)
    negate(LinkingInvariant,NegLinkingInvariant),
    weakest_precondition(AbstractInitialisation,NegLinkingInvariant,TempPred),
    negate(TempPred,NotTNotJ),
    weakest_precondition(ConcreteInitialisation,NotTNotJ,PO),
    negate(PO,CounterExPO),

    check_predicate(CounterExPO,State,Res),

    % result string and print
    ajoin(['Initialisation [T1] not [T] not J: Counter-Example found: ', Res],Result),
    write(Result), nl, translate:nested_print_bexpr(CounterExPO),
    write(State),nl.

operations_pos([],_I,_LI,_A,[]).
operations_pos([operation(Name,Results,_Parameters,Body)|FurtherOps],Invariant,LinkingInvariant,AbstractOperations,[PreRefPoResult,RefPoResult|ListOpPos]) :-
    % first find the matching abstract operation
    member(operation(Name,_AbsResults,_AbsParameters,AbsBody),AbstractOperations),

    % check if we have preconditions - otherwise we assume true
    (get_texpr_expr(Body,precondition(Precondition,RealBody)) -> true ; RealBody=Body, create_texpr(truth,pred,[],Precondition)),
    (get_texpr_expr(AbsBody,precondition(AbsPrecondition,AbsRealBody)) -> true ; AbsRealBody=AbsBody, create_texpr(truth,pred,[],AbsPrecondition)),

    % refinement po for the preconditions
    negate(AbsPrecondition,NegAbsPrecondition),
    conjunct_predicates([Invariant,LinkingInvariant,Precondition,NegAbsPrecondition],PreRefPo),

    check_predicate(PreRefPo,State,PreRefPoRes),

    % result string and print
    ajoin([Name, ' Precondition Refinement I & J & P => P1: Counter-Example found: ', PreRefPoRes],PreRefPoResult),
    write(PreRefPoResult), nl, translate:nested_print_bexpr(PreRefPo),
    write(State),nl,

    % for the refinement of operations, we need to rename existing output variables (i.e. prime them)
    prime_variables(Results,PrimedResults),
    list_renamings_to_primed_vars(Results,Renamings),
    rename_bt(RealBody,Renamings,RealBodyWithPrimedVariables),
    create_equals(Results,PrimedResults,EqualResultsPred),

    % refinement po for the operation
    % I & J & P => [S1[out'/out]] not S not (J & out'=out)
    conjunct_predicates([LinkingInvariant|EqualResultsPred],RHS),
    negate(RHS,NegRHS),
    weakest_precondition(AbsRealBody,NegRHS,RHS2),
    negate(RHS2,NegRHS2),
    weakest_precondition(RealBodyWithPrimedVariables,NegRHS2,RHS3),
    negate(RHS3,NegRHS3),
    conjunct_predicates([Invariant,LinkingInvariant,Precondition,NegRHS3],RefPo),

    check_predicate(RefPo,State2,RefPoRes),
    % result string and print
    ajoin([Name, ' Refinement I & J & P => S1 not S not J: Counter-Example found: ', RefPoRes],RefPoResult),
    write(RefPoResult), nl, translate:nested_print_bexpr(RefPo),
    write(State2),nl,

    operations_pos(FurtherOps,Invariant,LinkingInvariant,AbstractOperations,ListOpPos).

create_equals([],[],[]).
create_equals([Id|Ids],[Pred|Preds],Predicates) :-
    create_equals(Ids,Preds,SubPreds),
    create_texpr(equal(Id,Pred),pred,[],TPred),
    Predicates = [TPred|SubPreds].