% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(clpfd_interface,[post_constraint/1,  post_constraint/2, post_constraint2/2,
                            clpfd_neq_expr/2, clpfd_eq_expr/2,
                            clpfd_eq_div/3, clpfd_eq_guarded_div/3,
                            clpfd_eq/2, clpfd_neq/2,
                            clpfd_geq2/2,
                            clpfd_geq/3, clpfd_gt/3, clpfd_leq/3, clpfd_lt/3,
                            clpfd_inrange/3, clpfd_inrange/4,
                            clpfd_not_inrange/3, clpfd_not_in_non_empty_range/3,
                            clpfd_sum/3,
                            clpfd_inlist/2, clpfd_not_inlist/2, clpfd_reify_inlist/3,
                            clpfd_minimum/2, clpfd_maximum/2,
                            clpfd_domain/3,
                            clpfd_alldifferent/1,
                            %clpfd_labeling/1, % not used at the moment
                            clpfd_reverse_labeling/1,
                            is_64_bit_system/0, clpfd_overflow_error_message/0,
                            catch_clpfd_overflow_call1/1]).

:- use_module(library(clpfd)).


:- if((current_prolog_flag(version_data,sicstus(4,X,Y,_,_)),X<3,Y<3)).
% -------- patch provided by SICS -----------
% solves lost solution in CLP(FD) when calling:
%     when(nonvar(X),member(Y,[1,2])), X in 1..2, X in 2..3.
% patch provided 23rd June 2012 for SICS 4.2.1
%  bug cause ProB to miss solutions in e.g. following set comprehension:
% {xx | xx:{1|->TRUE,2|->TRUE,2|->FALSE} & (xx  :  {(2|->TRUE),(2|->FALSE),(3|->TRUE)} or xx  :  {(1|->TRUE),(1|->FALSE),(2|->TRUE),(2|->FALSE),(3|->TRUE),(3|->FALSE)})}
:- set_prolog_flag(redefine_warnings,proceed).
clpfd:propagate_interval_chk(Pruned, Min, Max) :-
	clpfd:'$fd_in_interval'(Pruned, Min, Max, 1), !, % 20120623
	clpfd:'$fd_evaluate_indexical'(RC, Global),
	clpfd:evaluate(RC, Global).
clpfd:propagate_interval_chk(Pruned, Min, Max) :-
	clpfd:arg_attribute(Pruned, _, in(Pruned,Min..Max), 1), !,
	fail.
clpfd:propagate_interval_chk(_, _, _).

clpfd:prune_and_propagate_chk(Pruned, Set) :-
	clpfd:'$fd_in_set'(Pruned, Set, 1), !, % 20120623
	clpfd:'$fd_evaluate_indexical'(RC, Global),
	clpfd:evaluate(RC, Global).
clpfd:prune_and_propagate_chk(Pruned, Set) :-
	clpfd:arg_attribute(Pruned, _, in_set(Pruned,Set), 1), !,
	fail.
clpfd:prune_and_propagate_chk(_, _).
%:- set_prolog_flag(redefine_warnings,on).
% ---------- end patch ---------------
:- endif.


:- use_module(typechecker).
:- use_module(tools).
:- use_module(preferences).
:- use_module(library(timeout)).
:- use_module(error_manager).
:- use_module(self_check).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'Provide interface to CLP(FD), with Prolog fallback in case of errors or when CLP(FD) turned off.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- meta_predicate post_constraint(0).
:- meta_predicate post_constraint(0,0).
:- meta_predicate post_constraint2(0,*).
:- meta_predicate time_out_constraint(0,0).


% Note: clpfd_eq is sometimes called with arithmetic expressions (treatment of plus/minus/...)
clpfd_eq(E1,E2) :- number(E1),number(E2),!,E1=E2.
clpfd_eq(E1,E2) :- post_constraint( E1 #= E2 ).
clpfd_neq(E1,E2) :- number(E1),number(E2),!,E1\=E2.
clpfd_neq(E1,E2) :- % print(neq(E1,E2)),nl,
   post_constraint( E1 #\= E2 , dif(E1,E2)).
  % clpfd_domain(E1,D1,D2), clpfd_domain(E2,B1,B2), print(dom(D1:D2,B1:B2)),nl.

% this will force Y to be nonzero
clpfd_eq_div(E1,X,Y) :- post_constraint( E1 #= X/Y, prolog_eq(E1,X//Y)).
% this waits until we are sure Y is not zero
clpfd_eq_guarded_div(E1,X,Y) :- 
    post_constraint3((E1 #= X/YY,MY in 0..1, Y #= MY*YY), prolog_eq(E1,X//Y),Posted), % Y can either be 0 or equal to YY
    % this is better than the previous solution:
   % post_constraint(Y#\=0 #=> YY#=Y, prolog_eq(Y,YY)).
   % however, x:44..77 & E : 1..5555 & E = x/y & y:0..4 only restricts E to 1..77, not 11..77
   % what is still missing is propagating upper bounds of Y to YY
   % Y in 0..4, M in 0..1, YY in 1..77, Y #=M*YY
   % with this propagation below, we at least obtain propagation for x:44..77 & E : 1..5555 & y:0..4 & E = x/y 
   % what we need is continously propagating the bounds from Y to YY (but not vice versa for 0)
   % maybe we can achieve this using indexicals ??
   (Posted==true
     -> fd_max(Y,YUp), (number(YUp) -> YY #=< YUp ; true),
        fd_min(Y,YLow), (number(YLow) -> YY #>= YLow ; true)
     ;  true).
   
% do we need to instantiate YY if Y becomes 0 ??
%:- block copy_divisor(-,?).
%copy_divisor(Y,Y).


% neq takes expressions as arguments
clpfd_eq_expr(E1,E2) :- simple(E1),simple(E2),!, E1=E2.
clpfd_eq_expr(E1,E2) :- simple(E1),computable_arith_expression(E2),!, E1 is E2.
clpfd_eq_expr(E1,E2) :- simple(E2),computable_arith_expression(E1),!, E2 is E1.
:- if(current_prolog_flag(dialect,sicstus)).
:- if(current_prolog_flag(version_data,sicstus(4,2,_,_,_))).
clpfd_eq_expr(E1,E2) :- post_constraint( E1 #= E2, prolog_eq(E1,E2)).
:- else.
clpfd_eq_expr(E1,E2) :- prolog_eq(E1,E2),
  post_constraint( E1 #= E2). % Used instead of line below because of serious bug in SICStus 4.1.3
                              % BUG: X*X #= 100, X #= 9.  succeeds !
:- endif.                       % sicstus 4.1.3
:- else.
clpfd_eq_expr(E1,E2) :- post_constraint( E1 #= E2, prolog_eq(E1,E2)).
:- endif.                       % sicstus
:- assert_must_succeed((clpfd_interface:prolog_eq(X,2+1),X==3)).
:- assert_must_succeed((clpfd_interface:prolog_eq(4-1,X),X==3)).
:- assert_must_succeed((clpfd_interface:prolog_eq(5-2,(2*2)-Y),Y=1)).
:- assert_must_fail((clpfd_interface:prolog_eq(5-2,(2*2)-Y),Y=0)).
prolog_eq(A,B) :- simple(A),!, when(ground(B), A is B).
prolog_eq(A,B) :- simple(B),!, when(ground(A), B is A).
prolog_eq(A,B) :- when((ground(A),ground(B)),=:=(A,B)).

computable_arith_expression(X) :- number(X),!.
computable_arith_expression(X) :- var(X),!,fail.
computable_arith_expression(-(X)) :- computable_arith_expression(X).
computable_arith_expression(+(X,Y)) :- computable_arith_expression(X),computable_arith_expression(Y).
computable_arith_expression(-(X,Y)) :- computable_arith_expression(X),computable_arith_expression(Y).
computable_arith_expression(*(X,Y)) :- computable_arith_expression(X),computable_arith_expression(Y).

clpfd_neq_expr(E1,E2) :-  number(E1),number(E2),!,E1\=E2.
clpfd_neq_expr(E1,E2) :-  %print(neq_expr(E1,E2)),nl,
   post_constraint( E1 #\= E2, prolog_neq(E1,E2)).

:- assert_must_succeed((clpfd_interface:prolog_neq(X,2+1),X=2)).
:- assert_must_fail((clpfd_interface:prolog_neq(X,2+1),X=3)).
:- assert_must_succeed((clpfd_interface:prolog_neq(3,X),X=2)).
:- assert_must_succeed((clpfd_interface:prolog_neq(5-2,(2*2)-Y),Y=2)).
prolog_neq(A,B) :- simple(A),simple(B),!, dif(A,B).
prolog_neq(A,B) :- when((ground(A),ground(B)),=\=(A,B)).

:- assert_must_succeed((clpfd_interface:clpfd_leq_expr(3,X+1),X=3)).
:- assert_must_fail((clpfd_interface:clpfd_leq_expr(2,X-1),X=2)).
clpfd_leq_expr(E1,E2) :- post_constraint( E1 #=< E2, prolog_leq(E1,E2)).
prolog_leq(A,B) :- when((ground(A),ground(B)),=<(A,B)).

/* TO DO: do custom equalities div, plus, ...
mydiv(X, Y, Z) :-
	(Y#=0 #\/ (Y#\=0 #/\ X/Y #= Z)). */

% a version of geq which avoids posting constraints if possible
clpfd_geq2(E1,E2) :- number(E1),number(E2),!, E1 >= E2.
clpfd_geq2(E1,E2) :- post_constraint( E1 #>= E2 ).

clpfd_geq(E1,E2,Posted) :- post_constraint2( E1 #>= E2 , Posted).
clpfd_gt(E1,E2,Posted) :-  post_constraint2( E1 #> E2  , Posted).
clpfd_leq(E1,E2,Posted) :- post_constraint2( E1 #=< E2 , Posted).
clpfd_lt(E1,E2,Posted) :-  post_constraint2( E1 #< E2  , Posted).

clpfd_inrange(X,Low,Up) :- clpfd_inrange(X,Low,Up,_Posted).
clpfd_inrange(X,Low,Up,Posted) :-
  (Low==Up -> X=Low
   ; (number(Low),number(Up))
      -> Low =< Up, % otherwise interval empty
         (number(X) -> X >= Low, X =< Up, Posted=true %, Low =<Up
                     ; post_constraint2(X in Low..Up, Posted))  %  Note: X in 1..Y, Y in 0..3. fails ! Range needs to be constant
    ;  post_constraint2( (X #>=Low #/\ X #=< Up),Posted ) ).

% assumes Low =< Up !
clpfd_not_in_non_empty_range(X,Low,Up) :-
   post_constraint((X #<Low #\/ X #> Up),outside_range_blocking(X,Low,Up)).
   
:- block outside_range_blocking(-,?,?).
outside_range_blocking(X,Low,Up) :-  (X<Low ; X>Up).

% assumes Low =< Up !
clpfd_not_inrange(X,Low,Up) :-
   post_constraint((X #<Low #\/ X #> Up #\/ Up #< Low),not_in_range_blocking(X,Low,Up)).

:- block not_in_range_blocking(-,?,?),not_in_range_blocking(?,-,?).
not_in_range_blocking(X,Y,Z) :- 
     (X<Y -> true ; not_in_range_blocking2(X,Y,Z)).
:- block not_in_range_blocking2(?,?,-).
not_in_range_blocking2(X,Y,Z) :- (Z<Y -> true ; Z<X).


clpfd_sum(List,Sum,Posted) :- post_constraint2( sum(List,'#=',Sum),Posted).

% assert that a variable is one of the numbers in the FDList
clpfd_inlist(El,FDList) :-
   post_constraint( (list_to_fdset(FDList,FDSET), El in_set FDSET)).
clpfd_not_inlist(El,FDList) :-
   post_constraint( (list_to_fdset(FDList,FDSET), fdset_complement(FDSET,C),El in_set C)).

clpfd_maximum(El,List) :- %print(maximum(El,List)),nl,
                          post_constraint( maximum(El,List)). %, print(done(El)).
clpfd_minimum(El,List) :- post_constraint( minimum(El,List)).

/*
on_exception(E,list_to_fdset([0,8589934592],R),true)
E= error(domain_error(integer_list,[0,8589934592]),domain_error(list_to_fdset([0,8589934592],_A),1,integer_list,[0,8589934592]))
domain_error
*/


clpfd_reify_inlist(El,FDList,FDRes) :- 
   post_constraint( (list_to_fdset(FDList,FDSET), (El in_set FDSET) #<=> FDRes)).



post_constraint(_) :- preferences:preference(use_clpfd_solver,false),!.
post_constraint(C) :-   % print(posting(C)),nl, %
  on_exception(error(Err,_),time_out_constraint(C,true), check_error(Err,C) ).
    % Note: this will only catch exceptions that occur while posting !
    %  CLPFD integer overflow exceptions can still happen after posting :-( 


post_constraint2(_,Posted) :-
   preferences:preference(use_clpfd_solver,false),!,Posted=false.
post_constraint2(C,Posted) :-   % print(posting(C)),nl, %
  on_exception(error(Err,_),time_out_constraint(C,(Posted=false)), (check_error(Err,C),Posted=false) ),
  (Posted==false -> true ; Posted=true).

check_error(representation_error('CLPFD integer overflow'),C) :- !,print_message('Ignoring constraint (integer overflow):'),
                 print_message(C).
check_error(domain_error(_Type,_Val),C) :- !,  print_message('Ignoring constraint (domain error):'),
                 % print(Type), print(' : '), print(Val), print(' : '),
                 print_message(C).
check_error(Err,C) :- print_message('Ignoring constraint (unknown error):'),
                 print(Err), print(' : '), print_message(C).


:- assert_must_succeed((print('Is 64-bit: '),(clpfd_interface:is_64_bit_system -> print('Yes') ; print('NO')),nl)).

is_64_bit_system :- on_exception(_, _X #= 268435456, fail). % 268435456 = 2^28, will generate overflow in 32-bit system


clpfd_overflow_error_message :- clpfd_interface:is_64_bit_system, !,
  add_error(probcli,'A CLPFD integer overflow occurred.\nTurn CLPFD off (Advanced Preferences).').
clpfd_overflow_error_message :-
  add_error(probcli,'A CLPFD integer overflow occurred.\nTurn CLPFD off (Advanced Preferences) or use a 64 bit version of ProB.').


% post_constraint3(CLPFD,PrologBackup,Posted)
post_constraint3(_,C,Posted) :- preferences:preference(use_clpfd_solver,false),!,Posted=false,call(C).
post_constraint3(C,PrologBackup,Posted) :-   % print(posting(C,PrologBackup)),nl,
  on_exception(error(Err,_),
               time_out_constraint((C,Posted=true),(Posted=false,PrologBackup)),
               (check_error(Err,C),Posted=false,call(PrologBackup)) ).
  
% post_constraint(CLPFD,PrologBackup)
post_constraint(_,C) :- preferences:preference(use_clpfd_solver,false),!,call(C).
post_constraint(C,PrologBackup) :-   % print(posting(C,PrologBackup)),nl,
  on_exception(error(Err,_),time_out_constraint(C,PrologBackup), (check_error(Err,C),call(PrologBackup)) ).

time_out_constraint(C,PrologBackup) :-  %% debug:new_pp(C,PP), %%
  %%preferences:preference(time_out,CurTO), TO is 800 + (CurTO//100), time_out(C,TO,T),
  time_out(C,1200,T),  %% debug:new_sol(C,PP), %%
  (T==time_out -> print_message('Timeout when posting constraint:'), print_message(C),
                  (preferences:preference(fail_if_clpfd_timeout,true)
                    -> print_message('Assuming unsatisfiable !'),fail
                    ;  call(PrologBackup))
               ; true).


:- assert_must_succeed((clpfd_interface:clpfd_domain(3,L,U), L==3, U==3)).
clpfd_domain(Var,Low,Up) :- var(Var),!,
   (preferences:preference(use_clpfd_solver,false) 
     -> Low=inf,Up=sup
     ;  fd_min(Var,Low), fd_max(Var,Up)
    ).
        %fd_dom(Var, Low..Up)). fd_dom will sometimes return {1} \/ 2..3 or something like that !
clpfd_domain(X,X,X).

:- assert_must_succeed(clpfd_interface:clpfd_alldifferent([1,2,3])).
:- assert_must_succeed(clpfd_interface:all_dif([1,2,3],[])).
:- assert_must_fail(clpfd_interface:clpfd_alldifferent([1,2,3,2])).
:- assert_must_fail(clpfd_interface:all_dif([1,2,3,2],[])).

clpfd_alldifferent(L) :- post_constraint(all_different(L),all_dif(L,[])).

all_dif([],_).
all_dif([H|T],All) :- all_dif_aux(All,H), all_dif(T,[H|All]).

all_dif_aux([],_).
all_dif_aux([H|T],X) :- dif(H,X), all_dif_aux(T,X).

/*
clpfd_labeling(Variables) :- 
  (preferences:preference(use_clpfd_solver,false)  -> true
    ; labeling([ffc,enum],Variables)).  %ff, enum
*/

:- use_module(library(lists),[reverse/2]).
% label a list of variables which was reversed:
clpfd_reverse_labeling(Variables) :- var(Variables),!,
  add_internal_error('Must be list of variables: ',clpfd_reverse_labeling(Variables)).
clpfd_reverse_labeling(Variables) :- !,
   reverse(Variables,RevV), labeling([ffc,enum],RevV). % ffc,enum
/*
clpfd_reverse_labeling(Variables) :- 
% Note: is now also called if use_clpfd_solver preference is false as CLP(FD) used for carrier SETS
  reverse_and_filter(Variables,[],LV,[],FilteredVariables),
  labeling([ffc,enum],LV), %ff,ffc step,enum,bisect  -> we could control them by a preference
  labeling([ffc,enum],FilteredVariables). % TO DO: pass back to kernel_waitflags and give other things priority



reverse_and_filter([],Acc,Acc,FA,FA).
reverse_and_filter([H|T],Acc,Res,FAcc,FRes) :-
    (useless_clpfd_variable(H,_G) -> print(useless(H,_G)),nl,
        reverse_and_filter(T,Acc,Res,[H|FAcc],FRes)
      ; reverse_and_filter(T,[H|Acc],Res,FAcc,FRes)).

useless_clpfd_variable(H,G) :- var(H), frozen(H,G), %print(frozen(H,G)),nl,
   useless_goal(G,H).

useless_goal(true,_).
useless_goal(clpfd:(X in _A.._B),Var) :- Var==X. % only goal is the enumerator itself
*/

:- meta_predicate catch_clpfd_overflow_call1(0).
catch_clpfd_overflow_call1(Call) :- 
  on_exception(error(representation_error('CLPFD integer overflow'),_),
               Call,
               ( clpfd_interface:clpfd_overflow_error_message, fail)).



/*
Testing for how large the integers can be:
| ?- X is integer(2**30), Y#=X+1.
! Representation error in argument 2 of user:'t=u+c'/3
! CLPFD integer overflow
! goal:  't=u+c'(_102,1073741824,1)
| ?- X is integer(2**30), Y#=X.
X = 1073741824,
Y = 1073741824 ? 
*/

/* grounding_wait_flags
 :::
enumerate_int
| ?- on_exception(E,X in 0..2147483646, print(E)).
error(representation_error(CLPFD integer overflow),representation_error(_710 in 0..2147483646,2,CLPFD integer overflow))
E = error(representation_error('CLPFD integer overflow'),representation_error(_A in 0..2147483646,2,'CLPFD integer overflow')) ? 
yes

X in 0..9, X #>3, fd_dom(X,R).
R = 4..9,
X in 4..9 ? 
yes
| ?- X in 0..9, X #>3, fd_dom(X,R), R=.. L.
L = [..,4,9],
R = 4..9,
X in 4..9 ? 
yes
| ?- fd_dom(X,R).
R = inf..sup ? 
yes
| ?- X in 0..2147483647, X#<2, X#>0.
! Representation error in argument 2 of user:in/2
! CLPFD integer overflow
! goal:  _97 in 0..2147483647
| ?- X in 0..214748364.
X in 0..214748364 ? 
yes
| ?- X in 0..2147483646.
! Representation error in argument 2 of user:in/2
! CLPFD integer overflow
! goal:  _97 in 0..2147483646
| ?- X in 0..10, X #>2147483646.
! Representation error in argument 1 of user:'x=<y IND'/2
! CLPFD integer overflow
! goal:  'x=<y IND'(2147483647,_98)
*/
