% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(error_manager,[add_warning/2, add_warning/3, add_warning/4,
           reset_errors/0, add_error/2, add_error/3, add_error/4,
           add_error_and_fail/3, add_error_and_fail/2,
           add_error_fail/3,
           add_failed_call_error/1, add_internal_error/2, add_internal_error/4,
           add_message/2, add_message/3,
           add_all_perrors/1, add_all_perrors/2, add_all_perrors/3,
           get_all_errors/1, get_all_errors_and_reset/1,
           get_error/2,
           get_error_span_linecol/4,
           position_is_not_in_main_file/1,
           real_error_occurred/0, reset_real_error_occured/0,
           test_error_occured/1, count_errors_occured/2,
           
           set_error_context/1, clear_error_context/0, get_error_context/1,
           
           extract_line_col/5, extract_all_line_col/5,
           
           trace_upon_error/0, turn_trace_upon_error_on/0,turn_trace_upon_error_off/0,
           turn_show_error_source_on/0, turn_show_error_source_off/0,
           
            enter_new_error_scope/1, exit_error_scope/2,
            add_new_event_in_error_scope/1, event_occured_in_error_scope/1
            ]).

:- use_module(library(lists)).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'The error_manager controls error messages. Error messages can be added, retrieved, counted,etc. The module also takes care of extracting source spans from error messages.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

%:- use_module(translate).
:- use_module(debug, [debug_mode/1]).


% ERROR SCOPING

:- dynamic cur_block_level/1.
cur_block_level(0).
:- volatile error_occured_at_level/1.
:- dynamic error_occured_at_level/1.

enter_new_error_scope(Level) :- retract(cur_block_level(L)), L1 is L+1, assert(cur_block_level(L1)),
  Level=L1.
exit_error_scope(Level,ErrorOcc) :- retract(cur_block_level(CurL)),
   (CurL = Level -> NewL is CurL-1
    ; Level < CurL -> NewL is Level-1,  print('*** exiting multiple error blocks ***'),nl
    ; NewL is CurL -> print('*** block already exited ***'),nl
   ), assert(cur_block_level(NewL)),
   copy_error_scope_events(CurL,NewL),
   (retract(error_occured_at_level(CurL))
    -> ErrorOcc=true,
       (error_occured_at_level(NewL) -> true ; assert(error_occured_at_level(NewL)))
    ; ErrorOcc=false).

assert_real_error_in_current_block :- cur_block_level(L),L>0,
  \+ error_occured_at_level(L), assert(error_occured_at_level(L)),!.
assert_real_error_in_current_block.


reset_error_scopes :- retractall(cur_block_level(_)), assert(cur_block_level(0)),
   retractall(error_occured_at_level(_)),
   retractall(specific_event_occured_at_level(_,_)).

:- volatile specific_event_occured_at_level/2.
:- dynamic specific_event_occured_at_level/2.

% can be used to keep track of whether specific events occured within an error_scope
add_new_event_in_error_scope(Event) :- cur_block_level(L),
   (specific_event_occured_at_level(L,Event) -> fail
     ; assert(specific_event_occured_at_level(L,Event))).

event_occured_in_error_scope(Event) :- cur_block_level(L),specific_event_occured_at_level(L,Event).

copy_error_scope_events(FromLevel,ToLevel) :- retract(specific_event_occured_at_level(FromLevel,Event)),
   \+ specific_event_occured_at_level(ToLevel,Event),
   % print(copying_event(FromLevel,ToLevel,Event)),nl,
   assert(specific_event_occured_at_level(ToLevel,Event)),fail.
copy_error_scope_events(_,_).

% -----------------------

:- volatile logged_error/4, real_error_occurred/0, logged_error_spans/2.
:- dynamic logged_error/4.
:- dynamic real_error_occurred/0.
:- dynamic logged_error_spans/2.

reset_real_error_occured :- retractall(real_error_occurred).
reset_errors :- retractall(real_error_occurred),
                retractall(logged_error(_,_,_,_)),retractall(logged_error_spans(_,_)),
                reset_error_scopes,
                clear_error_context.

reset_error_span_if_required :-
	(logged_error(_,_,_,_) -> true
		; retractall(logged_error_spans(_,_))).


:- meta_predicate add_failed_call_error(0).
:- meta_predicate add_internal_error(-,0).

add_failed_call_error(Call) :- !,
   add_internal_error('Internal Error; call failed: ',Call).

add_internal_error(Module,Msg,Term,Span) :-
   add_error(internal_error(Module),Msg,Term,Span).
add_internal_error(Msg,Module:Call) :- !, 
   add_error(internal_error(Module),Msg,Call).
add_internal_error(Msg,Call) :-
   add_internal_error(Msg,unknown_module:Call).
   
add_error(Source,ErrMsg,ErrTerm,Span) :-
   add_error2(Source,ErrMsg,ErrTerm,Span),
   assert(logged_error_spans(Source,Span)),
   print_error_span(Span).

print_error_span(unknown) :- !.
print_error_span(Span) :-
   extract_line_col(Span,StartLine,StartCol,_EndLine,_EndCol), !,
   (extract_file_name(Span,Filename) ->
      ajoin(['Line: ',StartLine,' Column: ',StartCol, ' in file ',Filename],ErrMsg) 
      ;  ajoin(['Line: ',StartLine,' Column: ',StartCol],ErrMsg)
   ),print_error(ErrMsg).
print_error_span(Span) :- print_error(Span). % span not recognized !

add_error_fail(Source,ErrMsg, ErrTerm) :- add_error_and_fail(Source,ErrMsg, ErrTerm).
add_error_and_fail(Source,ErrMsg, ErrTerm) :- add_error(Source,ErrMsg, ErrTerm),fail.
add_error_and_fail(Source,ErrMsg) :- add_error(Source,ErrMsg),fail.

add_error(Source,ErrMsg, ErrTerm) :-
    add_error2(Source,ErrMsg, ErrTerm,unknown).
    
%:- use_module(translate,[translate_error_term/2]).

:- volatile trace_upon_error/0.
:- dynamic trace_upon_error/0.
turn_trace_upon_error_on :- (trace_upon_error -> true ; assert(trace_upon_error)).
turn_trace_upon_error_off :-  retractall(trace_upon_error).
:- dynamic show_error_source/0.
show_error_source. % True if error source and span to be printed
turn_show_error_source_on :- (show_error_source -> true ; assert(show_error_source)).
turn_show_error_source_off :-  retractall(show_error_source).

trace_if_user_wants_it :- (trace_upon_error -> trace ; true).

%:- use_module(translate, [translate_error_term/2, translate_context/2]).
add_error2(Source,ErrMsg, ErrTerm,Span) :- 
    (show_error_source -> print_error('An error occurred !') ; true),
    assert_real_error_occured,
    (show_error_source -> print_error(source(Source)) ; true),
    print_error(ErrMsg),
    translate:translate_error_term(ErrTerm,RString),
    print_error(RString),
    string_concatenate(ErrMsg,RString,S),
    safe_translate_context(Context,TContext),
    ((Context=[];\+show_error_source) -> true ; print_error(TContext)),
    reset_error_span_if_required,
    assertz_logged_error(Source,S,TContext,Span).


safe_translate_context(Context,Result) :- 
% avoid infinite loop when error occurs inside translate_context
    get_error_context(Context),
    set_error_context(translate_context),
    (translate:translate_context(Context,TContext)
      -> true ; TContext=Context),
    set_error_context(Context),
    Result=TContext.
 
      
add_error(Source,ErrMsg) :- print_error('An error occurred !'),
    assert_real_error_occured,
    add_error_or_warning(Source,ErrMsg,unknown).

add_error_or_warning(Source,ErrMsg,Span) :-
    (preferences:get_preference(strict_raise_warnings,true)
    -> assert_real_error_occured ; true),
    (show_error_source -> print_error(source(Source)) ; true),
    print_error(ErrMsg),
    safe_translate_context(Context,TContext),
    ((Context=[];\+show_error_source) -> true ; print_error(TContext)),
    reset_error_span_if_required,
    assertz_logged_error(Source,ErrMsg,TContext,Span).

assert_real_error_occured :- (real_error_occurred -> true ; assert(real_error_occurred)),
    assert_real_error_in_current_block.

assertz_logged_error(Source,S,TContext,Span) :-
    assertz(logged_error(Source,S,TContext,Span)),trace_if_user_wants_it.

add_warning(Source,ErrMsg, ErrTerm) :- 
   add_warning(Source,ErrMsg, ErrTerm,unknown).
add_warning(Source,ErrMsg, ErrTerm,Span) :- 
    translate:translate_error_term(ErrTerm,RString),
    string_concatenate(ErrMsg,RString,S),
    add_error_or_warning(warning(Source),S,Span).
add_warning(Source,Message) :- add_error_or_warning(warning(Source),Message,unknown).
   
add_message(Source,ErrMsg, ErrTerm) :- /* same as add_error but does not add err msg */
    translate:translate_error_term(ErrTerm,RString),
    string_concatenate(ErrMsg,RString,S),
    add_message(Source,S).
add_message(Source,ErrMsg) :- 
   (show_error_source -> print_error(source(Source)) ; true),
   print_error(ErrMsg).
   %assertz(logged_error(Source,ErrMsg,_,_)).

% adds a list of errors with position information
% [error(Msg,Pos)]
add_all_perrors(Errors) :- add_all_perrors(Errors,[],bmachine).
add_all_perrors(Errors,Files) :- add_all_perrors(Errors,Files,bmachine).
add_all_perrors(Errors,Files,Context) :- add_all_perrors(Errors,Files,Context,0,100).
add_all_perrors([],_,_,_,_).
add_all_perrors([H|Rest],_Filenames,ContextOfError,NrOfErrors,MaxErrors) :-
    NrOfErrors >= MaxErrors,!,
    length([H|Rest],Len),
    add_error(ContextOfError,'Reached maximium number of errors; number of errors not reported: ',Len).
add_all_perrors([error(Msg,Pos)|Rest],Filenames,ContextOfError,NrOfErrors,MaxErrors) :-
	( Pos = pos(_,Filenumber,Srow,Scol,Erow,Ecol) ->
	    Ecol1 is Ecol+1,
	    ( get_filename(Filenumber,Filenames,Filename) ->
		atom_concat('\n ',Filename,BrFilename),
		add_error(ContextOfError,Msg,BrFilename,src_span(Filename,Filenumber,Srow,Scol,Erow,Ecol1,[],[]))
	    ; otherwise ->
		add_error(ContextOfError,Msg,'',src_span(Srow,Scol,Erow,Ecol1,[],[])))
	; Pos = pos(Srow,Scol,Filename) ->
	    add_error(ContextOfError,Msg,'',src_position_with_filename(Srow,Scol,1,Filename)) 
	; Pos = pos(Srow,Scol) -> /* deprecated, should no longer appear */
	    print(deprecated_pos(Srow,Scol)),nl,
	    add_error(ContextOfError,Msg,'',src_position(Srow,Scol,none,1)) /* none=offset */
	; otherwise ->
	    add_error(ContextOfError,Msg,'',Pos)),
	N1 is NrOfErrors+1,
	add_all_perrors(Rest,Filenames,ContextOfError,N1,MaxErrors).

get_filename(Filenumber,Filenames,Filename) :-
	number(Filenumber),!,
	nth1(Filenumber,Filenames,Filename).
get_filename(N,F,R) :- add_error(get_filename,'Internal error: ',get_filename(N,F,R)),
   F=[R]. % if there is only one file then return that file

get_error(Source,ErrMsg) :-
   retract(logged_error(Source,ErrMsg,_,_)).
   
get_error(Msg) :-
    retract(logged_error(Source,ErrMsg,Context,Span)),
    (extract_line_col(Span,StartLine,StartCol,_EndLine,_EndCol)
      -> (string_concatenate(StartCol,';;',Sp0),
          string_concatenate(', Column: ',Sp0,Sp1),
          string_concatenate(StartLine,Sp1,Sp2),
          string_concatenate('; ### Line: ',Sp2,Sp3),
          (extract_file_name(Span,Filename)
            -> string_concatenate(Filename,Sp3,Sp4),
               string_concatenate('; ### File: ',Sp4,EndStr)
             ; EndStr = Sp3
          )
         )
      ;  (EndStr = ';;')
     ),
    ((Context='' ; Context =[])
      ->  string_concatenate(ErrMsg,EndStr,S0)
       ;  translate:translate_error_term(Context,CString),
          string_concatenate(CString,EndStr,CString2),
          string_concatenate(':  Context = ',CString2,CString3),
          string_concatenate(ErrMsg,CString3,S0)
     ),
    (debug:debug_mode(off) -> Msg = S0
      ;  string_concatenate(Source,': ',S1),
         string_concatenate(S1,S0,Msg)
     ).
      
      
get_all_errors(All) :-
    findall(Err,get_error(Err),All),
    All \= [], clear_error_context.

get_all_errors_and_reset(All) :- findall(Err,get_error(Err),All),reset_errors,All \= [].

test_error_occured(Source) :- logged_error(Source,_,_,_),real_error_occurred,!.

%:- use_module(library(lists),[length/2]). length is builtin

count_errors_occured(Source,NrOfErrors) :- 
  findall(1,logged_error(Source,_,_,_),Ls),
  length(Ls,NrOfErrors).
  
  
:- volatile current_context/1.
:- dynamic current_context/1.

/* add context information; will be added to any error messages */
/* will be removed upon backtracking */
%push_error_context(_Ctx) :- !.
set_error_context(Ctx) :- % print(ctx(Ctx)),nl,
  retractall(current_context(_)),
  asserta(current_context(Ctx)).
%push_error_context(_Ctx) :-
%  (retract(current_context(_Ctx)) -> true ; print_error(could_not_retract_context(_Ctx))).

clear_error_context :- %print(clearing_ctx),nl,
   retractall(current_context(_)).

%pop_error_context :- 
%  (retract(current_context(_Ctx)) -> true ; print_error(could_not_pop_context(_Ctx))).
  
get_error_context(C) :- findall(Ct,current_context(Ct),C).


get_error_span(Source,Span) :-
   retract(logged_error_spans(Source,Span)).
   

get_error_span_linecol(Line,Col,EndLine,EndCol) :-
    get_error_span(_Source,Span),  % looks at logged_error_spans
    extract_line_col(Span,Line,Col,EndLine,EndCol).

extract_line_col([],_,_,_,_) :- !,fail.
extract_line_col([H|T],Srow,Scol,Erow,Ecol) :- !,  
   (member(nodeid(N),[H|T])
     -> extract_line_col(N,Srow,Scol,Erow,Ecol) % we have a B Expression info field
   % TO DO: add rodinpos
    ; (member(N,[H|T]),extract_line_col(N,Srow,Scol,Erow,Ecol)) -> true
   ).
extract_line_col(b(_,_,Infos),Srow,Scol,Erow,Ecol) :- !,
  extract_line_col(Infos,Srow,Scol,Erow,Ecol).
extract_line_col(span_predicate(P,_,_),Srow,Scol,Erow,Ecol) :- !,
  extract_line_col(P,Srow,Scol,Erow,Ecol).
extract_line_col(pos(_,_Filenumber,Srow,Scol,Erow,Ecol),Srow,S1,Erow,E1) :-  number(Srow),
  % _Filenumber is an index into list provided by b_get_all_used_filenames
  number(Scol), number(Erow), number(Ecol),!,
  (Scol>0 -> S1 is Scol-1 ; S1 = Scol),  % the BParser starts column numbering at 1
  E1=Ecol. %Somehow the Ecol info is ok for Tcl
extract_line_col(unknown,_,_,_,_) :- !,fail.
extract_line_col(src_span(_Filenumber,_Filename,StartLine,SC,EndLine,EC,Offset,LenBytes),StartLine,StartCol,EndLine,EndCol)
 :- !, extract_line_col(src_span(StartLine,SC,EndLine,EC,Offset,LenBytes),StartLine,StartCol,EndLine,EndCol).
extract_line_col(src_span(StartLine,SC,EndLine,EC,_Off,[]),StartLine,StartCol,EndLine,EndCol)
 :-  !, StartCol is SC-1, EndCol is EC-1.
extract_line_col(src_span(SL,SC,EL,EC,_Off,_LenBytes),StartLine,StartCol,EndLine,EndCol) :- !,
   StartLine=SL,StartCol is SC-1,EndLine = EL, EndCol is EC-1.
 %:- !, StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine = StartLine. %% fix because new library generates erroneous EndLine/EndCol info
% EndCol is EC-1.
extract_line_col(span_info(_Info,Span),SL,SC,EL,EC) :- !, extract_line_col(Span,SL,SC,EL,EC).
extract_line_col(src_span_operator(SP1,SP2),SL,SC,EL,EC) :- !, 
 (extract_line_col(SP1,SL,SC,EL,EC) ; extract_line_col(SP2,SL,SC,EL,EC)).
extract_line_col(multi_span(SL,SC,EL,EC,_Off,_Len,TAIL),StartLine,StartCol,EndLine,EndCol) :- !,
  ((StartLine = SL, StartCol is SC-1, 
     EndLine = EL, EndCol is EC-1 
     )
   ; extract_line_col(TAIL,StartLine,StartCol,EndLine,EndCol)).
extract_line_col(src_position(StartLine,SC,_Off,LenBytes),StartLine,StartCol,EndLine,EndCol) :- !,
   StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine is StartLine.
extract_line_col(src_position_with_filename(StartLine,SC,LenBytes,_Filename),StartLine,StartCol,EndLine,EndCol) :- !,
   StartCol is SC-1, EndCol is StartCol+LenBytes, EndLine is StartLine.
extract_line_col(lineCol(StartLine,StartCol),StartLine,StartCol,EndLine,EndCol) :- !,
   EndLine=StartLine, EndCol=StartCol.
extract_line_col(Struct,StartLine,StartCol,EndLine,EndCol) :- 
    \+ atomic(Struct), 
    Struct =.. [_|Args], Args \= [],
      /* recursively look for error spans inside */
    member(Arg,Args),
    extract_line_col(Arg,StartLine,StartCol,EndLine,EndCol),!.
%extract_line_col(S,_,_,_,_) :- print('COULD NOT EXTRACT LINE/COL INFO: '), print(S),
%    functor(S,F,N), print(' '),print(F/N),nl,nl,fail.


/* For TCL TK: extract all line_col information in one go and in 4 lists */
extract_all_line_col(Span, StartLineList,StartColList,EndLineList,EndColList) :-
    findall(lc(SL,SC,EL,EC), extract_line_col(Span,SL,SC,EL,EC),All),
    %print(all(All,Span)),nl,
    split_lc(All,StartLineList,StartColList,EndLineList,EndColList).
    
split_lc([], [],[],[],[]).
split_lc([lc(SL,SC,EL,EC)|T],[SL|T1],[SC|T2],[EL|T3],[EC|T4]) :-
   split_lc(T,T1,T2,T3,T4).

%:- use_module(bmachine,[b_get_all_used_filenames/1,b_get_main_filenumber/1]).
% TO DO: merge with line_col and in Tcl/Tk use info to highlight proper file !
extract_file_name(pos(_,Filenumber,_,_,_,_), Filename) :-
  bmachine:b_get_all_used_filenames(FN), get_filename(Filenumber,FN,Filename),!.
extract_file_name(pos(_,Filenumber,_,_,_,_), Filenumber).
extract_file_name(src_span(_,Filename,_,_,_,_,_,_), Filename).
extract_file_name(src_position_with_filename(_,_,_,Filename), Filename).

position_is_not_in_main_file(pos(_,Filenumber,_,_,_,_)) :- % B position info
  bmachine:b_get_main_filenumber(MainFN),
  MainFN \= Filenumber.
position_is_not_in_main_file(src_span(_,Filename,_,_,_,_,_,_)) :- 
   specfile:currently_opened_file(Main), Main \= Filename.
position_is_not_in_main_file(src_position_with_filename(_,_,_,Filename)) :- 
   specfile:currently_opened_file(Main), Main \= Filename.
