% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(eval_strings,[
  add_last_expression_to_unit_tests/0, print_last_expression/0, print_last_info/0,
  toggle_eval_det/0,
  eval_string/2,eval_string/3,eval_string/4,eval_codes/6,
  eval_file/4,
  set_eval_dot_file/1]).

:- use_module(library(lists)).

:- use_module(error_manager).
:- use_module(tools).
:- use_module(debug).

:- use_module(module_information,[module_info/2]).
:- module_info(group,repl).
:- module_info(description,'Tools to evaluate B expressions and predicates passed as strings.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(bmachine,[b_get_machine_constants/1,b_get_machine_variables/1,
                        b_parse_machine_expression_from_codes/6,
                        b_parse_machine_predicate_from_codes_open/5]).
:- use_module(b_global_sets,[add_prob_deferred_set_elements_to_store/2]).

% GENERAL EVAL for Expressions & Predicates

eval_string(String,StringResult) :- eval_string(String,StringResult,_EnumWarning,_).
eval_string(String,StringResult,EnumWarning) :- eval_string(String,StringResult,EnumWarning,_).
eval_string(String,StringResult,EnumWarning,LocalState) :-  % String is an atom
   atom_codes(String,Codes),
   eval_codes(Codes,exists,StringResult,EnumWarning,LocalState,_).

eval_file(Filename,OuterQuantifier,StringResult,EnumWarning) :-
   read_string_from_file(Filename,Codes),
   eval_codes_with_time_out(Codes,OuterQuantifier,StringResult,EnumWarning).

:- use_module(library(timeout)).
:- use_module(preferences,[get_computed_preference/2]).
eval_codes_with_time_out(Codes,OuterQuantifier,StringResult,EnumWarning) :-
  get_computed_preference(debug_time_out,DTO), 
  %print(debug_time_out(DTO)),nl,
  time_out(eval_codes(Codes,OuterQuantifier,StringResult,EnumWarning,_,_),DTO,TimeOutRes),
  (TimeOutRes=time_out ->
    StringResult = '**** TIME-OUT ****', print(StringResult), print(' ('),print(DTO), print('ms)'),nl,
    EnumWarning = time_out
    ; true).
    


:- volatile current_codes/1.
:- dynamic current_codes/1.
set_current_codes(C) :- retractall(current_codes(_)), assert(current_codes(C)).
%eval_codes(C,_,Res,E,L) :- print(eval_codes(C)),nl,fail.
eval_codes(E,Q,R,En,LS,TypeInfo) :-
   set_error_context(eval_codes),
   eval_codes_aux(E,Q,R,En,LS,TypeInfo),
   clear_error_context.
eval_codes_aux([58,116,32|Expression],_,Res,EnumWarning,LocalState,typing) :- % :t Expression  Haskell like command
  % Should we always enable this command ?
        !, EnumWarning=false, LocalState=[],
        (b_parse_machine_expression_from_codes(Expression,[prob,variables],_Typed,Type,true,Error)
         -> (Error=none
             -> translate:pretty_type(Type,PrettyType),
               (kernel_objects:max_cardinality(Type,Card) -> true ; Card='??'),
                ajoin([PrettyType,' /* card=',Card,' */'],Res),
                print(Res),nl
             ;  (Error=type_error -> Res = 'TYPE ERROR' ; Res = 'SYNTAX ERROR'),
                 print('Not a valid expression'),nl)
        ;  print('Parsing failed'),nl,
           Res = 'SYNTAX ERROR').
eval_codes_aux(Expression,_,Res,EnumWarning,LocalState,TypeInfo) :- set_current_codes(Expression),
        debug_println(20,parsing_as_expression(Expression)),
        b_parse_machine_expression_from_codes(Expression,[prob,variables],Typed,Type,false,Error),
        debug_println(20,parse_result(Error,Type)),
        (Error=none ; Error=type_error),!,
        (Error=type_error
          -> show_error_pos,print('TYPE ERROR'),nl, 
             Res='TYPE ERROR' , EnumWarning=false, LocalState=[], TypeInfo=error
          ; (Type=pred -> eval_predicate(Typed,Res,EnumWarning,LocalState) % happens for DEFINITIONS ??
              ; eval_expression(Typed,Res,EnumWarning),LocalState=[]),
             extract_type_information(Typed,TypeInfo)
    	).
eval_codes_aux(Expression,OuterQuantifier,Res,EnumWarning,LocalState,TypeInfo) :-
        debug_println(20,parsing_as_predicate(Expression)),
        b_parse_machine_predicate_from_codes_open(OuterQuantifier,Expression,
                                                  [],[prob,variables],Typed),!,
        %print(Typed),nl,
        debug_println(20,eval_predicate(Typed)),
        eval_predicate(Typed,Res,EnumWarning,LocalState),
        extract_type_information(Typed,TypeInfo).
eval_codes_aux(_E,_,'SYNTAX ERROR',false,[],syntax_error) :- show_error_pos,
        print('Not a valid expression or predicate'),nl.


extract_type_information(b(exists(Parameters,_Typed),pred,_),exists(Parameters)) :- !.
extract_type_information(b(forall(Parameters,_LHS,_RHS),pred,_),forall(Parameters)) :- !.
extract_type_information(b(_,pred,_),predicate) :- !.
extract_type_information(b(_,T,_),expression(T)).

:- use_module(library(timeout)).
:- use_module(library(file_systems)).

% 
% 
%  eval_rule_file removed; now use: probcli -eval_rule_file /Users/leuschel/svn_root/NewProBPrivate/examples/B/Siemens/RuleValidation/Deploy/AssociativityXY_3_type.v /Users/leuschel/svn_root/NewProBPrivate/examples/B/Siemens/RuleValidation/MainRuleBaseFile.mch

% performance benchmark:
% time(eval_strings:test_parser('examples/Rules/sample.rule')).
% time(eval_strings:test_parser('examples/Rules/sample2.rule')).
% time(eval_strings:test_parser('examples/Rules/sudoku.rule')).
:- public test_parser/1.
test_parser(File) :- print(processing_rule_file(File)),nl,
   read_string_from_file(File,Codes),print(parsing),nl,
   b_parse_machine_predicate_from_codes_open(exists,Codes,[],[prob,variables],
                                             _Typed),
   print(done),nl.


% ----------
% EXPRESSIONS 
% ----------

:- use_module(state_space,[current_state_id/1]).
:- use_module(store).
:- use_module(translate).
:- use_module(b_interpreter).
:- use_module(eclipse_interface,[getState/2]).

:- meta_predicate probcli_clpfd_overflow_mnf_call1(0).
:- meta_predicate probcli_clpfd_overflow_call1(0).

get_cur_state(State) :-
    get_cur_state2(State1), add_prob_deferred_set_elements_to_store(State1,State).
get_cur_state2(EState) :- current_state_id(CurID), getState(CurID,EState),!.
get_cur_state2([]). % happens if no machine loaded

replace_expression_by_kodkod_if_enabled(Typed,NewExpression) :-
    (replace_expression_by_kodkod_if_enabled_aux(Typed,NewExpression) -> true
      ; NewExpression=Typed).
replace_expression_by_kodkod_if_enabled_aux(Typed,NewExpression) :-
    get_texpr_expr(Typed,comprehension_set(Ids,Pred)),
    replace_kodkod_if_enabled(Ids,Pred,NewPred),
    get_texpr_info(Typed,Info),get_texpr_type(Typed,Type),
    create_texpr(comprehension_set(Ids,NewPred),Type,Info,NewExpression).

eval_expression(Typed,RRes,EnumWarning) :- 
    get_cur_state(EState),
    eval_expression(EState,Typed,RRes,EnumWarning).

eval_expression(EState,Typed,RRes,EnumWarning) :-  % print('Start Eval Expression: '),nl,flush_output,
        enter_new_error_scope(ScopeID),
        replace_expression_by_kodkod_if_enabled(Typed,Typed2),
        probcli_clpfd_overflow_mnf_call1(b_interpreter:b_compute_expression(Typed2,[],EState,Value)),
        debug_println(20,normalising(Value)),
        !,
        ((store:normalise_value_for_var(eval_strings,Value,NValue),
         get_texpr_type(Typed2,ValueType),
         translate:translate_bvalue_with_type(NValue,ValueType,Result))
         -> set_last_expression(expr,Typed2,NValue)
           ; print('Could not normalise value:'),nl,
                   translate:translate:translate_bvalue(Value,Result)),
        get_enum_warning(EnumWarning),
        print('Expression Value = '),nl,display_and_set_result(ScopeID,Result,RRes).
eval_expression(_,_,'NOT-WELL-DEFINED',false) :- exit_error_scope(_ScopeID,_).


% ----------
% PREDICATES
% ----------

:- use_module(kernel_waitflags).
:- use_module(b_enumerate,[b_tighter_enumerate_values/2]).
:- use_module(bsyntaxtree).


test_bool_exists(EState, Parameters,Typed,LocalState,WF) :- 
        init_wait_flags(WF),
        b_interpreter:set_up_typed_localstate(Parameters,_FreshOutputVars,TypedVals,[],LocalState),
        b_enumerate:b_tighter_enumerate_values(TypedVals,WF),
        b_interpreter:b_test_boolean_expression(Typed,LocalState,EState,WF).


eval_predicate(ExTyped,RRes,EnumWarning,LocalState) :- 
        get_cur_state(State),
        eval_predicate(State,ExTyped,RRes,EnumWarning,LocalState).

:- use_module('kodkod/kodkod', [apply_kodkod/3]).
replace_kodkod_if_enabled(Parameters,Typed,NewPredicate) :-
    b_get_machine_constants(Constants),
    b_get_machine_variables(Variables),
    append([Parameters,Variables,Constants],Identifiers),
    apply_kodkod(Identifiers,Typed,NewPredicate).

:- use_module(partition_detection,[detect_all_partitions_in_predicate/2]).

eval_predicate(State,ExTyped,RRes,EnumWarning,LocalState) :- 
        ExTyped=b(exists(Parameters,Typed),pred,_I),!,
        %print('Existentially Quantified Predicate is '),flush_output,
        detect_all_partitions_in_predicate(Typed,Typed2), % detect set partitions
        replace_kodkod_if_enabled(Parameters,Typed2,NTyped),
        enter_new_error_scope(ScopeID),
        (probcli_clpfd_overflow_call1((test_bool_exists(State, Parameters,NTyped,LocalState,WF),
                                        eval_ground_wf(WF)))
          -> (eval_det -> Res = 'POSSIBLY TRUE'  ; Res = 'TRUE'),
             set_last_expression(pred,ExTyped,true)
          ;  Res = 'FALSE', set_last_expression(pred,ExTyped,false)),
        get_enum_warning(EnumWarning),
        print('Existentially Quantified Predicate over '), print_parameters(Parameters),
        print(' is '),display_and_set_result(ScopeID,Res,RRes),
         (Res='FALSE' -> true
          ; print('Solution: '),nl, display_solution(Parameters,LocalState)
         ).
         %,(eval_det -> portray_waitflags(WF),trace ; true).
eval_predicate(State, ExTyped,RRes,EnumWarning,LocalState) :- 
        ExTyped=b(forall(Parameters,TypedLHS,TypedRHS),pred,_I),!,
        %print('Universally Quantified Predicate is '),flush_output,
        enter_new_error_scope(ScopeID),
        create_texpr(negation(TypedRHS),pred,[try_smt],NegRHS),
        conjunct_predicates([TypedLHS,NegRHS],Conjunction),
        replace_kodkod_if_enabled(Parameters,Conjunction,NConjunction),
        %translate:print_bexpr(Conjunction),
        (probcli_clpfd_overflow_call1((test_bool_exists(State, Parameters,NConjunction,LocalState,WF),
                                        eval_ground_wf(WF)))
          -> (eval_det -> Res = 'POSSIBLY TRUE'  ; Res = 'FALSE'),
             set_last_expression(pred,ExTyped,false)
          ;  Res = 'TRUE', set_last_expression(pred,ExTyped,true)),
        get_enum_warning(EnumWarning),
        print('Universally Quantified Predicate over '), print_parameters(Parameters),
        print(' is '), display_and_set_result(ScopeID,Res,RRes),
         (Res='TRUE' -> true
          ; print('Counter example: '),nl, display_solution(Parameters,LocalState)
         ).
eval_predicate(State, Typed,RRes,EnumWarning,[]) :-
        enter_new_error_scope(ScopeID),
        debug_println(20,test_boolean_expression(Typed)),
        detect_all_partitions_in_predicate(Typed,Typed2), % detect set partitions
        replace_kodkod_if_enabled([],Typed2,NTyped),
        (probcli_clpfd_overflow_call1(b_interpreter:b_test_boolean_expression_wf(NTyped,[],State))
          -> Res='TRUE',set_last_expression(pred,Typed,true)
          ;  Res='FALSE',set_last_expression(pred,Typed,false)),
        get_enum_warning(EnumWarning),
        print('Predicate is '),display_and_set_result(ScopeID,Res,RRes).

get_enum_warning(EnumWarning) :-
   (event_occured_in_error_scope(enumeration_warning(_,_,_,_))
          -> EnumWarning=true ; EnumWarning=false).
display_and_set_result(ScopeID,Res,RRes) :-
     exit_error_scope(ScopeID,ErrOcc),ErrOcc=true,!,
     show_error_pos,
     print('NOT-WELL-DEFINED ('), print(Res), print(')'),nl,
     RRes = 'NOT-WELL-DEFINED'.
display_and_set_result(_,Res,RRes) :- print(Res), nl, RRes=Res.
:- use_module(graph_canon,[print_cstate_graph/2]).
display_solution(Parameters,LocalState) :-
     (eval_det -> print_solution_with_type(Parameters,LocalState)
      ; normalise_store(LocalState,NState)
       -> %translate:print_bstate(NState)
          print_solution_with_type(Parameters,NState),
          %visualize_graph:tcltk_print_state_as_graph_for_dot(NState,'~/Desktop/out.dot')
          (eval_dot_file(File) -> graph_canon:print_cstate_graph(NState,File) ; true)
        ; print('Could not normalise value(s):'),nl,translate:print_bstate(LocalState)
     ),nl.

:- volatile eval_dot_file/1.
:- dynamic eval_dot_file/1.

set_eval_dot_file(F) :- retractall(eval_dot_file(_)),
    assert(eval_dot_file(F)).

% print solution using type information
print_solution_with_type([],[]).
print_solution_with_type([Identifier|TT],[bind(Varname,Value)|VT]) :-
    def_get_texpr_id(Identifier,Varname), !,
    get_texpr_type(Identifier,Type),
    print('       '),print(Varname), print(' = '),
    translate:translate_bvalue_with_type(Value,Type,Result),
    print(Result),
    (TT=[] -> nl
            ; print(' &'),nl,print_solution_with_type(TT,VT)).
print_solution_with_type(P,State) :-
    add_internal_error('Illegal call: ', print_solution_with_type(P,State)).

:- use_module(bsyntaxtree,[def_get_texpr_id/2]).
print_parameters([]).
print_parameters([TID]) :- !, print_id(TID).
print_parameters([TID|T]) :- print_id(TID), print(','), print_parameters(T).
print_id(TID) :- def_get_texpr_id(TID,ID), print(ID).

% TOOLS 
% -----
:- use_module(clpfd_interface,[clpfd_overflow_error_message/0]).


probcli_clpfd_overflow_mnf_call1(Call) :- start_timer(T1,WT1),
  on_exception(error(representation_error('CLPFD integer overflow'),_),
               (Call->stop_timer(T1,WT1); stop_timer(T1,WT1),print('Expression not well-defined !'),nl,fail),
               ( stop_timer(T1,WT1),clpfd_interface:clpfd_overflow_error_message, fail)).

probcli_clpfd_overflow_call1(Call) :- start_timer(T1,WT1),
  on_exception(error(representation_error('CLPFD integer overflow'),_),
               (Call -> stop_timer(T1,WT1) ; stop_timer(T1,WT1),fail),
               ( stop_timer(T1,WT1),clpfd_interface:clpfd_overflow_error_message, fail)).

:- volatile last_eval_time/2.
:- dynamic last_eval_time/2.
start_timer(T1,WT1) :- retractall(last_eval_time(_,_)),
                   statistics(runtime,[T1,_]),
                   statistics(walltime,[WT1,_]).
stop_timer(T1,WT1) :- 
        statistics(runtime,[T2,_]), TotTime is T2-T1,
        statistics(walltime,[WT2,_]), WTotTime is WT2-WT1,
        retractall(last_eval_time(_,_)),
        debug_println(20,stopped_timer(TotTime,T2,T1)),
        assert(last_eval_time(TotTime,WTotTime)).


:- volatile eval_det/0.
:- dynamic eval_det/0.
toggle_eval_det :-
  (retract(eval_det) ->  print('% Going back to full enumeration')
    ; assert(eval_det),
      print('% Going to deterministic mode (for existentially quantified formulas)')),nl.

set_eval_det :- eval_det -> true ; assert(eval_det).
unset_eval_det :- retractall(eval_det).

eval_ground_wf(WF) :- (eval_det -> ground_det_wait_flag(WF) ; ground_wait_flags(WF)).





show_error_pos :-
  (get_error_span_linecol(Line,Col,EL,EC) -> show_error_pos(Line,Col,EL,EC) ; true).
show_error_pos(Line,Col,EL,EC) :- %print(pos(Line,Col,EL,EC)),nl,
   ((current_codes(E),get_line(E,Line,ErrLine))
     -> print('### '),atom_codes(N,ErrLine), print(N),nl
      ; true),
   print('### '), C1 is Col - 1, indent_ws(C1),
   (EL=Line,EC>Col -> Len is EC-Col ; Len=1),
   underline(Len),nl,
   (Line>1 -> print('### Line: '), print(Line),
              print(' Column: '), print(Col),nl
            ; true).
indent_ws(X) :- X<1,!.
indent_ws(X) :- print(' '), X1 is X-1, indent_ws(X1).
underline(0) :- !.
underline(N) :- N>0, print('^'), N1 is N-1, underline(N1).

get_line(_,Nr,R) :- Nr<1,!,R=[].
get_line([],_,R) :- !, R=[].
get_line(Codes,Nr,Res) :- is_newline(Codes,T),!, N1 is Nr-1, get_line(T,N1,Res).
get_line([H|T],1,Res) :- !, Res=[H|RT], get_line(T,1,RT).
get_line([_H|T],Nr,Res) :- get_line(T,Nr,Res).

is_newline([10,13|T],T).
is_newline([10|T],T).
is_newline([13|T],T).

:- volatile last_expression/3.
:- dynamic last_expression/3.
set_last_expression(Type,Expr,Value) :- retractall(last_expression(_,_,_)),
  assert(last_expression(Type,Expr,Value)).

:- use_module(bsyntaxtree,[get_texpr_type/2]).

print_last_info :- \+ last_expression(_,_Expr,_Value),!,
  print('Please evaluate an expression or predicate first.'),nl.
print_last_info :- last_expression(Type,Expr,_Value),
  % for insertion into unit tests
   print('% Type: '), print_quoted(Type),
   ((Type=expr,get_texpr_type(Expr,ET),pretty_type(ET,PrettyType))
     -> print(' : '), print(PrettyType),
        print(' [Card='),
        (kernel_objects:max_cardinality(ET,Card) -> print(Card) ; print('??')), print(']')
     ;  true),nl,

   print('% Eval Time: '),
   (last_eval_time(Time,WTime)
    -> print(Time), print(' ms ('), print(WTime),print(' ms walltime)')
     ; print('*UNKOWN*')),nl.

print_last_expression :- \+ last_expression(_,_Expr,_Value),!,
  print('Please evaluate an expression or predicate first.'),nl.
print_last_expression :- last_expression(_Type,Expr,Value),
   print_last_info,
   print('% '), translate:print_bexpr(Expr),nl,
   (Value = false -> print_quoted(must_fail_det(1,"",Expr)),print('.'),nl
    ; print_quoted(Expr),nl,
      print('% = '),print_quoted(Value),nl).

:- use_module(library(system)).
add_last_expression_to_unit_tests :- \+ last_expression(_,_Expr,_Value),!,
  print('Please evaluate an expression or predicate first.'),nl.
add_last_expression_to_unit_tests :- last_expression(Type,Expr,Value),
  open_unit_test_file(S), !, write(S,' & '), nl(S), write(S,'  ( '),
  (print_unit_test_assertion(Type,Expr,Value,S) -> true ; print('PRINTING FAILED'),nl),
  write(S,' ) '), datime(datime(Yr,Mon,Day,Hr,Min,_Sec)),
  format(S,'/* ~w/~w/~w ~w:~w */',[Day,Mon,Yr,Hr,Min]),
  nl(S), close(S).
add_last_expression_to_unit_tests :- print('Could not save expression to unit test file ($PROB_EX_DIR/B/Laws/REPL_UNIT_TESTS.def).'),nl.

open_unit_test_file(S) :-
  environ('PROB_EX_DIR',Dir),
  atom_concat(Dir,'/B/Laws/REPL_UNIT_TESTS.def',REPLFILE),
  print(opening(REPLFILE)),nl,
  my_open(REPLFILE,append,S).


my_open(File,Mode,S) :- on_exception(error(existence_error(_,_),E),
                   open(File,Mode,S), add_error_fail(my_open,'File does not exist: ',File:E)).

:- use_module(bsyntaxtree,[create_texpr/4]).
print_unit_test_assertion(pred,Typed,true,S) :-
   print_bexpr_stream(S,Typed).
print_unit_test_assertion(pred,Typed,false,S) :-
   create_texpr(negation(Typed),pred,[],Neg),
   print_bexpr_stream(S,Neg).
print_unit_test_assertion(expr,Typed,Val,S) :-
  print_bexpr_stream(S,Typed), write(S,' = '), print_bvalue_stream(S,Val).



%% DETERMINISTIC PROPAGATION UNIT TESTS
%% ------------------------------------

% check that certain predicates are determined to be failing without enumeration
% to do: also add positive tests: certain predicates succeed with instantiating all variables
% eval_strings:must_fail_tests.

must_fail_tests :- set_eval_det,
   must_fail_det(Nr,Str,ExTyped), check_failed(Nr,ExTyped),
   Str \= [],
   % print('PARSING: '),nl,name(SS,Str), print(SS),nl,
   ( b_parse_machine_predicate_from_codes_open(exists,Str,[],[prob,variables],NewTyped) ->
       check_same_ast(Nr,NewTyped,ExTyped) 
   ; otherwise -> 
       print('*** Could not parse: '), print(Nr), nl
       %add_error(must_fail_tests,'Could not parse saved string: ',Nr)
   ),
   fail.
must_fail_tests :- preferences:preference(use_clpfd_solver,true),
                   must_fail_clpfd_det(Nr,ExTyped), check_failed(Nr,ExTyped),fail.
must_fail_tests :- unset_eval_det.

check_failed(Nr,ExTyped) :- 
            print(Nr), print(' >>> '), translate:print_bexpr(ExTyped),nl,
            (eval_predicate(ExTyped,Res,_,_)
             -> (Res='FALSE' -> true ; 
                add_error(must_fail_tests,'Test has not failed deterministically: ',Nr))
             ;  add_error(must_fail_tests,'eval_predicate failed: ',Nr)
            ).

:- use_module(self_check).

:- assert_must_succeed(must_fail_tests).

check_same_ast(Nr,A,B) :- % check equivalent to stored AST, apart from pos
   remove_pos(A,RA),
   (RA=B -> true
      ; add_error(check_same_ast,'AST not identical: ',Nr),
        print(A),nl,print(B),nl,
        tools_printing:trace_unify(A,B)).
% remove position Info
remove_pos(X,R) :- var(X),!,R=X.
remove_pos(pos(_,_,_,_,_,_),R) :- !, R=pos(_,_,_,_,_,_).
remove_pos(X,R) :- atomic(X),!,R=X.
remove_pos(L,R) :- L=..[F|Args], l_remove_pos(Args,RA), R=..[F|RA].

l_remove_pos([],[]).
l_remove_pos([H|T],[RH|RT]) :- remove_pos(H,RH), l_remove_pos(T,RT).

% 
must_fail_det(1,"#(x,y,z).(x:BOOL & x=z & z=y & x /= y)",
b(exists([b(identifier(x),boolean,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(y),boolean,[nodeid(pos(8,1,4,5,4,5))]),b(identifier(z),boolean,[nodeid(pos(9,1,4,7,4,7))])],b(conjunct(b(conjunct(b(equal(b(identifier(x),boolean,[nodeid(pos(17,1,4,20,4,20))]),b(identifier(z),boolean,[nodeid(pos(18,1,4,22,4,22))])),pred,[nodeid(pos(16,1,4,20,4,22))]),b(equal(b(identifier(z),boolean,[nodeid(pos(20,1,4,26,4,26))]),b(identifier(y),boolean,[nodeid(pos(21,1,4,28,4,28))])),pred,[nodeid(pos(19,1,4,26,4,28))])),pred,[nodeid(pos(11,1,4,11,4,28))]),b(not_equal(b(identifier(x),boolean,[nodeid(pos(23,1,4,32,4,32))]),b(identifier(y),boolean,[nodeid(pos(24,1,4,37,4,37))])),pred,[nodeid(pos(22,1,4,32,4,37))])),pred,[nodeid(pos(10,1,4,11,4,37))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,38))])).

% 
must_fail_det(2,"#x.(x:BOOL & (x=FALSE <=> x=TRUE))", b(exists([b(identifier(x),boolean,[nodeid(pos(7,1,4,2,4,2))])],b(equivalence(b(equal(b(identifier(x),boolean,[nodeid(pos(14,1,4,15,4,15))]),b(boolean_false,boolean,[nodeid(pos(15,1,4,17,4,21))])),pred,[nodeid(pos(13,1,4,15,4,21))]),b(equal(b(identifier(x),boolean,[nodeid(pos(17,1,4,27,4,27))]),b(boolean_true,boolean,[nodeid(pos(18,1,4,29,4,32))])),pred,[nodeid(pos(16,1,4,27,4,32))])),pred,[nodeid(pos(12,1,4,15,4,32))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,34))])).

% 
must_fail_det(3,"#(b,c).(b:BOOL & b=c & b/=c)", b(exists([b(identifier(b),boolean,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(c),boolean,[nodeid(pos(8,1,4,5,4,5))])],b(conjunct(b(equal(b(identifier(b),boolean,[nodeid(pos(15,1,4,18,4,18))]),b(identifier(c),boolean,[nodeid(pos(16,1,4,20,4,20))])),pred,[nodeid(pos(14,1,4,18,4,20))]),b(not_equal(b(identifier(b),boolean,[nodeid(pos(18,1,4,24,4,24))]),b(identifier(c),boolean,[nodeid(pos(19,1,4,27,4,27))])),pred,[nodeid(pos(17,1,4,24,4,27))])),pred,[nodeid(pos(9,1,4,9,4,27))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,28))])).

% 
must_fail_det(4,"#(e,s).(s<:INT & e:s & e/:s)", b(exists([b(identifier(e),integer,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(s),set(integer),[nodeid(pos(8,1,4,5,4,5))])],b(conjunct(b(conjunct(b(subset(b(identifier(s),set(integer),[nodeid(pos(12,1,4,9,4,9))]),b(integer_set('INT'),set(integer),[nodeid(pos(13,1,4,12,4,14))])),pred,[nodeid(pos(11,1,4,9,4,14))]),b(member(b(identifier(e),integer,[nodeid(pos(15,1,4,18,4,18))]),b(identifier(s),set(integer),[nodeid(pos(16,1,4,20,4,20))])),pred,[nodeid(pos(14,1,4,18,4,20))])),pred,[nodeid(pos(10,1,4,9,4,20))]),b(not_member(b(identifier(e),integer,[nodeid(pos(18,1,4,24,4,24))]),b(identifier(s),set(integer),[nodeid(pos(19,1,4,27,4,27))])),pred,[nodeid(pos(17,1,4,24,4,27))])),pred,[nodeid(pos(9,1,4,9,4,27))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,28))])).

% 
must_fail_det(5,"#(x,y,z,v,xz).(x:BOOL & x/=y & v/=z & x=v & (x=z <=> xz=TRUE) & (xz=FALSE => x=TRUE) & (xz=FALSE => x=FALSE))", b(exists([b(identifier(x),boolean,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(y),boolean,[nodeid(pos(8,1,4,5,4,5))]),b(identifier(z),boolean,[nodeid(pos(9,1,4,7,4,7))]),b(identifier(v),boolean,[nodeid(pos(10,1,4,9,4,9))]),b(identifier(xz),boolean,[nodeid(pos(11,1,4,11,4,12))])],b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(x),boolean,[nodeid(pos(22,1,4,25,4,25))]),b(identifier(y),boolean,[nodeid(pos(23,1,4,28,4,28))])),pred,[nodeid(pos(21,1,4,25,4,28))]),b(not_equal(b(identifier(v),boolean,[nodeid(pos(25,1,4,32,4,32))]),b(identifier(z),boolean,[nodeid(pos(26,1,4,35,4,35))])),pred,[nodeid(pos(24,1,4,32,4,35))])),pred,[nodeid(pos(16,1,4,16,4,35))]),b(equal(b(identifier(x),boolean,[nodeid(pos(28,1,4,39,4,39))]),b(identifier(v),boolean,[nodeid(pos(29,1,4,41,4,41))])),pred,[nodeid(pos(27,1,4,39,4,41))])),pred,[nodeid(pos(15,1,4,16,4,41))]),b(equivalence(b(equal(b(identifier(x),boolean,[nodeid(pos(32,1,4,46,4,46))]),b(identifier(z),boolean,[nodeid(pos(33,1,4,48,4,48))])),pred,[nodeid(pos(31,1,4,46,4,48))]),b(equal(b(identifier(xz),boolean,[nodeid(pos(35,1,4,54,4,55))]),b(boolean_true,boolean,[nodeid(pos(36,1,4,57,4,60))])),pred,[nodeid(pos(34,1,4,54,4,60))])),pred,[nodeid(pos(30,1,4,46,4,60))])),pred,[nodeid(pos(14,1,4,16,4,60))]),b(implication(b(equal(b(identifier(xz),boolean,[nodeid(pos(39,1,4,66,4,67))]),b(boolean_false,boolean,[nodeid(pos(40,1,4,69,4,73))])),pred,[nodeid(pos(38,1,4,66,4,73))]),b(equal(b(identifier(x),boolean,[nodeid(pos(42,1,4,78,4,78))]),b(boolean_true,boolean,[nodeid(pos(43,1,4,80,4,83))])),pred,[nodeid(pos(41,1,4,78,4,83))])),pred,[nodeid(pos(37,1,4,66,4,83))])),pred,[nodeid(pos(13,1,4,16,4,83))]),b(implication(b(equal(b(identifier(xz),boolean,[nodeid(pos(46,1,4,89,4,90))]),b(boolean_false,boolean,[nodeid(pos(47,1,4,92,4,96))])),pred,[nodeid(pos(45,1,4,89,4,96))]),b(equal(b(identifier(x),boolean,[nodeid(pos(49,1,4,101,4,101))]),b(boolean_false,boolean,[nodeid(pos(50,1,4,103,4,107))])),pred,[nodeid(pos(48,1,4,101,4,107))])),pred,[nodeid(pos(44,1,4,89,4,107))])),pred,[nodeid(pos(12,1,4,16,4,107))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,109))])).

% 
must_fail_det(6,"#(x,xz).((x : BOOL & xz : BOOL) & (((x = TRUE) <=> not(x = TRUE)) <=> (xz = TRUE) & ((x = TRUE) <=> not(x = TRUE)) <=> (xz = FALSE)))", b(exists([b(identifier(x),boolean,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(xz),boolean,[nodeid(pos(8,1,4,5,4,6))])],b(conjunct(b(equivalence(b(equivalence(b(equal(b(identifier(x),boolean,[nodeid(pos(17,1,4,21,4,21))]),b(boolean_true,boolean,[nodeid(pos(18,1,4,23,4,26))])),pred,[nodeid(pos(16,1,4,21,4,26))]),b(negation(b(equal(b(identifier(x),boolean,[nodeid(pos(21,1,4,36,4,36))]),b(boolean_true,boolean,[nodeid(pos(22,1,4,38,4,41))])),pred,[nodeid(pos(20,1,4,36,4,41))])),pred,[nodeid(pos(19,1,4,32,4,42))])),pred,[nodeid(pos(15,1,4,21,4,42))]),b(equal(b(identifier(xz),boolean,[nodeid(pos(24,1,4,49,4,50))]),b(boolean_true,boolean,[nodeid(pos(25,1,4,52,4,55))])),pred,[nodeid(pos(23,1,4,49,4,55))])),pred,[nodeid(pos(14,1,4,21,4,55))]),b(equivalence(b(equivalence(b(equal(b(identifier(x),boolean,[nodeid(pos(29,1,4,62,4,62))]),b(boolean_true,boolean,[nodeid(pos(30,1,4,64,4,67))])),pred,[nodeid(pos(28,1,4,62,4,67))]),b(negation(b(equal(b(identifier(x),boolean,[nodeid(pos(33,1,4,77,4,77))]),b(boolean_true,boolean,[nodeid(pos(34,1,4,79,4,82))])),pred,[nodeid(pos(32,1,4,77,4,82))])),pred,[nodeid(pos(31,1,4,73,4,83))])),pred,[nodeid(pos(27,1,4,62,4,83))]),b(equal(b(identifier(xz),boolean,[nodeid(pos(36,1,4,90,4,91))]),b(boolean_false,boolean,[nodeid(pos(37,1,4,93,4,97))])),pred,[nodeid(pos(35,1,4,90,4,97))])),pred,[nodeid(pos(26,1,4,62,4,97))])),pred,[nodeid(pos(9,1,4,10,4,97))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,99))])).

% 
must_fail_det(7,"#(x,z,x2,xz).((((x : BOOL & z : BOOL) & x2 : BOOL) & xz : BOOL) & (((x2 /= z & x = x2) & ((x = TRUE) <=> (x2 = FALSE)) <=> (xz = TRUE)) & ((x = TRUE) <=> (x2 = FALSE)) <=> (xz = FALSE)))", b(exists([b(identifier(x),boolean,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(z),boolean,[nodeid(pos(8,1,4,5,4,5))]),b(identifier(x2),boolean,[nodeid(pos(9,1,4,7,4,8))]),b(identifier(xz),boolean,[nodeid(pos(10,1,4,10,4,11))])],b(conjunct(b(conjunct(b(conjunct(b(not_equal(b(identifier(x2),boolean,[nodeid(pos(19,1,4,24,4,25))]),b(identifier(z),boolean,[nodeid(pos(20,1,4,28,4,28))])),pred,[nodeid(pos(18,1,4,24,4,28))]),b(equal(b(identifier(x),boolean,[nodeid(pos(22,1,4,32,4,32))]),b(identifier(x2),boolean,[nodeid(pos(23,1,4,34,4,35))])),pred,[nodeid(pos(21,1,4,32,4,35))])),pred,[nodeid(pos(13,1,4,15,4,35))]),b(equivalence(b(equivalence(b(equal(b(identifier(x),boolean,[nodeid(pos(27,1,4,41,4,41))]),b(boolean_true,boolean,[nodeid(pos(28,1,4,43,4,46))])),pred,[nodeid(pos(26,1,4,41,4,46))]),b(equal(b(identifier(x2),boolean,[nodeid(pos(30,1,4,52,4,53))]),b(boolean_false,boolean,[nodeid(pos(31,1,4,55,4,59))])),pred,[nodeid(pos(29,1,4,52,4,59))])),pred,[nodeid(pos(25,1,4,41,4,59))]),b(equal(b(identifier(xz),boolean,[nodeid(pos(33,1,4,66,4,67))]),b(boolean_true,boolean,[nodeid(pos(34,1,4,69,4,72))])),pred,[nodeid(pos(32,1,4,66,4,72))])),pred,[nodeid(pos(24,1,4,41,4,72))])),pred,[nodeid(pos(12,1,4,15,4,72))]),b(equivalence(b(equivalence(b(equal(b(identifier(x),boolean,[nodeid(pos(38,1,4,80,4,80))]),b(boolean_true,boolean,[nodeid(pos(39,1,4,82,4,85))])),pred,[nodeid(pos(37,1,4,80,4,85))]),b(equal(b(identifier(x2),boolean,[nodeid(pos(41,1,4,91,4,92))]),b(boolean_false,boolean,[nodeid(pos(42,1,4,94,4,98))])),pred,[nodeid(pos(40,1,4,91,4,98))])),pred,[nodeid(pos(36,1,4,80,4,98))]),b(equal(b(identifier(xz),boolean,[nodeid(pos(44,1,4,105,4,106))]),b(boolean_false,boolean,[nodeid(pos(45,1,4,108,4,112))])),pred,[nodeid(pos(43,1,4,105,4,112))])),pred,[nodeid(pos(35,1,4,80,4,112))])),pred,[nodeid(pos(11,1,4,15,4,112))])),pred,[used_ids([]),
nodeid(pos(6,1,4,1,4,114))])).

% 
must_fail_det(8,"#(a,b).((a : BOOL & b : BOOL) & ((b = TRUE) <=> ((a = TRUE) <=> not(a = TRUE)) & (b = FALSE) <=> ((a = FALSE) <=> (a = TRUE))))", b(exists([b(identifier(a),boolean,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(b),boolean,[nodeid(pos(8,1,4,5,4,5))])],b(conjunct(b(equivalence(b(equal(b(identifier(b),boolean,[nodeid(pos(12,1,4,10,4,10))]),b(boolean_true,boolean,[nodeid(pos(13,1,4,12,4,15))])),pred,[nodeid(pos(11,1,4,10,4,15))]),b(equivalence(b(equal(b(identifier(a),boolean,[nodeid(pos(16,1,4,22,4,22))]),b(boolean_true,boolean,[nodeid(pos(17,1,4,24,4,27))])),pred,[nodeid(pos(15,1,4,22,4,27))]),b(negation(b(equal(b(identifier(a),boolean,[nodeid(pos(20,1,4,37,4,37))]),b(boolean_true,boolean,[nodeid(pos(21,1,4,39,4,42))])),pred,[nodeid(pos(19,1,4,37,4,42))])),pred,[nodeid(pos(18,1,4,33,4,43))])),pred,[nodeid(pos(14,1,4,22,4,43))])),pred,[nodeid(pos(10,1,4,10,4,43))]),b(equivalence(b(equal(b(identifier(b),boolean,[nodeid(pos(24,1,4,49,4,49))]),b(boolean_false,boolean,[nodeid(pos(25,1,4,51,4,55))])),pred,[nodeid(pos(23,1,4,49,4,55))]),b(equivalence(b(equal(b(identifier(a),boolean,[nodeid(pos(28,1,4,62,4,62))]),b(boolean_false,boolean,[nodeid(pos(29,1,4,64,4,68))])),pred,[nodeid(pos(27,1,4,62,4,68))]),b(equal(b(identifier(a),boolean,[nodeid(pos(31,1,4,74,4,74))]),b(boolean_true,boolean,[nodeid(pos(32,1,4,76,4,79))])),pred,[nodeid(pos(30,1,4,74,4,79))])),pred,[nodeid(pos(26,1,4,62,4,79))])),pred,[nodeid(pos(22,1,4,49,4,79))])),pred,[nodeid(pos(9,1,4,10,4,79))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,82))])).

% 
must_fail_det(9,"#(a,b,c).(((a : BOOL & b : BOOL) & c : BOOL) & ((((b = TRUE) <=> (a = TRUE => a = FALSE) & b = TRUE) & (a = FALSE => c = FALSE)) & (a = FALSE => c = TRUE)))", b(let_predicate([b(identifier(b),boolean,[nodeid(pos(8,1,4,5,4,5))])],[b(boolean_true,boolean,[nodeid(pos(38,1,4,96,4,99))])],b(exists([b(identifier(a),boolean,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(c),boolean,[nodeid(pos(9,1,4,7,4,7))])],b(conjunct(b(conjunct(b(equivalence(b(equal(b(identifier(b),boolean,[nodeid(pos(27,1,4,52,4,52))]),b(boolean_true,boolean,[nodeid(pos(28,1,4,56,4,59))])),pred,[nodeid(pos(26,1,4,52,4,59))]),b(implication(b(equal(b(identifier(a),boolean,[nodeid(pos(31,1,4,67,4,67))]),b(boolean_true,boolean,[nodeid(pos(32,1,4,71,4,74))])),pred,[nodeid(pos(30,1,4,67,4,74))]),b(equal(b(identifier(a),boolean,[nodeid(pos(34,1,4,79,4,79))]),b(boolean_false,boolean,[nodeid(pos(35,1,4,83,4,87))])),pred,[nodeid(pos(33,1,4,79,4,87))])),pred,[nodeid(pos(29,1,4,67,4,87))])),pred,[nodeid(pos(25,1,4,52,4,87))]),b(implication(b(equal(b(identifier(a),boolean,[nodeid(pos(41,1,4,105,4,105))]),b(boolean_false,boolean,[nodeid(pos(42,1,4,109,4,113))])),pred,[nodeid(pos(40,1,4,105,4,113))]),b(equal(b(identifier(c),boolean,[nodeid(pos(44,1,4,118,4,118))]),b(boolean_false,boolean,[nodeid(pos(45,1,4,122,4,126))])),pred,[nodeid(pos(43,1,4,118,4,126))])),pred,[nodeid(pos(39,1,4,105,4,126))])),pred,[]),b(implication(b(equal(b(identifier(a),boolean,[nodeid(pos(48,1,4,133,4,133))]),b(boolean_false,boolean,[nodeid(pos(49,1,4,137,4,141))])),pred,[nodeid(pos(47,1,4,133,4,141))]),b(equal(b(identifier(c),boolean,[nodeid(pos(51,1,4,146,4,146))]),b(boolean_true,boolean,[nodeid(pos(52,1,4,150,4,153))])),pred,[nodeid(pos(50,1,4,146,4,153))])),pred,[nodeid(pos(46,1,4,133,4,153))])),pred,[])),pred,[used_ids([b]),used_ids([b]),used_ids([a,b,c])])),pred,[nodeid(pos(6,1,4,1,4,156))])).

% 
must_fail_det(10,"#(a,b,c).(((a : BOOL & b : BOOL) & c : BOOL) & (((b = TRUE) <=> (a = TRUE => a = FALSE) & (c = TRUE) <=> (a = b)) & (c = TRUE) <=> not(a = b)))", b(exists([b(identifier(a),boolean,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(b),boolean,[nodeid(pos(8,1,4,5,4,5))]),b(identifier(c),boolean,[nodeid(pos(9,1,4,7,4,7))])],b(conjunct(b(conjunct(b(equivalence(b(equal(b(identifier(b),boolean,[nodeid(pos(14,1,4,12,4,12))]),b(boolean_true,boolean,[nodeid(pos(15,1,4,14,4,17))])),pred,[nodeid(pos(13,1,4,12,4,17))]),b(implication(b(equal(b(identifier(a),boolean,[nodeid(pos(18,1,4,25,4,25))]),b(boolean_true,boolean,[nodeid(pos(19,1,4,27,4,30))])),pred,[nodeid(pos(17,1,4,25,4,30))]),b(equal(b(identifier(a),boolean,[nodeid(pos(21,1,4,35,4,35))]),b(boolean_false,boolean,[nodeid(pos(22,1,4,37,4,41))])),pred,[nodeid(pos(20,1,4,35,4,41))])),pred,[nodeid(pos(16,1,4,25,4,41))])),pred,[nodeid(pos(12,1,4,12,4,41))]),b(equivalence(b(equal(b(identifier(c),boolean,[nodeid(pos(25,1,4,48,4,48))]),b(boolean_true,boolean,[nodeid(pos(26,1,4,50,4,53))])),pred,[nodeid(pos(24,1,4,48,4,53))]),b(equal(b(identifier(a),boolean,[nodeid(pos(28,1,4,60,4,60))]),b(identifier(b),boolean,[nodeid(pos(29,1,4,62,4,62))])),pred,[nodeid(pos(27,1,4,60,4,62))])),pred,[nodeid(pos(23,1,4,48,4,62))])),pred,[nodeid(pos(11,1,4,12,4,62))]),b(equivalence(b(equal(b(identifier(c),boolean,[nodeid(pos(32,1,4,69,4,69))]),b(boolean_true,boolean,[nodeid(pos(33,1,4,71,4,74))])),pred,[nodeid(pos(31,1,4,69,4,74))]),b(negation(b(equal(b(identifier(a),boolean,[nodeid(pos(36,1,4,84,4,84))]),b(identifier(b),boolean,[nodeid(pos(37,1,4,86,4,86))])),pred,[nodeid(pos(35,1,4,84,4,86))])),pred,[nodeid(pos(34,1,4,80,4,87))])),pred,[nodeid(pos(30,1,4,69,4,87))])),pred,[nodeid(pos(10,1,4,12,4,87))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,89))])).

% 
must_fail_det(11,"#(x,y,z).(((x : BOOL & y : BOOL) & z : BOOL) & ((TRUE:BOOL) => ((x : {TRUE}) <=> (y = TRUE) & (x = y) <=> (z = TRUE)) & x /= y <=> (z = TRUE)))", b(exists([b(identifier(x),boolean,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(y),boolean,[nodeid(pos(8,1,4,5,4,5))]),b(identifier(z),boolean,[nodeid(pos(9,1,4,7,4,7))])],b(implication(b(truth,pred,[nodeid(pos(19,1,4,29,4,34))]),b(conjunct(b(conjunct(b(equivalence(b(member(b(identifier(x),boolean,[nodeid(pos(26,1,4,40,4,40))]),b(set_extension([b(boolean_true,boolean,[nodeid(pos(28,1,4,43,4,46))])]),set(boolean),[nodeid(pos(27,1,4,42,4,47))])),pred,[nodeid(pos(25,1,4,40,4,47))]),b(equal(b(identifier(y),boolean,[nodeid(pos(30,1,4,53,4,53))]),b(boolean_true,boolean,[nodeid(pos(31,1,4,55,4,58))])),pred,[nodeid(pos(29,1,4,53,4,58))])),pred,[nodeid(pos(24,1,4,40,4,58))]),b(equivalence(b(equal(b(identifier(x),boolean,[nodeid(pos(34,1,4,64,4,64))]),b(identifier(y),boolean,[nodeid(pos(35,1,4,66,4,66))])),pred,[nodeid(pos(33,1,4,64,4,66))]),b(equal(b(identifier(z),boolean,[nodeid(pos(37,1,4,72,4,72))]),b(boolean_true,boolean,[nodeid(pos(38,1,4,74,4,77))])),pred,[nodeid(pos(36,1,4,72,4,77))])),pred,[nodeid(pos(32,1,4,64,4,77))])),pred,[nodeid(pos(23,1,4,40,4,77))]),b(equivalence(b(not_equal(b(identifier(x),boolean,[nodeid(pos(41,1,4,82,4,82))]),b(identifier(y),boolean,[nodeid(pos(42,1,4,85,4,85))])),pred,[nodeid(pos(40,1,4,82,4,85))]),b(equal(b(identifier(z),boolean,[nodeid(pos(44,1,4,90,4,90))]),b(boolean_true,boolean,[nodeid(pos(45,1,4,92,4,95))])),pred,[nodeid(pos(43,1,4,90,4,95))])),pred,[nodeid(pos(39,1,4,82,4,95))])),pred,[nodeid(pos(22,1,4,40,4,95))])),pred,[nodeid(pos(10,1,4,11,4,95))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,97))])).

 
must_fail_det(12,"SIGMA(x,y).(x : 1 .. 10 & y : {1,2}|x) = 111", b(equal(b(general_sum([b(identifier(x),integer,[nodeid(pos(0,0,0,0,0,0))]),b(identifier(y),integer,[nodeid(pos(0,0,0,0,0,0))])],b(conjunct(b(member(b(identifier(x),integer,[nodeid(pos(0,0,0,0,0,0))]),b(interval(b(integer(1),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(10),integer,[nodeid(pos(0,0,0,0,0,0))])),set(integer),[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))]),b(member(b(identifier(y),integer,[nodeid(pos(0,0,0,0,0,0))]),b(set_extension([b(integer(1),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(2),integer,[nodeid(pos(0,0,0,0,0,0))])]),set(integer),[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))]),b(identifier(x),integer,[nodeid(pos(0,0,0,0,0,0))])),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(111),integer,[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))])).

must_fail_det(13,"PI(x,y).(x : 1 .. 10 & y : {1,2}|x) = 111", b(equal(b(general_product([b(identifier(x),integer,[nodeid(pos(0,0,0,0,0,0))]),b(identifier(y),integer,[nodeid(pos(0,0,0,0,0,0))])],b(conjunct(b(member(b(identifier(x),integer,[nodeid(pos(0,0,0,0,0,0))]),b(interval(b(integer(1),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(10),integer,[nodeid(pos(0,0,0,0,0,0))])),set(integer),[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))]),b(member(b(identifier(y),integer,[nodeid(pos(0,0,0,0,0,0))]),b(set_extension([b(integer(1),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(2),integer,[nodeid(pos(0,0,0,0,0,0))])]),set(integer),[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))]),b(identifier(x),integer,[nodeid(pos(0,0,0,0,0,0))])),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(111),integer,[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))])).

/* 
must_fail_det(16,"union(ran(%x.((x <: {1,2,3,4} & card(x) > 0) & min(x) = 2|x))) = {1,2,3,4}",
b(equal(b(general_union(b(range(b(comprehension_set([b(identifier(x),set(integer),[nodeid(pos(0,0,0,0,0,0))]),b(identifier('_lambda_result_'),set(integer),[lambda_result])],b(conjunct(b(conjunct(b(conjunct(b(subset(b(identifier(x),set(integer),[nodeid(pos(0,0,0,0,0,0))]),b(set_extension([b(integer(1),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(2),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(3),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(4),integer,[nodeid(pos(0,0,0,0,0,0))])]),set(integer),[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))]),b(greater(b(card(b(identifier(x),set(integer),[nodeid(pos(0,0,0,0,0,0))])),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(0),integer,[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))])),pred,[nodeid(pos(0,0,0,0,0,0))]),b(equal(b(min(b(identifier(x),set(integer),[nodeid(pos(0,0,0,0,0,0))])),integer,[contains_wd_condition,nodeid(pos(0,0,0,0,0,0))]),b(integer(2),integer,[nodeid(pos(0,0,0,0,0,0))])),pred,[contains_wd_condition,nodeid(pos(0,0,0,0,0,0))])),pred,[contains_wd_condition,nodeid(pos(0,0,0,0,0,0))]),b(equal(b(identifier('_lambda_result_'),set(integer),[lambda_result]),b(identifier(x),set(integer),[nodeid(pos(0,0,0,0,0,0))])),pred,[])),pred,[contains_wd_condition])),set(couple(set(integer),set(integer))),[contains_wd_condition,was(lambda),generated(quantified_union)])),set(set(integer)),[contains_wd_condition,generated(quantified_union)])),set(integer),[contains_wd_condition,nodeid(pos(0,0,0,0,0,0))]),b(set_extension([b(integer(1),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(2),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(3),integer,[nodeid(pos(0,0,0,0,0,0))]),b(integer(4),integer,[nodeid(pos(0,0,0,0,0,0))])]),set(integer),[nodeid(pos(0,0,0,0,0,0))])),pred,[contains_wd_condition,nodeid(pos(0,0,0,0,0,0))])).
*/


% #(z,x,y).(((z : POW(POW(BOOL)) & x : POW(BOOL)) & y : INTEGER) & ((z = {{},{TRUE},{FALSE},{TRUE,FALSE}} & (x : z) <=> (y = 2)) & y > 3))
% check that we detect that z is a complete set, hence x:z and hence y=2
must_fail_det(14,"z = {{},{TRUE},{FALSE},{TRUE,FALSE}} & (x:z <=> y=2) & y>3 ", b(exists([b(identifier(z),set(set(boolean)),[]),b(identifier(x),set(boolean),[]),b(identifier(y),integer,[])],b(conjunct(b(conjunct(b(equal(b(identifier(z),set(set(boolean)),[nodeid(pos(nan,1,1,1,1,1))]),b(set_extension([b(empty_set,set(boolean),[nodeid(pos(nan,1,1,6,1,6))]),b(set_extension([b(boolean_true,boolean,[nodeid(pos(nan,1,1,10,1,10))])]),set(boolean),[nodeid(pos(nan,1,1,9,1,14))]),b(set_extension([b(boolean_false,boolean,[nodeid(pos(nan,1,1,17,1,17))])]),set(boolean),[nodeid(pos(nan,1,1,16,1,22))]),b(set_extension([b(boolean_true,boolean,[nodeid(pos(nan,1,1,25,1,25))]),b(boolean_false,boolean,[nodeid(pos(nan,1,1,30,1,30))])]),set(boolean),[nodeid(pos(nan,1,1,24,1,35))])]),set(set(boolean)),[nodeid(pos(nan,1,1,5,1,36))])),pred,[nodeid(pos(nan,1,1,1,1,36))]),b(equivalence(b(member(b(identifier(x),set(boolean),[nodeid(pos(nan,1,1,41,1,41))]),b(identifier(z),set(set(boolean)),[nodeid(pos(nan,1,1,43,1,43))])),pred,[nodeid(pos(nan,1,1,41,1,43))]),b(equal(b(identifier(y),integer,[nodeid(pos(nan,1,1,49,1,49))]),b(integer(2),integer,[nodeid(pos(nan,1,1,51,1,51))])),pred,[nodeid(pos(nan,1,1,49,1,51))])),pred,[nodeid(pos(nan,1,1,41,1,51))])),pred,[nodeid(pos(nan,1,1,1,1,52))]),b(greater(b(identifier(y),integer,[nodeid(pos(nan,1,1,56,1,56))]),b(integer(3),integer,[nodeid(pos(nan,1,1,58,1,58))])),pred,[nodeid(pos(nan,1,1,56,1,58))])),pred,[nodeid(pos(nan,1,1,1,1,58))])),pred,[used_ids([x,y,z])])).

% #(z,x,y).(((z : POW(BOOL * BOOL) & x : BOOL * BOOL) & y : INTEGER) & ((z = {TRUE |-> FALSE,TRUE |-> TRUE,FALSE |-> FALSE,FALSE |-> TRUE} & (x : z) <=> (y = 2)) & y > 3))
% check that we detect that z is a complete set, hence x:z and hence y=2
must_fail_det(15," z = {(TRUE,FALSE),(TRUE,TRUE),(FALSE,FALSE),(FALSE,TRUE)} & (x:z <=> y=2) & y>3 ", b(exists([b(identifier(z),set(couple(boolean,boolean)),[]),b(identifier(x),couple(boolean,boolean),[]),b(identifier(y),integer,[])],b(conjunct(b(conjunct(b(equal(b(identifier(z),set(couple(boolean,boolean)),[nodeid(pos(nan,1,1,1,1,1))]),b(set_extension([b(couple(b(boolean_true,boolean,[nodeid(pos(nan,1,1,7,1,7))]),b(boolean_false,boolean,[nodeid(pos(nan,1,1,12,1,12))])),couple(boolean,boolean),[nodeid(pos(nan,1,1,6,1,17))]),b(couple(b(boolean_true,boolean,[nodeid(pos(nan,1,1,20,1,20))]),b(boolean_true,boolean,[nodeid(pos(nan,1,1,25,1,25))])),couple(boolean,boolean),[nodeid(pos(nan,1,1,19,1,29))]),b(couple(b(boolean_false,boolean,[nodeid(pos(nan,1,1,32,1,32))]),b(boolean_false,boolean,[nodeid(pos(nan,1,1,38,1,38))])),couple(boolean,boolean),[nodeid(pos(nan,1,1,31,1,43))]),b(couple(b(boolean_false,boolean,[nodeid(pos(nan,1,1,46,1,46))]),b(boolean_true,boolean,[nodeid(pos(nan,1,1,52,1,52))])),couple(boolean,boolean),[nodeid(pos(nan,1,1,45,1,56))])]),set(couple(boolean,boolean)),[nodeid(pos(nan,1,1,5,1,57))])),pred,[nodeid(pos(nan,1,1,1,1,57))]),b(equivalence(b(member(b(identifier(x),couple(boolean,boolean),[nodeid(pos(nan,1,1,62,1,62))]),b(identifier(z),set(couple(boolean,boolean)),[nodeid(pos(nan,1,1,64,1,64))])),pred,[nodeid(pos(nan,1,1,62,1,64))]),b(equal(b(identifier(y),integer,[nodeid(pos(nan,1,1,70,1,70))]),b(integer(2),integer,[nodeid(pos(nan,1,1,72,1,72))])),pred,[nodeid(pos(nan,1,1,70,1,72))])),pred,[nodeid(pos(nan,1,1,62,1,72))])),pred,[nodeid(pos(nan,1,1,1,1,73))]),b(greater(b(identifier(y),integer,[nodeid(pos(nan,1,1,77,1,77))]),b(integer(3),integer,[nodeid(pos(nan,1,1,79,1,79))])),pred,[nodeid(pos(nan,1,1,77,1,79))])),pred,[nodeid(pos(nan,1,1,1,1,79))])),pred,[used_ids([x,y,z])])).

% #(z,x,y).(((z : POW(POW(BOOL)) & x : POW(POW(BOOL))) & y : INTEGER) & ((z = {{},{TRUE},{FALSE},{TRUE,FALSE}} & x <: z <=> (y = 2)) & y > 3))
must_fail_det(16,[],b(exists([b(identifier(z),set(set(boolean)),[]),b(identifier(x),set(set(boolean)),[]),b(identifier(y),integer,[])],b(conjunct(b(conjunct(b(equal(b(identifier(z),set(set(boolean)),[nodeid(pos(nan,1,1,1,1,1))]),b(set_extension([b(empty_set,set(boolean),[nodeid(pos(nan,1,1,6,1,6))]),b(set_extension([b(boolean_true,boolean,[nodeid(pos(nan,1,1,10,1,10))])]),set(boolean),[nodeid(pos(nan,1,1,9,1,14))]),b(set_extension([b(boolean_false,boolean,[nodeid(pos(nan,1,1,17,1,17))])]),set(boolean),[nodeid(pos(nan,1,1,16,1,22))]),b(set_extension([b(boolean_true,boolean,[nodeid(pos(nan,1,1,25,1,25))]),b(boolean_false,boolean,[nodeid(pos(nan,1,1,30,1,30))])]),set(boolean),[nodeid(pos(nan,1,1,24,1,35))])]),set(set(boolean)),[nodeid(pos(nan,1,1,5,1,36))])),pred,[nodeid(pos(nan,1,1,1,1,36))]),b(equivalence(b(subset(b(identifier(x),set(set(boolean)),[nodeid(pos(nan,1,1,41,1,41))]),b(identifier(z),set(set(boolean)),[nodeid(pos(nan,1,1,44,1,44))])),pred,[nodeid(pos(nan,1,1,41,1,44))]),b(equal(b(identifier(y),integer,[nodeid(pos(nan,1,1,50,1,50))]),b(integer(2),integer,[nodeid(pos(nan,1,1,52,1,52))])),pred,[nodeid(pos(nan,1,1,50,1,52))])),pred,[nodeid(pos(nan,1,1,41,1,52))])),pred,[nodeid(pos(nan,1,1,1,1,53))]),b(greater(b(identifier(y),integer,[nodeid(pos(nan,1,1,57,1,57))]),b(integer(3),integer,[nodeid(pos(nan,1,1,59,1,59))])),pred,[nodeid(pos(nan,1,1,57,1,59))])),pred,[nodeid(pos(nan,1,1,1,1,59))])),pred,[used_ids([x,y,z])])).

% Eval Time: 0 ms (0 ms walltime)
% #(x).(x : INTEGER & ([111,222,333,444,555] |>> {x} = [111,222,333,444] & x > 555))
must_fail_det(17,"#(x).(([111,222,333,444,555] |>> {x} = [111,222,333,444] & x > 555))", b(exists([b(identifier(x),integer,[nodeid(pos(nan,1,1,3,1,3))])],b(conjunct(b(equal(b(range_subtraction(b(sequence_extension([b(integer(111),integer,[nodeid(pos(nan,1,1,9,1,9))]),b(integer(222),integer,[nodeid(pos(nan,1,1,13,1,13))]),b(integer(333),integer,[nodeid(pos(nan,1,1,17,1,17))]),b(integer(444),integer,[nodeid(pos(nan,1,1,21,1,21))]),b(integer(555),integer,[nodeid(pos(nan,1,1,25,1,25))])]),set(couple(integer,integer)),[nodeid(pos(nan,1,1,8,1,28))]),b(set_extension([b(identifier(x),integer,[nodeid(pos(nan,1,1,35,1,35))])]),set(integer),[nodeid(pos(nan,1,1,34,1,36))])),set(couple(integer,integer)),[nodeid(pos(nan,1,1,8,1,36))]),b(sequence_extension([b(integer(111),integer,[nodeid(pos(nan,1,1,41,1,41))]),b(integer(222),integer,[nodeid(pos(nan,1,1,45,1,45))]),b(integer(333),integer,[nodeid(pos(nan,1,1,49,1,49))]),b(integer(444),integer,[nodeid(pos(nan,1,1,53,1,53))])]),set(couple(integer,integer)),[nodeid(pos(nan,1,1,40,1,56))])),pred,[nodeid(pos(nan,1,1,8,1,56))]),b(greater(b(identifier(x),integer,[nodeid(pos(nan,1,1,60,1,60))]),b(integer(555),integer,[nodeid(pos(nan,1,1,64,1,64))])),pred,[nodeid(pos(nan,1,1,60,1,64))])),pred,[nodeid(pos(nan,1,1,8,1,64))])),pred,[used_ids([]),nodeid(pos(nan,1,1,1,1,68))])).



% #(m,n).(m:30..100 & n:10..20 & m<n )
must_fail_clpfd_det(1,b(exists([b(identifier(m),integer,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(n),integer,[nodeid(pos(8,1,4,5,4,5))])],b(conjunct(b(conjunct(b(member(b(identifier(m),integer,[nodeid(pos(12,1,4,9,4,9))]),b(interval(b(integer(30),integer,[nodeid(pos(14,1,4,11,4,12))]),b(integer(100),integer,[nodeid(pos(15,1,4,15,4,17))])),set(integer),[nodeid(pos(13,1,4,11,4,17))])),pred,[nodeid(pos(11,1,4,9,4,17))]),b(member(b(identifier(n),integer,[nodeid(pos(17,1,4,21,4,21))]),b(interval(b(integer(10),integer,[nodeid(pos(19,1,4,23,4,24))]),b(integer(20),integer,[nodeid(pos(20,1,4,27,4,28))])),set(integer),[nodeid(pos(18,1,4,23,4,28))])),pred,[nodeid(pos(16,1,4,21,4,28))])),pred,[nodeid(pos(10,1,4,9,4,28))]),b(less(b(identifier(m),integer,[nodeid(pos(22,1,4,32,4,32))]),b(identifier(n),integer,[nodeid(pos(23,1,4,34,4,34))])),pred,[nodeid(pos(21,1,4,32,4,34))])),pred,[nodeid(pos(9,1,4,9,4,34))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,36))])).

%  #(m,n).(m:30..101 & n:10..20 & n*n=m & m/=100)
must_fail_clpfd_det(2,b(exists([b(identifier(m),integer,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(n),integer,[nodeid(pos(8,1,4,5,4,5))])],b(conjunct(b(conjunct(b(conjunct(b(member(b(identifier(m),integer,[nodeid(pos(13,1,4,9,4,9))]),b(interval(b(integer(30),integer,[nodeid(pos(15,1,4,11,4,12))]),b(integer(101),integer,[nodeid(pos(16,1,4,15,4,17))])),set(integer),[nodeid(pos(14,1,4,11,4,17))])),pred,[nodeid(pos(12,1,4,9,4,17))]),b(member(b(identifier(n),integer,[nodeid(pos(18,1,4,21,4,21))]),b(interval(b(integer(10),integer,[nodeid(pos(20,1,4,23,4,24))]),b(integer(20),integer,[nodeid(pos(21,1,4,27,4,28))])),set(integer),[nodeid(pos(19,1,4,23,4,28))])),pred,[nodeid(pos(17,1,4,21,4,28))])),pred,[nodeid(pos(11,1,4,9,4,28))]),b(equal(b(multiplication(b(identifier(n),integer,[nodeid(pos(24,1,4,32,4,32))]),b(identifier(n),integer,[nodeid(pos(25,1,4,34,4,34))])),integer,[nodeid(pos(23,1,4,32,4,34))]),b(identifier(m),integer,[nodeid(pos(26,1,4,36,4,36))])),pred,[nodeid(pos(22,1,4,32,4,36))])),pred,[nodeid(pos(10,1,4,9,4,36))]),b(not_equal(b(identifier(m),integer,[nodeid(pos(28,1,4,40,4,40))]),b(integer(100),integer,[nodeid(pos(29,1,4,43,4,45))])),pred,[nodeid(pos(27,1,4,40,4,45))])),pred,[nodeid(pos(9,1,4,9,4,45))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,46))])). 

% #(x,y).(y : BOOL & (x : NATURAL & (((x : {11}) <=> (y = TRUE) & x < 10) & (x : {12}) <=> (y = FALSE))))
must_fail_clpfd_det(3,b(exists([b(identifier(x),integer,[nodeid(pos(7,1,4,3,4,3))]),b(identifier(y),boolean,[nodeid(pos(8,1,4,5,4,5))])],b(conjunct(b(member(b(identifier(x),integer,[nodeid(pos(12,1,4,9,4,9))]),b(integer_set('NATURAL'),set(integer),[nodeid(pos(13,1,4,11,4,17))])),pred,[nodeid(pos(11,1,4,9,4,17))]),b(conjunct(b(conjunct(b(equivalence(b(member(b(identifier(x),integer,[nodeid(pos(21,1,4,32,4,32))]),b(set_extension([b(integer(11),integer,[nodeid(pos(23,1,4,35,4,36))])]),set(integer),[nodeid(pos(22,1,4,34,4,37))])),pred,[nodeid(pos(20,1,4,32,4,37))]),b(equal(b(identifier(y),boolean,[nodeid(pos(25,1,4,43,4,43))]),b(boolean_true,boolean,[nodeid(pos(26,1,4,45,4,48))])),pred,[nodeid(pos(24,1,4,43,4,48))])),pred,[nodeid(pos(19,1,4,32,4,48))]),b(less(b(identifier(x),integer,[nodeid(pos(28,1,4,53,4,53))]),b(integer(10),integer,[nodeid(pos(29,1,4,55,4,56))])),pred,[nodeid(pos(27,1,4,53,4,56))])),pred,[nodeid(pos(18,1,4,32,4,56))]),b(equivalence(b(member(b(identifier(x),integer,[nodeid(pos(32,1,4,61,4,61))]),b(set_extension([b(integer(12),integer,[nodeid(pos(34,1,4,64,4,65))])]),set(integer),[nodeid(pos(33,1,4,63,4,66))])),pred,[nodeid(pos(31,1,4,61,4,66))]),b(equal(b(identifier(y),boolean,[nodeid(pos(36,1,4,72,4,72))]),b(boolean_false,boolean,[nodeid(pos(37,1,4,74,4,78))])),pred,[nodeid(pos(35,1,4,72,4,78))])),pred,[nodeid(pos(30,1,4,61,4,78))])),pred,[nodeid(pos(17,1,4,32,4,78))])),pred,[nodeid(pos(9,1,4,9,4,78))])),pred,[used_ids([]),nodeid(pos(6,1,4,1,4,81))])).


% #(x,y).((x : INTEGER & y : INTEGER) & (({x} /<<: {11,12} <=> (y = 3) & x > 12) & y < 3))
must_fail_clpfd_det(4,b(exists([b(identifier(x),integer,[]),b(identifier(y),integer,[])],b(conjunct(b(conjunct(b(equivalence(b(not_subset_strict(b(set_extension([b(identifier(x),integer,[nodeid(pos(nan,1,1,2,1,2))])]),set(integer),[nodeid(pos(nan,1,1,1,1,3))]),b(set_extension([b(integer(11),integer,[nodeid(pos(nan,1,1,11,1,11))]),b(integer(12),integer,[nodeid(pos(nan,1,1,14,1,14))])]),set(integer),[nodeid(pos(nan,1,1,10,1,16))])),pred,[nodeid(pos(nan,1,1,1,1,16))]),b(equal(b(identifier(y),integer,[nodeid(pos(nan,1,1,22,1,22))]),b(integer(3),integer,[nodeid(pos(nan,1,1,24,1,24))])),pred,[nodeid(pos(nan,1,1,22,1,24))])),pred,[nodeid(pos(nan,1,1,1,1,24))]),b(greater(b(identifier(x),integer,[nodeid(pos(nan,1,1,28,1,28))]),b(integer(12),integer,[nodeid(pos(nan,1,1,30,1,30))])),pred,[nodeid(pos(nan,1,1,28,1,30))])),pred,[nodeid(pos(nan,1,1,1,1,30))]),b(less(b(identifier(y),integer,[nodeid(pos(nan,1,1,35,1,35))]),b(integer(3),integer,[nodeid(pos(nan,1,1,37,1,37))])),pred,[nodeid(pos(nan,1,1,35,1,37))])),pred,[nodeid(pos(nan,1,1,1,1,37))])),pred,[used_ids([x,y])])).

% #(x,y).((x : 5 .. 10000 & x / y = 19000) & y : 1 .. 100)
must_fail_clpfd_det(5,b(exists([b(identifier(x),integer,[]),b(identifier(y),integer,[])],b(conjunct(b(conjunct(b(member(b(identifier(x),integer,[nodeid(pos(nan,1,1,1,1,1))]),b(interval(b(integer(5),integer,[nodeid(pos(nan,1,1,3,1,3))]),b(integer(10000),integer,[nodeid(pos(nan,1,1,6,1,6))])),set(integer),[nodeid(pos(nan,1,1,3,1,6))])),pred,[nodeid(pos(nan,1,1,1,1,6))]),b(equal(b(div(b(identifier(x),integer,[nodeid(pos(nan,1,1,14,1,14))]),b(identifier(y),integer,[nodeid(pos(nan,1,1,16,1,16))])),integer,[contains_wd_condition,nodeid(pos(nan,1,1,14,1,16))]),b(integer(19000),integer,[nodeid(pos(nan,1,1,18,1,18))])),pred,[contains_wd_condition,nodeid(pos(nan,1,1,14,1,18))])),pred,[contains_wd_condition,nodeid(pos(nan,1,1,1,1,18))]),b(member(b(identifier(y),integer,[nodeid(pos(nan,1,1,26,1,26))]),b(interval(b(integer(1),integer,[nodeid(pos(nan,1,1,28,1,28))]),b(integer(100),integer,[nodeid(pos(nan,1,1,31,1,31))])),set(integer),[nodeid(pos(nan,1,1,28,1,31))])),pred,[nodeid(pos(nan,1,1,26,1,31))])),pred,[contains_wd_condition,nodeid(pos(nan,1,1,1,1,31))])),pred,[used_ids([x,y])])).

% #(x,y).((x : 500 .. 10000 & 499 / y = x) & y : 1 .. 100)
must_fail_clpfd_det(6,b(exists([b(identifier(x),integer,[]),b(identifier(y),integer,[])],b(conjunct(b(conjunct(b(member(b(identifier(x),integer,[nodeid(pos(nan,1,1,1,1,1))]),b(interval(b(integer(500),integer,[nodeid(pos(nan,1,1,3,1,3))]),b(integer(10000),integer,[nodeid(pos(nan,1,1,8,1,8))])),set(integer),[nodeid(pos(nan,1,1,3,1,8))])),pred,[nodeid(pos(nan,1,1,1,1,8))]),b(equal(b(div(b(integer(499),integer,[nodeid(pos(nan,1,1,16,1,16))]),b(identifier(y),integer,[nodeid(pos(nan,1,1,20,1,20))])),integer,[contains_wd_condition,nodeid(pos(nan,1,1,16,1,20))]),b(identifier(x),integer,[nodeid(pos(nan,1,1,22,1,22))])),pred,[contains_wd_condition,nodeid(pos(nan,1,1,16,1,22))])),pred,[contains_wd_condition,nodeid(pos(nan,1,1,1,1,22))]),b(member(b(identifier(y),integer,[nodeid(pos(nan,1,1,26,1,26))]),b(interval(b(integer(1),integer,[nodeid(pos(nan,1,1,28,1,28))]),b(integer(100),integer,[nodeid(pos(nan,1,1,31,1,31))])),set(integer),[nodeid(pos(nan,1,1,28,1,31))])),pred,[nodeid(pos(nan,1,1,26,1,31))])),pred,[contains_wd_condition,nodeid(pos(nan,1,1,1,1,31))])),pred,[used_ids([x,y])])).


