% (c) 2012-13 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(external_functions,[call_external_function/5, call_external_predicate/5,
                              call_external_substitution/6,
                              reset_external_functions/0,
                              external_function_argument_list/2,
                              check_if_predicates_were_evaluated/1]).

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(library(avl)).

% TO DO: add external Prolog files
% Should we do auto-conversion of B to Prolog types ,... ?

:- use_module(kernel_waitflags).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module provides a way to call external functions and predicates.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').


:- discontiguous expects_waitflag/1.
:- discontiguous expects_spaninfo/1.
:- discontiguous performs_io/1.
:- discontiguous does_not_modify_state/1.
:- discontiguous external_function_argument_list/2.

:- assert_must_succeed(( external_functions:call_external_function('COS',[int(0)],int(100),unknown,_WF) )).
:- assert_must_succeed(( external_functions:call_external_function('SIN',[int(0)],int(0),unknown,_WF) )).
% first a few hard-wired external functions for efficiency:
call_external_function('COS',[X],Value,_,_WF) :- !, 'COS'(X,Value). % on_exception ?
call_external_function('SIN',[X],Value,_,_WF) :- !, 'SIN'(X,Value).
call_external_function('TAN',[X],Value,_,_WF) :- !, 'TAN'(X,Value).
% now the generic case:
call_external_function(F,Args,Value,SpanInfo,WF) :-
   (expects_waitflag(F) -> E1=[WF] ; E1=[]),
   (expects_spaninfo(F) -> E2=[SpanInfo|E1] ; E2=E1),
   append(Args,[Value|E2],AV),
    % Check if there is better way to do this
   Call =.. [F|AV],
   %print('EXTERNAL FUNCTION CALL: '), print(F), print(' '), print(Call),nl,
   safe_call(Call,SpanInfo)
   %, print(result(Value)),nl.
   .

call_external_predicate(F,Args,PREDRES,SpanInfo,WF) :-
   (expects_waitflag(F) -> E1=[WF] ; E1=[]),
   (expects_spaninfo(F) -> E2=[SpanInfo|E1] ; E2=E1),
   append(Args,[PREDRES|E2],AV),
   Call =.. [F|AV],
   % print('EXTERNAL PREDICATE CALL: '), print(Call),nl, %%
   safe_call(Call,SpanInfo).

call_external_substitution(FunName,Args,InState,OutState,Info,WF) :-
 % TO DO: signal if substitution can be backtracked
   get_wait_flag_for_subst(FunName,WF,Flag),
   (does_not_modify_state(FunName) -> OutState=[] ; OutState = OutState2),
   call_external_substitution2(Flag,FunName,Args,InState,OutState2,Info,WF).

get_wait_flag_for_subst(FunName,WF,Flag) :- perform_directly(FunName),!,
   get_wait_flag1(WF,Flag). % ensure that precondition, ... have been evaluated
get_wait_flag_for_subst(FunName,WF,Flag) :- performs_io(FunName),!,
   get_enumeration_finished_wait_flag(WF,Flag). % ensure that precondition, ... have been evaluated
%   get_last_wait_flag(external(FunName),WF,Flag).
get_wait_flag_for_subst(FunName,WF,Flag) :- get_binary_choice_wait_flag(external(FunName),WF,Flag).

:- block call_external_substitution2(-,?,?,?,?,?,?).
call_external_substitution2(_,F,Args,InState,OutState,_Info,WF) :-
   (expects_waitflag(F) -> E1=[WF] ; E1=[]),
   (expects_spaninfo(F) -> E2=[SpanInfo|E1] ; E2=E1),
   append(Args,[InState,OutState|E2],AV),
   Call =.. [F|AV],
   % print('EXTERNAL SUBSTITUTION CALL: '), print(Call),nl, %%
   safe_call(Call,SpanInfo).
   
safe_call(X,SpanInfo) :- 
         on_exception(Exception, call(X),
                (add_exception_error(Exception,X,SpanInfo),fail)).

safe_is(Context,X,Y) :- safe_is(Context,X,Y,unknown).
safe_is(Context,X,Y,Span) :-   % print(safe_is(Context,X,Y)),nl,
         on_exception(Exception, X is Y,
                (add_exception_error(Exception,Context,Span),fail)). %, print(res(X)),nl.

add_exception_error(error(existence_error(procedure,external_functions:FUN),_),X,SpanInfo) :- !,
   add_error(external_functions,'### EXTERNAL function does not exist: ',FUN:X,SpanInfo).
add_exception_error(error(undefined,ERR),X,SpanInfo) :- 
   is_error(ERR),!,
   add_error(external_functions,'### Arithmetic operation not defined: ',X,SpanInfo).
add_exception_error(error(evaluation_error(undefined),ERR),X,SpanInfo) :- 
   is_error(ERR),!,
   add_error(external_functions,'### Arithmetic operation not defined: ',X,SpanInfo).
add_exception_error(error(evaluation_error(zero_divisor),ERR),X,SpanInfo) :- 
   is_error(ERR),!,
   add_error(external_functions,'### Division by 0 in arithmetic operation: ',X,SpanInfo).
add_exception_error(error(existence_error(source_sink,File),_),X,SpanInfo) :- !,
   add_error(external_functions,'### File does not exist (calling external function): ',File:X,SpanInfo).
add_exception_error(error(consistency_error(Msg,_,format_arguments),_),_X,SpanInfo) :- !,
   add_error(external_functions,'### Illegal format string (see SICStus Prolog reference manual for syntax): ',Msg,SpanInfo).
add_exception_error(Exception,X,SpanInfo) :-
   add_error(external_functions,'### Exception occured while calling external function: ',X:Exception,SpanInfo).

is_error(evaluation_error('is'(_,_),2,_,_)).
is_error(domain_error('is'(_,_),2,_,_)).

% STANDARD EXTERNAL FUNCTIONS


% -------------------------------

% MATHEMATICAL FUNCTIONS

:- public 'COS'/2.
:- block 'COS'(-,?). % redundant; we could add CLPFD constraints for R
'COS'(int(X),int(R)) :- cos2(100,X,R,unknown).
expects_spaninfo('COSx').
:- public 'COSx'/4.
:- block 'COSx'(-,?,?,?),'COSx'(?,-,?,?).
'COSx'(int(PrecisionMultiplier),int(X),int(R),Span) :- cos2(PrecisionMultiplier,X,R,Span).
:- block cos2(-,?,?,?),cos2(?,-,?,?).
cos2(Multiplier,X,R,Span) :- safe_is('COSx',R,round(Multiplier*cos(X/Multiplier)),Span).

:- public 'SIN'/2.
:- block 'SIN'(-,?). % redundant; we could add CLPFD constraints for R
'SIN'(int(X),int(R)) :- sin2(100,X,R,unknown).
expects_spaninfo('SINx').
:- public 'SINx'/4.
:- block 'SINx'(-,?,?,?),'SINx'(?,-,?,?).
'SINx'(int(PrecisionMultiplier),int(X),int(R),Span) :- %print('SINx'(PrecisionMultiplier,X,R)),nl,
  sin2(PrecisionMultiplier,X,R,Span).
:- block sin2(-,?,?,?),sin2(?,-,?,?).
sin2(Multiplier,X,R,Span) :- safe_is('SINx',R,round(Multiplier*sin(X/Multiplier)),Span).

:- public 'TAN'/2.
:- block 'TAN'(-,?). % redundant; we could add CLPFD constraints for R
'TAN'(int(X),int(R)) :- tan2(100,X,R,unknown).
expects_spaninfo('TANx').
:- public 'TANx'/4.
:- block 'TANx'(-,?,?,?),'TANx'(?,-,?,?).
'TANx'(int(PrecisionMultiplier),int(X),int(R),Span) :- tan2(PrecisionMultiplier,X,R,Span).
:- block tan2(?,-,?,?),tan2(-,?,?,?).
tan2(Multiplier,X,R,Span) :- safe_is('TANx',R,round(Multiplier*tan(X/Multiplier)),Span).


expects_spaninfo('PROLOG_FUN').
:- public 'PROLOG_FUN'/5.
'PROLOG_FUN'(string(Function),int(PrecisionMultiplier),int(Argument),int(Result),Span) :-
  prolog_fun2(Function,PrecisionMultiplier,Argument,Result,Span).
% Function can be:
% sin, cos, tan, cot, sinh, cosh, tanh, asin, acos, atan, acot, asinh, acosh, atanh, acoth, sqrt, log, exp
:- block prolog_fun2(?,-,?,?,?),prolog_fun2(-,?,?,?,?), prolog_fun2(?,?,-,?,?).
prolog_fun2(Function,Multiplier,X,R,Span) :-
   E =.. [Function,X/Multiplier],
   safe_is(Function,R,round(Multiplier*E),Span).

expects_spaninfo('LOGx').
:- public 'LOGx'/5.
:- block 'LOGx'(-,?,?,?,?),'LOGx'(?,-,?,?,?), 'LOGx'(?,?,-,?,?).
'LOGx'(int(PrecisionMultiplier),int(Base),int(X),int(R),Span) :- 
  log2(PrecisionMultiplier,Base,X,R,Span).
:- block log2(?,-,?,?,?),log2(-,?,?,?,?),log2(?,?,-,?,?).
log2(Multiplier,Base,X,R,Span) :- 
  safe_is('LOGx',R,round(Multiplier*log(Base/Multiplier,X/Multiplier)),Span).

:- public 'GCD'/3.
:- block 'GCD'(-,?,?),'GCD'(?,-,?).
'GCD'(int(X),int(Y),int(R)) :- gcd2(X,Y,R).
:- block gcd2(?,-,?),gcd2(-,?,?).
gcd2(X,Y,R) :- R is gcd(X,Y).

expects_spaninfo('MSB').
:- public 'MSB'/3.
:- block 'MSB'(-,?,?).
'MSB'(int(X),int(R),Span) :- msb2(X,R,Span). % most significant bit
:- block msb2(-,?,?).
msb2(X,R,Span) :- safe_is('MSB',R,msb(X),Span). % TO DO: could be made reversible by propagating constraint if R known

% -------------------------------

% THE CHOOSE operator (for TLA, recursive functions over sets,...)
% or choice operator from Abrial Book, page 60ff

:- use_module(custom_explicit_sets).
expects_waitflag('CHOOSE').
expects_spaninfo('CHOOSE').
:- public 'CHOOSE'/4.
:- block 'CHOOSE'(-,?,?,?).
'CHOOSE'([],_,Span,WF) :- 
   add_wd_error_span('CHOOSE applied to empty set: ',[],Span,WF).
'CHOOSE'(CS,H,_,_) :- is_interval_closure_or_integerset(CS,Low,Up),!, choose_interval(Low,Up,H).
'CHOOSE'([H|T],C,_,_) :- when(ground([H|T]),(sort([H|T],SL),SL=[C|_])).
'CHOOSE'(avl_set(A),H,_,_) :- avl_min(A,H).
'CHOOSE'(global_set(GS),H,_,_) :- H = fd(1,GS). % assume GS not empty, integer sets covered above
'CHOOSE'(closure(P,T,B),H,Span,WF) :- 
   custom_explicit_sets:expand_custom_set(closure(P,T,B),ER),
   'CHOOSE'(ER,H,Span,WF).

:- block choose_interval(-,?,?).
choose_interval(Low,_Up,H) :- number(Low),!,H=int(Low).
choose_interval(_Low,Up,H) :- choose_interval2(Up,H). % lower-bound is minus infinity
:- block choose_interval2(-,?).
choose_interval2(Up,H) :- number(Up),!,H=int(Up). % for those intervals we choose Up
choose_interval2(_,int(0)). % we have INTEGER or something equivalent; simply choose 0


% Prolog term order on encoding
:- public 'LESS'/2.
:- block 'LESS'(-,?), 'LESS'(?,-).
'LESS'(string(A),string(B)) :- !, less2(A,B).
'LESS'(int(A),int(B)) :- !, less2(A,B). % use CLPFD less ?
'LESS'(fd(A,G),fd(B,G)) :- !, less2(A,B). % use CLPFD less
'LESS'(pred_false,B) :- !, B=pred_true.
'LESS'(pred_true,_) :- !, fail.
% TO DO: expand finite closures, ...
'LESS'(X,Y) :- when((ground(X),ground(Y)), X @< Y).

:- block less2(-,?), less2(?,-).
less2(A,B) :- A @< B.

% -------------------------------

% a first prototype for implementing recursive functions
% still a bit clumsy to define recursive functions (using STRING parameter for names)
% a restriction is that we can only have one version per machine, i.e.
%  one can only define recursive function in the CONSTANTS section and only
%  in a deterministic fashion not depending on enumerated constants,...
% also: we still need to ensure that the recurisve closures are not expanded
%  this should probably go hand in hand with the symbolic pragma

reset_external_functions :-
    retractall(recursive_function(_,_)),
    retractall(evaluated_predicate(_)).

:- volatile recursive_function/2.
:- dynamic recursive_function/2.
 
expects_waitflag('REC').
expects_spaninfo('REC').
:- public 'REC'/5.
:- block 'REC'(?,-,?,?,?).
'REC'(string(FunName),ArgValue,Value,Span,WF) :- 
  % print('REC'(FunName,ArgValue)),nl,
   (recursive_function(FunName,Fun)
    -> bsets_clp:apply_to(Fun,ArgValue,Value,Span,WF)
    ;  add_error(external_functions,'### RECURSIVE function does not exist: ',FunName,Span),
       fail).

expects_spaninfo('RECx').
:- public 'RECx'/3.
:- block 'RECx'(-,?,?).
'RECx'(string(FunName),Result,Span) :- 
     % print('RECx'(FunName)),nl, %%
   (recursive_function(FunName,FunBody)
    -> kernel_objects:equal_object(Result,FunBody)
    ;  add_error(external_functions,'### RECURSIVE function does not exist: ',FunName,Span),
       fail).

:- use_module(custom_explicit_sets,[mark_closure_as_recursive/2]).
%expects_waitflag('REC_LET').
expects_spaninfo('REC_LET').
external_function_argument_list('REC_LET',[normal,symbolic]).
:- public 'REC_LET'/4.
'REC_LET'(string(FunName),FunBody,Result,Span) :-
   when(ground(FunBody),
       ( mark_closure_as_recursive(FunBody,SFunBody),  % what we actually need is that FunBody is evaluated already with a symbolic annotation ! 
       % TO DO: also marks as possibly being non-wd (infinite loop; avoid over-eager computation by b_interpreter_check)
         store_recursive_function(FunName,SFunBody,Span),
         kernel_objects:equal_object(Result,FunBody) % pass the argument through
         )).

store_recursive_function(FunName,FunBody,Span) :-
  print('STORING RECURSIVE FUNCTION : '), print(FunName),nl,  % print(FunBody),nl, %
  (recursive_function(FunName,OtherBody)
    -> (OtherBody=FunBody -> true
         ; add_error(external_functions,'### Multiple definitions of recursive function not yet allowed: ',FunName:FunBody:OtherBody,Span)
        )
    ;  assert(recursive_function(FunName,FunBody))
  ).

% -------------------------------

:- assert_must_succeed(( external_functions:call_external_function('STRING_APPEND',[string(a),string(b)],string(ab),unknown,_WF) )).
% STRING MANIPULATION FUNCTIONS
:- use_module(kernel_waitflags).
% function to append two strings, reversible
expects_waitflag('STRING_APPEND').
:- public 'STRING_APPEND'/4.
:- block 'STRING_APPEND'(-,?,-,?), 'STRING_APPEND'(?,-,-,?).
'STRING_APPEND'(string(A),string(B),string(C),WF) :- %print(app(A,B,C)),nl,
  app2(A,B,C,WF).
:- block app2(-,?,-,?), app2(?,-,-,?).
app2(A,B,C,_WF) :- %print(app2(A,B,C)),nl,
   nonvar(A),nonvar(B),!,atom_concat(A,B,CC), /* overcome bug in SICStus; delay unification */
   CC=C. %, print(app2(A,B,C)),nl.
app2(A,B,C,WF) :- atom_codes(C,CC), app3(A,B,CC,WF).
app3(A,B,CC,WF) :-
   (   nonvar(B) -> atom_codes(B,BB), append(AA,BB,CC), atom_codes(A,AA)
     ; nonvar(A) -> atom_codes(A,AA), append(AA,BB,CC), atom_codes(B,BB)
     ; length(CC,CLen), %print(enum(CC,CLen)),nl,
       get_wait_flag(CLen,'STRING_APPEND',WF,LWF),
       app4(A,B,CC,WF,LWF)
    ).
:- block app4(-,?,-,?,-), app4(?,-,-,?,-).
app4(A,B,CC,WF,_) :- (nonvar(A) ; nonvar(B)), !, % no need to enumerate
   app3(A,B,CC,WF).
app4(A,B,CC,_WF,_) :- %print(enumerating(CC)),nl,
   append(AA,BB,CC), % will be non-deterministic
   atom_codes(A,AA), atom_codes(B,BB).

:- public 'STRING_CHARS'/2.
:- block 'STRING_CHARS'(-,?).
% to do: make it reversible
'STRING_CHARS'(string(A),SeqRes) :- string_chars2(A,SeqRes).
:- block string_chars2(-,?).
string_chars2(A,SeqRes) :- atom_codes(A,AA), generate_char_seq(AA,1,CharSeq),
   equal_object_optimized(CharSeq,SeqRes).

generate_char_seq([],_,[]).
generate_char_seq([Code|T],Nr,[(int(Nr),string(CS))|TSeq]) :-
   atom_codes(CS,[Code]),
   N1 is Nr+1, generate_char_seq(T,N1,TSeq).

:- public 'STRING_CODES'/2.
:- block 'STRING_CODES'(-,-).
'STRING_CODES'(string(A),SeqRes) :- string_codes2(A,SeqRes).
:- block string_codes2(-,-).
string_codes2(A,SeqRes) :- nonvar(A),!,string_codes4(A,SeqRes).
string_codes2(A,SeqRes) :- SeqRes==[],!, empty_string_atom(A).
string_codes2(A,SeqRes) :- expand_custom_set_to_list(SeqRes,SeqList,_,string_codes2),
   when((nonvar(A);ground(SeqList)), string_codes3(A,SeqList)).
   
string_codes3(A,SeqRes) :- nonvar(A),!,string_codes4(A,SeqRes).
string_codes3(A,SeqRes) :- sort(SeqRes,SSeqRes),extract_codes(SSeqRes,1,Codes),
   atom_codes(A,Codes).
string_codes4(A,SeqRes) :- atom_codes(A,AA), generate_code_seq(AA,1,CodeSeq),
   equal_object_optimized(CodeSeq,SeqRes).

generate_code_seq([],_,[]).
generate_code_seq([Code|T],Nr,[(int(Nr),int(Code))|TSeq]) :-
   N1 is Nr+1, generate_code_seq(T,N1,TSeq).

extract_codes([],_,[]).
extract_codes([(int(Nr),int(Code))|T],N,[Code|CT]) :-
  (Nr==N -> true ; add_error(extract_codes,'Unexpected index: ',(Nr,N))),
  N1 is N+1, extract_codes(T,N1,CT).


empty_string(string('')).
empty_string_atom('').

:- assert_must_succeed(( external_functions:call_external_function('STRING_LENGTH',[string(a)],int(1),unknown,_WF) )).
% function to get length of a string
:- public 'STRING_LENGTH'/2.
'STRING_LENGTH'(SA,int(L)) :- greater_than_equal(int(L),int(0)),
   string_len2(SA,L).
:- block string_len2(-,-).
string_len2(SA,L)  :- L==0,!, empty_string(SA).
string_len2(string(A),L) :- string_len3(A,L).
:- block string_len3(-,-).
string_len3(A,L)  :- L==0,!, empty_string_atom(A).
% in case A is not known and L=1 we could enumerate chars ??
string_len3(A,L) :- string_len4(A,L).
:- block string_len4(-,?).
string_len4(A,L) :- atom_length(A,LL), LL=L. % delay unification due to bug in SICStus atom_length
% bug in SICStus: dif(X,1), atom_length(a,X) succeeds in 4.2.0 and 4.2.1

% function to split a string into a list of strings which were delimited by a separator
:- public 'STRING_SPLIT'/3.
:- block 'STRING_SPLIT'(-,?,?), 'STRING_SPLIT'(?,-,?).
'STRING_SPLIT'(string(A),string(B),R) :- string_split2(A,B,R).
:- block string_split2(-,?,?),string_split2(?,-,?).
string_split2(Atom,Separator,SplitAtomList) :-
  split_atom_string(Atom,Separator,List), % safe_call ?
  convert_prolog_to_b_list(List,SplitAtomList).
  % TO DO: make reversible to join strings; use expand_custom_set_to_list on third arg



:- assert_must_succeed((external_functions:split_atom_string('ef,g',',',R), R==[ef,g])).
:- assert_must_succeed((external_functions:split_atom_string('ab,cd,ef,g',',',R), R==['ab','cd',ef,g])).
:- assert_must_succeed((external_functions:split_atom_string('ab','a',R), R==['','b'])).
:- assert_must_succeed((external_functions:split_atom_string('"STR1"','',R), R==['"STR1"'])). 

split_atom_string(Atom,Sep,SplitList) :- 
  atom_chars(Sep,SepAscii),
  (SepAscii=[] -> SplitList = [Atom]
   ; atom_chars(Atom,ListAscii),split3(ListAscii,SepAscii,[],SplitList)).
:- use_module(library(lists)).
split3([],_,MatchSoFar,R) :- !,reverse(MatchSoFar,RM), atom_chars(Atom,RM),R=[Atom].
split3(List,Sep,MatchSoFar,Res) :-
   append(Sep,Tail,List),!, % we have a match with a separator
   reverse(MatchSoFar,RM), atom_chars(Atom,RM),
   Res=[Atom|R2], split3(Tail,Sep,[],R2).
split3([H|T],Sep,MatchSoFar,Res) :- % no match
   split3(T,Sep,[H|MatchSoFar],Res).
   

convert_prolog_to_b_list(PL,BL) :- convert_prolog_to_b_list_aux(PL,1,CPL),
  equal_object(CPL,BL).
  
convert_prolog_to_b_list_aux([],_,[]).
convert_prolog_to_b_list_aux([H|T],Index,[(int(Index),CH)|CT]) :-
  convert_prolog_to_b_term(H,CH),
  I1 is Index+1, convert_prolog_to_b_list_aux(T,I1,CT).

convert_prolog_to_b_term(N,R) :- number(N),!, R=int(N).
convert_prolog_to_b_term(A,R) :- atomic(A),!, R=string(A).
convert_prolog_to_b_term(A,R) :-
   add_internal_error('Illegal Prolog term: ',convert_prolog_to_b_term(A,R)), R=A.

% -------------------------------

% STRING FILE SYSTEM FUNCTIONS


:- use_module(kernel_objects).
:- use_module(library(file_systems)).
:- public 'FILES'/2.
:- block 'FILES'(-,?).
'FILES'(string(A),L) :- files2(A,L).
:- block files2(-,?).
files2(Dir,List) :- findall(string(F),file_member_of_directory(Dir,F,_FLong),L),
  equal_object_optimized(L,List).
:- public 'FULL_FILES'/2.
:- block 'FULL_FILES'(-,?).
'FULL_FILES'(string(A),L) :- full_files2(A,L).
:- block full_files2(-,?).
full_files2(Dir,List) :- findall(string(FLong),file_member_of_directory(Dir,_F,FLong),L),
  equal_object_optimized(L,List).


:- public 'DIRECTORIES'/2.
:- block 'DIRECTORIES'(-,?).
'DIRECTORIES'(string(A),L) :- dir2(A,L).
:- block dir2(-,?).
dir2(Dir,List) :- findall(string(F),directory_member_of_directory(Dir,F,_FLong),L),
  equal_object_optimized(L,List).
:- public 'FULL_DIRECTORIES'/2.
:- block 'FULL_DIRECTORIES'(-,?).
'FULL_DIRECTORIES'(string(A),L) :- full_dir2(A,L).
:- block full_dir2(-,?).
full_dir2(Dir,List) :- findall(string(FLong),directory_member_of_directory(Dir,_,FLong),L),
  equal_object_optimized(L,List).

:- public 'FILE_EXISTS'/2.
:- block 'FILE_EXISTS'(-,?).
'FILE_EXISTS'(string(A),PREDRES) :- file_exists2(A,PREDRES).
:- block file_exists2(-,?).
file_exists2(A,PREDRES) :- (file_exists(A) -> PREDRES=pred_true ; PREDRES=pred_false).
:- public 'DIRECTORY_EXISTS'/2.
:- block 'DIRECTORY_EXISTS'(-,?).
'DIRECTORY_EXISTS'(string(A),PREDRES) :- dir_exists2(A,PREDRES).
:- block dir_exists2(-,?).
dir_exists2(A,PREDRES) :- (directory_exists(A) -> PREDRES=pred_true ; PREDRES=pred_false).


:- public 'FILE_PROPERTY'/3.
:- block 'FILE_PROPERTY'(-,?,?), 'FILE_PROPERTY'(?,-,?).
'FILE_PROPERTY'(string(A),string(P),BoolRes) :-
   file_property2(A,P,BoolRes).
:- block file_property2(-,?,?), file_property2(?,-,?).
file_property2(File,Prop,BoolRes) :-
    file_property(File,Prop) -> BoolRes=pred_true ; BoolRes=pred_false.
% possible properties: readable, writable, executable

:- public 'FILE_PROPERTY_VALUE'/3.
:- block 'FILE_PROPERTY_VALUE'(-,?,?), 'FILE_PROPERTY_VALUE'(?,-,?).
'FILE_PROPERTY_VALUE'(string(A),string(P),int(R)) :-
   file_property_val2(A,P,R).
:- block file_property_val2(-,?,?), file_property_val2(?,-,?).
file_property_val2(File,Prop,Res) :-
    file_property(File,Prop,Val) -> Res=Val
         ; add_error(external_functions,'Illegal FILE_PROPERTY',Prop),Res=int(-1).
% possible properties:
%  size_in_bytes, create_timestamp, modify_timestamp, access_timestamp
% + on unix the follwing is also an integer: owner_user_id, owner_group_id

:- public 'DIRECTORY_PROPERTY'/3.
:- block 'DIRECTORY_PROPERTY'(-,?,?), 'DIRECTORY_PROPERTY'(?,-,?).
'DIRECTORY_PROPERTY'(string(A),string(P),BoolRes) :-
   dir_property2(A,P,BoolRes).
:- block dir_property2(-,?,?), dir_property2(?,-,?).
dir_property2(File,Prop,BoolRes) :-
    directory_property(File,Prop) -> BoolRes=pred_true ; BoolRes=pred_false.

:- public 'DIRECTORY_PROPERTY_VALUE'/3.
:- block 'DIRECTORY_PROPERTY_VALUE'(-,?,?), 'DIRECTORY_PROPERTY_VALUE'(?,-,?).
'DIRECTORY_PROPERTY_VALUE'(string(A),string(P),int(R)) :-
   dir_property_val2(A,P,R).
:- block dir_property_val2(-,?,?), dir_property_val2(?,-,?).
dir_property_val2(File,Prop,Res) :-
    directory_property(File,Prop,Val) -> Res=Val
           ; add_error(external_functions,'Illegal DIRECTORY_PROPERTY',Prop), Res=int(-1).
% --------------------------------

% I/O


% printf as predicate; can also be used as a function to BOOL
expects_spaninfo('fprintf').
:- public fprintf/5.
:- block fprintf(-,?,?,?,?), fprintf(?,-,?,?,?), fprintf(?,?,-,?,?).
fprintf(string(File),string(FormatString),Value,PredRes,Span) :- PredRes = pred_true,
   when((ground(FormatString),ground(Value)), 
   fprintf2(File,FormatString,Value,Span)).
expects_spaninfo('printf').

:- assert_must_succeed(external_functions:printf(string('test ok = ~w~n'),pred_true,pred_true,unknown)).
:- block printf(-,?,?,?), printf(?,-,?,?).
printf(string(FormatString),Value,PredRes,Span) :- PredRes = pred_true,
   when((ground(FormatString),ground(Value)), 
   fprintf2(user_output,FormatString,Value,Span)).
fprintf2(File,FormatString,Value,Span) :-
   translate:translate_bvalue(Value,TV),
   file_format(File,FormatString,[TV],Span).
% warning: format can call an argument as goal with ~@; this is dangerous
% and we should prohibit this
% Luckily PRINTF("@: <>~@<>~n","print(danger)") where print(danger)
% is arbitrary code is not a security exploit:
% our PRINTF translates the argument to an atom, i.e., one could
% only call Prolog predicates without arguments, anything else leads to an exception
% (even PRINTF("@: <>~@<>~n","nl") leads to an exception)

% vprintf: expression which returns argument as value 
expects_spaninfo('vprintf').
:- public vprintf/4.
vprintf(Format,Value,ValueRes,Span) :-
    printf(Format,Value,pred_true,Span),
    equal_object(Value,ValueRes). % we could generate a version which only instantiates when printing done
expects_spaninfo('fvprintf').
:- public fvprintf/5.
fvprintf(File,Format,Value,ValueRes,Span) :-
    fprintf(File,Format,Value,pred_true,Span),
    equal_object(Value,ValueRes). % we could generate a version which only instantiates when printing done

   
% PRINT a single value as substitution
performs_io('PRINT').
does_not_modify_state('PRINT').
:- public 'PRINT'/3.
:- block 'PRINT'(-,?,?), 'PRINT'(?,-,?).
'PRINT'(Val,_Env,OutEnv) :- %print('PRINT : '), print(Env),nl,
   print_value(Val),nl, OutEnv=[].

% PRINT a single value as substitution
perform_directly('PRINT_BT').
does_not_modify_state('PRINT_BT').
:- public 'PRINT_BT'/3.
:- block 'PRINT_BT'(-,?,?), 'PRINT_BT'(?,-,?).
'PRINT_BT'(Val,_Env,OutEnv) :- %print('PRINT : '), print(Env),nl,
   print_value(Val),nl, OutEnv=[].
'PRINT_BT'(Val,_Env,_OutEnv) :-
   print('BACKTRACKING: '),print_value(Val),nl, fail.

print_value(Val) :- translate:translate_bvalue(Val,TV), print(TV).

performs_io('DEBUG_PRINT_STATE').
does_not_modify_state('DEBUG_PRINT_STATE').
:- public 'DEBUG_PRINT_STATE'/3.
:- block 'DEBUG_PRINT_STATE'(-,?,?).
'DEBUG_PRINT_STATE'(Value,Env,OutEnv) :-
    print_value(Value),translate:print_bstate(Env),nl, OutEnv=[].

 % PRINTF as substitution
expects_spaninfo('PRINTF').
performs_io('PRINTF').
does_not_modify_state('PRINTF').
:- public 'PRINTF'/5.
:- block 'PRINTF'(-,?,?,?,?), 'PRINTF'(?,-,?,?,?), 'PRINTF'(?,?,-,?,?).
'PRINTF'(string(FormatString),Value,_Env,OutEnvModifications,Span) :-
  % print('PRINTF'(FormatString,Value,_Env)),nl,
   when((ground(FormatString),ground(Value)), 
         (fprintf2(user_output,FormatString,Value,Span),
          OutEnvModifications=[] /* substitution finished */
         )).
expects_spaninfo('FPRINTF').
performs_io('FPRINTF').
does_not_modify_state('FPRINTF').
:- public 'FPRINTF'/6.
:- block 'FPRINTF'(-,?,?,?,?,?), 'FPRINTF'(?,-,?,?,?,?), 'FPRINTF'(?,?,-,?,?,?).
'FPRINTF'(string(File),string(FormatString),Value,_Env,OutEnvModifications,Span) :-
   when((ground(FormatString),ground(Value)), 
         (fprintf2(File,FormatString,Value,Span),
          OutEnvModifications=[] /* substitution finished */
         )).

file_format(user_output,FormatString,Args,Span) :- !,
    safe_call(myformat(user_output,FormatString,Args),Span).
file_format(File,FormatString,Args,Span) :-
    open(File,append,Stream),
    safe_call(myformat(Stream,FormatString,Args),Span),
    close(Stream). % TO DO: call_cleanup

:- public myformat/3.
myformat(Stream,FormatString,Args) :-
  on_exception(error(consistency_error(_,_,format_arguments),_),
               format(Stream,FormatString,Args),
               % backup: the user probably has not provided ~w annotations:
               (format(Stream,FormatString,[]), format_args(Stream,Args))).
               
format_args(_Stream,[]).
format_args(Stream,[H|T]) :- format(Stream,'~w~n',H), format_args(Stream,T).

% --------------------------------------------

% RANDOM

:- assert_must_succeed(( external_functions:call_external_function('RANDOM',[int(0),int(1)],int(0),unknown,_WF) )).
:- use_module(library(random)).
% Generate a random number >= Low and < Up
% warning: this is not a mathematical function
:- public 'RANDOM'/3.
:- block 'RANDOM'(-,?,?), 'RANDOM'(?,-,?).
'RANDOM'(int(Low),int(Up),int(R)) :- random2(Low,Up,R).
:- block random2(-,?,?), random2(?,-,?).
random2(L,U,R) :- random(L, U, R).

:- use_module(library(system)).

:- assert_must_succeed(( external_functions:call_external_function('TIME',[string(year)],int(Y),unknown,_WF), number(Y), Y>2011 )).
:- assert_must_succeed(( external_functions:call_external_function('TIME',[string(month)],int(M),unknown,_WF), number(M), M>0, M<13 )).
:- public 'TIME'/2.
'TIME'(string(S),int(N)) :- time2(S,N).
:- block time2(-,?).
time2(now,N) :- !, now(N). % provide UNIX timesstamp
time2(year,N) :- !,datime(datime(N,_,_,_,_,_)).
time2(month,N) :- !,datime(datime(_,N,_,_,_,_)).
time2(day,N) :- !,datime(datime(_,_,N,_,_,_)).
time2(hour,N) :- !,datime(datime(_,_,_,N,_,_)).
time2(min,N) :- !,datime(datime(_,_,_,_,N,_)).
time2(sec,N) :- !,datime(datime(_,_,_,_,_,N)).
time2(OTHER,_) :- add_error(time2,'Illegal time field (must be now,year,month,day,hour,min,sec): ',OTHER),fail.

:- public 'TIMESTAMP_INFO'/3.
:- block 'TIMESTAMP_INFO'(-,-,-).
'TIMESTAMP_INFO'(string(S),int(Ticks),int(Res)) :- time_info2(S,Ticks,Res).
:- block time_info2(-,?,?), time_info2(?,-,?).
time_info2(year,Ticks,N) :- !,datime(Ticks,datime(N,_,_,_,_,_)).
time_info2(month,Ticks,N) :- !,datime(Ticks,datime(_,N,_,_,_,_)).
time_info2(day,Ticks,N) :- !,datime(Ticks,datime(_,_,N,_,_,_)).
time_info2(hour,Ticks,N) :- !,datime(Ticks,datime(_,_,_,N,_,_)).
time_info2(min,Ticks,N) :- !,datime(Ticks,datime(_,_,_,_,N,_)).
time_info2(sec,Ticks,N) :- !,datime(Ticks,datime(_,_,_,_,_,N)).
time_info2(OTHER,_,_) :- add_error(time_info2,'Illegal time field (must be now,year,month,day,hour,min,sec): ',OTHER),fail.

:- public 'TIMESTAMP'/7.
:- block 'TIMESTAMP'(-,-,-,-,-,-,-).
'TIMESTAMP'(int(Y),int(M),int(D),int(H),int(Min),int(S),int(Res)) :-
   timestamp2(Y,M,D,H,Min,S,Res).

:- block timestamp2(-,?,?,?,?,?,-), timestamp2(?,-,?,?,?,?,-), timestamp2(?,?,-,?,?,?,-),
         timestamp2(?,?,?,-,?,?,-), timestamp2(?,?,?,?,?,-,-).
timestamp2(Y,M,D,H,Min,S,ResTimeStamp) :- datime(ResTimeStamp, datime(Y,M,D,H,Min,S)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check if a given expression has already a ground value when
% the predicate is called

% the predicate is true when the value is ground, false otherwise
:- public 'IS_DETERMINED'/2.
'IS_DETERMINED'(Value,PredRes) :- ground(Value),!,PredRes=pred_true.
'IS_DETERMINED'(_Value,pred_false).

% other variant: the predicate is always true, the information if the variable is written
% to a file to check it later.
expects_spaninfo('IS_DETERMINED_INTO_FILE').
:- public 'IS_DETERMINED_INTO_FILE'/5.
'IS_DETERMINED_INTO_FILE'(FileValue,VariableValue,Value,PredRes,SpanInfo) :-
    check_value_is_string(FileValue,'IS_DETERMINED_INTO_FILE','file name',SpanInfo,File),
    check_value_is_string(VariableValue,'IS_DETERMINED_INTO_FILE','variable name',SpanInfo,Varname),
    ( ground(File) -> 
        open(File,write,S),
        call_cleanup( check_determined2(Value,S,Varname), close(S))
    ; otherwise ->
        add_error(external_functions,'### EXTERNAL Unspecified file in check_determined','',SpanInfo)),
    PredRes = pred_true.
check_determined2(Value,S,Varname) :-
    check_determined3(Value,Varname,Term),
    writeq(S,Term),write(S,'.\n').
check_determined3(Value,Name,Term) :- ground(Value),!,Term=determined_value(Name,Value).
check_determined3(_Value,Name,not_determined_value(Name)).

check_value_is_string(string(S),_PredName,_Location,_SpanInfo,String) :-
    !, String=S.
check_value_is_string(Value,PredName,Location,SpanInfo,_String) :-
    atom_concat('### EXTERNAL ',PredName,Msg1),
    atom_concat(Msg1,': Type error, expected string as ',Msg2),
    atom_concat(Msg2,Location,Msg),
    add_error(external_functions,Msg,Value,SpanInfo),
    fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% store the information that a predicate has actually been used
% We can check this later with check_if_predicates_were_evaluated/1
% below. The CLI offers a way to use the predicate.

:- use_module(tools, [assert_once/1]).
:- dynamic evaluated_predicate/1.
expects_spaninfo('IS_EVALUATED').
:- public 'IS_EVALUATED'/3.
'IS_EVALUATED'(Value,PredRes,SpanInfo) :-
    check_value_is_string(Value,'IS_EVALUATED',name,SpanInfo,Name),
    assert_once( evaluated_predicate(Name) ),
    PredRes = pred_true.

check_if_predicates_were_evaluated(List) :-
    maplist(check_if_predicate_was_evaluated,List).
check_if_predicate_was_evaluated(evaluated(P)) :-
    format('Checking if "~w" has been evaluated.~n',[P]),
    ( evaluated_predicate(P) -> true
    ; otherwise -> add_error(external_functions,'Predicate has not been evaluated', P)).
check_if_predicate_was_evaluated(not_evaluated(P)) :-
    format('Checking if "~w" has not been evaluated.~n',[P]),
    ( evaluated_predicate(P) -> add_error(external_functions,'Predicate has been evaluated', P)
    ; otherwise -> true).
