% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 
/* graph_canon.pl - graph isomorphism module 1/11/2005 */

:- module(graph_canon,   [
			print_cstate_graph/1, print_cstate_graph/2,
			initialise_nauty/0, clear_nauty/0,
			add_state_graph_to_nauty/2,
			reset_typing_info_for_symm/0]).


:- use_module(library(system)).
:- use_module(library(lists)).
:- use_module(library(ordsets)).
%:- use_module(library(gauge)).
:- use_module(library(clpfd)).
:- use_module(library(trees)).
:- use_module(library(assoc)).

:- use_module(error_manager).

%:- compile(sp4_compatibility).
:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,symmetry).
:- module_info(description,'Compute canonical forms of state graphs and store them using nauty C interface.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

%-------------- code profiling -------------------------------------------------
%setup_profiling :- prolog_flag(compiling,_,profiledcode),compile(graph_canon).
%profilegc :- view([graph_canon:_]).	
%update :- prolog_flag(compiling,_,profiledcode),compile(graph_canon).
%-------------- utilities etc --------------------------------------------------

	
% assert this predicate such that there's only one instance of it
assert_once(Functor,Args) :-
	append([Functor],Args,F_),
	Pred =.. F_,
	(call(Pred) ->
		true
		; assert(Pred)).
		
% swap an assertion for another
swap_assertion(Functor,Test,Args) :-
	append([Functor],Test,T),
	PredT =.. T,
	(call(PredT) ->
		retractall(PredT),
		assert_once(Functor,Args)
		; true).



:- dynamic non_symmetric_elements/1, % store non_symmetric_elements as a list
	   /* a set used in this session of ProB, with the following format
	      (Name,AllElements,SymmPartition,NonSymPartition),
	       where the non syms are the enumated elements used in this machine
	       , of type Name */
	   uses_constants/1, 	% yes or no, depending on whether this machine uses constants
	   constants/1.  	% the list of constants used by this machine, in the order
	   			% they are defined in the PROERTIES clause. If none are used, the empty list is given.
	  
uses_constants(no).

setup_uses_constants :-
	(user:machine(Mach),member(El1,Mach),member(El2,Mach),functor(El1,'ConcreteConstants',_),functor(El2,'Properties',_) ->
		swap_assertion(uses_constants,[_],[yes])
		; swap_assertion(uses_constants,[_],[no])).


	
% asserts the constants as a list, in the order they appear in 
%update_constants :-
%	retractall(constants(_)),reduce_graph_state_space:current_constants(CCV),
%	bmachine:b_get_properties_from_machine(Properties),
%	reduce_graph_state_space:extract_constants(Properties,CCN),
%	findall(bind(Name,Value),(member(Value,CCV),member(Name,CCN)),Res),
%	assert(constants(Res)).


/* given the properties clause of the bmachine, extract the names of the 
known constants, in the order of their declaration */
% currently works only for machines that do not use inheritance

% removed

% ----------------- setting up typing info for use in symm ---------------------
:- dynamic machine_free_variables/3.
machine_free_variables([],[],[]).

% update typing info for use in sym reduction -- do when machine loaded
reset_typing_info_for_symm :-
	reset_machine_free_variables,
	debug:debug_println(9,skipping(assert_typing_info)).

reset_machine_free_variables :-
	retractall(machine_free_variables(_,_,_)),
	bmachine:b_get_machine_constants(C),
	bmachine:b_get_machine_variables(V),
	append(C,V,B),
	assert(machine_free_variables(C,V,B)).

% ------------------------------------------------------------------------------
	
/* Assert all the elements of the sets used by this machine, for this session of ProB,
   and also those that are symmetric or not */
% assert_global_set_info :- better use b_global_sets now



/***************** CANONICAL LABELS FOR GRAPHS OF MULTIPLE RELATIONS***********/

	
/**********************************************/

/***************SEARCHING FOR CANONICAL*********/

		    
/**********************************************/

/*************UTILISING CANONICAL SEARCH*******/
/* required the changing of search4: such that the partition (the first found) which
   has the lowest num is retained and returned to the user. */

/**********************************************/
/******************************************************************************/

/******** STATE AS SINGLE GRAPH ***************/
:- use_module(specfile,[expand_const_and_vars/3]).
state_graph(root,R) :- !,R=[].
state_graph(concrete_constants(C),SG) :- !, state_graph(C,SG).
state_graph(const_and_vars(ConstID,S),SG) :- !,
    expand_const_and_vars(ConstID,S,Full), 
	reset_n,
	state_graph(Full,[],SG).
state_graph(S,SG) :-
	reset_n,
	state_graph(S,[],SG).
state_graph([],Final,Final).
state_graph([bind(VarName,SValue)|T],C,F) :-
    (custom_explicit_sets:is_custom_explicit_set(SValue,state_graph)
      -> custom_explicit_sets:try_expand_custom_set(SValue,Value)
      ;  Value=SValue
    ),
    (top_level_translate_node(Value,VarName,PG) -> true
      %print(translate(Value,VarName,PG)),nl
      ; add_error(state_graph,'Call failed: ',top_level_translate_node(Value,VarName,PG))),
	append(PG,C,C2),
	state_graph(T,C2,F).

top_level_translate_node(DataValue,VariableName,NewTransitions) :- 
% Input: DataValue
% Output: NewTransitions required to represent the structure
% In contrast to translate_inner_node, we do not generate new inner artificial nodes;
%    VariableName is used as an arc-label to encode the top-level information
   expand_data_value(DataValue,S1), 
	% check whether S1 is a relation/atom/set
   (relation(S1) -> grelation(S1,sg_root,VariableName,[],NewTransitions)
   ; (emptyset(S1) -> NewTransitions = [(sg_root,VariableName,sg_root)]
   %% [('$Empty_Set',VariableName,sg_root)]  has problems as we cannot distinguish between v={} and v={{}}
   %% [] has problems when we have set_up_constants and variables which are all initialised with
   %%  the empty set: then we get confused between set_up_constants & initialise_machine
   ; (set(S1) -> gset(S1,sg_root,VariableName,[],NewTransitions)
   ; (S1= (P1,P2)
	    ->   % a pair
	         translate_inner_node(P1,Node1,NewTrans1),
	         translate_inner_node(P2,Node2,NewTrans2),
	         append([(Node1,VariableName,Node2)|NewTrans1],NewTrans2,NewTransitions)
	    ;   % an atom
	        NewTransitions = [(S1,VariableName,sg_root)]
	   )
	 )
	 )
  ).


expand_data_value(DataValue,S1) :-
    (custom_explicit_sets:is_custom_explicit_set(DataValue,expand_data_value)
      -> custom_explicit_sets:try_expand_custom_set(DataValue,S1)
      ;  DataValue=S1
    ).
% TO DO: convert to Difference-List to get rid of all the appends below


gset([],_,_,F,F).
gset([S1|T],Parent,TransitionLabel,F,Fi) :-
	translate_inner_node(S1,C1,F2),
	append([(C1,TransitionLabel,Parent)|F2],F,F4),
	gset(T,Parent,TransitionLabel,F4,Fi).

grelation([],_,_,F,F).
grelation([(E1,E2)|T],Parent,VariableName,F,Fi) :-
    translate_inner_node(E1,C1,F2),
    translate_inner_node(E2,C2,F3),
	append([(C1,VariableName,C2)|F2],F,F5),
	append(F3,F5,F6),
	grelation(T,Parent,VariableName,F6,Fi).
	
		
translate_inner_node(DataValue,NodeinGraph,NewTransitions) :-  %% print(trans(DataValue)),nl,
% Input: DataValue
% Output: NodeinGraph representing DataValue + NewTransitions required to represent the structure
   expand_data_value(DataValue,S1), 
	% check whether S1 is a atom/set
   (emptyset(S1) -> NodeinGraph = '$Empty_Set', NewTransitions=[]
    ; (cn_value(S1,NodeinGraph) -> NewTransitions=[]
    ; (set(S1) ->
			new_sg_node(S1,NodeinGraph),
			gset(S1,NodeinGraph,'$to_el',[],NewTransitions)  % use the $to_el label to encode membership, using ':' would be more readable; but less efficient for symmetry (extra layers required) 
	; (S1= (P1,P2)
	    -> ( % a pair
	         translate_inner_node(P1,Node1,NewTrans1),
	         translate_inner_node(P2,Node2,NewTrans2),
	         new_sg_node(S1,NodeinGraph),
	         append([(NodeinGraph,'$from',Node1),
	                 (NodeinGraph,'$to_el',Node2)|NewTrans1],NewTrans2,NewTransitions)
	       )
	    ;  (% an atom
	        NewTransitions = [],
	        NodeinGraph = S1
	        )
	   ))
	 )
  ).

 
relation([(_,_)|_]).
emptyset([]).
set([]).
set([_|_]).

:- dynamic cn/1, cn_value/2.

reset_n :-
	retractall(cn(_)),retractall(cn_value(_,_)),asserta(cn(0)).
	
/* generate a new state graph node */
new_sg_node(Value,abs_cn(J,NestingCol)) :- retract(cn(I)), J is I+1,
	asserta(cn(J)), 
	get_abstract_node_type_colour(Value,NestingCol),
	assert(cn_value(Value,abs_cn(J,NestingCol))).

% for the moment: just compute the "depth/nesting" of the relation; we should do something more intelligent
get_abstract_node_type_colour([],C) :- !,C=1.
get_abstract_node_type_colour([H|_],C) :- !, get_abstract_node_type_colour(H,CH),
  C is CH+1.
get_abstract_node_type_colour((A,_),C) :- !, get_abstract_node_type_colour(A,CH),
  C is CH+1.
get_abstract_node_type_colour(_,0).

	
:- use_module(graph_iso_nauty).

print_cstate_graph(File) :-
	state_space:current_expression(_,State),
	print_cstate_graph(State,File).
print_cstate_graph(State,File) :-
	state_graph(State,GCE),
	%print(state_graph(GCE)),nl,
	(graph2dot(GCE,File) -> true
	  ; add_error(graph_canon,'Dot conversion failed: ',graph2dot(GCE,File))).
	%,print_nauty_graph(GCE).
	
/* ---------------------------- */

initialise_nauty :- nauty_init.
clear_nauty :- nauty_clear.

add_state_graph_to_nauty(StateExpression,Exists) :-
	(state_graph(StateExpression,SG)
	  -> add_nauty_graph(SG,Exists)
	  ; add_error(graph_canon,'Could not compute state_graph: ',StateExpression),fail).
    
/* ---------------------------- */
   
	
graph2dot(G,File) :- reset_dot,
	tell(File),
	print('digraph state {\n graph [fontsize=12]\nrankdir=TB;\n'),
	%print('digraph state {\n graph [page=\"8.5, 11\",ratio=fill,size=\"7.5,10\"];\n'),
	edges2dot(G),
	nodes2dot,
	findall(cluster(Ty,El),cluster(Ty,El),T),
	clusters2dot(T),
	print('}'),
	told.
:- use_module(b_global_sets).
cluster(Type,Elements) :- b_global_set(Type),  
  % compute all elements of global sets which are used in current state graph
  findall(Translation, (enum_global_type(El,Type), node_col(El,Translation,_,_)),Elements).
% maybe add boolean  ??
edges2dot([]).
edges2dot([(A,B,C)|T]) :-  /* edge from A -> C with label B */
	translate_node_value(A,AID),
	(B = abs_cn(BB,ColBB)
	  -> translate_abs_cn_node(BB,ColBB,EdgeLabel)
	   ; EdgeLabel = B),
	translate_node_value(C,CID),
	get_col_assoc(edge,EdgeLabel,EdgCol),
	format('\"~w\" -> \"~w\" [label = \"~w\", color = \"~w\"];\n',
	       [AID,CID,EdgeLabel,EdgCol]),
	edges2dot(T).

reset_dot :- retractall(col_association(_,_,_)),
  retractall(node_col(_,_,_,_)),
  retractall(cur_col(_,_)),
  assert(cur_col(node,0)), assert(cur_col(edge,99)).
  
translate_node_value(abs_cn(AA,ColAA), Translation) :- !,
        translate_abs_cn_node(AA,ColAA,Translation).
translate_node_value(sg_root,Translation) :- !,
        Translation = 'ROOT-NODE', add_node(sg_root,Translation).
translate_node_value(Val,Translation) :-
	    translate_bvalue_for_dot(Val,Translation),
	    add_node(Val,Translation).
	    
translate_bvalue_for_dot(string(S),Translation) :- !, 
   % normal quotes confuse dot
   %ajoin(['''''',S,''''''],Translation).
   ajoin(['\\"',S,'\\"'],Translation).
translate_bvalue_for_dot(Val,Translation) :- translate:translate_bvalue(Val,Translation).

nodes2dot :- node_col(_,Translation,Col,Shape),
	format('\"~w\" [color = \"~w\", style = \"filled, solid\", shape = \"~w\"]\n',[Translation,Col,Shape]),
	fail.
nodes2dot.

:- dynamic node_col/4.
add_node(X,_) :- node_col(X,_,_,_),!. % already processed
add_node(sg_root,Translation) :- 
    assert(node_col(sg_root,Translation,lightblue,diamond)).
add_node(fd(X,Type),Translation) :- !,
     get_col_assoc(node,Type,Col),
     assert(node_col(fd(X,Type),Translation,Col,box)).
add_node(int(X),Translation) :- !,
     Col = 'wheat3', %get_col_assoc(node,integer,Col),
     assert(node_col(int(X),Translation,Col,box)).
add_node(string(X),Translation) :- !,
     Col = 'khaki1', %'lavender' 'burlywood', %get_col_assoc(node,integer,Col),
     assert(node_col(string(X),Translation,Col,box)).
add_node(pred_true,T) :- !, assert(node_col(pred_true,T,'OliveDrab4',ellipse)).
add_node(pred_false,T) :- !, assert(node_col(pred_false,T,brown,ellipse)).
add_node(_Val,_Translation).

:- dynamic col_association/3. % store colours used for types, edge labels,...
get_col_assoc(Domain,T,Col) :- col_association(Domain,T,Col),!.
get_col_assoc(Domain,T,Col) :- get_next_col(Domain,Col),
  assert(col_association(Domain,T,Col)).
   
get_next_col(Type,Col) :- retract(cur_col(Type,N)),!, N1 is N+1,
  (default_colours(N1,Col) -> NX=N1 
   ; NX = 1, default_colours(1,Col)),
  assert(cur_col(Type,NX)).
get_next_col(_T,red).

:- dynamic cur_col/2.
cur_col(node,0).
cur_col(edge,99).
default_colours(1,'#efdf84').
default_colours(2,'#bdef6b').
default_colours(3,'#5863ee').
default_colours(4,'LightSteelBlue1').
default_colours(5,gray).

default_colours(100,firebrick).
default_colours(101,sienna).
default_colours(102,'SlateBlue4').
default_colours(103,black).

translate_abs_cn_node(AA,ColAA,Translation) :-
        tools:string_concatenate('_',AA,AID1),
		tools:string_concatenate(ColAA,AID1,AID2),
		tools:string_concatenate('abs',AID2,Translation).
   
print_sg_root :- 
  translate:translate_bvalue(sg_root,SGR),
              print('\"'),print(SGR),print('\" [shape=diamond,color=lightblue,style=filled];\n').
clusters2dot([]).
clusters2dot([cluster(Name,Elements)|T]) :-
	format('subgraph \"cluster_~w\" {node [style=filled,color=white]; label=\"~w\"; style=filled;color=lightgrey; ',[Name,Name]),
	nodes2dot(Elements),
	clusters2dot(T).
	
nodes2dot([]) :- print('}'),nl.
nodes2dot([H|T]) :-
	translate:translate_params_for_dot([H],HP),
	%format('~w; ',[HP]),
	tools:print_escaped(HP), print('; '),
	nodes2dot(T).


/**********************************************/

/****** CANONICAL LABELLING USING LEXICOGRAPHIC AND AUTOMORHISM PRUNING: *******
 ****** Makes use of Schreier-Sims data structure defined in Kreher's    *******
 ****** C.A.G.E.S book, and uses the techniques for pruning described in *******
 ****** the chapter 7.                                                   *******/

/* IF USE IN FUTURE; UPDATE TO ALLOW SET OF VERTICES TO BE PASSED TO IT AS WITH
   SEARCH4/7 */


/*******************************************************************
*/
% schreiersims.pl implementation of the schreier sims representation of an
% automorphism group taken from Kreher's Combinatorial Algorithms book. Provides
% method for membership test of a permutation in the group that should run in O(n^2)
% time. Plans are not to spend time on implementing a method for listing all permutations
% represented in the group -- since this is not required at the time of writing.

% THIS IS NOT CURRENTLY USED, BUT WAS ONCE USED TO MORE CLOSELY FOLLOW NAUTY.
% HOWEVER, USING SCHREIERSIMS STRUCTURES IMPLEMENTATION IS VERY LIST INTENSIVE
% RESULTING IN A POOR PERFORMANCE -- HENCE NOT USED NOW. FUTURE WORK MAY WANT
% TO LOOK AT ITS USE BELOW..


/*** END OF CANONICAL LABELLING USING LEXICOGRAPHIC AND AUTOMORPHISM PRUNING ***/