% (c) 2009-2011 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html

% This module contains arithmetic that also supports the symbol inf
% The operator infis/2 can be used analogously to the standard Prolog operator is/2.
% Currently supported arithmetic operations are +, -, *, min, max

:- module(inf_arith, [infis/2,infless/2, 
                      infgreaterequal/2, infgreater/2, infdec/2, infinc/2, infdec_if_positive/2, infmin/3, infmax/3,
                      block_inf_greater/2, block_inf_greater_equal/2
                     ]).

:- use_module(module_information).
:- module_info(group,kernel).
:- module_info(description,'This module contains arithmetic that also supports the symbol inf.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(self_check).

:- op( 700, xfx, infis).

Res infis Expr :- infis2(Expr,R), R=Res.


:- assert_must_succeed(( infis(R,22+inf), R==inf )).
:- assert_must_succeed(( infis(R,22+21), R==43 )).
:- assert_must_succeed(( infis(R,-(22+21)*(-1)), R==43 )).
:- assert_must_succeed(( infis(R,(-inf)+21), R==(-inf) )).
:- assert_must_succeed(( infis(R,22+(-inf)), R==(-inf) )).
:- assert_must_succeed(( infis(R,22-(-inf)), R==inf )).
:- assert_must_succeed(( infis(R,22-21), R==1 )).
:- assert_must_succeed(( infis(R,(-inf)-21), R==(-inf) )).
:- assert_must_succeed(( infis(R,inf-21), R==inf )).
:- assert_must_succeed(( infis(R,22*inf), R==inf )).
:- assert_must_succeed(( infis(R,inf//22), R==inf )).

infis2(A,   A) :- number(A),!.
infis2(inf, inf).
infis2(-A,  Res) :- infis2(A,Ac), infnegate(Ac,Res).
infis2(A+B, Res) :- infis2(A,Ac), infis2(B,Bc), infadd(Ac,Bc,Res).
infis2(A-B, Res) :- infis2(A,Ac), infis2(-B,Bc), infadd(Ac,Bc,Res).
infis2(A*B, Res) :- infis2(A,Ac), infis2(B,Bc), infmult(Ac,Bc,Res).
infis2(A//B, Res) :- infis2(A,Ac), infis2(B,Bc), infdiv(Ac,Bc,Res).
infis2(min(A,B), Res) :-
    infis2(A,Ac), infis2(B,Bc), infmin(Ac,Bc,Res).
infis2(max(A,B), Res) :-
    infis2(A,Ac), infis2(B,Bc), infmax(Ac,Bc,Res).

:- assert_must_succeed(( infisnegative(-2) )).
:- assert_must_succeed(( infisnegative(-inf) )).
:- assert_must_fail(( infisnegative(inf) )).
:- assert_must_fail(( infisnegative(0) )).

infisnegative(-inf) :- !.
infisnegative(inf) :- !,fail.
infisnegative(X) :- X<0.

infnegate(inf,-inf) :- !.
infnegate(-inf,inf) :- !.
infnegate(A,B) :- B is -A.

:- assert_must_succeed(( infadd(22,inf,R), R==inf )).
:- assert_must_succeed(( infadd(22,21,R), R==43 )).
:- assert_must_succeed(( infadd((-inf),21,R), R==(-inf) )).
infadd(A,B,R) :-
    ( infisnegative(A) ->
        infnegate(A,Apos),
        ( infisnegative(B) ->
            infnegate(B,Bpos),
            infadd(Apos,Bpos,Rneg),
            infnegate(Rneg,R)
        ; otherwise ->
            infsub2(B,Apos,R))
    ; otherwise ->
        ( infisnegative(B) ->
            infnegate(B,Bpos),            
            infsub2(A,Bpos,R)
        ; otherwise ->
            infadd2(A,B,R))).
infadd2(inf,_,inf) :- !.
infadd2(_,inf,inf) :- !.
infadd2(A,B,R) :- R is A+B.

% don't know what is inf - inf
infsub2(inf,B,inf) :- !, B \== inf.
infsub2(_,inf,-inf) :- !.
infsub2(A,B,R) :- R is A-B.

infmult(A,B,R) :-
    ( infisnegative(A) ->
        infnegate(A,Aneg),
        infmult(Aneg,B,Rneg),
        infnegate(Rneg,R)
    ; infisnegative(B) ->
        infnegate(B,Bneg),
        infmult(A,Bneg,Rneg),
        infnegate(Rneg,R)
    ; otherwise ->
        infmult2(A,B,R)).
infmult2(inf,_,inf) :- !.
infmult2(_,inf,inf) :- !.
infmult2(A,B,R) :- R is A*B.

infdiv(A,B,R) :-
    ( infisnegative(A) ->
        infnegate(A,Aneg),
        infdiv(Aneg,B,Rneg),
        infnegate(Rneg,R)
    ; infisnegative(B) ->
        infnegate(B,Bneg),
        infdiv(A,Bneg,Rneg),
        infnegate(Rneg,R)
    ; otherwise ->
        infdiv2(A,B,R)).
infdiv2(inf,B,inf) :- !,B \== inf.
infdiv2(_,inf,0) :- !.
infdiv2(A,B,R) :- R is A//B.

:- assert_must_succeed( infgreaterequal(inf,inf)).
:- assert_must_succeed( infgreaterequal(inf,100)).
:- assert_must_succeed( infgreaterequal(2000,100)).
:- assert_must_fail( infgreaterequal(2000,inf)).
:- assert_must_fail( infgreaterequal(1999,2000)).
infgreaterequal(A,B) :- (A=B -> true ; infgreater(A,B)).

infgreater(A,B) :- infless(B,A).
infless(inf,_) :- !,fail.
infless(_,-inf) :- !,fail.
infless(-inf,_X) :- !.
infless(_X,inf) :- !.
infless(A,B) :- A<B.

:- assert_must_succeed(( infmax(22,inf,R), R==inf )).
:- assert_must_succeed(( infmax(22,21,R), R==22 )).
:- assert_must_succeed(( infmax((-inf),21,R), R==21 )).
:- assert_must_succeed(( infmax((-inf),inf,R), R==inf )).

infmax(A,B,R) :-
    ( infgreater(A,B) -> A=R
    ; otherwise -> B=R).
    

:- assert_must_succeed(( infmin(22,inf,R), R==22 )).
:- assert_must_succeed(( infmin(22,21,R), R==21 )).
:- assert_must_succeed(( infmin((-inf),21,R), R==(-inf) )).
:- assert_must_succeed(( infmin((-inf),inf,R), R==(-inf) )).
infmin(A,B,R) :-
    ( infless(A,B) -> A=R
    ; otherwise -> B=R).

:- assert_must_succeed(( infdec(inf,R), R==inf )).
:- assert_must_succeed(( infdec(3,R), R==2 )).
:- assert_must_succeed(( infdec(-inf,R), R==(-inf) )).
infdec(inf,R) :- !, R = inf.
infdec(-inf,R) :- !, R = -inf.
infdec(N,N1) :- N1 is N-1.

:- assert_must_succeed(( infinc(inf,R), R==inf )).
:- assert_must_succeed(( infinc(3,R), R==4 )).
:- assert_must_succeed(( infinc(-inf,R), R==(-inf) )).
infinc(inf,R) :- !, R = inf.
infinc(-inf,R) :- !, R = -inf.
infinc(N,N1) :- N1 is N+1.

:- assert_must_succeed(( infdec_if_positive(inf,R), R==inf )).
:- assert_must_succeed(( infdec_if_positive(3,R), R==2 )).
infdec_if_positive(inf,R) :- !, R = inf.
infdec_if_positive(-inf,_) :- !, fail.
infdec_if_positive(N,N1) :- N>0, N1 is N-1.

% ------------------------ blocking versions --------------------


:- block block_inf_greater_equal(-,?), block_inf_greater_equal(?,-).
block_inf_greater_equal(X,Y) :- infgreaterequal(X,Y).


:- block block_inf_greater(-,?), block_inf_greater(?,-).
block_inf_greater(X,Y) :- infgreater(X,Y).
