% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(kernel_equality,[equality_objects_lwf/4, equality_objects/3, 
           equality_objects_with_type/4,
           empty_set_test/2, empty_cartesian_product/3, 
           or_eq_obj/3,
           get_cardinality_wait_flag/4, get_cardinality_wait_flag/5,
           get_cardinality_powset_wait_flag/3,
           get_cardinality_relation_over_wait_flag/6,
           check_and_remove/4,
           membership_test/3, cartesian_pair_test/4,
           in_nat_range_test/4,
           
           dif/3, integer_dif/2, bool_dif/2, eq_atomic/4,
           
           subset_test/4, subset_strict_test/4]).

:- use_module(kernel_objects,[equal_object/2, not_equal_object/2, cardinality_as_int_for_wf/2,
                              safe_mul/3, safe_add/3, safe_pow2/2,
                              empty_set/1, not_empty_set/1,
                              exhaustive_kernel_check/2, exhaustive_kernel_check/1,
                              exhaustive_kernel_fail_check/1]).

:- use_module(self_check).

:- use_module(debug).
:- use_module(tools).   

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides reified equality and membership predicates.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(typechecker).
:- use_module(error_manager).

% kernel_equality
% :- ensure_loaded(kernel_equality).

:- use_module(library(avl)).
:- use_module(custom_explicit_sets).
:- use_module(kernel_waitflags).

:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([int(2),int(1)],[int(2)],pred_false))). 
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([int(2),int(1)],[],pred_false))). 
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([int(2),int(1)],[int(1),int(2)],pred_true))). 
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([[int(2)],[int(1)]],[[int(1)],[int(2)]],pred_true))). 
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([string('"STR1"')],[string('"STR2"')],pred_false))). 
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([string('"STR2"'),string('"STR1"')],
                       [string('"STR1"'),string('"STR2"')],pred_true))). 
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([(int(1),int(2)),(int(1),int(3))],[(int(1),int(3)),(int(1),int(2))],pred_true))). 
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([(int(2),int(2)),(int(1),int(3))],[(int(2),int(3)),(int(1),int(2))],pred_false))). 
:- assert_must_succeed((equality_objects(A,B,R), A=[int(2),int(3)],B=[int(X),int(Y)],
                     X=4,Y=2, R==pred_false)).
:- assert_must_succeed((equality_objects(A,B,R), A=[int(2),int(3)],B=[int(X),int(Y)],
                     X=2,Y=4, R==pred_false)).
:- assert_must_succeed((equality_objects(A,B,R), A=[int(2),int(3)],B=[int(X),int(Y)],
                     X=3,Y=2, R==pred_true)).
:- assert_must_succeed((equality_objects(A,B,R), A=[int(2),int(3)],B=[int(X),int(Y)],
                     X=2,Y=3, R==pred_true)).
:- assert_must_succeed((equality_objects(A,B,R), A=int(2),B=int(X),
                     R=pred_true, X==2)).


/* equality_objects/3 is a function which compares two objects
   and returns either pred_true or pred_false in the third argument,
   as soon as it can be decided whether the objects are equal */


equality_objects_lwf(X,Y,Res,_LWF) :- X==Y,!,Res=pred_true.
equality_objects_lwf(X,Y,Res,LWF) :-
   equality_objects(X,Y,Res),
   force_equality_objects_lwf2(X,Y,Res,LWF).

:- block force_equality_objects_lwf2(?,?,-,-).
force_equality_objects_lwf2(X,Y,Res,_LWF) :-   /* TO : try avoid calling equal_object/not_equal_object */
   (  nonvar(Res) -> true
    ; (  %% print(forcing_equal(X,Y,_WF)),nl, %%
      % alternative would be to propagate LWF down to and_equality ?
         Res \== pred_false, equal_object(X,Y), % not required anymore as and_equality can propagate pred_true down ????
         Res=pred_true
       ; Res \== pred_true, not_equal_object(X,Y), Res=pred_false
      )).
  

:- block equality_objects_with_expansion(-,-,?), equality_objects_with_expansion(?,-,-), equality_objects_with_expansion(-,?,-).
% will expand second argument; use if first argument is a tail of a list (representing a set)
equality_objects_with_expansion(X,Y,R) :- 
   nonvar(Y),  is_custom_explicit_set(Y,equality_objects_with_expansion),!,
    expand_custom_set(Y,ExpSet),
    equality_objects(X,ExpSet,R).
equality_objects_with_expansion(X,Y,R) :- equality_objects(X,Y,R).


:- use_module(bool_pred).
% a version of equality_objects with type information available: allows to dispatch to boolean version
equality_objects_with_type(boolean,X,Y,R) :- !, bool_pred:bool_equality(X,Y,R).
equality_objects_with_type(_,X,Y,R) :- (X==Y -> R=pred_true ; equality_objectsb(X,Y,R)).

:- use_module(b_global_sets,[is_global_set_of_cardinality_one/2]).
equality_objects(X,Y,R) :- (X==Y -> R=pred_true ; equality_objectsb(X,Y,R)).
% This is in principle more precise: but entails about a 10 % penalty for make cruise
%equality_objects(X,Y,R) :- when((?=(X,Y);nonvar(X);nonvar(Y)),  (X==Y -> R=pred_true  ; equality_objectsb(X,Y,R))).
:- block equality_objectsb(-,-,-).
equality_objectsb(X,Y,R) :- 
   (nonvar(X) -> X=fd(XN,GS) ; nonvar(Y) -> Y=fd(XN,GS)),
   is_global_set_of_cardinality_one(GS,N),!, % there exist only one object of that type
   XN=N, Y=X, R=pred_true.
%equality_objectsb(X,Y,R) :- X==Y,!,R=pred_true. % only makes sense if all three are variables; cannot happen here; we could have int(X)==int(Y), but then code further below will trigger
equality_objectsb(X,Y,R) :- equality_objects0(X,Y,R).

%:- block %equality_objects0(-,-,?), 
%         equality_objects0(?,-,-), equality_objects0(-,?,-).
equality_objects0(X,Y,R) :- nonvar(R),!, % we know the result of the comparison
  ( R=pred_true  -> equal_object(X,Y)
  ; R=pred_false -> not_equal_object(X,Y)
  ; add_error_fail(equality_objects0,'Illegal result: ',equality_objects0(X,Y,R))
  ).
% now we know that either X or Y must be nonvar
equality_objects0(X,Y,R) :-
   (nonvar(Y)
     -> (X==[] -> eq_empty_set(Y,R)
          ;  equality_objects1(Y,X,R)
        )
     ;  equality_objects1(X,Y,R)
   ).

:- use_module(clpfd_interface,[post_constraint2/2]).
equality_int(X,Y,Res) :- nonvar(X),nonvar(Y),!,(X=Y -> Res=pred_true ; Res=pred_false).
equality_int(X,Y,Res) :- 
   clpfd_interface:post_constraint2( ((X#=Y) #<=> R01), Posted),
   (Posted = true
      -> prop_eq_01(Res,R01)
       ; eq_atomic(X,Y,integer,Res)).

% equality_objects1 assumes first argument is nonvar:
%%equality_objects1(X,Y,R) :- print(equality_objects1(X,Y,R)),nl,fail. %%
equality_objects1(pred_true /* bool_true */,X,Res) :- !, bool_equality(pred_true /* bool_true */,X,Res). %bool_eq_atomic(X,bool_true,pred_false /* bool_false */,Res).
equality_objects1(pred_false /* bool_false */,X,Res) :- !,bool_equality(pred_false /* bool_false */,X,Res). %bool_eq_atomic(X,bool_false,bool_true,Res).
equality_objects1(int(X),IY,Res) :- !, IY=int(Y), equality_int(X,Y,Res).
equality_objects1(term(X),TY,Res) :- !, TY=term(Y), eq_atomic(X,Y,term,Res).
equality_objects1(string(X),SY,Res) :- !, SY=string(Y), eq_atomic(X,Y,string,Res).
equality_objects1(fd(X,Type),FY,Res) :- !,  FY=fd(Y,Type), %print(equality_fd(X,Y,Type,Res)),nl,
   equality_fd_objects(X,Y,Type,Res).
equality_objects1(freeval(Id,Case1,Value1),FY,Res) :- !, FY=freeval(Id,Case2,Value2),
    and_equality(CaseRes,ValueRes,Res),
    eq_atomic(Case1,Case2,freeval_case,CaseRes),
    equality_freeval_content(CaseRes,Value1,Value2,ValueRes).
equality_objects1(rec(X),RY,Res) :- !, RY=rec(Y), equality_fields(X,Y,Res).
equality_objects1((X,Y),XY2,Res) :- !, XY2=(X2,Y2),
     and_equality(RY,RX,Res),
     equality_objects(X,X2,RX),
     (RX==pred_false -> true
       ; equality_objects(Y,Y2,RY)).
equality_objects1([],S,R) :- !,eq_empty_set(S,R).
%equality_objects1(X,[],R) :- !,eq_empty_set(X,R).
equality_objects1(X,Y,R) :- equality_objects2(X,Y,R).

:- block equality_freeval_content(-,?,?,?).
% if the case is different, we don't care anymore about the content
equality_freeval_content(pred_false,_Value1,_Value2,_).
equality_freeval_content(pred_true,Value1,Value2,Res) :-
    % if we do not wait until we now that the case is the same,
    % an internal type error might be raised
    equality_objects(Value1,Value2,Res).


:- block prop_eq_01(-,-).
% prop_eq_01(A,B) :- print(prop(A,B)),nl,fail.
prop_eq_01(P,B) :- var(P), !, prop_01_eq(B,P).
prop_eq_01(pred_true,1).
prop_eq_01(pred_false,0).
prop_01_eq(1,pred_true).
prop_01_eq(0,pred_false).

:- use_module(tools_printing,[print_term_summary/1]).
:- block equality_objects2(?,-,-). % also wait on second or third argument
equality_objects2(X,Y,Res) :- nonvar(Res),!,% we know the result of the comparison
  ( Res=pred_true  -> %% print(forcing_equal(X,Y)),nl,%%
                 equal_object(X,Y)
  ; Res=pred_false -> not_equal_object(X,Y)
  ; add_error_fail(equality_objects2,'Illegal result: ',equality_objects2(X,Y,Res))
  ).
equality_objects2([H|T],S,Res) :- !,eq_cons(S,H,T,Res). % move to equality_objects1 ?!
equality_objects2(avl_set(A),[H|T],Res) :- !, eq_cons(avl_set(A),H,T,Res).
equality_objects2(ES1,ES2,Res) :- equality_explicit_sets(ES1,ES2,ERes), !, Res=ERes.
equality_objects2(ES1,S,Res) :- is_custom_explicit_set(ES1,eq_cons),!, % move to custom_explicit_sets?
    expand_custom_set(ES1,ExpSet), equality_objects0(ExpSet,S,Res).
equality_objects2(X,Y,Res) :-
    nl,print_term_summary(unknown_equality_objects(X,Y,Res)),nl, % can happen with abort(.) terms
    (equal_object(X,Y),Res=pred_true
    ;
     not_equal_object(X,Y),Res=pred_false
    ).

:- use_module(b_global_sets,[b_fd_type/3]).
:- use_module(library(clpfd)).
equality_fd_objects(X,Y,_,Res) :- number(X),number(Y),!,
  (X=Y -> Res=pred_true ; Res= pred_false).
%equality_fd_objects(X,Y,_Type,Res) :- !,
%   ((X#=Y) #<=> R01), prop_eq_01(Res,R01).
% Is it really necessary to keep code below: clp(FD) will do this for us ?!
equality_fd_objects(X,Y,Type,Res) :-
    nonvar(Type),b_global_sets:b_fd_type(Type,LowBnd,UpBnd),!,
    (LowBnd=UpBnd
      -> X=Y,Res=pred_true /* only one object of Type exists */
      ;  (LowBnd+1 =:= UpBnd
          -> bin_eq_atomic(X,Y,Res,LowBnd,UpBnd) /* two objects exist */
          ;  eq_atomic(X,Y,global(Type),Res),
			  ((var(Res),(var(X);var(Y))) 
			   ->  X in LowBnd..UpBnd, Y in LowBnd..UpBnd, % still required ! even with fd_utils_clpfd ! (for v8 of Bosch CrCtl_Comb2_Final_mch.eventb)
				  ((X#=Y) #<=> R01), prop_eq_01(Res,R01)
			   ;  true
			 )
          )
    ).
equality_fd_objects(X,Y,Type,Res) :- 
    add_internal_error('Unknown type: ',equality_fd_objects(X,Y,Type,Res)),
    eq_atomic(X,Y,global(Type),Res).

:- block eq_empty_set(-,-).
eq_empty_set(X,R) :- var(X),!,
  (R=pred_true -> empty_set(X) ; not_empty_set(X)).
eq_empty_set([],R) :- !,R=pred_true.
eq_empty_set([_|_],R) :- !,R=pred_false.
eq_empty_set(CS,R) :- is_custom_explicit_set(CS,eq_empty_set),!,
   (is_empty_explicit_set(CS) -> R=pred_true
                               ; R=pred_false).
eq_empty_set(S,R) :- add_internal_error('Illegal Set: ',eq_empty_set(S,R)),fail.

%%eq_cons(S,H,T,Res) :- print(eq_cons(S,H,T,Res)),nl,fail.
eq_cons([],_,_,Res) :- !, Res=pred_false.
eq_cons([H2|T2],H1,T1,Res) :- !,
   check_and_remove([H2|T2],H1,NewSet2,RemoveSuccesful),
    %% print(removed(H1,H2,T2,NewSet2,RemoveSuccesful)), %%
   eq_cons2(RemoveSuccesful,T1,NewSet2,Res).
eq_cons(avl_set(A),H1,T1,Res) :- element_can_be_added_or_removed_to_avl(H1),!,
    (remove_element_from_explicit_set(avl_set(A),H1,RA)
      ->  equality_objects_with_expansion(T1,RA,Res)  %%% PROBLEM: T1 has to be in list form !
       ;  Res = pred_false
    ).
eq_cons(ES,H1,T1,Res) :- is_custom_explicit_set(ES,eq_cons),expand_custom_set(ES,ExpSet),
   eq_cons1(ExpSet,H1,T1,Res).

:- block eq_cons1(-,?,?,?).
eq_cons1(ExpSet,H1,T1,Res) :- eq_cons(ExpSet,H1,T1,Res).
   
:- block eq_cons2(-,?,?,?).
eq_cons2(not_successful,_T1,_NewSet2,pred_false).
eq_cons2(successful,T1,NewSet2,Res) :-  equality_objects(T1,NewSet2,Res).
   
:- block check_and_remove(-,?,?,?).
check_and_remove([],_,[],not_successful).
check_and_remove([H|T],ElementToRemove,ResultingSet,Res) :-
   equality_objects(H,ElementToRemove,RH),
   %%print(equality_objects_result(H,ElementToRemove,RH)),nl,
   check_and_remove2(RH,H,T,ElementToRemove,ResultingSet,Res).

:- block check_and_remove2(-,?,?,?,?,?).
check_and_remove2(pred_true, _H,T,_ElementToRemove,T,successful).
check_and_remove2(pred_false,H,T,ElementToRemove,[H|RT],Res) :-
   check_and_remove(T,ElementToRemove,RT,Res).


:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:and_equality(pred_false,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:and_equality(pred_true,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:and_equality(pred_true,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:and_equality(pred_true,pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:and_equality(pred_true,pred_false,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:and_equality(pred_false,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:and_equality(pred_false,pred_false,pred_true))).
:- block and_equality(-,-,-).
and_equality(X,Y,Res) :- 
   (   X==pred_true  -> Res=Y
     ; X==pred_false -> Res=pred_false
     ; Y==pred_true  -> Res=X
     ; Y==pred_false -> Res=pred_false
     ; Res==pred_true -> X=pred_true, Y=pred_true
     ; Res==pred_false -> and_equality2(X,Y,Res)  % we know that one of X,Y must be pred_false
     ; add_error_fail(and_equality,'Illegal call: ',and_equality(X,Y,Res))
   ).
:- block and_equality2(-,-,?).
and_equality2(X,Y,Res) :- 
   (   X==pred_true  -> Res=Y
     ; X==pred_false -> Res=pred_false
     ; Y==pred_true  -> Res=X
     ; Y==pred_false -> Res=pred_false
     ; add_error_fail(and_equality,'Illegal call: ',and_equality2(X,Y,Res))
   ).

/*
:- block bool_eq_atomic(-,?,?,-).
bool_eq_atomic(X,Y,OtherY,Res) :- nonvar(Res),!,
   (Res=pred_true -> X=Y ; X=OtherY).
bool_eq_atomic(X,Y,_,Res) :- X==Y -> Res=pred_true ; Res=pred_false. */


:- block bin_eq_atomic(-,?,-,?,?), bin_eq_atomic(?,-,-,?,?).
bin_eq_atomic(X,Y,Res,V1,V2) :- nonvar(Res),!,
   %print(forcing_bin_eq_atomic(X,Y,Res,V1,V2)),nl,
   (Res=pred_true -> X=Y
    ; %print(bin_dif(X,Y)),nl,
      ((X=V1,Y=V2) ; (X=V2,Y=V1))
    ). % difference with eq_atomic: we now know the value !
bin_eq_atomic(X,Y,Res,_,_) :-
  %print(triggering_bin_eq(X,Y,Res)), translate:print_frozen_info(Res),nl, 
  (X==Y -> Res=pred_true ; Res=pred_false).

eq_atomic(X,Y,_Type,Res) :- nonvar(X),nonvar(Y),!,(X=Y -> Res=pred_true ; Res=pred_false).
eq_atomic(X,Y,Type,Res) :- when((?=(X,Y);nonvar(Res)),eq_atomic2(X,Y,Type,Res)).

/*
Comment: % for integers CLPFD reification will not ground Res if X==Y,   need to do when((?=(X,Y)...
| ?- X in 1..10, Y in 1..10, (X#=Y  #<=> Z), X=Y.
Y = X,
X in 1..10,
Z in 0..1 ?  */

eq_atomic2(X,Y,Type,Res) :- nonvar(Res),!,  (Res=pred_true -> X=Y ; dif(Type,X,Y)).
eq_atomic2(X,Y,_Type,Res) :- (X=Y -> Res=pred_true ; Res=pred_false).



%:- use_module(fd_utils,[neq_fd/3]).
:- use_module(fd_utils_clpfd,[neq_fd/3]).
dif(boolean,A,B) :- !, bool_pred:negate(A,B). % was: bool_dif(A,B).
dif(integer,A,B) :- !, integer_dif(A,B).
dif(global(T),A,B) :- !, neq_fd(A,B,T).
dif(_,A,B) :- dif(A,B).

:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:integer_dif(0,1))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:integer_dif(1,1))).
:- use_module(clpfd_interface,[clpfd_neq/2]).
integer_dif(A,B) :- clpfd_neq(A,B). % clpfd_neq calls dif on overflow


:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:bool_dif(pred_false,pred_true))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:bool_dif(pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:bool_dif(pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:bool_dif(pred_false,pred_false))).
bool_dif(A,B) :- bool_pred:negate(A,B).

/*  old version:
:- block eq_atomic(-,?,-), eq_atomic(?,-,-).                      
eq_atomic(X,Y,Res) :- nonvar(Res),!,
   (Res=pred_true -> X=Y ; dif(X,Y)).
eq_atomic(X,Y,Res) :- X==Y -> Res=pred_true
                            ; Res=pred_false.
*/

:- block equality_fields(-,-,?).
equality_fields([],[],pred_true).
equality_fields([field(Name1,V1)|T1],[field(Name2,V2)|T2],Res) :- 
 % should we wait for Name1 or Name2 to become nonvar?
  check_field_name_compatibility(Name1,Name2),
  and_equality(RH,RT,Res),
  equality_objects(V1,V2,RH),
  (RH==pred_false -> true
                ; equality_fields(T1,T2,RT)).

check_field_name_compatibility(Name1,Name2) :- nonvar(Name1), nonvar(Name2), Name1 \= Name2, !,
   add_error_and_fail(equality_fields,'Incompatible fields: ',fields(Name1,Name2)).
check_field_name_compatibility(_,_).


empty_set_test(Set,EqRes) :- kernel_equality:eq_empty_set(Set,EqRes).
empty_cartesian_product(Set1,Set2,EqRes) :- 
   eq_empty_set(Set1,EqRes1),
   (EqRes1==pred_true -> EqRes=pred_true /* avoid inspecting Set2 */
    ; or_eq_obj(EqRes1,EqRes2,EqRes),eq_empty_set(Set2,EqRes2)).


:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:or_eq_obj(pred_false,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:or_eq_obj(pred_false,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:or_eq_obj(pred_true,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:or_eq_obj(pred_true,pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:or_eq_obj(pred_true,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:or_eq_obj(pred_false,pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:or_eq_obj(pred_false,pred_false,pred_true))).
:- block or_eq_obj(-,-,-).
% very similar to disjoin in b_interpreter_check; no waitflag to instantiate if result false
or_eq_obj(X,Y,XorY) :-
  ( (X==pred_true ; Y==pred_true) -> XorY=pred_true
   ; XorY==pred_false -> X=pred_false,Y=pred_false
   ; X==pred_false -> XorY = Y
   ; Y==pred_false -> XorY = X
   ; XorY==pred_true -> or_eq_obj2(X,Y)
   ; add_internal_error('Illegal call: ',or_eq_obj(X,Y,XorY)) ).
:- block or_eq_obj2(-,-).
or_eq_obj2(X,Y) :- ((X==pred_true; Y=pred_true) -> true
            ; X==pred_false -> Y=pred_true
            ; Y==pred_false -> X=pred_true
            ; add_internal_error('Illegal call: ',or_eq_obj2(X,Y))).

% ------------------------------------------------------------------


% not used at the moment:
%get_partial_subset_priority(Set,WF,LWF) :- get_partial_set_size(Set,Size),
%              safe_pow2(Size,Prio),
%              get_wait_flag(Prio,psubset(Set),WF,LWF).
%get_partial_set_priority(Set,WF,LWF) :- get_partial_set_size(Set,Size),
%              get_wait_flag(Size,pset(Set),WF,LWF).
%get_partial_set_size(X,R) :- var(X),!,R=3.
%get_partial_set_size([],R) :- !,R=0.
%get_partial_set_size([_|T],R) :- !,get_partial_set_size(T,RT), R is RT+1.
%get_partial_set_size(_,10).


get_cardinality_wait_flag(Set,Info,WF,LWF) :-
    get_cardinality_wait_flag(Set,Info,WF,_Card,LWF).
get_cardinality_wait_flag(Set,Info,WF,Card,LWF) :-
    cardinality_as_int_for_wf(Set,Card),
  % when(nonvar(X), (print(get_card_waitflag(X,Set,WF,LWF)),nl)),
    get_cardinality_wait_flag2(Card,Info,WF,LWF).
:- block get_cardinality_wait_flag2(-,?,?,?).
get_cardinality_wait_flag2(X,Info,WF,LWF) :- get_wait_flag(X,Info,WF,LWF).

get_cardinality_powset_wait_flag(Set,WF,LWF) :-
    cardinality_as_int_for_wf(Set,X),
    get_cardinality_powset_wait_flag2(X,WF,LWF).
:- block get_cardinality_powset_wait_flag2(-,?,?).
get_cardinality_powset_wait_flag2(X,WF,LWF) :-
     priority_pow2(X,X2), 
     %safe_mul(X2,10000,X3), % as generating subsets this way will generate repetitions
     %print(powset_waitflag(X,X2)),nl,
     get_wait_flag(X2,powset_card,WF,LWF).

priority_pow2(X,R) :- X==inf,!, R=10000010.  % TO DO: provide a separate module for combining priorities in a systematic way !
priority_pow2(X,RR) :- safe_pow2(X,R),
 (R==inf -> RR=10000010 ; RR=R).

get_cardinality_relation_over_wait_flag(Dom,Ran,WF,LWF,MaxCardOfRel,MaxNrOfRels) :-
    cardinality_as_int_for_wf(Dom,X),
    cardinality_as_int_for_wf(Ran,Y),
    get_cardinality_relation_over_wait_flag2(X,Y,WF,LWF,MaxCardOfRel,MaxNrOfRels).

:- block get_cardinality_relation_over_wait_flag2(-,?,?,?,?,?), get_cardinality_relation_over_wait_flag2(?,-,?,?,?,?).
get_cardinality_relation_over_wait_flag2(X,Y,WF,LWF,MaxCardOfRel,MaxNrOfRels1) :-
    safe_mul(X,Y,MaxCardOfRel),
    safe_pow2(MaxCardOfRel,MaxNrOfRels),
    safe_add(MaxNrOfRels,1,MaxNrOfRels1),  /* add 1 to give priority to tighter_enum */
   % print(get_rel_card(X,Y,XY2)),nl,
    get_wait_flag(MaxNrOfRels1,powset_rel_card,WF,LWF).

% ------------------------------------------------------------------

:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:membership_test([int(2),int(1)],int(2),pred_true))).  
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:membership_test([int(2),int(1)],int(1),pred_true))).  
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:membership_test([int(2),int(1),int(3),int(4),int(5)],int(4),pred_true))).  
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:membership_test([int(2),int(1)],int(3),pred_false))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:membership_test([],int(3),pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:membership_test([int(2),int(1)],int(3),pred_true))).
:- assert_must_succeed((membership_test([int(2),int(1)],int(2),R),R==pred_true)).
:- assert_must_succeed((membership_test([int(2),int(1)],int(X),R),var(R),X=1,R==pred_true)).
:- assert_must_succeed((membership_test([int(2),int(1)],int(3),R),R==pred_false)).
:- assert_must_succeed((membership_test(global_set('Name'),fd(1,'Name'),R),R==pred_true)).

:- assert_must_succeed((kernel_equality:membership_test(avl_set(node(rec([field(cardinality,int(2)),field(col,fd(2,setField1)),field(fill,fd(1,setField2)),field(shape,fd(1,setField1))]),true,0,empty,empty)),
 rec([field(cardinality,int(2)),field(col,fd(2,setField1)),field(fill,fd(1,setField2)),field(shape,fd(1,setField1))]),R), R==pred_true)).
:- assert_must_succeed((
kernel_equality:membership_test([rec([field(cardinality,int(2)),field(col,fd(2,setField1)),field(fill,fd(1,setField2)),field(shape,fd(1,setField1))])],
rec([field(cardinality,int(2)),field(col,fd(2,setField1)),field(fill,fd(1,setField2)),field(shape,fd(1,setField1))]),R), R==pred_true
)).
/* A function that instantiates last argument when membership test can be decided */
:- block membership_test(-,?,-).
membership_test(Set,X,Res) :- var(Set),!,
   (Res==pred_true -> Set=[X|_] % TO DO: normalise X first ?
    ; membership_not_el(Set,X)).
membership_test(Set,X,Res) :- 
    membership_test2(Set,X,Res).

:- block membership_not_el(-,?).
membership_not_el([],_) :- !.
membership_not_el([H|T],X) :- !,kernel_objects:not_equal_object(H,X), membership_not_el(T,X).
membership_not_el(avl_set(A),X) :- !,membership_custom_set(avl_set(A),X,pred_false). 
membership_not_el(CS,X) :- is_custom_explicit_set(CS,membership_test),!,
    membership_custom_set(CS,X,pred_false).

membership_test2([],_X,Res) :- !, Res=pred_false.
membership_test2([H|T],X,Res) :- !,
  (Res==pred_false
    -> kernel_objects:not_equal_object(H,X), membership_test(T,X,Res)
    ;  equality_objects(H,X,HXRes), 
       (T==[] -> Res=HXRes ; cons_el_of(HXRes,T,X,Res))).
membership_test2(avl_set(A),X,Res) :- !,membership_custom_set(avl_set(A),X,Res). 
membership_test2(CS,X,Res) :- is_custom_explicit_set(CS,membership_test),!,
    membership_custom_set(CS,X,Res).
membership_test2(Z,X,Res) :-
    add_error_fail(membership_test,'Not Set as first argument: ', membership_test(Z,X,Res)).



:- block cons_el_of(-,?,?,-).
cons_el_of(EQRES,T,X,MEMRES) :- var(EQRES),!, %print(cons_el_of(EQRES,T,X,MEMRES)),nl,
   (MEMRES=pred_false -> EQRES=pred_false, /* otherwise element of set */
                           membership_not_el(T,X)
      ; (T==[] -> print(forcing_cons_el_of(T,X,MEMRES)),nl,EQRES=pred_true
       ; cons_el_of_mem_obj(EQRES,T,X)) /* we have to wait for result of comparison with X */
    ).
cons_el_of(pred_true,_,_,pred_true).
cons_el_of(pred_false,T,X,Res) :- (Res==pred_true -> T=[_|_] ; true), membership_test(T,X,Res).

:- block cons_el_of_mem_obj(-,?,?).
cons_el_of_mem_obj(pred_true,_,_).
cons_el_of_mem_obj(pred_false,T,X) :- T=[_|_], membership_test(T,X,pred_true).


:- assert_must_succeed((kernel_equality:cartesian_pair_test((int(1),int(1)),
           [int(1),int(2)],[int(2),int(3)],R),R==pred_false)).
:- assert_must_succeed((kernel_equality:cartesian_pair_test((int(3),int(1)),
           [int(1),int(2)],[int(2),int(3)],R),R==pred_false)).
:- assert_must_succeed((kernel_equality:cartesian_pair_test((int(2),int(3)),
           [int(1),int(2)],[int(2),int(3)],R),R==pred_true)).
:- assert_must_fail((kernel_equality:cartesian_pair_test((int(1),int(3)),
           [int(1),int(2)],[int(2),int(3)],pred_false))).
 
cartesian_pair_test((X,Y),XType,YType,Res) :- check_cartesian_pair1(X,XType,Y,YType,Res).

:- block check_cartesian_pair1(-,?,-,?,-).
%check_cartesian_pair1(X,XType,Y,YType,Res) :- Res==pred_true,!,
%   is_cartesian_pair_wf((X,Y),XType,YType,WF).
%check_cartesian_pair1(X,XType,Y,YType,Res) :- Res==pred_false,!,
%   not_is_cartesian_pair((X,Y),XType,YType,WF).
check_cartesian_pair1(X,XType,Y,YType,Res) :- 
   ((var(X),nonvar(Y)) -> check_cartesian_pair2(Y,YType,X,XType,Res)
     ; check_cartesian_pair2(X,XType,Y,YType,Res)).
     
check_cartesian_pair2(X,XType,Y,YType,Res) :- prop_pred_true(Res,MemResX),
   membership_test(XType,X,MemResX),
   check_cartesian_pair2(MemResX,Y,YType,Res).

:- block check_cartesian_pair2(-,?,?,?).
check_cartesian_pair2(pred_true,Y,YType,Res) :- membership_test(YType,Y,Res).
check_cartesian_pair2(pred_false,_Y,_YType,pred_false).

:- block prop_pred_true(-,?).
prop_pred_true(pred_true,pred_true).
prop_pred_true(pred_false,_).





:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:in_nat_range_test(int(9),int(10),int(12),pred_false))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:in_nat_range_test(int(11),int(10),int(12),pred_true))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:in_nat_range_test(int(22),int(22),int(22),pred_true))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:in_nat_range_test(int(22),int(22),int(21),pred_false))). 
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:in_nat_range_test(int(9),int(10),int(12),pred_true))). 
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:in_nat_range_test(int(22),int(22),int(22),pred_false))). 
:- assert_must_succeed((in_nat_range_test(X,int(11),int(12),pred_false), X=int(10) )).
:- assert_must_succeed((kernel_equality:in_nat_range_test(X,int(11),int(12),pred_false), X=int(13) )).
:- assert_must_succeed((in_nat_range_test(X,int(11),int(12),pred_true), X=int(11) )).
:- assert_must_succeed((kernel_equality:in_nat_range_test(int(11),X,int(12),pred_true), X=int(11) )).
:- assert_must_succeed((in_nat_range_test(int(11),X,int(12),pred_false), X=int(12) )).
:- assert_must_succeed((kernel_equality:in_nat_range_test(int(11),int(10),int(12),R), R==pred_true )).
:- assert_must_fail((kernel_equality:in_nat_range_test(int(11),int(10),int(12),pred_false))).

in_nat_range_test(int(X),int(Y),int(Z),Res) :- in_nat_range_test1(X,Y,Z,Res).

in_nat_range_test1(X,Y,Z,Res) :- number(X),number(Y),number(Z),!, 
% for improved performance, e.g., inside large set differences with intervals
   (  X < Y -> Res = pred_false
    ; X > Z -> Res = pred_false
    ; Res = pred_true).
in_nat_range_test1(X,Y,Z,ResYZ) :-
   prop_eq_01(ResYZ,R01),
   % print(posting((X#>=Y #/\ X#=<Z) #<=> R01)),nl,
   clpfd_interface:post_constraint( (X#>=Y #/\ X#=<Z) #<=> R01), % reify constraint if CLP(FD) active
   % print(in_nat_range_test(X,Y,Z,R01,Res)),nl,
   %was : convert_eq_to_mem(RYZ,Res),
   and_equality(RY,RZ,ResYZ),
   geq_test(X,Y,RY),
   geq_test(Z,X,RZ).

:- block geq_test(?,-,-),geq_test(-,?,-).
geq_test(X,Y,Res) :-
   (Res==pred_true -> kernel_objects:less_than_equal_direct(Y,X)
    ; Res==pred_false -> kernel_objects:less_than_direct(X,Y)
    ; X >= Y -> Res=pred_true
    ; Res = pred_false
   ).



/* SUBSET CHECK REIFICATION */

:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2)],[],pred_false,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2),int(1)],[int(2)],pred_false,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:subset_test([int(2),int(1)],[int(2)],pred_true,_WF))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2),int(1)],[int(2),int(1)],pred_true,_WF))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2),int(1)],[int(2),int(1),int(3)],pred_true,_WF))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2),int(1)],global_set('NATURAL1'),pred_true,_WF))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([],[int(2),int(1)],pred_true,_WF))). 
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:subset_test([],[int(2),int(1)],pred_false,_WF))). 
:- assert_must_succeed((kernel_equality:subset_test([],[int(3)],R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(2),int(1)],[int(2),int(1),int(3)],R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(2),int(1),int(4)],[int(2),int(1),int(3)],R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(1)],global_set('NATURAL'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(1)],global_set('NATURAL1'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(1),int(0)],global_set('NATURAL1'),R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(1),int(0)],X,R,_WF),X=global_set('NATURAL1'),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(-11)],global_set('INTEGER'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(-11)],global_set('INTEGER'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test(global_set('INTEGER'),global_set('NATURAL1'),R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_test(global_set('NATURAL1'),global_set('NATURAL1'),R,_WF),R==pred_true)).

:- use_module(kernel_objects,[check_subset_of_wf/3, not_subset_of_wf/3,
                              check_subset_of_global_sets/2, both_global_sets/4]).

:- use_module(custom_explicit_sets,[expand_custom_set_to_list/4,is_definitely_maximal_set/1]).
:- block subset_test(-,-,-,?).
subset_test(_S1,S2,R,_WF) :-   % print_term_summary(subset_test(_S1,S2,R,WF)), %%
   is_definitely_maximal_set(S2),!,R=pred_true.
subset_test(S1,S2,R,WF) :-  subset_test0(S1,S2,R,WF).
  
:- block subset_test0(-,?,-,?).
subset_test0(S1,S2,R,WF) :-   % print_term_summary(subset_test0(S1,S2,R,WF)), %%
   nonvar(R),!,
   (R==pred_true -> check_subset_of_wf(S1,S2,WF)
    ; R==pred_false -> not_subset_of_wf(S1,S2,WF)
    ; add_error_fail(subset_test0,'Illegal result value: ',R)).
subset_test0([],_,R,_WF) :- !,R=pred_true.
subset_test0(S1,S2,R,WF) :- subset_test1(S1,S2,R,WF).


:- use_module(coverage_tools,['$NOT_COVERED'/1]).

:- block subset_test1(-,?,-,?),subset_test1(?,-,-,?).
subset_test1(S1,S2,R,WF) :- nonvar(R),!,
   (R==pred_true -> check_subset_of_wf(S1,S2,WF) ; not_subset_of_wf(S1,S2,WF)).
subset_test1([],_,R,_WF) :- !,R=pred_true, '$NOT_COVERED'('cannot be covered').  /* cannot be covered; emtpy set treated above */
subset_test1(Set1,Set2,R,_WF) :-  kernel_objects:both_global_sets(Set1,Set2,G1,G2),!, % also deals with two intervals
    % print(test_subset_of_global_sets(G1,G2,R)),nl,
    test_subset_of_global_sets(G1,G2,R).
subset_test1(_Set1,Set2,R,_WF) :- 
   is_definitely_maximal_set(Set2),!, % TO DO: avoid checking it again if have already checked it above in subset_test
   R=pred_true.
subset_test1(Set1,Set2,R,_WF) :- Set2==[],!, eq_empty_set(Set1,R).
subset_test1(Set1,Set2,R,WF) :- test_subset_of_explicit_set(Set1,Set2,R,WF,Code),!, 
   call(Code).
subset_test1(Set1,Set2,R,WF) :-
  expand_custom_set_to_list(Set1,ESet1,_,subset_test1),
  subset_test2(ESet1,Set2,R,WF).

:- block subset_test2(-,?,?,?).
subset_test2([],_,R,_WF) :- !, R=pred_true.
subset_test2([H|T],Set2,R,WF) :- 
   membership_test(Set2,H,MemRes),
   subset_test3(MemRes,T,Set2,R,WF).

:- block subset_test3(-,?,?,?,?).
subset_test3(pred_false,_T,_Set2,R,_WF) :- R=pred_false.
subset_test3(pred_true,T,Set2,R,WF)   :- subset_test2(T,Set2,R,WF).

%:- block test_subset_of_global_sets(-,?,?), not_subset_of_global_sets(?,-,?).
%test_subset_of_global_sets(interval(Low1,Up1),interval(Low2,Up2)) :- !, test_interval_subset(Low1,Up1,Low2,Up2).
test_subset_of_global_sets(G1,G2,R) :- %print(check_subset(G1,G2)),nl,
   when((ground(G1),ground(G2)),(check_subset_of_global_sets(G1,G2) -> R=pred_true ; R=pred_false)).

% TO DO: provide reified version, especially for intervals with variables !

% ---------------------------- strict subset reification


:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2)],[],pred_false,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2),int(1)],[int(2)],pred_false,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:subset_strict_test([int(2),int(1)],[int(2)],pred_true,_WF))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2),int(1)],[int(2),int(1)],pred_false,_WF))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2),int(1)],[int(2),int(1),int(3)],pred_true,_WF))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2),int(1)],global_set('NATURAL1'),pred_true,_WF))). 
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([],[int(2),int(1)],pred_true,_WF))). 
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:subset_strict_test([],[int(2),int(1)],pred_false,_WF))). 
:- assert_must_succeed((kernel_equality:subset_strict_test([],[int(3)],R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(2),int(1)],[int(2),int(1),int(3)],R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(2),int(1),int(4)],[int(2),int(1),int(3)],R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(1)],global_set('NATURAL'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(1)],global_set('NATURAL1'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(1),int(0)],global_set('NATURAL1'),R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(1),int(0)],X,R,_WF),X=global_set('NATURAL1'),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(-11)],global_set('INTEGER'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(-11)],global_set('INTEGER'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test(global_set('INTEGER'),global_set('NATURAL1'),R,_WF),R==pred_false)).

:- assert_must_succeed((kernel_equality:subset_strict_test(global_set('NATURAL1'),global_set('NATURAL1'),R,_WF),R==pred_false)).

:- block subset_strict_test(-,-,-,?).
subset_strict_test(_S1,S2,R,_WF) :-   % print_term_summary(subset_test(_S1,S2,R,WF)), %%
   S2==[],!,R=pred_false.
subset_strict_test(S1,S2,R,_WF) :-   % print_term_summary(subset_test(_S1,S2,R,WF)), %%
   S1==[],!,eq_empty_set(S2,Empty), bool_pred:negate(Empty,R).
subset_strict_test(S1,S2,R,WF) :-  subset_strict_test0(S1,S2,R,WF).
  
:- use_module(kernel_objects,[strict_subset_of_wf/3, not_strict_subset_of_wf/3]).
:- block subset_strict_test0(-,?,-,?).
subset_strict_test0(S1,S2,R,WF) :-   % print_term_summary(subset_test0(S1,S2,R,WF)), %%
   nonvar(R),!,
   (R==pred_true -> strict_subset_of_wf(S1,S2,WF)
    ; R==pred_false -> not_strict_subset_of_wf(S1,S2,WF)
    ; add_error_fail(subset_strict_test0,'Illegal result value: ',R)).
subset_strict_test0(S1,S2,R,WF) :-  
    subset_test(S1,S2,IsSubset,WF),
    (IsSubset==pred_false -> R=pred_false
     ; conjoin(IsSubset,NotEqual,R,WF), bool_pred:negate(NotEqual,Equal),
       equality_objects(S1,S2,Equal)
     ).
  
conjoin(A,B,AB,WF) :-
    b_interpreter_check:conjoin(A,B,AB,b(truth,pred,[]),b(truth,pred,[]),WF).

