% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 


:- module(kernel_waitflags,
   [init_wait_flags/1, waitflag0_is_set/1,
    get_wait_flag/3, get_wait_flag/4,
    get_wait_flag0/2, get_wait_flag1/2, 
    get_middle_wait_flag/3, get_last_wait_flag/3,
    get_idle_wait_flag/3,
    get_binary_choice_wait_flag/3,
    %refresh_wait_flag/4,
    add_fd_variable_for_labeling/4,  add_fd_variable_for_labeling/2,
    get_new_subsidiary_wait_flag/4,
    update_waitflag/4,
    ground_constraintprop_wait_flags/1,
    check_is_wait_flag/1,
    create_inner_wait_flags/2, copy_wf0e_wait_flags/2,
    copy_wf_start/2, copy_wf_finish/2,
    
    clone_wait_flags_from1/2, ground_wait_flag0/1,
    %create_wdguarded_wait_flags/4,
    %get_abort_error_flag/2,
    get_enumeration_starting_wait_flag/2,
    get_enumeration_finished_wait_flag/2, 
    ground_det_wait_flag/1, ground_wait_flag_to/2,
    ground_wait_flags/1, ground_inner_wait_flags/1,
    
    add_abort_error_span/5,
    add_wd_error_span/4,
    add_wd_error/3, add_wd_error_set_result/6,
    %add_wd_error_span_now/3,
    assert_must_abort_wf/2,
    
    portray_waitflags/1,
    
    set_silent/1]).
/* File: kernel_waitflags.pl */
/* Created: 17/3/06 by Michael Leuschel */

/*
Phases of the ProB Kernel:

 -> all Waitflags unbound
 
* Phase 0:  propagation of completely known ground values in a deterministic fashion
 
 -> WF0 is bound
 
* Phase 1:  deterministic propagation (not necessarily of completely known full values)
 
 -> WF1 is bound
 
* Phase 2:  boolean propagation (e.g., P or Q will now try P or Q)
 
 -> WF2 is bound
 
* Phase 3:  non-deterministic propagation (e.g, x: {1,2,3})
 
 -> Enumeration Starting WF is bound
 
* Enumeration Phase:  (later to be split into enumeration of finite and infinite types (e.g., integer)
 
  -> Enumeration Finished Waitflag bound
 
* Error Generation Phase: errors for, e.g., division by zero, f(x) with x outside domain of f,... are generated
   
*/

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(typechecker).
:- use_module(library(clpfd),[labeling/2, fd_size/2, fd_degree/2]).
:- use_module(tools).
:- use_module(coverage_tools).
:- use_module(preferences,[preference/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module manages the various choice points and enumerators of ProB, by maintaining the kernel Waitflags store.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- type wait_flag +--> (var ; wfx(any,var,any)).

:- use_module(library(avl)).
:- use_module(library(atts)).

:- attribute wf_store/1,wf_fdvars/1.

% the Call must succeed and generate one abort_error
:- meta_predicate assert_must_abort_wf(0,*).
assert_must_abort_wf(M:Call,WF) :- assert_must_succeed_any(M:(kernel_waitflags:assert_must_abort2_wf(M:Call,WF))).

assert_must_abort2_wf(_,_) :- get_error(well_definedness_error,M),
   add_internal_error('Abort error prior to call in assert_must_abort: ',M),fail.
assert_must_abort2_wf(Call,WF) :- (real_error_occurred -> RE=true ; RE=false),
  init_wait_flags(WF),
  call(Call), print(called(Call)),nl,
  (ground_wait_flags(WF)
  -> add_internal_error('Call did not fail after grounding WF: ',assert_must_abort2_wf(Call,WF)),fail
  ;  (get_error(well_definedness_error,_), % we have 1 abort_error
       print(got_well_definedness_error),nl,
       (RE=false -> reset_real_error_occured ; true)
     )
  ).

:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF), kernel_waitflags:check_is_wait_flag(WF) )).

check_is_wait_flag(WFX) :- var(WFX),!,
  add_internal_error('Variable as waitflag: ',check_is_wait_flag(WFX)).
check_is_wait_flag(WFX) :-
  WFX = wfx(_,Store,_)
    -> (get_atts(Store,+wf_store(_))
         -> true
          ; print('WF does not yet contain heap: '), print(WFX),nl)
            %add_internal_error('WF does not contain heap: ',check_is_wait_flag(WFX)))
    ;  add_internal_error('Illegal WF Format: ',check_is_wait_flag(WFX)).

verify_attributes(Var, Value, [] ) :-
  nonvar(Value) -> print('Trying to unify waitflag_heap_store'),nl
                 ;  print(verify_attributes(Var,Value)),nl.

init_wait_flags(wfx(_,_,_)).
%init_wait_flags(wfx(_,Store,_)) :- init_wait_flags_store(Store).
% we no longer do this because of performance, (see, e.g., LotsOfInvariants.mch) :
init_wait_flags_store(_).
%init_wait_flags_store(Store) :-
%	empty_avl(Heap),
%	put_atts(Store,+wf_store(Heap)),
%	put_atts(Store,+wf_fdvars([])).


:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
     kernel_waitflags:ground_wait_flag0(WF), kernel_waitflags:waitflag0_is_set(WF) )).
:- assert_must_fail(( kernel_waitflags:init_wait_flags(WF), kernel_waitflags:waitflag0_is_set(WF) )).
waitflag0_is_set(wfx(WF0,_Store,_EF)) :- ground(WF0).

get_new_subsidiary_wait_flag(OldWF,Info,WFX,NewWF) :-
   (ground(OldWF) /* then we have created a choice point with the old WF: double new WF prio */
      -> NewPrio is OldWF*2, get_wait_flag(NewPrio,Info,WFX,NewWF)
      ;  NewWF = OldWF
   ).

% not used at the moment:
%refresh_wait_flag(OldWF,Info,WFX,NewWF) :-
%   (ground(OldWF) /* then we have created a choice point with the old WF: double new WF prio */
%      -> get_wait_flag(OldWF,Info,WFX,NewWF)
%      ;  NewWF = OldWF
%   ).

:- volatile silent/0.
:- dynamic silent/0.
set_silent(true) :- !,(silent -> true ; assert(silent)).
set_silent(false) :- retractall(silent).
waitflag_not_init :-  silent -> true
   ; print('waitflag-store not initialised'),nl, '$NOT_COVERED'('This should not happen').

% adds a CLP(FD) variable with given domain Size, will be enumerated by labeling (even if clpfd_solver preference is false !)
% Size could be computed by fd_size(FDVariable,Size)
add_fd_variable_for_labeling(FDVariable,WF) :- fd_size(FDVariable,Size),
   (Size=sup -> get_wait_flag(inf,enumerate_int_wf(FDVariable),WF,LWF),
       kernel_objects:enumerate_int_wf(FDVariable,LWF)
     ; add_fd_variable_for_labeling(FDVariable,Size,WF,_LWF)
   ).

add_fd_variable_for_labeling(FDVariable,_,_WF,LWF) :- nonvar(FDVariable),!,LWF=1.
add_fd_variable_for_labeling(FDVariable,Size,WF,LWF) :-   % print(add_fd_variable(FDVariable,Size)),nl,
   %translate:print_frozen_info(FDVariable),nl, when(nonvar(FDVariable),(nl,print(fd_found(FDVariable,Size)),nl,nl,flush_output(user))),
   preferences:preference(use_smt_mode,false),!,
   get_wait_flag(Size,label_clpfd_vars([FDVariable]),WF,LWF).
add_fd_variable_for_labeling(FDVariable,_Size,WF,LWF) :- 
  % otherwise: store the FDVariable in the separate list of FD Variables (all mixed together)
  WF=wfx(_,Store,_EnumFinished),
  my_get_fdvars_atts(Store,FDVARS),
  %insert_fd_var(FDVARS,FDVariable,Size,LWF,NewFDVARS),
  NewFDVARS = [fd_var(FDVariable,LWF)|FDVARS],
  %  print(added_fd_variable(FDVariable,Size,NewFDVARS)),nl, %%
  put_atts(Store,+wf_fdvars(NewFDVARS)).



/*
insert_fd_var([],V,Sz,LWF,[fd_var(V,Sz,LWF)]).
insert_fd_var([fd_var(V2,Sz2,LWF2)|T],V1,Sz1,LWF1,Res) :-
   (Sz1 =< Sz2
     ->  Res= [fd_var(V1,Sz1,LWF1),fd_var(V2,Sz2,LWF2)|T]
     ;   Res= [fd_var(V2,Sz2,LWF2)|RR], insert_fd_var(T,V2,Sz1,LWF1,RR)
   ). */




my_get_fdvars_atts(Store,FDVARS) :-
   (get_atts(Store,+wf_fdvars(FDVARS)) -> true
	  ;  var(Store),FDVARS=[]
   ).
	   /* store not initialised; initialise it below but print msg */
	  %   ,waitflag_not_init).
	     
my_get_next_fdvar_to_enumerate([fd_var(V1,LWF1)|FDVARS],NextFDVar,FDPrio,LWF,RemainingFDVARS) :-
   get_fd_priority(V1,Prio1,FrozenSize),
   get_next_fdvar(FDVARS, V1,Prio1,FrozenSize,LWF1, NextFDVar,FDPrio,LWF,RemainingFDVARS),
   %(NextFDVar == V1 -> true ; print(changed_order( NextFDVar,FDPrio, RemainingFDVARS)),nl),
   % print(enumerating(NextFDVar,FDPrio,RemainingFDVARS)),nl,
   %translate:print_clpfd_variable(NextFDVar),nl,nl,
   %translate:l_print_frozen_info(RemainingFDVARS),nl,trace,
   FDPrio \= sup. % sup can appear if posting of an X in A..B constraint overflowed

has_higher_fd_priority(FDVAR,NewPrio,NewFrozenSize, _CurFDVAR,sup,_CurFrozenSize) :- !,
   get_fd_priority(FDVAR,NewPrio,NewFrozenSize).
has_higher_fd_priority(FDVAR,NewPrio,NewFrozenSize, _CurFDVAR,CurBestPrio,CurFrozenSize) :-
  my_get_fd_size(FDVAR,Size), Size \= sup,
  (Size<2 -> NewPrio=Size, NewPrio<CurBestPrio, NewFrozenSize=1
        ; Size<CurBestPrio,
          size_of_attached_goals(FDVAR,NewFrozenSize),
          (NewFrozenSize=0 -> NewPrio is Size*4, NewPrio<CurBestPrio
           ; NewPrio = Size)
        ; Size=CurBestPrio,  size_of_attached_goals(FDVAR,NewFrozenSize),
          %print(size(Size,NewFrozenSize,CurFrozenSize)),nl,
          NewFrozenSize > CurFrozenSize, % has more constraints pending on it
          % print(more_constraints(FDVAR,NewPrio,NewFrozenSize,_CurFDVAR,CurBestPrio,CurFrozenSize)),nl, %
          % translate:print_clpfd_variable(FDVAR),nl, translate:print_clpfd_variable(_CurFDVAR),nl,
          NewPrio=Size
   ).
get_fd_priority(FDVAR,Prio,FrozenSize) :-
  my_get_fd_size(FDVAR,Size),
  (Size=sup -> Prio=sup,FrozenSize=0
   ; Size<2 -> Prio=Size,FrozenSize=1
        ; size_of_attached_goals(FDVAR,FrozenSize),
          (FrozenSize=0 -> Prio is Size*4
           ; Prio = Size)
  ).
get_next_fdvar([], V,S,_FS,LWF, V,S,LWF,[]).
%get_next_fdvar([H|T], V,S,_FS,LWF, V,S,LWF,Rest) :- Cnt<1,!, Rest=[H|T].
get_next_fdvar([fd_var(V2,LWF2)|T], V1,Prio1,FrozenSize1,LWF1, ResV,ResP,ResLWF,ResRest) :-
  % C1 is Cnt-1, % count maximum number of elements to look at;
   (ground(V2)
    -> LWF2=0, get_next_fdvar(T, V1,Prio1,FrozenSize1,LWF1, ResV,ResP,ResLWF,ResRest)
    ;  (has_higher_fd_priority(V2,Prio2,FrozenSize2,V1,Prio1,FrozenSize1)
        -> ResRest=[fd_var(V1,LWF1)|RR],
           get_next_fdvar(T, V2,Prio2,FrozenSize2,LWF2, ResV,ResP,ResLWF,RR)
        ;  ResRest=[fd_var(V2,LWF2)|RR],
           get_next_fdvar(T, V1,Prio1,FrozenSize1,LWF1, ResV,ResP,ResLWF,RR)
       )
   ).
my_get_fd_size(FDVAR,ResSize) :- fd_size(FDVAR,Size),!, ResSize=Size.
  %(Size=sup -> ResSize = 10000000 /* same as for inf */  ; ResSize=Size).
my_get_fd_size(FDVAR,Size) :- add_internal_error('Could not get fd_size: ',my_get_fd_size(FDVAR,Size)), Size=sup.

size_of_attached_goals(FDVAR,Res) :- var(FDVAR),fd_degree(FDVAR,D),!,
  frozen(FDVAR,Goal), size_of_frozen_goal2(Goal,FDVAR,Size), Res is D+Size.
size_of_attached_goals(_,0).
size_of_frozen_goal2((A,B),Var,Sz) :- !, size_of_frozen_goal2(A,Var,SA), size_of_frozen_goal2(B,Var,SB), Sz is SA+SB.
size_of_frozen_goal2(clpfd:_,_,Sz) :- !, Sz=0.
size_of_frozen_goal2(kernel_objects:G,V,R) :- !, size_of_frozen_goal3(G,V,R).
size_of_frozen_goal2(true,_,R) :- !,R=0.
size_of_frozen_goal2(_,_,1).

size_of_frozen_goal3(safe_less_than_equal(_,X,Y),_,R) :- (number(X);number(Y)), !,R=0.
size_of_frozen_goal3(call_enumerate_int(X,_,_,_),Var,R) :- Var==X,!,R=0.
size_of_frozen_goal3(true,_,R) :- !,R=0.
size_of_frozen_goal3(_,_,1).

/*
%size_of_attached_goals(V,S) :- !, S=1. % comment in to not compute size of attached goals
oldsize_of_attached_goals(V,Size) :- var(V), frozen(V,Goal), size_of_frozen_goal(Goal,V,Size).%   ,print(size(Goal,V,Size)),nl.
size_of_frozen_goal((A,B),Var,Sz) :- !, size_of_frozen_goal(A,Var,SA), size_of_frozen_goal(B,Var,SB), Sz is SA+SB.
size_of_frozen_goal(clpfd:in(X,'..'(_A,_B)),Var,R) :- Var==X,!,R=0. % useless
size_of_frozen_goal(clpfd:in(_X,_),_,R) :- !,R=2.
%clpfd:(_1759931 in{4}\/{6}
size_of_frozen_goal(kernel_objects:safe_less_than_equal(_,X,Y),_,R) :- (number(X);number(Y)), !,R=0.
size_of_frozen_goal(kernel_objects:call_enumerate_int(X,_,_,_),Var,R) :- Var==X,!,R=0.
size_of_frozen_goal(true,_,R) :- !,R=0.
size_of_frozen_goal(_,_,1).
*/

% get a wait-flag that will be triggered first next time;
% can be used to ensure that all pending co-routines complete
get_idle_wait_flag(Info,wfx(_WF0,Store,_EnumFinished),LWF) :- !,
    get_waitflag_from_store(1,Info,Store,LWF).


get_wait_flag(Prio,Store,WF) :- get_wait_flag(Prio,'??',Store,WF).
get_wait_flag(0,_Info,wfx(WF0,_Store,_EnumFinished),WF) :- !, WF=WF0.
get_wait_flag(1,Info,wfx(WF0,Store,_EnumFinished),WF) :- !,
   (var(WF0) -> get_waitflag_from_store(1,Info,Store,WF) ; WF=1).
get_wait_flag(Prio,Info,wfx(_,Store,EnumFinished),WF) :- 
   ground(EnumFinished) -> WF=Prio /* enumeration has finished: return a ground WF */
    ; get_waitflag_from_store(Prio,Info,Store,WF).
get_waitflag_from_store(Prio,Info,Store,WF) :-
   % print(get_waitflag_from_store(Prio,Info,Store,WF)),nl,
    (Prio=inf -> RealPrio = float(10000000)
       % , print(infinite_priority(Prio,Info)),nl %%
      ; RealPrio=float(Prio)),
	my_get_wf_store_atts(Store,Heap),
	(avl_fetch(RealPrio,Heap,wf(WF1,OldInfo))
	 -> (Info=label_clpfd_vars(Vars)
	      ->  merge_clpfd_vars(OldInfo,Vars,NewInfo),
	          % print(updating(RealPrio,NewInfo)),nl,
	          avl_store(RealPrio,Heap,wf(WF1,NewInfo),NewHeap),
	          put_atts(Store,+wf_store(NewHeap))
	       ;  true % NewHeap=Heap  %,print(reusing(RealPrio,WF1,_OldInfo,Info)),nl
	    )
	  ; avl_store(RealPrio,Heap,wf(WF1,Info),NewHeap),
	    put_atts(Store,+wf_store(NewHeap))
	),!,
	%avl_size(NewHeap,Sz),
	%(Sz=10 -> print(heap_size(10)),nl ; true),
	%(Sz=1000 -> print(heap_size(1000)),nl, portray_avl(NewHeap) ; true),
	WF=WF1.
get_waitflag_from_store(Prio,Info,Store,WF) :- 
    add_internal_error('Error getting Waitflag: ', get_waitflag_from_store(Prio,Info,Store,WF)),fail.


merge_clpfd_vars(label_clpfd_vars(OldVars),NewVars,label_clpfd_vars(Res)) :- !,
   append(NewVars,OldVars,Res).
merge_clpfd_vars(_OldInfo,NewVars,label_clpfd_vars(NewVars)).

my_get_non_empty_wf_store_atts(Store,Heap) :- get_atts(Store,+wf_store(Heap)).
my_get_wf_store_atts(Store,Heap) :-
   (get_atts(Store,+wf_store(Heap)) -> true
	  ;  var(Store),empty_avl(Heap)
	).
	%,  /* store not initialised; initialise it below but print msg */
	%     waitflag_not_init).


ground_wait_flag0(wfx(WF0,_,_)) :- grd_wf0(WF0).

ground_wait_flags(wfx(WF0,Store,EnumFinish)) :-  !,
    grd_wf0(WF0), 
    ground_waitflags_store(Store),
    grd_ef(EnumFinish).
ground_wait_flags(W) :-
    add_internal_error('Waitflags in wrong format: ',ground_wait_flags(W)).

grd_wf0(E) :- var(E) -> E=0 %%,print(grd_wf0_done),nl
                     ; (E==0 -> true ; add_internal_error('Illegal WF0 Waitflag value: ',grd_wf0(E))).
grd_ef(E) :- var(E) -> E=0
             ; ((E==0;E==wd_guarded) -> true ; add_internal_error('Illegal EF Waitflag value: ',grd_ef(E))).

/* currently unused: 
create_wdguarded_wait_flags(wfx(WF0,Store,EnumFinish),Res,Expected,
                            wfx(WF0,Store,LocalEnumFinish)) :-
  %% if Res=Expected then the computations of the inner waitflags should be well-defined;
  %%   otherwise they may not 
    copy_wfe(EnumFinish,Res,Expected,LocalEnumFinish).
:- block copy_wfe(-,?,?,?), copy_wfe(?,-,?,?).
copy_wfe(EnumFinish,Res,Expected,LocalEnumFinish) :-
  (Res==Expected -> LocalEnumFinish=EnumFinish
   ; LocalEnumFinish=wd_guarded %% branch is pruned by well-definedness condition 
   ).
*/

ground_constraintprop_wait_flags(wfx(WF0,Store,_WFE)) :- 
  %% print(ground_constraintprop_wait_flags(WF0,Store,_WFE)),nl, %%
  !,grd_wf0(WF0),% print(finished_grounding_wf0),nl,portray_waitflags_store(Store),
  ground_waitflags_store(Store).
ground_constraintprop_wait_flags(W) :-
   add_internal_error('Waitflags in wrong format',ground_constraintprop_wait_flags(W)).

:- use_module(clpfd_interface,[clpfd_reverse_labeling/1]).

ground_waitflags_store(Store) :-  % portray_waitflags_store(Store), %
  (preferences:preference(use_smt_mode,true)
    -> ground_waitflags_store_clpfd(Store)
    ;  ground_waitflags_store_normal(Store)).

ground_waitflags_store_normal(Store) :-
			my_get_non_empty_wf_store_atts(Store,Heap),!, % portray_waitflags_store(Store),
		 %% portray_avl(Heap), %%
			( avl_del_min(Heap,Prio,wf(WF,Info),NewHeap) ->          % we have non-ground waitflags with a matching priority
				put_atts(Store,+wf_store(NewHeap)),   % write the modified heap back
        %% spy_waitflags(grounding(Prio,Info),Store), %%
				ground_prio(Prio,WF,Info),
				ground_waitflags_store_normal(Store)         % continue recursively
			; otherwise ->                % heap is empty, nothing more to do
			   %% print(finished_grounding_waitflags_store),nl, %%
			    true).
ground_waitflags_store_normal(_Store). % no wf_store attribute -> empty heap

ground_prio(Prio,WF,Info) :-
		        (Info = label_clpfd_vars(Variables)
		         ->  % print(labeling(WF,Prio,Variables)),nl, translate:l_print_frozen_info(Variables),nl,
		             clpfd_reverse_labeling(Variables) %, print(labeled(Variables)),nl
		           ; true),
		        % print(ground(Prio,WF,Info)),nl, translate:print_frozen_info(WF),nl,
				WF = Prio.                              % ground waitflag
	            % print(grounded(Prio,Info)),nl,nl, %%

% mix ProB labeling with CLP(FD) labeling
ground_waitflags_store_clpfd(Store) :-
			my_get_wf_store_atts(Store,Heap),
			my_get_fdvars_atts(Store,FDVars),!, 
			( avl_del_min(Heap,Prio,wf(WF,_Info),NewHeap) ->          % we have non-ground waitflags with a matching priority
			 ((my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,FDPrio,LWF,RemainingFDVARS),
			   FDPrio=<Prio)
			   -> enum_fd_variable_store(Store,NextFDVar,FDPrio,LWF,RemainingFDVARS)
			   ;  put_atts(Store,+wf_store(NewHeap)),   % write the modified heap back
			      % print(grounding(Prio,_Info)),nl,
				  WF = Prio,                            % ground waitflag
			      ground_waitflags_store_clpfd(Store)         % continue recursively
	          )
			; otherwise ->
	           (my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,FDPrio,LWF,RemainingFDVARS)
			      -> enum_fd_variable_store(Store,NextFDVar,FDPrio,LWF,RemainingFDVARS)
			      ; true  % heap and FDVars list is empty, nothing more to do
			    )
			).
ground_waitflags_store_clpfd(Store) :- 
   add_internal_error('Illegal Waitflag: ',ground_waitflags_store_clpfd(Store)),fail.

enum_fd_variable_store(Store,NextFDVar,FDPrio,LWF,RemainingFDVARS) :-
   put_atts(Store,+wf_fdvars(RemainingFDVARS)),
   % print(labeling(NextFDVar,FDPrio,LWF,RemainingFDVARS)),nl, %
   labeling([],[NextFDVar]), LWF=FDPrio, ground_waitflags_store_clpfd(Store).

:- if(1=0).
% small utility to trace through waitflag enumeration
:- dynamic spying_on/1.
spying_on(1).
spy_waitflags(_,_) :- \+ spying_on(_) , !.
spy_waitflags(Msg,Store) :- retract(spying_on(Nr)),print(Msg),nl,
   (Nr=1 
    -> assert(spying_on(Nr)),
       print(' (j,t,p,q) ==> '),read_term(T,[]),
       action(T,Store)
    ;  N1 is Nr-1, assert(spying_on(N1))
   ).
action(t,_Store) :- !,trace.
action(p,Store) :- !,portray_waitflags_store(Store), spy_waitflags('',Store).
action(q,_Store) :- !,retractall(spying_on).
action(Nr,_) :- number(Nr),!, retractall(spying_on(_)), assert(spying_on(Nr)).
action(_,_).
:- endif.

:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF), kernel_waitflags:portray_waitflags(WF),
      kernel_waitflags:get_wait_flag(2,'Test',WF,WF2), when(nonvar(WF2),print(wf2)),
      clpfd:in(FD,'..'(1,2)),kernel_waitflags:add_fd_variable_for_labeling(FD,WF),
      kernel_waitflags:portray_waitflags(WF), kernel_waitflags:ground_wait_flags(WF)  )).
% PORTRAYING Waitflags
portray_waitflags(wfx(WF0,Store,WFE)) :- flush_output,
            print('*WAITFLAG STORE: '),print(wfx(WF0,Store,WFE)),nl,
            (var(WF0) -> print('> WF0 :: '), portray_attached_goals(WF0),nl ; true),
            portray_waitflags_store(Store),!,
            print('*END WAITFLAG STORE'),nl.
			%my_portray_avl(Heap),nl.
portray_waitflags(Store) :- 
   add_internal_error('Illegal Waitflag: ',portray_waitflags(Store)).

portray_fd_vars([]) :- !, '$NOT_COVERED'('This is debugging code!').
portray_fd_vars([fd_var(V,_)|T]) :- print('> '), print(V), nonvar(V),!, nl,
   portray_fd_vars(T), '$NOT_COVERED'('This is debugging code!').
portray_fd_vars([fd_var(V,_)|T]) :- my_get_fd_size(V,Size), !, print(' : '), print(Size),
   portray_attached_goals(V),
   portray_fd_vars(T),
   '$NOT_COVERED'('This is debugging code!').
portray_fd_vars(X) :- print('> *** UNKNOWN FDVARS: '), print(X), nl,
   '$NOT_COVERED'('This is debugging code!').

portray_waitflags_store(Store) :-
            my_get_fdvars_atts(Store,FDList), 
            (FDList=[] -> true ; print('FDVARS:'),nl,portray_fd_vars(FDList),nl),
			my_get_wf_store_atts(Store,Heap),
			avl_to_list(Heap,List), portray_avl_els(List).

portray_avl_els([]).
portray_avl_els([Prio-wf(LWF,Info)|T]) :-
  print('> '),print(Prio), print(' : '), print(LWF), print(' : '), portray_info(Info),nl,
  portray_attached_goals(LWF),
  portray_avl_els(T).
  
portray_info(label_clpfd_vars(Vars)) :- print('label_clpfd_vars :: '), translate:l_print_frozen_info(Vars).
portray_info(I) :- print(I).

portray_attached_goals(LWF) :- frozen(LWF,Goal), print('   :: '), portray_goal(Goal),nl.

:- use_module(tools_printing,[print_term_summary/1]).
portray_goal((A,B)) :- !, portray_goal(A), print(','), print('     '), portray_goal(B).
portray_goal(A) :- print_term_summary(A).

:- public my_portray_avl/1.
my_portray_avl(V) :- var(V), !, add_internal_error('Variable: ',my_portray_avl(V)).
my_portray_avl(V) :- portray_avl(V).

:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF), 
 kernel_waitflags:get_wait_flag1(WF,WF1), when(ground(WF1),X=a),
 kernel_waitflags:ground_wait_flags(WF),  X==a )).

 % create a copy of the waitflags with WF0 and WF0 as variable; should be finished with copy_wf_finish
 % can be used if you temporarily want to disable non-deterministic enumeration, ...
 copy_wf_start(wfx(_,WF,_),wfx(_,WF,_)).
 % just copy WF0 and WFE over to new waitflag
 copy_wf_finish(wfx(WF0,_,WFE),wfx(WF0,_,WFE)).
 

/* create a copy where the enumeration_finished waitflag is shared;
  the inner enumeration_finished wait flags should normally not be grounded */
%create_inner_wait_flags(wfx(_,_,WFE),wfx(WF0Inner,S,WFE)) :- !,init_wait_flags_store(S). % old version
create_inner_wait_flags(wfx(_,_,WFE),wfx(WF0Inner,S,WFEInner)) :- init_wait_flags_store(S),
   get_inner_enumeration_over_wait_flag(wfx(WF0Inner,S,WFEInner),LastWF),
   copy_wfe_to_inner(LastWF,WFE,WFEInner). % delay copying WFE at least until inner enumeration has been run
:- block copy_wfe_to_inner(-,?,?).
copy_wfe_to_inner(_,WFE,WFE).

ground_inner_wait_flags(wfx(WF0,Store,_EWF)) :- !,
 % does not ground enumeration finished waitflag
   grd_wf0(WF0), 
   ground_waitflags_store(Store).
ground_inner_wait_flags(W) :-
   add_internal_error('Inner waitflags in wrong format: ',W).

% copy WF0 and WFE, set up a new store for the rest
copy_wf0e_wait_flags(wfx(WF0,_,WFE),wfx(WF0,S,WFE)) :- init_wait_flags_store(S).


clone_wait_flags_from1(wfx(_WF0,S,WFE),wfx(_NewWF0,S,WFE)).
% creates a copy of the wait_flags, except for WF0; this needs to be grounded separately with grd_wf0/1


get_wait_flag0(wfx(WF0,_,_),WF0).
%get_wait_flag0(WFX,WF0) :-   get_wait_flag(0,WFX,WF0).  /* when this is ground you can do det. propagatons */
get_wait_flag1(WFX,WF1) :-   get_wait_flag(1,WFX,WF1). /* when this is ground you can do det. propagatons */

get_binary_choice_wait_flag(Info,WFX,WF2) :- % use 3 rather than 2 to give priority over enumeration choicepoints
  get_wait_flag(4,Info,WFX,WF2). /* when this is ground you can do binary propagatons */

get_middle_wait_flag(Info,WFX,WFn) :- get_wait_flag(10,Info,WFX,WFn).

get_last_wait_flag(Info,WFX,Res) :- get_wait_flag(999998,Info,WFX,Res).


get_enumeration_starting_wait_flag(WFX,WF2) :-   get_wait_flag(999999,WFX,WF2).
get_inner_enumeration_over_wait_flag(WFX,WF2) :-   get_wait_flag(1999999,inner_enumeration,WFX,WF2).

get_enumeration_finished_wait_flag(wfx(_,_,WFE),WFE) :- !.
get_enumeration_finished_wait_flag(W,E) :- 
   add_internal_error('Waitflags in wrong format: ',get_enumeration_finished_wait_flag(W,E)).
   

  
ground_det_wait_flag(wfx(WF0,Store,_)) :- %% print(ground_det_wait_flag),nl,
     grd_wf0(WF0),ground_waitflags_store_up_to(1,Store).

ground_wait_flag_to(wfx(WF0,Store,_),Limit) :- %% print(ground_det_wait_flag),nl,
  % TO DO: deal with label_clpfd_vars Info !!
     grd_wf0(WF0),ground_waitflags_store_up_to(Limit,Store).
ground_waitflags_store_up_to(Prio,Store) :- % print(prio(Prio)),nl,
	my_get_wf_store_atts(Store,Heap),  % portray_waitflags_store(Store), %
	( avl_min(Heap,Min,wf(WF,Info)) -> % heap is non-empty, first element is WF with priority Min
	    ( Min =< Prio ->          % we have non-ground waitflags with a matching priority
		((avl_del_min(Heap,AMin,_,NewHeap),AMin==Min)     % remove the first element
		 -> true ; add_internal_error('Could not remove min. in ground_waitflags_store_up_to: ',Min:Info:Heap), fail
		),
		put_atts(Store,+wf_store(NewHeap)),   % write the modified heap back
			 	%%  print(grounding2(Prio,_Info)),nl, %%
		ground_prio(Min,WF,Info),
		ground_waitflags_store_up_to(Prio,Store)         % continue recursively
	    ; otherwise ->            % no more matching waitflags present
		true)                 % stop here
	; otherwise ->                % heap is empty, nothing more to do
	    true).                    % stop here


update_waitflag(Prio,CurrentWaitflag,NewWaitFlag,WF) :-
  /* if the CurrentWaitflag is already ground and now a new WF with lower priority has been added to the store:
      create a new non-ground waitflag to give priority to new WF with lower priority value */
   (var(CurrentWaitflag) -> NewWaitFlag=CurrentWaitflag
     ; WF=wfx(_WF0,Store,_),my_get_wf_store_atts(Store,Heap),
       ((avl_min(Heap,Min,wf(_NewerWF,_Info)),(Prio=inf;Min<Prio))
         -> get_waitflag_from_store(Prio,updated_wf,Store,NewWaitFlag)  % could be optimized
           % ,print(updating_wf(Prio,CurrentWaitflag,NewWaitFlag,Min)),nl
         ;  NewWaitFlag=CurrentWaitflag
       )
   ).


% currently not used:
%/* same as get_enumeration_finished_wait_flag, but will fail if user does not want
%   to look for abort errors */
%get_abort_error_flag(wfx(_,_,WFE),WFE) :- !, preferences:preference(find_abort_values,true).
%get_abort_error_flag(W,_) :- 
%   add_internal_error(get_abor_error_flag,'Waitflags in wrong format: ',W).
    

add_wd_error(Msg,Term,WF) :- add_wd_error_span(Msg,Term,unknown,WF).
add_wd_error_set_result(Msg,Term,Result,ResultValue,Span,WF) :-
   add_abort_error7(well_definedness_error,Msg,Term,Result,ResultValue,Span,WF).
add_wd_error_span(Msg,Term,Span,WF) :- add_abort_error_span(well_definedness_error,Msg,Term,Span,WF).

add_abort_error_span(ErrType,Msg,Term,Span,WF) :- add_abort_error7(ErrType,Msg,Term,0,0,Span,WF).

add_abort_error7(ErrType,Msg,Term,Result,ResultValue,Span,WF) :-
    get_enumeration_finished_wait_flag(WF,AWF), 
    add_abort_error2(AWF,ErrType,Msg,Term,Result,ResultValue,Span).

% no longer used:
%add_wd_error_span_now(Msg,Term,Span) :- add_abort_error2(true,well_definedness_error,Msg,Term,0,0,Span).

:- use_module(state_space,
  [store_abort_error_for_context_state_if_possible/4]).
:- block add_abort_error2(-,?,?,?,?,?,?).
add_abort_error2(wd_guarded,_Err,_Msg,_Term,Result,ResultValue,_Span) :- !,% ignoring error message; no-wd problem
  % set Result to default ResultValue; useful to get rid of pending co-routines; result will not be used
  (Result=ResultValue -> true ; print(add_abort_error_failure(Result,ResultValue)),nl).
add_abort_error2(_AWF,ErrType,Msg,Term,_,_,Span) :-  % print(adding_abort_error(Msg)),nl,trace,
  % add_error(abort_error,Msg,Term),
    (store_abort_error_for_context_state_if_possible(ErrType,Msg,Term,Span)
      -> true ; add_error(add_abort_error2,'Could not store error: ',(Msg,Term),Span)),
    % print(done),nl,
    fail.

 :- assert_must_succeed(( kernel_waitflags:test_waitflags(1) )).
 :- assert_must_succeed(( kernel_waitflags:test_waitflags(2) )).
 
test_waitflags(1) :-
	init_wait_flags(Waitflags),
	get_wait_flag(10,Waitflags,WF1),
	when(nonvar(WF1),print('WF1 - Prio 10\n')),
	get_wait_flag(5,Waitflags,WF2),
	when(nonvar(WF2),(print('WF2 - Prio 5\n'),
	       get_wait_flag(6,Waitflags,WF4),
         	when(nonvar(WF4),print('WF4 - Prio 6\n')),
	       get_wait_flag(2,Waitflags,WF5),
	      when(nonvar(WF5),print('WF5 - Prio2\n')))),
        get_wait_flag(5,Waitflags,WF3),
	when(nonvar(WF3),print('WF3 - Prio 5\n')),
	ground_wait_flags(Waitflags),nl.
	
test_waitflags(2) :-
  init_wait_flags(Waitflags),
  get_wait_flag(10,Waitflags,WF1),
  when(nonvar(WF1),print('WF1 Prio 10\n')),
  get_wait_flag(5,Waitflags,WF2),
   when(nonvar(WF2),(print('WF2 - Prio 5\n'),
   get_wait_flag(6,Waitflags,WF4),
 	when(nonvar(WF4),print('WF4 - Prio 6\n')),
   get_wait_flag(2,Waitflags,WF5),
  when(nonvar(WF5),print('WF5 - Prio2\n')))),
  get_wait_flag(5,Waitflags,WF3),
  when(nonvar(WF3),print('WF3 - Prio 5\n')),
  ground_wait_flags(Waitflags),nl.

  