% (c) 2009-2010 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

/* ---------------------------------------- */
/* (c) Michael Leuschel, 2001-2011          */
/* All rights reserved.                     */
/* This code may not be distributed without */
/* written permission by Michael Leuschel   */
/* ---------------------------------------- */

:- module(kernel_waitflags,
   [init_wait_flags/1, waitflag0_is_set/1,
    get_wait_flag/3, get_wait_flag/4,
    get_wait_flag0/2, get_wait_flag1/2, 
    get_first_wait_flag/2, get_middle_wait_flag/3, get_last_wait_flag/3,
    get_binary_choice_wait_flag/3,
    add_fd_variable_for_labeling/4,  add_fd_variable_for_labeling/2,
    get_new_subsidiary_wait_flag/4,
    update_waitflag/4,
    ground_constraintprop_wait_flags/1,
    check_is_wait_flag/1,
    create_inner_wait_flags/2, 
    clone_wait_flags_from1/2, ground_wait_flag0/1,
    create_wdguarded_wait_flags/4,
    get_abort_error_flag/2,
    get_enumeration_starting_wait_flag/2,
    get_enumeration_finished_wait_flag/2, 
    ground_det_wait_flag/1, 
    ground_wait_flags/1,
    
    add_abort_error/2,
    add_abort_error/3,
    assert_must_abort_wf/2,
    
    portray_waitflags/1,
    
    set_silent/1]).
/* File: kernel_waitflags.pl */
/* Created: 17/3/06 by Michael Leuschel */

/*
Phases of the ProB Kernel:

 -> all Waitflags unbound
 
* Phase 0:  propagation of completely known ground values in a deterministic fashion
 
 -> WF0 is bound
 
* Phase 1:  deterministic propagation (not necessarily of completely known full values)
 
 -> WF1 is bound
 
* Phase 2:  boolean propagation (e.g., P or Q will now try P or Q)
 
 -> WF2 is bound
 
* Phase 3:  non-deterministic propagation (e.g, x: {1,2,3})
 
 -> Enumeration Starting WF is bound
 
* Enumeration Phase:  (later to be split into enumeration of finite and infinite types (e.g., integer)
 
  -> Enumeration Finished Waitflag bound
 
* Error Generation Phase: errors for, e.g., division by zero, f(x) with x outside domain of f,... are generated
   
*/

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(typechecker).

:- type wait_flag +--> (var ; wfx(any,var,any)).


% the Call must succeed and generate one abort_error
:- meta_predicate assert_must_abort_wf(0,*).
assert_must_abort_wf(Call,WF) :- assert_must_succeed_any(assert_must_abort2_wf(Call,WF)).

assert_must_abort2_wf(_,_) :- get_error(well_definedness_error,M),
   add_error(assert_must_abort,'Abort error prior to call: ',M),fail.
assert_must_abort2_wf(Call,WF) :- (real_error_occurred -> RE=true ; RE=false),
  init_wait_flags(WF),
  call(Call), print(called(Call)),nl,
  (ground_wait_flags(WF)
  -> add_error(assert_must_abort_wf,'Call did not fail after grounding WF: ',Call),fail
  ;  (get_error(well_definedness_error,_), % we have 1 abort_error
       print(got_well_definedness_error),nl,
       (RE=false -> reset_real_error_occured ; true)
     )
  ).


check_is_wait_flag(WFX) :-
  WFX = wfx(_,Store,_)
    -> (get_atts(Store,+wf_store(_)) -> true ; add_error(check_is_wait_flag,'WF does not contain heap: ',WFX))
    ;  add_error(check_is_wait_flag,'Illegal WF Format: ',WFX).

verify_attributes(Var, Value, true ) :-
  nonvar(Value) -> print('Trying to unify waitflag_heap_store'),nl
                 ;  print(verify_attributes(Var,Value)),nl.

:- use_module(library(avl)).
:- use_module(library(atts)).

:- attribute wf_store/1,wf_fdvars/1.

init_wait_flags(wfx(_,Store,_)) :- init_wait_flags_store(Store).

init_wait_flags_store(Store) :-
	empty_avl(Heap),
	put_atts(Store,+wf_store(Heap)),
	put_atts(Store,+wf_fdvars([])).

waitflag0_is_set(wfx(WF0,_Store,_EF)) :- ground(WF0).

get_new_subsidiary_wait_flag(OldWF,Info,WFX,NewWF) :-
   (ground(OldWF) /* then we have created a choice point with the old WF: double new WF prio */
      -> NewPrio is OldWF*2, get_wait_flag(NewPrio,Info,WFX,NewWF)
      ;  NewWF = OldWF
   ).

:- dynamic silent/0.
set_silent(true) :- !,(silent -> true ; assert(silent)).
set_silent(false) :- retractall(silent).
waitflag_not_init :-  silent -> true
   ; print('waitflag-store not initialised'),nl.

:- use_module(library(clpfd),[fd_size/2]).
% adds a CLP(FD) variable with given domain Size, will be enumerated by labeling (even if clpfd_solver preference is false !)
% Size could be computed by fd_size(FDVariable,Size)
add_fd_variable_for_labeling(FDVariable,WF) :- add_fd_variable_for_labeling(FDVariable,1,WF,_LWF).

add_fd_variable_for_labeling(FDVariable,_,_WF,LWF) :- nonvar(FDVariable),!,LWF=1.
add_fd_variable_for_labeling(FDVariable,_Size,WF,LWF) :-  WF=wfx(_,Store,_EnumFinished),
  my_get_fdvars_atts(Store,FDVARS),
  %insert_fd_var(FDVARS,FDVariable,Size,LWF,NewFDVARS),
  NewFDVARS = [fd_var(FDVariable,LWF)|FDVARS],
  %  print(added_fd_variable(FDVariable,Size,NewFDVARS)),nl, %%
  put_atts(Store,+wf_fdvars(NewFDVARS)).

/*
insert_fd_var([],V,Sz,LWF,[fd_var(V,Sz,LWF)]).
insert_fd_var([fd_var(V2,Sz2,LWF2)|T],V1,Sz1,LWF1,Res) :-
   (Sz1 =< Sz2
     ->  Res= [fd_var(V1,Sz1,LWF1),fd_var(V2,Sz2,LWF2)|T]
     ;   Res= [fd_var(V2,Sz2,LWF2)|RR], insert_fd_var(T,V2,Sz1,LWF1,RR)
   ). */




my_get_fdvars_atts(Store,FDVARS) :-
   (get_atts(Store,+wf_fdvars(FDVARS)) -> true
	  ;  var(Store),FDVARS=[],  /* store not initialised; initialise it below but print msg */
	     waitflag_not_init).
	     
my_get_next_fdvar_to_enumerate([fd_var(V1,LWF1)|FDVARS],NextFDVar,FDPrio,LWF,RemainingFDVARS) :-
   get_fd_priority(V1,Prio1,FrozenSize),
   get_next_fdvar(FDVARS, V1,Prio1,FrozenSize,LWF1, NextFDVar,FDPrio,LWF,RemainingFDVARS),
   %(NextFDVar == V1 -> true ; print(changed_order( NextFDVar,FDPrio, RemainingFDVARS)),nl),
   FDPrio \= sup. % sup can appear if posting of an X in A..B constraint overflowed

has_higher_fd_priority(FDVAR,NewPrio,NewFrozenSize, _CurFDVAR,sup,_CurFrozenSize) :- !,
   get_fd_priority(FDVAR,NewPrio,NewFrozenSize).
has_higher_fd_priority(FDVAR,NewPrio,NewFrozenSize, _CurFDVAR,CurBestPrio,CurFrozenSize) :-
  my_get_fd_size(FDVAR,Size), Size \= sup,
  (Size<2 -> NewPrio=Size, NewPrio<CurBestPrio, NewFrozenSize=1
        ; Size<CurBestPrio,
          size_of_attached_goals(FDVAR,NewFrozenSize),
          (NewFrozenSize=0 -> NewPrio is Size*4, NewPrio<CurBestPrio
           ; NewPrio = Size)
        ; Size=CurBestPrio,  size_of_attached_goals(FDVAR,NewFrozenSize),
          %print(size(Size,NewFrozenSize,CurFrozenSize)),nl,
          NewFrozenSize > CurFrozenSize, % has more constraints pending on it
           print(more_constraints(FDVAR,NewPrio,NewFrozenSize,_CurFDVAR,CurBestPrio,CurFrozenSize)),nl,
          NewPrio=Size
   ).
get_fd_priority(FDVAR,Prio,FrozenSize) :-
  my_get_fd_size(FDVAR,Size),
  (Size=sup -> Prio=sup,FrozenSize=0
   ; Size<2 -> Prio=Size,FrozenSize=1
        ; size_of_attached_goals(FDVAR,FrozenSize),
          (FrozenSize=0 -> Prio is Size*4
           ; Prio = Size)
  ).
get_next_fdvar([], V,S,_FS,LWF, V,S,LWF,[]).
%get_next_fdvar([H|T], V,S,_FS,LWF, V,S,LWF,Rest) :- Cnt<1,!, Rest=[H|T].
get_next_fdvar([fd_var(V2,LWF2)|T], V1,Prio1,FrozenSize1,LWF1, ResV,ResP,ResLWF,ResRest) :-
  % C1 is Cnt-1, % count maximum number of elements to look at;
   (ground(V2)
    -> LWF2=0, get_next_fdvar(T, V1,Prio1,FrozenSize1,LWF1, ResV,ResP,ResLWF,ResRest)
    ;  (has_higher_fd_priority(V2,Prio2,FrozenSize2,V1,Prio1,FrozenSize1)
        -> ResRest=[fd_var(V1,LWF1)|RR],
           get_next_fdvar(T, V2,Prio2,FrozenSize2,LWF2, ResV,ResP,ResLWF,RR)
        ;  ResRest=[fd_var(V2,LWF2)|RR],
           get_next_fdvar(T, V1,Prio1,FrozenSize1,LWF1, ResV,ResP,ResLWF,RR)
       )
   ).
my_get_fd_size(FDVAR,ResSize) :- fd_size(FDVAR,Size),!, ResSize=Size.
  %(Size=sup -> ResSize = 10000000 /* same as for inf */  ; ResSize=Size).
my_get_fd_size(FDVAR,_) :- add_error_fail(my_get_fd_size,'Could not get fd_size: ',FDVAR).

%size_of_attached_goals(V,S) :- !, S=1. % comment in to not compute size of attached goals
size_of_attached_goals(V,Size) :- var(V), frozen(V,Goal), size_of_frozen_goal(Goal,V,Size).%   ,print(size(Goal,V,Size)),nl.
size_of_frozen_goal((A,B),Var,Sz) :- !, size_of_frozen_goal(A,Var,SA), size_of_frozen_goal(B,Var,SB), Sz is SA+SB.
size_of_frozen_goal(clpfd:in(X,'..'(_A,_B)),Var,R) :- Var==X,!,R=0. % useless
size_of_frozen_goal(clpfd:in(_X,_),_,R) :- !,R=2.
%clpfd:(_1759931 in{4}\/{6}
size_of_frozen_goal(kernel_objects:safe_less_than_equal(_,X,Y),_,R) :- (number(X);number(Y)), !,R=0.
size_of_frozen_goal(kernel_objects:call_enumerate_int(X,_,_,_),Var,R) :- Var==X,!,R=0.
size_of_frozen_goal(true,_,R) :- !,R=0.
size_of_frozen_goal(_,_,1).





get_wait_flag(Prio,Store,WF) :- get_wait_flag(Prio,'?',Store,WF).
get_wait_flag(0,_Info,wfx(WF0,_Store,_EnumFinished),WF) :- !, WF=WF0.   /* comment in to store WaitFlag 0 separately and only ground once; advantage is to keep size of the heap down; disadvantage is that ground calls may not get priority over wf0 calls once wf0 has been grounded */
get_wait_flag(1,Info,wfx(WF0,Store,_EnumFinished),WF) :- !,
  (var(WF0) -> get_waitflag_from_store(1,Info,Store,WF) ; WF=1).
get_wait_flag(Prio,Info,wfx(_,Store,EnumFinished),WF) :- 
  ground(EnumFinished) -> WF=Prio /* enumeration has finished: return a ground WF */
   ; get_waitflag_from_store(Prio,Info,Store,WF).
get_waitflag_from_store(Prio,Info,Store,WF) :-
   % print(get_waitflag_from_store(Prio,Info,Store,WF)),nl,
    (Prio=inf -> RealPrio = 10000000
       % , print(infinite_priority(Prio,Info)),nl %%
      ; RealPrio=Prio),
	my_get_wf_store_atts(Store,Heap),
	(avl_fetch(RealPrio,Heap,wf(WF1,_OldInfo))
	 -> (Info=label_clpfd_vars(Vars)
	      ->  merge_clpfd_vars(_OldInfo,Vars,NewInfo),
	          % print(updating(RealPrio,NewInfo)),nl,
	          avl_store(RealPrio,Heap,wf(WF1,NewInfo),NewHeap)
	       ;  NewHeap=Heap  %,print(reusing(RealPrio,WF1,_OldInfo,Info)),nl
	    )
	  ; avl_store(RealPrio,Heap,wf(WF1,Info),NewHeap)
	),
	put_atts(Store,+wf_store(NewHeap)),!,
	%avl_size(NewHeap,Sz),
	%(Sz=10 -> print(heap_size(10)),nl ; true),
	%(Sz=100 -> print(heap_size(100)),nl ; true),
	%(Sz=1000 -> print(heap_size(1000)),nl, portray_avl(NewHeap) ; true),
	WF=WF1.
get_waitflag_from_store(Prio,Info,Store,WF) :- 
   add_error(get_waitflag_from_store,'Error getting Waitflag: ', get_waitflag_from_store(Prio,Info,Store,WF)),fail.


merge_clpfd_vars(label_clpfd_vars(OldVars),NewVars,label_clpfd_vars(Res)) :- !,
   append(NewVars,OldVars,Res).
merge_clpfd_vars(_OldInfo,NewVars,label_clpfd_vars(NewVars)).


my_get_wf_store_atts(Store,Heap) :-
   (get_atts(Store,+wf_store(Heap)) -> true
	  ;  var(Store),empty_avl(Heap),  /* store not initialised; initialise it below but print msg */
	     waitflag_not_init).


ground_wait_flag0(wfx(WF0,_,_)) :- grd_wf0(WF0).

ground_wait_flags(wfx(WF0,Store,EnumFinish)) :-  
    %% print(ground_wait_flags(wfx(WF0,Store,EnumFinish))),nl, %%
    grd_wf0(WF0), 
    ground_waitflags_store(Store),
    grd_ef(EnumFinish).

grd_wf0(E) :- var(E) -> E=0 %%,print(grd_wf0_done),nl
                     ; (E==0 -> true ; add_error(grd_ef,'Illegal WF0 Waitflag value: ',E)).
grd_ef(E) :- var(E) -> E=0
             ; ((E==0;E==wd_guarded) -> true ; add_error(grd_ef,'Illegal EF Waitflag value: ',E)).


create_wdguarded_wait_flags(wfx(WF0,Store,EnumFinish),Res,Expected,
                            wfx(WF0,Store,LocalEnumFinish)) :-
  /* if Res=Expected then the computations of the inner waitflags should be well-defined;
     otherwise they may not */
    copy_wfe(EnumFinish,Res,Expected,LocalEnumFinish).
:- block copy_wfe(-,?,?,?), copy_wfe(?,-,?,?).
copy_wfe(EnumFinish,Res,Expected,LocalEnumFinish) :-
  (Res==Expected -> LocalEnumFinish=EnumFinish
   ; LocalEnumFinish=wd_guarded /* branch is pruned by well-definedness condition */).

ground_constraintprop_wait_flags(wfx(WF0,Store,_WFE)) :- 
  %% print(ground_constraintprop_wait_flags(WF0,Store,_WFE)),nl, %%
  !,grd_wf0(WF0),ground_waitflags_store(Store).
ground_constraintprop_wait_flags(W) :-
   add_error(ground_constraintprop_wait_flags,'Waitflags in wrong format',W).

:- use_module(clpfd_interface,[clpfd_reverse_labeling/1]).
:- use_module(library(clpfd),[labeling/2, fd_size/2]).
ground_waitflags_store(Store) :-
			my_get_wf_store_atts(Store,Heap),my_get_fdvars_atts(Store,FDVars),!, 
			%portray_waitflags_store(Store),
			( avl_del_min(Heap,Prio,wf(WF,_Info),NewHeap) ->          % we have non-ground waitflags with a matching priority
			 ((my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,FDPrio,LWF,RemainingFDVARS),
			   FDPrio=<Prio)
			   -> enum_fd_variable_store(Store,NextFDVar,FDPrio,LWF,RemainingFDVARS)
			   ;  put_atts(Store,+wf_store(NewHeap)),   % write the modified heap back
			      % print(grounding(Prio,_Info)),nl,
				  WF = Prio,                            % ground waitflag
			      ground_waitflags_store(Store)         % continue recursively
	          )
			; otherwise ->
	           (my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,FDPrio,LWF,RemainingFDVARS)
			      -> enum_fd_variable_store(Store,NextFDVar,FDPrio,LWF,RemainingFDVARS)
			      ; true)
			).
ground_waitflags_store(Store) :- 
   add_error(ground_waitflags_store,'Illegal Waitflag: ',ground_waitflags_store(Store)),fail.

enum_fd_variable_store(Store,NextFDVar,FDPrio,LWF,RemainingFDVARS) :-
   put_atts(Store,+wf_fdvars(RemainingFDVARS)),  % print(labeling(NextFDVar,FDPrio,LWF,RemainingFDVARS)),nl, %
   labeling([],[NextFDVar]), LWF=FDPrio, ground_waitflags_store(Store).

portray_waitflags(wfx(WF0,Store,WFE)) :- flush_output,
            print(wfx(WF0,Store,WFE)),nl,
            my_get_wf_store_atts(Store,FDList), print(wf_fdvars(FDList)),nl,
            portray_waitflags_store(Store),!.
			%portray_avl(Heap),nl.
portray_waitflags(Store) :- 
   add_error(portray_waitflags,'Illegal Waitflag: ',portray_waitflags(Store)).

portray_waitflags_store(Store) :-
			my_get_wf_store_atts(Store,Heap),!,
			avl_to_list(Heap,List), portray_avl_els(List).

portray_avl_els([]).
portray_avl_els([Prio-wf(LWF,Info)|T]) :-
  print('> '),print(Prio), print(' : '), print(LWF), print(' : '), print(Info),nl,
  frozen(LWF,Goal), print('   :: '), portray_goal(Goal),nl,
  portray_avl_els(T).

:- use_module(tools_printing,[print_term_summary/1]).
portray_goal((A,B)) :- !, portray_goal(A), print(','), print('     '), portray_goal(B).
portray_goal(A) :- print_term_summary(A).

:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF), 
 kernel_waitflags:get_wait_flag1(WF,WF1), when(ground(WF1),X=a),
 kernel_waitflags:ground_wait_flags(WF),  X==a )).

 
 
 

create_inner_wait_flags(wfx(_,_,WFE),wfx(_,S,WFE)) :- init_wait_flags_store(S).
/* create a copy where the enumeration_finished waitflag is shared;
  the inner enumeration_finished wait flags should normally not be grounded */


clone_wait_flags_from1(wfx(_WF0,S,WFE),wfx(_NewWF0,S,WFE)).
% creates a copy of the wait_flags, except for WF0; this needs to be grounded separately with grd_wf0/1


get_wait_flag0(wfx(WF0,_,_),WF0).
%get_wait_flag0(WFX,WF0) :-   get_wait_flag(0,WFX,WF0).  /* when this is ground you can do det. propagatons */
get_wait_flag1(WFX,WF1) :-   get_wait_flag(1,WFX,WF1). /* when this is ground you can do det. propagatons */

get_binary_choice_wait_flag(Info,WFX,WF2) :- % use 3 rather than 2 to give priority over enumeration choicepoints
  get_wait_flag(4,Info,WFX,WF2). /* when this is ground you can do binary propagatons */

get_middle_wait_flag(Info,WFX,WFn) :- get_wait_flag(10,Info,WFX,WFn).
  
get_first_wait_flag(WFX,Res) :- get_wait_flag(0,WFX,Res).
get_last_wait_flag(Info,WFX,Res) :- get_wait_flag(999998,Info,WFX,Res).


get_enumeration_starting_wait_flag(WFX,WF2) :-   get_wait_flag(999999,WFX,WF2).

get_enumeration_finished_wait_flag(wfx(_,_,WFE),WFE) :- !.
get_enumeration_finished_wait_flag(W,_) :- 
   add_error(get_enumeration_finished_wait_flag,'Waitflags in wrong format: ',W).
   

  
ground_det_wait_flag(wfx(WF0,Store,_)) :- %% print(ground_det_wait_flag),nl,
     grd_wf0(WF0),ground_waitflags_store_up_to(1,Store).

ground_waitflags_store_up_to(Prio,Store) :-
	my_get_wf_store_atts(Store,Heap),  %portray_avl(Heap), %
	( avl_min(Heap,Min,wf(WF,_Info)) -> % heap is non-empty, first element is WF with priority Min
	    ( Min =< Prio ->          % we have non-ground waitflags with a matching priority
		((avl_del_min(Heap,_Min,_,NewHeap),_Min==Min)     % remove the first element
		 -> true ; add_error(ground_waitflags_store_up_to,'Could not remove min: ',Min:_Info:Heap), fail
		),
		put_atts(Store,+wf_store(NewHeap)),   % write the modified heap back
			 	%% print(grounding2(Prio,_Info)),nl, %%
		WF = Min,                              % ground waitflag
		ground_waitflags_store_up_to(Prio,Store)         % continue recursively
	    ; otherwise ->            % no more matching waitflags present
		true)                 % stop here
	; otherwise ->                % heap is empty, nothing more to do
	    true).                    % stop here


update_waitflag(Prio,CurrentWaitflag,NewWaitFlag,WF) :-
  /* if the CurrentWaitflag is already ground and now a new WF with lower priority has been added to the store:
      create a new non-ground waitflag to give priority to new WF with lower priority value */
   (var(CurrentWaitflag) -> NewWaitFlag=CurrentWaitflag
     ; WF=wfx(_WF0,Store,_),my_get_wf_store_atts(Store,Heap),
       ((avl_min(Heap,Min,wf(_NewerWF,_Info)),(Prio=inf;Min<Prio))
         -> get_waitflag_from_store(Prio,updated_wf,Store,NewWaitFlag)  % could be optimized
           % ,print(updating_wf(Prio,CurrentWaitflag,NewWaitFlag,Min)),nl
         ;  NewWaitFlag=CurrentWaitflag
       )
   ).


:- use_module(preferences,[preference/2]).
/* same as get_enumeration_finished_wait_flag, but will fail if user does not want
   to look for abort errors */
get_abort_error_flag(wfx(_,_,WFE),WFE) :- !, preferences:preference(find_abort_values,true).
get_abort_error_flag(W,_) :- 
   add_error(get_abor_error_flag,'Waitflags in wrong format: ',W).
    
   
add_abort_error(Msg,WF) :-  add_abort_error(Msg,'',WF).
add_abort_error(Msg,Term,WF) :- add_abort_error(Msg,Term,0,0,WF).
add_abort_error(Msg,Term,Result,ResultValue,WF) :-
    get_enumeration_finished_wait_flag(WF,AWF), 
    %% print(trying_add_error(Msg,Term,AWF)),nl,
    add_abort_error2(AWF,Msg,Term,Result,ResultValue).

:- use_module(state_space,
  [store_abort_error_for_context_state_if_possible/2]).
:- block add_abort_error2(-,?,?,?,?).
add_abort_error2(wd_guarded,_Msg,_Term,Result,ResultValue) :- !,% ignoring error message; no-wd problem
  % set Result to ResultValue; useful to get rid of pending co-routines
  (Result=ResultValue -> true ; print(add_abort_error_failure(Result,ResultValue)),nl).
add_abort_error2(_AWF,Msg,Term,_,_) :- %%  print(adding_abort_error(Msg)),nl,
  % add_error(abort_error,Msg,Term),
    (store_abort_error_for_context_state_if_possible(Msg,Term)
      -> true ; add_error(add_abort_error2,'Could not store error: ',(Msg,Term))),
   %% print(done),nl,
    fail.

  
 
 
 
test :-
	init_wait_flags(Waitflags),
	get_waitflag(10,Waitflags,WF1),
	when(nonvar(WF1),print('WF1 - Prio 10\n')),
	get_waitflag(5,Waitflags,WF2),
	when(nonvar(WF2),(print('WF2 - Prio 5\n'),
	       get_waitflag(6,Waitflags,WF4),
         	when(nonvar(WF4),print('WF4 - Prio 6\n')),
	       get_waitflag(2,Waitflags,WF5),
	      when(nonvar(WF5),print('WF5 - Prio2\n')))),
	get_waitflag(5,Waitflags,WF3),
	when(nonvar(WF3),print('WF3 - Prio 5\n')),
	ground_wait_flags(7,Waitflags).
	
test2 :-
init_wait_flags(Waitflags),
get_waitflag(10,Waitflags,WF1),
when(nonvar(WF1),print('WF1 Prio 10\n')),
get_waitflag(5,Waitflags,WF2),
when(nonvar(WF2),(print('WF2 - Prio 5\n'),
   get_waitflag(6,Waitflags,WF4),
 	when(nonvar(WF4),print('WF4 - Prio 6\n')),
   get_waitflag(2,Waitflags,WF5),
  when(nonvar(WF5),print('WF5 - Prio2\n')))),
get_waitflag(5,Waitflags,WF3),
when(nonvar(WF3),print('WF3 - Prio 5\n')),
ground_wait_flags(Waitflags).

  