:- module(kodkod2,
          [ apply_rewriting_rules/2
          , extract_used_types/4
          , create_atoms_for_types/6
          , create_identifier_relations/6
          , trigger_integer_decisions/1
          ]).

:- use_module(library(lists)).
:- use_module(library(avl)).
:- use_module(library(ordsets)).

:- use_module(kodkod_tools).
:- use_module(kodkod_integer_recalc).

:- use_module(probsrc(tools),[ajoin/2]).
:- use_module(probsrc(module_information),[module_info/2]).
:- use_module(probsrc(error_manager)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(bmachine)).
:- use_module(probsrc(translate), [pretty_type/2,translate_bexpression/2]).
:- use_module(probsrc(b_interpreter),[b_compute_expression/4]).
:- use_module(probsrc(b_global_sets),[all_elements_of_type/2,is_b_global_constant/3]).
:- use_module(probsrc(store),[empty_state/1]).
:- use_module(probsrc(b_interpreter),[b_compute_expression/4]).

:- module_info(group,kodkod).
:- module_info(description,'This module contains the several phases of a translation process to Kodkod.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Phase 1
% applying re-writing rules to make some statements translatable to Kodkod

apply_rewriting_rules(TExpr,NTExpr) :-
    remove_bt(TExpr,Expr,NExpr,NTExpr),
    get_texpr_info(TExpr,Info),
    apply_rewriting_rules_pre(Expr,Info,IExpr),
    syntaxtransformation(IExpr,Subs,_,NSubs,Expr1),
    apply_rewriting_rules_l(Subs,NSubs),
    apply_rewriting_rules_post(Expr1,Info,NExpr).
apply_rewriting_rules_pre(Expr,Info,NExpr) :-
    kodkod_rewrite_pre(Expr,Info,Expr1),!,
    apply_rewriting_rules_pre(Expr1,Info,NExpr).
apply_rewriting_rules_pre(Expr,_,Expr).
apply_rewriting_rules_post(Expr,Info,NExpr) :-
    kodkod_rewrite_post(Expr,Info,Expr1),!,
    apply_rewriting_rules_post(Expr1,Info,NExpr).
apply_rewriting_rules_post(Expr,_,Expr).

apply_rewriting_rules_l([],[]).
apply_rewriting_rules_l([Expr|Erest],[NExpr|Nrest]) :-
    apply_rewriting_rules(Expr,NExpr),
    apply_rewriting_rules_l(Erest,Nrest).

kodkod_rewrite_pre(not_equal(A,B),_,negation(Equal)) :- !,
    create_kk_tpred(equal(A,B),Equal).
kodkod_rewrite_pre(member(Elem,Powset),_,Result) :-
    get_texpr_expr(Powset,PExpr),
    get_texpr_info(Elem,ElemInfo),
    kodkod_rewrite_membership(PExpr,Elem,ElemInfo,Result),
    !.
kodkod_rewrite_pre(subset(Subset,Powset),_,forall([Elem],TPre,TCond)) :-
    get_texpr_info(Subset,Info),
    get_texpr_type(Subset,set(ElemType)),
    unique_identifier(TId),
    element_analysis(Info,ElemInfo),
    create_texpr(identifier(TId),ElemType,ElemInfo,Elem),
    get_texpr_expr(Powset,PExpr),
    kodkod_rewrite_membership(PExpr,Elem,ElemInfo,Cond),!,
    create_texpr(member(Elem,Subset),pred,[],TPre),
    create_texpr(Cond,pred,[],TCond).
kodkod_rewrite_post(interval(A,B),Info,comprehension_set([TmpId],Predicate)) :- !,
    element_analysis(Info,ElemInfo),
    unique_identifier(TId),
    create_texpr(identifier(TId),integer,ElemInfo,TmpId),
    create_texpr(less_equal(A,TmpId),pred,[],Left),
    create_texpr(less_equal(TmpId,B),pred,[],Right),
    conjunct_predicates([Left,Right],Predicate).
kodkod_rewrite_post(sequence_extension(Elems),_Info,set_extension(Pairs)) :- !,
    convert_elements_to_pairs(Elems,1,Pairs).
kodkod_rewrite_post(general_sum([TId],Pred,Expr),_Info,general_sum([TId],NPred,Expr)) :-
    get_texpr_id(TId,Id), get_texpr_id(TElem,Id),
    extract_finite_integer_range(TId,[],IRange),
    \+ get_texpr_expr(Pred,member(TElem,_TSet)),!,
    write('rewritten\n'),
    create_texpr(comprehension_set([TId],Pred),set(integer),[analysis([elem(interval):IRange])],
                 TCompSet),
    create_texpr(member(TId,TCompSet),pred,[],NPred).
convert_elements_to_pairs([],_,[]).
convert_elements_to_pairs([Elem|Erest],Index,[Pair|Prest]) :-
    convert_element_to_pair(Elem,Index,Pair),
    I2 is Index + 1,
    convert_elements_to_pairs(Erest,I2,Prest).
convert_element_to_pair(Elem,Index,Pair) :-
    Range = irange(Index,Index),
    create_texpr(integer(Index),integer,[analysis([interval:Range])],TIndex),
    get_texpr_type(Elem,Type),
    get_texpr_info(Elem,EInfo),
    findall(right(Scope):Info,
            ( memberchk(analysis(AInfos),EInfo),member(Scope:Info,AInfos)),
            RInfos),
    create_texpr(couple(TIndex,Elem),couple(integer,Type),[analysis([left(interval):Range|RInfos])],Pair).

kodkod_rewrite_membership(pow_subset(Set),Elem,_,subset(Elem,Set)).
kodkod_rewrite_membership(fin_subset(Set),Elem,_,subset(Elem,Set)).
kodkod_rewrite_membership(relations(A,B),Elem,_,subset(Elem,Cartesian)) :-
    create_kk_tpred(cartesian_product(A,B),Cartesian).
kodkod_rewrite_membership(pow1_subset(Set),Elem,ElemInfo,conjunct(Subset,Exists)) :-
    get_texpr_type(Set,set(Type)),
    create_texpr(identifier('___elem'),Type,[kodkod|ElemInfo],Id),
    create_kk_tpred(subset(Elem,Set),Subset),
    create_kk_tpred(exists([Id],Member),Exists),
    create_kk_tpred(member(Id,Set),Member).
kodkod_rewrite_membership(fin1_subset(Set),Elem,Info,Pred) :-
    kodkod_rewrite_membership(pow1_subset(Set),Elem,Info,Pred).
kodkod_rewrite_membership(interval(A,B),Elem,_,conjunct(Lower,Upper)) :-
    create_kk_tpred(less_equal(A,Elem),Lower),
    create_kk_tpred(less_equal(Elem,B),Upper).
create_kk_tpred(Pred,TPred) :-
    create_texpr(Pred,pred,[kodkod],TPred).

element_analysis(SetInfo,[analysis(MemberInfo)]) :-
    findall( S:R,
             (   member(analysis(Analysis), SetInfo),
                 member(elem(S):R,Analysis)),
             MemberInfo).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Phase 2
% extract all used types
% for integers, check if bitstrings are necessary and store
%  the needed intervals

% extract_used_types(+TExpr,-UsedTypes,-IntegerRange,-BitstringRange):
%   extracts from a syntax tree the list of used types, the
%   ranges of integers and the ranges of integers that occur in sets
extract_used_types(TExpr,UsedTypes,IntegerRange,BitstringRange) :-
    % first adapt the output of the interval analysis to our needs
    % (see the comments in kodkod_integer_recalc for details)
    integer_recalc(TExpr,RExpr),
    % now extract the used types
    extract_used_types1(RExpr,UsedTypes,IntegerRange,BitstringRange).

extract_used_types1(TExpr,UsedTypes,IntegerRange,BitstringRange) :-
    syntaxtraversion(TExpr,Expr,Type,_,Subs1,_),
    ( higher_order_exception(Expr,Subs) ->
        UsedTypesA=[],IntRangeA=none,BitRangeA=none
    ; extract_type_information(Type,TExpr,UsedTypesA,IntRangeA,BitRangeA) ->
        Subs1 = Subs
    ; otherwise ->
        throw(kodkod(unsupported_type(Type)))),
    extract_used_types_l(Subs,UsedTypesA,IntRangeA,BitRangeA,
                         UsedTypes,IntegerRange,BitstringRange).

extract_used_types_l([],Types,IntRange,BitRange,Types,IntRange,BitRange).
extract_used_types_l([Expr|Erest],TypesI,IntRangeI,BitRangeI,Types,IntRange,BitRange) :-
    extract_used_types1(Expr,TypesA,IntRangeA,BitRangeA),
    merge_integer_info(IntRangeI,IntRangeA,IntRangeB),
    merge_integer_info(BitRangeI,BitRangeA,BitRangeB),
    ord_union(TypesI,TypesA,TypesB),
    extract_used_types_l(Erest,TypesB,IntRangeB,BitRangeB,Types,IntRange,BitRange).

extract_type_information(Type,TExpr,UsedTypesA,IntRangeA,BitRangeA) :-
    extract_type_information2(Type,TExpr,[],UsedTypesA,IntRangeA,BitRangeA).
extract_type_information2(T,_,_,[T],none,none) :-
    basic_type(T).
extract_type_information2(pred,_,_,[],none,none).
extract_type_information2(integer,TExpr,Scope,[],IntRange,none) :-
    extract_finite_integer_range(TExpr,Scope,IntRange).
extract_type_information2(set(Type),TExpr,Scope,UsedTypes,none,BitRange) :-
    extract_type_information_set(Type,TExpr,[elem|Scope],UsedTypes,BitRange).
extract_type_information2(seq(Type),TExpr,Scope,UsedTypes,IntRange,BitRange) :-
    extract_type_information2(set(couple(integer,Type)),TExpr,Scope,UsedTypes,IntRange,BitRange).
extract_type_information2(couple(A,B),TExpr,Scope,UsedTypes,IntRange,BitRange) :-
    extract_type_information2(A,TExpr,[left|Scope],TypesA,IntRangeA,BitRangeA),
    extract_type_information2(B,TExpr,[right|Scope],TypesB,IntRangeB,BitRangeB),
    merge_integer_info(IntRangeA,IntRangeB,IntRange),
    merge_integer_info(BitRangeA,BitRangeB,BitRange),
    ord_union(TypesA,TypesB,UsedTypes).

extract_type_information_set(T,_,_,[T],none) :-
    basic_type(T).
extract_type_information_set(integer,TExpr,Scope,[],BitRange) :-
    extract_finite_integer_range(TExpr,Scope,BitRange).
extract_type_information_set(couple(A,B),TExpr,Scope,UsedTypes,BitRange) :-
    extract_type_information_set(A,TExpr,[left|Scope],TypesA,BitRangeA),
    extract_type_information_set(B,TExpr,[right|Scope],TypesB,BitRangeB),
    merge_integer_info(BitRangeA,BitRangeB,BitRange),
    ord_union(TypesA,TypesB,UsedTypes).

basic_type(boolean).
basic_type(global(_)).

higher_order_exception(member(Elem,THOSet),[Elem|Subsets]) :-
    get_texpr_expr(THOSet,HOSet),
    higher_order_exception2(HOSet,Subsets).
higher_order_exception2(total_function(Dom,Ran),[Dom,Ran]).
higher_order_exception2(total_injection(Dom,Ran),[Dom,Ran]).
higher_order_exception2(total_surjection(Dom,Ran),[Dom,Ran]).
higher_order_exception2(total_bijection(Dom,Ran),[Dom,Ran]).
higher_order_exception2(partial_function(Dom,Ran),[Dom,Ran]).
higher_order_exception2(partial_injection(Dom,Ran),[Dom,Ran]).
higher_order_exception2(partial_surjection(Dom,Ran),[Dom,Ran]).
higher_order_exception2(partial_bijection(Dom,Ran),[Dom,Ran]).
higher_order_exception2(total_relation(Dom,Ran),[Dom,Ran]).
higher_order_exception2(surjection_relation(Dom,Ran),[Dom,Ran]).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Phase 3
% create the relations for the used types

% create_atoms_for_types(+Types,+IntegerRange,+BitstringRange,
%                        -Size,-Typemap,-Idmap,-Valuemap):
%   Types is a list of used types
%   IntegerRange is the range of used integers
%   IntsetRange is the range of used integers that occur in a set and
%     have each to be modelled as an atom.
%   Typemap is an AVL map from used types to relations
%   Idmap is an AVL map from some constants (including __true and __false
%      for true and false) to a relation
%   Valuemap in an AVL mapping from B values to Kodkod tuples
create_atoms_for_types(Types,IntegerRange,IntsetRange,Typemap,Idmap,Valuemap) :-
    empty_avl(EmptyMap),
    create_atoms_for_intsets(IntsetRange,EmptyMap,TM1),
    create_atoms_for_integers(IntegerRange,TM1,TM2),
    create_atoms_for_types2(Types,TM2,EmptyMap,EmptyMap,Typemap,Idmap,Valuemap).

create_atoms_for_intsets(none,Map,Map).
create_atoms_for_intsets(irange(A,B),MapIn,MapOut) :-
    Size is B-A+1,
    intset_relation_kodkod_name(Relname),
    add_type_relation(integer,Relname,Size,intsetrel(Relname,A,B),MapIn,MapOut).

create_atoms_for_integers(none,Map,Map).
create_atoms_for_integers(irange(A,B),MapIn,MapOut) :-
    pow2integer_relation_kodkod_name(Relname),
    add_type_relation('__pow2integer',Relname,unknown,pow2rel(Relname,A,B),MapIn,MapOut).

add_type_relation(Type,RelationName,Size,Relspec,In,Out) :-
    avl_store(Type,In,relation(RelationName,Size,Relspec),Out).
add_standard_relation(BType,KType,Id,Size,In,Out) :-
    create_relation_for_type(Id,KType,Relspec),
    avl_store(BType,In,relation(Id,Size,Relspec),Out).
add_global_type_relation(GType,RelationPrefix,Size,Id,In,Out) :-
    unique_identifier(RelationPrefix,GType,Id),
    add_standard_relation(global(GType),Id,Id,Size,In,Out).

create_atoms_for_types2([],TypeMap,IdMap,ValueMap,TypeMap,IdMap,ValueMap).
create_atoms_for_types2([Type|Trest],TMin,IDMin,VMin,TMout,IDMout,VMout) :-
    create_atoms_for_type(Type,TMin,IDMin,VMin,TM,IDM,VM),
    create_atoms_for_types2(Trest,TM,IDM,VM,TMout,IDMout,VMout).

create_atoms_for_type(boolean,TMIn,IdIn,VMIn,TMOut,IdOut,VMOut) :-
    add_standard_relation(boolean,bool,bool,2,TMIn,TMOut),
    lookup_constant_value(boolean_true,boolean,VTrue),
    lookup_constant_value(boolean_false,boolean,VFalse),
    assign_values_to_atoms([VTrue,VFalse],boolean,VMIn,VMOut),
    assign_elements_to_atoms(['__kodkod__true','__kodkod__false'],bool,IdIn,IdOut).
create_atoms_for_type(global(G),TMIn,IdIn,VMIn,TMOut,IdOut,VMOut) :-
    all_elements_of_type(G,BValues),
    length(BValues,Size),
    add_global_type_relation(G,g,Size,RelId,TMIn,TMOut),
    assign_values_to_atoms(BValues,global(G),VMIn,VMOut),
    findall(E, is_b_global_constant(G,_,E), Elements),
    assign_elements_to_atoms(Elements,RelId,IdIn,IdOut).

lookup_constant_value(Expr,Type,Value) :-
    create_texpr(Expr,Type,[],TExpr),
    empty_state(Empty),
    b_compute_expression(TExpr,Empty,Empty,Value).

assign_values_to_atoms(BValues,Type,VMin,VMout) :-
    assign_values_to_atoms2(BValues,Type,0,VMin,VMout).
assign_values_to_atoms2([],_Type,_Index,ValueMap,ValueMap).
assign_values_to_atoms2([BValue|BVrest],Type,Index,VMin,VMout) :-
    assign_value_to_atom(BValue,Type,Index,VMin,VM),
    Index2 is Index+1,
    assign_values_to_atoms2(BVrest,Type,Index2,VM,VMout).
assign_value_to_atom(BValue,Type,Index,VMin,VMout) :-
    avl_store(Type:BValue,VMin,Index,VMout).

assign_elements_to_atoms(Elements,Type,IdMapIn,IdMapOut) :-
    assign_elements_to_atoms2(Elements,Type,0,IdMapIn,IdMapOut).
assign_elements_to_atoms2([],_Type,_Index,IdMap,IdMap).
assign_elements_to_atoms2([Elem|Erest],Type,Index,IdMapIn,IdMapOut) :-
    assign_element_to_atom(Elem,Type,Index,IdMapIn,IdMap),
    Index2 is Index+1,
    assign_elements_to_atoms2(Erest,Type,Index2,IdMap,IdMapOut).
assign_element_to_atom(Elem,Type,Index,In,Out) :-
    unique_identifier(e,Elem,Id),
    create_element_relation(Id,Type,Index,Rel),
    avl_store(Elem,In,Rel,Out).

create_relation_for_type(Id,Type,Rel) :-
    create_generic_relation(Id,[Type],set,exact,full,Rel).
create_element_relation(Id,Type,Index,Rel) :-
    create_generic_relation(Id,[Type],set,exact,[tuple([Index])],Rel).
create_generic_relation(Id,Types,SetOrSingle,ExactOrSubset,TupleSet,reldef(Id,Types,SetOrSingle,ExactOrSubset,TupleSet)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Phase 4
% create the relations for global constants or variables (i.e. those identifiers
% which are not introduced by quantifiers)

create_identifier_relations([],_Typemap,_Bits,IdMap,IdMap,[]).
create_identifier_relations([TId|Rest],Typemap,Bits,IdMapIn,IdMapOut,IntegerDecisions) :-
    create_identifier_relation(TId,Typemap,Bits,IdMapIn,IdMapInter,ID1),
    append(ID1,ID2,IntegerDecisions),
    create_identifier_relations(Rest,Typemap,Bits,IdMapInter,IdMapOut,ID2).

create_identifier_relation(TId,Typemap,Bits,In,Out,IntegerDecisions) :-
    get_texpr_id(TId,Id),
    get_texpr_type(TId,Type),
    get_texpr_info(TId,Info),
    unique_identifier(id,Id,RelId),
    is_set_or_single(Type,SetOrSingle,RestType),
    create_identifier_relation2(RestType,SetOrSingle,RelId,Info,Bits,Typemap,Rel,IntegerDecisions),
    avl_store(Id,In,Rel,Out).
create_identifier_relation2(integer,set,RelId,_Info,_Bits,_Typemap,Rel,[]) :-
    !,intset_relation_kodkod_name(Type),
    create_generic_relation(RelId,[Type],set,subset,full,Rel).
create_identifier_relation2(integer,single,RelId,Info,Bits,_Typemap,Rel,[Trigger]) :-
    !,delayed_integer_decision(Type,Trigger),
    % if the type does not fit into the bitset range, the choice is clear:
    % we have to use the pow2 representation
    ( in_bitrange(Info,Bits) -> true
    ; otherwise              -> pow2integer_relation_kodkod_name(Type)),
    create_generic_relation(RelId,[Type],single,subset,full,Rel).
create_identifier_relation2(RestType,SetOrSingle,RelId,_Info,_Bits,Typemap,Rel,[]) :-
    extract_ktypes(RestType,Typemap,Types),
    create_generic_relation(RelId,Types,SetOrSingle,subset,full,Rel).

in_bitrange(Info,irange(BLow,BUp)) :-
    integer_range(Info,Low,Up),
    BLow =< Low,
    Up =< BUp.
integer_range(Info,Lower,Upper) :-
    memberchk(analysis(A),Info),
    memberchk(interval:irange(Lower,Upper),A).

:- block delayed_integer_decision(-,-).
delayed_integer_decision(Type,1) :-
    ( nonvar(Type) ->
        ( pow2integer_relation_kodkod_name(Type); intset_relation_kodkod_name(Type) ),!
    ; otherwise ->
        ( pow2integer_relation_kodkod_name(Type); intset_relation_kodkod_name(Type) )).
trigger_integer_decisions([]).
trigger_integer_decisions([Trigger|Rest]) :-
    Trigger=1,trigger_integer_decisions(Rest).

is_set_or_single(seq(T),set,couple(integer,T)) :- !.
is_set_or_single(set(T),set,T) :- !.
is_set_or_single(T,single,T).

extract_ktypes(couple(A,B),Typemap,Types) :- !,
    extract_ktypes(A,Typemap,ATypes),
    append(ATypes,BTypes,Types),    
    extract_ktypes(B,Typemap,BTypes).
extract_ktypes(Type,Typemap,[Name]) :-
    avl_fetch(Type,Typemap,Relation),
    Relation = relation(Name,_Size,_Relspec).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Get all referenced relations in the Kodkod formula

remove_unused_relations(Typemap,Idmap,Valuemap,Formula,NTypemap,NIdmap,Valuemap) :-
    get_used_relations(Formula,UsedRels),
    remove_unused_idrelations(UsedRels,Idmap,NIdmap),
    get_used_types_in_rels(NIdmap,UsedRels,UsedRels2),
    remove_unused_types(Typemap,UsedRels2,NTypemap).


remove_unused_idrelations(UsedRels,Idmap,NIdmap) :-
    avl_to_list(Idmap,IdList),
    remove_unused_idrelations(IdList,UsedRels,NIdList),
    list_to_avl(NIdList,NIdmap).
remove_unused_idrelations([],_UsedRels,[]).
remove_unused_idrelations([B-Reldef|Rest],UsedRels,NIdList) :-
    Reldef = reldef(Id,_,_,_,_),
    ( memberchk(Id,UsedRels) -> NIdList = [B-Reldef|Nrest]
    ; otherwise -> NIdList = Nrest),
    remove_unused_idrelations(Rest,UsedRels,Nrest).

get_used_types_in_rels(Idmap,UsedRels,NewUsed) :-
    avl_to_list(Idmap,IdList),
    findall(T, (member(_-reldef(_,Ts,_,_,_),IdList),member(T,Ts)), Types),
    append(UsedRels,Types,U2),
    sort(U2,NewUsed).

remove_unused_types(Typemap,UsedRels,NTypemap) :-
    avl_to_list(Typemap,TypeList),
    remove_unused_types(TypeList,UsedRels,NTypeList),
    list_to_avl(NTypeList,NTypemap).
remove_unused_types([],_,[]).
remove_unused_types([T|Trest],UsedRels,NTypeList) :-
    T = relation(Id,_,_),
    ( memberchk(Id,UsedRels) -> NTypeList = [T|Nrest]
    ; otherwise -> NTypeList = Nrest),
    remove_unused_types(Trest,UsedRels,Nrest).

get_used_relations(K,UsedRels) :-
    get_used_relations2(K,U,[]),
    sort(U,UsedRels).
get_used_relations2(relref(Id)) --> !,[Id].
get_used_relations2(K) -->
    {compound(K),!,K =.. Args},get_used_relations_l(Args).
get_used_relations2(_K) --> [].
get_used_relations_l([]) --> [].
get_used_relations_l([K|Krest]) -->
    get_used_relations2(K),
    get_used_relations_l(Krest).
