% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

% LTL and CTL model checking

:- module(ltl,[ltl_model_check/4,ltl_model_check2/4,
               ctl_model_check/4, ltl_check_assertions/2, parse_ltlfile/2,
               ltl_model_check_with_ce/4,
               ltl_formula_structure/3,
               typecheck_temporal_formula/3]).

:- use_module(library(lists)).
:- use_module(library(avl)).
:- use_module(library(codesio)).

:- use_module(debug).
:- use_module(bmachine,[b_get_machine_operation/4,b_type_expression/5,
                        b_definition_prefixed/4, b_get_typed_definition/3]).
:- use_module(error_manager).
:- use_module(probcspsrc(haskell_csp),[channel/2, channel_type_list/2, filter_ltl_formulas_from_pragmas/2, parse_single_csp_expression/3]).
:- use_module(tools).
:- use_module(specfile,[animation_mode/1, csp_mode/0, csp_with_bz_mode/0,
                        state_corresponds_to_initialised_b_machine/2]).
:- use_module(self_check).

:- use_module(state_space,[set_max_nr_of_new_impl_trans_nodes/1,impl_trans_term/3,
                           current_state_id/1,visited_expression/2,
                           transition/3,transition/4,
                           state_space_add/2,retract_visited_expression/2,  % used for test cases
                           impl_trans_term_all/2,state_space_reset/0,
                           execute_id_trace_from_current/3]).

:- use_module(kernel_objects,[equal_object/2]).
:- use_module(b_interpreter,[b_compute_expression/4,b_test_boolean_expression_wf/3]).
:- use_module(bsyntaxtree).
:- use_module(parsercall,[parse_exprs_preds/4,call_ltl_parser/3]).
:- use_module(store,[empty_state/1,lookup_value/3,add_var_to_localstate/4]).

:- use_module(plugins).

:- use_module('promela/promela_tools',[parse_promela_strings/4,eval_promela/4]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,ltl).
:- module_info(description,'This module provides CTL and LTL model checking on models.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').


% import the low-level C-library wrapper
:- use_module(extension('ltlc/ltlc')).

:- meta_predicate printtime(0).

/* CTL Model Checker */

/* sample formula: ef(e(b(stop_Moat/0))) */

sat(true,_E,T,R) :- !, T=[], R=true.
sat(false,_E,T,R) :- !, T=[], R=false.
sat(not(F),E,T,Res) :- !, sat(F,E,T,Res1), (Res1=true ->Res=false ; Res=true).
sat(ap(P),E,[],Res) :- !, (check_ap(P,E) -> Res=true ; Res=false). /* elementary atomic proposition */
sat(and(F,G),E,T,Res) :- !,
   sat(F,E,T1,Res1), 
   (Res1=false -> Res=false,T=T1
               ;  sat(G,E,T,Res)).
sat(or(F,G),E,T,Res) :- !,
   sat(F,E,T1,Res1), %print(sat_or1(F,E,T1,Res1)),nl,
   (Res1=true ->  Res=true,T=T1
               ;  sat(G,E,T,Res) %, print(sat_or2(G,E,T,Res)),nl
    ).
sat(implies(A,B),E,T,Res) :- !, sat( or(not(A),B),E,T,Res).
sat(an(F),E,T,Res) :- !, sat(not(en(not(F))),E,T,Res).
sat(en(F),E,T,Res) :- !, /* exists next */
    ((trans(E,E2,_),sat(F,E2,T2,true))
      -> Res=true, T=[E2|T2]
      ;  Res=false, T=[]).
sat(action(OpSpec,F),E,T,Res) :- !, /* exists next with predicate on transition; ena renamed to action by preprocess_formula2 */
    ((trans(E,E2,Act),check_transition(OpSpec,Act,E,E2),sat(F,E2,T2,true))
      ->  Res=true, T=[E2|T2]
      ;   Res=false, T=[]).
sat(eu(F,G),E,T,Res) :- !, /* exists until */
    sat_eu(E,F,G,T,Res).
sat(ef(F),E,T,Res) :- !, /* exists future */
    sat(eu(true,F),E,T,Res).
sat(ag(F),E,T,Res) :- !,sat(not(ef(not(F))),E,T,Res).
sat(X,_State,_,_) :- add_internal_error('Uncovered CTL Formula: ',X),fail.
 
:- dynamic sat_eu_table/5.
sat_eu(E,F,G,T,R) :-
    empty_avl(Computing),  %print(start_eu(E,F,G,T,R)),nl,trace,
    sat_eu2(E,F,G,Computing,OutComputing,T,Res),  print(res(Res)),nl,
    ( Res = computing ->
        % we found only false and computing (i.e. cycles) nodes,
        % there is no true node to encounter
        assert(sat_eu_table(E,F,G,[],false)),
        % we know that all visited nodes must be false, too
        mark_all_computing_nodes_as_false(F,G,OutComputing),
        R=false
    ; otherwise ->
        R = Res).
% sat_eu2/7 is used recursively to find a reachable node where G is true
% A set of elements that have already been checked is passed (InComputing and
% OutComputing)
sat_eu2(E,_F,_G,InComputing,OutComputing,T,R) :-
    % the node has already been checked, Result is "computing"
    avl_fetch(E,InComputing),!,R=computing,T=[],OutComputing = InComputing.
sat_eu2(E,F,G,InComputing,OutComputing,T,R) :-
    % the value for this node has already been computed
    sat_eu_table(E,F,G,TT,Res),!, %print(sat_eu_table(E,F,G,TT,Res)),nl,
    R=Res,T=TT,OutComputing=InComputing.
sat_eu2(E,F,G,InComputing,OutComputing,T,Res) :- % print(sat_eu_eval(E,F,G)),nl,
    % the goal is true in this node
    Res=true,
    /* exists until */
    sat(G,E,T,true),!,
    % print(set_eu_g_true(G,T)),nl,
    assert(sat_eu_table(E,F,G,[],true)),
    InComputing = OutComputing.
sat_eu2(E,F,G,InComputing,OutComputing,T,Res) :- /* exists until */
    % the goal is not true
    % sat(not(G),E,_), /* added to avoid extra answers */
    sat(F,E,T1,Res1),
    % print(sat_eu_loop(E,F,T1,Res1)),nl,
    ( Res1=false ->
        % if F is false the result is false anyhow - we do not have to look into the future
        assert(sat_eu_table(E,F,G,[],false)),
        Res=false, T=T1, OutComputing=InComputing
    ; Res1=true ->
        avl_store(E,InComputing,true,Computing),
        sat_eu_step(E,F,G,Computing,OutComputing,T2,Res2),
        ( Res2 = true ->
            % we found a node with the goal, mark this as true
            assert(sat_eu_table(E,F,G,T2,true)),
            Res=true, T=T2
        ; Res2 = computing ->
            Res=computing, T=[]
        ; Res2 = false ->
            % we found only false nodes, mark this as false
            assert(sat_eu_table(E,F,G,[],false)),
            Res=false, T=[]
        )
    ).
sat_eu_step(E,F,G,InComputing,OutComputing,T,R) :-
    findall(N,trans(E,N,_),Nexts),
    sat_eu_step2(Nexts,F,G,InComputing,OutComputing,T,false,R).
sat_eu_step2([],_F,_G,InComputing,OutComputing,T,In,Out) :- !,In=Out,T=[],InComputing=OutComputing.
sat_eu_step2([Next|Nrest],F,G,InComputing,OutComputing,T,In,R) :-
    sat_eu2(Next,F,G,InComputing,Computing,T2,Res),
    ( Res=true  -> R=true,T=[Next|T2],OutComputing=Computing
    ; otherwise ->
        combine_result(In,Res,New),
        sat_eu_step2(Nrest,F,G,Computing,OutComputing,T,New,R)).
combine_result(false,false,R) :- !,R=false.
combine_result(_,_,computing).

mark_all_computing_nodes_as_false(F,G,Computing) :-
    avl_member(E,Computing),
    assert(sat_eu_table(E,F,G,[],false)),
    fail.
mark_all_computing_nodes_as_false(_F,_G,_Computing).
    
ctl_sat_for_all(Formula,States,Trace,Res) :-
    ctl_sat_for_all2(States,Formula,[],unknown,Trace1,Res1),
    ( Res1 = unknown -> Res=true, Trace=[]
    ; otherwise      -> Res=Res1, Trace=Trace1).
ctl_sat_for_all2([],_Formula,Trace,Res,Trace,Res).
ctl_sat_for_all2([State|Srest],Formula,PrevTrace,PrevRes,Trace,Res) :-
    sat(Formula,State,Trace1,Res1),
    NewTrace = [State|Trace1],
    ( Res1=false ->
        % Counter-example found, use it and stop
        Res=false, Trace=NewTrace
    ; PrevRes=unknown ->
        % First witness found, continue
        ctl_sat_for_all2(Srest,Formula,NewTrace,true,Trace,Res)
    ; otherwise ->
        % Another witness found, ignore it and use the old
        ctl_sat_for_all2(Srest,Formula,PrevTrace,true,Trace,Res)).
   
ctl_model_check2(Formula,Mode,InitStates,Res) :-
  retractall(sat_eu_table(_,_,_,_,_)),
  preprocess_formula(Formula,PFormula),
  ( printtime(ctl_sat_for_all(PFormula,InitStates,Trace,Res)) ->
      ( Res = true -> print('Witness found: ')
      ; otherwise  -> print('Counterexample found: ')),
      nl,print(Trace),nl,
      Trace = [Initstate|_],
      ( Mode=init -> set_trace_to_init(Initstate) % we need to go to the initialised state
                                                  % before the counterexample itself is shown
      ; otherwise -> true), 
      executetrace(Trace),
      print_size_of_table(ltl:sat_eu_table/5)
  ; otherwise ->
      add_internal_error('CTL model check failed: ',PFormula),
      print_size_of_table(ltl:sat_eu_table/5),fail
  ),nl.

ctl_model_check(Formula,MaxNodes,Mode,Res) :-
	( temporal_parser(Formula,ctl,Ctl) ->
	    true
	; otherwise ->
	    add_error(ctl,'CTL Parser failed: ',Formula),fail),
    set_max_nr_of_new_impl_trans_nodes(MaxNodes),
    select_ctl_mode(Mode,Ctl,Startnodes,Ctl2),
    ctl_model_check2(Ctl2,Mode,Startnodes,Res).
select_ctl_mode(starthere,Ctl,[Startnode],Ctl) :-
    !,current_state_id(Startnode).
select_ctl_mode(init,Ctl,Startnodes,Ctl) :-
    !,init_states(Startnodes).
select_ctl_mode(Mode,Ctl,[],Ctl) :-
    add_error(ltl,'Internal error: Unexpected mode', Mode),
    fail.

/* --------------------------- */


trans(A,B,Op) :- impl_trans_term(A,Op,B). /* will automatically compute as required */

:- use_module(eclipse_interface,[test_boolean_expression_in_node/2]).
% check an atomic proposition in a given node (i.e. state)
check_ap(bpred(Property),Node) :-
	!, test_boolean_expression_in_node(Node,Property).
check_ap(promela_pred(Predicate),Node) :-
	visited_expression(Node,State),!,
	eval_promela(Predicate,global,State,Result),!,
	( Result == 0    -> fail
	; number(Result) -> true
	; otherwise      -> add_error(ltl,'Promela eval returned a non-boolean value',Result),fail).
check_ap(enabled(Op),Node) :- !, check_enabled(Op,Node).
check_ap(deadlock,Node) :- !, \+ trans(Node,_,_).
check_ap(sink,Node) :- !,
	\+ ( trans(Node,Next,_),Next \= Node ).
check_ap(stateid(SpecNode),Node) :- !, SpecNode = Node.

check_enabled(Spec,SrcNode) :-
        check_transition(Spec,Transition,SrcNode,DestNode),
        trans(SrcNode,DestNode,Transition),!.

% check_transition(Spec,Transition,SrcNode,DstNode)
%   Checks if a transition fulfils the specification, it fails if not.
%   Please note, that this predicate is called _before_ the actual transition
%   is known.
%  Spec: specification that encodes how the transition should look like
%        See below for the supported transitions
%  Transition: The transition itself as it is stored in the second arg of state_space:transition/3
%  SrcNode: The Id of the source state (first arg of transition/3)
%  DstNode: The Id of the destination state (third arg of transition/3)

:- assert_must_succeed( check_transition(bop(dummyop,2,0,[],[]),dummyop(int(1),int(2)),tsrc,tdst) ).
:- assert_must_fail( check_transition(bop(dummyop,2,0,[],[]),otherop(int(1),int(2)),tsrc,tdst) ).
:- assert_must_fail( check_transition(bop(dummyop,2,0,[],[]),dummyop(int(1)),tsrc,tdst) ).
:- assert_must_succeed(call_cleanup(( empty_state(E),
                                      state_space_add(tsrc,E),
                                      create_texpr(integer(2),integer,[],I2),
                                      check_transition(bop(dummyop,2,0,[2/I2],[]),
                                                       dummyop(int(1),int(2)),tsrc,tdst)),
                                    retract_visited_expression(tsrc,_))).
:- assert_must_fail(call_cleanup(( empty_state(E),
                                   state_space_add(tsrc,E),
                                   create_texpr(integer(2),integer,[],I2),
                                   check_transition(bop(dummyop,2,0,[1/I2],[]),
                                                    dummyop(int(1),int(2)),tsrc,tdst)),
                                 retract_visited_expression(tsrc,_))).
:- assert_must_succeed( check_transition(bop(dummyop,1,1,[],[]),'-->'(dummyop(int(1)),[int(2)]),tsrc,tdst) ).
:- assert_must_fail( check_transition(bop(dummyop,1,1,[],[]),dummyop(int(1)),tsrc,tdst) ).
:- assert_must_succeed(call_cleanup(( empty_state(E),
                                      state_space_add(tsrc,E),
                                      create_texpr(integer(2),integer,[],I2),
                                      check_transition(bop(dummyop,1,1,[],[1/I2]),
                                                       '-->'(dummyop(int(1)),[int(2)]),tsrc,tdst)),
                                    retract_visited_expression(tsrc,_))).
:- assert_must_fail(call_cleanup(( empty_state(E),
                                   state_space_add(tsrc,E),
                                   create_texpr(integer(2),integer,[],I2),
                                   check_transition(bop(dummyop,1,1,[],[1/I2]),
                                                    '-->'(dummyop(int(1)),[int(1)]),tsrc,tdst)),
                                 retract_visited_expression(tsrc,_))).
                       
% bop(Name,NumberArgs,NumberResults,ArgPatterns,ResPatterns):
%   The transition is a B operation with name Name, NumberArgs parameters and NumberResults
%   result values.
%   ArgPatterns is a list of Terms of the form Index/Expr where Index is the
%     number of the parameter that should have the value of the expression Expr.
%   ResPatterns is a list analogous to ArgPatterns on the result
check_transition(bop(Name,NumberArgs,NumberResults,ArgPatterns,ResPatterns),
                 Transition,SrcNode,_DstNode) :-
    !, functor(Op,Name,NumberArgs),
    ( NumberResults == 0 -> Transition = Op
    ; otherwise          ->
        Transition = '-->'(Op,ReturnValues),
        length(ReturnValues,NumberResults),
        check_bop_res_patterns(ResPatterns,SrcNode,State,ReturnValues)),
    check_bop_arg_patterns(ArgPatterns,SrcNode,State,Op).
check_transition(btrans(TransPred),Transition,SrcNode,DstNode) :-
    !, check_eventb_transition(TransPred,Transition,SrcNode,DstNode).
check_transition(csp(Channel,Arity,Args),Transition,_SrcNode,_DstNode) :-
    !,translate_csp_event(Channel,Arity,Args,Transition,false).
    %print(translate_csp_event(Channel,Arity,Args,Transition,_SrcNode,_DstNode)),nl.
check_transition(xtl(Op,Arity,PatternArgs),Transition,_SrcNode,_DstNode) :-
    !, check_xtl_transition(Op,Arity,PatternArgs,Transition).
check_transition(plugin_transition_check(Id,Predicate),Transition,_SrcNode,_DstNode) :-
    !, plugin_evaluate_transition_predicate(Id,Predicate,Transition).
check_transition(Spec,_Trans,_SrcNode,_DstNode) :-
    functor(Spec,Name,Arity), add_error(ltl, 'Unknown transition specification ', Name/Arity), fail.

:- block check_xtl_transition(?,?,?,-).
check_xtl_transition(Op,Arity,PatternArgs,Transition) :-
	( Arity=0 ->
	    functor(Transition,Op,_) /* in this case: pattern match any operation with right number of args */
	;   %functor(Transition,Op,Arity),
        Transition =.. [Op|Args],
	    check_xtl_arguments(Args,PatternArgs)
	).

check_bop_arg_patterns([],_,_,_).
check_bop_arg_patterns([P|Prest],Node,State,Op) :-
    check_bop_arg_patterns2([P|Prest],Node,State,Op).
:- block check_bop_arg_patterns2(?,?,?,-).
check_bop_arg_patterns2([],_,_,_).
check_bop_arg_patterns2([Index/Expr|Prest],Node,State,Op) :-
    check_bop_arg_pattern(Index,Expr,Node,State,Op),
    check_bop_arg_patterns2(Prest,Node,State,Op).
check_bop_arg_pattern(Index,Expr,Node,State,Op) :-
    arg(Index,Op,OpValue),  % exception if Index is expression like 1+1 <----
    check_b_operation_value(OpValue,Expr,Node,State).

check_bop_res_patterns([],_,_,_).
check_bop_res_patterns([P|Prest],Node,State,OpResult) :-
    check_bop_res_patterns2([P|Prest],Node,State,OpResult).
:- block check_bop_res_patterns2(?,?,?,-).
check_bop_res_patterns2([],_,_,_).
check_bop_res_patterns2([Index/Expr|Prest],Node,State,OpResult) :-
    check_bop_res_pattern(Index,Expr,Node,State,OpResult),
    check_bop_res_patterns2(Prest,Node,State,OpResult).
check_bop_res_pattern(Index,Expr,Node,State,OpResult) :-
    nth1(Index,OpResult,OpValue),
    check_b_operation_value(OpValue,Expr,Node,State).

check_b_operation_value(OpValue,Expr,Node,State) :-
    ( var(State) ->
        visited_expression(Node,RawState),
        state_corresponds_to_initialised_b_machine(RawState,State)
    ; otherwise -> true),
    empty_state(LS),
    b_compute_expression(Expr,LS,State,PatternValue),
    equal_object(PatternValue,OpValue),!.
    

:- block check_eventb_transition(?,-,?,?).
check_eventb_transition(event(Name),Transition,_SrcNode,_DstNode) :- !,
    functor(Transition,Name,_).
check_eventb_transition(event(Name,Predicate,PoststateAccess),Transition,SrcNode,DstNode) :- !,
    functor(Transition,Name,_),
    visited_expression(SrcNode,CurState),
    state_corresponds_to_initialised_b_machine(CurState,CurBState),
    parameter_values(Transition,LocalState1),
    add_post_state_values(PoststateAccess,DstNode,LocalState1,LocalState),
    b_test_boolean_expression_wf(Predicate,LocalState,CurBState).

add_post_state_values([],_DstNode,State,State) :- !.
add_post_state_values(Access,Node,InState,OutState) :-
    visited_expression(Node,DstState1),
    state_corresponds_to_initialised_b_machine(DstState1,DstState),
    add_post_state_values2(Access,DstState,InState,OutState).
add_post_state_values2([],_DstState,State,State).
add_post_state_values2([poststate(Orig,Primed)|Prest],DstState,InState,OutState) :-
    lookup_value(Orig,DstState,Value),
    add_var_to_localstate(Primed,Value,InState,InterState),
    add_post_state_values2(Prest,DstState,InterState,OutState).

parameter_values(Transition,LocalState) :-
    Transition =.. [Name|Arguments],
    b_get_machine_operation(Name,_,Parameters,_),
    empty_state(Empty),
    add_parameter_values2(Parameters,Arguments,Empty,LocalState).
add_parameter_values2([],[],State,State).
add_parameter_values2([TParameter|Prest],[Argument|Arest],InState,OutState) :-
    def_get_texpr_id(TParameter,Parameter),
    add_var_to_localstate(Parameter,Argument,InState,InterState),
    add_parameter_values2(Prest,Arest,InterState,OutState).

/*
:- assert_must_succeed(( ltl:split_operation('ch.2.3','.',ch,A), A=[2,3] )).
:- assert_must_succeed(( ltl:split_operation('ch.2.3','.',ch,[2,3]) )).
:- assert_must_succeed(( ltl:split_operation('ch','.',ch,[]) )).
:- assert_must_succeed(( ltl:split_operation('ch.','.',ch,[A]), A='' )).
split_operation(OpAtom, Sep,  OpName, OpArgs) :-
   atom_codes(OpAtom,ListAscii), atom_codes(Sep,[SepACode]),
   split_op(ListAscii,SepACode,HeadA, ArgsA),
   atom_codes(Head,HeadA), OpName=Head,
   l_name(Args,ArgsA), Args=OpArgs.
*/

l_name([],[]).
l_name([H|T],[AH|AT]) :- name(H,AH),  % using name to generate numbers if possible
                               l_name(T,AT).

split_op([],_,[],[]). % no separator found
split_op([H|T],Sep,OpName,Args) :-
   (H=Sep -> Args=[FirstArg|OtherArgs],OpName=[],split_args(T,Sep,FirstArg,OtherArgs)
    ; OpName = [H|CT], split_op(T,Sep,CT,Args)).

split_args([],_Sep,[],[]).
split_args([H|T],Sep,FirstArg,OtherArgs) :-
   (H=Sep -> FirstArg=[], OtherArgs=[NextArg|TA],split_args(T,Sep,NextArg,TA)
    ; FirstArg = [H|CT], split_args(T,Sep,CT,OtherArgs)).

:- assert_must_succeed((ltl: mergeArgsToDotTuple([1,2,'<bob,alice>','{bob}'],Res), Res = '1.2.<bob,alice>.{bob}')).
:- assert_must_succeed((ltl: mergeArgsToDotTuple(['(bob,alice)'],Res), Res = '(bob,alice)')).
:- assert_must_succeed((ltl: mergeArgsToDotTuple([],''))).

mergeArgsToDotTuple(List,Res) :- mergeArgsToDotTuple(List,'',Res).

mergeArgsToDotTuple([],Str,Str).
mergeArgsToDotTuple([H|T],Str,Res) :-
        (T\=[] -> string_concatenate(H,'.',HRes) ; HRes=H),
         string_concatenate(Str,HRes,StrRes),mergeArgsToDotTuple(T,StrRes,Res).

splitToSingleArgs(dotTuple(List),List).
splitToSingleArgs(X,[X]).

check_xtl_arguments([],[]).
check_xtl_arguments([],[H|T]) :-
   print('### Extra arguments in LTL/CTL pattern: '),print([H|T]),nl,fail.
check_xtl_arguments([H|T],[]) :- print(ltl_check_xtl_arguments_ignoring([H|T])),nl.
check_xtl_arguments([H|T],[PatH|PatT]) :-
    check_xtl_arg(PatH,H,T,RemArgs),
    check_xtl_arguments(RemArgs,PatT).

check_xtl_arg('_',_,TArgs,RemArgs) :- !,RemArgs=TArgs.
check_xtl_arg('?',_,TArgs,RemArgs) :- !,RemArgs=TArgs.
check_xtl_arg(X,Y,TArgs,RemArgs) :- atomic(Y),!,X=Y,RemArgs=TArgs.
check_xtl_arg(X,Y,TArgs,RemArgs) :- nonvar(Y),!,Y=..[Functor|YArgs], Functor=X, 
  append(YArgs,TArgs,RemArgs).
check_xtl_arg(X,Y,TArgs,RemArgs) :- var(Y),
   print(variable_for_ltl_pattern(X)),nl,
   X=Y,RemArgs=TArgs.

:- assert_must_succeed(translate_csp_event(c,3,[a,10,20],io([in(a),in(int(10)),in(int(20))],c,src),true)).
:- assert_must_succeed(translate_csp_event(c,3,[a,10,20],io([out(a),out(int(10)),out(int(20))],c,src),true)).
:- assert_must_fail(translate_csp_event(c,3,[a,1,20],io([in(a),in(int(10)),in(int(20))],c,src),true)).
:- assert_must_succeed(translate_csp_event(c,3,[a,'_','_'],io([in(a),in(int(10)),in(int(20))],c,src),true)).
:- assert_must_succeed(translate_csp_event(c,3,[a,'?','?'],io([in(a),in(int(10)),in(int(20))],c,src),true)).
:- assert_must_succeed(translate_csp_event(c,3,['_'],io([in(a),in(int(10)),in(int(20))],c,src),true)).
:- assert_must_succeed(translate_csp_event(c,0,[],io([],c,src),true)).
:- assert_must_succeed(translate_csp_event(tick,0,[],tick(_),true)).
:- assert_must_succeed(translate_csp_event(tau,0,[],tau(_),true)).
:- assert_must_fail(translate_csp_event(cz,1,[tupleExp([s,s])],io([list([alice,bob])],send,_span),true)).
:- assert_must_fail(translate_csp_event(cz,1,[listExp([s,s])],io([list([alice,bob])],send,_span),true)).
:- assert_must_succeed(translate_csp_event(send,1,[tupleExp([s,s])],io([na_tuple([s,s])],send,_span),true)).
:- assert_must_fail(ltl:check_csp_arguments([setValue([alice,bob]),tupleExp([s,t,u])], [setValue([alice,bob]),in(na_tuple([s,t]))],_)).
:- assert_must_succeed(ltl:check_csp_arguments([setValue([alice,bob]),tupleExp([s,t,u])], [setValue([alice,bob]),in(na_tuple([s,t,u]))],_)).
:- assert_must_succeed(ltl:check_csp_arguments([setValue([bob,alice]),tupleExp([s,t,u])], [setValue([alice,bob]),in(na_tuple([s,t,u]))],_)).

translate_csp_event(tick,0,[],R,_) :- !, R=tick(_).
translate_csp_event(tau,0,[],R,_) :- !, R=tau(_).
translate_csp_event(Event,Arity,PatArgs,io(Vals,Event,_SrcSpan),Test) :- !,
	(channel(Event,_),\+Test ->
		((channel_type_list(Event,List),length(List,Len),Len==Arity) -> 
		    check_csp_arguments(PatArgs,Vals,Event)
		; otherwise -> add_internal_error('Internal Error: Missing/extra arguments in CSP LTL/CTL pattern: ',Event:PatArgs)
		)
	; Test -> check_csp_arguments(PatArgs,Vals,Event)
	;otherwise -> add_internal_error('Internal Error: Channel not defined: ',Event/Arity)).

:- block check_csp_arguments(?,-,?).
check_csp_arguments([],[],_).
check_csp_arguments([],[H|T],Event) :-
  print('* Missing arguments in CSP LTL/CTL pattern (use .?): '),print(Event),
  print(' : '),print([H|T]),nl.
check_csp_arguments([H|T],[],Event) :- 
  add_error(check_csp_arguments,'Extra arguments in CSP LTL/CTL pattern: ',Event:[H|T]),fail.
check_csp_arguments([H|T],[PatH|PatT],Event) :-
  check_csp_arg(H,PatH),!,check_csp_arguments(T,PatT,Event).

check_csp_arg('_',_) :- !.
check_csp_arg('?',_) :- !.
check_csp_arg(X,in(Y)) :- !,check_csp_arg(X,Y).
check_csp_arg(X,out(Y)) :- !,check_csp_arg(X,Y).
check_csp_arg(list(L),list(L1)) :- !,l_check_csp_arg(L,L1,list).
check_csp_arg(setValue(L),setValue(L1)) :- !,sort(L,R),sort(L1,R1),l_check_csp_arg(R,R1,setValue). % elements order is not important by set expressions
check_csp_arg(tupleExp(L),na_tuple(L1)) :- !,l_check_csp_arg(L,L1,na_tuple).
check_csp_arg(X,Y) :- eq_csp_arg(Y,X).
eq_csp_arg(X,Y) :- eq_csp_val(X,Y).
eq_csp_val(int(X),int(Y)) :- !, Y=X.
eq_csp_val(int(X),Y) :- !,Y=X.
eq_csp_val(X,X).

l_check_csp_arg([],[],_):- !.
l_check_csp_arg([],[H|T],Functor) :- add_error(check_csp_arguments,'Missing elements in CSP LTL pattern: ',Functor:[H|T]),fail.
l_check_csp_arg([H|T],[H1|T1],Functor) :- check_csp_arg(H,H1),!,l_check_csp_arg(T,T1,Functor).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get LTL formulas from definitions
:- dynamic ltl_formula_cache/2.

get_ltl_formulas(Names,Strings,Formulas) :-
	(animation_mode(b) -> get_ltl_formula_strings(Names,Strings)
	; animation_mode(cspm) -> filter_ltl_formulas_from_pragmas(Names,Strings)
        ; animation_mode(csp_and_b) -> get_ltl_formula_strings(BNames,BStrings),
                                       filter_ltl_formulas_from_pragmas(CSPNames,CSPStrings),
					append(BNames,CSPNames,Names),append(BStrings,CSPStrings,Strings)
				       ),
	( ltl_formula_cache(Strings,Formulas) -> true
	; otherwise ->
	    retractall(ltl_formula_cache(_,_)),
	    temporal_parser_l(Strings,ltl,Formulas),
	    assert(ltl_formula_cache(Strings,Formulas))).

get_ltl_formula_strings(Names,Formulas) :-
	findall(ltl_assertion(Name,F),get_ltl_formula_string(Name,F),Assertions),
	findall(N,member(ltl_assertion(N,_),Assertions),Names),
	findall(F,member(ltl_assertion(_,F),Assertions),Formulas).	
get_ltl_formula_string(Tail,S) :-
	b_definition_prefixed(_,'ASSERT_LTL',Tail,DefName),
	b_get_typed_definition(DefName,[],F),
	( get_texpr_expr(F,string(S)) -> true
	; get_texpr_pos(F,Pos) ->
	    add_all_perrors([error('Expected String for LTL formula',Pos)]),
	    fail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check all LTL assertions in a machine

ltl_check_assertions(Limit,Outcome) :-
	get_ltl_formulas(Names,Strings,Formulas),
	extract_expectations(Names,Exp),
	ltl_check_assertions2(Names,Strings,Formulas,Exp,Limit,Outcomes),
	merge_outcomes(Outcomes,Outcome).
	
ltl_check_assertions2([],[],[],[],_,[]).
ltl_check_assertions2([Name|Nrest],[String|Srest],[Formula|Frest],[Exp|Erest],Limit,[Outcome|Orest]) :-
	print('Checking LTL formula '),print(Name),print(': '),print(String),nl,
	flush_output(user_output),
	catch( ltl_model_check2(Formula,Limit,init,Status),
	       E,
	       Status = exception(E)),
	( Exp=pass, Status=ok -> Outcome=pass, print('no counter-example found, test passed\n')
	; Exp=fail, Status=no -> Outcome=pass, print('counter-example found, test passed\n')
	; Status=exception(E) ->
	    add_error(ltl,'Caught exception',E),Outcome=error
	; Exp=pass, Status=no ->
	    Outcome=fail, add_error(ltl, 'Expected correct formula, but found counter-example')
	; Exp=fail, Status=ok ->
	    Outcome=fail, add_error(ltl, 'Expected counter-example, but model checker found none')
	; otherwise ->
	    add_error(ltl,'Unexpected result',(Exp,Status)),Outcome=fail),
	ltl_check_assertions2(Nrest,Srest,Frest,Erest,Limit,Orest).
	
extract_expectations([],[]).
extract_expectations([N|Nrest],[E|Erest]) :-
	name(N,Codes),
	( append(_,"FAIL",Codes) -> E=fail; E=pass),
	extract_expectations(Nrest,Erest).

merge_outcomes(List,Result) :-
	merge_outcomes2(List,no_tests,Result).
merge_outcomes2([],R,R).
merge_outcomes2([O|Rest],P,R) :-
	merge_outcomes3(P,O,I),
	merge_outcomes2(Rest,I,R).
merge_outcomes3(no_tests,X,X) :- !.
merge_outcomes3(_,error,error) :- !.
merge_outcomes3(P,fail,fail) :- P\=error,!.
merge_outcomes3(P,pass,P).
 
%*******************************************************************************

% ltl_model_check(FormulaAsAtom,Max,Mode,Status):
%  FormulaAsAtom: The LTL formula as an atom
%  Max: The maximum number of states that should be newly calculated
%  Mode: Where to start the search in state space
%     starthere: start the evaluation of the formula in the current state
%     init: start in the (possible several) initialisation states of the model
%     checkhere: start the search as in init, but check the formula F in the
%        current state by constructing a new formula (current -> F)
%  Status:
%     no: the model does not fulfil the formula, a counter-example is shown by
%         modifying the history
%     ok: the model does fulfil the formula
%     incomplete: the search was stopped (usually because the limit Max was reached)
%         before finding a counter-example or guarantee the absent of a counter-example
ltl_model_check(FormulaAsAtom,Max,Mode,Status) :- print_message(parsing_ltl_formula),
	( temporal_parser(FormulaAsAtom,ltl,Ltl) ->
	    true
	; otherwise ->
	    add_error(ltl,'LTL Parser failed: ',FormulaAsAtom),fail),
	ltl_model_check2(Ltl,Max,Mode,Status).
	
ltl_model_check2(Ltl,Max,Mode,Status) :-
    print_message(initialising_ltlc),
	print(starting_model_checking),nl,
    ( Ltl=syntax_error ->
        Status=syntax_error
    ; otherwise ->
        ltl_model_check3(Ltl,Max,Mode,Result),
        convert_result(Result,Mode,Status)).

ltl_model_check3(Ltl,Max,Mode,Result) :- 
    call_c_initialisation,
    set_max_nr_of_new_impl_trans_nodes(Max),
    select_ltl_mode(Mode,Ltl,Startnodes,Ltl2),
    preprocess_formula(Ltl2,Ltl3),
    ltl_model_check_core(Startnodes,Ltl3,Result).


ltl_model_check_core([],_Ltl,nostart) :- !.
ltl_model_check_core(Startnodes,Ltl,Result) :-
    c_ltl_modelcheck(Ltl, Startnodes, Result, ltl:cltl_callback).

call_c_initialisation :- ltlc_init,!.
call_c_initialisation :- 
    add_error(ltl,'Initialisation of LTL-C module failed'),
    fail.

select_ltl_mode(starthere,Ltl,[Startnode],not(Ltl)) :-
    !,current_state_id(Startnode).
select_ltl_mode(init,Ltl,Startnodes,not(Ltl)) :-
    !,init_states(Startnodes).
select_ltl_mode(checkhere,Ltl,Startnodes,not(globally(implies(ap(current),Ltl)))) :-
    !,init_states(Startnodes).
select_ltl_mode(Mode,Ltl,[],Ltl) :-
    add_error(ltl,'Internal error: Unexpected mode', Mode),
    fail.

convert_result(nostart,_,nostart) :- !.
convert_result(none,_,ok) :- !.
convert_result(incomplete,_,incomplete) :- !.
convert_result(model(CE,_Entry),Mode,no) :- !,
    translate_ctrace(CE,Op,N,Dest),
    ( (Mode=init; Mode=checkhere) ->
        ce_get_first_state(CE,UsedInitState),
        set_trace_to_init(UsedInitState)
    ; otherwise ->
        true),
    execute_id_trace_from_current(Dest,Op,N).
convert_result(Result,_,_) :-
    add_error(ltl,'Internal error: Unexpected result', Result),
    fail.

ce_get_first_state([atom(State,_,_)|_],State).

translate_ctrace(Atoms,OpIds,StateIds,Dest) :-
    translate_ctrace2(Atoms,beginning,OpIds,StateIds,Dest).
translate_ctrace2([atom(StartNode,_,OpId)|RestAtoms],_PreviousState,Ops,Nodes,Dest) :-
    ( OpId = none -> Ops=[],Nodes=[],Dest=StartNode
    ; otherwise ->
        transition(StartNode,_Op,OpId,NextNode),
        Ops=[OpId|OpRest], Nodes=[NextNode|NextRest], 
        translate_ctrace2(RestAtoms,NextNode,OpRest,NextRest,Dest)).
translate_ctrace2([],Dest,[],[],Dest).

printtime(C) :-
    statistics(runtime,[Start|_]),
    (call(C) -> Res=ok ; Res=failure),
    statistics(runtime,[End|_]),
    T is End - Start,
    format('CTL check took ~3d seconds~n', [T]),
    Res=ok.

% currently not called: nor is unparsed_ap called 
translate_preds(In,Out) :-
	unparsed_ap_l(In,Out,Translation,[]),
	parse_preds_exprs(Translation).
	
parse_preds_exprs([]) :- !. % nothing to parse
parse_preds_exprs(Translation) :-
	split_preds(Translation,PredStrings,Preds,ExprStrings,Exprs),
	animation_mode(AnimMode),
	parse_preds_exprs2(AnimMode,PredStrings,Preds,ExprStrings,Exprs).

parse_preds_exprs2(promela,PredStrings,Preds,ExprStrings,Exprs) :- !,
	annotate_ap_type(Preds,promela_pred,PromPreds),
	parse_promela_strings(ExprStrings,PredStrings,Exprs,PromPreds).

parse_preds_exprs2(Mode,PredStrings,Preds,ExprStrings,Exprs) :-
    is_registered_plugin(Mode),!,
    plugin_parse(ExprStrings,PredStrings,[],Exprs,Preds,[]).   % <--- first argument ID is missing ! Should it be Mode ??

% default: parse the predicates and expressions as B 
parse_preds_exprs2(_,PredStrings,Preds,ExprStringsWithTypes,Exprs) :-
	annotate_ap_type(Preds,bpred,BPreds),
	split_exprs_types(ExprStringsWithTypes,ExprStrings,Types),
	parse_exprs_preds(ExprStrings,UExprs,PredStrings,UPreds),
	type_exprs(UExprs,Types,Exprs,EErrors),
	type_preds(UPreds,BPreds,PErrors),
	append(EErrors,PErrors,Errors),
	add_all_perrors(Errors,[],type_error).

split_exprs_types([],[],[]).
split_exprs_types([Expr/Type|ETRest],[Expr|ERest],[Type|TRest]) :-
	split_exprs_types(ETRest,ERest,TRest).

type_exprs([],[],[],[]).
type_exprs([Expr|Erest],[Type|TypeRest],[Typed|Trest],Errors) :-
	b_type_expression(Expr,[variables],Type,Typed,LErrors),
	append(LErrors,ErrorRest,Errors),
	type_exprs(Erest,TypeRest,Trest,ErrorRest).
type_preds([],[],[]).
type_preds([Pred|Prest],[Typed|Trest],Errors) :-
	b_type_expression(Pred,[variables],pred,Typed,LErrors),
	append(LErrors,ErrorRest,Errors),
	type_preds(Prest,Trest,ErrorRest).

annotate_ap_type([],_,[]).
annotate_ap_type([AP|APrest],Type,[R|Rrest]) :-
	functor(AP,Type,1),arg(1,AP,R),
	annotate_ap_type(APrest,Type,Rrest).



split_preds([],[],[],[],[]).
split_preds([pred(A,B)|Rest],[A|ARest],[B|BRest],X,Y) :- !,split_preds(Rest,ARest,BRest,X,Y).
split_preds([expr(A,T,B)|Rest],X,Y,[A/T|ARest],[B|BRest]) :- split_preds(Rest,X,Y,ARest,BRest).

% replacing unparsed/1 by atomic propositions
unparsed_ap(ap(P),ap(P)) --> !.
unparsed_ap(action(unparsed_b(OpName/Arity/Types,Exprs)),action(b(OpName/Arity,Es))) --> !,
	add_unparsed_exprs(Exprs,Types,Es).
unparsed_ap(action(b(OpName/Arity/_Types)),action(b(OpName/Arity))) --> !.
unparsed_ap(action(Op),action(Op)) --> !.
unparsed_ap(unparsed(Unparsed),ap(P)) --> !,[pred(Unparsed,P)].
unparsed_ap(F,N) --> {F =..[Functor|Args]},unparsed_ap_l(Args,NArgs),{N=..[Functor|NArgs]}.
unparsed_ap_l([],[]) --> [].
unparsed_ap_l([F|Rest],[N|NRest]) --> unparsed_ap(F,N), unparsed_ap_l(Rest,NRest).

add_unparsed_exprs([],[],[]) --> [].
add_unparsed_exprs([expr(P,U)|UR],[T|TR],[expr(P,E)|ER]) -->
	[expr(U,T,E)], add_unparsed_exprs(UR,TR,ER).

%*******************************************************************************
% set current trace

executetrace(Trace) :-
	translate_trace(Trace,OpIds,Ids,Dest),
	print(exec(Dest,OpIds,Ids)),nl,
	execute_id_trace_from_current(Dest,OpIds,Ids).

translate_trace([Start,Id|RestNodes],[OpId|OpRest],[Id|IdRest],Dest) :-
	!,state_space:transition(Start,_Op,OpId,Id),
	translate_trace([Id|RestNodes],OpRest,IdRest,Dest).
translate_trace([Dest],[],[],Dest).


%*******************************************************************************
% find initialised states
% TODO: does NOT work with XTL (CSP?) and constants

init_states(Init) :-
	animation_mode(Mode),
	( is_registered_plugin(Mode) ->
        findall(I,find_plugin_inits(root,Mode,_,I),Init)
    ; mode_cst_init(Mode) ->
	    findall(I,find_init(root,I,_),Init)
	; mode_one_step(Mode) ->
	    next_states(root,Init)
	; otherwise ->
	    fail).

mode_cst_init(b).
mode_cst_init(z).
mode_cst_init(csp_and_b).

mode_one_step(csp).
mode_one_step(cspm).
mode_one_step(xtl).
mode_one_step(promela).

next_states(Start,States) :-
	impl_trans_term_all(Start,Ops),
	findall(S, member(op(_Id,_,S),Ops), States).

find_init(Start,Init,Trace) :- Start==Init,!,Trace=[]. % usually called with Start=Init=root
find_init(Start,Init,[State|Rest]) :-
	impl_trans_term(Start,O,State),
	find_init2(O,State,Init,Rest).
find_init2(O,Init,Init,[]) :-
	has_functor_and_maybe_tau(O,'$initialise_machine').
find_init2(O,State,Init,Path) :-
    has_functor_and_maybe_tau(O,'$setup_constants'),
	find_init(State,Init,Path).
find_init2(start_cspm_MAIN,State,Init,Path) :-
    find_init(State,Init,Path).
    

% has_functor_and_maybe_tau(Term,Functor)
% checks if Term has the form "Functor(...)" or "tau(Functor(...))"
% this is used for CSP||B specification where the initialisation is wrapped with
% in a tau operator
has_functor_and_maybe_tau(tau(Term),Functor) :-
    has_functor_and_maybe_tau(Term,Functor),!.
has_functor_and_maybe_tau(Term,Functor) :-
    functor(Term,Functor,_).


find_plugin_inits(Start,Id,[Start|Rest],StateId) :-
    visited_expression(Start,State),
    ( plugin_is_initialised_state(Id,State) ->
        StateId = Start,
        Rest = []
    ; otherwise ->
        impl_trans_term(Start,_,Next),
        find_plugin_inits(Next,Id,Rest,StateId)).
	
set_trace_to_init(Initstate) :-
	animation_mode(Mode),
	find_trace_to_init(Mode,Initstate,Trace),!,
	state_space_reset,
	executetrace(Trace).
find_trace_to_init(Mode,Target,[root,Target]) :-
	mode_one_step(Mode).
find_trace_to_init(Mode,Target,[root|Trace]) :-
	mode_cst_init(Mode),
	find_init(root,Target,Trace).
find_trace_to_init(Mode,Target,Trace) :-
    is_registered_plugin(Mode),
    find_plugin_inits(root,Mode,Trace,Target).
                  

%*******************************************************************************
% do some preprocessing on a formula:
%  - insert current state id
%  - rewrite CSP declarations in some cases

preprocess_formula(In,Out) :-
    ( preprocess_formula2(In,Out) -> true
    ; otherwise ->
        add_error(ltl,'preprocessing formula failed',In),fail).

preprocess_formula2(ap(A),ap(B)) :- !,preprocess_ap(A,B).
preprocess_formula2(action(A),action(B)) :- !,preprocess_tp(A,B).
preprocess_formula2(ena(SpecA,FA),action(SpecB,FB)) :-
    !,preprocess_tp(SpecA,SpecB),preprocess_formula2(FA,FB).
preprocess_formula2(Term,NewTerm) :-
	functor(Term,Functor,Arity),
	functor(NewTerm,Functor,Arity),
	preprocess_formula_args(Arity,Term,NewTerm).

preprocess_formula_args(0,_,_) :- !.
preprocess_formula_args(N,Term,NewTerm) :-
	arg(N,Term,Arg),
	arg(N,NewTerm,NewArg),
	preprocess_formula2(Arg,NewArg),
	N2 is N-1,
	preprocess_formula_args(N2,Term,NewTerm).

preprocess_ap(current,stateid(StateId)) :- !,current_state_id(StateId).
preprocess_ap(enabled(A),enabled(B)) :- !,preprocess_tp(A,B).
preprocess_ap(AP,AP).

preprocess_tp(csp(Channel,Arity,CspArgs),
              bop(Channel,NumberArgs,NumberResults,ArgPatterns,ResPatterns)) :-
    csp_with_bz_mode,
    b_get_machine_operation(Channel,Results,Parameters,_Body),
    !,
    % check if more arguments are specified that allowed
    length(Parameters,NumberArgs),
    length(Results,NumberResults),
    MaxNumberAllowed is NumberArgs + NumberResults,
    ( Arity > MaxNumberAllowed ->
        add_error(ltl,'Too many arguments for LTL/CTL channel', Channel),fail
    ; Arity =< NumberResults ->
        ParamArgs = CspArgs,ResArgs = []
    ; otherwise ->
        MinArgs is min(NumberArgs,Arity),
        append_length(ParamArgs,ResArgs,CspArgs,MinArgs)),
    csp_arguments(ParamArgs,1,Parameters,ArgPatterns,Exprs1),
    csp_arguments(ResArgs,1,Results,ResPatterns,Exprs2),
    append(Exprs1,Exprs2,Exprs), 
    parse_preds_exprs(Exprs).
preprocess_tp(TP,TP).

csp_arguments([],_Index,_Parameters,[],[]).
csp_arguments([P|ParamArgs],Index,[Param|ParamRest],ArgPatterns,Exprs) :-
    ( is_csp_placeholder(P) ->
        ArgPatterns = ArgPatternsRest,
        Exprs = ExprsRest
    ; otherwise ->
        ArgPatterns = [Index/ArgPattern|ArgPatternsRest],
        Exprs = [Expr|ExprsRest],
        csp_argument(P,Param,ArgPattern,Expr)),
    I2 is Index+1,
    csp_arguments(ParamArgs,I2,ParamRest,ArgPatternsRest,ExprsRest).
csp_argument(Argument,Parameter,ParsedExpr,expr(ToParse,Type,ParsedExpr)) :-
    get_texpr_type(Parameter,Type),
    atom_codes(Argument,ToParse).

is_csp_placeholder('_').
is_csp_placeholder('?').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% callback from LTLc:

% cltl_callback(+Node, -Aps, -Out, +Queries, +Specs):
%  Callback predicate that evaluates for a state "Node" all atomic
%  propositions and returns a list of outgoing transitions.
%  It evaluates if each of the transitions met the propositions on transistions.
%  "Queries" is a list of atomic propositions (ap), "Aps" is a lists where
%  each entry is 0 resp 1 if the corresponding atomic proposition is true resp. false.
%  "Specs" is a list of propositions on transitions (tp).
%  "Out" is a list of outgoing transitions, for details, see below check_tps/4.
cltl_callback(Node, Aps, Out, Queries, Specs) :- 
	check_aps(Queries,Node,Aps),     % generate a list of 0s and 1s, depending on whether the AP is met or not
	impl_trans_term_all(Node,Trans), % find all outgoing transitions
	check_tps(Trans,Specs,Node,Out). % evaluate the TP on each transition and generate a list of transitions, each
                                     %   annotated with 0s and 1s, depending on wheterh the TP is met or not

% check list of atomic propositions (ap)
check_aps([],_,[]).
check_aps([Query|QRest],Node,[Result|RRest]) :-
	(check_ap(Query,Node) -> !,Result=1; Result=0),
	check_aps(QRest,Node,RRest).

% check list of transition properties:
% check_tps(+Transitions, +Specifications, +State, -Result):
%  "Transitions" is a list of terms of the form op(Id,Operation,NextState)
%          each term represents a transition from "State" to "NextState",
%          with the id "Id" and operation "Operation"
%  "Specifications" is a list of propositions on transistions (tp)
%  "Result" is a list where each element corresponds to a transition in "Transitions"
%          each entry has the form "trans(Next,Id,ResVector)" where "Next" and "Id"
%          are the destination state and the id of the transition and
%          "ResVector" a list where each element corresponds to a proposition (tp)
%          in "Specifications". It is 0 resp. 1 if the proposition is false resp. true.
check_tps([],_,_,[]).
check_tps([op(Id,Op,Next)|OpRest],Specs,Node,[trans(Next,Id,Tp)|TpRest]) :-
	check_tps2(Specs,Node,Next,Op,Tp),
	check_tps(OpRest,Specs,Node,TpRest).
check_tps2([],_,_,_,[]).
check_tps2([Spec|SRest],SrcNode,DstNode,Op,[Result|RRest]) :-
	(check_transition(Spec,Op,SrcNode,DstNode) -> !,Result=1; Result=0),
	check_tps2(SRest,SrcNode,DstNode,Op,RRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parse a file with several LTL formulas:

parse_ltlfile(Filename, Formulas) :-
	( parse_ltlfile2(Filename,Formulas) -> true
	; otherwise ->
	    add_error(ltl, 'Error while parsing the LTL formula file')).
parse_ltlfile2(Filename, Formulas) :-
	read_file(Filename,Codes),
	parse_lines(Lines,Codes,[]),
	parse_sections(Raws,Lines,[]),
	parse_file_formulas(Raws,Formulas).
parse_file_formulas(In,Out) :-
	extract_formulas(In,Strings,Parsed,Out),
	temporal_parser_l(Strings,ltl,Parsed).

extract_formulas([],[],[],[]).
extract_formulas([formula(Name,Raw)|Rin],[Raw|Rraw],
		 [Parsed|Rparsed],[formula(Name,Parsed)|Rout]) :-
	extract_formulas(Rin,Rraw,Rparsed,Rout).

/*
parse_file2([]) --> [].
parse_file2([F|Rest]) -->
	parse_section(F),
	parse_file2(Rest). */

parse_sections([formula(Name,F)|Rest]) -->
    munch_comments,
	[header(Name)],
	parse_section_formulas("",FCodes),!,
	{ atom_codes(F,FCodes) },
	parse_sections(Rest).
parse_sections([]) --> [].

munch_comments --> [comment(_)],!,munch_comments.
munch_comments --> [].

parse_section_formulas(Prev,F) -->
    munch_comments,
	[formula(Snippet)],!,
	{(Prev==[] -> S2=Snippet; append("\n",Snippet,S2) ),append(Prev,S2,F2)},
	parse_section_formulas(F2,F).
parse_section_formulas(F,F) --> munch_comments,[].

parse_lines([]) --> [].
parse_lines([L|Rest]) -->
	parse_line(L),
	parse_lines(Rest).

parse_line(comment(C)) -->
	pl_spaces,
	"#",!,
	pl_until_newline(C).
parse_line(comment("")) -->
	pl_spaces,
	pl_newline,!.
parse_line(header(Title)) -->
	pl_title(Title),
	pl_spaces, pl_newline, !.
parse_line(formula([C|F])) -->
	% no empty formulas
	[C],
	pl_until_newline(F).

pl_title(Title) -->
	"[", pl_spaces, pl_title2(T), {atom_codes(Title,T)}.
pl_title2([]) --> pl_spaces,"]",!.
pl_title2([]) --> pl_newline,!,{fail}. % maybe an exception?
pl_title2([C|Rest]) --> [C],pl_title2(Rest).

pl_spaces --> " ",!,pl_spaces.
pl_spaces --> "\t",!,pl_spaces.
pl_spaces --> [].

pl_until_newline([]) --> pl_newline,!.
pl_until_newline([C|Rest]) --> [C],!,pl_until_newline(Rest).
pl_until_newline([]) --> [].

pl_newline --> "\n\r",!.
pl_newline --> "\n",!.
pl_newline --> "\r".

read_file(Filename,Codes) :-
	open(Filename,read,S),
	read_codes(S,Codes),
	close(S).
read_codes(S,Codes) :-
	get_code(S,Code),
	( Code < 0 ->
	    Codes = []
	; otherwise ->
	    Codes = [Code|Rest],
	    read_codes(S,Rest)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parse an LTL or CTL formula
% Mode: ltl or ctl
temporal_parser(Formula,Mode,Result) :-
    temporal_parser_l([Formula],Mode,[Result]).

temporal_parser_l(Formulas,Mode,Result) :-
    % call the Java parser for LTL
    call_ltl_parser(Formulas,Mode,Res1),
    % check if everything was fine
    check_syntax_result_l(Res1,Parsed,ParseFlag,Mode), ParseFlag=ok,
    % typecheck atomic propositions and such things
    typecheck_temporal_formula_l(Parsed,Result,TypeFlag), TypeFlag=ok.

check_syntax_result_l([],[],_,_).
check_syntax_result_l([R|Rrest],[F|Frest],Res,Mode) :-
    check_syntax_result(R,F,Res,Mode),
    check_syntax_result_l(Rrest,Frest,Res,Mode).
check_syntax_result(ltl(F),F,_,_Mode).
check_syntax_result(syntax_error(E),_,error,Mode) :-
    add_error(Mode,'Syntax error: ',E).

typecheck_temporal_formula_l([],[],_).
typecheck_temporal_formula_l([F|Frest],[R|Rrest],Res) :-
    typecheck_temporal_formula(F,R,Res),
    typecheck_temporal_formula_l(Frest,Rrest,Res).
typecheck_temporal_formula(ap(AP),ap(Typed),Res) :-
    !,typecheck_ap(AP,Typed,Res).
typecheck_temporal_formula(action(TP),action(Typed),Res) :-
    !,typecheck_tp(TP,Typed,Res).
typecheck_temporal_formula(ena(TP,F),ena(Typed,R),Res) :-  % used in CTL only
    !,typecheck_tp(TP,Typed,Res),
    typecheck_temporal_formula(F,R,Res).
typecheck_temporal_formula(F,R,Res) :-
    F =.. [Functor|InArgs],
    same_length(InArgs,OutArgs),
    R =.. [Functor|OutArgs],
    typecheck_temporal_formula_l(InArgs,OutArgs,Res).

typecheck_ap(BasicAP,AP,_Res) :- is_basic_ap(BasicAP),!,AP=BasicAP.
typecheck_ap(enabled(TP),enabled(R),Res) :-
    !,typecheck_tp(TP,R,Res).
typecheck_ap(bpred(Raw),bpred(Typed),Res) :-
    b_type_expression(Raw,[variables],pred,Typed,Errors),
    add_all_perrors(Errors,[],typecheck_atomic_property),
    ( Errors == [] -> true
    ; otherwise    -> Res=error).
typecheck_ap(unparsed_pred(Atom),promela_pred(Predicate),ResultFlag) :-
    animation_mode(promela),!,
    atom_codes(Atom,Codes),
	( parse_promela_strings([],[Codes],[],[Predicate]) ->
        true
    ; otherwise ->
        add_error(ltl, 'Failed to parse Promela predicate ', Atom),
        ResultFlag = error).
typecheck_ap(unparsed_pred(Atom),plugin_predicate_check(Id,Predicate),ResultFlag) :-
    active_plugin(Id),!,
    atom_codes(Atom,Codes),
    ( plugin_parse(Id,[],[Codes],[],[],[Predicate],[]) -> true
    ; otherwise                                        -> ResultFlag=error).

is_basic_ap(deadlock).
is_basic_ap(sink).
is_basic_ap(stateid(_)).

typecheck_tp(unparsed_trans(Atom),plugin_transition_check(Id,Predicate),ResultFlag) :-
    active_plugin(Id),!,
    atom_codes(Atom,Codes),
    ( plugin_parse(Id,[],[],[Codes],[],[],[Predicate]) -> true
    ; otherwise                                        -> ResultFlag=error).
typecheck_tp(unparsed_trans(Op),xtl(Opname,Arity,Args),ResultFlag) :-
    animation_mode(xtl), !,
    ( split_atom(Op,['.'],[Opname|Args]) -> 
        length(Args,Arity)
    ; otherwise ->
        add_error(ltl,'Failed to parse operation specification '),
        ResultFlag = error).
typecheck_tp(unparsed_trans(Opname),csp(Channel,Arity,ResArgs),ResultFlag) :-
    csp_mode, !,
    ((atom_codes(Opname,L),tools: exact_member(95,L)) -> 
	% 95 for '_' (_ always matches)
	ParserFlag = false
    ; otherwise ->
	ParserFlag = true
    ),
    ( (Opname = tick) -> Channel=Opname,Args = [], ResArgs=[]
    ; (Opname = tau)  -> Channel=Opname,Args = [], ResArgs=[]
    ; (split_atom(Opname,['.','!'],[Channel|Args]), channel(Channel,_)) ->
	( (channel_type_list(Channel, TypeList),length(TypeList,Arity),length(Args,Arity)) -> 
		( (Args = [])  -> 
			ResArgs = []
		; otherwise ->
			mergeArgsToDotTuple([Channel|Args],DotTupleArgs),
			(ParserFlag -> 
				parse_single_csp_expression(DotTupleArgs,ParsedArgs,ltl),
				splitToSingleArgs(ParsedArgs,[Channel|ResArgs])
			; otherwise -> 
				ResArgs = Args
			)
		)
	; otherwise -> add_error(ltl,'Number of arguments not correct ', Opname),ResultFlag=error
	)
    ; otherwise ->
        add_error(ltl,'Channel does not exist: ',Opname),
        ResultFlag = error, Args=[]
    ).
typecheck_tp(bop(oppattern(_Pos,Name,Args)),Result,ResultFlag) :-
    ( b_get_machine_operation(Name,Results,Parameters,_) ->
        Result = bop(Name,NumberArgs,NumberResults,ArgPatterns,ResPatterns),
        length(Parameters,NumberArgs), length(Results,NumberResults),
        % Pattern matching for results is not yet supported
        ResPatterns = [],
        ( Args=[] ->
            ArgPatterns = [],
            ResPatterns = []
        ; length(Args,NumberArgs) ->
            create_b_op_patterns(Args,Parameters,ArgPatterns,ResultFlag)
        ; otherwise ->
            add_error(ltl,'Number of arguments not correct', Name),ResultFlag=error)
    ; csp_mode, Args=[] ->
        print(unparsed_trans(Name)),nl,
        typecheck_tp(unparsed_trans(Name),Result,ResultFlag)
    ; otherwise ->
        add_error(ltl,'Could not find operation ', Name),ResultFlag=error).
typecheck_tp(btrans(event(Name)),btrans(event(Name)),_ResultFlag) :- !.
typecheck_tp(btrans(event(Name,RawPredicate)),btrans(event(Name,TPredicate,Poststate)),ResultFlag) :-
    ( b_get_machine_operation(Name,_,_Parameters,_) ->
        ( bmachine:type_with_errors(RawPredicate,[prepost,operation(Name)],pred,TPredicate) ->
            find_poststate_access(TPredicate,Poststate)
        ; otherwise -> ResultFlag=error)
    ; otherwise ->
        add_error(ltl,'Could not find operation ', Name),ResultFlag=error).

find_poststate_access(TExpr,Poststate) :-
    find_poststate_access2(TExpr,P,[]),
    sort(P,Poststate).
find_poststate_access2(TExpr) -->
    {syntaxtraversion(TExpr,Expr,_,Infos,Subs,_)},
    find_poststate_access3(Expr,Infos),
    find_poststate_access_l(Subs).
find_poststate_access3(identifier(Id),Infos) -->
    {memberchk(poststate,Infos),!,atom_concat(Orig,'\'',Id)},
    [poststate(Orig,Id)].
find_poststate_access3(_,_) --> [].
find_poststate_access_l([]) --> !.
find_poststate_access_l([TExpr|Rest]) -->
    find_poststate_access2(TExpr),
    find_poststate_access_l(Rest).
        

create_b_op_patterns(Args,Parameters,ArgPatterns,Res) :-
    create_b_op_patterns2(Args,Parameters,ArgPatterns,1,Res).
create_b_op_patterns2([],[],[],_,_).
create_b_op_patterns2([Arg|ARest],[Param|ParamRest],Patterns,Index,Res) :-
    create_b_op_pattern(Arg,Param,Pattern,Res),
    ( var(Pattern) -> Patterns = PRest
    ; otherwise -> Patterns = [Index/Pattern|PRest]),
    Index2 is Index+1,
    create_b_op_patterns2(ARest,ParamRest,PRest,Index2,Res).
create_b_op_pattern(undef(_Pos),_,_,_).
create_b_op_pattern(def(_Pos,Raw),TParam,Typed,Res) :-
    get_texpr_type(TParam,Type),
    b_type_expression(Raw,[variables],Type,Typed,Errors),
    add_all_perrors(Errors,[],typecheck_operation_argument),
    ( Errors == [] -> true
    ; otherwise    -> Res=error).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%

ltl_formula_structure(Formula,Atomics,Structure) :-
    c_ltl_aptp(Formula,ApTp),
    pp_formulas(ApTp,Atomics),
    ltl_structure(Formula,ApTp,Structure).
ltl_structure(ap(AP),Subformulas,ap(N)) :-
    !,nth0(N,Subformulas,ap(AP)),!.
ltl_structure(action(TP),Subformulas,tp(N)) :-
    !,nth0(N,Subformulas,action(TP)),!.
ltl_structure(F,ApTp,N) :-
    F =.. [Functor|Args],
    same_length(Args,NewArgs),
    N =.. [Functor|NewArgs],
    ltl_structure_l(Args,ApTp,NewArgs).
ltl_structure_l([],_,[]).
ltl_structure_l([F|Frest],ApTp,[N|Nrest]) :-
    ltl_structure(F,ApTp,N),
    ltl_structure_l(Frest,ApTp,Nrest).

ltl_model_check_with_ce(Formula,Max,Mode,Result) :-
    ltl_model_check3(Formula,Max,Mode,R1),
    ce_result(R1,Formula,Result).

ce_result(nostart,_,nostart) :- !.
ce_result(none,_,ok) :- !.
ce_result(incomplete,_,incomplete) :- !.
ce_result(model(CE,LoopEntry),Formula,counterexample(Counterexample,LoopEntry,PathToCE)) :- !,
    ce_result_initpath(CE,PathToCE),
    c_ltl_aptp(Formula,ApTp),
    length(ApTp,N),
    convert_counterexample(CE,N,Counterexample).
ce_result(Result,_,_) :-
    add_error(ltl,'Internal error: Unexpected result', Result),
    fail.

ce_result_initpath(CE,Path) :-
    ce_get_first_state(CE,InitState),
    find_init(root,InitState,StatePath),
    find_operation_ids(StatePath,root,Path).
find_operation_ids([],_,[]).
find_operation_ids([Dst|Srest],Src,[(Id,Op,Dst)|Irest]) :-
    transition(Src,Op,Id,Dst),
    find_operation_ids(Srest,Dst,Irest).

/* just restrict the evaluation bits to the atomic formulas */
convert_counterexample([],_N,[]).
convert_counterexample([atom(State,Evals,NextOp) |Arest],N,
                       [atom(State,NEvals,OpTuple)|Erest]) :-
    prefix_length(Evals,NEvals,N),
    ( NextOp = none          -> OpTuple = none
    ; otherwise              -> transition(State,Action,NextOp,Dst),OpTuple=(NextOp,Action,Dst)),
    convert_counterexample(Arest,N,Erest).

:- use_module(probsrc(translate)).
pp_op(Id,PP) :-
    transition(_Src,Action,Id,_Dst),!,
    translate:translate_event(Action,PP).
    
pp_formulas([],[]).
pp_formulas([ApTp|Rest],[PP|PPrest]) :-
    ( pp_formula(ApTp,PP) -> true
    ; otherwise -> PP = '???'),
    pp_formulas(Rest,PPrest).
pp_formula(ap(AP),ap(PP)) :-
    pp_ap(AP,PP).
pp_formula(action(AP),tp(PP)) :-
    pp_tp(AP,PP).

pp_ap(bpred(B),PP) :- !,
    translate:translate_bexpression(B,PP1),
    ajoin(['{',PP1,'}'],PP).
pp_ap(enabled(TP),PP) :- !,
    pp_tp2(TP,TPP),
    ajoin(['e(',TPP,')'],PP).
pp_ap(Expr,PP) :-
    write_to_codes(Expr,Codes),
    atom_codes(PP,Codes).

pp_tp(T,PP) :-
    pp_tp2(T,PP1),
    ajoin(['[',PP1,']'],PP).
pp_tp2(btrans(event(Name)),Name) :- !.
pp_tp2(btrans(event(Name,Predicate,_)),PP) :- !,
    translate:translate_bexpression(Predicate,PP1),
    ajoin([Name,' | ',PP1],PP).
pp_tp2(_,'???').
