% (c) 2009-2012 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(parsercall, [load_b_machine/2,parse_exprs_preds/4,parse_exprs_preds/5,
                       load_b_machine_probfile/2, % a way to directly load a .prob file
                       get_parser_version/1,call_ltl_parser/3,
                       call_tla2b_parser/1, tla2b_filename/2,
                       call_promela_parser/1, promela_prolog_filename/2]).

:- use_module(tools).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module takes care of calling the Java B Parser if necessary.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).
:- use_module(library(process)).
:- use_module(library(file_systems)).
:- use_module(library(codesio)).

:- use_module(bparser(parser_call),[loadMachineHierarchy/2,loadMachineHierarchyPedantic/2]).
:- use_module(error_manager,[add_error/2,add_error/3,add_failed_call_error/1]).
:- use_module(self_check).
:- use_module(preferences).
:- use_module(tools,[ajoin/2,host_platform/1, split_filename/3, libraries_as_classpath/2, print_error/1]).
:- use_module(debug).

:- dynamic java_parser_version/1.
:- dynamic parser_location/1.

load_b_machine(Filename, Machine) :-
	get_preference(selected_bparser,Parser),
	(Parser=java   -> load_b_machine_java(Filename,Machine)
	;Parser=prolog -> loadMachineHierarchy(Filename,Machine)
	;Parser=all    -> loadMachineHierarchyPedantic(Filename,Machine)
        ; otherwise ->  add_error(bmachine,'Unknown alternative for preference selected_bparser ',Parser),fail
	).
	    
% throws parse_errors(Errors) in case of syntax errors
load_b_machine_java(Filename, Machine) :- %print(load_b_machine(Filename)),nl,
	prob_filename(Filename, ProBFile),
	debug_println(9,prob_filename(Filename,ProBFile)),
	( dont_need_compilation(Filename,ProBFile,Machine,LoadResult) ->
	    % no compilation is needed
	    (LoadResult=loaded -> debug_println(9,'.prob file up-to-date')
	      ; print_error('*** Loading failed *** '),nl,fail)
	; otherwise ->
	    debug_println(9,'Calling Parser'),
	    (debug:debug_mode(on)
	     -> time(call_console_parser(Filename,ProBFile))
	     ;  call_console_parser(Filename,ProBFile)
	    ),
	    load_b_machine_probfile(ProBFile, Machine)
	).

% converts the filename of the machine into a filename for the parsed machine
prob_filename(Filename, ProB) :-
	split_filename(Filename,Basename,_Extension),
	safe_atom_chars(Basename,BasenameC,prob_filename1),
	append(BasenameC,['.','p','r','o','b'],ProBC),
	safe_atom_chars(ProB,ProBC,prob_filename2),!.
prob_filename(Filename, ProB) :-
    add_failed_call_error(prob_filename(Filename,ProB)),fail.

check_version(ProBFile, Version, Machine) :-
    (Version='$UNKNOWN' -> print('*** Unknown parser version number. Trying to load parsed file.'),nl
                          /* keep CheckVersionNr as variable */
     ; CheckVersionNr=Version),
    read_machine_from_file(ProBFile,parser_version(CheckVersionNr),Machine).


read_machine_from_file(ProBFile,VersionTerm,complete_machine(Main,Machines,Files,Pragmas)) :-
    open(ProBFile, read, S),
    call_cleanup(( read(S,VersionTerm),
                   read(S,classical_b(Main,Files)), read_machines(S,Machines,Pragmas)),
                 close(S)).

load_b_machine_probfile(ProBFile, Machine) :-
	debug_println(9,consulting(ProBFile)),
    ( read_machine_from_file(ProBFile,_VersionTerm,Machine) -> true
    ; otherwise ->
        add_error(parsercall,'Failed to read parser output: wrong format?',ProBFile),fail).

read_machines(S,Machines,Pragmas) :-
    read(S,Term),
    ( Term == end_of_file ->
        Machines = [],
        Pragmas = []
    ; Term = machine(M) ->
        Machines = [M|Mrest],
        read_machines(S,Mrest,Pragmas)
    ; Term = pragma(_NodeID,_PragmaName,_Arguments,_Warnings,_FileNo,_SLine,_SCol,_ELine,_ECol) ->
        append([Term],Prest,Pragmas),
        read_machines(S,Machines,Prest)
    ; Term = global_pragma(_PragmaName,_Arguments,_Warnings,_FileNo,_Sline,_SCol,_ELine,_ECol,_ASTNodes) ->
        append([Term],Prest,Pragmas),
        read_machines(S,Machines,Prest)
    ).

% checks if all files that the prob-file depends on
% are older than the prob-file
dont_need_compilation(Filename,ProB,Machine,LoadResult) :-
	file_exists(Filename),
	catch( (   get_parser_version(Version),
			   file_property(Filename, modify_timestamp, BTime),
			   debug_print(9,'Parser Version: '), debug_println(9,Version),
			   debug_println(9,' Filename Timestamp: '), debug_print(9,Filename), debug_print(9,' : '),debug_println(9,BTime),
			   file_exists(ProB), % the .prob file exists
			   file_property(ProB, modify_timestamp, ProbTime),
			   debug_println(9,' .prob Timestamp: '), debug_print(9,ProB), debug_print(9,' : '),debug_println(9,ProbTime),
			   % print(time_stamps(ProB,ProbTime,Filename,BTime)),nl,
			   BTime < ProbTime, % .mch file is older than .prob file
			   check_version(ProB, Version, Machine),
			   Machine = complete_machine(_,_,Files,_Pragmas),
			   % print(checking_all_older(Files,ProbTime)),nl,
			   all_older(Files, ProbTime),
			   LoadResult = loaded),
	       error(A,B),
	       (debug_println(9,error(A,B)),
	        (A=resource_error(memory)
	          -> add_error(load_b_machine,'Insufficient memory to load file.\nTry starting up ProB with more memory\n (e.g., setting GLOBALSTKSIZE=500M before starting ProB).'),
	             LoadResult=A
	          ; fail))).
all_older([],_).
all_older([File|Rest],PTime) :- % print(checking(File)),nl,trace,
    (file_exists(File) -> true ; print('File does not exist anymore: '), print(File),nl,fail),
	file_property(File, modify_timestamp, BTime),
	debug_print(9,' Subsidiary File Timestamp: '), debug_print(9,File), debug_print(9,' : '),debug_println(9,BTime),
	( BTime < PTime -> true
    ; print('File has changed : '), print(File), print(' : '), print(BTime),nl,fail ),
	all_older(Rest,PTime).

:- use_module(system_call).

get_jvm_options(XOpts) :- get_preference(use_large_jvm_for_parser,true),!,
    XOpts = ['-Xms300m','-Xmx1500m','-Xss5m'].
get_jvm_options(['-Xmx512m','-Xss2m']).

call_console_parser(Filename, ProB) :-
	get_java_command(JavaCmd,Classpath),
	replace_windows_path_backslashes(Filename,FilenameW),
	replace_windows_path_backslashes(ProB,ProBW),

    get_jvm_options(XOpts),
    append(XOpts, ['-cp',Classpath,'de.prob.cliparser.CliBParser',
                      %% '-v', % comment in for verbose
				      '-prolog','-lineno','-out',ProBW,FilenameW], FullOpts),
	debug_println(9,system_call_parser(JavaCmd,FullOpts)),
	%print('java '), print(FullOpts),nl,
	system_call(JavaCmd, FullOpts,Text,JExit),
	( JExit = exit(0) ->
	    true
	; memberchk(JExit,[exit(253),exit(254),exit(255),exit(-1),exit(-2),exit(-3)]) ->
	    catch( read_from_codes(Text,Term),
		   _,
		   (name(T,Text),Term = exception(T))),
	    ( Term = io_exception(Msg) ->
		    add_error(parsercall,'Error while opening the B file: ',Msg)
	    ; Term = preparse_exception(Tokens,Msg) ->
		    findall(error(Msg,Pos),member(Pos,Tokens),Errors1),
			( Errors1 = [] ->
			    /* if there are no error tokens: try and extract pos. info from text Msg */
				Errors = [error(Msg,Pos)],
				extract_position_info(Msg,Pos)
			; otherwise ->
				Errors = Errors1),
		     throw(parse_errors(Errors))
	    ; Term = parse_exception(Pos,Msg) ->
		     throw(parse_errors([error(Msg,Pos)]))
	    ; Term = exception(Msg) ->
		     %add_error(bmachine,'Error in the classical B parser: ', Msg),
		     extract_position_info(Msg,Pos), throw(parse_errors([error(Msg,Pos)]))
	    ; otherwise ->
		     name(Msg,Text),
		     %add_error(bmachine,'Error in the classical B parser: ', Msg),
		     extract_position_info(Msg,Pos), throw(parse_errors([error(Msg,Pos)]))
		),
	    fail
	; otherwise ->
	    name(T,Text),
	    add_error(bmachine,'Error while parsing machine: ',T),
	    fail).

extract_position_info(Text,pos(Row,Col,unknown)) :-
	atom_codes(Text,Codes),
	epi(Row,Col,Codes,_),!.
extract_position_info(_Text,none).
epi(Row,Col) --> "[", epi_number(Row),",",epi_number(Col),"]".
epi_number(N) --> epi_numbers(Chars),{name(N,Chars)}.
epi_numbers([C|Rest]) --> [C],{is_number(C)},epi_numbers2(Rest).
epi_numbers2([C|Rest]) --> [C],{is_number(C),!},epi_numbers2(Rest).
epi_numbers2("") --> !.
is_number(C) :- member(C,"0123456789"),!.


get_parser_version(Version) :- java_parser_version(Version),!.
get_parser_version(Version) :-
	get_version_from_parser(Version),
	assert(java_parser_version(Version)),!.
get_parser_version(_Vers) :- add_error(get_parser_version,'Could not get parser version'),fail.

get_version_from_parser(Version) :-
	get_java_command(JavaCmd,Classpath),
	debug_println(9,calling_parser_version(JavaCmd,Classpath)),
	process_create(JavaCmd, ['-cp',Classpath,'de.prob.cliparser.CliBParser','-version'],
		       [process(Java),stdout(pipe(JStdOut))]),
	read_line(JStdOut,Text),
	process_wait(Java,JExit),
	( JExit = exit(0) ->
	    ( Text = end_of_file ->
		add_error(parsercall, 'parser returned no version number'),
		fail
	    ; otherwise ->
		length(Text,Length),
		( Length < 100 -> atom_codes(Version,Text)
		; otherwise ->
		    prefix_length(Text,Prefix,100),
		    append(Prefix,"...",Full),
		    name(Msg,Full),
		    add_error(parsercall, 'parser returned unexpected string', Msg),
		    fail))
	; otherwise ->
	    add_error(parsercall, 'parser failed when trying to get version number'),
	    Version='$UNKNOWN').

get_java_command(JavaCmdW,ClasspathW) :-
    FileList = ['probcliparser.jar'], 
    (parser_location(Classpath) -> true;
	         otherwise -> libraries_as_classpath(FileList,Classpath)), 
	get_java_command_path(JavaCmdW),
	replace_windows_path_backslashes(Classpath,ClasspathW).

get_java_command_path(JavaCmdW) :- 
	absolute_file_name(path(java),
				JavaCmd,
				[access(exist),extensions(['.exe','']),solutions(all),file_errors(fail)]),
	
	replace_windows_path_backslashes(JavaCmd,JavaCmdW),!.
get_java_command_path(_JavaCmd) :-
    add_error(get_java_command_path,'Could not get path to java command. Make sure a Java Runtime is installed',''),
    fail.

call_ltl_parser([], _CtlOrLtl, []) :- !.
call_ltl_parser(Formulas, CtlOrLtl, Result) :-
    get_ltl_lang_spec(LangSpec),print(LangSpec),nl,
    generate_formula_codes(Formulas,Codes,""),
    write_string_to_file('ltl_formulas.txt',LtlFilename,Codes),

    get_mode_opt(CtlOrLtl,ModeOpt),
	get_java_command(JavaCmd,Classpath),
    Opts = ['-cp',Classpath,'de.prob.cliparser.LtlConsoleParser',ModeOpt,'-lang',LangSpec,LtlFilename],
    
	print(calling_ltl_parser(JavaCmd,Opts)),nl,
    process_create(JavaCmd, Opts,[process(Process),stdout(pipe(JStdOut))]),
    read(JStdOut,Result),
	process_wait(Process,_ExitCode).

get_mode_opt(ltl,'-ltl').
get_mode_opt(ctl,'-ctl').

:- use_module(specfile,[b_or_z_mode/0,csp_with_bz_mode/0]).
get_ltl_lang_spec('B,none') :- csp_with_bz_mode,!.
get_ltl_lang_spec('B') :- b_or_z_mode,!.
get_ltl_lang_spec(none).

generate_formula_codes([F]) -->
    !,generate_formula_codes2(F),"\n".
generate_formula_codes([F|Rest]) -->
    generate_formula_codes2(F), "###\n",
    generate_formula_codes(Rest).
generate_formula_codes2(F) -->
    write_to_codes(F).


% ---------------------------

% call the TLA2B parser to translate a TLA file into a B machine
call_tla2b_parser(TLAFile) :-
   get_java_command_path(JavaCmdW),
   replace_windows_path_backslashes(TLAFile,TLAFileW),
   get_jvm_options(XOpts),
   absolute_file_name(prob_lib('TLA2B.jar'),TLA2BJAR),
   append(XOpts,['-jar',TLA2BJAR,TLAFileW],FullOpts),
   debug_println(9,calling_java(FullOpts)),
   (my_system_call(JavaCmdW, FullOpts,tla2b_parser) -> true
     ; print_error('Be sure to download TLA2B.zip from http://nightly.cobra.cs.uni-duesseldorf.de/tla,'),
       print_error('then unzip the archive and put TLA2B.jar into the ProB lib/ directory.'),
       fail).

tla2b_filename(TLAFile,BFile) :-
	split_filename(TLAFile,Basename,_Extension),
	atom_chars(Basename,BasenameC),
	append(BasenameC,['_',t,l,a,'.','m','c','h'],TLABC),
	atom_chars(BFile,TLABC),!.
tla2b_filename(Filename, BFile) :-
    add_failed_call_error(tla2b_filename(Filename,BFile)),fail.
     

my_system_call(Command,Options,Origin) :-
   system_call(Command,Options,Text,Exit),
   treat_exit_code(Exit,Command,Options,Text,Origin).
treat_exit_code(exit(0),_,_,_,_) :- !.
treat_exit_code(Exit,Command,Options,Text,Origin) :-
   debug_println(9,treat_exit_code(Exit,Command,Options,Text,Origin)),
   catch( read_from_codes(Text,Term),
		   _,
		   (name(T,Text),Term = T)), !,
	print_error(Term),
	add_error(Origin,'Error occured during sytem_call: ',Term),
	fail.

    
% replace backslashes by forward slashes in atoms, only under windows
% TODO(DP,14.8.2008): Check if still needed with SICStus 4.0.4 and if
% everybody uses that version
replace_windows_path_backslashes(Old,New) :-
	( host_platform(windows) ->
	    name(Old,OldStr),
	    replace_string_backslashes(OldStr,NewStr),
	    name(New,NewStr)
	; otherwise ->
	    Old=New).

replace_string_backslashes([],[]) :- !.
replace_string_backslashes([C|OldRest],[N|NewRest]) :- !,
	( C=92 /* backslash */ ->
	    N=47 /* forward slash */
	; otherwise ->
	    N=C ),
	replace_string_backslashes(OldRest,NewRest).
replace_string_backslashes(X,Y) :-
  add_error(parsercall,'Illegal call: ',replace_string_backslashes(X,Y)),
  fail.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% parse expressions and predicates

parse_exprs_preds(ExprStrings,Exprs,PredStrings,Preds) :-
    parse_exprs_preds(ExprStrings,Exprs,PredStrings,Preds,[]).
parse_exprs_preds(ExprStrings,Exprs,PredStrings,Preds,DefStrings) :-
	make_dummy_machine(ExprStrings,PredStrings,DefStrings,Machinetext),
	write_string_to_file('prob_parser_temp.mch',Filename,Machinetext),
	%% print('FILE:'),nl, atom_codes(NN,Machinetext),print(NN),nl, %%
	prob_filename(Filename,ProBFilename),
	catch( ( load_b_machine(Filename,PMachine),
		     extract_preds_exprs(PMachine,Preds,Exprs)),
	         parse_errors(Errors),
	       ( %absolute_file_name(Filename,ABS), print(abs(ABS)),nl,
	         delete_file(ProBFilename), delete_file(Filename),%% 
	         %print(errors(Errors)),nl,
		     map_source_positions(Errors,NewErrors),
		     throw(parse_errors(NewErrors)))),
	delete_file(Filename),
	delete_file(ProBFilename).

write_string_to_file(BaseName,Filename,Machine) :-
	open(temp(BaseName),write,S,[if_exists(generate_unique_name)]),
	stream_property(S, file_name(Filename)),
	print_string(Machine,S),
	close(S).
print_string([],_).
print_string([C|Rest],S) :- put_code(S,C),print_string(Rest,S).

extract_preds_exprs(complete_machine(_,[abstract_machine(_,_,_,M)],_,_Pragmas),Preds,Exprs) :-
	findall(P,extract_pred(M,P),Preds),
	findall(E,extract_expr(M,E),Exprs).
extract_pred(Machine,P) :-
	member(operations(_,Ops),Machine),
	member(operation(_,_Id,_Res,_Params,precondition(_,P,_)),Ops).
extract_expr(Machine,E) :-
	member(operations(_,Ops),Machine),
	member(operation(_,_Id,_Res,_Params,block(_,assign(_,_Lhs,Rhs))),Ops),
	member(E,Rhs).

map_source_positions([],[]).
map_source_positions([error(Msg,Pos)|ORest],[error(Msg,NewPos)|NRest]) :-
	( Pos = pos(Row,Col,FileName)
	 -> NRow is Row-3,  % there are 3 header lines in dummy machine ;  was ?? (Row-2)//2,
	    NewPos = pos(NRow,Col,FileName)
	; otherwise -> Pos=NewPos),
	map_source_positions(ORest,NRest).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% writing machine with multiple predicates

make_dummy_machine(Exprs,Preds,Defs,M) :-
	make_dummy_machine_cfg(Preds,Exprs,Defs,M,[]),!.
make_dummy_machine(Exprs,Preds,Defs,M) :- add_error(parser_call,'Call failed: ',make_dummy_machine(Exprs,Preds,Defs,M)),fail.

make_dummy_machine_cfg(Preds,Exprs,Defs) -->
	"MACHINE ltldummy\n",
	make_dummy_machine_defs(Defs),
	make_dummy_machine_ops(Preds,Exprs),
	"END\n".
make_dummy_machine_defs([]) --> !.
make_dummy_machine_defs([Def|T]) --> "DEFINITIONS\n", write_string(Def), make_dummy_machine_defs2(T).
make_dummy_machine_defs2([]) --> !.
make_dummy_machine_defs2([Def|T]) --> ";\n", write_string(Def), make_dummy_machine_defs(T).

make_dummy_machine_ops([],[]) --> !.
make_dummy_machine_ops(Preds,Exprs) -->
	"OPERATIONS\n",
	make_dummy_machine_exprs(Exprs,Preds,1),
	make_dummy_machine_preds(Preds,1).
make_dummy_machine_exprs([],_,_) --> !.
make_dummy_machine_exprs([Expr|Rest],Preds,N) -->
	"  expr", write_number(N), "= BEGIN r:=(\n",
	write_string(Expr), ") END",
	( {Rest=[],Preds=[]} -> "" ; ";" ),
	"\n",
	{N2 is N+1},
	make_dummy_machine_exprs(Rest,Preds,N2).
make_dummy_machine_preds([],_) --> "".
make_dummy_machine_preds([Ap|Rest],N) -->
	"  pred", write_number(N), "= PRE (\n",
	write_string(Ap),
	") THEN skip END",
	( {Rest=[]} -> "" ; ";" ),
	"\n",
	{N1 is N + 1},
	make_dummy_machine_preds(Rest,N1).
write_number(N) --> {name(N,NStr)}, write_string(NStr).		  
write_string([],In,In).
write_string([C|Rest],[C|In],Out) :- write_string(Rest,In,Out).

   
% ---------------------------
:- use_module(tools,[runtime_application_path/1]).
% call the Promela parser to translate a .pml file into a Prolog AST
call_promela_parser(PMLFile) :-
   promela_prolog_filename(PMLFile,PrologFile),
   (file_exists(PrologFile) -> delete_file(PrologFile) ; true),
   get_java_command_path(JavaCmdW),
   replace_windows_path_backslashes(PMLFile,PMLFileW),
   get_jvm_options(XOpts),
   absolute_file_name(prob_lib('Promela.jar'),PromelaJAR),
   runtime_application_path(AppPath),
   append(XOpts,['-jar',PromelaJAR,PMLFileW,AppPath],FullOpts),
   debug_println(9,calling_java(FullOpts)),
   my_system_call(JavaCmdW, FullOpts,promela_parser).

promela_prolog_filename(PMLFile,PrologFile) :-
	atom_chars(PMLFile,BasenameC),
	append(BasenameC,['.','p','l'],NewC),
	atom_chars(PrologFile,NewC),!.
promela_prolog_filename(Filename, PrologFile) :-
    add_failed_call_error(promela_filename(Filename,PrologFile)),fail.
    
