% (c) 2012-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(pe, [pe/2, 
               partially_evaluate/2, % partially evaluate a call marked as unfold
               partially_evaluate_and_load/2,
               partially_evaluate_fail_detect/2,
               specialized_predicate/2,
               generalize_call/3,
               clear_specialized_code/0,
               % annotations:
                unfold/1, unfold/2,
                execute/1, execute/2
              ]).

% written by Michael Leuschel
% started 15/12/2012


% a partial evaluator implemented using term expansion 
% the original program is copied unmodified, interspersed with discontiguous directives
% and unfolder clauses in the style of Logen (see Theory and Practice of Logic Programming 4(1):pp. 139-191.)

% A difference with Logen is that the term expander works directly on the source program:
%  the goal is to only selectively annotate certain points of the source program.

% Annotations are:
%  - mark predicates to be unfolded with 
%     :- unfold p(_,...).
%  - mark predicates to be fully executed with 
%     :- execute p(_,...).
%  - mark predicates to be fully executed upon a condition with:
%     :- execute(p(_,...),Condition).

% Commands:
%  - clear specialized code with clear_specialized_code
%  - specialize a predicate with partially_evaluate(Pred,R)
%      (R is the residual call to be used to call the specialized code)
% TO DO:
%  - proper module treatment [partially done]
%  - asserting and calling specialized clauses [partially done]
%  - generate memoization points to avoid backpropagation and clause duplication
%  - introduce memo annotation and treatment if necessary


:- use_module(builtin_tools).

:- multifile user:term_expansion/6.

:- use_module(library(lists)).
:- use_module(library(timeout)).


:- op(1150, fx, unfold).

%:- meta_predicate unfold(0).
unfold(MCall,Condition) :-
   get_module_context(MCall,Module,Call),
   print(unfold(Module:Call)),nl,
   assert(is_unfolded(Module,Call,Condition)). % Condition ignored at the moment
:- dynamic is_unfolded/3.
unfold(MCall) :- unfold(MCall,true).

is_unfolded(M:C) :- !, is_unfolded(M,C,_).
is_unfolded(C) :- prolog_load_context(module,M),
    is_unfolded(M,C,_).

% check if this call is unfolded in this or some other module (which could be imported)
is_unfolded_somewhere(MCall,AdaptedCall) :-
   get_module_context(MCall,Module,Call),
   is_unfolded(OtherModule,Call,_),
   (Module=OtherModule -> AdaptedCall=Module:Call
     ; AdaptedCall = OtherModule:Call,
       print('*** assuming '), print(OtherModule:Call), print(' imported in '),
       print(Module),nl
       % TO DO: keep track of imports and check that this is indeed the case !!
    ).

:- op(1150, fx, execute).
% register conditions under which some predicates should be executed directly
:- dynamic execute_predicate/3.
execute(MCall) :- 
   get_module_context(MCall,Module,Call),
   print(execute(Module,Call)),nl,
   assert(execute_predicate(Module,Call,true)).
execute(MCall,Condition) :- 
   get_module_context(MCall,Module,Call),
   print(execute(Module,Call,Condition)),nl,
   assert(execute_predicate(Module,Call,Condition)).

execute_predicate_in_load_context(Call,Cond) :- prolog_load_context(module,M),
   execute_predicate_in_context(M,Call,Cond).
execute_predicate_in_context(_,M:Call,Cond) :- execute_predicate(M,Call,Cond).
execute_predicate_in_context(Module,Call,Cond) :- execute_predicate(Module,Call,Cond).

add_module_qualifier(M:C,R) :- !, R=M:C. % already qualified
add_module_qualifier(C,M:C) :- prolog_load_context(module,M).

get_module_context(M:C,RM,RC) :- !, RM=M,RC=C. % already qualified
get_module_context(Call,M,Call) :- prolog_load_context(module,M).

:- dynamic block_declaration/4.
% TO DO: extract predicate and use as index --> block_declaration/2
% ?? Do we need a meta_predicate annotation ??
add_block_declaration(A) :- prolog_load_context(module,M),
   %print(block_declaration(A,M,TP)),nl,
   (flatten_block(A,Pred,N,BlockList)
     -> functor(Call,Pred,N),
        assert(block_declaration(Call,M,BlockList,A))
         %,print(block_declaration(Call,M,BlockList,A)),nl
      ;  print('### could not parse block declaration: '), 
         prolog_load_context(term_position,TP), % TP=$stream_position(?,?,LineNumberBefore,?)
         print((TP:A)),nl
    ).
get_block_declaration(M:Call,BlockList,A) :-
   block_declaration(Call,M,BlockList,A).

% transform a block declaration into a list of argument pattern lists
flatten_block((A,B),Pred,N,X) :- !, flatten_block(A,Pred,N,XA),
   flatten_block(B,Pred,N,XB),
   append(XA,XB,X).
flatten_block(Call,Pred,N,X) :- Call =.. [Pred|Args], length(Args,N),
   maplist(get_block_pattern,Args,BPat),
   X = [ BPat ].
% - is kept, everything else is translated to variable matching anything
get_block_pattern(X,Pat) :- (X = '-' -> Pat = '-' ; Pat = _).

% check if a call is blocked:
call_is_blocked(_M:Call,BlockList) :- !, call_is_blocked(Call,BlockList).
call_is_blocked(Call,BlockList) :-
   %print(checking_block(Call,BlockList)),nl,
   Call =.. [_Pred|Args],
   maplist(get_arg_pattern,Args,Pat),
   (member(Pat,BlockList)
    -> print(blocking(Call,Pat)),nl
     ; %print(non_blocking(Pat)),nl,
       fail
   ).

get_arg_pattern(X,Pat) :- (nonvar(X) -> Pat = nonvar ; Pat = '-').


% NOTE: block declarations can also be extracted using predicate_property
% | ?- predicate_property(test_pe:myapp(A,B,C),P).
% P = (block myapp(-,?,-)) ? 


% now: generate unfolder predicates for those predicates marked as unfoldable
user:term_expansion(':-'(block(A)), Lay1, Tokens1,
                    ':-'(block(A)), Lay1, [partial_evaluator | Tokens1]) :- 
    nonmember(partial_evaluator, Tokens1), % do not expand if already expanded
    !,
    add_block_declaration(A).
user:term_expansion(H1, Lay1, Tokens1, 
     [(:- discontiguous(P1/A1)),H1, % a copy of the original fact
      (:- discontiguous(P2/A2)),H2],  % the corresponding unfolder clause
      Lay1, [partial_evaluator | Tokens1]) :-
    %print(check(H1,Tokens1)),nl,
    nonmember(partial_evaluator, Tokens1), % do not expand if already expanded
    is_unfolded(H1),!, % fact
    get_predicate_arity(H1,_,P1,A1),
    generate_unfolder_call(H1,true,H2),
    print('  unfold fact: '), print(H2),nl,
    get_predicate_arity(H2,_,P2,A2).
user:term_expansion(':-'(H1,B1), Lay1, Tokens1, 
     [(:- discontiguous(P1/A1)),':-'(H1,B1), % a copy of the original clause
      (:- discontiguous(P2/A2)),':-'(H2,PP_B2)],  % the corresponding unfolder clause
      Lay1, [partial_evaluator | Tokens1]) :-
    nonmember(partial_evaluator, Tokens1), % do not expand if already expanded
    is_unfolded(H1),!, % rule
    get_predicate_arity(H1,_,P1,A1),
    generate_unfolder_call(H1,Code,H2), 
    compile_body(B1,B2,Code),
    post_process(B2,PP_B2),
    print(' ***  unfold rule: '),nl, portray_clause((H2:-PP_B2)),
    get_predicate_arity(H2,_,P2,A2).

get_predicate_arity(X,M,P,A) :- var(X),!,
   print('*** get_predicate_arity of variable: '),
   print(get_predicate_arity(X,M,P,A)),nl,fail.
get_predicate_arity(':'(Module,Call),Module,Pred,Arity) :- !,
   functor(Call,Pred,Arity).
get_predicate_arity(Call,Module,Pred,Arity) :-
   prolog_load_context(module,Module),
   functor(Call,Pred,Arity).

% generate an unfolding call
generate_unfolder_call(Module:Call,CodeArg,Module:NewUnfoldingCall) :- !,
    generate_unfolder2(Call,CodeArg,NewUnfoldingCall).
generate_unfolder_call(Call,CodeArg,Module:NewUnfoldingCall) :- !,
    prolog_load_context(module,Module),
    generate_unfolder2(Call,CodeArg,NewUnfoldingCall).
generate_unfolder2(Call,CodeArg,NewUnfoldingCall) :-
    Call =.. [F|Args],
    append(Args,[CodeArg],NewArgs),
    atom_concat(F,'$unfold',NewF),
    NewUnfoldingCall =.. [NewF|NewArgs].

% specialize the call built-in
specialize_call(Module,Call,Code) :- var(Call),!, Code=call(Module:Call).
specialize_call(_Module,Call,Code) :- 
    executable_builtin(Call),!,call(Call), Code=true.
specialize_call(_Module,Call,Code) :- 
    is_builtin(Call),!, Code=Call.
specialize_call(Module,Call,Code) :- %print(try_exec(Module,Call,Code)),nl,
    execute_predicate_in_context(Module,Call,Condition), % print(cond(Condition)),nl,
    call(Condition),!, % we can execute the predicate
    %print(executing(Call)),nl,
    call(Module:Call),Code=true.
% TO DO: register unfolder predicates and look if we can unfold the call
specialize_call(Module,Call,Module:Call).

% compile the body of a clause; translating calls to unfoldable predicates
compile_body(B1,B2,Code) :- var(B1),!, prolog_load_context(module,M),
    B2=pe:specialize_call(M,B1,Code).
compile_body(call(A),B2,Code) :- !, compile_body(A,B2,Code).
compile_body((A1,B1),(A2,B2),(C1,C2)) :- !, % specialize the conjunction
    compile_body(A1,A2,C1), compile_body(B1,B2,C2).
compile_body(B1,B2,Code) :- is_unfolded_somewhere(B1,MB1),!, 
    %VB2=if(B2,true,(print(unfold_failed(MB1)),nl,fail)),  %% comment in if you want to see failed execute calls
    (get_block_declaration(MB1,BlockList,_A)
      ->  generate_unfolder_call(MB1,UnfCode,UnfB2),
           B2 = (pe:call_is_blocked(MB1,BlockList)
                   -> %Code=MB1 % just residualize; maybe do this if all args are variables or unfold condition is false
                      partially_evaluate_fail_detect(MB1,Code)
                   ; Code=UnfCode, UnfB2)
      ;  generate_unfolder_call(MB1,Code,B2)
    ).
compile_body(B1,B2,Code) :- execute_predicate_in_load_context(B1,true),!,
    add_module_qualifier(B1,MB1),
    B2=MB1,
    %B2=if(MB1,true,(print(exec_failed(MB1)),nl,fail)),  %% comment in if you want to see failed execute calls
    Code=true.
compile_body(B1,B2,Code) :- execute_predicate_in_load_context(B1,Condition),!,
    add_module_qualifier(B1,MB1),
    B2=(call(Condition) -> call(MB1),Code=true ; Code=MB1).
compile_body(B1,B2,Code) :- is_builtin(B1), !, B2 = builtin_tools:specialize_builtin(B1,Code).
compile_body((Tst->Thn;Els),B2,Code) :- !,
     protected_compile_body(Tst,Tst2,C0),
     protected_compile_body(Thn,T2,C1), protected_compile_body(Els,E2,C2),
     compile_body(Thn,UPT2,UPC1), /* unprotected version for static if */
     compile_body(Els,UPE2,UPC2),
     (\+ could_be_decided(C0)
       -> %nl,nl,print(undecidable(C0,Tst,Tst2)),nl,nl,
          B2 = (Tst2, T2,E2),   Code=(C0->C1;C2)  % always dynamic if
       ;  B2 = (Tst2, pe:post_process(C0,PPC0),
                     ( PPC0==fail  -> UPE2,Code=UPC2 % static if; we can use an improved unprotected compile 
                      ; PPC0==true -> UPT2,Code=UPC1 % ditto 
                      ; T2,E2, Code=(PPC0->C1;C2)))  % dynamic if
      ).
compile_body(if(Tst,Thn,Els),B2,Code) :- !,
     protected_compile_body(Tst,Tst2,C0),
     protected_compile_body(Thn,T2,C1), protected_compile_body(Els,E2,C2),
     compile_body(Thn,UPT2,UPC1), /* unprotected version for static if */
     compile_body(Els,UPE2,UPC2),
     (\+ could_be_decided(C0)
       -> B2 = (Tst2, T2,E2),   Code=if(C0,C1,C2)
       ;  B2 = (Tst2, pe:post_process(C0,PPC0),
                      (PPC0==fail   -> UPE2,Code=UPC2  /* static if; we can use unprotected compile */
                       ; PPC0==true -> UPT2,Code=UPC1  /* ditto */
                       ; T2,E2, Code=if(PPC0,C1,C2)))
     ).
compile_body((A;B),B2,Code) :- !,
     protected_compile_body(A,T2,C1), protected_compile_body(B,E2,C2),
     B2 = (T2,E2), Code = (C1 ; C2).
compile_body((A->B),B2,Code) :- !,
     protected_compile_body(A,T2,C1), protected_compile_body(B,E2,C2),
     compile_body(B,UPE2,UPC2), % in case A simplifies to true
     B2 = (T2, pe:post_process(C1,PPC1),
           (PPC1==true -> UPE2, Code=UPC2 ; PPC1\==false, E2, Code = (PPC1 -> C2))).
compile_body(\+(A),B2,Code) :- !,
     protected_compile_body(A,B2,C1), Code = \+(C1).
compile_body(on_exception(E,A,B),S1,Code) :- !, 
     protected_compile_body(A,AS,AC), protected_compile_body(B,BS,BC),
     S1 = (on_exception(E,AS,AC=throw(E)),BS),
     Code = on_exception(E,AC,BC).
compile_body(time_out(A,TO,TRes),B2,Code) :- !,
     protected_compile_body(A,B2,C1), Code = time_out(C1,TO,TRes).
%compile_body(!,B2,Code) :- !, B2=true, Code=!. % dynamic cut
compile_body(!,B2,Code) :- !, B2=!, Code=true. % static cut
% to do: X^P, call_cleanup, try, call_residue_vars, iterators
compile_body(M:B1,B2,Code) :- !,B2=true, Code=M:B1.
compile_body(B1,B2,Code) :- prolog_load_context(module,M),
     % add module qualifier if missing as residual clauses reside in pe module
     B2=true, Code=M:B1.

% check if a piece of code could be decided to be true or fail at compile time
could_be_decided(X) :- var(X),!.
could_be_decided(true). could_be_decided(fail).
could_be_decided((A,_)) :- could_be_decided(A).
could_be_decided((_;_)).

:- use_module(library(terms),[term_variables/2]). 


% compile a body in such a way that we the specializer does not instantiate nor fails
protected_compile_body((A1,B1),(A2,B2),(C1,C2)) :- !,
   protected_compile_body(A1,A2,C1), protected_compile_body(B1,B2,C2).
protected_compile_body(call(A),B2,Code) :- !, protected_compile_body(A,B2,Code).
protected_compile_body(B1,Spec,Code) :- is_unfolded_somewhere(B1,MB1),!, 
    print(protected_unfold(MB1)),nl,
    Spec=partially_evaluate_fail_detect(MB1,Code). % memoize; return residual call as Code
    % TO DO: return fail if all clauses fail
protected_compile_body(B1,B2,Code) :- execute_predicate_in_load_context(B1,Condition),!,
    add_module_qualifier(B1,MB1),
    B2=((ground(B1),call(Condition))  /* we can execute and no bindings can be generated */
    % Sometimes it is also safe to do this even if B1 is not ground, namely if the variables are all free; e.g., the call  bmachine:b_machine_temp_predicate(Q)
        -> (call(MB1) -> Code=true ; Code=fail)
        ; (call(Condition), \+ pe:is_builtin(B1)) % it is an executable user-built-in
            ->  terms:term_variables(MB1,Vars), copy_term((MB1,Vars),(CMB1,CVars)),
                (call(CMB1)
                   ->  Code = (Vars=CVars) % we could seperated out into a conjunction of =
                   ; Code=fail)
        ; Code=MB1).
protected_compile_body((Tst->Thn;Els),B2,Code) :- !,
     protected_compile_body(Tst,Tst2,C0),
     protected_compile_body(Thn,T2,C1), protected_compile_body(Els,E2,C2),
     (\+ could_be_decided(C0)
       -> %nl,nl,print(undecidable(C0,Tst,Tst2)),nl,nl,
          B2 = (Tst2, T2,E2),   Code=(C0->C1;C2)  % always dynamic if
       ;  B2 = (Tst2, pe:post_process(C0,PPC0),
                     ( PPC0==fail  -> E2,Code=C2 % static if  
                      ; PPC0==true -> T2,Code=C1 % ditto 
                      ; T2,E2, Code=(PPC0->C1;C2)))  % dynamic if
      ).
protected_compile_body(if(Tst,Thn,Els),B2,Code) :- !,
     protected_compile_body(Tst,Tst2,C0),
     protected_compile_body(Thn,T2,C1), protected_compile_body(Els,E2,C2),
     (\+ could_be_decided(C0)
       -> B2 = (Tst2, T2,E2),   Code=if(C0,C1,C2)
       ;  B2 = (Tst2, pe:post_process(C0,PPC0),
                      (PPC0==fail   -> E2,Code=C2  /* static if */
                       ; PPC0==true -> T2,Code=C1  /* ditto */
                       ; T2,E2, Code=if(PPC0,C1,C2)))
     ).
protected_compile_body((A->B),B2,Code) :- !,
     protected_compile_body(A,T2,C1), protected_compile_body(B,E2,C2),
     B2 = (T2, pe:post_process(C1,PPC1),
           (PPC1==true -> E2, Code=C2
             ; PPC1==false -> Code=false
             ; E2, Code = (PPC1 -> C2))).
protected_compile_body(Body,Spec,Code) :- \+ can_instantiate_or_fail(Body),!,
   %print(is_safe(Body)),nl,
   compile_body(Body,Spec,Code).
protected_compile_body(Body,Spec,Code) :- is_builtin(Body),!,
   Spec =(builtin_tools:is_decidable_builtin(Body)
           -> (Body -> Code=true ; Code=fail)
           ; Code=Body).
protected_compile_body(Body,PSpec,PCode) :-
   compile_body(Body,BSpec,BCode),
   % better: create a memoisation point
   PSpec = (terms:term_variables(Body,VB),findall((VB:-BCode),BSpec,Clauses)),
   PCode = (member((VB:-C),Clauses),call(C)).



% check when compile_body could instantiate the goal or fail during specialization time
can_instantiate_or_fail(V) :- var(V),!,fail.
can_instantiate_or_fail(true) :- !,fail.
can_instantiate_or_fail(otherwise) :- !,fail.
can_instantiate_or_fail(B) :- is_builtin(B).
can_instantiate_or_fail(B) :- is_unfolded_somewhere(B,_).
can_instantiate_or_fail(B) :- execute_predicate_in_load_context(B,_).
% below cannot instantiante; or can they ??
%can_instantiate_or_fail((A;B)) :- can_instantiate_or_fail(A) ; can_instantiate_or_fail(B).
%can_instantiate_or_fail((A->B)) :- can_instantiate_or_fail(A) ; can_instantiate_or_fail(B).
%can_instantiate_or_fail(\+(_)).


% remove useless true literals; remove everything after a fail
post_process(X,Y) :- var(X),!,Y=X.
post_process((A,B),Res) :- !,post_process(A,PA),
   (PA==true -> post_process(B,Res)
    ; PA==fail -> Res = fail
    ; post_process(B,PB),
       (PB==true -> Res=PA
        ; (PA==!, PB=(!,_)) -> Res = PB % remove two cuts that follow each other
        ; Res = (PA,PB))
   ).
post_process((A;B),Res) :- !,post_process(A,PA),
   (PA==fail -> post_process(B,Res)
    ; post_process(B,PB),
       (PB==fail -> Res=PA ; Res = (PA;PB))
   ).
post_process((A->B),Res) :- !,post_process(A,PA),
   (PA==fail -> Res=fail
    ; PA==true -> post_process(B,Res)
    ; post_process(B,PB),
      Res = (PA->PB)
   ).
post_process(M:W,Res) :- nonvar(W),W=when(A,B),!, Res=M:R,post_process(when(A,B),R). % also do this for other constructs ??
post_process(when(A,B),Res) :- !,post_process(A,PA), post_process(B,PB),
   (PA==true -> Res=PB
    ; PA==fail -> print('*** unsatisfiable when: '), print(when(A,B)),nl,
      Res = when(A,PB)
    ; Res = when(PA,PB)
   ).
post_process(member(_,L),Res) :- L==[],!, Res = fail. % appears in protected_compile
post_process(BI,S) :- post_process_builtin(BI,SBI),!, S=SBI.
post_process(call(X),S) :- nonvar(X),!, post_process(X,S).
post_process(time_out(X,TO,TRes),Res) :- !,  % to do: check if library(timeout) loaded
    post_process(X,PX),
    Res = timeout:time_out(PX,TO,TRes).
post_process(X,X).


:- meta_predicate pe(0,*).
unfold_call(Call,PPCode) :- %print(unfolding(Call)),nl,
   generate_unfolder_call(Call,Code,NC),!,
   call(NC),
   post_process(Code,PPCode).
unfold_call(Call,Code) :- 
    print('*** Trying to unfold unregistered call : '), print(Call),nl,
    Code=true, call(Call).

pe(C,Clauses) :- 
   findall((C :- Code), unfold_call(C,Code), Clauses).

:- dynamic specialized_predicate/2.
:- meta_predicate partially_evaluate(0,*).
:- meta_predicate partially_evaluate(0,*,*).
:- meta_predicate partially_evaluate_and_load(0,*).
% partially_evaluate(Call, ResidualCallToCallSpecializedCode)
partially_evaluate_and_load(Call,Res) :- partially_evaluate(Call,Res,Reused),
   (Reused=true -> true ; load_partially_evaluated_code).

partially_evaluate(Call,Res) :- partially_evaluate(Call,Res,_Reused).
partially_evaluate(Call,Res,Reused) :- 
   %copy_term(Call,Copy),numbervars(Copy,0,_), % no longer used; expensive for large terms
   term_variables(Call,Vars1),
   specialized_predicate(Call,Entry),
   % Check that call has not been instantiated; i.e., is an instance of call already specialized
   term_variables(Call,Vars2), Vars1==Vars2,
   %\+ \+ (Call = Copy), 
   !,
   Res=pe:Entry, Reused=true.
partially_evaluate(Call,pe:Entry,false) :- generate_entry(Call,Entry),
   assert(specialized_predicate(Call,Entry)),
   %print(partially_evaluating(Call,Entry)),nl,
   start_specialization(Entry),
   % TO DO: generate block declarations !
   ((get_block_declaration(Call,BlockList,BlockDecl),
     call_is_blocked(Call,BlockList))
     -> instantiated_block_pattern(Call,OrigCallPat),
        %print(generating_block_declaration(Call,OrigCallPat,Entry,BlockDecl)),nl,
        adapt_block_declaration(BlockDecl,OrigCallPat,Entry,NewBlockDecl),
        portray_clause(':-'(block(NewBlockDecl))),
        (NewBlockDecl=fail
           -> % block optimized away; should not happen if call_is_blocked succeeds ??
              print('### call can never block'),nl
           ;  prolog:'$block'(pe:NewBlockDecl,compiled)  % SICSTUS hook to add block declarations at runtime; see SPRM 12990
        )
     ;  true),
   pe_aux(Call,Entry).
pe_aux(Call,Entry) :-
   unfold_call(Call,Code),
   assert_specialized_clause(Entry,Code),
   fail.
pe_aux(_Call,_Entry) :- true. 
  %print(finished_partially_evaluating(_Call,_Entry)),nl.

% generalize a call according to a binding-type description
generalize_call(dynamic,_,_) :- !.
generalize_call(ground_dynamic,X,_V) :-
  ground(X),!. % TO DO: add info to _V that ground
generalize_call(static,X,GX) :- !, GX=X.
generalize_call(atomic,X,GX) :- atomic(X),!, GX=X.
generalize_call(ground,X,GX) :- ground(X),!, GX=X.
generalize_call(list(BT),L,GL) :- !,list_gen_call(L,BT,GL).
generalize_call(nonvar,X,GX) :- !, functor(X,F,N), functor(GX,F,N).
generalize_call(term(F,BT),X,GX) :-  X =.. [F|XArgs], !,
   l_generalize_call(BT,XArgs,GXArgs), 
   GX =.. [F|GXArgs].
generalize_call(BT,X,GX) :- 
    print('*** Nonmatching binding type: '), print(BT),nl,
    print('*** Value: '), print(X),nl,
    GX=X.

l_generalize_call([],[],GX) :- !, GX=[].
l_generalize_call([BType|BT],[X|TX],[GX|TGX]) :-
     generalize_call(BType,X,GX),
     l_generalize_call(BT,TX,TGX).

list_gen_call([],_,GL) :- !, GL=[].
list_gen_call([H|T],BType,[GH|GT]) :- !, generalize_call(BType,H,GH), list_gen_call(T,BType,GT).
list_gen_call(L,T,GL) :- print('*** Nonmatching list binding-type: '), print(list_gen_call(L,T,GL)),nl,
    GL=L.

:- use_module(library(terms),[free_of_var/2]).
% first version of a specific optimizer to remove redundant calls;
% TO DO : generalize based on assertions :- redundant(p(X), free_var(X)).
post_process_specialized_clause((A,B),Before,After,Res) :- !,
    post_process_specialized_clause(A,Before,(B,After),PA),
    post_process_specialized_clause(B,(PA,Before),After,PB),
    (PA==true -> Res = PB
     ; PB==true -> Res = PA
     ; Res = (PA,PB)).
post_process_specialized_clause(b_interpreter:ground_wait_flags(W),Before,After,true) :-
    nonvar(W), W=wfx(A,B,C),
    % suboptimal implementation; remove ground_wait_flags if arg fresh variable
    var(A),var(B),var(C),
    free_of_var(A,Before), free_of_var(B,Before), free_of_var(C,Before),
    free_of_var(A,After), free_of_var(B,After), free_of_var(C,After),
    !.
%post_process_specialized_clause(kernel_objects:less_than(int(X),int(Y)),_,_,Code) :- !, Code = '<'(X,Y).
%post_process_specialized_clause(kernel_objects:less_than_equal(int(X),int(Y)),_,_,Code) :- !, Code = '=<'(X,Y).
%post_process_specialized_clause(kernel_mappings:not_eq_wf(int(A), int(C), _, _),_,_,Code) :- !, Code = dif(A,C).
post_process_specialized_clause(A,_,_,A). % :- print(pp(A)),nl.

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(\+ environ(partially_evaluate_compile, true)).
start_specialization(Entry) :-
    assert( ( Entry :- fail ) ), % make an entry for the predicate; in case no clauses generated by unfolding
    retract( ( Entry :- fail) ).
assert_specialized_clause(Head,Body) :-
   post_process_specialized_clause(Body,Head,true,NewBody),
   assert( (Head :- NewBody ) ), 
   portray_clause( (Head :- NewBody ) ).
clear_specialized_clauses :- retract(specialized_predicate(Call,Entry)),
   print(clearing(Call,Entry)),nl,
   retract((Entry :- _Body)),fail.
clear_specialized_clauses.
load_partially_evaluated_code. % already asserted
:- else.
start_specialization(Entry) :- print(start_specialization(Entry)),nl,
   % TO DO: only do this if necessary in load_partially_evaluated_code
   assert_specialized_clause(Entry,fail).
assert_specialized_clause(Head,Body) :-
   open('pe_cache.pl',write,S),
   portray_clause(S, (Head :- Body ) ),
   close(S).
clear_specialized_clauses :- print('% Clearing specialized code'),nl,
   open('pe_cache.pl',append,S),
   format(S,'/* PE CACHE FILE */~n',[]), close(S).
load_partially_evaluated_code :- print('% LOADING : pe_cache.pl'),nl,
   %compile('pe_cache.pl').
   ensure_loaded('pe_cache.pl').
:- endif.

:- meta_predicate partially_evaluate_fail_detect(0,*).
partially_evaluate_fail_detect(Call,Result) :-
   partially_evaluate(Call,ResCall),
   (\+ (clause(ResCall,B), B\=fail) -> Result=fail
     % TO DO: check if only a single clause exists with no instantiation in head  and no block-> replace
     ; Result = ResCall).
clear_specialized_code :- clear_specialized_clauses,
   reset_pe_pred_count.

:- dynamic pe_pred_count/1.
% count number of specialized predicates generated
pe_pred_count(0).
reset_pe_pred_count :- retractall(pe_pred_count(_)), assert(pe_pred_count(0)).
get_new_pred(P,NP) :- retract(pe_pred_count(C)),
   C1 is C+1, assert(pe_pred_count(C1)),
   number_codes(C,NC), atom_codes(A,NC),
   atom_concat(P,A,NP).

% generate an unfolding call
generate_entry(_Module:Call,NewSpecializedCall) :- !,
    generate_entry(Call,NewSpecializedCall).
generate_entry(Call,NewSpecializedCall) :-
    Call =.. [F|Args],
    term_variables(Args,Vars), % TO DO: generate number 
    atom_concat(F,'$spec_',FS),
    get_new_pred(FS,NewF),
    NewSpecializedCall =.. [NewF|Vars].

:- use_module(library(plunit)).
:- begin_tests(block_pe).
test(pat1) :- instantiated_block_pattern(p(f(X,Y),a,Z),R), R==p('?','?',Z).
test(pat2) :- instantiated_block_pattern(p(V,f(X,X),a,V),R), R==p(V,'?','?',V).
:- end_tests(block_pe).

% replace nonvar args by '?'
instantiated_block_pattern(_:Call,Pattern) :- !, instantiated_block_pattern(Call,Pattern).
instantiated_block_pattern(Call,Pattern) :-
    Call =.. [Pred|Args],
    maplist(inst_block_pat_aux,Args,PatArgs),
    Pattern =.. [Pred|PatArgs].
inst_block_pat_aux(Arg,Pat) :- (nonvar(Arg) -> Pat = '?' ; Pat=Arg).

adapt_block_declaration((B1,B2),OrigPat,ResidualCall,Result) :- !,
    adapt_block_declaration(B1,OrigPat,ResidualCall,RB1),
    adapt_block_declaration(B2,OrigPat,ResidualCall,RB2),
    (RB1 = false -> Result = RB2 ; RB2 = false -> Result = RB1
      ; Result = (RB1,RB2)).
adapt_block_declaration(Block,OrigPat,ResidualCall,Result) :-
    copy_term((OrigPat,ResidualCall),(O,R)),
    %print(check(Block,O,R)),nl,
    (Block=O % will instantiate some variables of R by '-'
       -> %print(match(O,R)),nl,
          R =.. [Pred|RArgs],
          maplist(inst_var('?'),RArgs,BRArgs), % replace remaining vars by '?'
          Result =.. [Pred|BRArgs]
        ; Result = false
     ).
inst_var(Val,H,Res) :- (var(H) -> Res=Val ; Res=H).
