% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Pragma handling

:- module(pragmas, [ store_pragmas/1
                   , apply_pragmas/5
                   , unhandled_pragma/4
                   , global_pragma/2
		   , apply_pragmas_to_event_b_machine/3
		   , extract_pragmas_from_event_b_extensions/2
                   ]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module handles pragmas loaded with a b machine').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).

:- use_module(debug).
:- use_module(error_manager).
:- use_module(bmachine_structure).

:- dynamic present_pragma/4, unhandled_pragma/4, global_pragma/2.

store_pragmas(Pragmas) :-
    retractall(present_pragma(_,_,_,_)),
    retractall(unhandled_pragma(_,_,_,_)),
    retractall(global_pragma(_,_)),
    maplist(store_pragma2,Pragmas).

store_pragma2(pragma(NodeID,PragmaName,Arguments,Warnings,FileNo,SLine,SCol,ELine,ECol)) :- !,
    store_pragma3(NodeID,PragmaName,Arguments,pos(NodeID,FileNo,SLine,SCol,ELine,ECol)),
    (Warnings = [] -> true
     ; add_warning(pos(FileNo,SLine,SCol,ELine,ECol),'Pragma with Warnings: ',[PragmaName,Arguments,Warnings])).

store_pragma2(global_pragma(PragmaName,Arguments,Warnings,FileNo,SLine,SCol,ELine,ECol,_ASTNodes)) :- !,
    assert(global_pragma(PragmaName,Arguments)),
    (Warnings = []
     -> (global_pragma(PragmaName) -> true
         ; add_warning(pragma,'Unknown global pragma: ',[PragmaName,Arguments],pos(FileNo,SLine,SCol,ELine,ECol)))
     ; add_warning(pragma,'Global pragma with Warnings: ',[PragmaName,Arguments,Warnings]),pos(FileNo,SLine,SCol,ELine,ECol)).
store_pragma2(P) :- add_internal_error('Unknown pragma: ',store_pragma2(P)).

store_pragma3(Id,Functor,Args,Pos) :-
    debug_println(9,pragma(Id,Functor,Args,Pos)),
    assert(present_pragma(Id,Functor,Args,Pos)),
    assert(unhandled_pragma(Id,Functor,Args,Pos)).


% Lookup all pragmas associated to a position (and remove them from the list)
lookup_pragmas(pos(Id,_FileNo,_,_,_,_),Pragmas) :-
    present_pragma(Id,_,_,_),!, % only call findall if we really have at least one pragma for this position; speeds up eg. flat200-90.mch
    findall(pragma(P,Args,Pos), present_pragma(Id,P,Args,Pos), Pragmas),
    %print(pragmas(Id,Pragmas)),nl,
    retractall(unhandled_pragma(Id,_,_,_)).
lookup_pragmas(_,[]).

apply_pragmas(Pos,TExpr,Type,Infos1,Infos) :-
    % find all pragmas that are associated with this position
    lookup_pragmas(Pos,Pragmas),
    % and check each of it (Infos and Infos1 are like
    apply_pragmas2_l(Pragmas,TExpr,Type,Pos,Infos,Infos1).
apply_pragmas2_l([],_TExpr,_Type,_Pos,Infos,Infos).
apply_pragmas2_l([Pragma|Prest],TExpr,Type,Pos,Infos,Infos1) :-
    apply_pragmas2(Pragma,TExpr,Type,Pos,Infos,Infos2),
    apply_pragmas2_l(Prest,TExpr,Type,Pos,Infos2,Infos1).
apply_pragmas2(pragma(Name,Arguments,_),TExpr,Type,_Pos,ResInfos,Infos) :-
    % ok, a pragma rule is defined
    pragma_rule(Name,Arguments,TExpr,Type,Info),!, ResInfos=[Info|Infos].
apply_pragmas2(pragma(Name,Arguments,Pos),_TExpr,_Type,_Pos,Infos,Infos) :-
    % encountered a pragma that is not even recognised by the parser
    add_warning(pragma,'Ignored pragma: ',[Name|Arguments],Pos).


% pragma rules: pragma(Pragma, Expr, Type, Info)
%pragma_rule(X,E,T,I) :- print(pragma_rule(X,E,T,I)),nl,fail.
pragma_rule(label,IdList, _, _, label(IdList)) :- IdList = [_|_].
pragma_rule(unit,Unit, _, _, unit(Unit)).
pragma_rule(inferred_unit,Unit, _, _, inferred_unit(Unit)).
pragma_rule(conversion,[], _, _, conversion).
pragma_rule(symbolic,[],_,_,prob_annotation('SYMBOLIC')).

% pragma which apply to the entire file; they do not need a location
global_pragma(generated).
global_pragma(unit_alias).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Event-B
extract_pragmas_from_event_b_extensions(Ext,Pragmas) :-
    findall(P,
	    (member(P,Ext), P=pragma(_T,_S,_A,_C)),
	    Pragmas).

% currently, there are only 2 possible pragmas:
% - attach a unit to a variable
% - attach a unit to a constant
apply_pragmas_to_event_b_machine(Pragmas,Machine,ResultingMachine) :-
    get_section(abstract_constants,Machine,AbsConstants),
    maplist(apply_event_b_pragmas(Pragmas),AbsConstants,NewAbsConstants),
    get_section(concrete_constants,Machine,ConcConstants),
    maplist(apply_event_b_pragmas(Pragmas),ConcConstants,NewConcConstants),
    get_section(abstract_variables,Machine,AbsVars),
    maplist(apply_event_b_pragmas(Pragmas),AbsVars,NewAbsVars),
    get_section(concrete_variables,Machine,ConcVars),
    maplist(apply_event_b_pragmas(Pragmas),ConcVars,NewConcVars),
    write_section(abstract_constants,NewAbsConstants,Machine,TmpMachine1),
    write_section(concrete_constants,NewConcConstants,TmpMachine1,TmpMachine2),
    write_section(abstract_variables,NewAbsVars,TmpMachine2,TmpMachine3),
    write_section(concrete_variables,NewConcVars,TmpMachine3,ResultingMachine).


apply_event_b_pragmas(Pragmas,b(identifier(Name),Type,Infos),b(identifier(Name),Type,Infos2)) :-
    member(pragma(PragmaType,_Machine,Name,Content),Pragmas)
    -> pragma_rule(PragmaType,Content,b(identifier(Name),Type,Infos),Type,AdditionalArgument),
       Infos2 = [AdditionalArgument|Infos]
    ;  Infos2 = Infos.

    