% (c) 2009-2011 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(pref_definitions,[b_get_preferences_from_machine/0,
         b_get_important_preferences_from_raw_machine/1,
         b_get_definition_string_list_from_machine/2]).

:- use_module(module_information).
:- module_info(group,tools).
:- module_info(description,'Tools for extracting information from DEFINITIONS.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(tools,[ajoin/2,print_message/1]).
:- use_module(bmachine,[b_definition_prefixed/4,b_get_typed_definition/3]).
:- use_module(preferences,[temporary_set_preference/2,eclipse_preference/2,preference_val_type/2]).
:- use_module(btypechecker,[unify_types_werrors/5]).
:- use_module(bsyntaxtree).
:- use_module(b_interpreter,[b_compute_expression/4]).
:- use_module(debug,[debug_println/2]).
% get and set important preferences from a raw, untyped machine (as they
% may influence type checking,...); 
% example: IGNORE_PRJ_TYPES, WARN_IF_DEFINITION_HIDES_VARIABLE
b_get_important_preferences_from_raw_machine(RawMachine) :-
    member(definitions(_Pos,Defs),RawMachine),
    member(expression_definition(_DPOS,DefName,[],RawVALUE),Defs),
    atom_concat('SET_PREF_',PREFString,DefName),
    eclipse_preference(PREFString,Pref),
    preference_val_type(Pref,PType),
    PType=bool, % at the moment only boolean preferences supported
    convert_raw_pref_val(PType,RawVALUE,PVal),
    %print(pref(Pref,PType,PVal)),nl,
   (temporary_set_preference(Pref,PVal)
         -> debug_println(9,set_important_pref(PREFString,Pref,PVal))
         ;  add_error(pref_definitions,'Could not set DEFINITION preference to specified value: ',(PREFString:PVal))
   ),
    fail.
b_get_important_preferences_from_raw_machine(_).  

convert_raw_pref_val(bool,boolean_true(_),true).
convert_raw_pref_val(bool,boolean_false(_),false).
%convert_raw_pref_val(int,integer(_,V),V).
%convert_raw_pref_val(nat,integer(_,V),V). % range checking done in temporary_set_preference
%convert_raw_pref_val(nat1,integer(_,V),V).
%convert_raw_pref_val(neg,integer(_,V),V).
%convert_raw_pref_val(range(_,_),integer(_,V),V).
%convert_raw_pref_val(string(_),string(_,S),S).

b_get_preferences_from_machine :-
   b_definition_prefixed(_,'SET_PREF_',String,DefName),
   eclipse_preference(String,Pref),
   b_get_typed_definition(DefName,[],TExpr),
   preference_val_type(Pref,PType),
   convert_pref_val(PType,TExpr,PVal),
   print_message(set_pref(String,Pref,PVal)),
   (temporary_set_preference(Pref,PVal)
     -> true  /* TO DO: only set the Preference temporarily ! */
     ;  add_error(pref_definitions,'Could not set DEFINITION preference: ',(String:PVal)) 
   ),
   fail.
b_get_preferences_from_machine.  

convert_pref_val(int,TExpr,Val) :- !,
	check_type(integer,TExpr),
	~~mnf( b_interpreter:b_compute_expression(TExpr,[],[],int(Val)) ).
convert_pref_val(nat,TExpr,Val) :- !,
	convert_pref_val(int,TExpr,Val),
	( (ground(Val),Val >= 0) -> true
	; otherwise -> error(TExpr,'Value must be a natural (>= 0)')).
convert_pref_val(neg,TExpr,Val) :- !,
	convert_pref_val(int,TExpr,Val),
	( (ground(Val),Val =< 0) -> true
	; otherwise -> error(TExpr,'Value must be non-positive (<= 0)')).
convert_pref_val(nat1,TExpr,Val) :- !,
	convert_pref_val(int,TExpr,Val),
	( (ground(Val),Val >= 1) -> true
	; otherwise -> error(TExpr,'Value must be a natural >= 1')).
convert_pref_val(range(Low,Up),TExpr,Val) :-!,
	convert_pref_val(int,TExpr,Val),
	( ground(Val),Val >= Low, Val =< Up -> true
	; otherwise -> ajoin(['Value must be in the interval [',Low,',',Up,']'],Msg),
	    error(TExpr,Msg)).
convert_pref_val(bool,TExpr,Val) :- !,
	check_type(boolean, TExpr),
	~~mnf( b_compute_expression(TExpr,[],[],BVal) ),
	( BVal = pred_true /* bool_true */ -> Val = true
	; BVal = pred_false /* bool_false */ -> Val = false
	; otherwise -> add_error(pref_definitions,'Unexpected result (neither true nor false)',BVal)).
convert_pref_val(string,TExpr,Val) :- !,
	check_type(string, TExpr),
	~~mnf( get_texpr_expr(TExpr,string(Val)) ).
convert_pref_val(PrefType,_,_) :-
	add_error(pref_definitions,'Unsupported preference type in definitions',PrefType),fail.

error(TExpr,Msg) :-
	get_texpr_pos(TExpr,Pos),
	add_all_perrors([error(Msg,Pos)]),
	fail.

check_type(Type,TExpr) :-
	get_texpr_type(TExpr,EType),
	get_texpr_pos(TExpr,Pos),
	unify_types_werrors(Type,EType,Pos,Errors,[]),
	( Errors = [] -> true
	; otherwise   -> add_all_perrors(Errors),fail).

% -------------------


:- use_module(tools,[safe_atom_codes/2, string_concatenate/3]).
:- use_module(bmachine,[b_definition_prefixed/4,b_get_definition_string_from_machine/2]).
	
   /* for tcltk: */
 b_get_definition_string_list_from_machine(Def_Name_Prefix,Result) :-
    atom_codes(Def_Name_Prefix,Codes), append(Codes,_,CodesPrefix),
    findall(S, (b_get_definition_string_from_machine(Def_Name,DS),
                safe_atom_codes(Def_Name,CodesPrefix), /* check that name begins with the same characters as Def_Name_Prefix */
                string_concatenate(DS,';',S)), Result).
                