% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(preferences,
    [init_preferences/0,
     tcltk_get_preference/2, get_computed_preference/2,
     get_preference/2, get_preferences_list/1, get_preferences_list/2,
     set_preference/2,
     temporary_set_preference/2, reset_temporary_preference/1, reset_all_temporary_preferences/0,
     preference_category/2, preference_description/2, preference_val_type/2, preference_default_value/2,
     preference/2,
     get_ith_preference_type/2, get_ith_preference_value/2, set_ith_preference_value/2,
     get_ith_preference_type/3, get_ith_preference_value/3, set_ith_preference_value/3,
     load_preferences/1,save_preferences/1,
     backup_preferences/0, revert_preferences/0, reset_to_defaults/1, reset_for_benchmarking/0,
     set_preference_group/2,
     activate_recent_documents/0,deactivate_recent_documents/0,
     
     add_recent_document/1,  /* use to add a file to recent documents list */
     get_recent_documents/1, clear_recent_documents/0,
     add_element_to_history/3, get_history_elements_list/3,

     eclipse_preference/2, print_eclipse_prefs/0,
     advanced_eclipse_preference/2, obsolete_eclipse_preference/1,
     print_non_default_preferences/0
    ]).
    
:- use_module(tools).
:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module manages the various preferences of ProB. Preferences can be grouped, saved and restored and validated from file.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(debug).
:- use_module(self_check).
:- use_module(error_manager).   
:- use_module(library(lists)).

:- dynamic change_flag/0.
assert_change_flag :- assert(change_flag).

init_preferences :- retractall(change_flag),reset_to_defaults, check_preferences.

check_preferences :- 
   %print('% '), preference_description(P,_), print(P), print(' '),
   preferences:preference_default_value(P,_),
   ((preferences:preference(P,Val),
     preferences:preference_val_type(P,T),
     preferences:preference_category(P,_))
     -> preferences:is_of_type(Val,T)
     ;  add_error(check_preferences,'Internal error for preference: ',P)),
   fail.
check_preferences. % :- print_message('Preferences Checked').   


load_preferences(File) :- print_message(load_preferences(File)),
   safe_on_exception(_Exc,see(File),
       (print_message('### cannot open preference file'),print_message(File),fail)),
   read_preferences, seen.

read_preferences :-  clear_dynamic_preference_predicates,
   read_preferences2.
read_preferences2 :-
   read(Term), Term \= end_of_file, !, %print(pref(Term)),nl,
   ((Term = preference(Pref,NVal))
      -> (volatile_preference(Pref) -> true
          ; retract(preference(Pref,_)) -> assert(preference(Pref,NVal))
          ; obsolete_preference(Pref) -> print_message(obsolete(Pref))
          ; add_message(load_preferences,'Unknown preference in preference file: ',Pref)
         )
      ;  Term = ':-'(dynamic(_)) -> true 
      ;  dynamic_preference_predicate(Term) -> assert(Term)
      ;  (Term = (Head:-fail),dynamic_preference_predicate(Head)) -> true
      ;  add_error(load_preferences,'Unrecognised:',Term) 
    ),
    read_preferences2.

read_preferences2 :- retractall(change_flag), recompute_nr_of_recent_documents,
                     recompute_nr_of_searched_and_replaced_patterns,recompute_nr_of_eval_history_elements.

clear_dynamic_preference_predicates :- dynamic_preference_predicate(X), retractall(X),fail.
clear_dynamic_preference_predicates.

% declare which dynamic predicates are stored in Preference file
dynamic_preference_predicate(recent_documents(_)).
dynamic_preference_predicate(searched_patterns(_)).
dynamic_preference_predicate(replaced_patterns(_)).
dynamic_preference_predicate(eval_history_elements(_)).
dynamic_preference_predicate(eval_csp_history_elements(_)).
dynamic_preference_predicate(checked_ltl_formulas(_)).

save_preferences(File) :- change_flag,!,
   print_message(save_preferences(File)),
   tell(File), 
   (print_preferences -> true ; true),
   told,
   retractall(change_flag).
save_preferences(File) :- print_message(preferences_unchanged(File)).

/* 
print_preferences_for_saving :- 
   preference_default_value(Pref,_),
   preference(Pref,CurVal),
   (temp_pref(Pref,OldVal,CurVal) -> V=OldVal ; V=CurVal),
   write_term(preference(V),[quoted(true)]),print('.'),nl,
   fail.
print_preferences_for_saving :- print_dynamic_preferences. */

:- assert_must_succeed(print_preferences).
print_preferences :- print_dynamic_pred(preferences,preference,2),
   print_dynamic_preferences.

print_dynamic_preferences :-
   dynamic_preference_predicate(P), functor(P,F,N),
   print_dynamic_pred(preferences,F,N),fail.
print_dynamic_preferences.  

:- assert_must_succeed(print_non_default_preferences).
print_non_default_preferences :- 
  print('======================='),nl,
  print('Non-Default Preferences:'),nl,
  preference(Pref,Val),
  (preference_default_value(Pref,Val) -> true
    ; print(Pref), print(' == '), print(Val),nl
  ),fail.
print_non_default_preferences :-
  print('======================='),nl.

:- dynamic change_flag_backup/0.
:- dynamic pref_backup/2.
  
backup_preferences :- retractall(pref_backup(_,_)),retractall(change_flag_backup),
                     preference(P,Val), assert(pref_backup(P,Val)),fail.
backup_preferences :- (change_flag -> assert(change_flag_backup) ; true).

revert_preferences :- \+(pref_backup(_,_)),!, print_error('No previously backed-up preferences in revert_preferences!').
revert_preferences :- retractall(preference(_,_)),pref_backup(P,Val), assert(preference(P,Val)),fail.
revert_preferences :- retractall(change_flag), (change_flag_backup -> assert_change_flag ; true).

reset_to_defaults :- reset_to_defaults(_).

reset_to_defaults(Category) :- preference_category(Pref,Category),
   preference_default_value(Pref,Val),
   retractall(preference(Pref,_)),
   assert(preference(Pref,Val)),
   fail.
reset_to_defaults(_) :- assert_change_flag.


reset_for_benchmarking :-
   set_preference(maxNrOfInitialisations,3),
   set_preference(maxNrOfEnablingsPerOperation,10),
   set_preference(maxint,4),
   set_preference(minint,-1),
   set_preference(time_out,3500),
   set_preference(globalsets_fdrange,3),
   set_preference(show_initialisation_arguments,true),
   set_preference(show_eventb_any_arguments,false),
   %set_preference(animate_skip_operations,true),
   set_preference(cspm_animate_all_processes_without_arguments,false),
   set_preference(cspm_animate_all_processes,false).
  % set_preference(convert_closures_into_explicit_form_for_store,true).
   
:- assert_must_succeed( \+((eclipse_preference(_EP,P), \+preference_default_value(P,_Val)))).
eclipse_preference('MAXINT',maxint).
eclipse_preference('MININT',minint).
eclipse_preference('DEFAULT_SETSIZE',globalsets_fdrange).
eclipse_preference('MAX_INITIALISATIONS',maxNrOfInitialisations).
eclipse_preference('MAX_OPERATIONS',maxNrOfEnablingsPerOperation).
%eclipse_preference('ANIMATE_SKIP_OPERATIONS',animate_skip_operations).
%eclipse_preference('EXPAND_CLOSURES_FOR_STATE',convert_closures_into_explicit_form_for_store).
eclipse_preference('SYMBOLIC',convert_comprehension_sets_into_closures).
eclipse_preference('CLPFD',use_clpfd_solver).
eclipse_preference('SMT',use_smt_mode).
eclipse_preference('STATIC_ORDERING',use_static_ordering).
eclipse_preference('COMPRESSION',use_state_packing).
%eclipse_preference('SYMMETRY_MARKERS',use_symmetry_marker_hash).
%eclipse_preference('PERMUTATION_FLOODING',enable_permutation_reducation).
eclipse_preference('SYMMETRY_MODE',symmetry_mode).
eclipse_preference('TIME_OUT',time_out).
eclipse_preference('PROOF_INFO',use_po).
eclipse_preference('TRY_FIND_ABORT',find_abort_values). % maybe rename into TRY_FIND_WD_ERROR ?
eclipse_preference('NUMBER_OF_ANIMATED_ABSTRACTIONS',number_animated_abstractions).
eclipse_preference('ALLOW_INCOMPLETE_SETUP_CONSTANTS',allow_incomplete_partial_setup_constants).
eclipse_preference('PARTITION_PROPERTIES',partition_predicates).
eclipse_preference('USE_RECORD_CONSTRUCTION',use_record_construction).
eclipse_preference('KODKOD',try_kodkod_on_load).
eclipse_preference('KODKOD_ONLY_FULL',kodkod_for_components).
eclipse_preference('MEMO',use_closure_expansion_memoization).
eclipse_preference('STRICT_RAISE_WARNINGS',strict_raise_warnings).

eclipse_preference(A,B) :- advanced_eclipse_preference(A,B).

advanced_eclipse_preference('OPERATION_REUSE',try_operation_reuse).
advanced_eclipse_preference('SHOW_EVENTB_ANY_VALUES',show_eventb_any_arguments).
advanced_eclipse_preference('RANDOMISE_OPERATION_ORDER',randomise_operation_order).
advanced_eclipse_preference('EXPAND_FORALL_UPTO',expand_forall_upto).
advanced_eclipse_preference('MAX_DISPLAY_SET',expand_avl_upto).
advanced_eclipse_preference('WARN_WHEN_EXPANDING_INFINITE_CLOSURES',warn_when_expanding_infinite_closures). 
advanced_eclipse_preference('WARN_IF_DEFINITION_HIDES_VARIABLE',warn_if_definition_hides_variable).  
advanced_eclipse_preference('TRACE_INFO',provide_trace_information).
advanced_eclipse_preference('DOUBLE_EVALUATION',double_evaluation_when_analysing).
%advanced_eclipse_preference('RECURSIVE',allow_recursive_closures).
advanced_eclipse_preference('IGNORE_HASH_COLLISIONS',ignore_hash_collisions).
advanced_eclipse_preference('FORGET_STATE_SPACE',forget_state_space).
advanced_eclipse_preference('LARGE_PARSER_JVM',use_large_jvm_for_parser).
advanced_eclipse_preference('CSP_STRIP_SOURCE_LOC',cspm_strip_source_location).
advanced_eclipse_preference('NEGATED_INVARIANT_CHECKING',do_neginvariant_checking).


advanced_eclipse_preference('DOT_PROP',dot_print_node_properties).
advanced_eclipse_preference('DOT_IDS',dot_print_node_ids).
advanced_eclipse_preference('DOT_INFO',dot_print_node_info).
advanced_eclipse_preference('DOT_LOOPS',dot_print_self_loops).
advanced_eclipse_preference('DOT_FUNCTIONS',dot_print_functions).
advanced_eclipse_preference('DOT_ROOT',dot_print_root).
advanced_eclipse_preference('DOT_LEAVES',dot_print_leaves).
advanced_eclipse_preference('DOT_ARGUMENTS',dot_print_action_arguments).
advanced_eclipse_preference('CHECK_AST',check_prolog_b_ast).
advanced_eclipse_preference('DOT_COLOR_ARC',dot_normal_arc_colour).
advanced_eclipse_preference('DOT_COLOR_NODE',dot_normal_node_colour).
advanced_eclipse_preference('DOT_COLOR_NODE_ERROR',dot_invariant_violated_node_colour).
advanced_eclipse_preference('DOT_COLOR_NODE_OPEN',dot_open_node_colour).
advanced_eclipse_preference('DOT_COLOR_NODE_GOAL',dot_goal_node_colour).
advanced_eclipse_preference('DOT_PEN_WIDTH',dot_node_penwidth).
advanced_eclipse_preference('BPARSER',selected_bparser).
advanced_eclipse_preference('BOOL_AS_PREDICATE',bool_expression_as_predicate).
advanced_eclipse_preference('IGNORE_PRJ_TYPES',ignore_prj_types).

obsolete_eclipse_preference('EXPAND_CLOSURES_FOR_STATE').

print_eclipse_prefs :- eclipse_preference(X,P),
   print('     '),print(X), print(' : '),
   preference_val_type(P,VT),
   print(VT), print(' ==> '),
   preference_description(P,Desc),
   print(Desc),nl,
   fail.
print_eclipse_prefs.

:- dynamic preference/2.

preference_default_value(maxint,3). % 2147483647
preference_default_value(minint,-1). %-2147483648
preference_default_value(number_animated_abstractions, 20).
preference_default_value(globalsets_fdrange,2).
%preference_default_value(do_invariant_checking,true).
preference_default_value(do_neginvariant_checking,false).
%preference_default_value(do_full_invariant_checking,true).
%preference_default_value(use_new_kernel,false).
preference_default_value(find_abort_values,false).
preference_default_value(treat_outermost_pre_as_select,true).
preference_default_value(require_operations_to_assign_all_outputs,false).
preference_default_value(cspm_animate_all_processes_without_arguments,false).
preference_default_value(cspm_animate_all_processes,false).
%preference_default_value(cspm_display_inout,false).
preference_default_value(cspm_detailed_animation,false).
preference_default_value(cspm_strip_source_location,false).
preference_default_value(csp_event_visible_to_user,false).
%preference_default_value(cspm_allow_incomplete_records,false).
%preference_default_value(findOnlyOneSol,false).
preference_default_value(maxNrOfInitialisations,4).
preference_default_value(maxNrOfEnablingsPerOperation,10).
preference_default_value(time_out,2500).
preference_default_value(debug_time_out_factor,5).
preference_default_value(use_large_jvm_for_parser,false).
preference_default_value(show_initialisation_arguments,true).
%preference_default_value(animate_skip_operations,true).
preference_default_value(show_eventb_any_arguments,false).
preference_default_value(convert_comprehension_sets_into_closures,false).  /* <------------  was true */
preference_default_value(use_clpfd_solver,false).
preference_default_value(use_smt_mode,false).
preference_default_value(fail_if_clpfd_timeout,false).
%preference_default_value(allow_recursive_closures,false).
preference_default_value(use_static_ordering,false).
preference_default_value(use_state_packing,false).
preference_default_value(provide_trace_information,false).
preference_default_value(partition_predicates,true).
preference_default_value(use_closure_expansion_memoization,false).
preference_default_value(allow_incomplete_partial_setup_constants,false).
preference_default_value(warn_if_definition_hides_variable,true).
preference_default_value(warn_when_expanding_infinite_closures,5).
preference_default_value(strict_raise_warnings,true).
preference_default_value(ignore_prj_types,false).
%preference_default_value(never_convert_closures_into_explicit_form,false).
%preference_default_value(convert_types_into_closures,true).
%preference_default_value(convert_closures_into_explicit_form_for_store,true). % <--- maybe remove preference
%preference_default_value(allow_empty_global_sets,false).
%preference_default_value(test_case_gen_stop_at_first_error,false).
preference_default_value(use_tk_custom_state_viewer,true).
preference_default_value(use_small_window,false).
preference_default_value(tk_show_source_pane,true).
preference_default_value(ignore_hash_collisions,false).
preference_default_value(forget_state_space,false).
%preference_default_value(enable_permutation_reducation,false).
%preference_default_value(use_symmetry_marker_hash,false).
%preference_default_value(enable_canonical_symmetry,false).
%preference_default_value(use_nauty_symmetry_reduction,false).
preference_default_value(symmetry_mode,off).
preference_default_value(selected_bparser,java).
preference_default_value(bool_expression_as_predicate,false).
preference_default_value(eventtrace,false).
preference_default_value(store_event_transinfo,false).
preference_default_value(try_kodkod_on_load,false).
preference_default_value(kodkod_for_components,true).
preference_default_value(use_record_construction,true). % was false


preference_default_value(check_prolog_b_ast,false).
preference_default_value(dot_print_action_arguments,true).
preference_default_value(dot_print_functions,false).
preference_default_value(dot_print_self_loops,true).
preference_default_value(dot_print_arc_colors,true).
preference_default_value(dot_print_root,true).
preference_default_value(dot_print_leaves,true).
preference_default_value(dot_print_node_ids,false).
preference_default_value(dot_print_use_constants,true).
preference_default_value(dot_print_node_info,true).
preference_default_value(dot_use_ps_viewer,false).
preference_default_value(dot_use_alterate_dot_viewer,false).
preference_default_value(path_to_dot,'dot').
preference_default_value(path_to_dotty,'dotty').
preference_default_value(path_to_dotty2,'dotty').
%preference_default_value(path_to_dotty,'/Applications/OmniGraffle/OmniGraffle.app/Contents/MacOS/OmniGraffle').
preference_default_value(path_to_ps_viewer,'open -a /Applications/Preview.app') :- host_platform(darwin).
preference_default_value(path_to_ps_viewer,'C:/Program Files/Ghostgum/gsview/gsview32') :- host_platform(windows).
preference_default_value(path_to_ps_viewer,'gv') :- host_platform(linux) ; host_platform(unknown).
%preference_default_value(path_to_ps_viewer,'open -a /Applications/MacGhostViewX_Folder/MacGhostViewX.app').
preference_default_value(dot_root_shape,invtriangle).
preference_default_value(dot_node_penwidth,2).
preference_default_value(dot_normal_node_shape,box). % egg,ellipse
preference_default_value(dot_current_node_shape,doubleoctagon).
preference_default_value(dot_open_node_colour,'#F4E3C1'). %'#E88C59' '#C96D37'
preference_default_value(dot_scope_limit_node_colour,gray).
preference_default_value(dot_normal_node_colour,'#99BF38').
preference_default_value(dot_invariant_violated_node_colour,'#FF3800'). %'#FF4409'
preference_default_value(dot_fill_normal_nodes,false).
preference_default_value(dot_normal_arc_colour,'#006391').
preference_default_value(dot_node_font_size,12).
preference_default_value(dot_edge_font_size,12).
preference_default_value(dot_colour_goal_nodes,true).
preference_default_value(dot_goal_node_colour,orange).
preference_default_value(dot_print_node_properties,false).

preference_default_value(font_size,'Courier 12').
preference_default_value(font_name,'Courier').
preference_default_value(font_size_only,12).
preference_default_value(use_font_size_for_columns,false).
preference_default_value(allow_source_code_editing,true).
preference_default_value(show_line_numbers,false).
preference_default_value(highlight_brackets,true).
preference_default_value(highlight_current_line,true).
preference_default_value(ask_when_content_changed,false).
preference_default_value(user_not_beginner,false).
preference_default_value(user_is_an_expert_with_accessto_source_distribution,false).
preference_default_value(default_to_runtime_type_checking_on_startup_for_expert,false).
preference_default_value(path_to_probe,'C:/probe/probe-1.30-windows/probe.exe') :- host_platform(windows).
preference_default_value(path_to_probe,'more') :- host_platform(X), X\==windows.
preference_default_value(path_to_spin,'xspin').
preference_default_value(path_to_text_editor,'emacs').
preference_default_value(path_to_fuzz,'fuzz').
preference_default_value(path_to_fdr,'fdr2').
preference_default_value(path_to_csp_typechecker,'/Users/leuschel/svn_root/ProB/cia/haskell/fecsp/typeCheckCSP').
preference_default_value(path_to_bcomp,'bcomp').
preference_default_value(number_of_recent_documents,15).
preference_default_value(number_of_searched_patterns,9).
preference_default_value(number_of_replaced_patterns,9).
preference_default_value(number_of_eval_history_elements,50).
preference_default_value(number_of_eval_csp_history_elements,50).
preference_default_value(number_of_checked_ltl_formulas,20).

preference_default_value(do_syntax_highlighting,true).
preference_default_value(sh_type_colour,darkred).
preference_default_value(sh_logical_colour,tomato).
preference_default_value(sh_assignments_colour,darkblue).
preference_default_value(sh_operators,'#5883be').
preference_default_value(sh_top_level_keywords,blue).
preference_default_value(sh_control_keywords,darkviolet).
%preference_default_value(sh_includes_keywords,darkslateblue).
preference_default_value(sh_comments,'#784e50').
preference_default_value(sh_unsupported_background,yellow).

preference_default_value(expand_forall_upto,100).
preference_default_value(double_evaluation_when_analysing,true).
preference_default_value(expand_avl_upto,100).

preference_default_value(allow_new_ops_in_refinement,false).
preference_default_value(use_po,true).
preference_default_value(try_operation_reuse,false).
preference_default_value(randomise_operation_order,false).
preference_default_value(set_rand,false).
preference_default_value(rand_seed,2342452).

preference_default_value(machines_path,Dir) :-
   catch(absolute_file_name(examples('.'),Dir),_,Dir='.').

% the following causes problems when loading ProB:
%:- use_module(plugins,[plugin_preference_default_value/2]).
preference_default_value(X,Y) :- plugins:plugin_preference_default_value(X,Y).

/* ------------------------- */


:- assert_must_succeed( \+((preference_description(P,_D),\+preference_default_value(P,_Val)))).
:- assert_must_succeed( \+((preference_default_value(P,_D), \+preference_description(P,_Val)))).
preference_description(maxNrOfInitialisations,'Max Number of Initialisations Computed').
preference_description(maxNrOfEnablingsPerOperation,'Max Number of Enablings per Operation Computed').
preference_description(time_out,'Time out for computing enabled transitions (in ms)').
preference_description(globalsets_fdrange,'Size of unspecified deferred sets in SETS section').
preference_description(maxint,'MAXINT, used for expressions such as xx::NAT (2147483647 for 4 byte ints)').
preference_description(minint,'MININT, used for expressions such as xx::INT (-2147483648 for 4 byte ints)'). % Atelier-B uses 2147483647
preference_description(number_animated_abstractions,'How many levels of refined models are animated by default').
%preference_description(findOnlyOneSol,'Find only one solution per Operation or Initialisation').
preference_description(debug_time_out_factor,'Debug Properties: Multiplying Factor for Timeout').
preference_description(use_large_jvm_for_parser,'Use large Heap and Stack size for Java Parser').
%preference_description(do_invariant_checking,'Check invariant').
preference_description(do_neginvariant_checking,'Check negation of invariant').
%preference_description(do_full_invariant_checking,'Check types when checking invariant').
%preference_description(use_new_kernel,'Use new Kernel (only some basic operations supported, only fore PRE and INITs)').
preference_description(find_abort_values,'Try more aggressively to detect not well-defined expressions (e.g. applying function outside of domain), may slow down animator').
preference_description(cspm_animate_all_processes,'Animate all CSP-M processes (also processes with arguments)').
preference_description(cspm_animate_all_processes_without_arguments,'Animate all CSP-M processes without arguments (not just MAIN)').
%preference_description(cspm_display_inout,'Display unsynchronised inputs/outputs as ?/!').
preference_description(cspm_detailed_animation,'Provide more fine-grained source-location animation for CSP (can slow model checking down)').
preference_description(cspm_strip_source_location,'Strip CSP source-location information (for RHS of processes, except for internal choice)').
preference_description(csp_event_visible_to_user,'Return the CSP event as visible event to the user (option only relevant for models in \'CSP || B\' mode).').
%preference_description(cspm_allow_incomplete_records,'Allow construction of incomplete records (e.g., for Casper files)').
preference_description(use_clpfd_solver,'Use CLP(FD) solver for B integers (restricts range to -2^28..2^28-1 on 32 bit machines)').
preference_description(convert_comprehension_sets_into_closures,'Lazy expansion of lambdas and set comprehensions (SYMBOLIC)').
preference_description(use_smt_mode,'Enable SMT-Mode (aggressive treatment of : and /: inside predicates)'). % will be set automatically in CBC mode; should normally not be set by user
preference_description(fail_if_clpfd_timeout,'Assume failure if posting a CLP(FD) constraint times-out').
%preference_description(allow_recursive_closures,'Lazy expansion of *Recursive* set Comprehensions and lambdas').
preference_description(use_static_ordering,'Use static ordering to enumerate constants which occur in most PROPERTIES first').
preference_description(use_state_packing,'Use more aggressive COMPRESSION when storing states').
preference_description(provide_trace_information,'Provide various tracing information on the terminal/console.').
preference_description(partition_predicates,'Partition predicates (PROPERTIES) into components').
preference_description(use_closure_expansion_memoization,'MEMOize expansions of set comprehensions and lambda abstractions.').
preference_description(allow_incomplete_partial_setup_constants,'Allow ProB to proceed even if only part of the CONSTANTS have been found.').
preference_description(warn_if_definition_hides_variable,'Warn if a DEFINITION hides a variable with the same name').
preference_description(warn_when_expanding_infinite_closures,'Warn when expanding infinite closures if MAXINT larger than: ').
preference_description(strict_raise_warnings,'Treat warnings as errors.').
preference_description(ignore_prj_types,'Ignore types of prj1 and prj2 (i.e., prj1({1},{1})(2,3) is considered well-defined').

%preference_description(never_convert_closures_into_explicit_form,'Never convert closures back into explicit form (only use for debugging !)').
%preference_description(convert_types_into_closures,'Lazy expansion of type expressions (POW,-->,...)').
%preference_description(convert_closures_into_explicit_form_for_store,'Convert lazy form back into explicit form for Variables, Constants, Operation Arguments').
%preference_description(allow_empty_global_sets,'Allow empty global SETs').
%preference_description(test_case_gen_stop_at_first_error,'Stop at first error in test case generation').
preference_description(use_tk_custom_state_viewer,'Use custom Tk viewer to represent state of B Machine').
preference_description(use_small_window,'Use a small main window for ProB (useful on small screens)').
preference_description(tk_show_source_pane,'Show the B Source Code').
preference_description(ignore_hash_collisions,'Ignore Hash Collisions (if true not all states may be computed, visited states are not memorised !)').
preference_description(forget_state_space,'Do not remember state space (mainly useful in conjunction with Ignore Hash Collisions)').
%preference_description(enable_permutation_reducation,'Enable Permutation Flooding').
%preference_description(use_symmetry_marker_hash,'Use Symmetry Hash Marker').
%preference_description(enable_canonical_symmetry,'Enable Symmetry Reduction by Canonical Form Computation').
%preference_description(use_nauty_symmetry_reduction,'Use Nauty ((c) McKay) C Library for Symmetry Reduction with Canonical Form Computation').
preference_description(symmetry_mode,'Symmetry Mode: off,flood,canon,nauty,hash').
preference_description(selected_bparser,'Parser used for B Machines: java, prolog, haskell, all').
preference_description(bool_expression_as_predicate,'Extended Predicate Syntax (allow p&q as syntactic sugar for (p=TRUE) & (q=TRUE), only possible in REPL)' ).

preference_description(eventtrace,'Store trace for each event (may cause large memory requirement)').
preference_description(store_event_transinfo,'Store Event-B Event-Info for each event (may cause larger memory requirement)').
preference_description(try_kodkod_on_load,'Apply translation to Kodkod on PROPERTIES when loading machines').
preference_description(kodkod_for_components,'Translate only complete components to Kodkod').
preference_description(use_record_construction,'Records: Check if axioms/properties describe a record pattern').
preference_description(show_eventb_any_arguments,'Show top-level ANY variable values of B Operations without parameters as parameters').
preference_description(show_initialisation_arguments,'Show effect of initialisation and setup constants in operation name').
%preference_description(animate_skip_operations,'Animate operations which are skip or PRE C THEN skip').
preference_description(treat_outermost_pre_as_select,'Treat outermost PRE as SELECT').
preference_description(require_operations_to_assign_all_outputs,'Require operations to assign to all outputs').

preference_description(check_prolog_b_ast,'Check the abstract syntax tree for B machines').
preference_description(dot_print_action_arguments,'Print Operation Arguments').
preference_description(dot_print_functions,'Print Functions (operations that return values w/o changing state)').
preference_description(dot_print_self_loops,'Print Self-Loops').
preference_description(dot_print_leaves,'Print Leaves').
preference_description(dot_print_arc_colors,'Colour the arcs (disable for import in OmniGraffle)').
preference_description(dot_print_root,'Print Root Node').
preference_description(dot_print_node_ids,'Print Node Identifiers').
preference_description(dot_print_node_info,'Print Node information (values of variables)').
preference_description(dot_print_use_constants,'Use constants for deferred set values if possible').
preference_description(dot_use_ps_viewer,'Use Postscript Viewer').
preference_description(dot_use_alterate_dot_viewer,'Use alternate Dot viewer').
preference_description(path_to_dot,'Path/Command for dot program (using "/" even on Windows. Try avoiding spaces in path)').
preference_description(path_to_dotty,'Path/Command for dot viewer (e.g., dotty)').
preference_description(path_to_dotty2,'Path/Command for alternate dot viewer').
preference_description(path_to_ps_viewer,'Path/Command for Postscript viewer').
preference_description(dot_root_shape,'Shape for root node (invtriangle,triangle, ellipse, box, diamond,...)').
preference_description(dot_normal_node_shape,'Shape for normal node').
preference_description(dot_current_node_shape,'Shape for current node').
preference_description(dot_open_node_colour,'Colour for open node (red,green,blue,gray,... or rgb hex #ff00ff) or adjacent to an open node if print leaves is off').
preference_description(dot_scope_limit_node_colour,'Colour for open, but uninteresting nodes (red,green,blue,gray,... or rgb hex #ff00ff)').
preference_description(dot_normal_node_colour,'Colour for normal node').
preference_description(dot_node_penwidth,'Pen width for nodes').
preference_description(dot_invariant_violated_node_colour,'Colour for invariant violation').
preference_description(dot_fill_normal_nodes,'Fill normal nodes with color').
preference_description(dot_normal_arc_colour,'Colour for normal arcs').
preference_description(dot_node_font_size,'Font Size for Node labels').
preference_description(dot_edge_font_size,'Font Size for Edge labels').
preference_description(dot_colour_goal_nodes,'Colour the nodes that satisfy the GOAL predicate').
preference_description(dot_goal_node_colour,'Colour for nodes satisfying the GOAL predicate').
preference_description(dot_print_node_properties,'Show properties of node').

preference_description(font_size,'Font size in main window').
preference_description(font_name,'Font name in text editor').
preference_description(font_size_only,'Font size in text editor').
preference_description(use_font_size_for_columns,'Use font size also for columns (you need to reselect font size + Courier will be used)').
preference_description(allow_source_code_editing,'Allow editing and saving of main B Machine file').
preference_description(show_line_numbers,'Show line numbers in main window').
preference_description(highlight_brackets,'Highlights brackets').
preference_description(highlight_current_line,'Highlights current line').
preference_description(ask_when_content_changed,'Asking when file content changed from external program').
preference_description(user_not_beginner,'User is not a beginner').
preference_description(user_is_an_expert_with_accessto_source_distribution,'User is expert with access to source distribution of ProB').
preference_description(default_to_runtime_type_checking_on_startup_for_expert,'When starting up (in expert mode) turn runtime type checking on').
preference_description(path_to_probe,'Path to Probe or other tool to view .csp files').
preference_description(path_to_spin,'Path to Spin or other tool to view .prom files').
preference_description(path_to_text_editor,'Path to External Text Editor').
preference_description(path_to_fuzz,'Path to the fuzz tool for ProZ').
preference_description(path_to_fdr,'Path to the FDR2 tool').
preference_description(path_to_csp_typechecker,'Path to a CSP Type Checking tool').
preference_description(path_to_bcomp,'Path to a bcomp (Atelier B Parser and Type Checker )').
preference_description(number_of_recent_documents,'Number of recently opened file names memorised').
preference_description(number_of_searched_patterns,'Number of recently searched patterns in text').
preference_description(number_of_replaced_patterns,'Number of recently replaced patterns in text').
preference_description(number_of_eval_history_elements,'Number of recently runs of commands in the eval window').
preference_description(number_of_eval_csp_history_elements,'Number of recently runs of commands in the eval window').
preference_description(number_of_checked_ltl_formulas, 'Number of recently checked ltl formulas').

preference_description(do_syntax_highlighting,'Enable Syntax Colour Highlighting').
preference_description(sh_type_colour,'Colour of typing constructs: NAT,POW,... (red,green,... or rgb hex #ff00ff)').
preference_description(sh_logical_colour,'Colour of logical operators: &,or,...').
preference_description(sh_assignments_colour,'Colour of assignments: :=, <--').
preference_description(sh_operators,'Colour of normal operators ran,dom,...').
preference_description(sh_top_level_keywords,'Colour of top level keywords: MACHINE, INVARIANT,...').
preference_description(sh_control_keywords,'Colour of control keywords: IF, CASE,...').
%preference_description(sh_includes_keywords,'Colour of inclusion related keywords: USES,...').
preference_description(sh_comments,'Colour of comments /* ... */').
preference_description(sh_unsupported_background,'Background colour for unsupported features (closure,...)').

preference_description(expand_forall_upto,'When analysing predicates: max. domain size for expansion of forall:').
preference_description(double_evaluation_when_analysing,'Evaluate PREDICATES positively and negatively when analysing:').
preference_description(expand_avl_upto,'Max size for pretty-printing sets (-1 means no limit) :').


preference_description(allow_new_ops_in_refinement,'Allow introduction of new operations in refinements').

preference_description(use_po,'Use PROOF information to restrict invariant checking to affected clauses.').
%preference_description(use_po,'Restrict invariant checking to affected clauses. Also remove clauses that are proven (EventB)').
preference_description(try_operation_reuse,'Try and reuse previously computed operation effects in B/Event-B').
preference_description(randomise_operation_order,'Randomise order of operations when computing successor states').
preference_description(set_rand,'Use a fix seed for random number generator. Allows to reproduce results.').
preference_description(rand_seed,'Seed for Random Number Generator').

preference_description(machines_path,'Last open directory').

%:- use_module(plugins,[plugin_preference_description/2]).
preference_description(X,Y) :- plugins:plugin_preference_description(X,Y).


/* ------------------------- */


%preference_update_action(do_full_invariant_checking,_) :- !, b_types_precompile.
%preference_change_requires_reload(enable_canonical_symmetry,true) :- !, .
%preference_update_action(use_nauty_symmetry_reduction,true)  :- !, graph_iso_nauty:nauty_init.
%:- use_module(probsrc(bmachine),[bmachine_is_precompiled/0]).
%:- use_module(probsrc(graph_iso_nauty),[nauty_init/0]).
preference_update_action(symmetry_mode,nauty)  :- !,
       (bmachine:bmachine_is_precompiled -> graph_iso_nauty:nauty_init ; true).
preference_update_action(_,_).

/* ------------------------- */

:- assert_must_succeed( \+((preference_val_type(P,_D),\+preference_default_value(P,_Val)))).
:- assert_must_succeed( \+((preference_default_value(P,_D), \+preference_val_type(P,_Val)))).
preference_val_type(maxint,nat).
preference_val_type(minint,neg).
preference_val_type(number_animated_abstractions,nat).
preference_val_type(globalsets_fdrange,nat).
%preference_val_type(do_invariant_checking,bool).
preference_val_type(do_neginvariant_checking,bool).
%preference_val_type(do_full_invariant_checking,bool).
%preference_val_type(use_new_kernel,bool).
preference_val_type(find_abort_values,bool).
preference_val_type(treat_outermost_pre_as_select,bool).
preference_val_type(require_operations_to_assign_all_outputs,bool).
preference_val_type(cspm_animate_all_processes_without_arguments,bool).
preference_val_type(cspm_animate_all_processes,bool).
%preference_val_type(cspm_display_inout,bool).
preference_val_type(cspm_detailed_animation,bool).
preference_val_type(cspm_strip_source_location,bool).
preference_val_type(csp_event_visible_to_user,bool).
%preference_val_type(cspm_allow_incomplete_records,bool).
%preference_val_type(findOnlyOneSol,bool).
preference_val_type(show_initialisation_arguments,bool).
%preference_val_type(animate_skip_operations,bool).
preference_val_type(show_eventb_any_arguments,bool).
preference_val_type(maxNrOfInitialisations,nat).
preference_val_type(maxNrOfEnablingsPerOperation,nat).
preference_val_type(time_out,nat1).
preference_val_type(debug_time_out_factor,nat1).
preference_val_type(use_large_jvm_for_parser,bool).
preference_val_type(convert_comprehension_sets_into_closures,bool).
preference_val_type(use_clpfd_solver,bool).
preference_val_type(use_smt_mode,bool).
preference_val_type(fail_if_clpfd_timeout,bool).
%preference_val_type(allow_recursive_closures,bool).
preference_val_type(use_static_ordering,bool).
preference_val_type(use_state_packing,bool).
preference_val_type(provide_trace_information,bool).
preference_val_type(partition_predicates,bool).
preference_val_type(use_closure_expansion_memoization,bool).
preference_val_type(allow_incomplete_partial_setup_constants,bool).
preference_val_type(warn_if_definition_hides_variable,bool).
preference_val_type(warn_when_expanding_infinite_closures,int).
preference_val_type(strict_raise_warnings,bool).
preference_val_type(ignore_prj_types,bool).
%preference_val_type(never_convert_closures_into_explicit_form,bool).
%preference_val_type(convert_types_into_closures,bool).
%preference_val_type(convert_closures_into_explicit_form_for_store,bool).
%preference_val_type(test_case_gen_stop_at_first_error,bool).
%preference_val_type(allow_empty_global_sets,bool).
preference_val_type(use_tk_custom_state_viewer,bool).
preference_val_type(use_small_window,bool).
preference_val_type(tk_show_source_pane,bool).
preference_val_type(ignore_hash_collisions,bool).
preference_val_type(forget_state_space,bool).
%preference_val_type(enable_permutation_reducation,bool).
%preference_val_type(use_symmetry_marker_hash,bool).
%preference_val_type(enable_canonical_symmetry,bool).
%preference_val_type(use_nauty_symmetry_reduction,bool).
preference_val_type(symmetry_mode,[off,flood,nauty,hash]).
preference_val_type(selected_bparser,[java,prolog,all]). %haskell
preference_val_type(bool_expression_as_predicate,bool).
preference_val_type(eventtrace,bool).
preference_val_type(store_event_transinfo,bool).
preference_val_type(try_kodkod_on_load,bool).
preference_val_type(kodkod_for_components,bool).
preference_val_type(use_record_construction,bool).

preference_val_type(dot_print_node_ids,bool).
preference_val_type(dot_print_use_constants,bool).
preference_val_type(dot_print_node_info,bool).
preference_val_type(check_prolog_b_ast,bool).
preference_val_type(dot_print_action_arguments,bool).
preference_val_type(dot_print_functions,bool).
preference_val_type(dot_print_self_loops,bool).
preference_val_type(dot_print_root,bool).
preference_val_type(dot_print_leaves,bool).
preference_val_type(dot_print_arc_colors,bool).
preference_val_type(dot_use_ps_viewer,bool).
preference_val_type(dot_use_alterate_dot_viewer,bool).
preference_val_type(path_to_dot,path).
preference_val_type(path_to_dotty,path).
preference_val_type(path_to_dotty2,path).
preference_val_type(path_to_ps_viewer,path).
preference_val_type(dot_root_shape,dot_shape).
preference_val_type(dot_node_penwidth,nat).
preference_val_type(dot_normal_node_shape,dot_shape).
preference_val_type(dot_current_node_shape,dot_shape).
preference_val_type(dot_open_node_colour,rgb_color).
preference_val_type(dot_scope_limit_node_colour,rgb_color).
preference_val_type(dot_normal_node_colour,rgb_color).
preference_val_type(dot_invariant_violated_node_colour,rgb_color).
preference_val_type(dot_fill_normal_nodes,bool).
preference_val_type(dot_normal_arc_colour,rgb_color).
preference_val_type(dot_node_font_size,nat1).
preference_val_type(dot_edge_font_size,nat1).
preference_val_type(dot_colour_goal_nodes,bool).
preference_val_type(dot_goal_node_colour,rgb_color).
preference_val_type(dot_print_node_properties,bool).

preference_val_type(font_name,string).
preference_val_type(font_size_only,nat1).
preference_val_type(font_size,string).
preference_val_type(use_font_size_for_columns,bool).
preference_val_type(allow_source_code_editing,bool).
preference_val_type(show_line_numbers,bool).
preference_val_type(highlight_brackets,bool).
preference_val_type(highlight_current_line,bool).
preference_val_type(ask_when_content_changed,bool).

preference_val_type(user_not_beginner,bool).
preference_val_type(user_is_an_expert_with_accessto_source_distribution,bool).
preference_val_type(default_to_runtime_type_checking_on_startup_for_expert,bool).
preference_val_type(path_to_probe,path).
preference_val_type(path_to_fuzz,path).
preference_val_type(path_to_fdr,path).
preference_val_type(path_to_csp_typechecker,path).
preference_val_type(path_to_bcomp,path). % Atelier B parser & type checker
preference_val_type(path_to_spin,path).
preference_val_type(path_to_text_editor,path).
preference_val_type(number_of_recent_documents,nat).
preference_val_type(number_of_searched_patterns,nat).
preference_val_type(number_of_replaced_patterns,nat).
preference_val_type(number_of_eval_history_elements,nat).
preference_val_type(number_of_eval_csp_history_elements,nat).
preference_val_type(number_of_checked_ltl_formulas,nat).

preference_val_type(do_syntax_highlighting,bool).
preference_val_type(sh_type_colour,rgb_color).
preference_val_type(sh_logical_colour,rgb_color).
preference_val_type(sh_assignments_colour,rgb_color).
preference_val_type(sh_operators,rgb_color).
preference_val_type(sh_top_level_keywords,rgb_color).
preference_val_type(sh_control_keywords,rgb_color).
%preference_val_type(sh_includes_keywords,rgb_color).
preference_val_type(sh_comments,rgb_color).
preference_val_type(sh_unsupported_background,rgb_color).

preference_val_type(expand_forall_upto,nat).
preference_val_type(double_evaluation_when_analysing,bool).
preference_val_type(expand_avl_upto,int).

preference_val_type(allow_new_ops_in_refinement,bool).
preference_val_type(use_po,bool).
preference_val_type(try_operation_reuse,bool).
preference_val_type(randomise_operation_order,bool).
preference_val_type(set_rand,bool).
preference_val_type(rand_seed,nat).

preference_val_type(machines_path,path). %directory_path). % preference not so critical; default value may not exist

%:- use_module(plugins,[plugin_preference_val_type/2]).
preference_val_type(X,Y) :- plugins:plugin_preference_val_type(X,Y).


/* ------------------------- */


:- assert_must_succeed( \+((preference_category(P,_D),\+preference_default_value(P,_Val)))).
:- assert_must_succeed( \+((preference_default_value(P,_D), \+preference_category(P,_Val)))).
preference_category(maxint,animation).
preference_category(minint,animation).
preference_category(globalsets_fdrange,animation).
%preference_category(do_invariant_checking,advanced).
preference_category(do_neginvariant_checking,advanced).
%preference_category(do_full_invariant_checking,animation).
%preference_category(use_new_kernel,advanced).
preference_category(find_abort_values,animation).
preference_category(treat_outermost_pre_as_select,animation).
preference_category(require_operations_to_assign_all_outputs,animation).
preference_category(cspm_animate_all_processes_without_arguments,csp_prefs).
preference_category(cspm_animate_all_processes,csp_prefs).
%preference_category(cspm_display_inout,csp_prefs).
preference_category(cspm_detailed_animation,csp_prefs).
preference_category(cspm_strip_source_location,csp_prefs).
preference_category(csp_event_visible_to_user,csp_prefs).
%preference_category(cspm_allow_incomplete_records,csp_prefs).
%preference_category(findOnlyOneSol,advanced).
preference_category(show_eventb_any_arguments,animation).
preference_category(maxNrOfInitialisations,animation).
preference_category(maxNrOfEnablingsPerOperation,animation).
preference_category(time_out,animation).
preference_category(number_animated_abstractions,animation).
preference_category(show_initialisation_arguments,animation).
%preference_category(animate_skip_operations,animation).
preference_category(debug_time_out_factor,advanced).
preference_category(use_large_jvm_for_parser,advanced).
preference_category(convert_comprehension_sets_into_closures,animation).
preference_category(use_clpfd_solver,animation).
preference_category(use_smt_mode,advanced).
preference_category(fail_if_clpfd_timeout,advanced).
%preference_category(allow_recursive_closures,hidden).
preference_category(use_static_ordering,advanced).
preference_category(use_state_packing,advanced).
preference_category(provide_trace_information,advanced).
preference_category(partition_predicates,advanced).
preference_category(use_closure_expansion_memoization,advanced).
preference_category(allow_incomplete_partial_setup_constants,advanced).
preference_category(warn_if_definition_hides_variable,advanced).
preference_category(warn_when_expanding_infinite_closures,advanced).
preference_category(strict_raise_warnings,advanced).
preference_category(ignore_prj_types,advanced).
%preference_category(never_convert_closures_into_explicit_form,hidden).
%preference_category(convert_types_into_closures,animation).
%preference_category(convert_closures_into_explicit_form_for_store,animation).
%preference_category(allow_empty_global_sets,advanced).
preference_category(ignore_hash_collisions,advanced).
preference_category(forget_state_space,advanced).
%preference_category(enable_permutation_reducation,advanced).
%preference_category(use_symmetry_marker_hash,advanced).
%preference_category(enable_canonical_symmetry,advanced).
%preference_category(use_nauty_symmetry_reduction,advanced).
preference_category(symmetry_mode,hidden).
preference_category(selected_bparser,advanced).
preference_category(bool_expression_as_predicate,advanced).
preference_category(eventtrace,animation).
preference_category(store_event_transinfo,hidden). % automatically set by probcli
preference_category(try_kodkod_on_load,hidden).
preference_category(kodkod_for_components,advanced).
preference_category(use_record_construction,advanced).
                    
%preference_category(test_case_gen_stop_at_first_error,advanced).
preference_category(use_tk_custom_state_viewer,hidden). % actually : if user wants to turn it off; simply do not define ANIMATION_FUNCTION
preference_category(use_small_window,animation).
preference_category(tk_show_source_pane,animation).


preference_category(check_prolog_b_ast,advanced).

preference_category(dot_print_node_ids,dot).
preference_category(dot_print_use_constants,bool).
preference_category(dot_print_node_info,dot).
preference_category(dot_print_action_arguments,dot).
preference_category(dot_print_functions,dot).
preference_category(dot_print_self_loops,dot).
preference_category(dot_print_root,dot).
preference_category(dot_print_leaves,dot).
preference_category(dot_print_arc_colors,dot).
preference_category(dot_use_ps_viewer,hidden).
preference_category(dot_use_alterate_dot_viewer,hidden).
preference_category(path_to_dotty,dot).
preference_category(path_to_dotty2,dot).
preference_category(path_to_dot,dot).
preference_category(path_to_ps_viewer,dot).
preference_category(dot_root_shape,dot).
preference_category(dot_node_penwidth,dot).
preference_category(dot_normal_node_shape,dot).
preference_category(dot_current_node_shape,dot).
preference_category(dot_open_node_colour,dot).
preference_category(dot_scope_limit_node_colour,dot).
preference_category(dot_normal_node_colour,dot).
preference_category(dot_invariant_violated_node_colour,dot).
preference_category(dot_fill_normal_nodes,dot).
preference_category(dot_normal_arc_colour,dot).
%PJT Start
%preference_category(dot_font_size,hidden).
preference_category(dot_node_font_size,dot).
preference_category(dot_edge_font_size,dot).
preference_category(dot_colour_goal_nodes,dot).
preference_category(dot_goal_node_colour,dot).
preference_category(dot_print_node_properties,dot).

%PJT End
preference_category(font_size,gui_prefs).
preference_category(font_name,gui_prefs).
preference_category(font_size_only,gui_prefs).
preference_category(use_font_size_for_columns,advanced).
preference_category(allow_source_code_editing,gui_prefs).
preference_category(show_line_numbers,gui_prefs).
preference_category(highlight_brackets,gui_prefs).
preference_category(highlight_current_line,gui_prefs).
preference_category(ask_when_content_changed,gui_prefs).
preference_category(user_not_beginner,gui_prefs).

preference_category(user_is_an_expert_with_accessto_source_distribution,hidden).
preference_category(default_to_runtime_type_checking_on_startup_for_expert,hidden).
preference_category(path_to_probe,csp_prefs).
preference_category(path_to_fuzz,advanced).
preference_category(path_to_fdr,csp_prefs).
preference_category(path_to_csp_typechecker,csp_prefs).
preference_category(path_to_bcomp,advanced).
preference_category(path_to_spin,advanced).
preference_category(path_to_text_editor,advanced).
preference_category(number_of_recent_documents,advanced).
preference_category(number_of_searched_patterns,advanced).
preference_category(number_of_replaced_patterns,advanced).
preference_category(number_of_eval_history_elements,advanced).
preference_category(number_of_eval_csp_history_elements,advanced).
preference_category(number_of_checked_ltl_formulas,advanced).
preference_category(allow_new_ops_in_refinement,advanced).
preference_category(use_po,advanced).
preference_category(try_operation_reuse,advanced).
preference_category(randomise_operation_order,advanced).
preference_category(set_rand,advanced).
preference_category(rand_seed,advanced).

preference_category(do_syntax_highlighting,syntax_highlighting).
preference_category(sh_type_colour,syntax_highlighting).
preference_category(sh_logical_colour,syntax_highlighting).
preference_category(sh_assignments_colour,syntax_highlighting).
preference_category(sh_operators,syntax_highlighting).
preference_category(sh_top_level_keywords,syntax_highlighting).
preference_category(sh_control_keywords,syntax_highlighting).
%preference_category(sh_includes_keywords,syntax_highlighting).
preference_category(sh_comments,syntax_highlighting).
preference_category(sh_unsupported_background,syntax_highlighting).

preference_category(expand_forall_upto,advanced).
preference_category(double_evaluation_when_analysing,advanced).
preference_category(expand_avl_upto,advanced).

preference_category(machines_path,hidden).

%:- use_module(plugins,[plugin_preference_category/2]).
preference_category(X,Y) :- plugins:plugin_preference_category(X,Y).


/* ------------------------- */

invalid_preference_value(maxint,X,'MAXINT smaller than MININT') :-
   get_preference(minint,MININT), X <MININT.
invalid_preference_value(minint,X,'MININT larger than MAXINT') :-
   get_preference(maxint,MAXINT), X >MAXINT.

/* ------------------------- */
:- assert_must_succeed( \+((preferences:obsolete_preference(P),preference_default_value(P,_Val)))).
obsolete_preference(animate_skip_operations).
obsolete_preference(debug_try_minimise).
obsolete_preference(convert_types_into_closures).
obsolete_preference(do_invariant_checking).
obsolete_preference(use_avl_trees_for_sets).
obsolete_preference(dot_font_size).
obsolete_preference(only_label_base_types).
obsolete_preference(use_new_kernel).
obsolete_preference(findOnlyOneSol).
obsolete_preference(enable_permutation_reducation).
obsolete_preference(use_symmetry_marker_hash).
obsolete_preference(enable_canonical_symmetry).
obsolete_preference(use_nauty_symmetry_reduction).
obsolete_preference(do_full_invariant_checking).
obsolete_preference(test_case_gen_stop_at_first_error).
obsolete_preference(use_failure_refinement).
obsolete_preference(do_on_the_fly_refinement).
obsolete_preference(never_convert_closures_into_explicit_form).
obsolete_preference(allow_empty_global_sets).
obsolete_preference(multilevel_animation).
obsolete_preference(remove_multilevel_prefix).
obsolete_preference(cspm_allow_incomplete_records).
obsolete_preference(cspm_display_inout).
obsolete_preference(allow_recursive_closures).
obsolete_preference(absint_abstract_domain_module).
obsolete_preference(absint_use_widening).
obsolete_preference(absint_soft_widening).
obsolete_preference(absint_extrapolation_threshold).
obsolete_preference(convert_closures_into_explicit_form_for_store).

% might be re-included later
obsolete_preference(plugin(absint,use_widening)).
obsolete_preference(plugin(absint,soft_widening)).
obsolete_preference(plugin(absint,extrapolation_threshold)).
obsolete_preference(plugin(absint,retract_widening_related_states)).



volatile_preference(store_event_transinfo). % not necessary to store in file; value will not be changed when loading pref file
/* ------------------------- */

is_of_type(true,bool).
is_of_type(false,bool).

is_of_type(X,neg) :- number(X), X=<0.
is_of_type(X,nat) :- number(X), X>=0.
is_of_type(X,nat1) :- number(X), X>0.
is_of_type(X,range(Low,Up)) :- number(X), X>=Low, X=<Up.
is_of_type(X,int) :- number(X).
is_of_type(X,string) :- atomic(X).
is_of_type(X,path) :- atomic(X). % some paths are files, some directorys (.app) and some aliases
is_of_type(X,file_path) :- atomic(X),check_file_exists(X).
is_of_type(X,directory_path) :- atomic(X),check_directory_exists(X).
is_of_type(X,term) :- nonvar(X).
is_of_type(X,ListOfCsts) :- ListOfCsts=[_|_], member(X,ListOfCsts).

is_of_type(X,dot_shape) :- is_of_type(X,[triangle,ellipse,box,diamond,hexagon,octagon,house,
                                         invtriangle,invhouse,invtrapez,doubleoctagon,egg]).
                                         
is_of_type(X,rgb_color) :-
   member(X,[red,green,blue,yellow,orange,black,white,gray,brown,violet,darkred,tomato,darkblue,
             darkgray,darkviolet,darkslateblue,lightblue,lightgray,maroon2,olivedrab2,
             steelblue,chartreuse3,chartreuse4])
   ;
   (atomic(X),name(X,[35,H1,H2,H3,H4,H5,H6]),
     hex(H1),hex(H2),hex(H3),hex(H4),hex(H5),hex(H6)). /* should be of the form #ff00ff */

hex(X) :- atomic(X), number(X), ((X<103,X>96) ; (X<71,X>64) ; (X>47,X<58)).

:- use_module(library(file_systems)).
check_file_exists(X) :- file_exists(X) -> true ; add_message(preferences,'Warning file for preference does not exist: ',X).
check_directory_exists(X) :-
  directory_exists(X) -> true ; add_message(preferences,'Warning directory for preference does not exist: ',X).
:- public check_path_exists/1.
check_path_exists(X) :- (file_exists(X) -> true
 ; directory_exists(X) -> true
 ; add_message(preferences,'Warning path for preference does not exist: ',X)).

:- dynamic temp_pref/3.
% use to temporarily set preferences;
% changes can be undone with reset_temporary_preference 
% change will not be saved if reset_temporary_preference not called (e.g., due to bug/timeout/Ctrl-C)
temporary_set_preference(Pref,Val) :- preference(Pref,OldVal),!,
   (OldVal=Val -> true
     ;  set_preference(Pref,Val),
        (retract(temp_pref(Pref,RealOldVal,_))
           -> assert(temp_pref(Pref,RealOldVal,Val)) % store information that new value is Val
           ;  assert(temp_pref(Pref,OldVal,Val))
        ),
        debug_println(9,temp_set_pref(Pref,Val))
    ).
reset_temporary_preference(Pref) :-
    (retract(temp_pref(Pref,OldVal,NewVal))
      -> (preference(Pref,NewVal)
            -> set_preference(Pref,OldVal), debug_println(9,reset_temp_pref(Pref,OldVal))
            ;  true /* user has explicitly changed the preference; then do not revert */
         )
     ; true).

reset_all_temporary_preferences :- temp_pref(Pref,_,_),reset_temporary_preference(Pref),fail.
reset_all_temporary_preferences.


set_preference(Pref,Val) :-
   (preference_val_type(Pref,Type)
     -> (is_of_type(Val,Type)
         -> (invalid_preference_value(Pref,Val,Msg)
              -> print_error('New Preference value is invalid'),
                 print_error(Pref), print_error(Val),
                 print_error(Msg),
                 print_error('No changes performed!'),
                 fail
              ; true
            )
         ;  print_error('New Preference value of incorrect type: '),
            print_error(Pref), print_error(Val),
             %Val =.. L, print_error(val(L)),
             %print_error(expected(Type)),
            print_error('No changes performed!'),
            fail
         )
      ; print_error('Do not know type of preference'), print_error(Pref)
   ),
   retractall(temp_pref(Pref,_,_)), % user has overriden temporary choice
   (retract(preference(Pref,OldVal)) -> New=false ; New=true),
   assert(preference(Pref,Val)),
   ((New=false,OldVal=Val) -> true ; 
                   debug_println(9,update_pref(Pref,OldVal,Val)),
                   assert_change_flag,
                   preference_update_action(Pref,Val)).
   

tcltk_get_preference(Pref,Val) :- /* same as get_preference, but convert non-atomic into atoms */
   get_preference(Pref,V),
   (atomic(V) -> Val=V
              ;  translate_term_into_atom(V,Val)
   ).  

get_preference(Pref,Val) :-
   (preference(Pref,V)
      -> Val=V
      ; (preference_default_value(Pref,V)
         -> Val=V
         ;  print_error('Invalid Preference in get_preference:'),
            print_error(Pref),
            Val = 0
         )
   ).

% rules for computed/derived preferences:
get_computed_preference(debug_time_out,DTimeOut) :-
          preference(time_out,CurTO), 
		  preference(debug_time_out_factor,TFactor), !,
		  DTimeOut is CurTO * TFactor.
get_computed_preference(convert_closures_into_explicit_form_for_store,Value) :-
          preference(convert_comprehension_sets_into_closures,SymbolicV),!,
          (SymbolicV=true -> Value=false ; Value=true).
get_computed_preference(X,_) :- print_error('Invalid Preference in get_computed_preference:'),
            print_error(X).


cur_pref_string(Pref,String) :-
   preference_description(Pref,Desc),
   string_concatenate(Desc,';',String).
 /*  preference(Pref,Val),
   string_concatenate(Val,';',V1),
   string_concatenate(' = ',V1,V2),
   string_concatenate(Desc,V2,String). */
   
   /* useful predicates for Tcl/Tk listbox interaction: */
get_preferences_list(List) :-
   findall(S,cur_pref_string(_,S),List).
get_preferences_list(Category,List) :-
   findall(S,(cur_pref_string(Pref,S),preference_category(Pref,Category)),List).
   
   
get_ith_preference_type(Pos,Type) :- get_ith_preference_type(_,Pos,Type).
get_ith_preference_type(Category,Pos,Type) :-
    findall(S, (preference_description(Pref,_Desc),preference_category(Pref,Category),preference_val_type(Pref,S)), List),
    (nth1(Pos,List,FType) -> Type=FType
                              ; (add_error(get_ith_preference_type,'Could not find preference: ',nth1(Pos,List)),fail)).
    
get_ith_preference_value(Pos,Val) :- get_ith_preference_value(_,Pos,Val).
get_ith_preference_value(Category,Pos,Val) :-
    findall(S, (preference_description(Pref,_Desc),preference_category(Pref,Category),tcltk_get_preference(Pref,S)), List),
    (nth1(Pos,List,FVal) -> Val=FVal
                             ; (add_error(get_ith_preference_value,'Could not find preference: ',nth1(Pos,List)),fail)).
    
set_ith_preference_value(Pos,Val) :- set_ith_preference_value(_,Pos,Val).
set_ith_preference_value(Category,Pos,Val) :-
    findall(S, (preference_description(S,_Desc),preference_category(S,Category)), List),
    (nth1(Pos,List,Pref) -> set_preference(Pref,Val)
                             ;  add_error(set_ith_preference_value,'Could not find preference: ',nth1(Pos,List)),fail).


/* ------------------------- */

:- dynamic nr_of_recent_documents/1.
:- dynamic recent_documents/1.
nr_of_recent_documents(0).

recompute_nr_of_recent_documents :-
  retractall(nr_of_recent_documents(_)),
  findall(X,recent_documents(X),List),
  length(List,Nr),
  assert(nr_of_recent_documents(Nr)).
  
  
:- dynamic add_recent_documents_active/0.
add_recent_documents_active.

activate_recent_documents :- (add_recent_documents_active -> true ; assert(add_recent_documents_active)).

deactivate_recent_documents :- retractall(add_recent_documents_active).

/* use to add a file to recent documents list */
add_recent_document(_) :- \+ add_recent_documents_active, !.
add_recent_document(Document) :-
   (retract(nr_of_recent_documents(Nr)) -> true ; Nr=0),
   (retract(recent_documents(Document))
     -> N1 =Nr ; N1 is Nr+1
   ),
   (preference(number_of_recent_documents,Max) -> true ; Max=10),
   (N1>Max -> (retract(recent_documents(_)),assert(nr_of_recent_documents(Nr)))
           ;   assert(nr_of_recent_documents(N1))),
   assertz(recent_documents(Document)),
   assert_change_flag.
     
get_recent_documents(list(Res)) :-
   findall(X,recent_documents(X),List),
   reverse(List,Res).

clear_recent_documents :-
  retractall(recent_documents(_)),
  retractall(nr_of_recent_documents(_)),
  assert(nr_of_recent_documents(0)).


/* use to add a string to the searched patterns list */
:- dynamic number_of_searched_patterns/1.
:- dynamic number_of_replaced_patterns/1.
:- dynamic number_of_eval_history_elements/1.
:- dynamic number_of_eval_csp_history_elements/1.
:- dynamic number_of_checked_ltl_formulas/1.
:- dynamic searched_patterns/1.
:- dynamic replaced_patterns/1.
:- dynamic eval_history_elements/1.
:- dynamic eval_csp_history_elements/1.
:- dynamic checked_ltl_formulas/1.

number_of_searched_patterns(0).
number_of_replaced_patterns(0).
number_of_eval_history_elements(0).
number_of_eval_csp_history_elements(0).
number_of_checked_ltl_formulas(0).

add_element_to_history(Predicate,Pattern,MaxNumb) :-
   merge_prefix_with_suffix(number_of_,Predicate,Name),
   get_predicate_term_with_single_argument(Name,Nr,TermNum),
   (retract(TermNum) -> true ; Nr=0),
   get_predicate_term_with_single_argument(Predicate,Pattern,Term), 
   (retract(Term) -> N1=Nr ; N1 is Nr+1),
   (preference(TermNum,Max) -> true ; Max=MaxNumb),
   get_predicate_term_with_single_argument(Predicate,_,Term1),    
   (N1>Max -> (retract(Term1),assert(TermNum))
            ;    get_predicate_term_with_single_argument(Name,N1,TermNum1),
                 assert(TermNum1)),assertz(Term),
   assert_change_flag.

get_history_elements_list(Predicate,Reversed,list(Res)) :-
   get_predicate_term_with_single_argument(Predicate,X,Term),
   findall(X,Term,List),
   (Reversed=true -> reverse(List,Res); Res=List).


recompute_nr_of_searched_and_replaced_patterns :-
  retractall(number_of_searched_patterns(_)),retractall(number_of_replaced_patterns(_)),
  findall(X,searched_patterns(X),List),findall(Y,replaced_patterns(Y),List2),
  length(List,Nr),length(List2,Nr2),
  assert(number_of_searched_patterns(Nr)),assert(number_of_replaced_patterns(Nr2)).

recompute_nr_of_eval_history_elements :- 
  retractall(number_of_eval_history_elements(_)),retractall(number_of_eval_csp_history_elements(_)),
  findall(B,eval_history_elements(B),BList),findall(C,eval_csp_history_elements(C),CList),
  length(BList,Nr1),length(CList,Nr2),assert(number_of_eval_history_elements(Nr1)),
  assert(number_of_eval_csp_history_elements(Nr2)).

recompute_nr_of_checked_ltl_formulas :-
  retractall(number_of_checked_ltl_formulas(_)),
  findall(F,checke_ltl_formulas(F),List),
  length(List,Nr),assert(number_of_checked_ltl_formulas(Nr)).

merge_prefix_with_suffix(Prefix,Ending,Name) :- 
   name(Prefix,Pr),name(Ending,End),
   append(Pr,End,CodeName),name(Name,CodeName).

get_predicate_term_with_single_argument(Name,Arg,Term) :- 
   functor(Term,Name,1),arg(1,Term,Arg).



/* ------------------------- */

set_preference_group(Group,_) :- \+ preference_group_description(Group,_),!,
   add_error(set_preference_mode,'Preference mode does not exist: ',Group).
set_preference_group(Group,default) :- preference_group_val(Group,Scheme,_,_),!,
   ( preference_group_val(Group,Scheme,Pref,_),preference_default_value(Pref,Val),
     set_preference(Pref,Val),fail
     ; true).
set_preference_group(Group,Scheme) :- preference_group_val(Group,Scheme,Pref,Val),
   set_preference(Pref,Val),fail.
set_preference_group(_,_).


preference_group_description(integer,'Values for MAXINT and MININT').
preference_group_description(dot_colors,'Colours for Dot graphs').

preference_group_val(integer,int32,maxint,2147483647). % 2147483647
preference_group_val(integer,int32,minint,-2147483648). %-2147483648

preference_group_val(dot_colors,classic,dot_normal_node_colour, olivedrab2).
preference_group_val(dot_colors,classic,dot_normal_arc_colour,steelblue).
preference_group_val(dot_colors,classic,dot_open_node_colour,red).
preference_group_val(dot_colors,classic,dot_goal_node_colour,orange).
preference_group_val(dot_colors,classic,dot_invariant_violated_node_colour,maroon2).

preference_group_val(dot_colors,winter,dot_normal_node_colour, '#425955').
preference_group_val(dot_colors,winter,dot_normal_arc_colour,'#1B1D26').
preference_group_val(dot_colors,winter,dot_open_node_colour,'#BFBD9F').
preference_group_val(dot_colors,winter,dot_goal_node_colour,'#CC8604').
preference_group_val(dot_colors,winter,dot_invariant_violated_node_colour,'#9C080C').

preference_group_val(dot_colors,dreams,dot_normal_node_colour, '#5D7359').
preference_group_val(dot_colors,dreams,dot_invariant_violated_node_colour,'#661C0E').
preference_group_val(dot_colors,dreams,dot_normal_arc_colour,'#8C5430').
preference_group_val(dot_colors,dreams,dot_open_node_colour,'#D6AA5C').
preference_group_val(dot_colors,dreams,dot_goal_node_colour,'#FFC800').

