% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(record_detection,[replace_sets_by_records/2]).


:- use_module(tools).
:- use_module(self_check).

:- use_module(module_information,[module_info/2]).
:- module_info(group,typechecker).
:- module_info(description,'This module detects bijections between deferred sets and cartesian products, and compiles them away.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').


:- use_module(library(lists)).

:- use_module(bmachine_structure).
:- use_module(bsyntaxtree).
:- use_module(preferences,[get_preference/2]).
:- use_module(b_ast_cleanup,[clean_up/3]).

replace_sets_by_records(Machine,ResultMachine) :-
    % replace until a fixpoint is reached
    replace_sets_by_records2(Machine,RMachine),!,
    replace_sets_by_records(RMachine,ResultMachine).
replace_sets_by_records(Machine,Machine).

replace_sets_by_records2(Machine,ResultMachine) :-
    get_preference(use_record_construction,true),
    has_deferred_sets(Machine),
    get_section(deferred_sets,Machine,DeferredSets),
    select_section(properties,Properties,NewProperties,Machine,ConsMachine),
    select_constructor_axiom(Properties,DeferredSets,Set,Constructor,Domain,RestList1),
    % now we now that we have detected a record
    replace_sets_by_records3(NewProperties,ConsMachine,Set,Constructor,Domain,RestList1,ResultMachine).

:- use_module(error_manager,[add_internal_error/2]).
replace_sets_by_records3(NewProperties,ConsMachine,Set,Constructor,Domain,RestList1,ResultMachine) :-
    create_constructor_definition(Constructor,Domain,ConsDef),
    create_recordset_definition(Set,Domain,SetDef),
    create_optional_field_access(Set,Domain,Constructor,RestList1,RestList),
    conjunct_predicates([ConsDef,SetDef|RestList],NewProperties),
    move_deferred_set(Set,ConsMachine,SetMachine),
    get_texpr_type(Constructor,set(couple(Type,_))),
    replace_type_in_machine(Set,Type,SetMachine,ResultMachine),
    !.
replace_sets_by_records3(NewProperties,ConsMachine,Set,Constructor,Domain,RestList1,ResultMachine) :-
    add_internal_error('Replacing record failed',replace_sets_by_records3(NewProperties,ConsMachine,Set,Constructor,Domain,RestList1,ResultMachine)),fail.

has_deferred_sets(Machine) :-
    get_section(deferred_sets,Machine,DefSets),
    DefSets = [_|_].

select_constructor_axiom(Properties,Sets,Set,Constructor,Domain,RestList) :-
    conjunction_to_list(Properties,PList),
    select(Prop,PList,RestList),
    is_constructor_axiom(Prop,Sets,Set,Constructor,Domain).

create_constructor_definition(Constructor,Domain,ConsDef) :-
    create_texpr(equal(Constructor,Identity),pred,[],ConsDef),
    get_texpr_type(Constructor,Type),
    create_texpr(identity(Domain),Type,[],Identity).

create_recordset_definition(SetName,Domain,SetDef) :-
    create_texpr(equal(Set,Domain),pred,[],SetDef),
    get_texpr_type(Domain,Type),
    create_texpr(identifier(SetName),Type,[],Set).

move_deferred_set(Set,OldMachine,NewMachine) :-
    select_section(deferred_sets,Sets,NewSets,OldMachine,Machine1),
    get_texpr_id(TSet,Set),
    selectchk(TSet,Sets,NewSets),
    select_section(concrete_constants,Constants,[TSet|Constants],Machine1,NewMachine).

is_constructor_axiom(TAxiom,Sets,GType,TConstructor,TDom) :-
    % The axiom has the form c : Dom >->> Set, where
    % c and Set are identifiers
    get_texpr_expr(TAxiom,member(TConstructor,TBijection)),
    get_texpr_expr(TBijection,total_bijection(TDom,TSet)),
    get_texpr_type(TSet,set(global(GType))),
    get_texpr_id(TSet,GType),
    % check if the set is a deferred set, not an enumerated set
    get_texpr_id(SetTest,GType),memberchk(SetTest,Sets),
    % in Dom, there should be no reference to Set
    no_reference(TDom,GType).

no_reference(TExpr,Type) :-
    syntaxtraversion(TExpr,Expr,_,_,Subs,_),
    no_reference2(Expr,Type),
    no_reference_l(Subs,Type).
no_reference2(identifier(Type),Type) :- !,fail.
no_reference2(_,_).
no_reference_l([],_).
no_reference_l([Expr|Rest],Type) :-
    no_reference(Expr,Type),
    no_reference_l(Rest,Type).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
create_optional_field_access(SetName,Domain,Constructor,Properties,RestList) :-
    get_texpr_id(Set,SetName),
    find_and_remove_field_declaration(Set,Domain,Constructor,Fields,Properties,P4),
    !,
    replace_update_definitions(Fields,Set,Constructor,P4,P5),
    create_field_definitions(Fields,Domain,FieldDefs),
    append(FieldDefs,P5,RestList).
create_optional_field_access(_Set,_Domain,_Constructor,Properties,Properties).

find_and_remove_field_declaration(Set,Domain,Constructor,Fields,Pin,Pout) :-
    % F1 >< ... >< Fn : Set -->> Domain
    bsyntax_pattern(member(FieldProduct,total_surjection(-Set,-Domain)), SurjMemb),
    selectchk(SurjMemb,Pin,P1),
    % the injection is only defined for closed records
    % F1 >< ... >< Fn : Set >-> Domain
    bsyntax_pattern(member(-FieldProduct,total_injection(-Set,-Domain)), InjMemb),
    selectchk(InjMemb,P1,P2),
    access_locations(FieldProduct,Domain,Fields),
    % for each field accessor fx, there should be a declaration
    % fx : Record --> Domain, remove them
    remove_field_declarations(Fields,Set,P2,P3),
    % for each field accessor fx, there should be a predicate
    % !(v1,...,...), remove it
    remove_field_quantifiers(Fields,Constructor,P3,Pout).

% find the field accessor functions
access_locations(TExpr,TDomain,Fields) :-
    access_locations1(TExpr,TDomain,[],Fields).
access_locations1(TExpr,TDomain,Path,Fields) :-
    get_texpr_expr(TExpr,Expr),
    access_locations2(Expr,TExpr,TDomain,Path,Fields).
access_locations2(identifier(_Id),TExpr,TDomain,Path,[fieldloc(TExpr,TDomain,RPath)]) :- rev(Path,RPath).
access_locations2(direct_product(F1,F2),_TExpr,TCart,Path,Fields) :-
    get_texpr_expr(TCart,cartesian_product(D1,D2)),
    access_locations1(F1,D1,[left |Path],Fields1),
    access_locations1(F2,D2,[right|Path],Fields2),
    append(Fields1,Fields2,Fields).

create_field_definitions([],_RecordDomain,[]).
create_field_definitions([fieldloc(AccessorId,Domain,Path)|FRest],RecordDomain,[Equal|ERest]) :-
    create_field_definition(AccessorId,Domain,Path,RecordDomain,Equal),
    create_field_definitions(FRest,RecordDomain,ERest).
create_field_definition(AccessorId,Domain,Path,RecordDomain,Equal) :-
    create_texpr(equal(AccessorId,AccessorDef),pred,[],Equal),
    % the accessor function maps a record to the domain
    get_texpr_type(RecordDomain,set(RecordTupleType)),
    create_accessor_function(Domain,Path,RecordTupleType,AccessorDef).
create_accessor_function(Domain,Path,RecordTupleType,AccessDefinition) :-
    % fieldAccess = {r,v | v = prj...(r)}
    create_texpr(identifier(r),RecordTupleType,[],RecordVar),
    get_texpr_type(Domain,set(FieldType)),
    create_texpr(identifier(v),FieldType,[],ValueVar),
    create_texpr(equal(ValueVar,Projection),pred,[],Equals),
    create_projection(Path,RecordVar,_,Projection),
    AccessorType = set(couple(RecordTupleType,FieldType)),
    create_texpr(comprehension_set([RecordVar,ValueVar],Equals),AccessorType,[],AccessDefinition).
create_projection([],RecordVar,InnerType,RecordVar) :- get_texpr_type(RecordVar,InnerType).
create_projection([Loc|Path],RecordVar,Type,Projection) :-
    ( Loc==left  -> Expr=first_of_pair(Rest), Type=A
    ; Loc==right -> Expr=second_of_pair(Rest),Type=B),
    create_texpr(Expr,Type,[],Projection),
    create_projection(Path,RecordVar,couple(A,B),Rest).


% remove fieldx : RecSet --> Domx
remove_field_declarations([],_RecSet,Predicates,Predicates).
remove_field_declarations([fieldloc(AccessorId,Domain,_Path)|Frest],RecSet,Pin,Pout) :-
    remove_field_declaration(AccessorId,Domain,RecSet,Pin,Pinter),
    remove_field_declarations(Frest,RecSet,Pinter,Pout).
remove_field_declaration(AccessorId,Domain,RecSet,Pin,Pout) :-
    bsyntax_pattern(member(-AccessorId,total_function(-RecSet,-Domain)),Pattern),
    selectchk(Pattern,Pin,Pout).

% remove !(f1,...,fn). ( f1:T1 & ... & fn:Tn => fieldx(make_rec(f1|->f2|->f3)) = fx)
remove_field_quantifiers(Fields,Constructor,Pin,Pout) :-
    remove_field_quantifiers2(Fields,0,Constructor,Pin,Pout).
    
remove_field_quantifiers2(Fields,N,Constructor,Pin,Pout) :-
    ( nth0(N,Fields,fieldloc(Field,_Dom,_Path)) ->
        remove_field_quantifier(Field,N,Fields,Constructor,Pin,Pinter),
        N2 is N+1,
        remove_field_quantifiers2(Fields,N2,Constructor,Pinter,Pout)
    ; otherwise ->
        Pin=Pout).
remove_field_quantifier(Field,N,Fields,Constructor,Pin,Pout) :-
    make_field_identifiers(Fields,FSkels),
    get_texpr_ids(FSkels,Ids),
    nth0(N,FSkels,Ref),
    bsyntax_pattern(forall(FSkels,_,equal(function(-Field,function(-Constructor,ConstArg)),-Ref)),ForAll),
    select(ForAll,Pin,Pout),
    are_mappings(ConstArg,Ids),!.
are_mappings(TExpr,Ids) :-
    are_mappings2(TExpr,Ids,[]).
are_mappings2(TExpr) -->
    {get_texpr_expr(TExpr,Expr)},
    are_mappings3(Expr).
are_mappings3(identifier(Id)) --> [Id].
are_mappings3(couple(A,B)) -->
    are_mappings2(A),
    are_mappings2(B).

make_field_identifiers([],[]).
make_field_identifiers([FieldAcc|FRest],[Id|Irest]) :-
    make_field_identifier(FieldAcc,Id),
    make_field_identifiers(FRest,Irest).
make_field_identifier(fieldloc(FieldAcc,_,_),Id) :-
    get_texpr_type(FieldAcc,set(couple(_,FType))),
    create_texpr(identifier(_),FType,_,Id).

replace_update_definitions(Fields,RecSet,Constructor,Pin,Pout) :-
    replace_update_definitions2(Fields,Fields,RecSet,Constructor,Pin,Pout).
replace_update_definitions2([],_Fields,_RecSet,_Constructor,Predicates,Predicates).
replace_update_definitions2([fieldloc(AccessorId,FDomain,Path)|Frest],Fields,RecSet,
                            Constructor,Pin,Pout) :-
    replace_update_definition(AccessorId,FDomain,Path,Fields,RecSet,Constructor,Pin,Pinter),
    replace_update_definitions2(Frest,Fields,RecSet,Constructor,Pinter,Pout).

% Identify two predicates of the form
%   updatex : Record ** Domain --> Record
%   !(Rec,New).(_ => updatex(Rec,New) = Constructor(..))
% and replace them by
%   updatex = {i,o | #(n,f1,...,fn).(i=(f1,...,fn)|->n & o=(f1,...,n,..,fn))}
replace_update_definition(FieldAccessorId,FieldDomain,Path,Fields,
                          RecSet,Constructor,Pin,[Equal|Pout]) :-
    % Updatex : Record ** Domain --> Record
    bsyntax_pattern(member(Update,total_function(cartesian_product(-RecSet,-FieldDomain),-RecSet)),Memb),
    select(Memb,Pin,P1),
    % !(Rec,New).(_ => Updatex(Rec,New) = Constructor(..))
    % TODO: Implication and constructor arguments are missing
    bsyntax_pattern(forall([Rec,New],_,equal(function(-Update,couple(-Rec,-New)),
                                             function(-Constructor,_))), ForAll),
    select(ForAll,P1,Pout),
    !,
    % Updatex = {i,o | #(i,r,f1,...,fn).(i=r|->n & ) }
    create_texpr(equal(Update,Function),pred,[],Equal),
    create_update_function(FieldAccessorId,FieldDomain,Path,Fields,Function).

create_update_function(FieldAccessorId,FieldDomain,Path,Fields,Function) :-
    % {i,o | #n,f1,...,fn. ( i=( (f1,...,fn) |-> n ) ) & o=(f1,...,n,...,fn) } 
    create_texpr(comprehension_set([In,Out],Pred),set(couple(InType,RecType)),[],CompSet),
    get_texpr_type(FieldAccessorId,set(couple(RecType,FieldType))),
    InType = couple(RecType,FieldType),
    create_texpr(identifier(update__in),InType,[],In),
    create_texpr(identifier(update__out),RecType,[],Out),

    % #(n,f1,...,fn).(...&...)
    create_exists([NewValue|FieldVars],ExistsPred,Pred),
    create_var_for_field(FieldAccessorId,FieldDomain,'fn__',NewValue),
    maplist(create_field_var,Fields,FieldVars),
    create_texpr(conjunct(InEqual,OutEqual),pred,[],ExistsPred),
    
    % i = ( (f1,...,fn) |-> n )
    create_texpr(equal(In,InPair),pred,[],InEqual),
    create_texpr(couple(OldFields,NewValue),InType,[],InPair),
    create_record_tuple(Fields,FieldVars,OldFields),

    % o = (f1,...,n,...,fn)
    create_texpr(equal(Out,NewFields),pred,[],OutEqual),
    update_record_tuple(Path,NewValue,OldFields,NewFields),

    % we need to call the clean-up functions because exists(...) needs more
    % information about which identifier are used and that information
    % is added by clean_up/3.
    clean_up(CompSet,[],Function).

% for a fields, create an identifier with the corresponding type
create_field_var(fieldloc(AccessorId,Domain,_Path),TId) :-
    create_var_for_field(AccessorId,Domain,'f__',TId).
create_var_for_field(AccessorId,Domain,Prefix,TId) :-
    get_texpr_id(AccessorId,FId),
    get_texpr_type(Domain,set(Type)),
    atom_concat(Prefix,FId,VId),
    create_texpr(identifier(VId),Type,[],TId).

% for a list of fields and identifiers f1,...,fn, create a tuple (f1,...,fn)
create_record_tuple([],[],_).
create_record_tuple([fieldloc(_,_,Path)|Frest],[FieldVar|Vrest],Tuple) :-
    create_record_tuple2(Path,FieldVar,Tuple),
    create_record_tuple(Frest,Vrest,Tuple).
create_record_tuple2([],Expr,Expr).
create_record_tuple2([Loc|Path],Expr,Tuple) :-
    ( Loc==left  -> create_texpr(couple(Sub,_),couple(SubType,_),[],Tuple)
    ; Loc==right -> create_texpr(couple(_,Sub),couple(_,SubType),[],Tuple)),
    get_texpr_type(Sub,SubType),
    create_record_tuple2(Path,Expr,Sub).

% in a tuple representing a record, replace one field
update_record_tuple([],NewValue,_OldFields,NewValue).
update_record_tuple([Loc|Path],NewValue,OldFields,NewFields) :-
    get_texpr_expr(OldFields,couple(A,B)),
    get_texpr_type(OldFields,Type),
    create_texpr(couple(X,Y),Type,[],NewFields),
    ( Loc==left  -> X=NewSub,B=OldSub,Y=B
    ; Loc==right -> Y=NewSub,A=OldSub,X=A),
    update_record_tuple(Path,NewValue,OldSub,NewSub).


    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
replace_type_in_machine(T1,T2) -->
    replace_type_in_machine2([abstract_constants,concrete_constants,
                   abstract_variables,concrete_variables,
                   promoted, unpromoted,
                   constraints, properties, invariant, assertions,
                   initialisation, operation_bodies],
                  T1,T2).
replace_type_in_machine2([], _, _) --> !.
replace_type_in_machine2([Sec|Srest], T1, T2) -->
    replace_type_in_machine3(Sec,T1,T2),
    replace_type_in_machine2(Srest,T1,T2).
replace_type_in_machine3(Sec,T1,T2,In,Out) :-
    select_section_texprs(Sec,Old,New,In,Out),
    replace_type_in_exprs(Old,T1,T2,New).
replace_type_in_expr(TExpr,T1,T2,New) :-
    get_texpr_expr(TExpr,Expr),
    get_texpr_type(TExpr,EType),
    get_texpr_info(TExpr,Info),
    replace_type(EType,T1,T2,NType),
    create_texpr(NExpr,NType,Info,New),
    syntaxtransformation(Expr,Subs,_Names,NSubs,NExpr),
    replace_type_in_exprs(Subs,T1,T2,NSubs).
replace_type_in_exprs([],_T1,_T2,[]).
replace_type_in_exprs([Sub|Irest],T1,T2,[NSub|Orest]) :-
    replace_type_in_expr(Sub,T1,T2,NSub),
    replace_type_in_exprs(Irest,T1,T2,Orest).

replace_type(global(Type),Type,NewType,Result) :- !, NewType=Result.
replace_type(couple(A,B),T1,T2,couple(RA,RB)) :-
    !, replace_type(A,T1,T2,RA), replace_type(B,T1,T2,RB).
replace_type(set(A),T1,T2,set(RA)) :-
    !, replace_type(A,T1,T2,RA).
replace_type(seq(A),T1,T2,seq(RA)) :-
    !, replace_type(A,T1,T2,RA).
replace_type(T,_,_,T).