% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(sap,[explore_and_generate_testcases/7,
               create_testcase_path/5, create_testcase_path/3,
               cbc_gen_test_cases/5,
               test_generation_by_xml_description/1,
               tcl_create_local_testsuite/5,
               tcl_get_events_preselected/2,
               tcl_generate_cb_testcases/5,
               tcl_generate_testcases/6
              ]).

:- use_module(library(lists)).
:- use_module(library(xml)).
:- use_module(library(timeout)).

:- use_module(error_manager).
:- use_module(state_space).
:- use_module(bmachine).
:- use_module(b_enumerate,[b_tighter_enumerate_values/2]).
:- use_module(b_state_model_check,[set_up_transition/7,set_up_initialisation/5]).
:- use_module(preferences,[set_preference/2,get_preference/2]).
:- use_module(specfile,[expand_const_and_vars_to_full_store/2,expand_to_constants_and_variables/3]).
:- use_module(translate,[translate_bvalue/2]).
:- use_module(eclipse_interface,[set_eclipse_preference/2]).
:- use_module(kernel_objects,[not_equal_object/2]).

:- use_module(bsyntaxtree).
:- use_module(tools).

:- use_module(extension('user_signal/user_signal'), [user_interruptable_call_det/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,test_generation).
:- module_info(description,'This module contains code to generate test cases based on Event-B or Classical B models').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

% This is quite a hack, TODO: move the init_states/1 predicate to
% another place
:- use_module(ltl).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% some code for information in the GUI

% get_events_preselected(-Events,-Preselect):
%   Events: a list of all events
%   Preselect: a list of numbers, each an index (starting with 0) of which event in Events should
%     be preselected because it is already covered in the state space
% Both lists are wrapped by list(...) to use it from Tcl
tcl_get_events_preselected(list(Events),list(Preselect)) :-
    get_all_events(Events),
    select_covered_operations(Events,0,Preselect).
select_covered_operations([],_,[]).
select_covered_operations([Event|Erest],N,Preselect) :-
    ( operation_name_not_yet_covered(Event) ->
        Preselect = PSrest
    ; otherwise ->
        Preselect = [N|PSrest]),
    N2 is N+1,
    select_covered_operations(Erest,N2,PSrest).

% returns a list of all events (their namese) in the model
get_all_events(Events) :-
    findall(E,b_top_level_operation(E),Events).

% tcl_generate_testcases(+EventSelection,+TargetString,+MaxDepth,+MaxNewNodes,+Testfile,-NumberOfTestcases):
%  EventSelection: a list of numbers containing all indeces of events that should be covered
%  TargetString: the predicate that identifies target nodes as string (atom)
%  MaxDepth: the maximum depth of the search
%  MaxNewNodes: the maximum number of newly explored nodes
%  Testfile: the file where we store the found test cases
%  NumberOfTestcases: The number of test cases that where found
tcl_generate_testcases(EventSelection,TargetString,MaxDepth,MaxNewNodes,Testfile,NumberOfTestcases) :-
    % convert the list of event indeces into a list of events
    get_selected_events(EventSelection,AllEvents,Events),
    b_parse_machine_predicate(TargetString,Target),
    set_max_nr_of_new_impl_trans_nodes(MaxNewNodes),
    safe_get_inits(Inits),
    ( explore_covered_statespace(Inits, Events, Target, MaxDepth, _ExploredDepth) ->
        true
    ; otherwise ->
        show_uncovered_events_as_errors),
    generate_test_cases(Inits,Testcases),
    length(Testcases,NumberOfTestcases),
    testcase_output(Testfile,AllEvents,Events,Testcases).

% get_selected_events(+TkNumbers,-ListOfAllEvents,-SelectedEvents)
%    Input: TkNumbers: numbers starting at 0; comes from Tk Selection Box
%    Output: ListOfAllEvents
%    Output: SelectedEvents: Name of Selected Events
get_selected_events(Selection,AllEvents,Events) :-
    get_all_events(AllEvents),
    get_selected_events2(Selection,AllEvents,Events).
get_selected_events2([],_AllEvents,[]).
get_selected_events2([Selection|Srest],AllEvents,[Event|Erest]) :-
    nth0(Selection,AllEvents,Event),!,
    get_selected_events2(Srest,AllEvents,Erest).

safe_get_inits(Inits) :-
    ltl:init_states(Inits),
    ( Inits = [] -> add_error(sap,'no initial states found'),fail
    ; otherwise -> true).

% this predicate is called if the breadth-first search fails
% it shows an error for every event, for which we did not find
% an appropiate path
show_uncovered_events_as_errors :-
    target_needed(Event),
    ajoin(['No test path found for event ',Event],Msg),
    add_error(sap,Msg),
    fail.
show_uncovered_events_as_errors.

get_uncovered_events(Uncovered) :-
    findall(Event, target_needed(Event), Uncovered).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

explore_and_generate_testcases(InEvents,Target,MaxDepth,MaxNewNodes,
                               Testfile,NumberOfTestcases,UncoveredEvents) :-
    % If no event is specified to be covered, we assume that all events
    % should be covered
    ( InEvents = []    ->
        write('No events to cover specified, assuming all'),nl,
        get_all_events(Events)
    ; InEvents = [_|_] -> InEvents = Events),
    statistics(walltime, [Start|_]),
    set_max_nr_of_new_impl_trans_nodes(MaxNewNodes),
    safe_get_inits(Inits),
    ( explore_covered_statespace(Inits, Events, Target, MaxDepth, _ExploredDepth) ->
        UncoveredEvents = []
    ; otherwise ->
        get_uncovered_events(UncoveredEvents)),
    generate_test_cases(Inits, Testcases),
    length(Testcases, NumberOfTestcases),
    get_all_events(AllEvents),
    testcase_output(Testfile,AllEvents,Events,Testcases),
    statistics(walltime, [End|_]),
    Runtime is End - Start,
    format('Runtime for test case generation: ~w ms~n',[Runtime]).
    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate test cases from an explored state-space


% generate_test_cases(+Inits,-TestCases):
%  Inits: List of initial states
%  TestCases: list of test cases, each of the form global_test_case(Trace),
%    where Trace is a list of events
generate_test_cases(Inits,TestCases) :-
    ( setof(T, generate_test_case(Inits,T), TestCases) ->
        true
    ; otherwise ->
        TestCases = []).
% find a single test case by first choosing one event and one initial state
generate_test_case(Inits,global_test_case(Trace)) :-
    fail, % SAP wants all test cases, not just for each event the shortest,
    % so ignore this old code
    minimum_path_length(Event,MaxSteps),
    member(Init,Inits),
    % now find a path that can serve as a test case
    test_case(Event,Init,MaxSteps,Trace).
generate_test_case(Inits,global_test_case(Trace)) :-
    find_global_maximum(MaxSteps),
    member(Init,Inits),
    find_way_to_target(Init,MaxSteps,Trace),
    contains_an_event_to_test(Trace).

% find the maximum of the minimum mapth length
find_global_maximum(MaxSteps) :-
    findall(M, minimum_path_length(_Event,M), [First|Maxis]),
    find_global_maximum2(Maxis,First,MaxSteps).
find_global_maximum2([],Max,Max).
find_global_maximum2([H|T],A,Max) :- B is max(H,A), find_global_maximum2(T,B,Max).

contains_an_event_to_test(Trace) :-
    member(Event,Trace),wanted_event(Event),!.
    

% find a single test case for a single event and an initial state
test_case(Event,Init,MaxSteps,Trace) :-
    % first find a path where Event occurs
    find_way_to_event(Init,MaxSteps,Event,Trace,TraceB,State,RestSteps),
    % then find a path from that state to a target state
    find_way_to_target(State,RestSteps,TraceB).

% find_way_to_event(+A,+Max,+Event,-Trace,TraceR,-B,-RestSteps):
%   A: node where we start the search
%   Max: maximum number of steps in the path
%   Event: the event we are searching for
%   Trace: a trace to the state, the last event is Event, the list
%     has the variable TraceR as tail
%   B: the state after Event occurs
%   RestSteps: the number of remaining steps (i.e. RestSteps=Max-length(Trace))
find_way_to_event(A,Max,Event,[Operation|Trace],TraceR,B,RestSteps) :-
    dec(Max,Max2),
    do_event(A,Operation,C),
    ( Event = Operation ->
        Trace = TraceR,
        B = C,
        RestSteps = Max2
    ; otherwise ->
        find_way_to_event(C,Max2,Event,Trace,TraceR,B,RestSteps)).

find_way_to_target(A,_Max,[]) :-
    target_node(A),!.
find_way_to_target(A,Max,[Operation|Trace]) :-
    dec(Max,Max2),
    do_event(A,Operation,B),
    find_way_to_target(B,Max2,Trace).

% decrease N, but never become negative
dec(N,N2) :- N>0, N2 is N-1.

do_event(A,Event,B) :-
    transition(A,FullOperation,B),
    functor(FullOperation,Event,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% explore the state-space so that it is usable for test-case generation

% bfs_node(NodeId,Depth): open nodes in the state space, for breadth first search,
%   the order is relevant, Depth is the distance from the initialisation nodes
%   (later we call this the breadt-first FIFO queue)
% predecessor(NodeId,PreNodeId): store predecessors of a node
% target_reachable(NodeId,TargetDist): A target state is reachable from this state
% event_dest(NodeId,Event): One of the events we want to cover leads to this node
% node_covered(NodeId,Depth): This node was encountered before by the search at depth Depth.
% target_needed(Event): We need to find a final from a state that is the destination of the event,
%  in the beginning, each given event is stored here
% wanted_event(Event): Event if one of the events we are interested in
%  in the beginning, each given event is stored here
% target_node(NodeId): The node is a target node

:- volatile bfs_node/2, predecessor/2, target_reachable/2, event_dest/2.
:- volatile node_covered/2, target_needed/1, wanted_event/1.
:- volatile target_node/1, minimum_path_length/2.
:- dynamic bfs_node/2, predecessor/2, target_reachable/2, event_dest/2.
:- dynamic node_covered/2, target_needed/1, wanted_event/1.
:- dynamic target_node/1, minimum_path_length/2.

% explore_covered_statespace(+Inits,+Events,+Target,-ExploredDepth):
%  Inits: a list of initialisation states
%  Events: the list of events that should be covered
%  Target: the predicate that is used to identify a target state
%  MaxDepth: the maximum depth that will be explored
%  ExploredDepth: the depth until the state-space was explored
% This predicate explores a part of the state-space by breadt-first search
% that contains for each event a path where the event occurs and that ends
% in a target state.
% Additionally, the state-space is explored to a fixed depth.
% (i.e. if we needed a state that is reachable by 4 steps, than all states
%  that are reachable by 4 steps are explored with their transitions and
%  successor states)
explore_covered_statespace(Inits,Events,Target,MaxDepth,ExploredDepth) :-
    % initialise the data structures
    reset_bfs(Events),

    % put initial states into the breadt-first FIFO queue
    assert_bfs_nodes(Inits,0,_,_),
    % check if the initialised states are target states
    mark_target_nodes(Inits,Target),

    % do the breadt-first search
    explore(Target,MaxDepth,-1,ExploredDepth).

% explore(+Target,+PrevExploredDepth,-ExploredDepth):
%  Target: the predicate that is used to identify target nodes
%  MaxDepth: the maximum depth that will be explored
%  PrevExploredDepth: the deepest node that is explored before calling this predicate
%  ExploredDepth: the deepest node that was explored by the search
explore(Target,_MaxDepth,ExploredDepth,ExploredDepth) :-
    % we do not have to find any more targets, we're finished
    \+ target_needed(_),!,
    % we did a breadth-first search until a certain depth, now we
    % complete the exploration of the state space to that depth
    complete_depth(Target,ExploredDepth).
explore(Target,MaxDepth,_,ExploredDepth) :-
    % get a node from the breadth-first FIFO queue
    retract( bfs_node(Node,Depth) ),!,
    % and explore it
    explore_node(Node, Target, Depth),
    !, % just to be safe that no back-tracking occurs
    Depth =< MaxDepth,
    explore(Target,MaxDepth,Depth,ExploredDepth).

% explore_node(+Node,+Target,+Depth):
%  Node: the node that should be explored
%  Target: the predicate that is used to identify target nodes
%  Depth: depth of the given node
explore_node(Node,Target,Depth) :-
    % newly encountered nodes are one level deeper
    NewDepth is Depth+1,
    % get all transitions from this node to successor nodes
    impl_trans_term_all(Node,Transitions),
    % check if there are events we are interested in and
    % get the successor nodes
    analyse_transitions(Transitions,Successors),
    % find out wich of the successor nodes are new and
    % store those in the breadt-first FIFO queue
    assert_bfs_nodes(Successors,NewDepth,NewNodes,OldNodes),
    % store that this node is a predecessor node of
    % all the successor nodes encountered
    assert_predecessors(Successors, Node),
    % find out which of the new nodes are target nodes
    % and in such a case propagate the information that a target
    % note is reachable back into the network.
    mark_target_nodes(NewNodes,Target),
    mark_target_nodes_by_old(OldNodes,Node).

mark_target_nodes_by_old([],_Node).
mark_target_nodes_by_old([PotentialTargetNode|Rest],Node) :-
    ( target_reachable(PotentialTargetNode,DistToTarget) ->
        NewDist is DistToTarget + 1,
        update_minimal_path_for_events(Node,PotentialTargetNode,DistToTarget),
        assert_target_reachable(Node,NewDist)
    ; otherwise ->
        true ),
    mark_target_nodes_by_old(Rest,Node).
 

% mark_target_nodes(+Nodes,+Target).
%  Nodes: a list of node IDs
%  Target: the predicate to check whether the node is a target node
%  PreDepth: the depth of the node that led to this node, 
% checks for each node in the given list if it is a target
% node, and if yes then propagate the information that the
% target is reachable to all predecessor nodes.
:- use_module(eclipse_interface,[test_boolean_expression_in_node/2]).
mark_target_nodes([],_Target).
mark_target_nodes([Node|NRest],Target) :-
    ( test_boolean_expression_in_node(Node,Target) ->
        assert( target_node(Node) ),
        assert_target_reachable(Node,1)
    ; otherwise -> true ),
    mark_target_nodes(NRest,Target).

% for all transitions, check if the resp. event is interesting
% and if yes, mark the successor node.
% and return the list of successor nodes.
analyse_transitions([],[]).
analyse_transitions([op(_Id,Operation,Succ)|Orest],[Succ|Srest]) :-
    % the arguments of the event are not interesting to us
    functor(Operation,Event,_),
    % if Event is one of the events we monitor, we mark
    % the destination node as a node from where we look for
    % a path to a target node.
    ( wanted_event(Event) ->
        assert_event_dest(Succ,Event)
    ; otherwise -> true),
    analyse_transitions(Orest,Srest).

% store the information, that the Node is reached by Event.
% We need this, because we need such a node fro
assert_event_dest(Node,Event) :-
    ( event_dest(Node,Event) -> true   % information was already present, ignore
    ; otherwise ->
        assert( event_dest(Node,Event) ),
        % if a target node can be reached from Node, we have a path in the explored
        % state-space that includes the Event and ends in a target node.
        ( target_reachable(Node,_) ->
            ( retract( target_needed(Event) ) -> true; true)
        ; otherwise -> true) ).

% reset all data structures needed for the breadth-first search
reset_bfs(Events) :-
    retractall(bfs_node(_,_)),
    retractall(predecessor(_,_)),
    retractall(target_reachable(_,_)),
    retractall(node_covered(_,_)),
    retractall(event_dest(_,_)),
    retractall(target_needed(_)),
    retractall(wanted_event(_)),
    retractall(target_node(_)),
    retractall(minimum_path_length(_,_)),
    reset_bfs2(Events).
reset_bfs2([]).
reset_bfs2([Event|Rest]) :-
    assert( target_needed(Event) ),
    assert( wanted_event(Event) ),
    reset_bfs2(Rest).

% assert_bfs_nodes(+Nodes,+Depth,-NewNodes):
%  Nodes: a list of nodes
%  Depth: the current depth of the search, this is stored for finding
%    out later until which depth the search has been done
%  NewNodes: a sublist of Nodes, that includes all nodes that have not been
%    encountered before
% This predicate checks if the given nodes are already encountered and if
%  not, it stores them into the breadt-first FIFO queue
assert_bfs_nodes([],_,[],[]).
assert_bfs_nodes([Node|Rest],Depth,NewNodes,OldNodes) :-
    ( node_covered(Node,_NodeDepth) -> % we already encountered Node, ignore
        NewNodes = NewRest,
        OldNodes = [Node|OldRest]
    ; otherwise -> % this is a new node
        NewNodes = [Node|NewRest],
        OldNodes = OldRest,
        assert( node_covered(Node,Depth) ), % mark as covered
        assertz( bfs_node(Node,Depth) )),   % put it into the FIFO queue for the breadth-first search
    assert_bfs_nodes(Rest,Depth,NewRest,OldRest).

% assert_predecessors(+Nodes,+Predecessor):
%  Nodes: a list of nodes
%  Predecessor: the predecessor node, a node from where a transition
%    leads to each of the nodes in Nodes
% This predicate stores for each node in Nodes that Predecessor is a
% predecessor of the node.
% This information is needed to be able to propagate information of
% a reachable target state back into the graph.
assert_predecessors([],_).
assert_predecessors([Node|Rest],Pre) :-
    ( predecessor(Node,Pre) -> true % information already there, ignore
    ; otherwise ->
        assert( predecessor(Node,Pre) )),
    assert_predecessors(Rest,Pre).

% store the information that a target node is reachable from a node
% propagate this information to predecessor nodes if necessary
assert_target_reachable(Node,DistToTarget) :-
    ( target_reachable(Node,OldDistToTarget) ->
        ( OldDistToTarget =< DistToTarget ->
            true                % information already there, ignore
        ; otherwise ->
            retract( target_reachable(Node,_) ),
            assert_target_reachable2(Node,DistToTarget)
        )
    ; otherwise ->
        assert_target_reachable2(Node,DistToTarget)
    ).
assert_target_reachable2(Node,DistToTarget) :-
    assert( target_reachable(Node,DistToTarget) ), fail.
assert_target_reachable2(Node,_DistToTarget) :-
    % the node can be reached by Event, so we know
    % that we have a path in the explored state-space that contains
    % Event a ends in a target node.
    event_dest(Node,Event),
    retract( target_needed(Event) ),
    % (this is a fail-loop)
    fail.
assert_target_reachable2(Node,DistToTarget) :-
    % propagate the information to the predecessors
    % (if we can reach a target node from this node, we also can
    %  reach the target node from each predecessor node.)
    predecessor(Node,Pre),
    update_minimal_path_for_events(Pre,Node,DistToTarget),
    NewDistToTarget is DistToTarget + 1,
    assert_target_reachable(Pre,NewDistToTarget),
    % (this is a fail-loop)
    fail.
assert_target_reachable2(_,_).

% Pre: node where the event starts
% Post: node where the events leads to
% DistToTarget: number of steps to reach the target when starting in the
%      pre node.
% For each event that should be covered, we store the minimal length of the
% path needed to cover that event.
% We store and compute that information while going back from the target node
% to set the target_reachable flag for states.
% To do that, we add the distance to the target to the depth of the node. The
% encountered minimum for each event is stored in minimum_path_length/2.
update_minimal_path_for_events(Pre,Post,DistToTarget) :-
    do_event(Pre,Event,Post),
    wanted_event(Event),
    node_covered(Pre,NodeDepth),
    MinLength is DistToTarget + NodeDepth,
    ( minimum_path_length(Event,Old) ->
        Old > MinLength,
        retractall( minimum_path_length(Event,_) ),
        assert( minimum_path_length(Event,MinLength) )
    ; otherwise ->
        assert( minimum_path_length(Event,MinLength) )),
    fail.
update_minimal_path_for_events(_Pre,_Post,_DistToTarget).

% explore the state-space to the given depth
complete_depth(Target,Depth) :-
    retract( bfs_node(Node,NodeDepth) ),!,
    ( NodeDepth =< Depth ->
        explore_node(Node,Target,Depth),
        complete_depth(Target,Depth)
    ; otherwise ->
        % the next nodes are deeper, just revert the previous retract
        % (this might not be necessary, because the data may not be used
        %  anymore after this, but just to keep it cleaner)
        asserta( bfs_node(Node,NodeDepth) )).
complete_depth(_Target,_Depth).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% XML output

testcase_output(Filename,_AllEvents,_SelectedEvents,Testcases) :-
    open_xml_testcase(Filename),
    % write_selected_events(AllEvents,SelectedEvents),
    write_testcases(Testcases,1),
    close_xml_testcase.

:- use_module(tools_io).
open_xml_testcase(Filename) :-
    safe_open_file(Filename,write,_Stream,[alias(xml_testcase),encoding('UTF-8')]),
    xml_write('<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<extended_test_suite>\n').
close_xml_testcase :-
    xml_write('</extended_test_suite>\n'),
    close(xml_testcase).

/*
write_selected_events(AllEvents,Events) :-
    xml_write('  <global_events>\n'),
    write_selected_events2(AllEvents,Events),
    xml_write('  </global_events>\n').
write_selected_events2([],_Selected).
write_selected_events2([Event|Rest],Selected) :-
    ( member(Event,Selected) -> Sel = true
    ; otherwise -> Sel = false),
    xml_write('    <event name=\"'),xml_write(Event),
    xml_write('\" covered=\"'),xml_write(Sel),
    xml_write('\"/>\n'),
    write_selected_events2(Rest,Selected).
    */


write_testcases([],_).
write_testcases([Testcase|TCrest],N) :-
    write_testcase(Testcase,N),
    N2 is N+1,
    write_testcases(TCrest,N2).

write_testcase(global_test_case(Trace),N) :-
    number_codes(N,Id),
    write_testcase2(Trace,[],Id).
write_testcase(global_test_case(Id,Trace),_N) :-
    write_testcase2(Trace,[],Id).
write_testcase(local_test_case(Id,Global,Local),_N) :-
    write_testcase2(Global,[Local],Id).

write_testcase2(GlobalTrace,LocalTraces,Id) :-
    xml_write('  <test_case id=\"'),put_codes(Id,xml_testcase),
    xml_write('\">\n'),
    write_trace(global,GlobalTrace),
    write_traces(LocalTraces,local),
    xml_write('  </test_case>\n').
write_traces([],_Type).
write_traces([Trace|Trest],Type) :-
    write_trace(Type,Trace),
    write_traces(Trest,Type).
write_trace(Type,Trace) :-
    xml_write('    <'),xml_write(Type),xml_write('>\n'),
    write_trace2(Trace,1),
    xml_write('    </'),xml_write(Type),xml_write('>\n').
write_trace2([],_).
write_trace2([Step|Trace],N) :-
    write_step(Step,N),
    N2 is N+1,
    write_trace2(Trace,N2).

write_step(Step,N) :-
    xml_write('      <step id=\"'),
    xml_write(N),
    xml_write('\">'),
    xml_write(Step),
    xml_write('</step>\n').

xml_write(Atom) :-
    write(xml_testcase, Atom).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% local test case generation

:- volatile is_local_event/1, is_global_event/2.
:- dynamic is_local_event/1, is_global_event/2.

tcl_create_local_testsuite(GlobalFile,LocalFile,MaxNewNodes,NumSat,NumUnsat) :-
    read_xml_testcases(GlobalFile,Global),
    set_max_nr_of_new_impl_trans_nodes(MaxNewNodes),
    process_testsuite(Global,Local,Unsatisfied),
    length(Local,NumSat),length(Unsatisfied,NumUnsat),
    append(Local,Unsatisfied,Testcases),
    testcase_output(LocalFile,[],[],Testcases).

process_testsuite(GlobalTestsuite,LocalTestsuite,Unsatisfied) :-
    reset_local_search,
    safe_get_inits(Inits),
    process_testsuite2(GlobalTestsuite,Inits,LocalTestsuite,Unsatisfied).
process_testsuite2([],_Inits,[],[]).
process_testsuite2([Global|Grest],Inits,LocalTestsuite,Unsatisfied) :-
    get_testcase_id(Global,Id),
    ( find_local_testcase(Global,Inits,Local) ->
        print('++ found local testcase: '),print(Id),nl,
        LocalTestsuite = [Local|Lnext],
        Unsatisfied = Unext
    ; otherwise ->
        print('--    no local testcase: '),print(Id),nl,
        LocalTestsuite = Lnext,
        Unsatisfied = [Global|Unext]),
    process_testsuite2(Grest,Inits,Lnext,Unext).

get_testcase_id(global_test_case(Id,_),I) :- atom_codes(I,Id).

% find_local_testcase(+Global,+Inits,-Local):
%   Global: a global test case
%   Inits: list of initial states
%   Local: a local test case
% search for a local trace of a global test case in the current model.
% A local trace consists of refined events of events of the global model,
% interspered by local events (i.e. events that refine skip)
find_local_testcase(global_test_case(Id,Global),Inits,local_test_case(Id,Global,Local)) :-
    member(Init,Inits),
    find_local_trace(Global,Init,Local).

find_local_trace([],_State,[]).
find_local_trace([GlobalEvent|Grest],State,LocalTrace) :-
    % do a number of local steps (without loop) until the global event
    find_local_path_to_global_event(State,GlobalEvent,[State],LocalTrace,RTrace,Dest),
    % continue with the next global event
    find_local_trace(Grest,Dest,RTrace).

% find_local_path_to_global_event(+State,+GlobalEvent,+History,-Trace,?RTrace,-End):
%   State: the state where the search starts
%   GlobalEvent: the global event that we are searching for, until reaching it, only
%     local events are taken
%   History: the states that we encountered before in this search, to prevend loops
%   Trace: the resulting trace, the last element is the global event, the tail of the
%     list is uninstantiated
%   RTrace: the tail of Trace
%   End: the last state in the search, the global event leads into this state
:- use_module(state_space,[compute_transitions_if_necessary/1]).
find_local_path_to_global_event(State,GlobalEvent,History,[Event|Trace],RTrace,End) :-
    % make sure this node is explored
    state_space:compute_transitions_if_necessary(State),
    % do an (non-deterministic) step
    do_event(State,Event,Dest),
    ( is_local_event(Event) ->
        % local events can be taken as long we do not loop
        \+ member(Dest,History),
        % continue the search with the new state
        find_local_path_to_global_event(Dest,GlobalEvent,[Dest|History],Trace,RTrace,End)
    ; is_global_event(Event,GlobalEvent) ->
        % if we encounter the global event we search for, we can stop
        Trace = RTrace,
        Dest = End).

% initial the data structures for the search
reset_local_search :-
    get_all_events(Events),
    retractall(is_local_event(_)),
    retractall(is_global_event(_,_)),
    assert_local_events(Events).

% for each event, we store whether it's local or global,
% and if it's global, we store the refined event
assert_local_events([]).
assert_local_events([Event|Erest]) :-
    get_operation_info(Event,Info),
    ( member(refines(Ref),Info) ->
        assert_abstract_events(Event,Ref)
    ; otherwise -> assert( is_local_event(Event) )),
    assert_local_events(Erest).
assert_abstract_events(Event,Refines) :-
    member(Abstract,Refines), assert( is_global_event(Event,Abstract) ),
    % (fail-loop)
    fail.
assert_abstract_events(_,_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% read XML testcases
read_xml_testcases(Filename,TestCases) :-
    read_string_from_file(Filename,Codes),
    xml_parse(Codes,xml(_Atts,Content)),
    member(element(extended_test_suite,_,XmlTestSuite),Content),
    extract_testcases(XmlTestSuite,TestCases).
extract_testcases(Content,TestCases) :-
    findall(TC, extract_testcase(Content,TC), TestCases).
extract_testcase(Content,global_test_case(Id,GlobalTrace)) :-
    member(element(test_case,Atts,EContent),Content),
    member(id=Id, Atts),
    member(element(global,_,GContent),EContent),
    findall(Step,
            ( member(element(step,_,[pcdata(P)]),GContent),atom_codes(Step,P) ),
            GlobalTrace).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dot debugger

:- use_module(library(process),[process_create/3,process_wait/2]).
:- public sapdebug/0.
sapdebug :-
    process_create(path(dot),
                   ['-T','pdf','-o','sap_debug.pdf'],
                   [stdin(pipe(S)),process(P)]),
    write(S,'digraph sap_debug {\n'),
    sapwrite_nodes(S),
    write(S,'\n'),
    sapwrite_edges(S),
    write(S,'}\n'),
    close(S),
    process_wait(P,_).

sapwrite_nodes(S) :-
    node_covered(Node,_),
    ( bfs_node(Node,_) -> Shape = triangle; Shape = box ),
    ( target_reachable(Node,_) -> Style = ',filled'; Style = ''),
    write_l(S,['  ',Node,' [shape=',Shape,', style=\"rounded',Style,'\"];\n']),
    fail.
sapwrite_nodes(_).

sapwrite_edges(S) :-
    predecessor(Post,Pre),
    write_l(S,['  ',Pre,' -> ',Post,';']),
    fail.
sapwrite_edges(_).

write_l(_S,[]) :- !.
write_l(S,[H|T]) :-
    write(S,H),write_l(S,T).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% test case generation for a given list of event by
% constraint solving

:- use_module(b_interpreter).
:- use_module(b_interpreter_eventb).
:- use_module(kernel_waitflags).
:- use_module(store).
:- use_module(state_space).

create_testcase_path(Events,Timeout,Result) :-
   create_testcase_path(init,Events,b(truth,pred,[]),Timeout,Result).
create_testcase_path(INIT,Events,Timeout,Result) :-
   create_testcase_path(INIT,Events,b(truth,pred,[]),Timeout,Result).
create_testcase_path(INIT,Events,EndPredicate,Timeout,Result) :-
    % allow user interrupts during the search, just fail if that happens
    testcase_path_timeout(INIT,Timeout,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,R),
    ( R==ok ->
        add_constants_to_state_space(ConstantsState,TestStates,ConstTestStates,SetupSequence,StartStateId),
        add_operations_to_state_space(Operations,StartStateId,ConstTestStates,TransInfos,OpSequence),
        append(SetupSequence,OpSequence,Result)
    ; otherwise -> Result = R).

testcase_path_timeout(INIT,0,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,R) :-
    !,testcase_path_interrupt(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,R).
testcase_path_timeout(INIT,Timeout,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,Result) :-
    time_out(testcase_path_interrupt(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,R),
             Timeout,
             T),
    ( T==success -> Result=R
    ; otherwise  -> Result=timeout).
testcase_path_interrupt(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos,Result) :-
    user_interruptable_call_det(testcase_path(INIT,Events,EndPredicate,ConstantsState,Operations,TestStates,TransInfos),
                                Interrupt),
    ( Interrupt==ok -> Result=ok
    ; otherwise     -> Result=interrupt).

add_constants_to_state_space(ConstantsState,TestStates,TestStates,[],root) :-
    empty_state(ConstantsState),!.
add_constants_to_state_space(ConstantsState,TestStates,ConstTestStates,[Tuple],ConstantsID) :-
    Tuple = (SetupID,Setup,root,ConstantsID),
    specfile:create_setup_constants_operation(ConstantsState,Setup),
    user:tcltk_add_new_transition_transid(root,Setup,ConstantsID,concrete_constants(ConstantsState),[],SetupID),
    remove_constants(TestStates,ConstantsState,ConstantsID,ConstTestStates).
remove_constants([],_,_,[]).
remove_constants([InState|Irest],Constants,ConstId,[const_and_vars(ConstId,OutState)|Orest]) :-
    append(OutState,Constants,InState),!,
    remove_constants(Irest,Constants,ConstId,Orest).


add_operations_to_state_space([],_,[],[],[]).
add_operations_to_state_space([Operation|Orest],PrevId,[State|Srest],[TransInfo|Trest],[Tuple|Irest]) :-
    Tuple = (OpId,Operation,PrevId,NextId),
    user:tcltk_add_new_transition_transid(PrevId,Operation,NextId,State,TransInfo,OpId),
    add_operations_to_state_space(Orest,NextId,Srest,Trest,Irest).

testcase_path(INITIALISE,Events,EndPredicate,NormalisedConstantsState,OPS,StateSequence,TINFOS) :-
    init_wait_flags(WF),
    (INITIALISE=init
     -> OPS =  [OpInit|Operations], TINFOS = [ITransInfo|OpTransInfos],
        tc_initialise(ConstantsState,InitialState,OpInit,ITransInfo,WF)
     ;  OPS = Operations, TINFOS = OpTransInfos,
        set_up_a_valid_state(ConstantsState,InitialState,WF),
        (INITIALISE = pred(InitialPredicate) -> eval_predicate(InitialPredicate,InitialState,WF)
         ; true)
    ),
    set_up_events(Events,InitialState,ConstantsState,Operations,States,FinalState,OpTransInfos,WF),
    eval_predicate(EndPredicate,FinalState,WF),
    ground_wait_flags(WF),
    normalise_states([InitialState|States],StateSequence),
    normalise_state(ConstantsState,NormalisedConstantsState).

eval_predicate(Predicate,State,WF) :-
    empty_state(EmptyState),
    b_test_boolean_expression(Predicate,EmptyState,State,WF).

set_up_events([],State,_ConstState,[],[],State,[],_WF).
set_up_events([Event|Erest],InState,ConstantState,[Operation|Orest],
              [InterState|Srest],OutState,[TransInfo|Trest],WF) :-
    set_up_transition(Event,Operation,ConstantState,InState,InterState,TransInfo,WF),
    set_up_events(Erest,InterState,ConstantState,Orest,Srest,OutState,Trest,WF).

tc_initialise(ConstantState,InitialState,Initialisation,TransInfo,WF) :-
    set_up_constants_state(ConstantState,WF),
    set_up_initialisation(ConstantState,InitialState,Initialisation,TransInfo,WF). % TO DO: here I would like to be able to set up the invariant instead, if requested by another parameter
    
set_up_constants_state(ConstantState,WF) :-
    b_get_properties_from_machine(Properties),
    b_get_machine_constants(Constants),
    empty_state(EmptyState),
    set_up_typed_localstate(Constants,_FreshVars,TypedVals,EmptyState,ConstantState),
    b_tighter_enumerate_values(TypedVals,WF),
    b_test_boolean_expression(Properties,EmptyState,ConstantState,WF).

% set up a valid state satisfying the invariant
set_up_a_valid_state(ConstantState,ValidState,WF) :- 
    set_up_constants_state(ConstantState,WF),
    b_get_machine_variables(Variables),
    % create_target_state(Variables,Values,ConstantsState,ValidState,WF),
    set_up_typed_localstate(Variables,_Values,TypedVals,ConstantState,ValidState),
    b_tighter_enumerate_values(TypedVals,WF),
    b_get_invariant_from_machine(Invariant),
    empty_state(EmptyState),
    b_test_boolean_expression(Invariant,EmptyState,ValidState,WF).
    

normalise_states([],[]).
normalise_states([I|Irest],[O|Orest]) :-
    normalise_state(I,O),
    normalise_states(Irest,Orest).
normalise_state(I,O) :-
    extract_vars(I,VI,VO,O),
    normalise_store(VI,VO).
extract_vars(const_and_vars(ConstID,IV),IV,OV,const_and_vars(ConstID,OV)) :- !.
extract_vars(IV,IV,OV,OV).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% constraint based test case generation

tcl_generate_cb_testcases(EventSelection,TargetString,MaxDepth,Filename,list(Uncovered)) :-
    % first translate Tk EventSelection Number list into Events
    get_selected_events(EventSelection,_AllEvents,Events),
    % parse target predicate for test-case sequences
    parse_optional_target(TargetString,Target),
    cbc_gen_test_cases(Events,Target,MaxDepth,Filename,Uncovered).

% parse a predicate string; if empty return truth
parse_optional_target(TargetString,Target) :-
    is_empty_string(TargetString),!,create_texpr(truth,pred,[],Target).
parse_optional_target(TargetString,Target) :-
    b_parse_machine_predicate(TargetString,Target).

is_empty_string(Atom) :-
    atom_codes(Atom,Codes),
    is_empty_string2(Codes).
is_empty_string2([]).
is_empty_string2([32|Codes]) :- is_empty_string2(Codes).

:- volatile cb_path/3, cb_uncovered_event/1.
:- volatile cb_try_state_caching/0.
:- dynamic cb_path/3, cb_uncovered_event/1.
:- dynamic cb_try_state_caching/0.

% constraint-based test case generation
cbc_gen_test_cases(EventsToCover,TargetPredicate,MaxLength,Filename,UncoveredEvents) :-
    get_preference(time_out,Timeout),
    % TO DO: factor the following 4 lines of code for reuse
    get_preference(use_smt_mode,Smt),
    set_preference(use_smt_mode,true),
    get_preference(use_clpfd_solver,CLPFD),
    set_preference(use_clpfd_solver,true), % we also enable CLPFD
    open_xml_testcase(Filename),
    call_cleanup( (cb_test_cases_init_search(EventsToCover),
                      cb_timed_testcases(TargetPredicate, MaxLength, Timeout)), 
                  (close_xml_testcase,
                      set_preference(use_smt_mode,Smt),
                      set_preference(use_clpfd_solver,CLPFD))),
    findall(E,cb_uncovered_event(E),UncoveredEvents).

cb_timed_testcases(TargetPred,MaxLengthOfTestCase,Timeout) :-
    statistics(walltime, [Start|_]),
    InitialLength = 1,
    cb_test_cases2(InitialLength,TargetPred,MaxLengthOfTestCase,Timeout), 
    statistics(walltime, [End|_]),
    Runtime is End - Start,
    write_results(Runtime).

cb_test_cases2(CurLengthOfTestCase,_TargetPred,_MaxLengthOfTestCase,_Timeout) :-
    \+ cb_has_uncovered_events,!, L1 is CurLengthOfTestCase-1,
    format('~nEverything covered, Length=~w~n',[L1]).
cb_test_cases2(CurLengthOfTestCase,_TargetPred,MaxLengthOfTestCase,_Timeout) :-
    CurLengthOfTestCase > MaxLengthOfTestCase,!,
    format('~nLimit Reached: ~w~n',[MaxLengthOfTestCase]).
cb_test_cases2(CurLengthOfTestCase,TargetPred,MaxLengthOfTestCase,Timeout) :-
    % now try to find a path of length CurLengthOfTestCase
    cb_find_paths(CurLengthOfTestCase,TargetPred,MaxLengthOfTestCase,Timeout),
    !,
    L2 is CurLengthOfTestCase+1,
    cb_test_cases2(L2,TargetPred,MaxLengthOfTestCase,Timeout).
cb_test_cases2(CurLengthOfTestCase,_TargetPred,_MaxLengthOfTestCase,_Timeout) :-
    % either everything covered or MaxLength reached or no way to extend current paths
    format('~nCould not generate a path of length ~w~n',[CurLengthOfTestCase]).

% reset datastructures for CBC test case generation
cb_test_cases_init_search(Events) :-
    reset_counters,
    retractall(cb_path(_,_,_)), % retract all found paths
    assert(cb_path(0,[],no_state)), % generate an intial path
    % set up all coverage targets:
    retractall(cb_uncovered_event(_)), 
    (Events=all -> cb_test_cases_init_all
     ; maplist(assert_cb_uncovered_event,Events)
     ).
cb_test_cases_init_all :- b_get_machine_operation(Event,_,_,_),
    assert_cb_uncovered_event(Event), fail.
cb_test_cases_init_all.  
assert_cb_uncovered_event(E) :- assert(cb_uncovered_event(E)).

cb_find_paths(Length,TargetPredicate,MaxLength,Timeout) :-
    format('~nCBC test search: starting length ~w~n',[Length]),
    % set up template for Path = PrefixPath ^ [Event]
    cb_prefix_trace(Length,PreLength,PrefixPath,Event,Path),
    % choose any existing prefix of length PreLength for which a CBC solution was found:
    cb_path(PreLength,PrefixPath,LastStateId), write('*'),flush_output,
    % check if full coverage not yet achieved:
    cb_has_uncovered_events,
    % now choose any event; uncovered ones first:
    b_ordered_get_operation(Event),  write('.'),flush_output,
    % statistics information: store that we attempt another path of length Length:
    count_path(Length),
    (Length>=MaxLength -> Result=testcase, % no need to find partial solutions now
        cb_uncovered_event(Event) % only set up paths for Events we want to cover
      ; true
      ),
    % and try to set up that path PrefixPath ^ [Event]
    cb_set_up_path(Path,Event,TargetPredicate,LastStateId,Timeout,Result),
    % now store info that we have found a solution for length Length with Result
    count_found(Length,Result),
    % fail-loop
    fail.
cb_find_paths(Length,_TargetPredicate,_MaxLengthOfTestCase,_Timeout) :-
    % success if a path of this length was generated, fail otherwise
    cb_path(Length,_,_).

% enumerate events with uncovered first:
b_ordered_get_operation(Event) :- 
   findall(Event,b_get_machine_operation(Event,_,_,_),AllEvents),
   include(cb_uncovered_event,AllEvents,Uncovered),
   exclude(cb_uncovered_event,AllEvents,Covered),
   % Note: we do a findall and exclude/include first to avoid side-effects from modifying cb_uncovered_event
   %  and possibly forgetting an event
   %print(uncovered(Uncovered)),nl,
   (member(Event,Uncovered)
    ;
     member(Event,Covered)).

cb_prefix_trace(Length,PreLength,PrefixPath,Event,Path) :-
    PreLength is Length-1,
    length(PrefixPath,PreLength),append(PrefixPath,[Event],Path).

cb_has_uncovered_events :-
    cb_uncovered_event(_),!.

cb_set_up_events(Trace,InitialState,ConstantsState,FinalState,DetailedTrace,WF) :-
    DetailedTrace = dt1(States,Operations,OpTransInfos),
    set_up_events(Trace,InitialState,ConstantsState,Operations,States,FinalState,OpTransInfos,WF).

combine_detail_infos(ConstantsState,InitialState,Initialisation,InitInfo,DT1,DT) :-
    DT1 = dt1(States,Operations,OpTransInfos),
    DT = dt(ConstantsState,
            [InitialState|States],
            [Initialisation|Operations],
            [InitInfo|OpTransInfos]).

cb_set_up_path(Trace,Event,TargetPredicate,StateId,Timeout,Result) :-
   % StateID is the last state of the previously found path Trace (without Event)
    cb_try_state_caching, % if true: try to use model checking paths found; currently disabled
    StateId \== no_state,
    lookup_state(StateId,ConstantState,InState),
    time_out( (init_wait_flags(WF),
                  set_up_transition(Event,Operation,ConstantState,InState,OutState,TransInfo,WF),
                  cb_finalize_path(TargetPredicate,OutState,WF,LResult)),
              Timeout,
              TO_Success),
    ( TO_Success = time_out ->
        cb_write_timeout(Timeout,Trace),fail
    ; otherwise ->
        cb_store_single_step_in_statespace(ConstantState,OutState,StateId,Operation,TransInfo,LastStateId,Sequence),
        cb_new_step(Trace,LResult,LastStateId,Sequence),
        % if we did not find a test case, just a normal step, we need to fall back
        % to constraint solving on the full path
        !,
        ( LResult==testcase -> Result=testcase
        ; cb_set_up_path(Trace,Event,TargetPredicate,no_state,Timeout,testcase) -> Result=testcase
        ; otherwise -> Result = path)).
cb_set_up_path(Trace,_Event,TargetPredicate,_StateId,Timeout,Result) :-
    time_out( (init_wait_flags(WF),
                  % set up constants and initialisation:
                  tc_initialise(ConstantsState,InitialState,Initialisation,TransInfo,WF),
                  combine_detail_infos(ConstantsState,InitialState,Initialisation,TransInfo,DT1,DT),
                  cb_set_up_events(Trace,InitialState,ConstantsState,FinalState,DT1,WF),
                  % either find a real testcase where TargetPredicate holds at end
                  % or find a path where TargetPredicate does not hold at end:
                  cb_finalize_path(TargetPredicate,FinalState,WF,Result),!),
              Timeout,
              TO_Success),
    ( TO_Success = time_out ->
        cb_write_timeout(Timeout,Trace),fail
    ; otherwise ->
        cb_store_path_in_statespace(DT,LastStateId,Sequence),
        cb_new_step(Trace,Result,LastStateId,Sequence)).
cb_new_step(Trace,Result,LastStateId,Sequence) :-
    cb_save_path(Trace,LastStateId),
    ( Result == testcase -> cb_found_test_case(Trace,Sequence)
    ; otherwise          -> true).

lookup_state(StateId,ConstStore,VarStore) :-
    visited_expression(StateId,Store),
    lookup_state2(Store,ConstStore,VarStore).
lookup_state2(const_and_vars(ConstId,VarStore),ConstStore,VarStore) :-
    !,visited_expression(ConstId,concrete_constants(ConstStore)).
lookup_state2(Store,Empty,Store) :- empty_state(Empty).

cb_finalize_path(TargetPredicate,_FinalState,WF,testcase) :-
    % if the target predicate is just true, we can store every found path
    % directly as a possible test trace, too.
    is_truth(TargetPredicate),!, % avoid ground_wait_flags twice if path infeasible
    ground_wait_flags(WF),!.
cb_finalize_path(TargetPredicate,FinalState,WF,testcase) :-
    % if the target predicate is fulfilled in the last state, we
    % can store the solution as a test trace.
    eval_predicate(TargetPredicate,FinalState,WF),
    ground_wait_flags(WF),!.
cb_finalize_path(_TargetPredicate,_FinalState,WF,path) :-
    % if the path is not feasible with the target predicate applied to the
    % last state, it still might be feasible as a starting point for other traces.
    ground_wait_flags(WF),!.

cb_found_test_case(Trace,Sequence) :-
    % mark all events in Trace as covered
    cb_mark_all_covered(Trace),
    cb_get_initialised_state(Sequence,InitState),
    % Sequence is a list of (_Id,Op,Src,Dst)
    cb_write_test_case(Sequence,InitState).

cb_get_initialised_state([(_Id,_Op,Src,_Dst)|_],InitState) :-
    visited_expression(Src,Store),
    expand_const_and_vars_to_full_store(Store,InitState).

cb_mark_all_covered([]).
cb_mark_all_covered([Event|Rest]) :-
    ( retract(cb_uncovered_event(Event)) ->
        nl,write('CBC test search: covered event '),write(Event),nl
    ; otherwise -> true),
    cb_mark_all_covered(Rest).

cb_write_test_case(Sequence,InitState) :-
    xml_write('  <test_case>\n'),
    cb_xml_initialisation(InitState),
    cb_xml_sequence(Sequence),
    xml_write('  </test_case>\n'),
    flush_output(xml_testcase).
cb_xml_initialisation(InitState) :-
    xml_write('    <initialisation>\n'),
    b_get_machine_constants(Constants),cb_xml_init2(InitState,Constants,constant),
    b_get_machine_variables(Variables),cb_xml_init2(InitState,Variables,variable),
    xml_write('    </initialisation>\n').
cb_xml_init2(Store,Identifiers,Type) :-
    get_texpr_id(TId,Id),
    member(TId,Identifiers),
    lookup_value(Id,Store,BValue),
    cb_xml_value(Id,Type,BValue),
    fail.
cb_xml_init2(_Store,_Identifiers,_Type).
cb_xml_value(Id,Type,BValue) :-
    translate_bvalue(BValue,SValue),
    xml_write('      <value type=\"'),
    xml_write(Type),
    xml_write('\" name=\"'),xml_write(Id),xml_write('\">'),
    xml_write(SValue),xml_write('</value>\n').
cb_xml_sequence([]).
cb_xml_sequence([(_Id,Op,Src,Dst)|Rest]) :-
    functor(Op,Name,_),
    cb_xml_step(Name,Op,Src,Dst),
    cb_xml_sequence(Rest).
cb_xml_step(Event,Op,SrcId,DstId) :-
    b_get_machine_operation(Event,_,Parameters,_),
    Parameters = [_|_],!,
    xml_write('    <step name=\"'),xml_write(Event),xml_write('\">\n'),
    Op =.. [_|Args],
    maplist(cb_xml_parameter,Parameters,Args),
    expand_state(SrcId,Src),
    expand_state(DstId,Dst),
    maplist(cb_xml_modified(Src),Dst),
    xml_write('    </step>\n').
cb_xml_step(Event,_,_Src,_Dst) :-
    xml_write('    <step name=\"'),xml_write(Event),xml_write('\" />\n').
cb_xml_parameter(Parameter,BValue) :-
    get_texpr_id(Parameter,Id),
    cb_xml_value_element(value,Id,BValue).
cb_xml_modified(Src,bind(Name,NewValue)) :-
    cb_var_is_modified(Name,NewValue,Src),!,
    cb_xml_value_element(modified,Name,NewValue).
cb_xml_modified(_,_).
cb_var_is_modified(Name,NewValue,Src) :-
    memberchk(bind(Name,OldValue),Src),!,
    not_equal_object(OldValue,NewValue).
cb_var_is_modified(Name,_NewValue,Src) :-
    nonmember(bind(Name,_OldValue),Src).

cb_xml_value_element(Element,Name,Value) :-
    translate_bvalue(Value,SValue),
    xml_write('      <'),xml_write(Element),
    xml_write(' name=\"'),xml_write(Name),xml_write('\">'),
    xml_write(SValue),
    xml_write('</'),xml_write(Element),xml_write('>\n').    

expand_state(Id,VariableBindings) :-
    visited_expression(Id,State),
    expand_to_constants_and_variables(State,_ConstantBindings,VariableBindings).

cb_write_timeout(Timeout,Events) :-
    length(Events,Length),
    count_timeout(Length),
    write(timeout),flush_output,
    xml_write('  <!-- timeout ('),xml_write(Timeout),xml_write(') ms) for events '),
    cb_write_timeouts2(Events),
    xml_write(' --!>\n'),flush_output(xml_testcase).
cb_write_timeouts2([Event]) :-
    !,xml_write(Event).
cb_write_timeouts2([Event|Erest]) :-
    xml_write(Event),xml_write(','),cb_write_timeouts2(Erest).

cb_save_path(Trace,LastStateId) :-
    length(Trace,CurDepth),
    assert(cb_path(CurDepth,Trace,LastStateId)).

cb_store_path_in_statespace(dt(ConstantsState1,States1,Operations,TransInfos),LastStateId,RealOperations) :-
    normalise_states([ConstantsState1|States1],[ConstantsState|States]),
    add_constants_to_state_space(ConstantsState,States,ConstStates,_SetupSequence,StartStateId),
    add_operations_to_state_space(Operations,StartStateId,ConstStates,TransInfos,OpSequence),
    OpSequence = [_InitOp|RealOperations],
    last(OpSequence,(_,_,_,LastStateId)).

cb_store_single_step_in_statespace(ConstantsState,State1,PrevStateId,Operation,TransInfo,NewStateId,[Tuple]) :-
    normalise_state(State1,State2),
    ( cb_find_constants_state(ConstantsState,ConstId) ->
        remove_constants([State2],ConstantsState,ConstId,[State])
    ; otherwise ->
        State = State2),
    add_operations_to_state_space([Operation],PrevStateId,[State],[TransInfo],[Tuple]),
    Tuple = (_,_,_,NewStateId).

cb_find_constants_state(ConstantsState,_ConstId) :-
    empty_state(ConstantsState),!,fail.
cb_find_constants_state(ConstantsState,ConstId) :-
    visited_expression(ConstId,concrete_constants(ConstantsState)),!.

test_generation_by_xml_description(Filename) :-
    read_test_generation_description(Filename,Description),
    execute_test_generation(Description).
execute_test_generation(cb_test_description(Events,TargetString,MaxDepth,Preferences,Filename,StateCaching)) :-
    parse_optional_target(TargetString,Target),
    maplist(set_ec_preference,Preferences),
    retractall(cb_try_state_caching),
    ( StateCaching = true -> assert(cb_try_state_caching); otherwise -> true),
    cbc_gen_test_cases(Events,Target,MaxDepth,Filename,_Uncovered).
set_ec_preference(pref(Name,Value)) :-
    set_eclipse_preference(Name,Value).

read_test_generation_description(Filename,Description) :-
    read_string_from_file(Filename,Codes),
    xml_parse(Codes,xml(_Atts,Content)),
    ( read_test_generation_description2(Content,Description) ->
        true
    ; otherwise ->
        add_error(sap,'error while reading test description file'),fail).
read_test_generation_description2(Xml,cb_test_description(Events,Target,MaxDepth,Preferences,File,StateCaching)) :-
    expect_xml_element('test-generation-description',Xml,Content),!,
    expect_xml_element('output-file',Content,[pcdata(FileCodes)]),atom_codes(File,FileCodes),
    get_xml_events(Content,Events),
    expect_xml_element(parameters,Content,XmlParameters),
    % maximum depth in parameters section
    expect_xml_element('maximum-depth',XmlParameters,[pcdata(MaxDepthCodes)]),number_codes(MaxDepth,MaxDepthCodes),
    % optional target predicate
    ( memberchk(element(target,_,[pcdata(Tcodes)]),Content) -> atom_codes(Target,Tcodes)
    ; otherwise -> Target = ''),
    % set preference in parameters section
    findall(PL,member(element(preference,PL,_),XmlParameters),PrefAttLists),
    maplist(convert_xml_preference,PrefAttLists,Preferences),
    % check if state caching should be used
    ( memberchk(element('state-caching',_,_),Content) -> StateCaching = true
    ; otherwise -> StateCaching = false).
convert_xml_preference(PAL,pref(Name,Value)) :-
    ( memberchk(name=CName,PAL) -> atom_codes(Name,CName)
    ; otherwise -> add_error(sap,'Missing preference name'),fail),
    ( memberchk(value=CValue,PAL) -> atom_codes(Value,CValue)
    ; otherwise -> add_error(sap,'Missing preference value'),fail).
get_xml_events(Content,Events) :-
    memberchk(element('event-coverage',_,EventElements),Content),!,
    findall(E,(member(element(event,_,[pcdata(Ecodes)]),EventElements),atom_codes(E,Ecodes)),Events).
get_xml_events(_Content,Events) :-
    get_all_events(Events).


expect_xml_element(ElemName,Xml,Elements) :-
    ( memberchk(element(ElemName,_,Es),Xml) -> Elements=Es
    ; otherwise ->
        atom_concat('did not find xml element ',ElemName,Msg),
        add_error(sap,Msg),fail).
              
:- volatile timeouts/2, path_count/2, solution_count/2, testcase_count/2.
:- dynamic timeouts/2, path_count/2, solution_count/2, testcase_count/2.

reset_counters :-
    retractall(timeouts(_,_)),
    retractall(path_count(_,_)),
    retractall(solution_count(_,_)),
    retractall(testcase_count(_,_)).

count(Prefix,Length) :-
    functor(C,Prefix,2),arg(1,C,Length),arg(2,C,Count),
    ( retract(C) -> true
    ; otherwise  -> Count = 0),
    NCount is Count+1,
    functor(N,Prefix,2),arg(1,N,Length),arg(2,N,NCount),
    assert(N).

count_timeout(Length) :- count(timeouts,Length).
count_path(Length) :- count(path_count,Length).
count_found(Length,Result) :-
    ( Result==testcase -> count(testcase_count,Length)
    ; otherwise -> true),
    count(solution_count,Length).

write_result_item(Desc,Db) :-
    format('~w: ',[Desc]),
    functor(C,Db,2),arg(1,C,Length),arg(2,C,Num),
    findall( count(Length,Num), call(C), UResults),
    sort(UResults,Results),
    write_result_item2(Results,Total),
    format('total: ~w~n',[Total]).

write_result_item2([],0).
write_result_item2([count(Length,Num)|Rest],Total) :-
    format('~w:~w, ',[Length,Num]),
    write_result_item2(Rest,SumRest),
    Total is SumRest + Num.

write_results(Runtime) :-
    write_result_item('Timeouts',timeouts),
    write_result_item('Checked paths (Depth:Paths)',path_count),
    write_result_item('Found paths',solution_count),
    write_result_item('Found test cases', testcase_count),
    format('Runtime: ~w ms~n',[Runtime]).
