% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(self_check,
    [assert_pre/2, assert_post/2,pp_mnf/1, pp_cll/1, mnf/1, mnf/2,
     mnf_det/1, det_call/1, force_det_call/1,
     assert_must_succeed/1,assert_must_fail/1,
     assert_must_succeed_multiple/1, assert_must_succeed_any/1,
     must_fail/1, must_succeed/1, must_succeed_without_residue/1,
     
     run_time_type_check_mode/1,
     turn_on_run_time_type_checks/0,turn_off_run_time_type_checks/0,
     
     get_module_list/1,
   
     perform_self_check/0, perform_self_check/1,
     perform_verbose_self_check/0, perform_verbose_self_check/1,
     display_self_checks/0,

     disable_interaction_on_errors/0,
     
     time_out_call/2, time_out_with_factor_call/3, time_out_call/1,
     
     check_deterministic/1, check_det/1,
     
     list_skeleton/1
    
     ]).

:- use_module(debugging_calls).
%:- disable_debugging_calls.

:- use_module(module_information,[module_info/2]).
:- module_info(group,infrastructure).
:- module_info(description,'This module provides predicates to define and run unit tests.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').


% first code which asserts pre-, post-conditions and self-checks

% first a copy of print_error; so that this bit of self-check does not rely on other modules
my_print_error(Error) :-
	current_output(X),
	set_output(user_error),
	(var(Error) 
	  -> print_message(error,'_')
	  ;  (write('! '),write_term(Error,[max_depth(50),numbervars(true)]),nl)),
	set_output(X).

:- dynamic pre_condition/2, post_condition/2.
:- meta_predicate assert_pre(0,0).
assert_pre(X,Pre) :-  % print(adding_pre(X)),nl,
    ((nonvar(X),nonvar(Pre))
       -> (retractall(pre_condition(X,_)),assert(pre_condition(X,Pre)))
       ;  (my_print_error('### illegal variable(s) in: '),
           my_print_error(assert_pre(X,Pre)),fail)
    ).
:- meta_predicate assert_post(0,0).
assert_post(X,Post) :- 
    ((nonvar(X),nonvar(Post))
       -> (retractall(post_condition(X,_)),assert(post_condition(X,Post)))
       ;  (my_print_error('### illegal variable(s) in: '),
           my_print_error(assert_post(X,Post)),fail)
    ).

:- dynamic self_check/2.
:- dynamic self_check_module/1.
:- volatile self_check/2, self_check_module/1. % this means that self-check cannot be run in the compiled versions !

% has a strange behaviour; investigate :- meta_predicate self_check(:,*). 

%add_self_check(_,_) :- !. % comment in to not include self-checks in distribution
add_self_check(Module,X) :-
    (self_check_module(Module) -> true; 
                     otherwise -> assert(self_check_module(Module))),
    (nonvar(X)
                   -> (self_check(X,Module) -> true ;
                                  otherwise -> assert(self_check(X,Module))); 
         otherwise -> my_print_error('### trying to assert variable as self_check: '),my_print_error(X),fail
    ).


:- meta_predicate assert_must_succeed_any(0).
:- meta_predicate assert_must_succeed(0).
:- meta_predicate assert_must_succeed_multiple(0).
:- meta_predicate assert_must_fail(0).
%:- meta_predicate kernel_waitflags:assert_must_abort_wf(0,*).

assert_must_succeed_any(M:X) :-
    add_self_check(M,must_succeed(M:X)).
assert_must_succeed(M:X) :-
    add_self_check(M,must_succeed_without_residue(M:X)).
assert_must_succeed_multiple(M:X) :-
    add_self_check(M,must_succeed_multiple_without_residue(M:X)).
assert_must_fail(M:X) :-
    add_self_check(M,must_fail(M:X)).


% ------------------------

:- use_module(gensym).
:- use_module(debug).
:- use_module(tools_printing).   
:- use_module(typechecker).
:- use_module(error_manager).
:- use_module(junit_tests).
:- use_module(library(lists)).

:- use_module(tools,[print_message/1,call_residue/2, predicate_functor/3, safe_numbervars/3]).
/* Example use:

:- assert_pre(user:b_fd_type(_G,_L,_U),true).
:- assert_post(user:b_fd_type(G,L,U),(atomic(G),(integer(L),integer(U)))).

*/



 :- use_module(library(timeout)).
 :- use_module(preferences).
 
:- meta_predicate time_out_with_factor_call(0,*,0).



time_out_with_factor_call(Call,Factor,Res) :-
  preferences:get_preference(time_out,CurTO),
  TO is integer(CurTO*Factor),
  %%print(time_out(Call,TO)),nl,
  statistics(runtime,[Start,_]),
  time_out(Call,TO,TimeOutRes),
  (TimeOutRes=time_out
    -> statistics(runtime,[Stop,_]),
       predicate_functor(Call,F,N),
       add_message(self_check,'Time out occurred: ',F/N), 
       print_term_summary(Call),
       Diff is Stop-Start,
       print(Diff), print(' ms runtime (time_out = '), print(TO), print(' ms)'),nl,
       call(Res)
    ;  true).

:- meta_predicate time_out_call(0,0).
time_out_call(Call,Res) :-
  preferences:get_preference(time_out,CurTO),
  time_out(Call,CurTO,TimeOutRes),
  (TimeOutRes=time_out
    -> predicate_functor(Call,F,N),
       add_message(self_check,'Time out occurred: ',F/N), 
       print_term_summary(Call),
       call(Res)
    ;  true).

:- meta_predicate time_out_call(0).
time_out_call(Call) :-
   time_out_call(Call,add_error_and_fail(time_out_call,'*** TIMEOUT occured: ',Call)).
   
/* some auxilary predicates that can be used for pre,post conditions: */

% used in some pre- and post-conditions
list_skeleton(X) :- 
  (nonvar(X),list_skel2(X) -> true
              ; (my_print_error('### not list skeleton: '), my_print_error(X),fail)).
list_skel2([]).
list_skel2([_H|T]) :- nonvar(T), list_skel2(T).



:- dynamic run_time_type_check_mode/1.
run_time_type_check_mode(on).

turn_on_run_time_type_checks :- 
      retract(run_time_type_check_mode(_)),
      assert(run_time_type_check_mode(on)).
      
turn_off_run_time_type_checks :- 
      retract(run_time_type_check_mode(_)),
      assert(run_time_type_check_mode(off)).


/* ===================================================== */

:- dynamic prepost_no_error_so_far/0.

prepost_no_error_so_far. % :- fail. 
/* as errors are now displayed by error_manager,  no need for interaction ?? */

disable_interaction_on_errors :-
	retractall(prepost_no_error_so_far).


/* ===================================================== */

:- register_debugging_call pp_mnf(*), pp_cll(*), mnf(*), mnf(-,*), det_call(*).
:- meta_predicate pp_mnf(0), pp_cll(0), mnf(0), mnf(-,0), det_call(0).

pp_mnf(X) :- prepost_mnf_call(X).
pp_cll(X) :- prepost_call(X).
mnf(X) :- mnf_call(X).
mnf(ProgramPoint,X) :- mnf_call_with_pp(ProgramPoint,X).


/* ===================================================== */


:- use_module(library(timeout)).

rt_timeout_call(Call) :- run_time_type_check_mode(off),!,call(Call).
rt_timeout_call(Call) :- time_out(Call,5000,TimeOutRes), 
  (TimeOutRes = success -> true ; (add_error(self_check,'### TIMEOUT: ',Call),fail)).

prepost_call(X) :-
	/* print(pre(X)),nl, */
	(verify_pre(X) -> true ; my_print_error(verify_pre_failed(X))),
	rt_timeout_call(X),
	/* print(post(X)),nl, */
	(verify_post(X) -> true ; my_print_error(verify_post_failed(X))).

verify_pre(Call) :-
  (run_time_type_check_mode(off) -> true
   ;
	(pre_condition(Call,Pre)
	->( (\+(rt_timeout_call(Pre))
		-> (get_predicate_arity(Call,Pred,Arity),
		    add_error(verify_pre,'### PRE-CONDITION ERROR OCCURED: ',Pred/Arity),
	   	    add_error(verify_pre,'### CALL: ',Call),nl,
            prepost_user_interaction
		   )
		;  (true)
	     )
	  )
	; (get_predicate_arity(Call,Pred,Arity),
	   add_error(verify_pre,'### No PRE-CONDITION for ',Pred/Arity),
	   print_term_summary(Call),
	   prepost_user_interaction )
	)).

verify_post(Call) :-
  (run_time_type_check_mode(off) -> true
   ;
	(post_condition(Call,Post)
	->( (\+(rt_timeout_call(Post))
		-> (get_predicate_arity(Call,Pred,Arity),
		    add_error(verify_post,'### POST-CONDITION ERROR OCCURED: ',Pred/Arity),
	   	    add_error(verify_post,'### CALL: ',Call),nl,
                    prepost_user_interaction
		   )
		;  (true)
	     )
	  )
	; (get_predicate_arity(Call,Pred,Arity),
	   add_error(verify_post,'### No POST-CONDITION for ',Pred/Arity),
	   prepost_user_interaction )
	)).


prepost_user_interaction :- prepost_no_error_so_far,!,
       my_print_error('### => Stop at next error (y/n/halt) => '),
       read(Answer),
       ((Answer='y') -> true ; (Answer='halt' -> halt ; retract(prepost_no_error_so_far))).
prepost_user_interaction.


get_predicate_arity(':'(Module,Call),Res,Arity) :- !,Res=Module:Pred,
   functor(Call,Pred,Arity).
get_predicate_arity(Call,Pred,Arity) :-
   functor(Call,Pred,Arity).
   
/* ===================================================== */

prepost_mnf_call(X) :- % print(prepost_mnf_call(X)),nl,
  (run_time_type_check_mode(off)
   -> call(X)
   ;  if(prepost_call(X),true,
        (add_error(mnf,'### WARNING CALL HAS FAILED: ',X),
	     print_call(X),
         prepost_user_interaction,
         fail)
       )
  ).
	
mnf_call(X) :-
  (run_time_type_check_mode(off)
   -> call(X)
   ;  if(call(X),true,
        (add_error(mnf,'### WARNING CALL HAS FAILED: ',X),
	     print_call(X),
         prepost_user_interaction,
         fail)
       )
  ).
	
print_call(X) :-
	     print('### Call: '),print_quoted_with_max_depth(X,10),nl,
	     print('### Summary: '),print_term_summary(X),nl,
	     print('### Predicate: '), 
	     (X = ':'(M,C) -> (print(M),print(':')) ; C=X),
	     functor(C,F,N),
	     print(F), print('/'), print(N),nl.
 
	
mnf_call_with_pp(ProgramPoint,X) :-
  (run_time_type_check_mode(off)
   -> call(X)
   ;  if(rt_timeout_call(X),true,
        (add_error(mnf,'### WARNING CALL HAS FAILED: ',(X,ProgramPoint)),
	     print_call(X),
	     print('### at program point:'),print(ProgramPoint),nl,
         prepost_user_interaction,
         fail)
       )
  ).

:- volatile found_det_sol/4.
:- dynamic found_det_sol/4.

mnf_det(X) :-
  (run_time_type_check_mode(off)
   -> call(X)
   ;  (get_functor_module(X,Module,F,Arity),
       retractall(found_det_sol(F,Arity,Module,_)),  /* cannot be nested for same predicate !*/
       copy_term(X,CopyX),
       mnf_call(X),
       (found_det_sol(F,Arity,Module,Previous) ->
          (add_error(mnf,'### WARNING CALL HAS MULTIPLE SOLUTIONS: ',CopyX),nl,
	       print('### Solution 1: '),print_quoted_with_max_depth(Previous,60),nl,nl,
	       print('### Solution 2: '),print_quoted_with_max_depth(X,60),nl,nl,
           prepost_user_interaction)
         ; assert(found_det_sol(F,Arity,Module,X))
       )
      )
  ).
  
  % det_call(';'(X=a,X=b)).
get_functor_module(Module:Call,Module,F,Arity) :- !, functor(Call,F,Arity).
get_functor_module(Call,unknown,F,Arity) :- functor(Call,F,Arity).
  
det_call(X) :- 
  (run_time_type_check_mode(off)
   -> call(X)
   ;  force_det_call(X)
  ).
  
:- meta_predicate residue_check_call(-).
residue_check_call(X) :-
    call_residue(X,CallResidue),
        ((CallResidue=[]) -> true
          ;  (add_error(must_succeed_without_residue,'### Call has residue: ',X),
              add_error(must_succeed_without_residue,'### Residue: ',CallResidue))
        ).

force_det_call(X) :- 
       get_functor_module(X,Module,F,Arity),
       retractall(found_det_sol(F,Arity,Module,_)),  /* cannot be nested for same predicate !*/
       copy_term(X,CopyX),
       %print(det_call(X)),nl,
       residue_check_call(X), %  was rt_timeout_call(X)
       (found_det_sol(F,Arity,Module,Previous) ->
          (add_error(mnf,'### WARNING CALL HAS MULTIPLE SOLUTIONS: ',CopyX),nl,
	       print('### Solution 1: '),print_quoted_with_max_depth(Previous,60),nl,nl,
	       print('### Solution 2: '),print_quoted_with_max_depth(X,60),nl,nl,
           prepost_user_interaction)
         ; assert(found_det_sol(F,Arity,Module,X))
       ).

/* ===================================================== */

:- meta_predicate must_fail(0).
must_fail(X) :-
	copy_term(X,Y),
	rt_timeout_call(X),!,
	(safe_numbervars(Y,0,_) -> true ; true),
	add_error(must_fail,'### Self-Check Failed !!! Call succeeded: ',Y),
	my_print_error('### The call: '),
	my_print_error(Y),
	my_print_error('### should have failed but succeeded with:'),
	my_print_error(X).
must_fail(_X).


:- meta_predicate must_succeed(0).

must_succeed(X) :-  %print(must_suceed(X)),nl,
	\+ rt_timeout_call(X),!,
	(safe_numbervars(X,0,_) -> true ; true),
	add_error(must_succeed,'### Self-Check Failed !!! Call failed: ',X).
%	my_print_error('### The call: '),
%	my_print_error(X),
%	my_print_error('### should have succeeded but failed !').
must_succeed(_X). %  :- print(ok_must_succeed(_X)),nl.

:- volatile found_must_succeed_sol/1, found_id/1.
:- dynamic found_must_succeed_sol/1, found_id/1.
found_id(0).
get_found_id(Nr) :- retract(found_id(Nr)), N1 is Nr+1, assert(found_id(N1)).
reset_found_id :- retractall(found_must_succeed_sol(_)),
 retractall(found_id(_)), assert(found_id(0)).

:- meta_predicate must_succeed_without_residue(0).
must_succeed_without_residue(X) :- get_found_id(ID),
	must_succeed(residue_check_call(X)),
	retractall(found_must_succeed_sol(ID)),
	%residue_check_call(X),
	(found_must_succeed_sol(ID)
	  -> (add_error(must_succeed_without_residue,'### Self-Check has multiple solutions: ',X),
	      !)
	  ;  (assert(found_must_succeed_sol(ID)),fail)
	 ).
must_succeed_without_residue(_).

:- meta_predicate must_succeed_multiple_without_residue(0).
:- public must_succeed_multiple_without_residue/1. % used by assert_must_succeed_multiple
must_succeed_multiple_without_residue(X) :- get_found_id(ID),
	must_succeed(X),
	retractall(found_must_succeed_sol(ID)),
	call_residue(X,CallResidue),
	((CallResidue=[]) -> true
	  ;  (add_error(must_succeed_multiple_without_residue,'### Self-Check has residue: ',X),
	      add_error(must_succeed_multiple_without_residue,'### Residue: ',CallResidue))
	),
	(found_must_succeed_sol(ID)
	  -> (true,!)
	  ;  (assert(found_must_succeed_sol(ID)),fail)
	 ).
must_succeed_multiple_without_residue(X) :- 
    add_error(must_succeed_multiple_without_residue,
              '### Self-Check did not succeed multiple times: ',X).


:- meta_predicate safe_call(0).

safe_call(X) :- on_exception(Exception,
                call(X),
                (print(exception(X,Exception)),nl,nl,
                 add_error(safe_call,'### Exception occured during self-check: ',X:Exception))).

:- use_module(preferences).
:- dynamic starttime/1.

get_module_list(ML) :- findall(Module,self_check_module(Module),Modules), sort(Modules,ML).

perform_self_check :- perform_self_check(_).





check_module(Module,Results) :- 
   findall(testcase(Module,X), self_check(X,Module), Calls),
   maplist(check_test_case, Calls, Results).

check_test_case(testcase(Module,Call), Verdict) :-
   set_error_context(self_check(Module,Call)),
   statistics(runtime,_),
   print('.'),
   safe_call(Call),
   flush_output,
   statistics(runtime,[_,T]),
   (T>20 -> print('*'),print(T),print('*') ; true),
   (T>2000 -> nl, print(Module), print(' : '), print(Call),nl ; true),
   (junit_mode(_) -> (get_all_errors_and_reset(Errors) -> V=error(Errors)
                                           ; otherwise -> V=pass)
   ; otherwise -> true),
   create_junit_result(Call,'Selfcheck', Module, T, V, Verdict).

perform_self_check(Module) :- nl, reset_found_id,
    ( nonvar(Module) -> true;
           otherwise -> self_check_module(Module)),
   print_message('---------------------'),
   print_message('---------------------'),
   print_message(('Performing Self-Check for Module ':Module)),
   print_message('---------------------'),
   backup_preferences,
   print_message('Setting minint=-1, maxint=3'),
   set_preference(minint,-1),
   set_preference(maxint,3),
   kernel_waitflags:set_silent(true),
   flush_output,
   print('% '),
   retractall(starttime(_)),
   statistics(runtime,[Start,_]), assert(starttime(Start)),
   check_module(Module,TestResults),
   print_junit(TestResults,Module),
   fail.
perform_self_check(_) :- 
   statistics(runtime,[End,_]),
   starttime(Start), Tot is End-Start,
   nl, print('Time: '), print(Tot), print(' ms'),
   nl, revert_preferences, clear_error_context,
   kernel_waitflags:set_silent(false),
   (error_manager:test_error_occured(_)
     -> print_message('Self-Check FAILED !!!'),fail
     ;  true),
   print_message('Self-Check Successful'),
   print_message('-------------------'), 
   flush_output.
   
   
perform_verbose_self_check :- perform_verbose_self_check(_).

perform_verbose_self_check(Module) :- nl, reset_found_id,
( nonvar(Module) -> true;
          otherwise -> self_check_module(Module)),
  print_message('---------------------'),
  print_message(('Performing Self-Check for Module ',Module)),
  flush_output,
   self_check(X,Module),
   print(X), flush_output,
   set_error_context(self_check(Module,X)),
   statistics(runtime,_),
   safe_call(X),
   statistics(runtime,[_,T]),
   print(' ---> '), print(T), print(' ms'),nl,
   ((T>20) -> (print_message(warning,'### LONG !'),nl) ; true),
   fail.

perform_verbose_self_check(_) :- nl,
   (error_manager:test_error_occured(_)
     -> print_message('Self-Check FAILED !!!'),fail
     ;  true),
   print_message('Self-Check Successful'),
   print_message('---------------------'),
   flush_output.

display_self_checks :- display_self_checks(_).

/* self_check:display_self_checks */
display_self_checks(Module) :- 
( nonvar(Module) -> true;
          otherwise -> self_check_module(Module)),
  print_message('---------------------'),
  print_message(('Performing Self-Check for Module ',Module)),
flush_output,
   self_check(X,Module),
   print_message(X),
   fail.
display_self_checks(_) :- nl,
   print_message('---------------------'), flush_output.

% ------------------------

:- dynamic det_counter/1.
det_counter(0).
:- meta_predicate check_deterministic(0).
check_deterministic(Call) :-
   retract(det_counter(X)), X1 is X+1, assert(det_counter(X1)),
   check_det2(Call,X).
:- meta_predicate check_det(0).
check_det(Call) :- check_deterministic(Call).

:- volatile calling/1.
:- dynamic calling/1.
check_det2(Call,X) :- assert(calling(X)),
  call(Call),
  (retract(calling(X)) -> true
    ; nl,
      print('### Call has multiple solutions: '),nl,
      print('### '), print(Call),nl,
      nl
   ).
check_det2(Call,X) :- (retract(calling(X)) -> print(fails(Call)),nl ; fail).
