% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

% disable clauses are not together warning, because
% cli_testcase/4 and cli_testcase/5 are sorted by id
:- set_prolog_flag(discontiguous_warnings,off).

:- ensure_loaded(prob_cli).

:- module(test_runner, [
                        test_repl/0,
                        run_last_test/0, last/0,
                        run_tests_by_id/1, id/1, ids/1,
                        run_tests_all/0, run_tests_all_except/1,
                        run_tests_by_category/1, category/1,
                        makefile_by_category/2, copy/1,
                        full_flush/0,
                        sanity_check/0,
                        verbose/0, vv/0,
                        cache/0 % enable caching
                       ]).

:- use_module(library(system)).
:- use_module(library(lists)).

:- use_module(junit_tests,[print_junit/2,set_junit_dir/1,create_junit_result/6]).
:- use_module(testcases).
:- use_module(tools).
:- use_module(system_call).

:- use_module(module_information).
:- module_info(group,testing).
:- module_info(description,'This module runs the tests stored in testcases.pl.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- dynamic test_started/1.

prob_junit(PROB_JUNIT) :-
    environ('PROB_JUNIT_DIR', Var), !,
    PROB_JUNIT = ['-junit', Var].
prob_junit([]).

dontstop(Var) :-
    environ('DONTSTOP', Var2), !, Var=Var2.
dontstop(false).

halt1_allowed :- \+ repl_mode,
   prob_junit([]), dontstop(false).

ignore_timeouts(Var) :-
    environ('IGNORE_TIMEOUTS', Var2), !, Var=Var2.
ignore_timeouts(false).

:- use_module(library(file_systems)).
% call updir from within Spider, when you are in the src subdirectory
set_cur_dir :- (directory_exists('./examples') -> true /* we seem to be in the right directory */
          ; directory_exists('./cia') -> current_directory(_Old,'./..')
          ; print('Please start tests from ProB directory or ProB/src directory'),nl),
   print('Current directory: '),
   current_directory(CD), print(CD),nl.

reset_test_runner :- set_cur_dir, retractall(test_failed(_)).

ids(Nr) :- show_command(Nr).
show_command(Nr) :- filtered_cli_testcase(Nr,_,_,L,Desc),
  print('# '), print(Desc),nl, l_pr([probcli|L]).
l_pr([]) :- nl.
l_pr([H|T]) :- print(' '),print(H), l_pr(T).


sanity_check :-
   filtered_cli_testcase(Id,Categories,_,Args,_Comm1),
   \+ member('-strict',Args),
   print('*** test does not use -strict option: '), print_test(Id,Categories),nl,fail. 
sanity_check :-
   cli_testcase_diff_check_output(Id,_F1,_F2),
   (filtered_cli_testcase(Id,Categories,_I1,_A1,_Comm1)
     ->  (Categories = [_] -> true
           ; print('*** multiple categories for test with output file: '),
             print_test(Id,Categories),
             nl  /* can generate race conditions when run in parallel */
          )
     ;   print('*** no cli_testcase for id: '), print(Id),nl
    ),fail.
sanity_check.    

print_test(Id,Comment) :- print(Id),print(' : '), print(Comment).

% run the latest test (with highest id)
last :- run_last_test.
run_last_test :- get_last_test_id(Id),
  run_tests_by_id(Id).

get_last_test_id(Id) :- set_last_test_id, last_test_id(Id).

set_last_test_id :-
   retractall(last_test_id(_)), assert(last_test_id(-1)),
   cli_testcase(Id,C1,I1,A1,Comm1),
   ((cli_testcase(Id,C2,I2,A2,Comm2),(C1,I1,A1,Comm1) \= (C2,I2,A2,Comm2))
     -> nl,print('*** multiple entries for test id: '), print(Id),nl
     ; true),
   update_last(Id),
   fail.
set_last_test_id.

:- volatile last_test_id/1.
:- dynamic last_test_id/1.
update_last(Id) :- number(Id),!, % debug_print(9,Id), debug_print(9,' '),
   retract(last_test_id(L)),
   (Id>L -> Max = Id ; Max = L),
   assert(last_test_id(Max)).
update_last(Id) :- nl,print('*** test id not a number: '), print(Id),nl.

% run a test with a specific id
id(X) :- run_tests_by_id(X).
run_tests_by_id(X) :- reset_test_runner,
  run_tests_by_id_aux(X),
  check_failed.
run_tests_by_id_aux([]).
run_tests_by_id_aux(From-To) :- !,
    (From>To -> true
     ; run_tests_by_id_aux(From), F1 is From+1,
	   run_tests_by_id_aux(F1-To)).
run_tests_by_id_aux([Id | Ids]) :- !,
    run_tests_by_id_aux(Id),
	run_tests_by_id_aux(Ids).
run_tests_by_id_aux(Id) :-
	filtered_cli_testcase(Id,TestCategories,TestInfos,Arguments,Comment), !,
	run_list([testcase(Id,TestCategories,TestInfos,Arguments,Comment)]).
run_tests_by_id_aux(Id) :-
	\+(filtered_cli_testcase(Id,_Category,_Infos,_Arguments,_Comment)), !,
	format(user_output,'*** No testcase with ID ~w found~n', [Id]).

run_tests_all :- run_tests_all_except([]).

:- use_module(library(ordsets)).
run_tests_all_except(Categories) :- reset_test_runner,
    sort(Categories,SC),
    cli_testcase(Id,TestCategories,Infos,Arguments,Comment), sort(TestCategories,TC),
    \+ ord_intersect(TC,SC),
    run_list([testcase(Id,TestCategories,Infos,Arguments,Comment)]),
    fail.
run_tests_all_except(_) :- check_failed.

% run all tests with a specific category
category(X) :- run_tests_by_category(X).
run_tests_by_category(X) :- reset_test_runner,
  run_tests_by_category_aux(X),
  check_failed.
run_tests_by_category_aux([]).
run_tests_by_category_aux([Cat | Cats]) :-
	run_tests_by_category_aux(Cat),
	run_tests_by_category_aux(Cats).
run_tests_by_category_aux(Cat) :-
	findall(testcase(Id, _Category, Infos, Arguments, Comment),
	  (filtered_cli_testcase(Id, TestCategories, Infos, Arguments, Comment),  member(Cat, TestCategories)), List),
	(List=[] -> format(user_output,'*** No testcase with category ~w found~n', [Cat])
	 ; run_list(List)).

% Generate a Makefile for all listed Categories
% example: makefile_by_category('AlstomMakefile',[animate,history,sptxt,rel_fnc])
% the Makefile can be run to perform the listed tests on a compiled version of probcli
makefile_by_category(File, Categories) :-
    (is_list(Categories) -> Cats = Categories ; Cats = [Categories]),
    open(File, write, Stream),
    format(Stream, 'test_all:', []),
    makefile_write_categories(Stream, Cats),
    makefile_by_category_aux(Stream, Cats),
    close(Stream).
makefile_by_category_aux(_Stream, []).
makefile_by_category_aux(Stream, [Cat | Cats]) :-
    makefile_by_category_single(Stream, Cat),
    makefile_by_category_aux(Stream, Cats).
makefile_by_category_single(Stream, Cat) :-
    findall(testcase(Id, _Category, Infos, Arguments, Comment),
    (filtered_cli_testcase(Id, TestCategories, Infos, Arguments, Comment),  member(Cat, TestCategories)), List),
    (List=[] -> format(user_output,'*** No testcase with category ~w found~n', [Cat])
             ;  format(Stream, '~n~w:~n', [Cat]), makefile_write_calls(Stream, List)).

makefile_write_categories(Stream, []) :-
    format(Stream, '~n', []).
makefile_write_categories(Stream, [Cat | Cats]) :-
    format(Stream, ' ~a', [Cat]),
    makefile_write_categories(Stream, Cats).

makefile_write_calls(_Stream, []).
makefile_write_calls(Stream, [testcase(Id, _TestCategories, Arguments, Comment) | Tests]) :-
    (Comment='' -> true ; format(Stream,'\techo \"Test ~w : ~w\"~n',[Id,Comment])),
    format(Stream, '\t./probcli', []),
    makefile_write_arguments(Stream, Arguments),
    makefile_write_diff(Stream, Id),
    makefile_write_calls(Stream, Tests).

makefile_write_arguments(Stream, []) :-
    format(Stream, '~n', []).
makefile_write_arguments(Stream, [Arg | Args]) :-
    format(Stream, ' ~w', [Arg]),
    (quote_next_arg(Arg) -> makefile_write_arguments_quoted(Stream,Args)
      ; makefile_write_arguments(Stream, Args)).

quote_next_arg('-goal'). % this will contain spaces ,... surround in "" for shell

makefile_write_arguments_quoted(Stream, []) :-
    format(Stream, '~n', []).
makefile_write_arguments_quoted(Stream, [Arg | Args]) :-
    format(Stream, ' \"~w\"', [Arg]),
    makefile_write_arguments(Stream, Args).

makefile_write_diff(Stream, ID) :-
    cli_testcase_diff_check_output(ID, File1, File2),
    format(Stream, '\tdiff -b ~w ~w~n', [File1, File2]),
    fail.
makefile_write_diff(_Stream, _ID).

% -------------------------

copy(Cat) :- (Cat=[_|_] -> C=Cat ; C=[Cat]),
   generate_copy_commands(C,'testarchive/').

:- use_module(b_trace_checking,[get_default_trace_file/2]).
generate_copy_commands(Categories,Dest) :-
    filtered_cli_testcase(ID, TestCategories, _Infos, Arguments, _Comment),
    non_empty_inter(Categories,TestCategories), %print(ID),nl,
    Arguments=[MainFile|_], generate_copy_command(MainFile,Dest), % print(MainFile),nl,
    additional_testcase_file(ID,MainFile,Arguments,ExtraFile),
    generate_copy_command(ExtraFile,Dest),
    fail.
generate_copy_commands(_,_).

additional_testcase_file(ID,_,_,EFile) :- extra_testcase_file(ID,EFile).
additional_testcase_file(_ID,File,Arguments,TFile) :- member('-t',Arguments),
    get_default_trace_file(File,TFile).
additional_testcase_file(ID,_,_,RefFile2) :- cli_testcase_diff_check_output(ID,_File1,RefFile2).

non_empty_inter(A,B) :- member(X,A), member(X,B),!.

:- use_module(tools,[get_parent_directory/2,string_concatenate/3]).
generate_copy_command(File,Dest) :-
    file_exists(File),
    get_parent_directory(File,Dir),!,
    string_concatenate(Dest,Dir,DestDir),
    string_concatenate(Dest,File,DestFile),
    format(user_output,'\tmkdir -p ~w~n',[DestDir]),
    get_command_path(mkdir,MkCmdPath),
    system_call(MkCmdPath, ['-p',DestDir],_Text1,_JExit1),
    format(user_output,'\tcp ~w ~w~n',[File,DestFile]),
    get_command_path(cp,CpCmdPath),
    system_call(CpCmdPath, [File,DestFile],_Text2,_JExit2).
generate_copy_command(_,_).

% -------------------------

:- dynamic test_failed/1.

check_failed :- test_failed(X),!,print_failed_tests,
   print('Use the following command to run individual tests: '),nl,
   print('   sicstus -l src/test_runner.pl --goal "id('),print(X),print(')."'),nl,
   (halt1_allowed -> halt(1) ; print('halt(1) :: junit / dontstop mode'),nl).
check_failed :- print('All tests successful.'),nl.
print_failed_tests :- print('** Failed tests: '),
   test_failed(X), print(X), print(' '), lookup_test_description(X,Desc), print(Desc), nl, fail.
print_failed_tests :- nl.

lookup_test_description(Id,Desc) :- filtered_cli_testcase(Id,_,_,_,Desc).

run_list(List) :- maplist(run_single_testcase,List).


cache :- add_additional_arguments(['-cache','/Users/leuschel/svn_root/NewProB/examples/cache/']).
verbose :- add_additional_arguments(['-v']).
vv :- add_additional_arguments(['-vv']).

:- volatile additional_arguments/1.
:- dynamic additional_arguments/1.
add_additional_arguments(List) :-
  (retract(additional_arguments(Old)) -> true ; Old=[]),
  append(Old,List,New),
  assert(additional_arguments(New)).

skip(Infos) :-
    member(skip,Infos)
    ; member(conditional_skip(Callable), Infos), call(Callable).

run_single_testcase(testcase(Id,_Category,Infos,Arguments,_Comment)) :-
    full_flush,
    ( skip(Infos)
      -> format(user_output,'Skipping test ~w ~n', [Id]), print_junit_skip(Arguments)
      ;  format(user_output,'Running test ~w ~n',[Id]),
         print('testcase: probcli '), print_args(Arguments),nl,
         clear_diff_output(Id),
	 prob_junit(JUnit),
         append(Arguments, JUnit, Arguments1),
         (additional_arguments(ExtraArgs)
           -> append(ExtraArgs,Arguments1,Arguments2) ; Arguments1=Arguments2),
         (ignore_timeouts(true), modify_timeout(Arguments2,Arguments3) ; Arguments3 = Arguments2),
         print('executed: probcli '), print_args(Arguments3),nl,
	 full_flush,
	 statistics(runtime,[Start,_]),
	 retractall(test_started(_)), assert(test_started(Start)),
	 %catch(user:go_cli_with_junit_check(Arguments3), Exception, ExcOccured=true), !,
	 catch(user:go_cli(Arguments3), Exception, true), !,
	 (
	   Exception == halt(0)    -> check_diff_output(Id) ;
	   var(Exception)          -> check_diff_output(Id) ;
	   Exception = error(X,Y)  -> (halt1_allowed -> format(user_output,'error on test execution: error(~w,~w)',[X,Y]), halt(1) ; test_failed(Id,Exception)) ;
	   Exception == halt(1)    -> test_failed(Id, Exception) ;
	   otherwise               -> format(user_output,'Exception not caught in test_runner: ~w~n',[Exception]), halt(1)
	 )
    ), !,
    user:reset_cli,
    full_flush.

test_took(X) :-
    statistics(runtime,[End,_]),
    test_started(Start),
    X is End - Start.

print_junit_skip(Arguments) :-
    prob_junit(['-junit', Dir])
    -> set_junit_dir(Dir), test_took(X),
       create_junit_result(Arguments,'Integration Tests','Integration_Test',X,skip,Result),
       print_junit([Result],'Integration_Test')
    ;  true.

% if the test expects a time_out error, the timeout is not expanded
% otherwise, timeout is increased to allow coverage analysis / junit / etc to finish
modify_timeout(OldOptions,OldOptions) :- sublist(OldOptions,['-expcterr','time_out'],_,_,_), !.
modify_timeout([],['-p','TIME_OUT','150000']).
modify_timeout(['-p','TIME_OUT',OLD|T],['-p','TIME_OUT',NewT|T]) :- 
   tools:arg_is_number(OLD,OLDT),
   % preferences:preference_default_value(time_out,DEFAULT),
   % OLDT < DEFAULT,!, % we explicitly set a TIME_OUT lower than the default value
   NewT is OLDT * 60. % 150000 is a factor of 60 higher than 2500
modify_timeout([H|T],[H|MT]) :- modify_timeout(T,MT).

full_flush :- flush_output(user_output), flush_output(user_error).

print_args([]).
print_args([H|T]) :- print(H), print(' '), print_args(T).

test_failed(Id,Msg) :-
  test_took(X),
  create_junit_result(Id,'Integration Tests','Integration_Test',X,error(['Test with Id ',Id,'failed:',Msg]),Result),
  print_junit([Result],'Integration_Test'),
  format(user_output,'*** Test ~w FAILED: ~w~n', [Id,Msg]),
  assert(test_failed(Id)).
test_failed(Id,Msg1,Msg2) :-
  test_took(X),
  create_junit_result(Id,'Integration Tests','Integration_Test',X,error(['Test with Id',Id,'failed:',Msg1,Msg2]),Result),
  print_junit([Result],'Integration_Test'),
  format(user_output,'*** Test ~w FAILED: ~w~w~n', [Id,Msg1,Msg2]),
  assert(test_failed(Id)).

diff_failed(Id,F1,F2,OutputText) :-
  test_took(X),
  atom_codes(OutputTextAtom,OutputText),
  ErrMsg = ['Diff for test with Id\n',Id,'\nfailed:\n','Output file\n',F1,'\ndoes not correspond to stored version\n',F2,'\nOutput of Diff:\n',OutputTextAtom],
  create_junit_result(Id,'Diff Checking','Diff_checking',X,
		      error(ErrMsg),
		      Result),
  print_junit([Result],'Diff_checking'),
  format(user_output,'*** Test ~w FAILED: Diff failed:~nOutput file ~w~ndoes not correspond to stored version~n~w~n', [Id,F1,F2]),
  format(user_output,'~s~n',[OutputText]),
  assert(test_failed(Id)).

:- use_module(library(file_systems)).

clear_diff_output(Id) :- % clear all files that should be generated
  cli_testcase_diff_check_output(Id,GeneratedFile,_StoredReferenceFile),
  file_exists(GeneratedFile),
  format(user_output,'% Deleting: ~w~n',[GeneratedFile]),
  delete_file(GeneratedFile),fail.
clear_diff_output(_).

check_diff_output(Id) :-
  findall(diff(Id,GeneratedFile,StoredReferenceFile),
          cli_testcase_diff_check_output(Id,GeneratedFile,StoredReferenceFile),
	  ListOfDiffsToCheck),
  maplist(check_diff_output2, ListOfDiffsToCheck).

check_diff_output2(diff(Id,GeneratedFile,StoredReferenceFile)) :-
  \+ file_exists(GeneratedFile)       -> test_failed(Id,'Output file does not exist:',GeneratedFile) ;
  \+ file_exists(StoredReferenceFile) -> test_failed(Id,'Stored file does not exist:',StoredReferenceFile) ;
  otherwise                           -> diff(Id,GeneratedFile,StoredReferenceFile).

diff(Id,F1,F2) :-
  format(user_output,'% Checking: diff ~w ~w~n',[F1,F2]),
  get_command_path(diff,CmdPath),
  print(calling(CmdPath)),nl,
  system_call(CmdPath,['-b',F1,F2],OutputText,_ErrText,JExit), % use -q for quiet
  ( JExit = exit(0) -> true ; diff_failed(Id,F1,F2,OutputText)).

filtered_cli_testcase(ID,CategoriesOut,Infos,Arguments,Comment) :-
    (environ('PARALLEL', 1)
    -> cli_testcase(ID,[H|_T],Infos,Arguments,Comment),
       CategoriesOut = [H]
    ;  cli_testcase(ID,CategoriesOut,Infos,Arguments,Comment)).
    
:- volatile repl_mode/0.
:- dynamic repl_mode/0.

% a minimal shell to execute tests:
test_repl :- assert(repl_mode), test_repl_loop, retractall(repl_mode).
test_repl_loop :- print('TEST ==> '),read(T), test_eval(T), !, test_repl_loop.
test_repl_loop.

test_eval(N) :- number(N),!, id(N).
test_eval(last) :- !, run_last_test.
test_eval(N-M) :- number(N), number(M),!, id(N-M).
test_eval(v) :- !,verbose.
test_eval(verbose) :- !,verbose.
test_eval(cache) :- !,print('Enabling cache'),nl,
    cache.
test_eval(sanity_check) :- !, sanity_check.
test_eval(sc) :- !, sanity_check.
test_eval(vv) :- !,vv.
test_eval(q) :- !,fail.
test_eval(x) :- !,halt.
test_eval(quit) :- !,fail.
test_eval(halt) :- !,fail.
test_eval(cat) :- !, print('Categories: '),
   findall(Cat,
	  (filtered_cli_testcase(_, TestCategories, _, _, _),  member(Cat, TestCategories)), List),
   sort(List,SL), print(SL),nl.
test_eval(cat(Category)) :- !,
   run_tests_by_category(Category).
test_eval(Category) :- valid_category(Category),!,
   run_tests_by_category(Category).
test_eval(_) :- print('*** Unknown command'), nl,
     print(' Commands:   Nr,  Nr-Nr,  last, q, x, v, vv '),nl.
    
valid_category(Cat) :-
   filtered_cli_testcase(_Id, TestCategories, _Infos, _Arguments, _Comment), 
   member(Cat, TestCategories).
