% (c) 2009-2013 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(value_persistance, [set_storage_directory/1,unset_storage_directory/0,
                              save_constants/1,load_constants/2,
                              cache_is_activated/0, clear_applicable_flag/0,
                              load_partial_constants/3,
                              initialise_operation_caching/0,
                              lookup_cached_transitions/3,
                              add_new_transitions_to_cache/1]).


:- use_module(library(lists)).
:- use_module(library(file_systems)).
:- use_module(library(fastrw)).

:- use_module(tools).
:- use_module(error_manager,       [add_error/2,add_error/3]).
:- use_module(state_space,         [transition/3,visited_expression/2,
                                    time_out_for_node/1,not_all_transitions_added/1,
                                    time_out_for_node/2,transition/4,
                                    transition_info/2,max_reached_for_node/1]).
:- use_module(succeed_max,         [max_reached/1]).
:- use_module(preferences).
:- use_module(b_machine_hierarchy).
:- use_module(specfile,            [animation_mode/1, animation_minor_mode/1,
                                    state_corresponds_to_initialised_b_machine/2]).
:- use_module(store,               [lookup_value/3]).
:- use_module(kernel_objects,      [equal_object/2]).
:- use_module(bmachine,            [b_get_machine_refinement_hierarchy/1,
                                    get_operation_info/2]).
%:- use_module(bmachine_structure,  [get_section/3]).
:- use_module(b_global_sets,       [is_b_global_constant/3,b_global_set/1]).
:- use_module(bsyntaxtree).
:- use_module(extension('probhash/probhash')).

:- use_module(module_information).
:- module_info(group,interpreter).
:- module_info(description,'This module is responsible for caching constants and operations on disk').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- dynamic storage_directory/1.
% storage_directory('/home/plagge/svn/alstomprob/examples/cache/').
:- volatile storage_directory/1.
:- dynamic is_applicable_internal/1.
:- volatile is_applicable_internal/1.

/**********************************************************************/
/* common for loading and saving: */
/**********************************************************************/

cache_is_activated :-
    storage_directory(_).

cache_is_applicable :-
    is_applicable_internal(Applicable),!,Applicable=true.
cache_is_applicable :-
    cache_is_activated,
    % b_mode,  % only for B models
    animation_mode(b),
    \+ animation_minor_mode(_),
    b_get_machine_refinement_hierarchy([_]), % only if there is no refinement - we cache no refinements
    preferences:get_preference(symmetry_mode,off), % with symmetry on: you cannot simply patch operation updates
    assert_once( is_applicable_internal(true) ),!.
cache_is_applicable :-
    assert_once( is_applicable_internal(false) ),fail.

set_storage_directory(Dir) :-
    unset_storage_directory,
    (absolute_file_name(Dir,AbsDir,[file_type(directory),access(exist),fileerrors(fail)])
     -> assert(storage_directory(AbsDir))
     ;  add_error(cache,'Could not set cache directory: ',Dir)).

unset_storage_directory :-
    retractall(storage_directory(_)),
    clear_applicable_flag.

% should be called if e.g. animation mode information changes (csp_and_b mode)
clear_applicable_flag :- 
    retractall(is_applicable_internal(_)).

%:- use_module(version, [revision/1]).
:- use_module(version,[version/4]).
get_revision_info(info(V1,V2,V3,V4,ValuePersistanceRevision)) :-
    % revision(Revision), % revision not used: it differs between probcli and ProB Tcl/Tk
    version(V1,V2,V3,V4),
    ValuePersistanceRevision = '$Rev$'.

collect_computation_parameters(Name,[machine_name(Name)
                                    ,prob_revision(Revision)
                                    %,machine_hash(Hash)
                                    ,constants_signature(ConstSig)
                                    |Prefs]) :-
    find_relevant_preferences(Prefs),
    get_revision_info(Revision),
    %machine_hash(Name,Hash),
    compute_constants_signature(Name,ConstSig).

/* returns a list of Pref/Value pairs for each preference that is regarded
   a relevant when computing values for constants. See is_relevant_preference/2
   below */
find_relevant_preferences(Prefs) :-
    findall(preference(Pref,Value), relevant_preference(Pref,Value), Unsorted),
    sort(Unsorted,Prefs).

relevant_preference(Pref,Value) :-
    preference_category(Pref,Category),
    ( is_relevant_preference(Pref,Category) -> true; fail), % just to introduce a local cut
    get_preference(Pref,Value).

/* is_relevant_preference(+Preference,+Category)
   is used to specify those preferencees whose values are stored in the constants file.
   If such a preference changes, the constants have to be recomputed */

is_relevant_preference(Pref,_) :- nonvar(Pref),
    irrelevant_preference(Pref),!,fail.
is_relevant_preference(_Pref,animation).
is_relevant_preference(_Pref,hidden).
is_relevant_preference(_Pref,advanced).

% number_animated_abstractions : only relevant for Event-B + if refinement
irrelevant_preference(expand_avl_upto). % only affects pretty-printing
irrelevant_preference(bool_expression_as_predicate). % only used inside REPL/eval
irrelevant_preference(use_large_jvm_for_parser).
irrelevant_preference(expand_forall_upto). % only relevant in predicate analysis
irrelevant_preference(time_out). % if time_out occurs: no re-use anyway
irrelevant_preference(kodkod_for_components) :- preference(try_kodkod_on_load,false).
irrelevant_preference(do_neginvariant_checking). % just indicates whether invariant evaluated doubly
irrelevant_preference(provide_trace_information). % just provides feedback; should not change values
irrelevant_preference(use_closure_expansion_memoization). % only influences performance
irrelevant_preference(warn_when_expanding_infinite_closures). % only generates extra warnings
irrelevant_preference(use_tk_custom_state_viewer).
irrelevant_preference(use_small_window).
irrelevant_preference(tk_show_source_pane).
irrelevant_preference(check_prolog_b_ast).
irrelevant_preference(use_font_size_for_columns).
irrelevant_preference(user_is_an_expert_with_accessto_source_distribution).
irrelevant_preference(default_to_runtime_type_checking_on_startup_for_expert).
irrelevant_preference(path_to_fuzz). %% ??
irrelevant_preference(path_to_bcomp).
irrelevant_preference(path_to_spin).
irrelevant_preference(path_to_text_editor).
irrelevant_preference(number_of_recent_documents).
irrelevant_preference(number_of_searched_patterns).
irrelevant_preference(number_of_replaced_patterns).
irrelevant_preference(number_of_eval_history_elements).
irrelevant_preference(number_of_eval_csp_history_elements).
irrelevant_preference(number_of_checked_ltl_formulas).
irrelevant_preference(machines_path).


compute_constants_signature(Name,constsig(Name,PropHash,Sublist)) :-
    properties_hash(Name,PropHash),
    findall(Sig,
            ( referenced_machine_with_constants_or_properties(Name,RefName),
                compute_constants_signature(RefName,Sig)),
            SubSigs),
    sort(SubSigs,Sublist).

referenced_machine_with_constants_or_properties(OrigName,RefName) :-
    referenced_machine(OrigName,RefName,_,_),
    machine_has_constants_or_properties_trans(RefName).

machine_has_constants_or_properties_trans(Name) :-
    machine_has_constants_or_properties(Name),!.
machine_has_constants_or_properties_trans(Name) :-
    referenced_machine(Name,Ref,_,_),
    machine_has_constants_or_properties_trans(Ref),!.
machine_has_constants_or_properties(Name) :-
    machine_identifiers(Name,_Params,Sets,_AVars,_CVars,AConsts,CConsts),
    ( Sets=[_|_] ; AConsts==[_|_] ; CConsts=[_|_]),!.

referenced_machine(SrcName,DstName,Type,Prefix) :-
    machine_references(SrcName,Refs),
    member(ref(Type,DstName,Prefix),Refs).

/**********************************************************************/
/* saving values of constants */
/**********************************************************************/
:- use_module(state_packing).

save_constants(MaxReached) :-
    cache_is_applicable,
    \+ constants_loaded_from_file, % we have just loaded the constants from file; no need to re-write them
    main_machine_name(Name), % this fails e.g. in Z mode
    !,
    ( save_constants2(Name,MaxReached) -> true
    ; otherwise ->
        add_error(value_persistance,'Storing constants failed.')).
save_constants(_MaxReached).

save_constants2(Name,MaxReached) :-
    statistics(runtime,[Start,_]),
    statistics(walltime,[WStart,_]),
    save_constants_for_machine(Name,MaxReached),
    statistics(runtime,[Stop,_]), Time is Stop - Start,
    statistics(walltime,[WStop,_]), WTime is WStop - WStart,
    %% tools:print_memory_used_wo_gc,
    debug_format('value caching: storing constants for ~w (maxr: ~w) (~w [~w] ms).~n',
           [Name,MaxReached,Time,WTime]).

save_constants_for_machine(Name,MaxReached) :-
    find_constants(Name,Constants),
    Constants = [_|_],   % at least one constant exists, check before computing findall !
    find_constants_stores(Stores),
    is_usable_data(Constants,Stores),!,
    debug_println(9,saving_constants(Name)),
    maplist(generate_bindings(Constants),Stores,ConstantBindings),
    debug_println(9,collect_computation_parameters(Name)),
    collect_computation_parameters(Name,CompParams),
    debug_println(9,save_values_into_file(Name)),
    save_values_into_file(Name,CompParams,ConstantBindings,MaxReached).
save_constants_for_machine(Name,_MaxReached) :-
    print('No constants stored for '),print(Name),nl.

find_constants(Name,Constants) :-
    machine_identifiers(Name,Params,_Sets,_AVars,_CVars,AbstractConstants,ConcreteConstants),
    ( Params=[_|_]
     -> add_warning(value_persistance,'Parameters not yet supported.'), % ok to just issue warning ?? or should we generate error ?
        fail
    ; otherwise -> true),
    append(AbstractConstants,ConcreteConstants,RawConstants),
    maplist(get_raw_identifier,RawConstants,Unsorted),
    sort(Unsorted,Constants).
get_raw_identifier(identifier(_Pos,Id),Id).

/* Returns the stores of values in the state space that contain values for contants */
find_constants_stores(Stores) :-
    findall(S,
            ( transition(root,Trans,I),
                functor(Trans,'$setup_constants',_),
                visited_expression(I,concrete_constants(S))
                ),  
            Stores).

/* is_usable_data/2 succeeds if there are constants that are ready to store */
is_usable_data(Constants,Stores) :-
    % at least one constant exists
    Constants = [_|_],
    % at least one state with computed constants was generated
    Stores = [_|_],
    % all constants have been computed
    \+ not_all_transitions_added(root),
    % no time-out in the root node
    \+ time_out_for_node(root).

%generate_bindings(Constants,Store,ConstantsBinding) :-
%    maplist(generate_binding2(Store),Constants,ConstantsBinding).
%generate_binding2(Store,Constant,bind(Constant,Value)) :-
%    lookup_value(Constant,Store,Value).
generate_bindings(_Constants,Store,ConstantsBinding) :-
    % TODO: review this
    Store = ConstantsBinding.

save_values_into_file(Name,CompParams,ConstantBindings,MaxReached) :-
    storage_file_name(constants,Name,Filename),
    open(Filename,write,S,[type(binary)]),
    call_cleanup(save_values_into_file2(S,CompParams,ConstantBindings,MaxReached),
                 close(S)).
save_values_into_file2(S,CompParams,ConstantBindings,MaxReached) :-
    fast_write(S,comp_parameters(CompParams)),
    fast_write(S,maximum_reached(MaxReached)),
    maplist(write_bind(S,constants),ConstantBindings),
    fast_write(S,end_of_file).
write_bind(S,Type,Store) :-
    preferences:get_preference(use_state_packing,PREF),
    preferences:set_preference(use_state_packing,false),
    state_packing:pack_state(Store,PS), % does not seem to bring major performance benefit; but reduces file size
     % print(packed(Store,PS)),nl,
    preferences:set_preference(use_state_packing,PREF),
    fast_write(S,values(Type,PS)).
 %   fast_write(S,values(Type)), fast_write_list(PS,S).  % should we do something like this to reduce memory consumption ?

%fast_write_list([],S) :- fast_write(S,end_of_list).
%fast_write_list('$bind_var'(Var,Val,T),S) :- fast_write(S,bind(Var,Val)), fast_write_list(T,S).

/* compute the full file name for the stored constants for the machine
   the predicate uses the specified directory where constants values
   should be stored */
storage_file_name(Type,MachineName,FullName) :-
    file_suffix(Type,Suffix),
    storage_directory(Directory),
    atom_concat(MachineName,Suffix,Name),
    atom_concat(Directory,Name,FullName).

file_suffix(constants,'.probcst').
file_suffix(operations,'.probops').
file_suffix(operations_data,'.opdata').


/**********************************************************************/
/* load constants                                                     */
/**********************************************************************/

/* load_constants/2 returns a list of stores if the constants have been
   saved before for the same computation parameters.
   The predicate fails if no constants can be loaded.
   MaxReached is 'true' if the maximum number of computed solutions was reached
   or 'false' otherwise.
*/
:- dynamic constants_loaded_from_file/0.
:- volatile constants_loaded_from_file/0.
load_constants(ConstantStores,MaxReached) :-
    cache_is_applicable, % check if storing/re-using constants is activated
    main_machine_name(Name),
    machine_file_exists(Name,FullName),
    open(FullName,read,S,[type(binary)]),
    call_cleanup(load_constants2(S,Name,ConstantStores,MaxReached),
                 close(S)),!.
load_constants2(S,Name,ConstantStores,MaxReached) :-
    load_parameters(S,StoredParameters),                    % if the relevant computation parameters
    collect_computation_parameters(Name,CurrentParameters), % are different, we cannot re-use the
    StoredParameters = CurrentParameters,                   % values stored in the file
    load_max_reached(S,MaxReached),
    load_constant_values(S,ConstantStores),
    debug_format('value caching: re-using previously computed constants.~n',[]),
    assert_once( constants_loaded_from_file ).

machine_file_exists(Name,FullName) :-
    cache_file_exists(constants,Name,FullName).
cache_file_exists(Type,Name,FullName) :-
    storage_file_name(Type,Name,FullName),
    file_exists(FullName).

load_parameters(S,StoredParameters) :-
    safe_constants_read(S,comp_parameters(StoredParameters)).

load_max_reached(S,MR) :-
    safe_constants_read(S,maximum_reached(MR)).

load_constant_values(S,Stores) :-
    safe_constants_read(S,Term),
    ( Term = end_of_file ->
        Stores = []
    ; Term = values(constants,Store) ->
    	preferences:get_preference(use_state_packing,PREF),
    	preferences:set_preference(use_state_packing,false),
        state_packing:unpack_state(Store,UPS),
        % print(unpacked(Store,UPS)),nl,
    	preferences:set_preference(use_state_packing,PREF),
        Stores = [UPS|Rest],
        load_constant_values(S,Rest)).

safe_constants_read(S,Term) :-
    read_syntax_safe(S,T,Error),
    ( Error==true ->
        write('Invalid cache file content. File skipped.\n'),
        fail
    ; otherwise ->
        Term = T).

load_partial_constants(ConstantsBindings,InProperties,OutProperties) :-
    % TODO: - check if parameters / constraints have been used (what are the implications?)
    cache_is_applicable, % check if storing/re-using constants is activated
    main_machine_name(Name),
    findall( partial(MName,NewBindings),
             try_to_reuse_referenced_machine(Name,MName,NewBindings),
             PSol),
    PSol = [_|_],!,
    remove_evaluated_properties(PSol,InProperties,OutProperties),
    select_and_apply_binding(PSol,ConstantsBindings).
load_partial_constants(_ConstantsBindings,Properties,Properties).

/**********************************************************************/

:- meta_predicate open_cache_file(+,+,-,0).
open_cache_file(Type,Name,Stream,Call) :-
    cache_file_exists(Type,Name,FullName),
    open(FullName,read,Stream,[type(binary)]),
    call_cleanup( (call(Call),!),close(Stream)).

/**********************************************************************/

try_to_reuse_referenced_machine(OrigName,Name,Bindings) :-
    referenced_machine_with_constants_or_properties(OrigName,RefName),
    try_to_reuse_referenced_machine2(RefName,Name,Bindings).
try_to_reuse_referenced_machine2(RefName,Name,Bindings) :-
    % try to re-use values of a machine directly referenced by OrigName that
    % contains constants
    open_cache_file(constants,RefName,S,try_load_values2(RefName,S,Bindings1)),!,
    RefName = Name,
    Bindings1 = Bindings.
try_to_reuse_referenced_machine2(RefName,Name,Bindings) :-
    % using the directly referenced machine failed, using a machine
    % that is referenced by the referenced machine
    try_to_reuse_referenced_machine(RefName,Name,Bindings).
try_load_values2(RefName,S,Bindings) :-
    check_if_machine_has_no_parameters(RefName),
    load_parameters(S,StoredParameters),
    check_if_all_constants_computed(RefName,S),
    collect_computation_parameters(RefName,CurrentParameters),
    compare_computation_parameters(StoredParameters,CurrentParameters),
    load_constant_values(S,Bindings),
    debug_format('value caching: re-using previously computed values of machine ~w.~n',[RefName]).

check_if_machine_has_no_parameters(RefName) :-
    machine_identifiers(RefName,Params,_Sets,_AVars,_CVars,_AConsts,_CConsts),
    ( Params=[] -> true
    ; otherwise ->
        format('value caching: constants of referenced machine ~w are not used because it uses parameters',[Params]),
        fail).

compare_computation_parameters([],[]) :- !.
compare_computation_parameters([CH|CT],[SH|ST]) :-
    !,compare_computation_parameters(CH,SH),
    compare_computation_parameters(CT,ST).
compare_computation_parameters(C,C) :- !.
compare_computation_parameters(C,S) :-
    format('parameter difference: current is ~w, stored is ~w~n',[C,S]),
    fail.

check_if_all_constants_computed(_RefName,S) :-
    % make sure that all solutions have been generated
    load_max_reached(S,false),!.
check_if_all_constants_computed(RefName,_S) :-
    format('value caching: constants for referenced machine ~w not used because not all solutions were computed.~n',[RefName]),
    fail.

select_and_apply_binding(Solutions,Store) :-
    maplist(select_solutions,Solutions,Bindings),
    maplist(select_and_apply_binding2(Store),Bindings).
select_and_apply_binding2(Store,Binding) :-
    maplist(apply_binding(Store),Binding).
apply_binding(Store,bind(Id,Value)) :-
    lookup_value(Id,Store,SValue),
    equal_object(SValue,Value).
select_solutions(partial(_Name,Bindings),Binding) :-
    member(Binding,Bindings).

:- use_module(bmachine,[b_filenumber/3]).
machine_file_number(Name,Nr) :- b_filenumber(Name,Ext,Nr),
  Ext \= 'def'.

:- use_module(debug).
% remove the properties from those machines whose constants have already been set from file
remove_evaluated_properties(PSol,InProperties,OutProperties) :-
    findall(Name,member(partial(Name,_),PSol),UnsortedNames),
    sort(UnsortedNames,Names),
    maplist(machine_file_number,Names,FileNumbers),
    debug_println(9,filtering(Names,FileNumbers)),
    conjunction_to_list(InProperties,InAsList),
    % exclude all properties who are in one of the files for which we have restored (all) constant values
    exclude(belongs_to_file(FileNumbers),InAsList,OutAsList), % this is exclude/3 from library(lists)
    conjunct_predicates(OutAsList,OutProperties).
belongs_to_file(FileNumbers,TPred) :-
    get_texpr_info(TPred,Info),
    memberchk(nodeid(pos(_,FilePos,_,_,_,_)),Info),
    memberchk(FilePos,FileNumbers).

/**********************************************************************/

:- dynamic loadable_operation/5, storable_operation/4, stored_transition/4, index_file_opened/1.
:- volatile loadable_operation/5, storable_operation/4, stored_transition/4, index_file_opened/1.

initialise_operation_caching :-
    clear_applicable_flag,
    cache_is_applicable,
    main_machine_name(MachName), % if we have no machine name, fail silently (for default machine)
    !,
    % check wich operations are cachable and prepare information for them
    debug_println(9,initialise_operation_caching),
    retract_all_op_cache_data,
    try_to_load_operations_cache_for_all_machines,
    get_operation_names(MachName,OperationNames),
    maplist(init_operation_cache(MachName),OperationNames),
    save_operations_cache_survey(MachName,OperationNames),
    open_data_file(MachName).
initialise_operation_caching :-
    retract_all_op_cache_data.

retract_all_op_cache_data :-
    retractall(constants_loaded_from_file),
    retractall(loadable_operation(_,_,_,_,_)),
    retractall(storable_operation(_,_,_,_)),
    retractall(stored_transition(_,_,_,_)),
    ( index_file_opened(S) ->
        catch( close(S), _, true ),
        retractall(index_file_opened(_))
    ; otherwise -> true).

get_operation_names(MachName,OperationNames) :-
    machine_operations(MachName,RawOperationNames),
    maplist(get_raw_identifier,RawOperationNames,OperationNames).

try_to_load_operations_cache_for_all_machines :-
    find_loadable_machines(LoadableMachines),
    maplist(try_to_load_operations_cache,LoadableMachines).

try_to_load_operations_cache(loadable(_Type,MachName)) :-
    get_operations_computation_parameters(MachName,OpParams),
    open_cache_file(operations,MachName,S,read_cached_operations(S,MachName,OpParams,Error)),!,
    ( Error==true ->
        format('value caching: corrupted index file for machine ~w, deleting file.~n',[MachName]),
        delete_cache_file(MachName)
    ; otherwise   -> true).
try_to_load_operations_cache(loadable(main,MachName)) :-
    % delete the data file if one exists -- only for main machine
    delete_cache_file(MachName),!.
try_to_load_operations_cache(_Loadable).

delete_cache_file(MachName) :-
    cache_file_exists(operations_data,MachName,Filename),!,
    delete_file(Filename).

read_cached_operations(S,MachName,CurrentOpParams,SyntaxError) :-
    read_syntax_safe(S,StoredOpParams,SyntaxError),
    ( SyntaxError==true -> true
    ; CurrentOpParams = StoredOpParams ->
        read_cached_operations2(S,MachName,SyntaxError)
    ; otherwise ->
        write('value caching: general computations parameters have changed, no re-use of stored operations\n'),
        debug_params(CurrentOpParams,StoredOpParams),
        fail).
read_cached_operations2(S,MachName,SyntaxError) :-
    read_syntax_safe(S,Term,SyntaxError),
    ( SyntaxError==true -> true
    ; Term == end_of_file -> true
    ; read_cached_operations3(Term,MachName) ->
        read_cached_operations2(S,MachName,SyntaxError)
    ; otherwise ->
        functor(Term,F,A),
        debug_format('value caching: unrecognised entry in cache file for ~w: ~w/~w~n',[MachName,F,A]),
        fail
    ).
debug_params(op_comp_parameters(A),op_comp_parameters(B)) :- !,
  debug_params(A,B).
debug_params(preference(P,A),preference(P,B)) :- !,
  (A=B -> true ; print(P), print(':'), print(A/B), print(' ')).
debug_params([],[]) :- !,nl.
debug_params([Cur|CT],[Stored|ST]) :- !,
  debug_params(Cur,Stored),
  debug_params(CT,ST).
debug_params(A,B) :- (A=B -> true ; print(A/B),nl).

debug_format(T,F) :- (debug_mode(on) -> format(T,F) ; true).

:- use_module(error_manager).
read_syntax_safe(S,Term,Error) :-
    catch( fast_read(S,Term1),
           error(E,_),
           ( E=syntax_error(_)         -> Error = true
           ; E=permission_error(_,_,_) -> Term1 = end_of_file
           ; E=consistency_error(_,_,_)
               -> add_error('Consistency error when reading from stream: ',E),
                  Error = true
           ; otherwise
               -> add_error('Unknown error when reading from stream: ',E),
                  throw(E))
         ),
    Term1 = Term.

read_cached_operations3(operation(OpName,StoredHash),MachName) :-
    ( operation_hash(MachName,OpName,CurrentHash) ->
        ( CurrentHash = StoredHash ->
            find_matching_operations(OpName,MachName,PrefixedOpNames),
            ( PrefixedOpNames = [] ->
                format('value caching: no matching internal operation found for ~w in ~w.~n',
                       [OpName,MachName])
            ; otherwise ->
                maplist(store_loadable_operation(OpName,MachName),PrefixedOpNames) )
        ; otherwise ->
            format('value caching: operation ~w has changed.~n',[OpName])
        )
    ; otherwise ->
        format('value caching: unrecognised operation ~w.~n',[OpName])
    ).
read_cached_operations3(transition(OpName,Hash,Index),MachName) :-
    ( loadable_operation(_,OpName,MachName,_InputPattern,_OutputPattern) ->
        compute_short_hash(Hash,Short),
        assert( stored_transition(Short,Hash,OpName,Index) )
    ; otherwise ->
        % ignore
        true
    ).

store_loadable_operation(_OpName,_MachName,PrefixedOpName) :-
    loadable_operation(PrefixedOpName,_,_,_,_),!.
store_loadable_operation(OpName,MachName,PrefixedOpName) :-
    find_operation_input_output(PrefixedOpName,Input,Output) ->
    create_output_pattern(Output,OutputPattern),
    lookup_pattern(Input,InputPattern),!,
    assert( loadable_operation(PrefixedOpName,OpName,MachName,InputPattern,OutputPattern) ).
store_loadable_operation(_OpName,_MachName,PrefixedOpName) :-
    format('value caching: unable to generate input/output pattern for operation ~w.~n',
           [PrefixedOpName]).

find_matching_operations(Operation,MachName,FullNames) :-
    findall(O, is_matching_operation(Operation,MachName,O), FullNames).
is_matching_operation(Operation,MachName,Operation) :-
    % there is an operation with exactly the same name
    operation_with_machine(Operation,MachName).
is_matching_operation(Operation,MachName,FullName) :-
    atom_concat('.',Operation,DottedOp),
    operation_with_machine(FullName,MachName),
    atom_concat(_,DottedOp,FullName).
operation_with_machine(OpName,MachName) :-
    get_operation_info(OpName,Infos),
    ( memberchk(origin(Origin),Infos) ->
        last(Origin,_/MachName)
    ; otherwise ->
        main_machine_name(MachName)).

create_output_pattern(Ids,Values/Bindings) :-
    maplist(create_output_pattern2,Ids,Values,Bindings).
create_output_pattern2(Id,Value,bind(Id,Value)).
            
init_operation_cache(MachName,OpName) :-
    storable_operation(OpName,MachName,_InputPattern,_Output),!. % ignore, already registered
init_operation_cache(MachName,OpName) :-
    ( find_operation_input_output(OpName,Input,Output) ->
        lookup_pattern(Input,InputPattern),
        debug_println(9,find_operation_input_output(OpName,Input,Output)),
        assert( storable_operation(OpName,MachName,InputPattern,Output) ),
        % Operations of the main machine are never prefixed
        create_output_pattern(Output,OutputPattern),
        assert( loadable_operation(OpName,OpName,MachName,InputPattern,OutputPattern) )
    ; otherwise ->
        format('value caching: unable to generate input/output pattern for operation ~w.~n',
               [OpName])
    ).

get_operations_computation_parameters(Name,op_comp_parameters(OpParams)) :-
    find_relevant_preferences(Prefs),
    get_revision_info(Revision),
    OpParams = [machine_name(Name),prob_revision(Revision)|Prefs].
    
lookup_cached_transitions(PrefixedOpName,InState,OutputBindings) :-
    cache_is_applicable,
    loadable_operation(PrefixedOpName,OpName,MachName,InputPattern,OutputPattern),
    sort(InState,SortedInState),
    % statistics(runtime,[Start,_]), statistics(walltime,[WStart,_]),
    hash_input_values(OpName,MachName,InputPattern,SortedInState,ShortHash,InputHash),
    stored_transition(ShortHash,InputHash,OpName,Index),!,
    debug_format('value caching: re-using values for operation ~w.~n',[OpName]),
    ( load_solutions(MachName,Index,Solutions) -> true
        %length(Solutions,Sn),
        %format('value caching: re-using ~w solutions for operation ~w of machine ~w.~n',
        %       [Sn,OpName,MachName])
    ; otherwise ->
        format('value caching: re-using values for operation ~w failed.~n',[OpName]),
        fail),
    %print(solution(Solutions)),nl, print(pattern(OutputPattern)),nl,
    % OutputPattern is of the form [V1,...]/[bind(v1,V1),...]
    maplist(create_output_binding(OutputPattern),Solutions,OutputBindings).
    %statistics(runtime,[Stop,_]), Time is Stop - Start,
    %statistics(walltime,[WStop,_]), WTime is WStop - WStart,
    %tools:print_memory_used_wo_gc,
    %format('caching time: ~w (~w)~n',[Time,WTime]).
load_solutions(MachName,Index,Solutions) :-
    storage_file_name(operations_data,MachName,FilenameData),
    open(FilenameData,read,S,[type(binary),reposition(true)]),
    call_cleanup((seek(S,Index,bof,_),
                  fast_read(S,Solutions)),
                 close(S)).
create_output_binding(Pattern,transition(Param,Result,New,Info),
                      transition(Param,Result,Update,Info)) :-
    copy_term(Pattern,New/Update).
    
select_values(Ids,State,Values) :- sort(State,SState),
   select_values_sorted(Ids,SState,Values).
% Note: as Identifiers are sorted then we can avoid quadratic complexity here and scan State only once for all Ids
select_values_sorted(Ids,SState,Values) :-
   (select_values_sorted_aux(Ids,SState,Values) -> true
      ; print('Lookup: '),print(Ids), nl,
        print('State:  '), maplist(get_bind_id,SState,SIds), print(SIds),nl,fail).
select_values_sorted_aux([],_,[]).
select_values_sorted_aux([Id|Irest],State,[Value|Vrest]) :- 
   ( ordered_select(State,Id,StoredValue,RestState) ->
       Value=StoredValue,
       select_values_sorted_aux(Irest,RestState,Vrest)
    ; otherwise ->
        ajoin(['Looking up "', Id, '" failed.'], Msg),
        ( maplist(get_bind_id,State,Ids) -> true
        ; otherwise -> Ids = '<<cannot print state ids>>'),
        add_error(value_persistance,Msg,Ids),
        fail
    ).
ordered_select([bind(Id,Value)|T],Id,Value,T).
ordered_select([_|T],Id,Value,R) :- ordered_select(T,Id,Value,R).
get_bind_id(bind(Id,_),Id).

hash_input_values(OpName,MachName,InputPattern,SortedInState,Short,Long) :-
    %%statistics(runtime,[Start,_]), statistics(walltime,[WStart,_]),
    ( hash_input_values2(OpName,MachName,InputPattern,SortedInState,Short,Long) ->
        true
          %, statistics(runtime,[Stop,_]), Time is Stop - Start,
          %  statistics(walltime,[WStop,_]), WTime is WStop - WStart,
          %  tools:print_memory_used_wo_gc,
          %  format('hashing time for ~w:~w = ~w (~w)~n',[OpName,MachName,Time,WTime])
    ; add_failed_call_error(hash_input_values2(OpName,MachName,InputPattern,SortedInState,Short,Long)),
        fail).

hash_input_values2(OpName,MachName,InputPattern,SortedInState,Short,Long) :-
    copy_term(InputPattern,lookup_pattern(SortedVars,SortedInputValues,_,InputValues)),
    select_values_sorted(SortedVars,SortedInState,SortedInputValues),
    raw_sha_hash(OpName/MachName/InputValues,Long),
    compute_short_hash(Long,Short).
compute_short_hash([A,B,C,D|_],Short) :-
    Short is A<<24 + B<<16 + C<<8 + D.

save_operations_cache_survey(MachName,OperationNames) :-
    get_operations_computation_parameters(MachName,OpParams),
    storage_file_name(operations,MachName,FileName),
    open(FileName,write,S,[type(binary)]),
    fast_write(S,OpParams),
    save_operation_info(S,MachName,OperationNames),
    close(S).

save_operation_info(S,MachName,OperationNames) :-
    member(OpName,OperationNames),
    storable_operation(OpName,MachName,_InputPattern,_Output),
    operation_hash(MachName,OpName,OpHash),
    fast_write(S,operation(OpName,OpHash)),
    stored_transition(_Short,InputHash,OpName,Index),
    fast_write(S,transition(OpName,InputHash,Index)),
    fail.
save_operation_info(_S,_MachName,_OperationNames).

find_operation_input_output(OpName,Input,Output) :-
    get_operation_info(OpName,Info),
    memberchk(reads(Input1),Info),
    % remove references to global sets and their elements
    get_all_sets_and_enum_ids(SEIds),
    remove_all(Input1,SEIds,Input),
    memberchk(modifies(Output),Info).

get_all_sets_and_enum_ids(SEIds) :-
    findall(Id,
            ( is_b_global_constant(_,_,Id)
            ; b_global_set(Id)),
            SEIds).

add_new_transitions_to_cache(InStateId) :-
    cache_is_applicable,!,
    ( get_all_max_reached_operations(MaxReached),
        add_new_transitions_to_cache1(InStateId,MaxReached) ->
        true
    ; otherwise ->
        add_failed_call_error(add_new_transitions_to_cache1(InStateId,_MaxReached))).
add_new_transitions_to_cache(_InStateId).

get_all_max_reached_operations(MaxReached) :-
    findall(Op, max_reached(Op), MaxReached).

add_new_transitions_to_cache1(InStateId,MaxReached) :-
    visited_expression(InStateId,InState1),
    state_corresponds_to_initialised_b_machine(InState1,InState),
    main_machine_name(MachName),
    sort(InState,SortedInState),
    storable_operation(OpName,MachName,InputPattern,Output),
    \+ time_out_for_node(InStateId,OpName),
    nonmember(OpName,MaxReached),
    add_new_transitions_to_cache2(InStateId,SortedInState,OpName,InputPattern,Output,MachName),
    fail.
add_new_transitions_to_cache1(_InStateId,_MaxReached) :-
    flush_data.

add_new_transitions_to_cache2(InStateId,SortedInState,OpName,InputPattern,Output,MachName) :-
    hash_input_values(OpName,MachName,InputPattern,SortedInState,ShortHash,InputHash),
    \+ stored_transition(ShortHash,InputHash,_,_), % Is there already a cache entry?
    find_all_transitions(InStateId,OpName,Output,Transitions),
    length(Transitions,Length),
    %format('value caching: storing ~w transitions for operation ~w in state ~w.~n',
    %       [Length,OpName,InStateId]),
    %print(Transitions),nl,
    statistics(runtime,[Start,_]),
    statistics(walltime,[WStart,_]),
    store_transition_into_cache(OpName,ShortHash,InputHash,Transitions,MachName),
    statistics(runtime,[Stop,_]), Time is Stop - Start,
    statistics(walltime,[WStop,_]), WTime is WStop - WStart,
    %%tools:print_memory_used_wo_gc,
    debug_format('value caching: storing ~w transitions for operation ~w in state ~w (~w [~w] ms).~n',
           [Length,OpName,InStateId,Time,WTime]).   

store_transition_into_cache(OpName,ShortHash,Hash,Transitions,MachName) :-
    storage_file_name(operations_data,MachName,FilenameData),
    open(FilenameData,read,SR,[type(binary),reposition(true)]),
    call_cleanup(seek(SR,0,eof,Index),
                 close(SR)),
    open(FilenameData,append,SW,[type(binary)]),
    call_cleanup(fast_write(SW,Transitions),
                 close(SW)),
    assert( stored_transition(ShortHash,Hash,OpName,Index) ),
    index_file_opened(OI),
    fast_write(OI,transition(OpName,Hash,Index)).

find_all_transitions(InStateId,OpName,Output,Transitions) :-
    findall( transition(ParaValues,ResultValues,Updates,TransInfo),
             find_transition(InStateId,OpName,Output,ParaValues,ResultValues,Updates,TransInfo),
             Transitions).
find_transition(InStateId,OpName,Output,ParaValues,ResultValues,Updates,TransInfo) :-
    transition(InStateId,Operation,TransId,OutStateId),
    operation_name(Operation,OpName,ParaValues,ResultValues),
    findall(TI,transition_info(TransId,TI),TransInfo),
    visited_expression(OutStateId,OutState1), % we could use packed_visited_expression here ?
    state_corresponds_to_initialised_b_machine(OutState1,OutState),
    select_values(Output,OutState,Updates).
operation_name('-->'(Operation,ResultValues),OpName,ParaValues,ResultValues) :-
    !,operation_name(Operation,OpName,ParaValues,_).
operation_name(Operation,OpName,ParaValues,[]) :-
    Operation =.. [OpName|ParaValues].

open_data_file(Name) :-
    ( index_file_opened(S) ->
        close(S),retractall(index_file_opened(_))
    ; otherwise -> true ),
    storage_file_name(operations,Name,FilenameIndex),
    open(FilenameIndex,append,FI,[type(binary)]),
    assert(index_file_opened(FI)),
    storage_file_name(operations_data,Name,FilenameData),
    ( file_exists(FilenameData) -> true
    ; otherwise ->
        open(FilenameData,write,SD,[type(binary)]),
        close(SD)).

flush_data :-
    index_file_opened(S),
    flush_output(S).

find_loadable_machines(Solutions) :-
    findall(loadable(Type,Machine),
            find_loadable_machines2(Type,Machine),
            Solutions).
find_loadable_machines2(Type,Machine) :-
    main_machine_name(Main),
    find_loadable_machines_sees_includes(Main,Type,Machine).
find_loadable_machines_sees_includes(Main,main,Main).
find_loadable_machines_sees_includes(Main,sub,Seen) :-
    machine_sees(Main,Seen).
find_loadable_machines_sees_includes(Start,sub,Included) :-
    find_loadable_machines_includes(Start,Included).
find_loadable_machines_includes(Start,Included) :-
    machine_includes(Start,_Prefix1,M),
    ( Included=M
    ; find_loadable_machines_includes(M,Included)).

machine_sees(MachA,MachB) :-
    machine_references(MachA,References),
    member(ref(Type,MachB,''),References),
    memberchk(Type,[sees,uses]).
machine_includes(MachA,Prefix,MachB) :-
    machine_references(MachA,References),
    member(ref(Type,MachB,Prefix),References),
    memberchk(Type,[includes,extends]).

% lookup_pattern/2:
% To enable a fast lookup of values in a state, we have to sort the variables.
% On the other hand, we must not sort the variables when storing them, their order
% is significant.
% This predicate generates a "lookup_pattern" that sorts the variables and their
% corresponding values and enables to map the sorted values to the original order
% by a simple unification.
lookup_pattern(Variables,lookup_pattern(SortedVariables,SortedValues,Variables,Values)) :-
    maplist(variable_value_pair,Variables,VarValues,Values),
    sort(VarValues,SortedVarValues),
    maplist(variable_value_pair,SortedVariables,SortedVarValues,SortedValues).
variable_value_pair(I,I/V,V).
