:- module(weakest_preconditions, [weakest_precondition/3]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,cbc).
:- module_info(description,'This module provides weakest precondition calculation for B and Event-B operations.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

:- use_module(library(lists)).
:- use_module(library(sets), [subset/2]).

:- use_module(bsyntaxtree, [get_texpr_expr/2,
			    get_texpr_id/2, get_texpr_ids/2,
			    get_texpr_type/2,
			    create_texpr/4,
			    create_forall/3,
    			    conjunct_predicates/2,
			    replace_id_by_expr/4,
                            find_typed_identifier_uses/3]).
:- use_module(b_state_model_check, [negate/2]).
:- use_module(btypechecker, [get_accessed_vars/4]).
:- use_module(bmachine_eventb, [prime_variables/2]).


weakest_precondition(Action,Pred,WP) :-
    get_texpr_expr(Action,Expr),
    weakest_precondition2(Expr,Pred,WP).

% TODO: while
% any statement. see schneider p. 126
weakest_precondition2(any(Parameters,PreCond,Body),Pred,WP) :- !,
    weakest_precondition(Body,Pred,InnerWP),
    create_texpr(implication(PreCond,InnerWP),pred,[],UnquantifiedWP),
    create_forall(Parameters,UnquantifiedWP,WP).
weakest_precondition2(assign(TIds,Exprs),Pred,WP) :- !,
    function_assignment_to_overwrite(TIds,Exprs,TIds2,Exprs2),
    get_texpr_ids(TIds2,Ids),
    replace_all_id_by_expr(Ids,Exprs2,Pred,WP).
weakest_precondition2(assign_single_id(TId,Expr),Pred,WP) :- !,
    function_assignment_to_overwrite([TId],[Expr],[TId2],[Expr2]),
    get_texpr_id(TId2,Id),
    replace_id_by_expr(Pred,Id,Expr2,WP).
% nondeterministic choice. see schneider p. 126
weakest_precondition2(becomes_element_of(IDs,Set),Pred,WP) :- !,
    create_members(Set,IDs,MemberOf),
    create_texpr(implication(MemberOf,Pred),pred,[],InnerWP),
    create_forall(IDs,InnerWP,WP).
% nondet choice of action. see schneider p. 129
weakest_precondition2(choice(Choices),Pred,WP) :- !,
    choice_wps(Choices,Pred,WPs),
    conjunct_predicates(WPs,WP).
weakest_precondition2(if(IfList),Pred,WP) :- !,
    if_wp(IfList,Pred,WP).
% let statement: local variable with value. see schneider p. 125
weakest_precondition2(let(Parameters,Defs,Body),Pred,WP) :- !,
    weakest_precondition(Body,Pred,InnerWP),
    create_texpr(implication(Defs,InnerWP),pred,[],UnquantifiedWP),
    create_forall(Parameters,UnquantifiedWP,WP).
weakest_precondition2(parallel(Actions),Pred,WP) :- !,
    merge_parallels(Actions,Action),
    weakest_precondition(Action,Pred,WP).
weakest_precondition2(precondition(Condition,Action),Pred,WP) :- !,
    weakest_precondition(Action,Pred,InnerWP),
    conjunct_predicates([Condition,InnerWP],WP).
% there are two kinds of selects: just whens and whens/else
% the else case just collects all guards and negates them
% see schneider p. 132
weakest_precondition2(select(Whens),Pred,WP) :- !,
    select_whens_wps(Whens,Pred,WPs),
    conjunct_predicates(WPs,WP).
weakest_precondition2(select(Whens,Else),Pred,WP) :- !,
    select_whens_wps(Whens,Pred,WPs),
    conjunct_predicates(WPs,WhensWP),
    % collect all guards and negate them
    % => precondition for else-case
    findall(PC,(member(SW,Whens),get_texpr_expr(SW,select_when(PC,_))), AllPreconds),
    maplist(negate,AllPreconds,NegAllPreconds),
    conjunct_predicates(NegAllPreconds,ElsePrecond),
    % imply [T_n]S
    weakest_precondition(Else,Pred,InnerWP),
    create_texpr(implication(ElsePrecond,InnerWP),pred,[],ElseWP),
    conjunct_predicates([WhensWP,ElseWP],WP).
weakest_precondition2(sequence(Actions),Pred,WP) :- !,
    sequence_wp(Actions,Pred,WP).
weakest_precondition2(skip,Pred,Pred) :- !.
weakest_precondition2(var(IDs,Subst),Pred,WP) :- !,
    weakest_precondition(Subst,Pred,InnerWP),
    create_forall(IDs,InnerWP,WP).
weakest_precondition2(while(P,S,I,V),R,WP) :- !,
    % [while P DO S INVARIANT I VARIANT V END] R
    % see abrial, b-book, p. 395:
    % there are five predicates. their conjunction is equal to the weakest precondition
    % first, create the list of all variables modified by the Stmt (used in forall)
    find_typed_identifier_uses(S,[],AllIdentifiersInStmt),
    get_accessed_vars(S,[],ModifiedVarsInStmt,_Reads),
    % the identifiers in ModifiedVarsInStmt are not typed. fetch the types from the list of all identifiers
    % TypedModifiedVarsInStmt = X
    get_texpr_ids(TypedModifiedVarsInStmt,ModifiedVarsInStmt),
    subset(TypedModifiedVarsInStmt,AllIdentifiersInStmt),

    ForAlls = TypedModifiedVarsInStmt,

    % first obligation: the invariant
    WP1 = I,
    
    % second obligation: forall x in X: (I & P => [S]I)
    conjunct_predicates([I,P],LHS2),
    weakest_precondition(S,I,RHS2),
    create_texpr(implication(LHS2,RHS2),pred,[],Inner2),
    create_forall(ForAlls,Inner2,WP2),

    % third obligation: forall x in X: (I => v in N)
    create_texpr(member(V,b(integer_set('NATURAL'),set(integer),[])),pred,[],RHS3),
    create_texpr(implication(I,RHS3),pred,[],Inner3),
    create_forall(ForAlls,Inner3,WP3),

    % fourth obligation: forall x in X: I & P => [gamma:=V][S](V<gamma)
    create_texpr(identifier('gamma\''),integer,[],Gamma),
    create_texpr(assign_single_id(Gamma,V),subst,[],AssignVariantToGamma),
    create_texpr(less(V,Gamma),pred,[],VLessGamma),
    weakest_precondition(S,VLessGamma,RHS4T),
    weakest_precondition(AssignVariantToGamma,RHS4T,RHS4),
    LHS4 = LHS2,
    create_texpr(implication(LHS4,RHS4),pred,[],Inner4),
    create_forall(ForAlls,Inner4,WP4),

    % fifth obligation: forall x in X: (I & not P) => R
    negate(P,NotP), conjunct_predicates([I,NotP],LHS5),
    create_texpr(implication(LHS5,R),pred,[],Inner5),
    create_forall(ForAlls,Inner5,WP5),

    % conjunct them all together
    conjunct_predicates([WP1,WP2,WP3,WP4,WP5],WP).
weakest_precondition2(Action,Pred,_WeakestPrecondition) :- !,
    add_error_fail(weakest_precondition,'weakest-precondition implementation missing: ',[Action,Pred]).

choice_wps([],_Pred,[]).
choice_wps([Choice|Choices],Pred,[ChoiceWP|WPs]) :-
    weakest_precondition(Choice,Pred,ChoiceWP),
    choice_wp(Choices,Pred,WPs).

% see schneider p. 38
% [if E then S else T end] P = (E => [S]P) & (not E => [T]P)
if_wp([IF|IFs],Pred,WP) :-
    get_texpr_expr(IF,if_elsif(Test,Body)),
    % if-part
    weakest_precondition(Body,Pred,InnerWP),
    create_texpr(implication(Test,InnerWP),pred,[],IfWP),
    % if there are further else/elseif parts, create pos
    (if_wp(IFs,Pred,SubWP)
    -> negate(Test,NegTest),
       create_texpr(implication(NegTest,SubWP),pred,[],ElseWP),
       conjunct_predicates([IfWP,ElseWP],WP)
    ;  WP = IfWP).

merge_parallels([X],X).
merge_parallels([A1,A2|As],X) :-
    merge_parallels2(A1,A2,AMerged), !,
    merge_parallels([AMerged|As],X).
merge_parallels([A1,A2|As],X) :-
    merge_parallels2(A2,A1,AMerged), !,
    merge_parallels([AMerged|As],X).
merge_parallels(A1,A2,_) :-
    add_error_fail(merge_parallels,'merging of parallel executions failed (implementation missing?): ',[A1,A2]).

merge_parallels2(A1,A2,AMerged) :-
    get_texpr_expr(A1,A1Expr), get_texpr_expr(A2,A2Expr),
    merge_parallels_normalize(A1Expr,A1ExprN),
    merge_parallels_normalize(A2Expr,A2ExprN),
    merge_parallels3(A1ExprN,A2ExprN,AMergedExpr),
    create_texpr(AMergedExpr,subst,[],AMerged).

merge_parallels3(assign(IDs1,Exprs1),assign(IDs2,Exprs2),
		 assign(IDs3,Exprs3)) :-
	append(IDs1,IDs2,IDs3), append(Exprs1,Exprs2,Exprs3).
merge_parallels3(any(Parameters,PreCond,Body),SomeSubstitution,any(Parameters,PreCond,NewBody)) :-
    create_texpr(SomeSubstitution,subst,[],SomeSubstitutionTExpr),
    create_texpr(parallel([Body,SomeSubstitutionTExpr]),subst,[],NewBody).

merge_parallels_normalize(becomes_element_of(IDs,Set), any(Parameters,PreCond,Body)) :-
    prime_variables(IDs,Parameters), create_members(Set,Parameters,PreCond),
    create_texpr(assign(IDs,Parameters),subst,[],Body).
    
merge_parallels_normalize(assign_single_id(ID,Expr),assign([ID],[Expr])) :- !.		
merge_parallels_normalize(X,X).

sequence_wp(Actions,Pred,WP) :-
    reverse(Actions,ReversedActions),
    sequence_wp2(ReversedActions,Pred,WP).
sequence_wp2([],Pred,Pred).
sequence_wp2([Action|Actions],PredIn,PredOut) :-
    weakest_precondition(Action,PredIn,PredT),
    sequence_wp2(Actions,PredT,PredOut).

select_whens_wps([],_Pred,[]).
select_whens_wps([When|Whens],Pred,[WP|WPs]) :-
    get_texpr_expr(When,select_when(PreCond, Body)),
    weakest_precondition(Body,Pred,InnerWP),
    create_texpr(implication(PreCond,InnerWP),pred,[],WP),
    select_whens_wps(Whens,Pred,WPs).

create_members(Set,IDs,MemberOf) :-
    maplist(create_member(Set),IDs,Preds),
    conjunct_predicates(Preds,MemberOf).
create_member(Set,ID,Res) :-
    create_texpr(member(ID,Set),pred,[],Res).

replace_all_id_by_expr([],[],Pred,Pred).
replace_all_id_by_expr([Id|Ids],[Expr|Exprs],Pred,Result) :-
    replace_id_by_expr(Pred,Id,Expr,TmpRes),
    replace_all_id_by_expr(Ids,Exprs,TmpRes,Result).

function_assignment_to_overwrite([],[],[],[]).
function_assignment_to_overwrite([b(identifier(ID),Type,Infos)|IdOrFuns],[Expr|Exprs],[b(identifier(ID),Type,Infos)|IdOuts],[Expr|ExprOuts]) :-
    function_assignment_to_overwrite(IdOrFuns,Exprs,IdOuts,ExprOuts).
function_assignment_to_overwrite([b(function(Function,Argument),_Type,_Infos)|IdOrFuns],[Expr|Exprs],[Function|IdOuts],[ExprOut|ExprOuts]) :-
    get_texpr_type(Argument,ArgumentType), get_texpr_type(Expr,ExprType),
    get_texpr_type(Function,FunctionType),
    CoupleType = couple(ArgumentType,ExprType),
    create_texpr(couple(Argument,Expr),CoupleType,[],OverwriteCouple),
    create_texpr(set_extension([OverwriteCouple]),set(CoupleType),[],CoupleInSet),
    create_texpr(overwrite(Function,CoupleInSet),FunctionType,[],ExprOut),
    function_assignment_to_overwrite(IdOrFuns,Exprs,IdOuts,ExprOuts).