% (c) 2009 Lehrstuhl fuer Softwaretechnik und Programmiersprachen, 
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html) 

:- module(fuzzfile, [load_fuzzfile/1, reset_fuzzfile/0, get_z_definitions/1]).

:- dynamic current_fuzzfile/2.

:- use_module(library(lists)).

:- use_module(probsrc(tools)).
:- use_module(probsrc(error_manager)).

:- use_module(probsrc(module_information)).
:- module_info(group,proz).
:- module_info(description,'This module includes a parser for files generated by the fuzz type checker for Z specifications.').
:- module_info(revision,'$Rev$').
:- module_info(lastchanged,'$LastChangedDate$').

fuzz_file_empty :- \+ current_fuzzfile(_,_),!.
fuzz_file_empty :- current_fuzzfile(_,[]).

reset_fuzzfile :- retractall(current_fuzzfile(_,_)).

get_z_definitions(Defs) :-
	current_fuzzfile(_,Defs).



load_fuzzfile(File) :-
    reset_fuzzfile,

    (readfile(File,Content) -> true;
	add_error(fuzzfile,'Error while reading file', File),fail),

    (zparser(Defs,Content,[]) -> true;
	add_error(fuzzfile,'Error while parsing fuzz file', File), fail),

    assert(current_fuzzfile(File,Defs)).


%*******************************************************************************
% read file content
readfile(Filename,Content) :-
	open(Filename,read,Stream),
	readbytes(Stream,Content),
	close(Stream).
readbytes(Stream,Res) :-
	get_code(Stream,C),
	(C= -1 -> Res=[] ; Res=[C|Rest],readbytes(Stream,Rest)).


%*******************************************************************************
% Parser
zparser(Items) --> wss, zparser2(Items).
zparser2([I|Rest]) --> zitem(I), !, zparser(Rest).
zparser2([]) --> [].

wss --> [C], {member(C," \t\n\r")}, !, wss.
wss --> "".

zitem(Item) --> "(", !, zitem2(Item).
zitem(String) --> identifier(String).

zitem2([]) --> ")",!.
zitem2(Item) --> itemname(IName), wss, zitem3(IName,Item), wss, ")",!.

zitem3('LIST',List) --> !,zparser(List).
zitem3('NAME',name(Name,Deco)) --> !, identifier(Name), wss, string(Deco).
zitem3('REF',ref(Name,Params)) --> !,
	linenumber(_), wss, zitem(Name), wss, zitem(Params).
zitem3('SREF',sref(Name,Deco,Params,Renamings)) --> !,
	linenumber(_), wss, identifier(Name), wss, string(Deco),
	wss, zitem(Params), wss, zitem(Renamings).
zitem3('NUMBER',number(N)) --> !,
	linenumber(_), wss, string(N).
zitem3(IName,Item) --> linenumber(_), zparser(Items),
	{to_lower(IName,LIName), Item =.. [LIName | Items]}.

linenumber(Number) --> digit(D), linenumber2(Rest), {name(Number,[D|Rest])}.
linenumber2([D|Rest]) --> digit(D),!,linenumber2(Rest).
linenumber2([]) --> [].

itemname(N) --> letter(L), itemname2(Rest), {name(N,[L|Rest])}.
itemname2([L|Rest]) --> alphanum(L), !, itemname2(Rest).
itemname2([]) --> [].

string(Atom) --> [34], stringcontent(String), [34], {name(Atom,String)}.

stringcontent([C|Rest]) --> [C], {C \= 34}, !, stringcontent(Rest).
stringcontent([]) --> [].

identifier(Id) --> [34], idcontent(String), {atom_codes(Id,String)}.

idcontent(String) --> [C], idcontent2(C, String).
idcontent2(34, []) --> !. % closing quotation mark
idcontent2(92, [D|String]) --> % backslash
	[C],{deescape(C,D)},!,idcontent(String).
idcontent2(C, [C|String]) --> idcontent(String).

deescape(95,95). % underscore

alphanum(A) --> [A],{(is_digit(A)->true;is_letter(A))}.

digit(D) --> [D],{is_digit(D)}.
letter(L) --> [L], {is_letter(L)}.

is_letter(L) :-
	member(L,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"),!.
is_digit(L) :-
	member(L,"0123456789"),!.

to_lower(UItem,LItem) :- name(UItem,Up),to_lower2(Up,Low),name(LItem,Low).
to_lower2([],[]).
to_lower2([Up|URest],[Low|LRest]) :-
	(nth0(N,"ABCDEFGHIJKLMNOPQRSTUVWXYZ",Up) ->
	    nth0(N,"abcdefghijklmnopqrstuvwxyz",Low);
	    Up = Low),
	to_lower2(URest,LRest).

